
#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 1953
 * transforms:
 *    ('imul', 'a', '#b(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b'))
 *    ('imul', 'a@8', 128) => ('ishl', 'a', 7)
 *    ('imul', 'a@16', 32768) => ('ishl', 'a', 15)
 *    ('imul', 'a@32', 2147483648) => ('ishl', 'a', 31)
 *    ('imul', 'a@64', 9223372036854775808) => ('ishl', 'a', 63)
 *    ('imul', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b'))))
 *    ('ishl', 'a', '#b') => ('imul', 'a', ('ishl', 1, 'b'))
 *    ('imul@64', 'a', '#b(is_bitcount2)') => ('iadd', ('ishl', 'a', ('ufind_msb', 'b')), ('ishl', 'a', ('find_lsb', 'b')))
 *    ('unpack_64_2x32_split_x', ('imul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b')
 *    ('unpack_64_2x32_split_x', ('umul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b')
 *    ('imul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('imul_high', 'a', 'b'))
 *    ('umul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('umul_high', 'a', 'b'))
 *    ('udiv', 'a', 1) => a
 *    ('idiv', 'a', 1) => a
 *    ('umod', 'a', 1) => 0
 *    ('imod', 'a', 1) => 0
 *    ('imod', 'a', -1) => 0
 *    ('irem', 'a', 1) => 0
 *    ('irem', 'a', -1) => 0
 *    ('udiv', 'a', '#b(is_pos_power_of_two)') => ('ushr', 'a', ('find_lsb', 'b'))
 *    ('idiv', 'a', '#b(is_pos_power_of_two)') => ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', 'b')))
 *    ('idiv', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', ('iabs', 'b')))))
 *    ('umod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1))
 *    ('imod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1))
 *    ('imod', 'a', '#b(is_neg_power_of_two)') => ('bcsel', ('ieq', ('ior', 'a', 'b'), 'b'), 0, ('ior', 'a', 'b'))
 *    ('irem', 'a', '#b(is_pos_power_of_two)') => ('isub', 'a', ('iand', ('bcsel', ('ilt', 'a', 0), ('iadd', 'a', ('isub', 'b', 1)), 'a'), ('ineg', 'b')))
 *    ('irem', 'a', '#b(is_neg_power_of_two)') => ('irem', 'a', ('iabs', 'b'))
 *    ('~fneg', ('fneg', 'a')) => a
 *    ('ineg', ('ineg', 'a')) => a
 *    ('fabs', ('fneg', 'a')) => ('fabs', 'a')
 *    ('fabs', ('u2f', 'a')) => ('u2f', 'a')
 *    ('iabs', ('iabs', 'a')) => ('iabs', 'a')
 *    ('iabs', ('ineg', 'a')) => ('iabs', 'a')
 *    ('f2b', ('fneg', 'a')) => ('f2b', 'a')
 *    ('i2b', ('ineg', 'a')) => ('i2b', 'a')
 *    ('~fadd', 'a', 0.0) => a
 *    ('fadd(is_only_used_as_float)', 'a@16', 0.0) => a
 *    ('fadd(is_only_used_as_float)', 'a@32', 0.0) => a
 *    ('iadd', 'a', 0) => a
 *    ('iadd_sat', 'a', 0) => a
 *    ('isub_sat', 'a', 0) => a
 *    ('uadd_sat', 'a', 0) => a
 *    ('usub_sat', 'a', 0) => a
 *    ('usadd_4x8_vc4', 'a', 0) => a
 *    ('usadd_4x8_vc4', 'a', -1) => -1
 *    ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmulz', 'a', 'b'), ('fmulz', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c'))
 *    ('~ffma', 'a', 'b', ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffma', 'a', 'b', ('fmul(is_used_once)', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmul(is_used_once)', 'a', 'b'), ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffma', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('ffma', 'a', 'c', 'd'))
 *    ('~ffmaz', 'a', 'b', ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffmaz', 'a', 'b', ('fmulz(is_used_once)', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c'))
 *    ('~fadd', ('fmulz(is_used_once)', 'a', 'b'), ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd')
 *    ('~ffmaz', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('ffmaz', 'a', 'c', 'd'))
 *    ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c'))
 *    ('iand', ('ior', 'a', 'b'), ('ior', 'a', 'c')) => ('ior', 'a', ('iand', 'b', 'c'))
 *    ('ior', ('iand', 'a', 'b'), ('iand', 'a', 'c')) => ('iand', 'a', ('ior', 'b', 'c'))
 *    ('~fadd', ('fneg', 'a'), 'a') => 0.0
 *    ('iadd', ('ineg', 'a'), 'a') => 0
 *    ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b
 *    ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b
 *    ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b
 *    ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b
 *    ('fadd', ('fsat', 'a'), ('fsat', ('fneg', 'a'))) => ('fsat', ('fabs', 'a'))
 *    ('~fmul', 'a', 0.0) => 0.0
 *    ('fmul', 'a@16', 0.0) => 0.0
 *    ('fmul', 'a@32', 0.0) => 0.0
 *    ('fmulz', 'a', 0.0) => 0.0
 *    ('fmulz', 'a', 'b(is_finite_not_zero)') => ('fmul', 'a', 'b')
 *    ('fmulz', 'a(is_finite)', 'b(is_finite)') => ('fmul', 'a', 'b')
 *    ('fmulz', 'a', 'a') => ('fmul', 'a', 'a')
 *    ('ffmaz', 'a', 'b(is_finite_not_zero)', 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', 'a(is_finite)', 'b(is_finite)', 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', 'a', 'a', 'b') => ('ffma', 'a', 'a', 'b')
 *    ('imul', 'a', 0) => 0
 *    ('umul_unorm_4x8_vc4', 'a', 0) => 0
 *    ('umul_unorm_4x8_vc4', 'a', -1) => a
 *    ('~fmul', 'a', 1.0) => a
 *    ('~fmulz', 'a', 1.0) => a
 *    ('fmul(is_only_used_as_float)', 'a', 1.0) => a
 *    ('imul', 'a', 1) => a
 *    ('fmul', 'a', -1.0) => ('fneg', 'a')
 *    ('imul', 'a', -1) => ('ineg', 'a')
 *    ('fmul', ('fsign', 'a'), ('fmul', 'a', 'a')) => ('fmul', ('fabs', 'a'), 'a')
 *    ('fmul', ('fmul', ('fsign', 'a'), 'a'), 'a') => ('fmul', ('fabs', 'a'), 'a')
 *    ('~ffma', 0.0, 'a', 'b') => b
 *    ('ffma@16(is_only_used_as_float)', 0.0, 'a', 'b') => b
 *    ('ffma@32(is_only_used_as_float)', 0.0, 'a', 'b') => b
 *    ('ffmaz', 0.0, 'a', 'b') => ('fadd', 0.0, 'b')
 *    ('~ffma', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffma@16', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffma@32', 'a', 'b', 0.0) => ('fmul', 'a', 'b')
 *    ('ffmaz', 'a', 'b', 0.0) => ('fmulz', 'a', 'b')
 *    ('ffma', 1.0, 'a', 'b') => ('fadd', 'a', 'b')
 *    ('ffmaz', 1.0, 'a', 'b') => ('fadd', 'a', 'b')
 *    ('ffma', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b')
 *    ('ffmaz', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b')
 *    ('~ffma', '#a', '#b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffmaz', '#a', '#b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~flrp', 'a', 'b', 0.0) => a
 *    ('~flrp', 'a', 'b', 1.0) => b
 *    ('~flrp', 'a', 'a', 'b') => a
 *    ('~flrp', 0.0, 'a', 'b') => ('fmul', 'a', 'b')
 *    ('~flrp', 'a', ('fadd(is_used_once)', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('sdot_4x8_iadd', 'a', 0, 'b') => b
 *    ('udot_4x8_uadd', 'a', 0, 'b') => b
 *    ('sdot_4x8_iadd_sat', 'a', 0, 'b') => b
 *    ('udot_4x8_uadd_sat', 'a', 0, 'b') => b
 *    ('sdot_2x16_iadd', 'a', 0, 'b') => b
 *    ('udot_2x16_uadd', 'a', 0, 'b') => b
 *    ('sdot_2x16_iadd_sat', 'a', 0, 'b') => b
 *    ('udot_2x16_uadd_sat', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd', 0, 'a', 'b') => b
 *    ('sudot_4x8_iadd_sat', 'a', 0, 'b') => b
 *    ('sudot_4x8_iadd_sat', 0, 'a', 'b') => b
 *    ('iadd', ('sdot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('udot_4x8_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_4x8_uadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sudot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sudot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sdot_2x16_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_2x16_iadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('udot_2x16_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_2x16_uadd', 'a', 'b', ('iadd', 'c', 'd'))
 *    ('iadd', ('sdot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('udot_4x8_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_4x8_uadd', 'a', 'b', 'c')
 *    ('iadd', ('sudot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('sudot_4x8_iadd', 'a', 'b(is_not_const)', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c')
 *    ('iadd', ('sdot_2x16_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_2x16_iadd', 'a', 'b', 'c')
 *    ('iadd', ('udot_2x16_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_2x16_uadd', 'a', 'b', 'c')
 *    ('sdot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('udot_4x8_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_4x8_uadd', 'a', 'b', 0), 'c')
 *    ('sudot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sudot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('sdot_2x16_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_2x16_iadd', 'a', 'b', 0), 'c')
 *    ('udot_2x16_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_2x16_uadd', 'a', 'b', 0), 'c')
 *    ('sdot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('udot_4x8_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_4x8_uadd', 'a', 'b', 0), 'c')
 *    ('sudot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sudot_4x8_iadd', 'a', 'b', 0), 'c')
 *    ('sdot_2x16_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_2x16_iadd', 'a', 'b', 0), 'c')
 *    ('udot_2x16_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_2x16_uadd', 'a', 'b', 0), 'c')
 *    ('fmul@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b')) => ('fmulz', 'a', 'b')
 *    ('fmul@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)')) => ('fmulz', 'a', 'b')
 *    ('ffma@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('ffma@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('sdot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c')
 *    ('udot_4x8_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sudot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sdot_2x16_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c')
 *    ('udot_2x16_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c')
 *    ('udot_4x8_uadd_sat', 'a', 'b', '#c(is_ult_0xfffc07fc)') => ('udot_4x8_uadd', 'a', 'b', 'c')
 *    ('udot_4x8_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('sdot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c')
 *    ('sudot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c')
 *    ('udot_2x16_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c')
 *    ('sdot_2x16_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c')
 *    ('~flrp@16', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@16', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@16', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@16', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@16', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@16', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@16', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@16', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@16', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp@32', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@32', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@32', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@32', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@32', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@32', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@32', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@32', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp@64', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a')
 *    ('~flrp@64', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a')
 *    ('~flrp@64', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp@64', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a')
 *    ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c'))
 *    ('~fadd@64', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~fadd@64', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a')
 *    ('~ffma@64', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b'))
 *    ('~fadd@64', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a')
 *    ('~flrp', ('fmul(is_used_once)', 'a', 'b'), ('fmul(is_used_once)', 'a', 'c'), 'd') => ('fmul', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('~flrp', 'a', 0.0, 'c') => ('fadd', ('fmul', ('fneg', 'a'), 'c'), 'a')
 *    ('ftrunc', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a')))
 *    ('ffloor@16', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('ffloor@32', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('ffloor@64', 'a') => ('fsub', 'a', ('ffract', 'a'))
 *    ('fadd@16', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('fadd@32', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('fadd@64', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a')
 *    ('ffract@16', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('ffract@32', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('ffract@64', 'a') => ('fsub', 'a', ('ffloor', 'a'))
 *    ('fceil', 'a') => ('fneg', ('ffloor', ('fneg', 'a')))
 *    ('ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c')
 *    ('~ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c')
 *    ('~fmul', ('fadd', ('iand', ('ineg', ('b2i', 'a@bool')), ('fmul', 'b', 'c')), '#d'), '#e') => ('bcsel', 'a', ('fmul', ('fadd', ('fmul', 'b', 'c'), 'd'), 'e'), ('fmul', 'd', 'e'))
 *    ('fdph', 'a', 'b') => ('fdot4', ('vec4', 'a.x', 'a.y', 'a.z', 1.0), 'b')
 *    ('fdot4', ('vec4', 'a', 'b', 'c', 1.0), 'd') => ('fdph', ('vec3', 'a', 'b', 'c'), 'd')
 *    ('fdot4', ('vec4', 'a', 0.0, 0.0, 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot4', ('vec4', 'a', 'b', 0.0, 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c')
 *    ('fdot4', ('vec4', 'a', 'b', 'c', 0.0), 'd') => ('fdot3', ('vec3', 'a', 'b', 'c'), 'd')
 *    ('fdot3', ('vec3', 'a', 0.0, 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot3', ('vec3', 'a', 'b', 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c')
 *    ('fdot2', ('vec2', 'a', 0.0), 'b') => ('fmul', 'a', 'b')
 *    ('fdot2', 'a', 1.0) => ('fadd', 'a.x', 'a.y')
 *    ('fdot2', 'a', 'b') => ('fsum2', ('fmul', 'a', 'b'))
 *    ('fdot3', 'a', 'b') => ('fsum3', ('fmul', 'a', 'b'))
 *    ('fdot4', 'a', 'b') => ('fsum4', ('fmul', 'a', 'b'))
 *    ('fsum2', 'a') => ('fadd', 'a.x', 'a.y')
 *    ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a')))
 *    ('ishl', ('iadd', ('imul', 'a', '#b'), '#c'), '#d') => ('iadd', ('imul', 'a', ('ishl', 'b', 'd')), ('ishl', 'c', 'd'))
 *    ('ishl', ('imul', 'a', '#b'), '#c') => ('imul', 'a', ('ishl', 'b', 'c'))
 *    ('ishl@8', ('ishl@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ishl@8', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@8', ('ushr@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ushr@8', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@8', ('ishr@8', 'a', '#b'), '#c') => ('ishr@8', 'a', ('imin', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 7))
 *    ('ishl@16', ('ishl@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ishl@16', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@16', ('ushr@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ushr@16', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@16', ('ishr@16', 'a', '#b'), '#c') => ('ishr@16', 'a', ('imin', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 15))
 *    ('ishl@32', ('ishl@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ishl@32', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@32', ('ushr@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ushr@32', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@32', ('ishr@32', 'a', '#b'), '#c') => ('ishr@32', 'a', ('imin', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 31))
 *    ('ishl@64', ('ishl@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ishl@64', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ushr@64', ('ushr@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ushr@64', 'a', ('iadd', 'b', 'c')), 0)
 *    ('ishr@64', ('ishr@64', 'a', '#b'), '#c') => ('ishr@64', 'a', ('imin', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 63))
 *    ('ishl', ('ushr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b'))
 *    ('ishl', ('ishr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b'))
 *    ('ushr', ('ishl', 'a@8', '#b'), 'b') => ('iand', 'a', ('ushr', 255, 'b'))
 *    ('ishl', ('ushr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b'))
 *    ('ishl', ('ishr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b'))
 *    ('ushr', ('ishl', 'a@16', '#b'), 'b') => ('iand', 'a', ('ushr', 65535, 'b'))
 *    ('ishl', ('ushr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b'))
 *    ('ishl', ('ishr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b'))
 *    ('ushr', ('ishl', 'a@32', '#b'), 'b') => ('iand', 'a', ('ushr', 4294967295, 'b'))
 *    ('ishl', ('ushr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b'))
 *    ('ishl', ('ishr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b'))
 *    ('ushr', ('ishl', 'a@64', '#b'), 'b') => ('iand', 'a', ('ushr', 18446744073709551615, 'b'))
 *    ('iand', ('ishl', 'a@32', '#b(is_first_5_bits_uge_2)'), -4) => ('ishl', 'a', 'b')
 *    ('iand', ('imul', 'a', '#b(is_unsigned_multiple_of_4)'), -4) => ('imul', 'a', 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_2)'), 4294967294) => ('iadd', ('iand', 'a', 4294967294), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_4)'), 4294967292) => ('iadd', ('iand', 'a', 4294967292), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_8)'), 4294967288) => ('iadd', ('iand', 'a', 4294967288), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_16)'), 4294967280) => ('iadd', ('iand', 'a', 4294967280), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_32)'), 4294967264) => ('iadd', ('iand', 'a', 4294967264), 'b')
 *    ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_64)'), 4294967232) => ('iadd', ('iand', 'a', 4294967232), 'b')
 *    ('ishl@32', ('iand', 'a@32', 2147483647), 1) => ('ishl', 'a', 1)
 *    ('iand', ('ishl', 'a@32', 1), 4294967294) => ('ishl', 'a', 1)
 *    ('iand', ('ushr', 'a@32', 1), 2147483647) => ('ushr', 'a', 1)
 *    ('ishl@32', ('iand', 'a@32', 1073741823), 2) => ('ishl', 'a', 2)
 *    ('iand', ('ishl', 'a@32', 2), 4294967292) => ('ishl', 'a', 2)
 *    ('iand', ('ushr', 'a@32', 2), 1073741823) => ('ushr', 'a', 2)
 *    ('ishl@32', ('iand', 'a@32', 65535), 16) => ('ishl', 'a', 16)
 *    ('iand', ('ishl', 'a@32', 16), 4294901760) => ('ishl', 'a', 16)
 *    ('iand', ('ushr', 'a@32', 16), 65535) => ('ushr', 'a', 16)
 *    ('ishl@32', ('iand', 'a@32', 255), 24) => ('ishl', 'a', 24)
 *    ('iand', ('ishl', 'a@32', 24), 4278190080) => ('ishl', 'a', 24)
 *    ('iand', ('ushr', 'a@32', 24), 255) => ('ushr', 'a', 24)
 *    ('ishl', ('iadd', 'a', '#b'), '#c') => ('iadd', ('ishl', 'a', 'c'), ('ishl', 'b', 'c'))
 *    ('imul', ('iadd(is_used_once)', 'a', '#b'), '#c') => ('iadd', ('imul', 'a', 'c'), ('imul', 'b', 'c'))
 *    ('imul', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('imul', ('iadd', 'a', 'c'), 'd'), ('imul', 'b', 'd'))
 *    ('ishl', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('ishl', ('iadd', 'a', 'c'), 'd'), ('ishl', 'b', 'd'))
 *    ('inot', ('flt(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('fge', 'a', 'b')
 *    ('inot', ('fge(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('flt', 'a', 'b')
 *    ('inot', ('feq(is_used_once)', 'a', 'b')) => ('fneu', 'a', 'b')
 *    ('inot', ('fneu(is_used_once)', 'a', 'b')) => ('feq', 'a', 'b')
 *    ('inot', ('ilt(is_used_once)', 'a', 'b')) => ('ige', 'a', 'b')
 *    ('inot', ('ult(is_used_once)', 'a', 'b')) => ('uge', 'a', 'b')
 *    ('inot', ('ige(is_used_once)', 'a', 'b')) => ('ilt', 'a', 'b')
 *    ('inot', ('uge(is_used_once)', 'a', 'b')) => ('ult', 'a', 'b')
 *    ('inot', ('ieq(is_used_once)', 'a', 'b')) => ('ine', 'a', 'b')
 *    ('inot', ('ine(is_used_once)', 'a', 'b')) => ('ieq', 'a', 'b')
 *    ('iand', ('feq', 'a', 'b'), ('fneu', 'a', 'b')) => False
 *    ('iand', ('flt', 'a', 'b'), ('flt', 'b', 'a')) => False
 *    ('iand', ('ieq', 'a', 'b'), ('ine', 'a', 'b')) => False
 *    ('iand', ('ilt', 'a', 'b'), ('ilt', 'b', 'a')) => False
 *    ('iand', ('ult', 'a', 'b'), ('ult', 'b', 'a')) => False
 *    ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a')
 *    ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a')
 *    ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a')
 *    ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a')
 *    ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a')
 *    ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0)
 *    ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a')
 *    ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0)
 *    ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a')
 *    ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0)
 *    ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a')
 *    ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b')
 *    ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b')
 *    ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b')
 *    ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0)
 *    ('flt', 0.0, ('fsat(is_used_once)', 'a')) => ('flt', 0.0, 'a')
 *    ('fge', 0.0, ('b2f', 'a@1')) => ('inot', 'a')
 *    ('fge', ('fneg', ('b2f', 'a@1')), 0.0) => ('inot', 'a')
 *    ('fneu', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b')
 *    ('fneu', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b')
 *    ('fneu', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('ior', 'a', 'b')
 *    ('fneu', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('iand', 'a', 'b')
 *    ('fneu', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('iand', 'a', 'b')
 *    ('fneu', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ixor', 'a', 'b')
 *    ('fneu', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ixor', 'a', 'b')
 *    ('fneu', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ixor', 'a', 'b')
 *    ('feq', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b'))
 *    ('feq', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('iand', 'a', 'b'))
 *    ('feq', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('inot', ('iand', 'a', 'b'))
 *    ('feq', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ieq', 'a', 'b')
 *    ('feq', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ieq', 'a', 'b')
 *    ('feq', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ieq', 'a', 'b')
 *    ('flt', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('ior', 'a', 'b')
 *    ('flt', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('ior', 'a', 'b')
 *    ('fge', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('inot', ('ior', 'a', 'b'))
 *    ('fge', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b'))
 *    ('flt', 'a', ('fneg', 'a')) => ('flt', 'a', 0.0)
 *    ('fge', 'a', ('fneg', 'a')) => ('fge', 'a', 0.0)
 *    ('flt', ('fmin', 'c', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')))), 0.0) => ('ior', ('flt', 'c', 0.0), ('ior', 'a', 'b'))
 *    ('~flt', ('fadd', 'a', 'b'), 'a') => ('flt', 'b', 0.0)
 *    ('~fge', ('fadd', 'a', 'b'), 'a') => ('fge', 'b', 0.0)
 *    ('~feq', ('fadd', 'a', 'b'), 'a') => ('feq', 'b', 0.0)
 *    ('~fneu', ('fadd', 'a', 'b'), 'a') => ('fneu', 'b', 0.0)
 *    ('~flt', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('flt', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('flt', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~fge', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fge', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fge', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~feq', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('feq', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~feq', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('feq', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('~fneu', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fneu', 'a', ('fadd', 'c', ('fneg', 'b')))
 *    ('~fneu', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fneu', ('fneg', ('fadd', 'c', 'b')), 'a')
 *    ('ieq', ('iadd', 'a', 'b'), 'a') => ('ieq', 'b', 0)
 *    ('ine', ('iadd', 'a', 'b'), 'a') => ('ine', 'b', 0)
 *    ('feq', ('b2f', 'a@1'), 0.0) => ('inot', 'a')
 *    ('fneu', ('b2f', 'a@1'), 0.0) => a
 *    ('ieq', ('b2i', 'a@1'), 0) => ('inot', 'a')
 *    ('ine', ('b2i', 'a@1'), 0) => a
 *    ('fneu', ('u2f', 'a'), 0.0) => ('ine', 'a', 0)
 *    ('feq', ('u2f', 'a'), 0.0) => ('ieq', 'a', 0)
 *    ('fge', ('u2f', 'a'), 0.0) => True
 *    ('fge', 0.0, ('u2f', 'a')) => ('uge', 0, 'a')
 *    ('flt', ('u2f', 'a'), 0.0) => False
 *    ('flt', 0.0, ('u2f', 'a')) => ('ult', 0, 'a')
 *    ('fneu', ('i2f', 'a'), 0.0) => ('ine', 'a', 0)
 *    ('feq', ('i2f', 'a'), 0.0) => ('ieq', 'a', 0)
 *    ('fge', ('i2f', 'a'), 0.0) => ('ige', 'a', 0)
 *    ('fge', 0.0, ('i2f', 'a')) => ('ige', 0, 'a')
 *    ('flt', ('i2f', 'a'), 0.0) => ('ilt', 'a', 0)
 *    ('flt', 0.0, ('i2f', 'a')) => ('ilt', 0, 'a')
 *    ('~flt', 0.0, ('fabs', 'a')) => ('fneu', 'a', 0.0)
 *    ('~flt', ('fneg', ('fabs', 'a')), 0.0) => ('fneu', 'a', 0.0)
 *    ('fge', 0.0, ('fabs', 'a')) => ('feq', 'a', 0.0)
 *    ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0)
 *    ('iand', ('fge', 'a', 0.0), ('fge', 1.0, 'a')) => ('feq', 'a', ('fsat', 'a'))
 *    ('fmax', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('ior', 'a', 'b'))
 *    ('fmax', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('iand', 'a', 'b')))
 *    ('fmin', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b'))
 *    ('fmin', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('ior', 'a', 'b')))
 *    ('fmin', ('b2f', 'a@1'), '#b') => ('bcsel', 'a', ('fmin', 'b', 1.0), ('fmin', 'b', 0.0))
 *    ('flt', ('fadd(is_used_once)', 'a', ('fneg', 'b')), 0.0) => ('flt', 'a', 'b')
 *    ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0)
 *    ('~bcsel', ('flt', 'b', 'a'), 'b', 'a') => ('fmin', 'a', 'b')
 *    ('~bcsel', ('flt', 'a', 'b'), 'b', 'a') => ('fmax', 'a', 'b')
 *    ('~bcsel', ('fge', 'a', 'b'), 'b', 'a') => ('fmin', 'a', 'b')
 *    ('~bcsel', ('fge', 'b', 'a'), 'b', 'a') => ('fmax', 'a', 'b')
 *    ('bcsel', ('i2b', 'a'), 'b', 'c') => ('bcsel', ('ine', 'a', 0), 'b', 'c')
 *    ('bcsel', ('inot', 'a'), 'b', 'c') => ('bcsel', 'a', 'c', 'b')
 *    ('bcsel', 'a', ('bcsel', 'a', 'b', 'c'), 'd') => ('bcsel', 'a', 'b', 'd')
 *    ('bcsel', 'a', 'b', ('bcsel', 'a', 'c', 'd')) => ('bcsel', 'a', 'b', 'd')
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', True, 'b') => ('ior', 'a', 'b')
 *    ('bcsel', 'a', 'a', 'b') => ('ior', 'a', 'b')
 *    ('bcsel', 'a', 'b', False) => ('iand', 'a', 'b')
 *    ('bcsel', 'a', 'b', 'a') => ('iand', 'a', 'b')
 *    ('~fmin', 'a', 'a') => a
 *    ('~fmax', 'a', 'a') => a
 *    ('imin', 'a', 'a') => a
 *    ('imax', 'a', 'a') => a
 *    ('umin', 'a', 'a') => a
 *    ('umin', 'a', 0) => 0
 *    ('umin', 'a', -1) => a
 *    ('umax', 'a', 'a') => a
 *    ('umax', 'a', 0) => a
 *    ('umax', 'a', -1) => -1
 *    ('fmax', ('fmax', 'a', 'b'), 'b') => ('fmax', 'a', 'b')
 *    ('umax', ('umax', 'a', 'b'), 'b') => ('umax', 'a', 'b')
 *    ('imax', ('imax', 'a', 'b'), 'b') => ('imax', 'a', 'b')
 *    ('fmin', ('fmin', 'a', 'b'), 'b') => ('fmin', 'a', 'b')
 *    ('umin', ('umin', 'a', 'b'), 'b') => ('umin', 'a', 'b')
 *    ('imin', ('imin', 'a', 'b'), 'b') => ('imin', 'a', 'b')
 *    ('fmax', ('fmax', ('fmax', 'a', 'b'), 'c'), 'a') => ('fmax', ('fmax', 'a', 'b'), 'c')
 *    ('umax', ('umax', ('umax', 'a', 'b'), 'c'), 'a') => ('umax', ('umax', 'a', 'b'), 'c')
 *    ('imax', ('imax', ('imax', 'a', 'b'), 'c'), 'a') => ('imax', ('imax', 'a', 'b'), 'c')
 *    ('fmin', ('fmin', ('fmin', 'a', 'b'), 'c'), 'a') => ('fmin', ('fmin', 'a', 'b'), 'c')
 *    ('umin', ('umin', ('umin', 'a', 'b'), 'c'), 'a') => ('umin', ('umin', 'a', 'b'), 'c')
 *    ('imin', ('imin', ('imin', 'a', 'b'), 'c'), 'a') => ('imin', ('imin', 'a', 'b'), 'c')
 *    ('ieq', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ine', 'a', 'b')
 *    ('ieq', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ieq', 'a', 'b')
 *    ('ine', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ine', 'a', 'b')
 *    ('feq', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ieq', 'a', 'b')
 *    ('fneu', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ine', 'a', 'b')
 *    ('iand@8', 'a', ('inot', ('ishr', 'a', 7))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@8', 128), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@8', 128), 128) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@8', 128), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@8', 128), 128) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@8', 7), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@8', 7), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@8', 7), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@8', 7), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@8', 7), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@8', 7), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@8', 7), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@8', 7), -1) => ('ige', 'a', 0)
 *    ('iand@16', 'a', ('inot', ('ishr', 'a', 15))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@16', 32768), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@16', 32768), 32768) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@16', 32768), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@16', 32768), 32768) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@16', 15), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@16', 15), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@16', 15), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@16', 15), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@16', 15), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@16', 15), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@16', 15), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@16', 15), -1) => ('ige', 'a', 0)
 *    ('iand@32', 'a', ('inot', ('ishr', 'a', 31))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@32', 2147483648), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@32', 2147483648), 2147483648) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@32', 2147483648), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@32', 2147483648), 2147483648) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@32', 31), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@32', 31), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@32', 31), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@32', 31), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@32', 31), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@32', 31), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@32', 31), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@32', 31), -1) => ('ige', 'a', 0)
 *    ('iand@64', 'a', ('inot', ('ishr', 'a', 63))) => ('imax', 'a', 0)
 *    ('ieq', ('iand', 'a@64', 9223372036854775808), 0) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ige', 'a', 0)
 *    ('ine', ('iand', 'a@64', 9223372036854775808), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@64', 63), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ushr', 'a@64', 63), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ushr', 'a@64', 63), 1) => ('ilt', 'a', 0)
 *    ('ine', ('ushr', 'a@64', 63), 1) => ('ige', 'a', 0)
 *    ('ine', ('ishr', 'a@64', 63), 0) => ('ilt', 'a', 0)
 *    ('ieq', ('ishr', 'a@64', 63), 0) => ('ige', 'a', 0)
 *    ('ieq', ('ishr', 'a@64', 63), -1) => ('ilt', 'a', 0)
 *    ('ine', ('ishr', 'a@64', 63), -1) => ('ige', 'a', 0)
 *    ('fmin', 'a', ('fneg', 'a')) => ('fneg', ('fabs', 'a'))
 *    ('imin', 'a', ('ineg', 'a')) => ('ineg', ('iabs', 'a'))
 *    ('fmin', 'a', ('fneg', ('fabs', 'a'))) => ('fneg', ('fabs', 'a'))
 *    ('imin', 'a', ('ineg', ('iabs', 'a'))) => ('ineg', ('iabs', 'a'))
 *    ('~fmin', 'a', ('fabs', 'a')) => a
 *    ('imin', 'a', ('iabs', 'a')) => a
 *    ('~fmax', 'a', ('fneg', ('fabs', 'a'))) => a
 *    ('imax', 'a', ('ineg', ('iabs', 'a'))) => a
 *    ('fmax', 'a', ('fabs', 'a')) => ('fabs', 'a')
 *    ('imax', 'a', ('iabs', 'a')) => ('iabs', 'a')
 *    ('fmax', 'a', ('fneg', 'a')) => ('fabs', 'a')
 *    ('imax', 'a', ('ineg', 'a')) => ('iabs', 'a')
 *    ('~fmax', ('fabs', 'a'), 0.0) => ('fabs', 'a')
 *    ('fmin', ('fmax', 'a', 0.0), 1.0) => ('fsat', 'a')
 *    ('~fmax', ('fmin', 'a', 1.0), 0.0) => ('fsat', 'a')
 *    ('~fmin', ('fmax', 'a', -1.0), 0.0) => ('fneg', ('fsat', ('fneg', 'a')))
 *    ('~fmax', ('fmin', 'a', 0.0), -1.0) => ('fneg', ('fsat', ('fneg', 'a')))
 *    ('fsat', ('fsign', 'a')) => ('b2f', ('!flt', 0.0, 'a'))
 *    ('fsat', ('b2f', 'a')) => ('b2f', 'a')
 *    ('fsat', 'a') => ('fmin', ('fmax', 'a', 0.0), 1.0)
 *    ('fsat', ('fsat', 'a')) => ('fsat', 'a')
 *    ('fsat', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fsat', ('fadd', ('fneg', 'a'), ('fneg', 'b')))
 *    ('fsat', ('fneg(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fneg', 'a'), 'b'))
 *    ('fsat', ('fneg(is_used_once)', ('fmulz(is_used_once)', 'a', 'b'))) => ('fsat', ('fmulz', ('fneg', 'a'), 'b'))
 *    ('fsat', ('fabs(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fabs', 'a'), ('fabs', 'b')))
 *    ('fmin', ('fmax', ('fmin', ('fmax', 'a', 'b'), 'c'), 'b'), 'c') => ('fmin', ('fmax', 'a', 'b'), 'c')
 *    ('imin', ('imax', ('imin', ('imax', 'a', 'b'), 'c'), 'b'), 'c') => ('imin', ('imax', 'a', 'b'), 'c')
 *    ('umin', ('umax', ('umin', ('umax', 'a', 'b'), 'c'), 'b'), 'c') => ('umin', ('umax', 'a', 'b'), 'c')
 *    ('fmax', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmax', 'a', 'b'))
 *    ('~fmin', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmin', 'a', 'b'))
 *    ('fmax', ('fneg', ('fmin', 'b', 'a')), 'b') => ('fmax', ('fabs', 'b'), ('fneg', 'a'))
 *    ('fmin', ('fneg', ('fmax', 'b', 'a')), 'b') => ('fmin', ('fneg', ('fabs', 'b')), ('fneg', 'a'))
 *    ('fmax', ('fadd(is_used_once)', ('fneg', 'a(is_not_negative)'), '#b(is_zero_to_one)'), 0.0) => ('fsat', ('fadd', ('fneg', 'a'), 'b'))
 *    ('extract_u8', ('imin', ('imax', 'a', 0), 255), 0) => ('imin', ('imax', 'a', 0), 255)
 *    ('ior', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('!fmax', 'b', 'c'))
 *    ('ior', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c')
 *    ('ior', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('!fmin', 'b', 'c'))
 *    ('ior', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c')
 *    ('ior', ('flt', 'a', '#b'), ('flt', 'a', '#c')) => ('flt', 'a', ('!fmax', 'b', 'c'))
 *    ('ior', ('flt', '#a', 'c'), ('flt', '#b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c')
 *    ('ior', ('fge', 'a', '#b'), ('fge', 'a', '#c')) => ('fge', 'a', ('!fmin', 'b', 'c'))
 *    ('ior', ('fge', '#a', 'c'), ('fge', '#b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c')
 *    ('~iand', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('fmin', 'b', 'c'))
 *    ('~iand', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c')
 *    ('~iand', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('fmax', 'b', 'c'))
 *    ('~iand', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c')
 *    ('iand', ('flt', 'a', '#b(is_a_number)'), ('flt', 'a', '#c(is_a_number)')) => ('flt', 'a', ('fmin', 'b', 'c'))
 *    ('iand', ('flt', '#a(is_a_number)', 'c'), ('flt', '#b(is_a_number)', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c')
 *    ('iand', ('fge', 'a', '#b(is_a_number)'), ('fge', 'a', '#c(is_a_number)')) => ('fge', 'a', ('fmax', 'b', 'c'))
 *    ('iand', ('fge', '#a(is_a_number)', 'c'), ('fge', '#b(is_a_number)', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c')
 *    ('ior', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imax', 'b', 'c'))
 *    ('ior', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imin', 'a', 'b'), 'c')
 *    ('ior', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imin', 'b', 'c'))
 *    ('ior', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imax', 'a', 'b'), 'c')
 *    ('ior', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umax', 'b', 'c'))
 *    ('ior', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umin', 'a', 'b'), 'c')
 *    ('ior', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umin', 'b', 'c'))
 *    ('ior', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umax', 'a', 'b'), 'c')
 *    ('iand', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imin', 'b', 'c'))
 *    ('iand', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imax', 'a', 'b'), 'c')
 *    ('iand', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imax', 'b', 'c'))
 *    ('iand', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imin', 'a', 'b'), 'c')
 *    ('iand', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umin', 'b', 'c'))
 *    ('iand', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umax', 'a', 'b'), 'c')
 *    ('iand', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umax', 'b', 'c'))
 *    ('iand', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umin', 'a', 'b'), 'c')
 *    ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'c'), 'd'), ('flt', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt', 'a', 'c'), 'd'), ('flt(is_used_once)', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'b'), 'd'), ('flt', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd')
 *    ('ior', ('ior(is_used_once)', ('flt', 'a', 'b'), 'd'), ('flt(is_used_once)', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd')
 *    ('ior', ('flt', 'a(is_a_number)', 'b(is_a_number)'), ('flt', 'b', 'a')) => ('fneu', 'a', 'b')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('bcsel@16', ('feq', 'a', 0.0), 1.0, ('i2f16', ('iadd', ('b2i16', ('flt', 0.0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0.0)))))) => ('i2f16', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@16'), ('b2f', 'c@16')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f16', ('f2i', 'a@16')) => ('ftrunc', 'a')
 *    ('~f2i16', ('i2f', 'a@16')) => a
 *    ('~f2i16', ('u2f', 'a@16')) => a
 *    ('~f2u16', ('i2f', 'a@16')) => a
 *    ('~f2u16', ('u2f', 'a@16')) => a
 *    ('fadd', ('b2f16', ('flt', 0.0, 'a@16')), ('fneg', ('b2f16', ('flt', 'a@16', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i16', ('flt', 0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0)))) => ('f2i16', ('fsign', 'a'))
 *    ('f2f16', ('f2f32', 'a@16')) => a
 *    ('i2i16', ('i2i32', 'a@16')) => a
 *    ('u2u16', ('u2u32', 'a@16')) => a
 *    ('f2f16', ('b2f32', 'a@1')) => ('b2f16', 'a')
 *    ('i2i16', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('u2u16', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('f2u32', ('f2f32', 'a@16')) => ('f2u32', 'a')
 *    ('f2i32', ('f2f32', 'a@16')) => ('f2i32', 'a')
 *    ('f2f16', ('u2f32', 'a')) => ('u2f16', 'a')
 *    ('f2f16', ('i2f32', 'a')) => ('i2f16', 'a')
 *    ('u2f32', ('u2u32', 'a@16')) => ('u2f32', 'a')
 *    ('i2f32', ('i2i32', 'a@16')) => ('i2f32', 'a')
 *    ('f2f16', ('f2f64', 'a@16')) => a
 *    ('i2i16', ('i2i64', 'a@16')) => a
 *    ('u2u16', ('u2u64', 'a@16')) => a
 *    ('f2f16', ('b2f64', 'a@1')) => ('b2f16', 'a')
 *    ('i2i16', ('b2i64', 'a@1')) => ('b2i16', 'a')
 *    ('u2u16', ('b2i64', 'a@1')) => ('b2i16', 'a')
 *    ('f2u64', ('f2f64', 'a@16')) => ('f2u64', 'a')
 *    ('f2i64', ('f2f64', 'a@16')) => ('f2i64', 'a')
 *    ('f2f16', ('u2f64', 'a')) => ('u2f16', 'a')
 *    ('f2f16', ('i2f64', 'a')) => ('i2f16', 'a')
 *    ('u2f64', ('u2u64', 'a@16')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@16')) => ('i2f64', 'a')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('bcsel@32', ('feq', 'a', 0.0), 1.0, ('i2f32', ('iadd', ('b2i32', ('flt', 0.0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0.0)))))) => ('i2f32', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@32'), ('b2f', 'c@32')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f32', ('f2i', 'a@32')) => ('ftrunc', 'a')
 *    ('~f2i32', ('i2f', 'a@32')) => a
 *    ('~f2i32', ('u2f', 'a@32')) => a
 *    ('~f2u32', ('i2f', 'a@32')) => a
 *    ('~f2u32', ('u2f', 'a@32')) => a
 *    ('fadd', ('b2f32', ('flt', 0.0, 'a@32')), ('fneg', ('b2f32', ('flt', 'a@32', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i32', ('flt', 0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0)))) => ('f2i32', ('fsign', 'a'))
 *    ('f2f32', ('f2f64', 'a@32')) => a
 *    ('i2i32', ('i2i64', 'a@32')) => a
 *    ('u2u32', ('u2u64', 'a@32')) => a
 *    ('f2f32', ('b2f64', 'a@1')) => ('b2f32', 'a')
 *    ('i2i32', ('b2i64', 'a@1')) => ('b2i32', 'a')
 *    ('u2u32', ('b2i64', 'a@1')) => ('b2i32', 'a')
 *    ('f2u64', ('f2f64', 'a@32')) => ('f2u64', 'a')
 *    ('f2i64', ('f2f64', 'a@32')) => ('f2i64', 'a')
 *    ('f2f32', ('u2f64', 'a')) => ('u2f32', 'a')
 *    ('f2f32', ('i2f64', 'a')) => ('i2f32', 'a')
 *    ('u2f64', ('u2u64', 'a@32')) => ('u2f64', 'a')
 *    ('i2f64', ('i2i64', 'a@32')) => ('i2f64', 'a')
 *    ('ior', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('ior', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('ior', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('~iand', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b')))
 *    ('bcsel@64', ('feq', 'a', 0.0), 1.0, ('i2f64', ('iadd', ('b2i64', ('flt', 0.0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0.0)))))) => ('i2f64', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0)))))
 *    ('bcsel', 'a', ('b2f(is_used_once)', 'b@64'), ('b2f', 'c@64')) => ('b2f', ('bcsel', 'a', 'b', 'c'))
 *    ('~i2f64', ('f2i', 'a@64')) => ('ftrunc', 'a')
 *    ('~f2i64', ('i2f', 'a@64')) => a
 *    ('~f2i64', ('u2f', 'a@64')) => a
 *    ('~f2u64', ('i2f', 'a@64')) => a
 *    ('~f2u64', ('u2f', 'a@64')) => a
 *    ('fadd', ('b2f64', ('flt', 0.0, 'a@64')), ('fneg', ('b2f64', ('flt', 'a@64', 0.0)))) => ('fsign', 'a')
 *    ('iadd', ('b2i64', ('flt', 0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0)))) => ('f2i64', ('fsign', 'a'))
 *    ('f2fmp', ('u2f32', 'a@32')) => ('u2fmp', 'a')
 *    ('f2fmp', ('i2f32', 'a@32')) => ('i2fmp', 'a')
 *    ('f2u16', ('f2fmp', 'a@32')) => ('f2u16', 'a')
 *    ('f2i16', ('f2fmp', 'a@32')) => ('f2i16', 'a')
 *    ('i2imp', ('f2u32', 'a@32')) => ('f2ump', 'a')
 *    ('i2imp', ('f2i32', 'a@32')) => ('f2imp', 'a')
 *    ('u2f16', ('i2imp', 'a@32')) => ('u2f16', 'a')
 *    ('i2f16', ('i2imp', 'a@32')) => ('i2f16', 'a')
 *    ('iand', ('u2u16', ('u2u8', 'a@16')), '#b') => ('iand', 'a', ('iand', 'b', 255))
 *    ('u2u16', ('u2u8(is_used_once)', ('iand', 'a@16', '#b'))) => ('iand', 'a', ('iand', 'b', 255))
 *    ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('iand', 'a', 'b'))
 *    ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('iand', 'a', 'b'))
 *    ('iand', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('iand', 'a', 'c'), 'b')
 *    ('iand', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('iand', 'a', 'c'), 'b')
 *    ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ior', 'a', 'b'))
 *    ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ior', 'a', 'b'))
 *    ('ior', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ior', 'a', 'c'), 'b')
 *    ('ior', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ior', 'a', 'c'), 'b')
 *    ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ixor', 'a', 'b'))
 *    ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ixor', 'a', 'b'))
 *    ('ixor', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ixor', 'a', 'c'), 'b')
 *    ('ixor', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ixor', 'a', 'c'), 'b')
 *    ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i8', 'a@8')) => a
 *    ('ishl', 'a@8', ('iand', 7, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@8', ('iand', 7, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@8', ('iand', 7, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@8', ('ishl(is_used_once)', ('iand', 'b', 1), 2)) => ('ushr', 'a', ('ishl', 'b', 2))
 *    ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i16', 'a@16')) => a
 *    ('ishl', 'a@16', ('iand', 15, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@16', ('iand', 15, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@16', ('iand', 15, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@16', ('ishl(is_used_once)', ('iand', 'b', 1), 3)) => ('ushr', 'a', ('ishl', 'b', 3))
 *    ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i32', 'a@32')) => a
 *    ('ishl', 'a@32', ('iand', 31, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@32', ('iand', 31, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@32', ('iand', 31, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@32', ('ishl(is_used_once)', ('iand', 'b', 1), 4)) => ('ushr', 'a', ('ishl', 'b', 4))
 *    ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('ior', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('ior', 'a', 'b'), 0)
 *    ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umax', 'a', 'b'), 0)
 *    ('ior', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umax', 'a', 'b'), 0)
 *    ('ineg', ('b2i64', 'a@64')) => a
 *    ('ishl', 'a@64', ('iand', 63, 'b')) => ('ishl', 'a', 'b')
 *    ('ishr', 'a@64', ('iand', 63, 'b')) => ('ishr', 'a', 'b')
 *    ('ushr', 'a@64', ('iand', 63, 'b')) => ('ushr', 'a', 'b')
 *    ('ushr', 'a@64', ('ishl(is_used_once)', ('iand', 'b', 1), 5)) => ('ushr', 'a', ('ishl', 'b', 5))
 *    ('ior', ('ieq', 'a', 0), ('ieq', 'a', 1)) => ('uge', 1, 'a')
 *    ('ior', ('uge', 1, 'a'), ('ieq', 'a', 2)) => ('uge', 2, 'a')
 *    ('ior', ('uge', 2, 'a'), ('ieq', 'a', 3)) => ('uge', 3, 'a')
 *    ('ior', 'a', ('ieq', 'a', False)) => True
 *    ('ior', 'a', ('inot', 'a')) => -1
 *    ('ine', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ine', 'a', 'b')
 *    ('b2i', ('ine', 'a@1', 'b@1')) => ('b2i', ('ixor', 'a', 'b'))
 *    ('iand', ('inot', ('iand', ('ior', ('ieq', 'a', 0), 'b'), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', 'b', 'c')), ('ilt', 'a', 0))
 *    ('iand', ('inot', ('iand', ('ieq', ('umin', 'a', 'b'), 0), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', ('ieq', 'b', 0), 'c')), ('ilt', 'a', 0))
 *    ('flt', 'a', ('fmax', 'b', 'a')) => ('flt', 'a', 'b')
 *    ('flt', ('fmin', 'a', 'b'), 'a') => ('flt', 'b', 'a')
 *    ('~fge', 'a', ('fmin', 'b', 'a')) => True
 *    ('~fge', ('fmax', 'a', 'b'), 'a') => True
 *    ('flt', 'a', ('fmin', 'b', 'a')) => False
 *    ('flt', ('fmax', 'a', 'b'), 'a') => False
 *    ('~fge', 'a', ('fmax', 'b', 'a')) => ('fge', 'a', 'b')
 *    ('~fge', ('fmin', 'a', 'b'), 'a') => ('fge', 'b', 'a')
 *    ('ilt', 'a', ('imax', 'b', 'a')) => ('ilt', 'a', 'b')
 *    ('ilt', ('imin', 'a', 'b'), 'a') => ('ilt', 'b', 'a')
 *    ('ige', 'a', ('imin', 'b', 'a')) => True
 *    ('ige', ('imax', 'a', 'b'), 'a') => True
 *    ('ult', 'a', ('umax', 'b', 'a')) => ('ult', 'a', 'b')
 *    ('ult', ('umin', 'a', 'b'), 'a') => ('ult', 'b', 'a')
 *    ('uge', 'a', ('umin', 'b', 'a')) => True
 *    ('uge', ('umax', 'a', 'b'), 'a') => True
 *    ('ilt', 'a', ('imin', 'b', 'a')) => False
 *    ('ilt', ('imax', 'a', 'b'), 'a') => False
 *    ('ige', 'a', ('imax', 'b', 'a')) => ('ige', 'a', 'b')
 *    ('ige', ('imin', 'a', 'b'), 'a') => ('ige', 'b', 'a')
 *    ('ult', 'a', ('umin', 'b', 'a')) => False
 *    ('ult', ('umax', 'a', 'b'), 'a') => False
 *    ('uge', 'a', ('umax', 'b', 'a')) => ('uge', 'a', 'b')
 *    ('uge', ('umin', 'a', 'b'), 'a') => ('uge', 'b', 'a')
 *    ('ult', 'a', ('iand', 'b', 'a')) => False
 *    ('ult', ('ior', 'a', 'b'), 'a') => False
 *    ('uge', 'a', ('iand', 'b', 'a')) => True
 *    ('uge', ('ior', 'a', 'b'), 'a') => True
 *    ('ilt', '#a', ('imax', '#b', 'c')) => ('ior', ('ilt', 'a', 'b'), ('ilt', 'a', 'c'))
 *    ('ilt', ('imin', '#a', 'b'), '#c') => ('ior', ('ilt', 'a', 'c'), ('ilt', 'b', 'c'))
 *    ('ige', '#a', ('imin', '#b', 'c')) => ('ior', ('ige', 'a', 'b'), ('ige', 'a', 'c'))
 *    ('ige', ('imax', '#a', 'b'), '#c') => ('ior', ('ige', 'a', 'c'), ('ige', 'b', 'c'))
 *    ('ult', '#a', ('umax', '#b', 'c')) => ('ior', ('ult', 'a', 'b'), ('ult', 'a', 'c'))
 *    ('ult', ('umin', '#a', 'b'), '#c') => ('ior', ('ult', 'a', 'c'), ('ult', 'b', 'c'))
 *    ('uge', '#a', ('umin', '#b', 'c')) => ('ior', ('uge', 'a', 'b'), ('uge', 'a', 'c'))
 *    ('uge', ('umax', '#a', 'b'), '#c') => ('ior', ('uge', 'a', 'c'), ('uge', 'b', 'c'))
 *    ('ilt', '#a', ('imin', '#b', 'c')) => ('iand', ('ilt', 'a', 'b'), ('ilt', 'a', 'c'))
 *    ('ilt', ('imax', '#a', 'b'), '#c') => ('iand', ('ilt', 'a', 'c'), ('ilt', 'b', 'c'))
 *    ('ige', '#a', ('imax', '#b', 'c')) => ('iand', ('ige', 'a', 'b'), ('ige', 'a', 'c'))
 *    ('ige', ('imin', '#a', 'b'), '#c') => ('iand', ('ige', 'a', 'c'), ('ige', 'b', 'c'))
 *    ('ult', '#a', ('umin', '#b', 'c')) => ('iand', ('ult', 'a', 'b'), ('ult', 'a', 'c'))
 *    ('ult', ('umax', '#a', 'b'), '#c') => ('iand', ('ult', 'a', 'c'), ('ult', 'b', 'c'))
 *    ('uge', '#a', ('umax', '#b', 'c')) => ('iand', ('uge', 'a', 'b'), ('uge', 'a', 'c'))
 *    ('uge', ('umin', '#a', 'b'), '#c') => ('iand', ('uge', 'a', 'c'), ('uge', 'b', 'c'))
 *    ('bcsel', ('ilt', 'a', 0), ('ineg', ('ishr', 'a', 'b')), ('ishr', 'a', 'b')) => ('iabs', ('ishr', 'a', 'b'))
 *    ('iabs', ('ishr', ('iabs', 'a'), 'b')) => ('ishr', ('iabs', 'a'), 'b')
 *    ('fabs', ('slt', 'a', 'b')) => ('slt', 'a', 'b')
 *    ('fabs', ('sge', 'a', 'b')) => ('sge', 'a', 'b')
 *    ('fabs', ('seq', 'a', 'b')) => ('seq', 'a', 'b')
 *    ('fabs', ('sne', 'a', 'b')) => ('sne', 'a', 'b')
 *    ('slt', 'a', 'b') => ('b2f', ('flt', 'a', 'b'))
 *    ('sge', 'a', 'b') => ('b2f', ('fge', 'a', 'b'))
 *    ('seq', 'a', 'b') => ('b2f', ('feq', 'a', 'b'))
 *    ('sne', 'a', 'b') => ('b2f', ('fneu', 'a', 'b'))
 *    ('seq', ('seq', 'a', 'b'), 1.0) => ('seq', 'a', 'b')
 *    ('seq', ('sne', 'a', 'b'), 1.0) => ('sne', 'a', 'b')
 *    ('seq', ('slt', 'a', 'b'), 1.0) => ('slt', 'a', 'b')
 *    ('seq', ('sge', 'a', 'b'), 1.0) => ('sge', 'a', 'b')
 *    ('sne', ('seq', 'a', 'b'), 0.0) => ('seq', 'a', 'b')
 *    ('sne', ('sne', 'a', 'b'), 0.0) => ('sne', 'a', 'b')
 *    ('sne', ('slt', 'a', 'b'), 0.0) => ('slt', 'a', 'b')
 *    ('sne', ('sge', 'a', 'b'), 0.0) => ('sge', 'a', 'b')
 *    ('seq', ('seq', 'a', 'b'), 0.0) => ('sne', 'a', 'b')
 *    ('seq', ('sne', 'a', 'b'), 0.0) => ('seq', 'a', 'b')
 *    ('seq', ('slt', 'a', 'b'), 0.0) => ('sge', 'a', 'b')
 *    ('seq', ('sge', 'a', 'b'), 0.0) => ('slt', 'a', 'b')
 *    ('sne', ('seq', 'a', 'b'), 1.0) => ('sne', 'a', 'b')
 *    ('sne', ('sne', 'a', 'b'), 1.0) => ('seq', 'a', 'b')
 *    ('sne', ('slt', 'a', 'b'), 1.0) => ('sge', 'a', 'b')
 *    ('sne', ('sge', 'a', 'b'), 1.0) => ('slt', 'a', 'b')
 *    ('fall_equal2', 'a', 'b') => ('fmin', ('seq', 'a.x', 'b.x'), ('seq', 'a.y', 'b.y'))
 *    ('fall_equal3', 'a', 'b') => ('seq', ('fany_nequal3', 'a', 'b'), 0.0)
 *    ('fall_equal4', 'a', 'b') => ('seq', ('fany_nequal4', 'a', 'b'), 0.0)
 *    ('fany_nequal2', 'a', 'b') => ('fmax', ('sne', 'a.x', 'b.x'), ('sne', 'a.y', 'b.y'))
 *    ('fany_nequal3', 'a', 'b') => ('fsat', ('fdot3', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('fany_nequal4', 'a', 'b') => ('fsat', ('fdot4', ('sne', 'a', 'b'), ('sne', 'a', 'b')))
 *    ('ball_iequal2', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y'))
 *    ('ball_iequal3', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('ieq', 'a.z', 'b.z'))
 *    ('ball_iequal4', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('iand', ('ieq', 'a.z', 'b.z'), ('ieq', 'a.w', 'b.w')))
 *    ('bany_inequal2', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y'))
 *    ('bany_inequal3', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ine', 'a.z', 'b.z'))
 *    ('bany_inequal4', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ior', ('ine', 'a.z', 'b.z'), ('ine', 'a.w', 'b.w')))
 *    ('ball_fequal2', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y'))
 *    ('ball_fequal3', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('feq', 'a.z', 'b.z'))
 *    ('ball_fequal4', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('iand', ('feq', 'a.z', 'b.z'), ('feq', 'a.w', 'b.w')))
 *    ('bany_fnequal2', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y'))
 *    ('bany_fnequal3', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('fneu', 'a.z', 'b.z'))
 *    ('bany_fnequal4', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('ior', ('fneu', 'a.z', 'b.z'), ('fneu', 'a.w', 'b.w')))
 *    ('feq', ('seq', 'a', 'b'), 1.0) => ('feq', 'a', 'b')
 *    ('feq', ('sne', 'a', 'b'), 1.0) => ('fneu', 'a', 'b')
 *    ('feq', ('slt', 'a', 'b'), 1.0) => ('flt', 'a', 'b')
 *    ('feq', ('sge', 'a', 'b'), 1.0) => ('fge', 'a', 'b')
 *    ('fneu', ('seq', 'a', 'b'), 0.0) => ('feq', 'a', 'b')
 *    ('fneu', ('sne', 'a', 'b'), 0.0) => ('fneu', 'a', 'b')
 *    ('fneu', ('slt', 'a', 'b'), 0.0) => ('flt', 'a', 'b')
 *    ('fneu', ('sge', 'a', 'b'), 0.0) => ('fge', 'a', 'b')
 *    ('feq', ('seq', 'a', 'b'), 0.0) => ('fneu', 'a', 'b')
 *    ('feq', ('sne', 'a', 'b'), 0.0) => ('feq', 'a', 'b')
 *    ('feq', ('slt', 'a', 'b'), 0.0) => ('fge', 'a', 'b')
 *    ('feq', ('sge', 'a', 'b'), 0.0) => ('flt', 'a', 'b')
 *    ('fneu', ('seq', 'a', 'b'), 1.0) => ('fneu', 'a', 'b')
 *    ('fneu', ('sne', 'a', 'b'), 1.0) => ('feq', 'a', 'b')
 *    ('fneu', ('slt', 'a', 'b'), 1.0) => ('fge', 'a', 'b')
 *    ('fneu', ('sge', 'a', 'b'), 1.0) => ('flt', 'a', 'b')
 *    ('fneu', ('fneg', 'a'), 'a') => ('fneu', 'a', 0.0)
 *    ('feq', ('fneg', 'a'), 'a') => ('feq', 'a', 0.0)
 *    ('imul', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('iand', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b'))
 *    ('ior', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b'))
 *    ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b'))
 *    ('fsat', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('b2f', ('ior', 'a', 'b'))
 *    ('iand', 'a@bool16', 1.0) => ('b2f', 'a')
 *    ('iand', 'a@bool32', 1.0) => ('b2f', 'a')
 *    ('flt', ('fneg', ('b2f', 'a@1')), 0) => a
 *    ('ilt', 'a', 'a') => False
 *    ('ige', 'a', 'a') => True
 *    ('ieq', 'a', 'a') => True
 *    ('ine', 'a', 'a') => False
 *    ('ult', 'a', 'a') => False
 *    ('uge', 'a', 'a') => True
 *    ('flt', 'a', 'a') => False
 *    ('fge', 'a(is_a_number)', 'a') => True
 *    ('feq', 'a(is_a_number)', 'a') => True
 *    ('fneu', 'a(is_a_number)', 'a') => False
 *    ('iand', 'a', 'a') => a
 *    ('iand', 'a', -1) => a
 *    ('iand', 'a', 0) => 0
 *    ('ior', 'a', 'a') => a
 *    ('ior', 'a', 0) => a
 *    ('ior', 'a', True) => True
 *    ('ixor', 'a', 'a') => 0
 *    ('ixor', 'a', 0) => a
 *    ('ixor', 'a', ('ixor', 'a', 'b')) => b
 *    ('ixor', 'a', -1) => ('inot', 'a')
 *    ('inot', ('inot', 'a')) => a
 *    ('ior', ('iand', 'a', 'b'), 'b') => b
 *    ('ior', ('ior', 'a', 'b'), 'b') => ('ior', 'a', 'b')
 *    ('iand', ('ior', 'a', 'b'), 'b') => b
 *    ('iand', ('iand', 'a', 'b'), 'b') => ('iand', 'a', 'b')
 *    ('iand', ('inot', 'a'), ('inot', 'b')) => ('inot', ('ior', 'a', 'b'))
 *    ('ior', ('inot', 'a'), ('inot', 'b')) => ('inot', ('iand', 'a', 'b'))
 *    ('ishl', 0, 'a') => 0
 *    ('ishl', 'a', 0) => a
 *    ('ishr', 0, 'a') => 0
 *    ('ishr', -1, 'a') => -1
 *    ('ishr', 'a', 0) => a
 *    ('ushr', 0, 'a') => 0
 *    ('ushr', 'a', 0) => a
 *    ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('isub', 16, 'b'))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('urol', 'a', 'b')
 *    ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('isub', 32, 'b'))) => ('urol', 'a', 'b')
 *    ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('isub', 16, 'b'))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('uror', 'a', 'b')
 *    ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('isub', 32, 'b'))) => ('uror', 'a', 'b')
 *    ('urol@16', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 16, 'b')))
 *    ('urol@32', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 32, 'b')))
 *    ('uror@16', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 16, 'b')))
 *    ('uror@32', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 32, 'b')))
 *    ('~fexp2', ('flog2', 'a')) => a
 *    ('~flog2', ('fexp2', 'a')) => a
 *    ('fpow', 'a', 'b') => ('fexp2', ('fmul', ('flog2', 'a'), 'b'))
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 'b')) => ('fpow', 'a', 'b')
 *    ('~fexp2', ('fadd', ('fmul', ('flog2', 'a'), 'b'), ('fmul', ('flog2', 'c'), 'd'))) => ('~fmul', ('fpow', 'a', 'b'), ('fpow', 'c', 'd'))
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 0.5)) => ('fsqrt', 'a')
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 2.0)) => ('fmul', 'a', 'a')
 *    ('~fexp2', ('fmul', ('flog2', 'a'), 4.0)) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a'))
 *    ('~fpow', 'a', 1.0) => a
 *    ('~fpow', 'a', 2.0) => ('fmul', 'a', 'a')
 *    ('~fpow', 'a', 4.0) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a'))
 *    ('~fpow', 2.0, 'a') => ('fexp2', 'a')
 *    ('~fpow', ('fpow', 'a', 2.2), 0.454545) => a
 *    ('~fpow', ('fabs', ('fpow', 'a', 2.2)), 0.454545) => ('fabs', 'a')
 *    ('~fsqrt', ('fexp2', 'a')) => ('fexp2', ('fmul', 0.5, 'a'))
 *    ('~frcp', ('fexp2', 'a')) => ('fexp2', ('fneg', 'a'))
 *    ('~frsq', ('fexp2', 'a')) => ('fexp2', ('fmul', -0.5, 'a'))
 *    ('~flog2', ('fsqrt', 'a')) => ('fmul', 0.5, ('flog2', 'a'))
 *    ('~flog2', ('frcp', 'a')) => ('fneg', ('flog2', 'a'))
 *    ('~flog2', ('frsq', 'a')) => ('fmul', -0.5, ('flog2', 'a'))
 *    ('~flog2', ('fpow', 'a', 'b')) => ('fmul', 'b', ('flog2', 'a'))
 *    ('~fmul', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('fadd', 'a', 'b'))
 *    ('bcsel', ('flt', 'a', 0.0), 0.0, ('fsqrt', 'a')) => ('fsqrt', ('fmax', 'a', 0.0))
 *    ('~fmul', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a')
 *    ('~fmulz', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a')
 *    ('~fdiv', 1.0, 'a') => ('frcp', 'a')
 *    ('fdiv', 'a', 'b') => ('fmul', 'a', ('frcp', 'b'))
 *    ('~frcp', ('frcp', 'a')) => a
 *    ('~frcp', ('fsqrt', 'a')) => ('frsq', 'a')
 *    ('fsqrt', 'a') => ('frcp', ('frsq', 'a'))
 *    ('~frcp', ('frsq', 'a')) => ('fsqrt', 'a')
 *    ('fsin', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))
 *    ('fcos', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))
 *    ('i2b16(is_used_by_if)', 'a') => ('ine16', 'a', 0)
 *    ('i2b32(is_used_by_if)', 'a') => ('ine32', 'a', 0)
 *    ('i2b1(is_used_by_if)', 'a') => ('ine', 'a', 0)
 *    ('ieq', 'a', True) => a
 *    ('ine(is_not_used_by_if)', 'a', True) => ('inot', 'a')
 *    ('ine', 'a', False) => a
 *    ('ieq(is_not_used_by_if)', 'a', False) => ('inot', 'a')
 *    ('bcsel', 'a', True, False) => a
 *    ('bcsel', 'a', False, True) => ('inot', 'a')
 *    ('bcsel', True, 'b', 'c') => b
 *    ('bcsel', False, 'b', 'c') => c
 *    ('bcsel@16', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@16', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@16', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@16', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel@32', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@32', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@32', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@32', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel@64', 'a', 1.0, 0.0) => ('b2f', 'a')
 *    ('bcsel@64', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a'))
 *    ('bcsel@64', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a'))
 *    ('bcsel@64', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a')))
 *    ('bcsel', 'a', 'b', 'b') => b
 *    ('~fcsel', 'a', 'b', 'b') => b
 *    ('bcsel', 'a', -1, 0) => ('ineg', ('b2i', 'a@1'))
 *    ('bcsel', 'a', 0, -1) => ('ineg', ('b2i', ('inot', 'a')))
 *    ('bcsel', 'a', 1, 0) => ('b2i', 'a@1')
 *    ('bcsel', 'a', 0, 1) => ('b2i', ('inot', 'a'))
 *    ('iand', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('ior', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('ieq', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a')
 *    ('ieq', ('ineg', ('b2i', 'a@1')), -1) => a
 *    ('ine', ('ineg', ('b2i', 'a@1')), 0) => a
 *    ('ine', ('ineg', ('b2i', 'a@1')), -1) => ('inot', 'a')
 *    ('ige', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a')
 *    ('ilt', ('ineg', ('b2i', 'a@1')), 0) => a
 *    ('ult', 0, ('ineg', ('b2i', 'a@1'))) => a
 *    ('iand', ('ineg', ('b2i', 'a')), 1.0) => ('b2f', 'a')
 *    ('iand', ('ineg', ('b2i', 'a')), 1) => ('b2i', 'a')
 *    ('imax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('imin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('umax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b')))
 *    ('umin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b')))
 *    ('i2b16', ('b2i', 'a@16')) => a
 *    ('i2b32', ('b2i', 'a@32')) => a
 *    ('f2i', ('ftrunc', 'a')) => ('f2i', 'a')
 *    ('f2u', ('ftrunc', 'a')) => ('f2u', 'a')
 *    ('i2b', ('ineg', 'a')) => ('i2b', 'a')
 *    ('i2b', ('iabs', 'a')) => ('i2b', 'a')
 *    ('inot', ('f2b1', 'a')) => ('feq', 'a', 0.0)
 *    ('f2fmp', ('f2f32', 'a@16')) => a
 *    ('i2imp', ('i2i32', 'a@16')) => a
 *    ('i2imp', ('u2u32', 'a@16')) => a
 *    ('f2imp', ('f2f32', 'a@16')) => ('f2i16', 'a')
 *    ('f2ump', ('f2f32', 'a@16')) => ('f2u16', 'a')
 *    ('i2fmp', ('i2i32', 'a@16')) => ('i2f16', 'a')
 *    ('u2fmp', ('u2u32', 'a@16')) => ('u2f16', 'a')
 *    ('f2fmp', ('b2f32', 'a@1')) => ('b2f16', 'a')
 *    ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a')
 *    ('f2imp', ('b2f32', 'a@1')) => ('b2i16', 'a')
 *    ('f2ump', ('b2f32', 'a@1')) => ('b2i16', 'a')
 *    ('i2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a')
 *    ('u2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a')
 *    ('f2f32', ('f2fmp', 'a@32')) => a
 *    ('i2i32', ('i2imp', 'a@32')) => a
 *    ('u2u32', ('i2imp', 'a@32')) => a
 *    ('i2i32', ('f2imp', 'a@32')) => ('f2i32', 'a')
 *    ('u2u32', ('f2ump', 'a@32')) => ('f2u32', 'a')
 *    ('f2f32', ('i2fmp', 'a@32')) => ('i2f32', 'a')
 *    ('f2f32', ('u2fmp', 'a@32')) => ('u2f32', 'a')
 *    ('~f2f32', ('f2f64', 'a@32')) => a
 *    ('ffloor', 'a(is_integral)') => a
 *    ('fceil', 'a(is_integral)') => a
 *    ('ftrunc', 'a(is_integral)') => a
 *    ('fround_even', 'a(is_integral)') => a
 *    ('~ffract', 'a(is_integral)') => 0.0
 *    ('fabs', 'a(is_not_negative)') => a
 *    ('iabs', 'a(is_not_negative)') => a
 *    ('fsat', 'a(is_not_positive)') => 0.0
 *    ('~fmin', 'a(is_not_negative)', 1.0) => ('fsat', 'a')
 *    ('flt', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => False
 *    ('flt', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => False
 *    ('fmax', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0)
 *    ('fmax', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0)
 *    ('fneu', 'a(is_not_zero)', 0.0) => True
 *    ('feq', 'a(is_not_zero)', 0.0) => False
 *    ('fge', 'a(is_a_number_not_negative)', 'b(is_a_number_not_positive)') => True
 *    ('fge', 'a(is_not_positive)', 'b(is_gt_zero)') => False
 *    ('fge', 'a(is_lt_zero)', 'b(is_not_negative)') => False
 *    ('flt', 'a(is_not_negative)', 'b(is_not_positive)') => False
 *    ('flt', 'a(is_a_number_not_positive)', 'b(is_a_number_gt_zero)') => True
 *    ('flt', 'a(is_a_number_lt_zero)', 'b(is_a_number_not_negative)') => True
 *    ('ine', 'a(is_not_zero)', 0) => True
 *    ('ieq', 'a(is_not_zero)', 0) => False
 *    ('ige', 'a(is_not_negative)', 'b(is_not_positive)') => True
 *    ('ige', 'a(is_not_positive)', 'b(is_gt_zero)') => False
 *    ('ige', 'a(is_lt_zero)', 'b(is_not_negative)') => False
 *    ('ilt', 'a(is_not_negative)', 'b(is_not_positive)') => False
 *    ('ilt', 'a(is_not_positive)', 'b(is_gt_zero)') => True
 *    ('ilt', 'a(is_lt_zero)', 'b(is_not_negative)') => True
 *    ('ult', 0, 'a(is_gt_zero)') => True
 *    ('ult', 'a', 0) => False
 *    ('unpack_64_2x32_split_x', ('pack_64_2x32_split', 'a', 'b')) => a
 *    ('unpack_64_2x32_split_y', ('pack_64_2x32_split', 'a', 'b')) => b
 *    ('unpack_64_2x32_split_x', ('pack_64_2x32', 'a')) => a.x
 *    ('unpack_64_2x32_split_y', ('pack_64_2x32', 'a')) => a.y
 *    ('unpack_64_2x32', ('pack_64_2x32_split', 'a', 'b')) => ('vec2', 'a', 'b')
 *    ('unpack_64_2x32', ('pack_64_2x32', 'a')) => a
 *    ('unpack_double_2x32_dxil', ('pack_double_2x32_dxil', 'a')) => a
 *    ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')) => a
 *    ('pack_64_2x32', ('vec2', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a'))) => a
 *    ('pack_64_2x32', ('unpack_64_2x32', 'a')) => a
 *    ('pack_double_2x32_dxil', ('unpack_double_2x32_dxil', 'a')) => a
 *    ('iand', ('ieq', ('unpack_32_2x16_split_x', 'a'), '#b'), ('ieq', ('unpack_32_2x16_split_y', 'a'), '#c')) => ('ieq', 'a', ('pack_32_2x16_split', 'b', 'c'))
 *    ('ushr', 'a@16', 8) => ('extract_u8', 'a', 1)
 *    ('ushr', 'a@32', 24) => ('extract_u8', 'a', 3)
 *    ('ushr', 'a@64', 56) => ('extract_u8', 'a', 7)
 *    ('ishr', 'a@16', 8) => ('extract_i8', 'a', 1)
 *    ('ishr', 'a@32', 24) => ('extract_i8', 'a', 3)
 *    ('ishr', 'a@64', 56) => ('extract_i8', 'a', 7)
 *    ('iand', 255, 'a') => ('extract_u8', 'a', 0)
 *    ('u2u8', ('extract_u16', 'a', 1)) => ('u2u8', ('extract_u8', 'a', 2))
 *    ('u2u8', ('ushr', 'a', 8)) => ('u2u8', ('extract_u8', 'a', 1))
 *    ('i2i16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('i2i16', ('extract_i8', 'a', 'b'))
 *    ('u2u16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('u2u16', ('extract_u8', 'a', 'b'))
 *    ('ubfe', 'a', 0, 8) => ('extract_u8', 'a', 0)
 *    ('ubfe', 'a', 8, 8) => ('extract_u8', 'a', 1)
 *    ('ubfe', 'a', 16, 8) => ('extract_u8', 'a', 2)
 *    ('ubfe', 'a', 24, 8) => ('extract_u8', 'a', 3)
 *    ('ibfe', 'a', 0, 8) => ('extract_i8', 'a', 0)
 *    ('ibfe', 'a', 8, 8) => ('extract_i8', 'a', 1)
 *    ('ibfe', 'a', 16, 8) => ('extract_i8', 'a', 2)
 *    ('ibfe', 'a', 24, 8) => ('extract_i8', 'a', 3)
 *    ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('ushr', ('ishl', 'a@32', 16), 16) => ('extract_u16', 'a', 0)
 *    ('ushr', 'a@32', 16) => ('extract_u16', 'a', 1)
 *    ('ishr', ('ishl', 'a@32', 16), 16) => ('extract_i16', 'a', 0)
 *    ('ishr', 'a@32', 16) => ('extract_i16', 'a', 1)
 *    ('iand', 65535, 'a') => ('extract_u16', 'a', 0)
 *    ('ubfe', 'a', 0, 16) => ('extract_u16', 'a', 0)
 *    ('ubfe', 'a', 16, 16) => ('extract_u16', 'a', 1)
 *    ('ibfe', 'a', 0, 16) => ('extract_i16', 'a', 0)
 *    ('ibfe', 'a', 16, 16) => ('extract_i16', 'a', 1)
 *    ('ior', ('ishl', ('u2u32', 'a@8'), 24), ('ior', ('ishl', ('u2u32', 'b@8'), 16), ('ior', ('ishl', ('u2u32', 'c@8'), 8), ('u2u32', 'd@8')))) => ('pack_32_4x8', ('vec4', 'd', 'c', 'b', 'a'))
 *    ('extract_u16', ('extract_i16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b')
 *    ('extract_u16', ('extract_u16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b')
 *    ('pack_64_2x32_split', 'a', 'b') => ('ior', ('u2u64', 'a'), ('ishl', ('u2u64', 'b'), 32))
 *    ('pack_32_2x16_split', 'a', 'b') => ('ior', ('u2u32', 'a'), ('ishl', ('u2u32', 'b'), 16))
 *    ('unpack_64_2x32_split_x', 'a') => ('u2u32', 'a')
 *    ('unpack_64_2x32_split_y', 'a') => ('u2u32', ('ushr', 'a', 32))
 *    ('unpack_32_2x16_split_x', 'a') => ('u2u16', 'a')
 *    ('unpack_32_2x16_split_y', 'a') => ('u2u16', ('ushr', 'a', 16))
 *    ('unpack_half_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_half_2x16_split_x', 'a')
 *    ('unpack_32_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_32_2x16_split_x', 'a')
 *    ('unpack_64_2x32_split_x', ('iand', 'a', 4294967295)) => ('unpack_64_2x32_split_x', 'a')
 *    ('unpack_half_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_32_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_32_2x16_split_y', 'a')
 *    ('unpack_64_2x32_split_y', ('iand', 'a', 18446744069414584320)) => ('unpack_64_2x32_split_y', 'a')
 *    ('unpack_half_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_half_2x16_split_x', 'a')
 *    ('unpack_half_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_half_2x16_split_x', ('ushr', 'a', 16)) => ('unpack_half_2x16_split_y', 'a')
 *    ('unpack_32_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_32_2x16_split_x', 'a')
 *    ('unpack_32_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_32_2x16_split_y', 'a')
 *    ('ishl', ('pack_half_2x16', ('vec2', 'a', 0)), 16) => ('pack_half_2x16', ('vec2', 0, 'a'))
 *    ('ushr', ('pack_half_2x16', ('vec2', 0, 'a')), 16) => ('pack_half_2x16', ('vec2', 'a', 0))
 *    ('iadd', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b'))
 *    ('ior', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b'))
 *    ('ishl', ('pack_half_2x16_split', 'a', 0), 16) => ('pack_half_2x16_split', 0, 'a')
 *    ('ushr', ('pack_half_2x16_split', 0, 'a'), 16) => ('pack_half_2x16_split', 'a', 0)
 *    ('extract_u16', ('pack_half_2x16_split', 0, 'a'), 1) => ('pack_half_2x16_split', 'a', 0)
 *    ('iadd', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b')
 *    ('ior', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('i2i', 'a')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('i2i', 'b')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('i2i', 'c')
 *    ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('i2i', 'd')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('u2u', 'a')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('u2u', 'b')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('u2u', 'c')
 *    ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('u2u', 'd')
 *    ('extract_u8', ('ushr', 'a@16', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@32', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@32', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ushr', 'a@32', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ushr', 'a@64', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ushr', 'a@64', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ushr', 'a@64', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ushr', 'a@64', 32), 0) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ushr', 'a@64', 40), 0) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ushr', 'a@64', 48), 0) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ushr', 'a@64', 56), 0) => ('extract_u8', 'a', 7)
 *    ('extract_u8', ('ishr', 'a@16', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@32', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@32', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishr', 'a@32', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishr', 'a@64', 8), 0) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishr', 'a@64', 16), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishr', 'a@64', 24), 0) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishr', 'a@64', 32), 0) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ishr', 'a@64', 40), 0) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ishr', 'a@64', 48), 0) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ishr', 'a@64', 56), 0) => ('extract_u8', 'a', 7)
 *    ('extract_u8', ('extract_u16', 'a', 1), 0) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@16', 8), 1) => ('extract_u8', 'a', 0)
 *    ('extract_u8', ('ishl', 'a@32', 8), 3) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@32', 16), 3) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishl', 'a@32', 24), 3) => ('extract_u8', 'a', 0)
 *    ('extract_u8', ('ishl', 'a@64', 8), 7) => ('extract_u8', 'a', 6)
 *    ('extract_u8', ('ishl', 'a@64', 16), 7) => ('extract_u8', 'a', 5)
 *    ('extract_u8', ('ishl', 'a@64', 24), 7) => ('extract_u8', 'a', 4)
 *    ('extract_u8', ('ishl', 'a@64', 32), 7) => ('extract_u8', 'a', 3)
 *    ('extract_u8', ('ishl', 'a@64', 40), 7) => ('extract_u8', 'a', 2)
 *    ('extract_u8', ('ishl', 'a@64', 48), 7) => ('extract_u8', 'a', 1)
 *    ('extract_u8', ('ishl', 'a@64', 56), 7) => ('extract_u8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@16', 8), 1) => ('extract_i8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@32', 8), 3) => ('extract_i8', 'a', 2)
 *    ('extract_i8', ('ishl', 'a@32', 16), 3) => ('extract_i8', 'a', 1)
 *    ('extract_i8', ('ishl', 'a@32', 24), 3) => ('extract_i8', 'a', 0)
 *    ('extract_i8', ('ishl', 'a@64', 8), 7) => ('extract_i8', 'a', 6)
 *    ('extract_i8', ('ishl', 'a@64', 16), 7) => ('extract_i8', 'a', 5)
 *    ('extract_i8', ('ishl', 'a@64', 24), 7) => ('extract_i8', 'a', 4)
 *    ('extract_i8', ('ishl', 'a@64', 32), 7) => ('extract_i8', 'a', 3)
 *    ('extract_i8', ('ishl', 'a@64', 40), 7) => ('extract_i8', 'a', 2)
 *    ('extract_i8', ('ishl', 'a@64', 48), 7) => ('extract_i8', 'a', 1)
 *    ('extract_i8', ('ishl', 'a@64', 56), 7) => ('extract_i8', 'a', 0)
 *    ('ussub_4x8_vc4', 'a', 0) => a
 *    ('ussub_4x8_vc4', 'a', -1) => 0
 *    ('fsub', 'a', 'b') => ('fadd', 'a', ('fneg', 'b'))
 *    ('isub', 'a', 'b') => ('iadd', 'a', ('ineg', 'b'))
 *    ('uabs_usub', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b'))
 *    ('uabs_isub', 'a', 'b') => ('bcsel', ('ilt', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b'))
 *    ('fmul(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmul', 'a', 'b'))
 *    ('fmulz(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmulz', 'a', 'b'))
 *    ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('ffmaz', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('imul', ('ineg', 'a'), 'b') => ('ineg', ('imul', 'a', 'b'))
 *    ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('~fmulz(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmulz', ('fmulz', 'a', 'c'), 'b')
 *    ('~fmul(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)') => ('fmulz', ('fmul', 'a', 'c'), 'b')
 *    ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~ffma', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd')
 *    ('~ffma', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)', 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_fmul)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('ffma(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffma', 'a', 'b', 'd'), 'c')
 *    ('~fadd(is_used_once)', ('ffmaz(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffmaz', 'a', 'b', 'd'), 'c')
 *    ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('~fmul', '#a', ('fmul', 'b(is_not_const)', '#c')) => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('~fmulz', '#a', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmulz', 'a', 'c'), 'b')
 *    ('~fmul', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmul', 'a', 'c'), 'b')
 *    ('~ffma', '#a', ('fmul', 'b(is_not_const)', '#c'), 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', '#a', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd')
 *    ('~ffmaz', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd')
 *    ('imul', '#a', ('imul', 'b(is_not_const)', '#c')) => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~fadd', '#a', ('fadd', 'b(is_not_const)', '#c')) => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('~fadd', '#a', ('fneg', ('fadd', 'b(is_not_const)', '#c'))) => ('fadd', ('fadd', 'a', ('fneg', 'c')), ('fneg', 'b'))
 *    ('~fadd', '#a', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffma', 'b', 'c', ('fadd', 'a', 'd'))
 *    ('~fadd', '#a', ('fneg', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffma', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd')))
 *    ('~fadd', '#a', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffmaz', 'b', 'c', ('fadd', 'a', 'd'))
 *    ('~fadd', '#a', ('fneg', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffmaz', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd')))
 *    ('iadd', '#a', ('iadd', 'b(is_not_const)', '#c')) => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('iand', '#a', ('iand', 'b(is_not_const)', '#c')) => ('iand', ('iand', 'a', 'c'), 'b')
 *    ('ior', '#a', ('ior', 'b(is_not_const)', '#c')) => ('ior', ('ior', 'a', 'c'), 'b')
 *    ('ixor', '#a', ('ixor', 'b(is_not_const)', '#c')) => ('ixor', ('ixor', 'a', 'c'), 'b')
 *    ('~fadd', ('fadd(is_used_once)', 'a(is_fmul)', 'b(is_fmul)'), 'c(is_not_fmul)') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('idiv', ('imul(no_signed_wrap)', 'a', 'b'), 'b') => a
 *    ('bcsel', ('ige', ('find_lsb', 'a'), 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a')
 *    ('bcsel', ('ige', ('ifind_msb', 'a'), 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ige', ('ufind_msb', 'a'), 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('bcsel', ('ine', 'a', 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a')
 *    ('bcsel', ('ine', 'a', -1), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a')
 *    ('~fmul', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b')
 *    ('~fmul', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b'))
 *    ('~fmulz', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b')
 *    ('~fmulz', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b'))
 *    ('~bcsel', ('flt', 'a', 0.0), ('fneg', 'a'), 'a') => ('fabs', 'a')
 *    ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), 'd') => ('bcsel', ('iand', 'a', 'b'), 'c', 'd')
 *    ('bcsel', 'a', 'b', ('bcsel', 'c', 'b', 'd')) => ('bcsel', ('ior', 'a', 'c'), 'b', 'd')
 *    ('fmod', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ffloor', ('fdiv', 'a', 'b'))))
 *    ('frem', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ftrunc', ('fdiv', 'a', 'b'))))
 *    ('uadd_carry', 'a', 'b') => ('b2i', ('ult', ('iadd', 'a', 'b'), 'a'))
 *    ('usub_borrow', 'a', 'b') => ('b2i', ('ult', 'a', 'b'))
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bfi', ('bfm', 'bits', 'offset'), 'insert', 'base'))
 *    ('ihadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('uhadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('irhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('urhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('ihadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('uhadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('irhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1))
 *    ('urhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1))
 *    ('imul_32x16', 'a', 'b') => ('imul', 'a', ('extract_i16', 'b', 0))
 *    ('umul_32x16', 'a', 'b') => ('imul', 'a', ('extract_u16', 'b', 0))
 *    ('uadd_sat@64', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b'))
 *    ('uadd_sat', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b'))
 *    ('usub_sat', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b'))
 *    ('usub_sat@64', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b'))
 *    ('iadd_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), ('iadd', 'a', 'b'), 9223372036854775807))
 *    ('isub_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), ('isub', 'a', 'b'), 9223372036854775807))
 *    ('ilt', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ilt', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ige', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ige', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0)
 *    ('ilt', 'a@64', 0) => ('ilt', ('unpack_64_2x32_split_y', 'a'), 0)
 *    ('ige', 'a@64', 0) => ('ige', ('unpack_64_2x32_split_y', 'a'), 0)
 *    ('ine', 'a@64', 0) => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('ieq', 'a@64', 0) => ('ieq', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('ult', 0, 'a@64') => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0)
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('ior', ('iand', 'base', ('inot', ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))), ('iand', ('ishl', 'insert', 'offset'), ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))))
 *    ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bitfield_select', ('bfm', 'bits', 'offset'), ('ishl', 'insert', 'offset'), 'base'))
 *    ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ibfe', 'value', 'offset', 'bits'))
 *    ('ubitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ubfe', 'value', 'offset', 'bits'))
 *    ('bitfield_select', 'a', 'b', 0) => ('iand', 'a', 'b')
 *    ('bitfield_select', 'a', ('iand', 'a', 'b'), 'c') => ('bitfield_select', 'a', 'b', 'c')
 *    ('ubfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ubfe', 'value', 'offset', 'bits')
 *    ('ubfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ubfe', 'value', 'offset', 'bits')
 *    ('ibfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ibfe', 'value', 'offset', 'bits')
 *    ('ibfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ibfe', 'value', 'offset', 'bits')
 *    ('bfm', 'bits', ('iand', 31, 'offset')) => ('bfm', 'bits', 'offset')
 *    ('bfm', ('iand', 31, 'bits'), 'offset') => ('bfm', 'bits', 'offset')
 *    ('ubfe', 'a', 'b', 0) => 0
 *    ('ibfe', 'a', 'b', 0) => 0
 *    ('ubfe', 'a', 0, '#b') => ('iand', 'a', ('ushr', 4294967295, ('ineg', 'b')))
 *    ('b2i32', ('i2b', ('ubfe', 'a', 'b', 1))) => ('ubfe', 'a', 'b', 1)
 *    ('b2i32', ('i2b', ('ibfe', 'a', 'b', 1))) => ('ubfe', 'a', 'b', 1)
 *    ('ine', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ieq', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ine', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ieq', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0)
 *    ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ieq', 0, 'bits'), 0, ('ishr', ('ishl', 'value', ('isub', ('isub', 32, 'bits'), 'offset')), ('isub', 32, 'bits')))
 *    ('ubitfield_extract', 'value', 'offset', 'bits') => ('iand', ('ushr', 'value', 'offset'), ('bcsel', ('ieq', 'bits', 32), 4294967295, ('isub', ('ishl', 1, 'bits'), 1)))
 *    ('ifind_msb', 'value') => ('ufind_msb', ('bcsel', ('ilt', 'value', 0), ('inot', 'value'), 'value'))
 *    ('ifind_msb', 'value') => ('bcsel', ('ige', ('ifind_msb_rev', 'value'), 0), ('isub', 31, ('ifind_msb_rev', 'value')), ('ifind_msb_rev', 'value'))
 *    ('ufind_msb', 'value') => ('bcsel', ('ige', ('ufind_msb_rev', 'value'), 0), ('isub', 31, ('ufind_msb_rev', 'value')), ('ufind_msb_rev', 'value'))
 *    ('find_lsb', 'value') => ('ufind_msb', ('iand', 'value', ('ineg', 'value')))
 *    ('extract_i8', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 3, 'b'), 8)), 24)
 *    ('extract_u8', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 8)), 255)
 *    ('extract_i16', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 1, 'b'), 16)), 16)
 *    ('extract_u16', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 16)), 65535)
 *    ('pack_unorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 65535.0))))
 *    ('pack_unorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 255.0))))
 *    ('pack_snorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 32767.0))))
 *    ('pack_snorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 127.0))))
 *    ('unpack_unorm_2x16', 'v') => ('fdiv', ('u2f32', ('vec2', ('extract_u16', 'v', 0), ('extract_u16', 'v', 1))), 65535.0)
 *    ('unpack_unorm_4x8', 'v') => ('fdiv', ('u2f32', ('vec4', ('extract_u8', 'v', 0), ('extract_u8', 'v', 1), ('extract_u8', 'v', 2), ('extract_u8', 'v', 3))), 255.0)
 *    ('unpack_snorm_2x16', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec2', ('extract_i16', 'v', 0), ('extract_i16', 'v', 1))), 32767.0)))
 *    ('unpack_snorm_4x8', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec4', ('extract_i8', 'v', 0), ('extract_i8', 'v', 1), ('extract_i8', 'v', 2), ('extract_i8', 'v', 3))), 127.0)))
 *    ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('u2u32', ('f2f16', 'b')), 16), ('u2u32', ('f2f16', 'a')))
 *    ('unpack_half_2x16_split_x', 'a@32') => ('f2f32', ('u2u16', 'a'))
 *    ('unpack_half_2x16_split_y', 'a@32') => ('f2f32', ('u2u16', ('ushr', 'a', 16)))
 *    ('pack_32_2x16_split', 'a@16', 'b@16') => ('ior', ('ishl', ('u2u32', 'b'), 16), ('u2u32', 'a'))
 *    ('unpack_32_2x16_split_x', 'a@32') => ('u2u16', 'a')
 *    ('unpack_32_2x16_split_y', 'a@32') => ('u2u16', ('ushr', 'a', 16))
 *    ('isign', 'a') => ('imin', ('imax', 'a', -1), 1)
 *    ('imin', ('imax', 'a', -1), 1) => ('isign', 'a')
 *    ('imax', ('imin', 'a', 1), -1) => ('isign', 'a')
 *    ('fsign', 'a') => ('fsub', ('b2f', ('!flt', 0.0, 'a')), ('b2f', ('!flt', 'a', 0.0)))
 *    ('amul', 'a', 'b') => ('imul', 'a', 'b')
 *    ('umul24', 'a', 'b') => ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215))
 *    ('umad24', 'a', 'b', 'c') => ('iadd', ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)), 'c')
 *    ('imul24_relaxed', 'a', 'b') => ('imul24', 'a', 'b')
 *    ('imul24_relaxed', 'a', 'b') => ('imul', 'a', 'b')
 *    ('umad24_relaxed', 'a', 'b', 'c') => ('umad24', 'a', 'b', 'c')
 *    ('umad24_relaxed', 'a', 'b', 'c') => ('iadd', ('umul24_relaxed', 'a', 'b'), 'c')
 *    ('umul24_relaxed', 'a', 'b') => ('umul24', 'a', 'b')
 *    ('umul24_relaxed', 'a', 'b') => ('imul', 'a', 'b')
 *    ('imad24_ir3', 'a', 'b', 0) => ('imul24', 'a', 'b')
 *    ('imad24_ir3', 'a', 0, 'c') => c
 *    ('imad24_ir3', 'a', 1, 'c') => ('iadd', 'a', 'c')
 *    ('imad24_ir3', '#a', '#b', 'c') => ('iadd', ('imul', 'a', 'b'), 'c')
 *    ('imul24', 'a', '#b@32(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b'))
 *    ('imul24', 'a', '#b@32(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b'))))
 *    ('imul24', 'a', 0) => 0
 *    ('fcsel', ('slt', 0, 'a'), 'b', 'c') => ('fcsel_gt', 'a', 'b', 'c')
 *    ('fcsel', ('slt', 'a', 0), 'b', 'c') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c')
 *    ('fcsel', ('sge', 'a', 0), 'b', 'c') => ('fcsel_ge', 'a', 'b', 'c')
 *    ('fcsel', ('sge', 0, 'a'), 'b', 'c') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c')
 *    ('bcsel', ('ilt', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'b', 'c')
 *    ('bcsel', ('ilt', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'c', 'b')
 *    ('bcsel', ('ige', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'b', 'c')
 *    ('bcsel', ('ige', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'c', 'b')
 *    ('bcsel', ('flt', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_gt', 'a', 'b', 'c')
 *    ('bcsel', ('flt', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c')
 *    ('bcsel', ('fge', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_ge', 'a', 'b', 'c')
 *    ('bcsel', ('fge', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c')
 *    ('iadd_sat@8', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 127, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 128, ('iadd', 'a', 'b')))
 *    ('isub_sat@8', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 127, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 128, ('isub', 'a', 'b')))
 *    ('iadd_sat@16', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 32767, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 32768, ('iadd', 'a', 'b')))
 *    ('isub_sat@16', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 32767, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 32768, ('isub', 'a', 'b')))
 *    ('iadd_sat@32', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 2147483647, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 2147483648, ('iadd', 'a', 'b')))
 *    ('isub_sat@32', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 2147483647, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 2147483648, ('isub', 'a', 'b')))
 *    ('iadd_sat@64', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 9223372036854775807, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 9223372036854775808, ('iadd', 'a', 'b')))
 *    ('isub_sat@64', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 9223372036854775807, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 9223372036854775808, ('isub', 'a', 'b')))
 *    ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))
 *    ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('ior(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('feq', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('inot', ('iand(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('feq', 'a', 'b'), ('feq', 'c', 'd'))
 *    ('f2b1', ('b2f', 'a@1')) => a
 *    ('i2b1', ('b2i', 'a@1')) => a
 *    ('f2b8', ('b2f', 'a@8')) => a
 *    ('i2b8', ('b2i', 'a@8')) => a
 *    ('f2b16', ('b2f', 'a@16')) => a
 *    ('i2b16', ('b2i', 'a@16')) => a
 *    ('f2b32', ('b2f', 'a@32')) => a
 *    ('i2b32', ('b2i', 'a@32')) => a
 *    ('f2f', ('b2f', 'a')) => ('b2f', 'a')
 *    ('f2u', ('b2f', 'a')) => ('b2i', 'a')
 *    ('f2i', ('b2f', 'a')) => ('b2i', 'a')
 *    ('u2f', ('b2i', 'a')) => ('b2f', 'a')
 *    ('u2u', ('b2i', 'a')) => ('b2i', 'a')
 *    ('i2f', ('b2i', 'a')) => ('b2f', 'a')
 *    ('i2i', ('b2i', 'a')) => ('b2i', 'a')
 *    ('i2i1', 'a@1') => a
 *    ('i2i8', 'a@8') => a
 *    ('i2i16', 'a@16') => a
 *    ('i2i32', 'a@32') => a
 *    ('i2i64', 'a@64') => a
 *    ('u2u1', 'a@1') => a
 *    ('u2u8', 'a@8') => a
 *    ('u2u16', 'a@16') => a
 *    ('u2u32', 'a@32') => a
 *    ('u2u64', 'a@64') => a
 *    ('f2f16', 'a@16') => a
 *    ('f2f32', 'a@32') => a
 *    ('f2f64', 'a@64') => a
 *    ('b2b1', 'a@1') => a
 *    ('b2b8', 'a@8') => a
 *    ('b2b16', 'a@16') => a
 *    ('b2b32', 'a@32') => a
 *    ('i2i1', ('i2i8', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u8', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i8', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u8', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i16', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u16', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i16', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u16', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i32', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u32', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i32', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u32', 'a')) => ('u2u1', 'a')
 *    ('i2i1', ('i2i64', 'a')) => ('i2i1', 'a')
 *    ('i2i1', ('u2u64', 'a')) => ('u2u1', 'a')
 *    ('u2u1', ('i2i64', 'a')) => ('i2i1', 'a')
 *    ('u2u1', ('u2u64', 'a')) => ('u2u1', 'a')
 *    ('i2i8', ('i2i1', 'a@1')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u1', 'a@1')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i16', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u16', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i16', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u16', 'a')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i32', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u32', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i32', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u32', 'a')) => ('u2u8', 'a')
 *    ('i2i8', ('i2i64', 'a')) => ('i2i8', 'a')
 *    ('i2i8', ('u2u64', 'a')) => ('u2u8', 'a')
 *    ('u2u8', ('i2i64', 'a')) => ('i2i8', 'a')
 *    ('u2u8', ('u2u64', 'a')) => ('u2u8', 'a')
 *    ('i2i16', ('i2i1', 'a@1')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u1', 'a@1')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i8', 'a@1')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u8', 'a@1')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i8', 'a@8')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u8', 'a@8')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i32', 'a')) => ('i2i16', 'a')
 *    ('i2i16', ('u2u32', 'a')) => ('u2u16', 'a')
 *    ('u2u16', ('i2i32', 'a')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u32', 'a')) => ('u2u16', 'a')
 *    ('i2i16', ('i2i64', 'a')) => ('i2i16', 'a')
 *    ('i2i16', ('u2u64', 'a')) => ('u2u16', 'a')
 *    ('u2u16', ('i2i64', 'a')) => ('i2i16', 'a')
 *    ('u2u16', ('u2u64', 'a')) => ('u2u16', 'a')
 *    ('i2i32', ('i2i1', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u1', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i8', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u8', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i8', 'a@8')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u8', 'a@8')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@1')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@1')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@8')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@8')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i16', 'a@16')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u16', 'a@16')) => ('u2u32', 'a')
 *    ('i2i32', ('i2i64', 'a')) => ('i2i32', 'a')
 *    ('i2i32', ('u2u64', 'a')) => ('u2u32', 'a')
 *    ('u2u32', ('i2i64', 'a')) => ('i2i32', 'a')
 *    ('u2u32', ('u2u64', 'a')) => ('u2u32', 'a')
 *    ('i2i64', ('i2i1', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u1', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i8', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u8', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i8', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u8', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i16', 'a@16')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u16', 'a@16')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@1')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@1')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@8')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@8')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@16')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@16')) => ('u2u64', 'a')
 *    ('i2i64', ('i2i32', 'a@32')) => ('i2i64', 'a')
 *    ('u2u64', ('u2u32', 'a@32')) => ('u2u64', 'a')
 *    ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a')
 *    ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a')
 *    ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a')
 *    ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a')
 *    ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a')
 *    ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a')
 *    ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a')
 *    ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a')
 *    ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a')
 *    ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a')
 *    ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a')
 *    ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a')
 *    ('ieq', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i16', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i16', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i16', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i16', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i16', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i16', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i32', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i32', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i32', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i32', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i32', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i32', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i8', 'b'))
 *    ('ine', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i8', 'b'))
 *    ('ilt', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i8', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ige', ('i2i8', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i64', ('i2i8', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i64', ('i2i8', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b')))
 *    ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i32', ('i2i16', 'b')), 'b'))
 *    ('ine', ('i2i32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i32', ('i2i16', 'b')), 'b'))
 *    ('ilt', ('i2i32', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b'))))
 *    ('ilt', '#a', ('i2i32', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b')))
 *    ('ige', ('i2i32', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b'))))
 *    ('ige', '#a', ('i2i32', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i16', 'b'))
 *    ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i16', 'b'))
 *    ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i16', 'b'))
 *    ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i64', ('i2i16', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i64', ('i2i16', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b')))
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ieq', 'a', ('i2i32', 'b'))
 *    ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ine', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ige', 'a', ('i2i32', 'b'))
 *    ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ilt', 'a', ('i2i32', 'b'))
 *    ('ige', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a')
 *    ('ilt', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a')
 *    ('ieq', ('i2i64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('i2i32', 'b')), ('ieq', ('i2i64', ('i2i32', 'b')), 'b'))
 *    ('ine', ('i2i64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('i2i32', 'b')), ('ine', ('i2i64', ('i2i32', 'b')), 'b'))
 *    ('ilt', ('i2i64', 'a@32'), '#b') => ('iand', ('ilt', -2147483648, 'b'), ('ior', ('ilt', 2147483647, 'b'), ('ilt', 'a', ('i2i32', 'b'))))
 *    ('ilt', '#a', ('i2i64', 'b@32')) => ('iand', ('ilt', 'a', 2147483647), ('ior', ('ilt', 'a', -2147483648), ('ilt', ('i2i32', 'a'), 'b')))
 *    ('ige', ('i2i64', 'a@32'), '#b') => ('iand', ('ige', 2147483647, 'b'), ('ior', ('ige', -2147483648, 'b'), ('ige', 'a', ('i2i32', 'b'))))
 *    ('ige', '#a', ('i2i64', 'b@32')) => ('iand', ('ige', 'a', -2147483648), ('ior', ('ige', 'a', 2147483647), ('ige', ('i2i32', 'a'), 'b')))
 *    ('ieq', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u16', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u16', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u16', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u16', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u16', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u16', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u32', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u32', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u32', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u32', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u32', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u32', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u8', 'b'))
 *    ('ine', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u8', 'b'))
 *    ('ult', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u8', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('uge', ('u2u8', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('ult', ('u2u8', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u64', ('u2u8', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u64', ('u2u8', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b')))
 *    ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u32', ('u2u16', 'b')), 'b'))
 *    ('ine', ('u2u32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u32', ('u2u16', 'b')), 'b'))
 *    ('ult', ('u2u32', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b'))))
 *    ('ult', '#a', ('u2u32', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b')))
 *    ('uge', ('u2u32', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b'))))
 *    ('uge', '#a', ('u2u32', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u16', 'b'))
 *    ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u16', 'b'))
 *    ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u16', 'b'))
 *    ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u64', ('u2u16', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u64', ('u2u16', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b')))
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ieq', 'a', ('u2u32', 'b'))
 *    ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ine', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('uge', 'a', ('u2u32', 'b'))
 *    ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ult', 'a', ('u2u32', 'b'))
 *    ('uge', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a')
 *    ('ult', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a')
 *    ('ieq', ('u2u64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('u2u32', 'b')), ('ieq', ('u2u64', ('u2u32', 'b')), 'b'))
 *    ('ine', ('u2u64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('u2u32', 'b')), ('ine', ('u2u64', ('u2u32', 'b')), 'b'))
 *    ('ult', ('u2u64', 'a@32'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 4294967295, 'b'), ('ult', 'a', ('u2u32', 'b'))))
 *    ('ult', '#a', ('u2u64', 'b@32')) => ('iand', ('ult', 'a', 4294967295), ('ior', ('ult', 'a', 0), ('ult', ('u2u32', 'a'), 'b')))
 *    ('uge', ('u2u64', 'a@32'), '#b') => ('iand', ('uge', 4294967295, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u32', 'b'))))
 *    ('uge', '#a', ('u2u64', 'b@32')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 4294967295), ('uge', ('u2u32', 'a'), 'b')))
 *    ('feq', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('feq', 'a', ('f2f16', 'b'))
 *    ('fneu', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fneu', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fge', 'a', ('f2f16', 'b'))
 *    ('flt', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('flt', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('fge', ('f2f16', 'b'), 'a')
 *    ('flt', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('flt', ('f2f16', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f16', 'b'))
 *    ('fneu', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f16', 'b'))
 *    ('flt', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f16', 'b'))
 *    ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('fge', ('f2f16', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('flt', ('f2f16', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f32', 'b'))
 *    ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f32', 'b'))
 *    ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a')
 *    ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('feq', 'a', ('f2f32', 'b'))
 *    ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fneu', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fge', 'a', ('f2f32', 'b'))
 *    ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('flt', 'a', ('f2f32', 'b'))
 *    ('fge', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a')
 *    ('flt', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a')
 *    ('i2i32', ('iand', 'a@64', 4294967295)) => ('u2u32', 'a')
 *    ('i2i16', ('iand', 'a@32', 65535)) => ('u2u16', 'a')
 *    ('i2i16', ('iand', 'a@64', 65535)) => ('u2u16', 'a')
 *    ('i2i8', ('iand', 'a@16', 255)) => ('u2u8', 'a')
 *    ('i2i8', ('iand', 'a@32', 255)) => ('u2u8', 'a')
 *    ('i2i8', ('iand', 'a@64', 255)) => ('u2u8', 'a')
 *    ('u2u16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a
 *    ('extract_u8(only_lower_8_bits_used)', 'a@16', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a
 *    ('extract_i8(only_lower_8_bits_used)', 'a@16', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ixor', 'a', 'b')
 *    ('u2u32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a
 *    ('extract_u8(only_lower_8_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a
 *    ('extract_i8(only_lower_8_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('u2u32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a
 *    ('extract_u16(only_lower_16_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('extract_u16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('i2i32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a
 *    ('extract_i16(only_lower_16_bits_used)', 'a@32', 0) => a
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('bcsel(only_lower_16_bits_used)', 'c', ('extract_i16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b')
 *    ('iadd(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iadd', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b')
 *    ('imul(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('imul', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b')
 *    ('iand(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iand', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b')
 *    ('ior(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ior', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b')
 *    ('ixor(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ixor', 'a', 'b')
 *    ('ldexp@16', 'x', 'exp') => ('fmul', ('fmul', 'x', ('i2i16', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -30), 30), 1), 15), 10))), ('i2i16', ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -30), 30), ('ishr', ('imin', ('imax', 'exp', -30), 30), 1)), 15), 10)))
 *    ('ldexp@32', 'x', 'exp') => ('fmul', ('fmul', 'x', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -254), 254), 1), 127), 23)), ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -254), 254), ('ishr', ('imin', ('imax', 'exp', -254), 254), 1)), 127), 23))
 *    ('ldexp@64', 'x', 'exp') => ('fmul', ('fmul', 'x', ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1), 1023), 20))), ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -2046), 2046), ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1)), 1023), 20)))
 *    ('ior(many-comm-expr)', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 1431655765), 1), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 2863311530), 1)) => ('bitfield_reverse', 'x')
 *    ('ior(many-comm-expr)', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 4278255360), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 16711935)) => ('bitfield_reverse', 'x')
 *    ('ball_iequal2', ('ieq', 'a', 'b'), -1) => ('ball_iequal2', 'a', 'b')
 *    ('ball_iequal2', ('feq', 'a', 'b'), -1) => ('ball_fequal2', 'a', 'b')
 *    ('bany_inequal2', ('ine', 'a', 'b'), 0) => ('bany_inequal2', 'a', 'b')
 *    ('bany_inequal2', ('fneu', 'a', 'b'), 0) => ('bany_fnequal2', 'a', 'b')
 *    ('ball_iequal3', ('ieq', 'a', 'b'), -1) => ('ball_iequal3', 'a', 'b')
 *    ('ball_iequal3', ('feq', 'a', 'b'), -1) => ('ball_fequal3', 'a', 'b')
 *    ('bany_inequal3', ('ine', 'a', 'b'), 0) => ('bany_inequal3', 'a', 'b')
 *    ('bany_inequal3', ('fneu', 'a', 'b'), 0) => ('bany_fnequal3', 'a', 'b')
 *    ('ball_iequal4', ('ieq', 'a', 'b'), -1) => ('ball_iequal4', 'a', 'b')
 *    ('ball_iequal4', ('feq', 'a', 'b'), -1) => ('ball_fequal4', 'a', 'b')
 *    ('bany_inequal4', ('ine', 'a', 'b'), 0) => ('bany_inequal4', 'a', 'b')
 *    ('bany_inequal4', ('fneu', 'a', 'b'), 0) => ('bany_fnequal4', 'a', 'b')
 *    ('ball_iequal8', ('ieq', 'a', 'b'), -1) => ('ball_iequal8', 'a', 'b')
 *    ('ball_iequal8', ('feq', 'a', 'b'), -1) => ('ball_fequal8', 'a', 'b')
 *    ('bany_inequal8', ('ine', 'a', 'b'), 0) => ('bany_inequal8', 'a', 'b')
 *    ('bany_inequal8', ('fneu', 'a', 'b'), 0) => ('bany_fnequal8', 'a', 'b')
 *    ('ball_iequal16', ('ieq', 'a', 'b'), -1) => ('ball_iequal16', 'a', 'b')
 *    ('ball_iequal16', ('feq', 'a', 'b'), -1) => ('ball_fequal16', 'a', 'b')
 *    ('bany_inequal16', ('ine', 'a', 'b'), 0) => ('bany_inequal16', 'a', 'b')
 *    ('bany_inequal16', ('fneu', 'a', 'b'), 0) => ('bany_fnequal16', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('flt', 'a', 'b')) => ('!flt', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('flt', 'b', 'a')) => ('!flt', 'b', 'a')
 *    ('iand', ('feq', 'a', 'a'), ('fge', 'a', 'b')) => ('!fge', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('fge', 'b', 'a')) => ('!fge', 'b', 'a')
 *    ('iand', ('feq', 'a', 'a'), ('feq', 'a', 'b')) => ('!feq', 'a', 'b')
 *    ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'a')) => ('!feq', 'b', 'a')
 *    ('feq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('feq', 'b', 'd'), ('feq', 'c', 'd'))
 *    ('fneu', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fneu', 'b', 'd'), ('fneu', 'c', 'd'))
 *    ('ieq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ieq', 'b', 'd'), ('ieq', 'c', 'd'))
 *    ('ine', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ine', 'b', 'd'), ('ine', 'c', 'd'))
 *    ('flt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('flt', 'b', 'd'), ('flt', 'c', 'd'))
 *    ('flt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('flt', 'd', 'b'), ('flt', 'd', 'c'))
 *    ('fge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fge', 'b', 'd'), ('fge', 'c', 'd'))
 *    ('fge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('fge', 'd', 'b'), ('fge', 'd', 'c'))
 *    ('ilt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ilt', 'b', 'd'), ('ilt', 'c', 'd'))
 *    ('ilt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ilt', 'd', 'b'), ('ilt', 'd', 'c'))
 *    ('ige', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ige', 'b', 'd'), ('ige', 'c', 'd'))
 *    ('ige', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ige', 'd', 'b'), ('ige', 'd', 'c'))
 *    ('ult', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ult', 'b', 'd'), ('ult', 'c', 'd'))
 *    ('ult', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ult', 'd', 'b'), ('ult', 'd', 'c'))
 *    ('uge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('uge', 'b', 'd'), ('uge', 'c', 'd'))
 *    ('uge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('uge', 'd', 'b'), ('uge', 'd', 'c'))
 *    ('fadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fadd', 'b', 'd'), ('fadd', 'c', 'd'))
 *    ('fmul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmul', 'b', 'd'), ('fmul', 'c', 'd'))
 *    ('fmulz', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmulz', 'b', 'd'), ('fmulz', 'c', 'd'))
 *    ('iadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('iadd', 'b', 'd'), ('iadd', 'c', 'd'))
 *    ('imul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('imul', 'b', 'd'), ('imul', 'c', 'd'))
 *    ('fddx', 'a') => 0.0
 *    ('fddx_fine', 'a') => 0.0
 *    ('fddx_coarse', 'a') => 0.0
 *    ('fddy', 'a') => 0.0
 *    ('fddy_fine', 'a') => 0.0
 *    ('fddy_coarse', 'a') => 0.0
 *    ('umul_low', '#a(is_lower_half_zero)', 'b') => 0
 *    ('imadsh_mix16', '#a@32(is_lower_half_zero)', 'b@32', 'c@32') => c
 *    ('imadsh_mix16', 'a@32', '#b@32(is_upper_half_zero)', 'c@32') => c
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd')
 *    ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd')
 *    ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c(is_not_const)'), ('fmulz', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz', 'b', 'c(is_not_const)'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmulz', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c(is_not_const)'), ('fmul', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul', 'b', 'c(is_not_const)'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmul', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c'), ('iadd', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c(is_not_const)'), ('iadd', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd', 'b', 'c(is_not_const)'), ('iadd(is_used_once)', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iadd', 'b', 'c'), ('iadd(is_used_once)', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c'), ('imul', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c(is_not_const)'), ('imul', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul', 'b', 'c(is_not_const)'), ('imul(is_used_once)', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imul', 'b', 'c'), ('imul(is_used_once)', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c'), ('iand', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c(is_not_const)'), ('iand', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand', 'b', 'c(is_not_const)'), ('iand(is_used_once)', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('iand', 'b', 'c'), ('iand(is_used_once)', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c'), ('ior', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c(is_not_const)'), ('ior', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior', 'b', 'c(is_not_const)'), ('ior(is_used_once)', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ior', 'b', 'c'), ('ior(is_used_once)', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c'), ('ixor', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c(is_not_const)'), ('ixor', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor', 'b', 'c(is_not_const)'), ('ixor(is_used_once)', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ixor', 'b', 'c'), ('ixor(is_used_once)', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c'), ('fmin', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c(is_not_const)'), ('fmin', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin', 'b', 'c(is_not_const)'), ('fmin(is_used_once)', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmin', 'b', 'c'), ('fmin(is_used_once)', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c'), ('fmax', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c(is_not_const)'), ('fmax', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax', 'b', 'c(is_not_const)'), ('fmax(is_used_once)', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fmax', 'b', 'c'), ('fmax(is_used_once)', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c'), ('imin', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c(is_not_const)'), ('imin', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin', 'b', 'c(is_not_const)'), ('imin(is_used_once)', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imin', 'b', 'c'), ('imin(is_used_once)', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c'), ('imax', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c(is_not_const)'), ('imax', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax', 'b', 'c(is_not_const)'), ('imax(is_used_once)', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('imax', 'b', 'c'), ('imax(is_used_once)', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c'), ('umin', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c(is_not_const)'), ('umin', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin', 'b', 'c(is_not_const)'), ('umin(is_used_once)', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umin', 'b', 'c'), ('umin(is_used_once)', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c'), ('umax', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c(is_not_const)'), ('umax', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax', 'b', 'c(is_not_const)'), ('umax(is_used_once)', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('umax', 'b', 'c'), ('umax(is_used_once)', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c')
 *    ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c')
 *    ('bcsel', 'c', ('frcp(is_used_once)', 'a'), ('frcp(is_used_once)', 'b')) => ('frcp', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('frsq(is_used_once)', 'a'), ('frsq(is_used_once)', 'b')) => ('frsq', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsqrt(is_used_once)', 'a'), ('fsqrt(is_used_once)', 'b')) => ('fsqrt', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('flog2(is_used_once)', 'a'), ('flog2(is_used_once)', 'b')) => ('flog2', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin(is_used_once)', 'a'), ('fsin(is_used_once)', 'b')) => ('fsin', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos(is_used_once)', 'a'), ('fcos(is_used_once)', 'b')) => ('fcos', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsin_amd(is_used_once)', 'a'), ('fsin_amd(is_used_once)', 'b')) => ('fsin_amd', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fcos_amd(is_used_once)', 'a'), ('fcos_amd(is_used_once)', 'b')) => ('fcos_amd', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fneg(is_used_once)', 'a'), ('fneg(is_used_once)', 'b')) => ('fneg', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fabs(is_used_once)', 'a'), ('fabs(is_used_once)', 'b')) => ('fabs', ('bcsel', 'c', 'a', 'b'))
 *    ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b'))
 *    ('ineg', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('ineg', 'a'), ('ineg', 'b'))
 *    ('iabs', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('iabs', 'a'), ('iabs', 'b'))
 *    ('inot', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('inot', 'a'), ('inot', 'b'))
 *    ('isign', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('isign', 'a'), ('isign', 'b'))
 *    ('fisnormal', 'a@16') => ('ult', 4095, ('iadd', ('ishl', 'a', 1), 2048))
 *    ('fisnormal', 'a@32') => ('ult', 33554431, ('iadd', ('ishl', 'a', 1), 16777216))
 *    ('fisnormal', 'a@64') => ('ult', 18014398509481983, ('iadd', ('ishl', 'a', 1), 9007199254740992))
 *    ('f2f16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2f16', 'a'), ('f2f16', 'b'))
 *    ('f2fmp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2fmp', 'a'), ('f2fmp', 'b'))
 *    ('u2u16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('u2u16', 'a'), ('u2u16', 'b'))
 *    ('i2i16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2i16', 'a'), ('i2i16', 'b'))
 *    ('i2imp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2imp', 'a'), ('i2imp', 'b'))
 *    ('f2f16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'))
 *    ('f2fmp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))
 *    ('u2u16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'))
 *    ('i2i16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'))
 *    ('i2imp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'))
 *    ('f2f16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'), ('f2f16', 'd'))
 *    ('f2fmp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'), ('f2fmp', 'd'))
 *    ('u2u16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'), ('u2u16', 'd'))
 *    ('i2i16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'), ('i2i16', 'd'))
 *    ('i2imp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'), ('i2imp', 'd'))
 */


static const nir_search_value_union nir_opt_algebraic_values[] = {
   /* ('imul', 'a', '#b(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 4 },
      -1,
   } },

   /* ('imul', 'a@8', 128) => ('ishl', 'a', 7) */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x80 /* 128 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 6, 7 },
      -1,
   } },

   /* replace1_0 -> 6 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7 /* 7 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 9 },
      -1,
   } },

   /* ('imul', 'a@16', 32768) => ('ishl', 'a', 15) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x8000 /* 32768 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 11, 12 },
      -1,
   } },

   /* replace2_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf /* 15 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 14 },
      -1,
   } },

   /* ('imul', 'a@32', 2147483648) => ('ishl', 'a', 31) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x80000000 /* 2147483648 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 17 },
      -1,
   } },

   /* replace3_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1f /* 31 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 19 },
      -1,
   } },

   /* ('imul', 'a@64', 9223372036854775808) => ('ishl', 'a', 63) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x8000000000000000 /* 9223372036854775808 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 21, 22 },
      -1,
   } },

   /* replace4_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3f /* 63 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 24 },
      -1,
   } },

   /* ('imul', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b')))) */
   /* search5_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 26 },
      -1,
   } },

   /* replace5_0_0 -> 0 in the cache */
   /* replace5_0_1_0_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 28 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 29 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 30 },
      -1,
   } },

   /* ('ishl', 'a', '#b') => ('imul', 'a', ('ishl', 1, 'b')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 32, 33 },
      -1,
   } },

   /* replace6_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 35, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 37 },
      -1,
   } },

   /* ('imul@64', 'a', '#b(is_bitcount2)') => ('iadd', ('ishl', 'a', ('ufind_msb', 'b')), ('ishl', 'a', ('find_lsb', 'b'))) */
   /* search7_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 21, 39 },
      -1,
   } },

   /* replace7_0_0 -> 21 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 42 },
      -1,
   } },
   /* replace7_1_0 -> 21 in the cache */
   /* replace7_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 44 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 43, 45 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', ('imul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b') */
   /* search8_0_0 -> 16 in the cache */
   /* search8_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul_2x32_64,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 47 },
      -1,
   } },

   /* replace8_0 -> 16 in the cache */
   /* replace8_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', ('umul_2x32_64(is_used_once)', 'a', 'b')) => ('imul', 'a', 'b') */
   /* search9_0_0 -> 16 in the cache */
   /* search9_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umul_2x32_64,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 50 },
      -1,
   } },

   /* replace9_0 -> 16 in the cache */
   /* replace9_1 -> 36 in the cache */
   /* replace9 -> 49 in the cache */

   /* ('imul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('imul_high', 'a', 'b')) */
   /* search10_0 -> 16 in the cache */
   /* search10_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imul_2x32_64,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace10_0_0 -> 16 in the cache */
   /* replace10_0_1 -> 36 in the cache */
   /* replace10_0 -> 49 in the cache */
   /* replace10_1_0 -> 16 in the cache */
   /* replace10_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul_high,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 2,
      { 49, 53 },
      -1,
   } },

   /* ('umul_2x32_64', 'a', 'b') => ('pack_64_2x32_split', ('imul', 'a', 'b'), ('umul_high', 'a', 'b')) */
   /* search11_0 -> 16 in the cache */
   /* search11_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umul_2x32_64,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace11_0_0 -> 16 in the cache */
   /* replace11_0_1 -> 36 in the cache */
   /* replace11_0 -> 49 in the cache */
   /* replace11_1_0 -> 16 in the cache */
   /* replace11_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_high,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 2,
      { 49, 56 },
      -1,
   } },

   /* ('udiv', 'a', 1) => a */
   /* search12_0 -> 32 in the cache */
   /* search12_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_udiv,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace12 -> 32 in the cache */

   /* ('idiv', 'a', 1) => a */
   /* search13_0 -> 32 in the cache */
   /* search13_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace13 -> 32 in the cache */

   /* ('umod', 'a', 1) => 0 */
   /* search14_0 -> 32 in the cache */
   /* search14_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umod,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },

   /* ('imod', 'a', 1) => 0 */
   /* search15_0 -> 32 in the cache */
   /* search15_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace15 -> 61 in the cache */

   /* ('imod', 'a', -1) => 0 */
   /* search16_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 32, 63 },
      -1,
   } },

   /* replace16 -> 61 in the cache */

   /* ('irem', 'a', 1) => 0 */
   /* search17_0 -> 32 in the cache */
   /* search17_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 32, 35 },
      -1,
   } },

   /* replace17 -> 61 in the cache */

   /* ('irem', 'a', -1) => 0 */
   /* search18_0 -> 32 in the cache */
   /* search18_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 32, 63 },
      -1,
   } },

   /* replace18 -> 61 in the cache */

   /* ('udiv', 'a', '#b(is_pos_power_of_two)') => ('ushr', 'a', ('find_lsb', 'b')) */
   /* search19_0 -> 0 in the cache */
   /* search19_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_udiv,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace19_0 -> 0 in the cache */
   /* replace19_1_0 -> 3 in the cache */
   /* replace19_1 -> 4 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 0, 4 },
      -1,
   } },

   /* ('idiv', 'a', '#b(is_pos_power_of_two)') => ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', 'b'))) */
   /* search20_0 -> 0 in the cache */
   /* search20_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace20_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace20_1_1_0 -> 3 in the cache */
   /* replace20_1_1 -> 4 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 71, 4 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 70, 72 },
      -1,
   } },

   /* ('idiv', 'a', '#b(is_neg_power_of_two)') => ('ineg', ('imul', ('isign', 'a'), ('ushr', ('iabs', 'a'), ('find_lsb', ('iabs', 'b'))))) */
   /* search21_0 -> 0 in the cache */
   /* search21_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace21_0_0_0 -> 0 in the cache */
   /* replace21_0_0 -> 70 in the cache */
   /* replace21_0_1_0_0 -> 0 in the cache */
   /* replace21_0_1_0 -> 71 in the cache */
   /* replace21_0_1_1_0_0 -> 3 in the cache */
   /* replace21_0_1_1_0 -> 28 in the cache */
   /* replace21_0_1_1 -> 29 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 71, 29 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 70, 75 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 76 },
      -1,
   } },

   /* ('umod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1)) */
   /* search22_0 -> 0 in the cache */
   /* search22_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umod,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace22_0 -> 0 in the cache */
   /* replace22_1_0 -> 3 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3, 79 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 80 },
      -1,
   } },

   /* ('imod', 'a', '#b(is_pos_power_of_two)') => ('iand', 'a', ('isub', 'b', 1)) */
   /* search23_0 -> 0 in the cache */
   /* search23_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace23_0 -> 0 in the cache */
   /* replace23_1_0 -> 3 in the cache */
   /* replace23_1_1 -> 79 in the cache */
   /* replace23_1 -> 80 in the cache */
   /* replace23 -> 81 in the cache */

   /* ('imod', 'a', '#b(is_neg_power_of_two)') => ('bcsel', ('ieq', ('ior', 'a', 'b'), 'b'), 0, ('ior', 'a', 'b')) */
   /* search24_0 -> 0 in the cache */
   /* search24_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imod,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace24_0_0_0 -> 0 in the cache */
   /* replace24_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace24_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 84, 3 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   /* replace24_2_0 -> 0 in the cache */
   /* replace24_2_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 85, 86, 87 },
      -1,
   } },

   /* ('irem', 'a', '#b(is_pos_power_of_two)') => ('isub', 'a', ('iand', ('bcsel', ('ilt', 'a', 0), ('iadd', 'a', ('isub', 'b', 1)), 'a'), ('ineg', 'b'))) */
   /* search25_0 -> 0 in the cache */
   /* search25_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace25_0 -> 0 in the cache */
   /* replace25_1_0_0_0 -> 0 in the cache */
   /* replace25_1_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 86 },
      -1,
   } },
   /* replace25_1_0_1_0 -> 0 in the cache */
   /* replace25_1_0_1_1_0 -> 3 in the cache */
   /* replace25_1_0_1_1_1 -> 79 in the cache */
   /* replace25_1_0_1_1 -> 80 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 80 },
      -1,
   } },
   /* replace25_1_0_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 90, 91, 0 },
      -1,
   } },
   /* replace25_1_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 92, 93 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 0, 94 },
      -1,
   } },

   /* ('irem', 'a', '#b(is_neg_power_of_two)') => ('irem', 'a', ('iabs', 'b')) */
   /* search26_0 -> 0 in the cache */
   /* search26_1 -> 26 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 26 },
      -1,
   } },

   /* replace26_0 -> 0 in the cache */
   /* replace26_1_0 -> 3 in the cache */
   /* replace26_1 -> 28 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irem,
      -1, 0,
      { 0, 28 },
      -1,
   } },

   /* ('~fneg', ('fneg', 'a')) => a */
   /* search27_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace27 -> 32 in the cache */

   /* ('ineg', ('ineg', 'a')) => a */
   /* search28_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 100 },
      -1,
   } },

   /* replace28 -> 32 in the cache */

   /* ('fabs', ('fneg', 'a')) => ('fabs', 'a') */
   /* search29_0_0 -> 32 in the cache */
   /* search29_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace29_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('fabs', ('u2f', 'a')) => ('u2f', 'a') */
   /* search30_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 104 },
      -1,
   } },

   /* replace30_0 -> 32 in the cache */
   /* replace30 -> 104 in the cache */

   /* ('iabs', ('iabs', 'a')) => ('iabs', 'a') */
   /* search31_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 106 },
      -1,
   } },

   /* replace31_0 -> 32 in the cache */
   /* replace31 -> 106 in the cache */

   /* ('iabs', ('ineg', 'a')) => ('iabs', 'a') */
   /* search32_0_0 -> 32 in the cache */
   /* search32_0 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 100 },
      -1,
   } },

   /* replace32_0 -> 32 in the cache */
   /* replace32 -> 106 in the cache */

   /* ('f2b', ('fneg', 'a')) => ('f2b', 'a') */
   /* search33_0_0 -> 32 in the cache */
   /* search33_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2b,
      -1, 0,
      { 98 },
      -1,
   } },

   /* replace33_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2b,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2b', ('ineg', 'a')) => ('i2b', 'a') */
   /* search34_0_0 -> 32 in the cache */
   /* search34_0 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2b,
      -1, 0,
      { 100 },
      -1,
   } },

   /* replace34_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2b,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~fadd', 'a', 0.0) => a */
   /* search35_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 32, 113 },
      -1,
   } },

   /* replace35 -> 32 in the cache */

   /* ('fadd(is_only_used_as_float)', 'a@16', 0.0) => a */
   /* search36_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 115 },
      1,
   } },

   /* replace36 -> 11 in the cache */

   /* ('fadd(is_only_used_as_float)', 'a@32', 0.0) => a */
   /* search37_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 117 },
      1,
   } },

   /* replace37 -> 16 in the cache */

   /* ('iadd', 'a', 0) => a */
   /* search38_0 -> 32 in the cache */
   /* search38_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace38 -> 32 in the cache */

   /* ('iadd_sat', 'a', 0) => a */
   /* search39_0 -> 32 in the cache */
   /* search39_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace39 -> 32 in the cache */

   /* ('isub_sat', 'a', 0) => a */
   /* search40_0 -> 32 in the cache */
   /* search40_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace40 -> 32 in the cache */

   /* ('uadd_sat', 'a', 0) => a */
   /* search41_0 -> 32 in the cache */
   /* search41_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace41 -> 32 in the cache */

   /* ('usub_sat', 'a', 0) => a */
   /* search42_0 -> 32 in the cache */
   /* search42_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace42 -> 32 in the cache */

   /* ('usadd_4x8_vc4', 'a', 0) => a */
   /* search43_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usadd_4x8_vc4,
      0, 1,
      { 16, 124 },
      -1,
   } },

   /* replace43 -> 16 in the cache */

   /* ('usadd_4x8_vc4', 'a', -1) => -1 */
   /* search44_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_usadd_4x8_vc4,
      0, 1,
      { 16, 126 },
      -1,
   } },

   /* replace44 -> 126 in the cache */

   /* ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* search45_1_0 -> 128 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 130, 132 },
      -1,
   } },

   /* replace45_0 -> 128 in the cache */
   /* replace45_1_0 -> 129 in the cache */
   /* replace45_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 128, 134 },
      -1,
   } },

   /* ('~fadd', ('fmulz', 'a', 'b'), ('fmulz', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c')) */
   /* search46_0_0 -> 16 in the cache */
   /* search46_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* search46_1_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 16, 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 136, 138 },
      -1,
   } },

   /* replace46_0 -> 16 in the cache */
   /* replace46_1_0 -> 36 in the cache */
   /* replace46_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 16, 140 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search47_0 -> 0 in the cache */
   /* search47_1 -> 3 in the cache */
   /* search47_2_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 0, 142, 143 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 3, 144 },
      -1,
   } },

   /* replace47_0 -> 0 in the cache */
   /* replace47_1_0 -> 3 in the cache */
   /* replace47_1_1 -> 142 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3, 142 },
      -1,
   } },
   /* replace47_2 -> 143 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 146, 143 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', ('fmul(is_used_once)', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   /* search48_0 -> 0 in the cache */
   /* search48_1 -> 3 in the cache */
   /* search48_2_0 -> 0 in the cache */
   /* search48_2_1 -> 142 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 0, 142 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 0, 3, 148 },
      -1,
   } },

   /* replace48_0 -> 0 in the cache */
   /* replace48_1_0 -> 3 in the cache */
   /* replace48_1_1 -> 142 in the cache */
   /* replace48_1 -> 146 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 0, 146 },
      -1,
   } },

   /* ('~fadd', ('fmul(is_used_once)', 'a', 'b'), ('ffma(is_used_once)', 'a', 'c', 'd')) => ('ffma', 'a', ('fadd', 'b', 'c'), 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 151, 152 },
      0,
   } },
   /* search49_1_0 -> 151 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 151, 154, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 153, 156 },
      -1,
   } },

   /* replace49_0 -> 151 in the cache */
   /* replace49_1_0 -> 152 in the cache */
   /* replace49_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 152, 154 },
      -1,
   } },
   /* replace49_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 151, 158, 155 },
      -1,
   } },

   /* ('~ffma', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('ffma', 'a', 'c', 'd')) */
   /* search50_0 -> 151 in the cache */
   /* search50_1_0 -> 152 in the cache */
   /* search50_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 152, 154 },
      0,
   } },
   /* search50_2_0 -> 152 in the cache */
   /* search50_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 151, 160, 161 },
      -1,
   } },

   /* replace50_0 -> 152 in the cache */
   /* replace50_1_0 -> 151 in the cache */
   /* replace50_1_1 -> 154 in the cache */
   /* replace50_1_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 151, 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 152, 163 },
      -1,
   } },

   /* ('~ffmaz', 'a', 'b', ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search51_0 -> 16 in the cache */
   /* search51_1 -> 36 in the cache */
   /* search51_2_0 -> 16 in the cache */
   /* search51_2_1 -> 137 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 137, 165 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 36, 166 },
      -1,
   } },

   /* replace51_0 -> 16 in the cache */
   /* replace51_1_0 -> 36 in the cache */
   /* replace51_1_1 -> 137 in the cache */
   /* replace51_1 -> 140 in the cache */
   /* replace51_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 140, 165 },
      -1,
   } },

   /* ('~ffmaz', 'a', 'b', ('fmulz(is_used_once)', 'a', 'c')) => ('fmulz', 'a', ('fadd', 'b', 'c')) */
   /* search52_0 -> 16 in the cache */
   /* search52_1 -> 36 in the cache */
   /* search52_2_0 -> 16 in the cache */
   /* search52_2_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 137 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 16, 36, 169 },
      -1,
   } },

   /* replace52_0 -> 16 in the cache */
   /* replace52_1_0 -> 36 in the cache */
   /* replace52_1_1 -> 137 in the cache */
   /* replace52_1 -> 140 in the cache */
   /* replace52 -> 141 in the cache */

   /* ('~fadd', ('fmulz(is_used_once)', 'a', 'b'), ('ffmaz(is_used_once)', 'a', 'c', 'd')) => ('ffmaz', 'a', ('fadd', 'b', 'c'), 'd') */
   /* search53_0_0 -> 16 in the cache */
   /* search53_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 36 },
      0,
   } },
   /* search53_1_0 -> 16 in the cache */
   /* search53_1_1 -> 137 in the cache */
   /* search53_1_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 16, 137, 165 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 171, 172 },
      -1,
   } },

   /* replace53_0 -> 16 in the cache */
   /* replace53_1_0 -> 36 in the cache */
   /* replace53_1_1 -> 137 in the cache */
   /* replace53_1 -> 140 in the cache */
   /* replace53_2 -> 165 in the cache */
   /* replace53 -> 168 in the cache */

   /* ('~ffmaz', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('ffmaz', 'a', 'c', 'd')) */
   /* search54_0 -> 16 in the cache */
   /* search54_1_0 -> 36 in the cache */
   /* search54_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 137 },
      0,
   } },
   /* search54_2_0 -> 36 in the cache */
   /* search54_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 36, 165 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 16, 174, 175 },
      -1,
   } },

   /* replace54_0 -> 36 in the cache */
   /* replace54_1_0 -> 16 in the cache */
   /* replace54_1_1 -> 137 in the cache */
   /* replace54_1_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 137, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 36, 177 },
      -1,
   } },

   /* ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c')) */
   /* search55_0_0 -> 128 in the cache */
   /* search55_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* search55_1_0 -> 128 in the cache */
   /* search55_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 179, 180 },
      -1,
   } },

   /* replace55_0 -> 128 in the cache */
   /* replace55_1_0 -> 129 in the cache */
   /* replace55_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 128, 182 },
      -1,
   } },

   /* ('iand', ('ior', 'a', 'b'), ('ior', 'a', 'c')) => ('ior', 'a', ('iand', 'b', 'c')) */
   /* search56_0_0 -> 128 in the cache */
   /* search56_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* search56_1_0 -> 128 in the cache */
   /* search56_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 184, 185 },
      -1,
   } },

   /* replace56_0 -> 128 in the cache */
   /* replace56_1_0 -> 129 in the cache */
   /* replace56_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 128, 187 },
      -1,
   } },

   /* ('ior', ('iand', 'a', 'b'), ('iand', 'a', 'c')) => ('iand', 'a', ('ior', 'b', 'c')) */
   /* search57_0_0 -> 128 in the cache */
   /* search57_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* search57_1_0 -> 128 in the cache */
   /* search57_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 189, 190 },
      -1,
   } },

   /* replace57_0 -> 128 in the cache */
   /* replace57_1_0 -> 129 in the cache */
   /* replace57_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 128, 192 },
      -1,
   } },

   /* ('~fadd', ('fneg', 'a'), 'a') => 0.0 */
   /* search58_0_0 -> 32 in the cache */
   /* search58_0 -> 98 in the cache */
   /* search58_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 98, 32 },
      -1,
   } },

   /* replace58 -> 113 in the cache */

   /* ('iadd', ('ineg', 'a'), 'a') => 0 */
   /* search59_0_0 -> 32 in the cache */
   /* search59_0 -> 100 in the cache */
   /* search59_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 100, 32 },
      -1,
   } },

   /* replace59 -> 61 in the cache */

   /* ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b */
   /* search60_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search60_1_0 -> 0 in the cache */
   /* search60_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 196, 197 },
      -1,
   } },

   /* replace60 -> 3 in the cache */

   /* ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b */
   /* search61_0 -> 0 in the cache */
   /* search61_1_0_0 -> 0 in the cache */
   /* search61_1_0 -> 196 in the cache */
   /* search61_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 196, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 0, 199 },
      -1,
   } },

   /* replace61 -> 3 in the cache */

   /* ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b */
   /* search62_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search62_1_0 -> 0 in the cache */
   /* search62_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 201, 202 },
      -1,
   } },

   /* replace62 -> 3 in the cache */

   /* ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b */
   /* search63_0 -> 0 in the cache */
   /* search63_1_0_0 -> 0 in the cache */
   /* search63_1_0 -> 201 in the cache */
   /* search63_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 201, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 0, 204 },
      -1,
   } },

   /* replace63 -> 3 in the cache */

   /* ('fadd', ('fsat', 'a'), ('fsat', ('fneg', 'a'))) => ('fsat', ('fabs', 'a')) */
   /* search64_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search64_1_0_0 -> 32 in the cache */
   /* search64_1_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 98 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 206, 207 },
      -1,
   } },

   /* replace64_0_0 -> 32 in the cache */
   /* replace64_0 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 103 },
      -1,
   } },

   /* ('~fmul', 'a', 0.0) => 0.0 */
   /* search65_0 -> 32 in the cache */
   /* search65_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 113 },
      -1,
   } },

   /* replace65 -> 113 in the cache */

   /* ('fmul', 'a@16', 0.0) => 0.0 */
   /* search66_0 -> 11 in the cache */
   /* search66_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 115 },
      -1,
   } },

   /* replace66 -> 115 in the cache */

   /* ('fmul', 'a@32', 0.0) => 0.0 */
   /* search67_0 -> 16 in the cache */
   /* search67_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 117 },
      -1,
   } },

   /* replace67 -> 117 in the cache */

   /* ('fmulz', 'a', 0.0) => 0.0 */
   /* search68_0 -> 16 in the cache */
   /* search68_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 117 },
      -1,
   } },

   /* replace68 -> 117 in the cache */

   /* ('fmulz', 'a', 'b(is_finite_not_zero)') => ('fmul', 'a', 'b') */
   /* search69_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 214 },
      -1,
   } },

   /* replace69_0 -> 16 in the cache */
   /* replace69_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('fmulz', 'a(is_finite)', 'b(is_finite)') => ('fmul', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 217, 218 },
      -1,
   } },

   /* replace70_0 -> 16 in the cache */
   /* replace70_1 -> 36 in the cache */
   /* replace70 -> 216 in the cache */

   /* ('fmulz', 'a', 'a') => ('fmul', 'a', 'a') */
   /* search71_0 -> 16 in the cache */
   /* search71_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      -1, 0,
      { 16, 16 },
      -1,
   } },

   /* replace71_0 -> 16 in the cache */
   /* replace71_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 16, 16 },
      -1,
   } },

   /* ('ffmaz', 'a', 'b(is_finite_not_zero)', 'c') => ('ffma', 'a', 'b', 'c') */
   /* search72_0 -> 16 in the cache */
   /* search72_1 -> 214 in the cache */
   /* search72_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 214, 137 },
      -1,
   } },

   /* replace72_0 -> 16 in the cache */
   /* replace72_1 -> 36 in the cache */
   /* replace72_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* ('ffmaz', 'a(is_finite)', 'b(is_finite)', 'c') => ('ffma', 'a', 'b', 'c') */
   /* search73_0 -> 217 in the cache */
   /* search73_1 -> 218 in the cache */
   /* search73_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 217, 218, 137 },
      -1,
   } },

   /* replace73_0 -> 16 in the cache */
   /* replace73_1 -> 36 in the cache */
   /* replace73_2 -> 137 in the cache */
   /* replace73 -> 223 in the cache */

   /* ('ffmaz', 'a', 'a', 'b') => ('ffma', 'a', 'a', 'b') */
   /* search74_0 -> 16 in the cache */
   /* search74_1 -> 16 in the cache */
   /* search74_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      -1, 0,
      { 16, 16, 36 },
      -1,
   } },

   /* replace74_0 -> 16 in the cache */
   /* replace74_1 -> 16 in the cache */
   /* replace74_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      -1, 0,
      { 16, 16, 36 },
      -1,
   } },

   /* ('imul', 'a', 0) => 0 */
   /* search75_0 -> 32 in the cache */
   /* search75_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace75 -> 61 in the cache */

   /* ('umul_unorm_4x8_vc4', 'a', 0) => 0 */
   /* search76_0 -> 16 in the cache */
   /* search76_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_unorm_4x8_vc4,
      0, 1,
      { 16, 124 },
      -1,
   } },

   /* replace76 -> 124 in the cache */

   /* ('umul_unorm_4x8_vc4', 'a', -1) => a */
   /* search77_0 -> 16 in the cache */
   /* search77_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_unorm_4x8_vc4,
      0, 1,
      { 16, 126 },
      -1,
   } },

   /* replace77 -> 16 in the cache */

   /* ('~fmul', 'a', 1.0) => a */
   /* search78_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 230 },
      -1,
   } },

   /* replace78 -> 32 in the cache */

   /* ('~fmulz', 'a', 1.0) => a */
   /* search79_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 232 },
      -1,
   } },

   /* replace79 -> 16 in the cache */

   /* ('fmul(is_only_used_as_float)', 'a', 1.0) => a */
   /* search80_0 -> 32 in the cache */
   /* search80_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 230 },
      1,
   } },

   /* replace80 -> 32 in the cache */

   /* ('imul', 'a', 1) => a */
   /* search81_0 -> 32 in the cache */
   /* search81_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 35 },
      -1,
   } },

   /* replace81 -> 32 in the cache */

   /* ('fmul', 'a', -1.0) => ('fneg', 'a') */
   /* search82_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 32, 236 },
      -1,
   } },

   /* replace82_0 -> 32 in the cache */
   /* replace82 -> 98 in the cache */

   /* ('imul', 'a', -1) => ('ineg', 'a') */
   /* search83_0 -> 32 in the cache */
   /* search83_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace83_0 -> 32 in the cache */
   /* replace83 -> 100 in the cache */

   /* ('fmul', ('fsign', 'a'), ('fmul', 'a', 'a')) => ('fmul', ('fabs', 'a'), 'a') */
   /* search84_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search84_1_0 -> 32 in the cache */
   /* search84_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 32, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 239, 240 },
      -1,
   } },

   /* replace84_0_0 -> 32 in the cache */
   /* replace84_0 -> 103 in the cache */
   /* replace84_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 103, 32 },
      -1,
   } },

   /* ('fmul', ('fmul', ('fsign', 'a'), 'a'), 'a') => ('fmul', ('fabs', 'a'), 'a') */
   /* search85_0_0_0 -> 32 in the cache */
   /* search85_0_0 -> 239 in the cache */
   /* search85_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 239, 32 },
      -1,
   } },
   /* search85_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 243, 32 },
      -1,
   } },

   /* replace85_0_0 -> 32 in the cache */
   /* replace85_0 -> 103 in the cache */
   /* replace85_1 -> 32 in the cache */
   /* replace85 -> 242 in the cache */

   /* ('~ffma', 0.0, 'a', 'b') => b */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search86_1 -> 0 in the cache */
   /* search86_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 245, 0, 3 },
      -1,
   } },

   /* replace86 -> 3 in the cache */

   /* ('ffma@16(is_only_used_as_float)', 0.0, 'a', 'b') => b */
   /* search87_0 -> 115 in the cache */
   /* search87_1 -> 11 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 115, 11, 247 },
      1,
   } },

   /* replace87 -> 247 in the cache */

   /* ('ffma@32(is_only_used_as_float)', 0.0, 'a', 'b') => b */
   /* search88_0 -> 117 in the cache */
   /* search88_1 -> 16 in the cache */
   /* search88_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 117, 16, 36 },
      1,
   } },

   /* replace88 -> 36 in the cache */

   /* ('ffmaz', 0.0, 'a', 'b') => ('fadd', 0.0, 'b') */
   /* search89_0 -> 117 in the cache */
   /* search89_1 -> 16 in the cache */
   /* search89_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 117, 16, 36 },
      -1,
   } },

   /* replace89_0 -> 117 in the cache */
   /* replace89_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 117, 36 },
      -1,
   } },

   /* ('~ffma', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search90_0 -> 0 in the cache */
   /* search90_1 -> 3 in the cache */
   /* search90_2 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 0, 3, 245 },
      -1,
   } },

   /* replace90_0 -> 0 in the cache */
   /* replace90_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ffma@16', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search91_0 -> 11 in the cache */
   /* search91_1 -> 247 in the cache */
   /* search91_2 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 247, 115 },
      -1,
   } },

   /* replace91_0 -> 11 in the cache */
   /* replace91_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 247 },
      -1,
   } },

   /* ('ffma@32', 'a', 'b', 0.0) => ('fmul', 'a', 'b') */
   /* search92_0 -> 16 in the cache */
   /* search92_1 -> 36 in the cache */
   /* search92_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 117 },
      -1,
   } },

   /* replace92_0 -> 16 in the cache */
   /* replace92_1 -> 36 in the cache */
   /* replace92 -> 216 in the cache */

   /* ('ffmaz', 'a', 'b', 0.0) => ('fmulz', 'a', 'b') */
   /* search93_0 -> 16 in the cache */
   /* search93_1 -> 36 in the cache */
   /* search93_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 117 },
      -1,
   } },

   /* replace93_0 -> 16 in the cache */
   /* replace93_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('ffma', 1.0, 'a', 'b') => ('fadd', 'a', 'b') */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* search94_1 -> 0 in the cache */
   /* search94_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 259, 0, 3 },
      -1,
   } },

   /* replace94_0 -> 0 in the cache */
   /* replace94_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ffmaz', 1.0, 'a', 'b') => ('fadd', 'a', 'b') */
   /* search95_0 -> 232 in the cache */
   /* search95_1 -> 16 in the cache */
   /* search95_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 232, 16, 36 },
      -1,
   } },

   /* replace95_0 -> 16 in the cache */
   /* replace95_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('ffma', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b') */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   /* search96_1 -> 0 in the cache */
   /* search96_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 264, 0, 3 },
      -1,
   } },

   /* replace96_0_0 -> 0 in the cache */
   /* replace96_0 -> 201 in the cache */
   /* replace96_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 201, 3 },
      -1,
   } },

   /* ('ffmaz', -1.0, 'a', 'b') => ('fadd', ('fneg', 'a'), 'b') */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   /* search97_1 -> 16 in the cache */
   /* search97_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 267, 16, 36 },
      -1,
   } },

   /* replace97_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 16 },
      -1,
   } },
   /* replace97_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 269, 36 },
      -1,
   } },

   /* ('~ffma', '#a', '#b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search98_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 271, 272, 131 },
      -1,
   } },

   /* replace98_0_0 -> 128 in the cache */
   /* replace98_0_1 -> 129 in the cache */
   /* replace98_0 -> 130 in the cache */
   /* replace98_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 130, 131 },
      -1,
   } },

   /* ('~ffmaz', '#a', '#b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search99_1 -> 33 in the cache */
   /* search99_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 275, 33, 137 },
      -1,
   } },

   /* replace99_0_0 -> 16 in the cache */
   /* replace99_0_1 -> 36 in the cache */
   /* replace99_0 -> 136 in the cache */
   /* replace99_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 136, 137 },
      -1,
   } },

   /* ('~flrp', 'a', 'b', 0.0) => a */
   /* search100_0 -> 0 in the cache */
   /* search100_1 -> 3 in the cache */
   /* search100_2 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 3, 245 },
      -1,
   } },

   /* replace100 -> 0 in the cache */

   /* ('~flrp', 'a', 'b', 1.0) => b */
   /* search101_0 -> 0 in the cache */
   /* search101_1 -> 3 in the cache */
   /* search101_2 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 3, 259 },
      -1,
   } },

   /* replace101 -> 3 in the cache */

   /* ('~flrp', 'a', 'a', 'b') => a */
   /* search102_0 -> 0 in the cache */
   /* search102_1 -> 0 in the cache */
   /* search102_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 0, 3 },
      -1,
   } },

   /* replace102 -> 0 in the cache */

   /* ('~flrp', 0.0, 'a', 'b') => ('fmul', 'a', 'b') */
   /* search103_0 -> 245 in the cache */
   /* search103_1 -> 0 in the cache */
   /* search103_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 245, 0, 3 },
      -1,
   } },

   /* replace103_0 -> 0 in the cache */
   /* replace103_1 -> 3 in the cache */
   /* replace103 -> 253 in the cache */

   /* ('~flrp', 'a', ('fadd(is_used_once)', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search104_0 -> 128 in the cache */
   /* search104_1_0 -> 128 in the cache */
   /* search104_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 128, 129 },
      0,
   } },
   /* search104_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 128, 282, 131 },
      -1,
   } },

   /* replace104_0_0 -> 129 in the cache */
   /* replace104_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 129, 131 },
      -1,
   } },
   /* replace104_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 284, 128 },
      -1,
   } },

   /* ('sdot_4x8_iadd', 'a', 0, 'b') => b */
   /* search105_0 -> 16 in the cache */
   /* search105_1 -> 124 in the cache */
   /* search105_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace105 -> 36 in the cache */

   /* ('udot_4x8_uadd', 'a', 0, 'b') => b */
   /* search106_0 -> 16 in the cache */
   /* search106_1 -> 124 in the cache */
   /* search106_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace106 -> 36 in the cache */

   /* ('sdot_4x8_iadd_sat', 'a', 0, 'b') => b */
   /* search107_0 -> 16 in the cache */
   /* search107_1 -> 124 in the cache */
   /* search107_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace107 -> 36 in the cache */

   /* ('udot_4x8_uadd_sat', 'a', 0, 'b') => b */
   /* search108_0 -> 16 in the cache */
   /* search108_1 -> 124 in the cache */
   /* search108_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace108 -> 36 in the cache */

   /* ('sdot_2x16_iadd', 'a', 0, 'b') => b */
   /* search109_0 -> 16 in the cache */
   /* search109_1 -> 124 in the cache */
   /* search109_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace109 -> 36 in the cache */

   /* ('udot_2x16_uadd', 'a', 0, 'b') => b */
   /* search110_0 -> 16 in the cache */
   /* search110_1 -> 124 in the cache */
   /* search110_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace110 -> 36 in the cache */

   /* ('sdot_2x16_iadd_sat', 'a', 0, 'b') => b */
   /* search111_0 -> 16 in the cache */
   /* search111_1 -> 124 in the cache */
   /* search111_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace111 -> 36 in the cache */

   /* ('udot_2x16_uadd_sat', 'a', 0, 'b') => b */
   /* search112_0 -> 16 in the cache */
   /* search112_1 -> 124 in the cache */
   /* search112_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 16, 124, 36 },
      -1,
   } },

   /* replace112 -> 36 in the cache */

   /* ('sudot_4x8_iadd', 'a', 0, 'b') => b */
   /* search113_0 -> 16 in the cache */
   /* search113_1 -> 124 in the cache */
   /* search113_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 124, 36 },
      -1,
   } },

   /* replace113 -> 36 in the cache */

   /* ('sudot_4x8_iadd', 0, 'a', 'b') => b */
   /* search114_0 -> 124 in the cache */
   /* search114_1 -> 16 in the cache */
   /* search114_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 124, 16, 36 },
      -1,
   } },

   /* replace114 -> 36 in the cache */

   /* ('sudot_4x8_iadd_sat', 'a', 0, 'b') => b */
   /* search115_0 -> 16 in the cache */
   /* search115_1 -> 124 in the cache */
   /* search115_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 16, 124, 36 },
      -1,
   } },

   /* replace115 -> 36 in the cache */

   /* ('sudot_4x8_iadd_sat', 0, 'a', 'b') => b */
   /* search116_0 -> 124 in the cache */
   /* search116_1 -> 16 in the cache */
   /* search116_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 124, 16, 36 },
      -1,
   } },

   /* replace116 -> 36 in the cache */

   /* ('iadd', ('sdot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search117_0_0 -> 16 in the cache */
   /* search117_0_1 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 16, 36, 298 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 299, 300 },
      -1,
   } },

   /* replace117_0 -> 16 in the cache */
   /* replace117_1 -> 36 in the cache */
   /* replace117_2_0 -> 137 in the cache */
   /* replace117_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 137, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 2,
      { 16, 36, 302 },
      -1,
   } },

   /* ('iadd', ('udot_4x8_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_4x8_uadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search118_0_0 -> 16 in the cache */
   /* search118_0_1 -> 36 in the cache */
   /* search118_0_2 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 16, 36, 298 },
      0,
   } },
   /* search118_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 304, 300 },
      -1,
   } },

   /* replace118_0 -> 16 in the cache */
   /* replace118_1 -> 36 in the cache */
   /* replace118_2_0 -> 137 in the cache */
   /* replace118_2_1 -> 165 in the cache */
   /* replace118_2 -> 302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 2,
      { 16, 36, 302 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sudot_4x8_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search119_0_0 -> 16 in the cache */
   /* search119_0_1 -> 36 in the cache */
   /* search119_0_2 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 298 },
      0,
   } },
   /* search119_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 307, 300 },
      -1,
   } },

   /* replace119_0 -> 16 in the cache */
   /* replace119_1 -> 36 in the cache */
   /* replace119_2_0 -> 137 in the cache */
   /* replace119_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 137, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 1,
      { 16, 36, 309 },
      -1,
   } },

   /* ('iadd', ('sdot_2x16_iadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('sdot_2x16_iadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search120_0_0 -> 16 in the cache */
   /* search120_0_1 -> 36 in the cache */
   /* search120_0_2 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 16, 36, 298 },
      0,
   } },
   /* search120_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 311, 300 },
      -1,
   } },

   /* replace120_0 -> 16 in the cache */
   /* replace120_1 -> 36 in the cache */
   /* replace120_2_0 -> 137 in the cache */
   /* replace120_2_1 -> 165 in the cache */
   /* replace120_2 -> 302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 2,
      { 16, 36, 302 },
      -1,
   } },

   /* ('iadd', ('udot_2x16_uadd(is_used_once)', 'a', 'b', '#c'), '#d') => ('udot_2x16_uadd', 'a', 'b', ('iadd', 'c', 'd')) */
   /* search121_0_0 -> 16 in the cache */
   /* search121_0_1 -> 36 in the cache */
   /* search121_0_2 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 16, 36, 298 },
      0,
   } },
   /* search121_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 314, 300 },
      -1,
   } },

   /* replace121_0 -> 16 in the cache */
   /* replace121_1 -> 36 in the cache */
   /* replace121_2_0 -> 137 in the cache */
   /* replace121_2_1 -> 165 in the cache */
   /* replace121_2 -> 302 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 2,
      { 16, 36, 302 },
      -1,
   } },

   /* ('iadd', ('sdot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_4x8_iadd', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search122_0_1 -> 36 in the cache */
   /* search122_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 317, 36, 124 },
      -1,
   } },
   /* search122_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 318, 137 },
      -1,
   } },

   /* replace122_0 -> 16 in the cache */
   /* replace122_1 -> 36 in the cache */
   /* replace122_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* ('iadd', ('udot_4x8_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_4x8_uadd', 'a', 'b', 'c') */
   /* search123_0_0 -> 317 in the cache */
   /* search123_0_1 -> 36 in the cache */
   /* search123_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 317, 36, 124 },
      -1,
   } },
   /* search123_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 321, 137 },
      -1,
   } },

   /* replace123_0 -> 16 in the cache */
   /* replace123_1 -> 36 in the cache */
   /* replace123_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c') */
   /* search124_0_0 -> 317 in the cache */
   /* search124_0_1 -> 36 in the cache */
   /* search124_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 317, 36, 124 },
      -1,
   } },
   /* search124_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 324, 137 },
      -1,
   } },

   /* replace124_0 -> 16 in the cache */
   /* replace124_1 -> 36 in the cache */
   /* replace124_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },

   /* ('iadd', ('sudot_4x8_iadd', 'a', 'b(is_not_const)', 0), 'c') => ('sudot_4x8_iadd', 'a', 'b', 'c') */
   /* search125_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search125_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 327, 124 },
      -1,
   } },
   /* search125_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 328, 137 },
      -1,
   } },

   /* replace125_0 -> 16 in the cache */
   /* replace125_1 -> 36 in the cache */
   /* replace125_2 -> 137 in the cache */
   /* replace125 -> 326 in the cache */

   /* ('iadd', ('sdot_2x16_iadd', 'a(is_not_const)', 'b', 0), 'c') => ('sdot_2x16_iadd', 'a', 'b', 'c') */
   /* search126_0_0 -> 317 in the cache */
   /* search126_0_1 -> 36 in the cache */
   /* search126_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 317, 36, 124 },
      -1,
   } },
   /* search126_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 330, 137 },
      -1,
   } },

   /* replace126_0 -> 16 in the cache */
   /* replace126_1 -> 36 in the cache */
   /* replace126_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* ('iadd', ('udot_2x16_uadd', 'a(is_not_const)', 'b', 0), 'c') => ('udot_2x16_uadd', 'a', 'b', 'c') */
   /* search127_0_0 -> 317 in the cache */
   /* search127_0_1 -> 36 in the cache */
   /* search127_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 317, 36, 124 },
      -1,
   } },
   /* search127_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 333, 137 },
      -1,
   } },

   /* replace127_0 -> 16 in the cache */
   /* replace127_1 -> 36 in the cache */
   /* replace127_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* ('sdot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search128_0 -> 275 in the cache */
   /* search128_1 -> 33 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace128_0_0 -> 16 in the cache */
   /* replace128_0_1 -> 36 in the cache */
   /* replace128_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd,
      1, 1,
      { 16, 36, 124 },
      -1,
   } },
   /* replace128_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 338, 137 },
      -1,
   } },

   /* ('udot_4x8_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_4x8_uadd', 'a', 'b', 0), 'c') */
   /* search129_0 -> 275 in the cache */
   /* search129_1 -> 33 in the cache */
   /* search129_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace129_0_0 -> 16 in the cache */
   /* replace129_0_1 -> 36 in the cache */
   /* replace129_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd,
      1, 1,
      { 16, 36, 124 },
      -1,
   } },
   /* replace129_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 341, 137 },
      -1,
   } },

   /* ('sudot_4x8_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sudot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search130_0 -> 275 in the cache */
   /* search130_1 -> 33 in the cache */
   /* search130_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 275, 33, 336 },
      -1,
   } },

   /* replace130_0_0 -> 16 in the cache */
   /* replace130_0_1 -> 36 in the cache */
   /* replace130_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd,
      -1, 0,
      { 16, 36, 124 },
      -1,
   } },
   /* replace130_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 344, 137 },
      -1,
   } },

   /* ('sdot_2x16_iadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('sdot_2x16_iadd', 'a', 'b', 0), 'c') */
   /* search131_0 -> 275 in the cache */
   /* search131_1 -> 33 in the cache */
   /* search131_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace131_0_0 -> 16 in the cache */
   /* replace131_0_1 -> 36 in the cache */
   /* replace131_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd,
      1, 1,
      { 16, 36, 124 },
      -1,
   } },
   /* replace131_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 347, 137 },
      -1,
   } },

   /* ('udot_2x16_uadd', '#a', '#b', 'c(is_not_const)') => ('iadd', ('udot_2x16_uadd', 'a', 'b', 0), 'c') */
   /* search132_0 -> 275 in the cache */
   /* search132_1 -> 33 in the cache */
   /* search132_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace132_0_0 -> 16 in the cache */
   /* replace132_0_1 -> 36 in the cache */
   /* replace132_0_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd,
      1, 1,
      { 16, 36, 124 },
      -1,
   } },
   /* replace132_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 350, 137 },
      -1,
   } },

   /* ('sdot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search133_0 -> 275 in the cache */
   /* search133_1 -> 33 in the cache */
   /* search133_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace133_0_0 -> 16 in the cache */
   /* replace133_0_1 -> 36 in the cache */
   /* replace133_0_2 -> 124 in the cache */
   /* replace133_0 -> 338 in the cache */
   /* replace133_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 2,
      { 338, 137 },
      -1,
   } },

   /* ('udot_4x8_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_4x8_uadd', 'a', 'b', 0), 'c') */
   /* search134_0 -> 275 in the cache */
   /* search134_1 -> 33 in the cache */
   /* search134_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace134_0_0 -> 16 in the cache */
   /* replace134_0_1 -> 36 in the cache */
   /* replace134_0_2 -> 124 in the cache */
   /* replace134_0 -> 341 in the cache */
   /* replace134_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 2,
      { 341, 137 },
      -1,
   } },

   /* ('sudot_4x8_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sudot_4x8_iadd', 'a', 'b', 0), 'c') */
   /* search135_0 -> 275 in the cache */
   /* search135_1 -> 33 in the cache */
   /* search135_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 275, 33, 336 },
      -1,
   } },

   /* replace135_0_0 -> 16 in the cache */
   /* replace135_0_1 -> 36 in the cache */
   /* replace135_0_2 -> 124 in the cache */
   /* replace135_0 -> 344 in the cache */
   /* replace135_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 344, 137 },
      -1,
   } },

   /* ('sdot_2x16_iadd_sat', '#a', '#b', 'c(is_not_const)') => ('iadd_sat', ('sdot_2x16_iadd', 'a', 'b', 0), 'c') */
   /* search136_0 -> 275 in the cache */
   /* search136_1 -> 33 in the cache */
   /* search136_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace136_0_0 -> 16 in the cache */
   /* replace136_0_1 -> 36 in the cache */
   /* replace136_0_2 -> 124 in the cache */
   /* replace136_0 -> 347 in the cache */
   /* replace136_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 2,
      { 347, 137 },
      -1,
   } },

   /* ('udot_2x16_uadd_sat', '#a', '#b', 'c(is_not_const)') => ('uadd_sat', ('udot_2x16_uadd', 'a', 'b', 0), 'c') */
   /* search137_0 -> 275 in the cache */
   /* search137_1 -> 33 in the cache */
   /* search137_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 275, 33, 336 },
      -1,
   } },

   /* replace137_0_0 -> 16 in the cache */
   /* replace137_0_1 -> 36 in the cache */
   /* replace137_0_2 -> 124 in the cache */
   /* replace137_0 -> 350 in the cache */
   /* replace137_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 2,
      { 350, 137 },
      -1,
   } },

   /* ('fmul@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b')) => ('fmulz', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search138_0_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      1, 1,
      { 362, 117 },
      -1,
   } },
   /* search138_0_1 -> 117 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 363, 117, 364 },
      -1,
   } },
   /* search138_1_0_0 -> 364 in the cache */
   /* search138_1_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      2, 1,
      { 364, 117 },
      -1,
   } },
   /* search138_1_1 -> 117 in the cache */
   /* search138_1_2 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 366, 117, 362 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 365, 367 },
      -1,
   } },

   /* replace138_0 -> 364 in the cache */
   /* replace138_1 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 364, 362 },
      -1,
   } },

   /* ('fmul@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)')) => ('fmulz', 'a', 'b') */
   /* search139_0 -> 16 in the cache */
   /* search139_1_0_0 -> 16 in the cache */
   /* search139_1_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      true,
      nir_op_feq,
      1, 1,
      { 16, 117 },
      -1,
   } },
   /* search139_1_1 -> 117 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      6,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 370, 117, 371 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 16, 372 },
      -1,
   } },

   /* replace139_0 -> 16 in the cache */
   /* replace139_1 -> 36 in the cache */
   /* replace139 -> 258 in the cache */

   /* ('ffma@32', ('bcsel', ('feq', 'b', 0.0, 'ignore_exact'), 0.0, 'a'), ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search140_0_0_0 -> 362 in the cache */
   /* search140_0_0_1 -> 117 in the cache */
   /* search140_0_0 -> 363 in the cache */
   /* search140_0_1 -> 117 in the cache */
   /* search140_0_2 -> 364 in the cache */
   /* search140_0 -> 365 in the cache */
   /* search140_1_0_0 -> 364 in the cache */
   /* search140_1_0_1 -> 117 in the cache */
   /* search140_1_0 -> 366 in the cache */
   /* search140_1_1 -> 117 in the cache */
   /* search140_1_2 -> 362 in the cache */
   /* search140_1 -> 367 in the cache */
   /* search140_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 365, 367, 137 },
      -1,
   } },

   /* replace140_0 -> 364 in the cache */
   /* replace140_1 -> 362 in the cache */
   /* replace140_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 364, 362, 137 },
      -1,
   } },

   /* ('ffma@32', 'a', ('bcsel', ('feq', 'a', 0.0, 'ignore_exact'), 0.0, '#b(is_not_const_zero)'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search141_0 -> 16 in the cache */
   /* search141_1_0_0 -> 16 in the cache */
   /* search141_1_0_1 -> 117 in the cache */
   /* search141_1_0 -> 370 in the cache */
   /* search141_1_1 -> 117 in the cache */
   /* search141_1_2 -> 371 in the cache */
   /* search141_1 -> 372 in the cache */
   /* search141_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 16, 372, 137 },
      -1,
   } },

   /* replace141_0 -> 16 in the cache */
   /* replace141_1 -> 36 in the cache */
   /* replace141_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* ('sdot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c') */
   /* search142_0 -> 16 in the cache */
   /* search142_1 -> 36 in the cache */
   /* search142_2 -> 137 in the cache */
   /* search142 -> 320 in the cache */

   /* replace142_0_0_0_0_0 -> 16 in the cache */
   /* replace142_0_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* replace142_0_0_0_1_0 -> 36 in the cache */
   /* replace142_0_0_0_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 378, 379 },
      -1,
   } },
   /* replace142_0_0_1_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 381 },
      -1,
   } },
   /* replace142_0_0_1_1_0 -> 36 in the cache */
   /* replace142_0_0_1_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 382, 383 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 380, 384 },
      -1,
   } },
   /* replace142_0_1_0_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 386 },
      -1,
   } },
   /* replace142_0_1_0_1_0 -> 36 in the cache */
   /* replace142_0_1_0_1_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 387, 388 },
      -1,
   } },
   /* replace142_0_1_1_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 390 },
      -1,
   } },
   /* replace142_0_1_1_1_0 -> 36 in the cache */
   /* replace142_0_1_1_1_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 36, 390 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 391, 392 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 389, 393 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 385, 394 },
      -1,
   } },
   /* replace142_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 395, 137 },
      -1,
   } },

   /* ('udot_4x8_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search143_0 -> 16 in the cache */
   /* search143_1 -> 36 in the cache */
   /* search143_2 -> 137 in the cache */
   /* search143 -> 323 in the cache */

   /* replace143_0_0_0_0_0 -> 16 in the cache */
   /* replace143_0_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* replace143_0_0_0_1_0 -> 36 in the cache */
   /* replace143_0_0_0_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 397, 398 },
      -1,
   } },
   /* replace143_0_0_1_0_0 -> 16 in the cache */
   /* replace143_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 381 },
      -1,
   } },
   /* replace143_0_0_1_1_0 -> 36 in the cache */
   /* replace143_0_0_1_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 400, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 399, 402 },
      -1,
   } },
   /* replace143_0_1_0_0_0 -> 16 in the cache */
   /* replace143_0_1_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 386 },
      -1,
   } },
   /* replace143_0_1_0_1_0 -> 36 in the cache */
   /* replace143_0_1_0_1_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 404, 405 },
      -1,
   } },
   /* replace143_0_1_1_0_0 -> 16 in the cache */
   /* replace143_0_1_1_0_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 390 },
      -1,
   } },
   /* replace143_0_1_1_1_0 -> 36 in the cache */
   /* replace143_0_1_1_1_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 36, 390 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 407, 408 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 406, 409 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 403, 410 },
      -1,
   } },
   /* replace143_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 411, 137 },
      -1,
   } },

   /* ('sudot_4x8_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search144_0 -> 16 in the cache */
   /* search144_1 -> 36 in the cache */
   /* search144_2 -> 137 in the cache */
   /* search144 -> 326 in the cache */

   /* replace144_0_0_0_0_0 -> 16 in the cache */
   /* replace144_0_0_0_0_1 -> 124 in the cache */
   /* replace144_0_0_0_0 -> 378 in the cache */
   /* replace144_0_0_0_1_0 -> 36 in the cache */
   /* replace144_0_0_0_1_1 -> 124 in the cache */
   /* replace144_0_0_0_1 -> 398 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 378, 398 },
      -1,
   } },
   /* replace144_0_0_1_0_0 -> 16 in the cache */
   /* replace144_0_0_1_0_1 -> 381 in the cache */
   /* replace144_0_0_1_0 -> 382 in the cache */
   /* replace144_0_0_1_1_0 -> 36 in the cache */
   /* replace144_0_0_1_1_1 -> 381 in the cache */
   /* replace144_0_0_1_1 -> 401 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      4, 1,
      { 382, 401 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 413, 414 },
      -1,
   } },
   /* replace144_0_1_0_0_0 -> 16 in the cache */
   /* replace144_0_1_0_0_1 -> 386 in the cache */
   /* replace144_0_1_0_0 -> 387 in the cache */
   /* replace144_0_1_0_1_0 -> 36 in the cache */
   /* replace144_0_1_0_1_1 -> 386 in the cache */
   /* replace144_0_1_0_1 -> 405 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      6, 1,
      { 387, 405 },
      -1,
   } },
   /* replace144_0_1_1_0_0 -> 16 in the cache */
   /* replace144_0_1_1_0_1 -> 390 in the cache */
   /* replace144_0_1_1_0 -> 391 in the cache */
   /* replace144_0_1_1_1_0 -> 36 in the cache */
   /* replace144_0_1_1_1_1 -> 390 in the cache */
   /* replace144_0_1_1_1 -> 408 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      7, 1,
      { 391, 408 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 3,
      { 416, 417 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 7,
      { 415, 418 },
      -1,
   } },
   /* replace144_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 8,
      { 419, 137 },
      -1,
   } },

   /* ('sdot_2x16_iadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c') */
   /* search145_0 -> 16 in the cache */
   /* search145_1 -> 36 in the cache */
   /* search145_2 -> 137 in the cache */
   /* search145 -> 332 in the cache */

   /* replace145_0_0_0_0 -> 16 in the cache */
   /* replace145_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* replace145_0_0_1_0 -> 36 in the cache */
   /* replace145_0_0_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 36, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 421, 422 },
      -1,
   } },
   /* replace145_0_1_0_0 -> 16 in the cache */
   /* replace145_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 381 },
      -1,
   } },
   /* replace145_0_1_1_0 -> 36 in the cache */
   /* replace145_0_1_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 36, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 424, 425 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 423, 426 },
      -1,
   } },
   /* replace145_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 427, 137 },
      -1,
   } },

   /* ('udot_2x16_uadd', 'a', 'b', 'c') => ('iadd', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c') */
   /* search146_0 -> 16 in the cache */
   /* search146_1 -> 36 in the cache */
   /* search146_2 -> 137 in the cache */
   /* search146 -> 335 in the cache */

   /* replace146_0_0_0_0 -> 16 in the cache */
   /* replace146_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* replace146_0_0_1_0 -> 36 in the cache */
   /* replace146_0_0_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 36, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 429, 430 },
      -1,
   } },
   /* replace146_0_1_0_0 -> 16 in the cache */
   /* replace146_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 381 },
      -1,
   } },
   /* replace146_0_1_1_0 -> 36 in the cache */
   /* replace146_0_1_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 36, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 432, 433 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 431, 434 },
      -1,
   } },
   /* replace146_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 435, 137 },
      -1,
   } },

   /* ('udot_4x8_uadd_sat', 'a', 'b', '#c(is_ult_0xfffc07fc)') => ('udot_4x8_uadd', 'a', 'b', 'c') */
   /* search147_0 -> 16 in the cache */
   /* search147_1 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      7,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 36, 437 },
      -1,
   } },

   /* replace147_0 -> 16 in the cache */
   /* replace147_1 -> 36 in the cache */
   /* replace147_2 -> 137 in the cache */
   /* replace147 -> 323 in the cache */

   /* ('udot_4x8_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('iadd', ('imul', ('extract_u8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_u8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_u8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_u8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search148_0 -> 16 in the cache */
   /* search148_1 -> 36 in the cache */
   /* search148_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_4x8_uadd_sat,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace148_0_0_0_0_0 -> 16 in the cache */
   /* replace148_0_0_0_0_1 -> 124 in the cache */
   /* replace148_0_0_0_0 -> 397 in the cache */
   /* replace148_0_0_0_1_0 -> 36 in the cache */
   /* replace148_0_0_0_1_1 -> 124 in the cache */
   /* replace148_0_0_0_1 -> 398 in the cache */
   /* replace148_0_0_0 -> 399 in the cache */
   /* replace148_0_0_1_0_0 -> 16 in the cache */
   /* replace148_0_0_1_0_1 -> 381 in the cache */
   /* replace148_0_0_1_0 -> 400 in the cache */
   /* replace148_0_0_1_1_0 -> 36 in the cache */
   /* replace148_0_0_1_1_1 -> 381 in the cache */
   /* replace148_0_0_1_1 -> 401 in the cache */
   /* replace148_0_0_1 -> 402 in the cache */
   /* replace148_0_0 -> 403 in the cache */
   /* replace148_0_1_0_0_0 -> 16 in the cache */
   /* replace148_0_1_0_0_1 -> 386 in the cache */
   /* replace148_0_1_0_0 -> 404 in the cache */
   /* replace148_0_1_0_1_0 -> 36 in the cache */
   /* replace148_0_1_0_1_1 -> 386 in the cache */
   /* replace148_0_1_0_1 -> 405 in the cache */
   /* replace148_0_1_0 -> 406 in the cache */
   /* replace148_0_1_1_0_0 -> 16 in the cache */
   /* replace148_0_1_1_0_1 -> 390 in the cache */
   /* replace148_0_1_1_0 -> 407 in the cache */
   /* replace148_0_1_1_1_0 -> 36 in the cache */
   /* replace148_0_1_1_1_1 -> 390 in the cache */
   /* replace148_0_1_1_1 -> 408 in the cache */
   /* replace148_0_1_1 -> 409 in the cache */
   /* replace148_0_1 -> 410 in the cache */
   /* replace148_0 -> 411 in the cache */
   /* replace148_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 8,
      { 411, 137 },
      -1,
   } },

   /* ('sdot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_i8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_i8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_i8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_i8', 'b', 3)))), 'c') */
   /* search149_0 -> 16 in the cache */
   /* search149_1 -> 36 in the cache */
   /* search149_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_4x8_iadd_sat,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace149_0_0_0_0_0 -> 16 in the cache */
   /* replace149_0_0_0_0_1 -> 124 in the cache */
   /* replace149_0_0_0_0 -> 378 in the cache */
   /* replace149_0_0_0_1_0 -> 36 in the cache */
   /* replace149_0_0_0_1_1 -> 124 in the cache */
   /* replace149_0_0_0_1 -> 379 in the cache */
   /* replace149_0_0_0 -> 380 in the cache */
   /* replace149_0_0_1_0_0 -> 16 in the cache */
   /* replace149_0_0_1_0_1 -> 381 in the cache */
   /* replace149_0_0_1_0 -> 382 in the cache */
   /* replace149_0_0_1_1_0 -> 36 in the cache */
   /* replace149_0_0_1_1_1 -> 381 in the cache */
   /* replace149_0_0_1_1 -> 383 in the cache */
   /* replace149_0_0_1 -> 384 in the cache */
   /* replace149_0_0 -> 385 in the cache */
   /* replace149_0_1_0_0_0 -> 16 in the cache */
   /* replace149_0_1_0_0_1 -> 386 in the cache */
   /* replace149_0_1_0_0 -> 387 in the cache */
   /* replace149_0_1_0_1_0 -> 36 in the cache */
   /* replace149_0_1_0_1_1 -> 386 in the cache */
   /* replace149_0_1_0_1 -> 388 in the cache */
   /* replace149_0_1_0 -> 389 in the cache */
   /* replace149_0_1_1_0_0 -> 16 in the cache */
   /* replace149_0_1_1_0_1 -> 390 in the cache */
   /* replace149_0_1_1_0 -> 391 in the cache */
   /* replace149_0_1_1_1_0 -> 36 in the cache */
   /* replace149_0_1_1_1_1 -> 390 in the cache */
   /* replace149_0_1_1_1 -> 392 in the cache */
   /* replace149_0_1_1 -> 393 in the cache */
   /* replace149_0_1 -> 394 in the cache */
   /* replace149_0 -> 395 in the cache */
   /* replace149_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 8,
      { 395, 137 },
      -1,
   } },

   /* ('sudot_4x8_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('iadd', ('imul', ('extract_i8', 'a', 0), ('extract_u8', 'b', 0)), ('imul', ('extract_i8', 'a', 1), ('extract_u8', 'b', 1))), ('iadd', ('imul', ('extract_i8', 'a', 2), ('extract_u8', 'b', 2)), ('imul', ('extract_i8', 'a', 3), ('extract_u8', 'b', 3)))), 'c') */
   /* search150_0 -> 16 in the cache */
   /* search150_1 -> 36 in the cache */
   /* search150_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sudot_4x8_iadd_sat,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },

   /* replace150_0_0_0_0_0 -> 16 in the cache */
   /* replace150_0_0_0_0_1 -> 124 in the cache */
   /* replace150_0_0_0_0 -> 378 in the cache */
   /* replace150_0_0_0_1_0 -> 36 in the cache */
   /* replace150_0_0_0_1_1 -> 124 in the cache */
   /* replace150_0_0_0_1 -> 398 in the cache */
   /* replace150_0_0_0 -> 413 in the cache */
   /* replace150_0_0_1_0_0 -> 16 in the cache */
   /* replace150_0_0_1_0_1 -> 381 in the cache */
   /* replace150_0_0_1_0 -> 382 in the cache */
   /* replace150_0_0_1_1_0 -> 36 in the cache */
   /* replace150_0_0_1_1_1 -> 381 in the cache */
   /* replace150_0_0_1_1 -> 401 in the cache */
   /* replace150_0_0_1 -> 414 in the cache */
   /* replace150_0_0 -> 415 in the cache */
   /* replace150_0_1_0_0_0 -> 16 in the cache */
   /* replace150_0_1_0_0_1 -> 386 in the cache */
   /* replace150_0_1_0_0 -> 387 in the cache */
   /* replace150_0_1_0_1_0 -> 36 in the cache */
   /* replace150_0_1_0_1_1 -> 386 in the cache */
   /* replace150_0_1_0_1 -> 405 in the cache */
   /* replace150_0_1_0 -> 416 in the cache */
   /* replace150_0_1_1_0_0 -> 16 in the cache */
   /* replace150_0_1_1_0_1 -> 390 in the cache */
   /* replace150_0_1_1_0 -> 391 in the cache */
   /* replace150_0_1_1_1_0 -> 36 in the cache */
   /* replace150_0_1_1_1_1 -> 390 in the cache */
   /* replace150_0_1_1_1 -> 408 in the cache */
   /* replace150_0_1_1 -> 417 in the cache */
   /* replace150_0_1 -> 418 in the cache */
   /* replace150_0 -> 419 in the cache */
   /* replace150_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 8,
      { 419, 137 },
      -1,
   } },

   /* ('udot_2x16_uadd_sat', 'a', 'b', 'c') => ('uadd_sat', ('iadd', ('imul', ('extract_u16', 'a', 0), ('extract_u16', 'b', 0)), ('imul', ('extract_u16', 'a', 1), ('extract_u16', 'b', 1))), 'c') */
   /* search151_0 -> 16 in the cache */
   /* search151_1 -> 36 in the cache */
   /* search151_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_udot_2x16_uadd_sat,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace151_0_0_0_0 -> 16 in the cache */
   /* replace151_0_0_0_1 -> 124 in the cache */
   /* replace151_0_0_0 -> 429 in the cache */
   /* replace151_0_0_1_0 -> 36 in the cache */
   /* replace151_0_0_1_1 -> 124 in the cache */
   /* replace151_0_0_1 -> 430 in the cache */
   /* replace151_0_0 -> 431 in the cache */
   /* replace151_0_1_0_0 -> 16 in the cache */
   /* replace151_0_1_0_1 -> 381 in the cache */
   /* replace151_0_1_0 -> 432 in the cache */
   /* replace151_0_1_1_0 -> 36 in the cache */
   /* replace151_0_1_1_1 -> 381 in the cache */
   /* replace151_0_1_1 -> 433 in the cache */
   /* replace151_0_1 -> 434 in the cache */
   /* replace151_0 -> 435 in the cache */
   /* replace151_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 4,
      { 435, 137 },
      -1,
   } },

   /* ('sdot_2x16_iadd_sat', 'a', 'b', 'c') => ('iadd_sat', ('iadd', ('imul', ('extract_i16', 'a', 0), ('extract_i16', 'b', 0)), ('imul', ('extract_i16', 'a', 1), ('extract_i16', 'b', 1))), 'c') */
   /* search152_0 -> 16 in the cache */
   /* search152_1 -> 36 in the cache */
   /* search152_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sdot_2x16_iadd_sat,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace152_0_0_0_0 -> 16 in the cache */
   /* replace152_0_0_0_1 -> 124 in the cache */
   /* replace152_0_0_0 -> 421 in the cache */
   /* replace152_0_0_1_0 -> 36 in the cache */
   /* replace152_0_0_1_1 -> 124 in the cache */
   /* replace152_0_0_1 -> 422 in the cache */
   /* replace152_0_0 -> 423 in the cache */
   /* replace152_0_1_0_0 -> 16 in the cache */
   /* replace152_0_1_0_1 -> 381 in the cache */
   /* replace152_0_1_0 -> 424 in the cache */
   /* replace152_0_1_1_0 -> 36 in the cache */
   /* replace152_0_1_1_1 -> 381 in the cache */
   /* replace152_0_1_1 -> 425 in the cache */
   /* replace152_0_1 -> 426 in the cache */
   /* replace152_0 -> 427 in the cache */
   /* replace152_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 4,
      { 427, 137 },
      -1,
   } },

   /* ('~flrp@16', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search153_0 -> 11 in the cache */
   /* search153_1 -> 247 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 247, 450 },
      -1,
   } },

   /* replace153_0 -> 449 in the cache */
   /* replace153_1 -> 247 in the cache */
   /* replace153_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 449, 247, 11 },
      -1,
   } },

   /* ('~flrp@16', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search154_0 -> 11 in the cache */
   /* search154_1_0 -> 11 in the cache */
   /* search154_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 247 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 11, 453, 454 },
      -1,
   } },

   /* replace154_0_0 -> 247 in the cache */
   /* replace154_0_1 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 247, 454 },
      -1,
   } },
   /* replace154_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 456, 11 },
      -1,
   } },

   /* ('~flrp@16', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search155_0_0 -> 11 in the cache */
   /* search155_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 247 },
      0,
   } },
   /* search155_1_0 -> 11 in the cache */
   /* search155_1_1 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 11, 454 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 458, 459, 460 },
      -1,
   } },

   /* replace155_0_0 -> 247 in the cache */
   /* replace155_0_1 -> 454 in the cache */
   /* replace155_0_2 -> 460 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 247, 454, 460 },
      -1,
   } },
   /* replace155_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 462, 11 },
      -1,
   } },

   /* ('~flrp@16', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search156_0 -> 11 in the cache */
   /* search156_1_0 -> 11 in the cache */
   /* search156_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 11, 247 },
      0,
   } },
   /* search156_2 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 11, 464, 454 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace156_0_1 -> 247 in the cache */
   /* replace156_0_2 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 466, 247, 454 },
      -1,
   } },
   /* replace156_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 467, 11 },
      -1,
   } },

   /* ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search157_0_0 -> 11 in the cache */
   /* search157_0_1_0 -> 466 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 466, 470 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 471 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search157_1_1 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 473, 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 472, 474 },
      -1,
   } },

   /* replace157_0 -> 11 in the cache */
   /* replace157_1 -> 473 in the cache */
   /* replace157_2 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 473, 469 },
      -1,
   } },

   /* ('~fadd@16', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search158_0_0 -> 11 in the cache */
   /* search158_0_1_0_0 -> 466 in the cache */
   /* search158_0_1_0_1_0 -> 469 in the cache */
   /* search158_0_1_0_1 -> 470 in the cache */
   /* search158_0_1_0 -> 471 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 471 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 477 },
      -1,
   } },
   /* search158_1_0 -> 473 in the cache */
   /* search158_1_1_0 -> 469 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 473, 479 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 478, 480 },
      -1,
   } },

   /* replace158_0 -> 11 in the cache */
   /* replace158_1 -> 473 in the cache */
   /* replace158_2_0 -> 469 in the cache */
   /* replace158_2 -> 479 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 11, 473, 479 },
      -1,
   } },

   /* ('~fadd@16', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search159_0 -> 11 in the cache */
   /* search159_1_0 -> 469 in the cache */
   /* search159_1_1_0 -> 473 in the cache */
   /* search159_1_1_1_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 473, 483 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 469, 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 11, 485 },
      -1,
   } },

   /* replace159_0 -> 11 in the cache */
   /* replace159_1 -> 473 in the cache */
   /* replace159_2 -> 469 in the cache */
   /* replace159 -> 476 in the cache */

   /* ('~fadd@16', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search160_0_0 -> 11 in the cache */
   /* search160_0_1_0 -> 466 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 487 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 488 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 466, 489 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 11, 490 },
      -1,
   } },
   /* search160_1_0 -> 473 in the cache */
   /* search160_1_1_0 -> 487 in the cache */
   /* search160_1_1 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 473, 488 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 491, 492 },
      -1,
   } },

   /* replace160_0 -> 487 in the cache */
   /* replace160_1 -> 473 in the cache */
   /* replace160_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 487, 473, 11 },
      -1,
   } },

   /* ('~fadd@16', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search161_0 -> 11 in the cache */
   /* search161_1_0_0 -> 487 in the cache */
   /* search161_1_0 -> 488 in the cache */
   /* search161_1_1_0 -> 473 in the cache */
   /* search161_1_1_1_0 -> 11 in the cache */
   /* search161_1_1_1 -> 483 in the cache */
   /* search161_1_1 -> 484 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 488, 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 11, 495 },
      -1,
   } },

   /* replace161_0 -> 487 in the cache */
   /* replace161_1 -> 473 in the cache */
   /* replace161_2 -> 11 in the cache */
   /* replace161 -> 494 in the cache */

   /* ('~ffma@16', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search162_0 -> 11 in the cache */
   /* search162_1_0 -> 466 in the cache */
   /* search162_1_1_0_0 -> 487 in the cache */
   /* search162_1_1_0 -> 488 in the cache */
   /* search162_1_1 -> 489 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 466, 489 },
      -1,
   } },
   /* search162_2_0 -> 473 in the cache */
   /* search162_2_1_0 -> 487 in the cache */
   /* search162_2_1 -> 488 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 473, 488 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 11, 497, 498 },
      -1,
   } },

   /* replace162_0 -> 487 in the cache */
   /* replace162_1 -> 473 in the cache */
   /* replace162_2 -> 11 in the cache */
   /* replace162 -> 494 in the cache */

   /* ('~ffma@16', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search163_1_0 -> 487 in the cache */
   /* search163_1 -> 488 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 501 },
      -1,
   } },
   /* search163_2_1_0 -> 487 in the cache */
   /* search163_2_1 -> 488 in the cache */
   /* search163_2_2 -> 501 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 502, 488, 501 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 500, 488, 503 },
      -1,
   } },

   /* replace163_0 -> 487 in the cache */
   /* replace163_1 -> 500 in the cache */
   /* replace163_2 -> 501 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 487, 500, 501 },
      -1,
   } },

   /* ('~ffma@16', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 506 },
      -1,
   } },
   /* search164_1_0 -> 247 in the cache */
   /* search164_1_1_0 -> 501 in the cache */
   /* search164_1_1 -> 502 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 247, 502 },
      -1,
   } },
   /* search164_2 -> 501 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 507, 508, 501 },
      -1,
   } },

   /* replace164_0 -> 506 in the cache */
   /* replace164_1 -> 247 in the cache */
   /* replace164_2 -> 501 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 247, 501 },
      -1,
   } },

   /* ('~ffma@16', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search165_0_0 -> 506 in the cache */
   /* search165_0 -> 507 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 511 },
      -1,
   } },
   /* search165_1_1 -> 473 in the cache */
   /* search165_1_2 -> 460 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 512, 473, 460 },
      -1,
   } },
   /* search165_2_0 -> 511 in the cache */
   /* search165_2_1 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 511, 473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 507, 513, 514 },
      -1,
   } },

   /* replace165_0 -> 506 in the cache */
   /* replace165_1 -> 460 in the cache */
   /* replace165_2_0 -> 511 in the cache */
   /* replace165_2_1 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 511, 473 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 506, 460, 516 },
      -1,
   } },

   /* ('~fadd@16', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search166_0 -> 466 in the cache */
   /* search166_1_0_0_0 -> 466 in the cache */
   /* search166_1_0_0_1_0 -> 11 in the cache */
   /* search166_1_0_0_1 -> 483 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 466, 483 },
      -1,
   } },
   /* search166_1_0_1_0 -> 466 in the cache */
   /* search166_1_0_1_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 466, 519 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 518, 520 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 466, 522 },
      -1,
   } },

   /* replace166_0 -> 247 in the cache */
   /* replace166_1 -> 466 in the cache */
   /* replace166_2 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 247, 466, 11 },
      -1,
   } },

   /* ('~flrp@32', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search167_0 -> 16 in the cache */
   /* search167_1 -> 36 in the cache */
   /* search167_2_0 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 36, 525 },
      -1,
   } },

   /* replace167_0 -> 449 in the cache */
   /* replace167_1 -> 36 in the cache */
   /* replace167_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 449, 36, 16 },
      -1,
   } },

   /* ('~flrp@32', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search168_0 -> 16 in the cache */
   /* search168_1_0 -> 16 in the cache */
   /* search168_1_1 -> 36 in the cache */
   /* search168_1 -> 263 in the cache */
   /* search168_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 16, 263, 137 },
      -1,
   } },

   /* replace168_0_0 -> 36 in the cache */
   /* replace168_0_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 36, 137 },
      -1,
   } },
   /* replace168_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 529, 16 },
      -1,
   } },

   /* ('~flrp@32', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search169_0_0 -> 16 in the cache */
   /* search169_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 36 },
      0,
   } },
   /* search169_1_0 -> 16 in the cache */
   /* search169_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 137 },
      0,
   } },
   /* search169_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 531, 532, 165 },
      -1,
   } },

   /* replace169_0_0 -> 36 in the cache */
   /* replace169_0_1 -> 137 in the cache */
   /* replace169_0_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 36, 137, 165 },
      -1,
   } },
   /* replace169_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 534, 16 },
      -1,
   } },

   /* ('~flrp@32', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search170_0 -> 16 in the cache */
   /* search170_1_0 -> 16 in the cache */
   /* search170_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 16, 36 },
      0,
   } },
   /* search170_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 16, 536, 137 },
      -1,
   } },

   /* replace170_0_0 -> 232 in the cache */
   /* replace170_0_1 -> 36 in the cache */
   /* replace170_0_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 232, 36, 137 },
      -1,
   } },
   /* replace170_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 538, 16 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search171_0_0 -> 16 in the cache */
   /* search171_0_1_0 -> 232 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 540 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 232, 541 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 542 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search171_1_1 -> 540 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 544, 540 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 543, 545 },
      -1,
   } },

   /* replace171_0 -> 16 in the cache */
   /* replace171_1 -> 544 in the cache */
   /* replace171_2 -> 540 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 544, 540 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search172_0_0 -> 16 in the cache */
   /* search172_0_1_0_0 -> 232 in the cache */
   /* search172_0_1_0_1_0 -> 540 in the cache */
   /* search172_0_1_0_1 -> 541 in the cache */
   /* search172_0_1_0 -> 542 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 542 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 548 },
      -1,
   } },
   /* search172_1_0 -> 544 in the cache */
   /* search172_1_1_0 -> 540 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 540 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 544, 550 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 549, 551 },
      -1,
   } },

   /* replace172_0 -> 16 in the cache */
   /* replace172_1 -> 544 in the cache */
   /* replace172_2_0 -> 540 in the cache */
   /* replace172_2 -> 550 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 16, 544, 550 },
      -1,
   } },

   /* ('~fadd@32', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search173_0 -> 16 in the cache */
   /* search173_1_0 -> 540 in the cache */
   /* search173_1_1_0 -> 544 in the cache */
   /* search173_1_1_1_0 -> 16 in the cache */
   /* search173_1_1_1 -> 269 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 544, 269 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 540, 554 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 16, 555 },
      -1,
   } },

   /* replace173_0 -> 16 in the cache */
   /* replace173_1 -> 544 in the cache */
   /* replace173_2 -> 540 in the cache */
   /* replace173 -> 547 in the cache */

   /* ('~fadd@32', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search174_0_0 -> 16 in the cache */
   /* search174_0_1_0 -> 232 in the cache */
   /* search174_0_1_1_0_0 -> 487 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 487 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 232, 558 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 16, 559 },
      -1,
   } },
   /* search174_1_0 -> 544 in the cache */
   /* search174_1_1_0 -> 487 in the cache */
   /* search174_1_1 -> 557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 544, 557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 560, 561 },
      -1,
   } },

   /* replace174_0 -> 487 in the cache */
   /* replace174_1 -> 544 in the cache */
   /* replace174_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 487, 544, 16 },
      -1,
   } },

   /* ('~fadd@32', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search175_0 -> 16 in the cache */
   /* search175_1_0_0 -> 487 in the cache */
   /* search175_1_0 -> 557 in the cache */
   /* search175_1_1_0 -> 544 in the cache */
   /* search175_1_1_1_0 -> 16 in the cache */
   /* search175_1_1_1 -> 269 in the cache */
   /* search175_1_1 -> 554 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 557, 554 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 16, 564 },
      -1,
   } },

   /* replace175_0 -> 487 in the cache */
   /* replace175_1 -> 544 in the cache */
   /* replace175_2 -> 16 in the cache */
   /* replace175 -> 563 in the cache */

   /* ('~ffma@32', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search176_0 -> 16 in the cache */
   /* search176_1_0 -> 232 in the cache */
   /* search176_1_1_0_0 -> 487 in the cache */
   /* search176_1_1_0 -> 557 in the cache */
   /* search176_1_1 -> 558 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 232, 558 },
      -1,
   } },
   /* search176_2_0 -> 544 in the cache */
   /* search176_2_1_0 -> 487 in the cache */
   /* search176_2_1 -> 557 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 544, 557 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 16, 566, 567 },
      -1,
   } },

   /* replace176_0 -> 487 in the cache */
   /* replace176_1 -> 544 in the cache */
   /* replace176_2 -> 16 in the cache */
   /* replace176 -> 563 in the cache */

   /* ('~ffma@32', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   /* search177_0 -> 362 in the cache */
   /* search177_1_0 -> 487 in the cache */
   /* search177_1 -> 557 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 569 },
      -1,
   } },
   /* search177_2_1_0 -> 487 in the cache */
   /* search177_2_1 -> 557 in the cache */
   /* search177_2_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 570, 557, 569 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 362, 557, 571 },
      -1,
   } },

   /* replace177_0 -> 487 in the cache */
   /* replace177_1 -> 362 in the cache */
   /* replace177_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 487, 362, 569 },
      -1,
   } },

   /* ('~ffma@32', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   /* search178_0_0 -> 506 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 506 },
      -1,
   } },
   /* search178_1_0 -> 36 in the cache */
   /* search178_1_1_0 -> 569 in the cache */
   /* search178_1_1 -> 570 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 36, 570 },
      -1,
   } },
   /* search178_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 574, 575, 569 },
      -1,
   } },

   /* replace178_0 -> 506 in the cache */
   /* replace178_1 -> 36 in the cache */
   /* replace178_2 -> 569 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 36, 569 },
      -1,
   } },

   /* ('~ffma@32', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search179_0_0 -> 506 in the cache */
   /* search179_0 -> 574 in the cache */
   /* search179_1_0_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 364 },
      -1,
   } },
   /* search179_1_1 -> 544 in the cache */
   /* search179_1_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 578, 544, 165 },
      -1,
   } },
   /* search179_2_0 -> 364 in the cache */
   /* search179_2_1 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 364, 544 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 574, 579, 580 },
      -1,
   } },

   /* replace179_0 -> 506 in the cache */
   /* replace179_1 -> 165 in the cache */
   /* replace179_2_0 -> 364 in the cache */
   /* replace179_2_1 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 364, 544 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 506, 165, 582 },
      -1,
   } },

   /* ('~fadd@32', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search180_0 -> 232 in the cache */
   /* search180_1_0_0_0 -> 232 in the cache */
   /* search180_1_0_0_1_0 -> 16 in the cache */
   /* search180_1_0_0_1 -> 269 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 232, 269 },
      -1,
   } },
   /* search180_1_0_1_0 -> 232 in the cache */
   /* search180_1_0_1_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 232, 585 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 584, 586 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 587 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 232, 588 },
      -1,
   } },

   /* replace180_0 -> 36 in the cache */
   /* replace180_1 -> 232 in the cache */
   /* replace180_2 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 36, 232, 16 },
      -1,
   } },

   /* ('~flrp@64', 'a', 'b', ('b2f', 'c@1')) => ('bcsel', 'c', 'b', 'a') */
   /* search181_0 -> 21 in the cache */
   /* search181_1 -> 41 in the cache */
   /* search181_2_0 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 41, 591 },
      -1,
   } },

   /* replace181_0 -> 449 in the cache */
   /* replace181_1 -> 41 in the cache */
   /* replace181_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 449, 41, 21 },
      -1,
   } },

   /* ('~flrp@64', 'a', ('fadd', 'a', 'b'), 'c') => ('fadd', ('fmul', 'b', 'c'), 'a') */
   /* search182_0 -> 21 in the cache */
   /* search182_1_0 -> 21 in the cache */
   /* search182_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 41 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 21, 594, 595 },
      -1,
   } },

   /* replace182_0_0 -> 41 in the cache */
   /* replace182_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 41, 595 },
      -1,
   } },
   /* replace182_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 597, 21 },
      -1,
   } },

   /* ('~flrp@64', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search183_0_0 -> 21 in the cache */
   /* search183_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search183_1_0 -> 21 in the cache */
   /* search183_1_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 21, 595 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 599, 600, 601 },
      -1,
   } },

   /* replace183_0_0 -> 41 in the cache */
   /* replace183_0_1 -> 595 in the cache */
   /* replace183_0_2 -> 601 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 41, 595, 601 },
      -1,
   } },
   /* replace183_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 603, 21 },
      -1,
   } },

   /* ('~flrp@64', 'a', ('fmul(is_used_once)', 'a', 'b'), 'c') => ('fmul', ('flrp', 1.0, 'b', 'c'), 'a') */
   /* search184_0 -> 21 in the cache */
   /* search184_1_0 -> 21 in the cache */
   /* search184_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search184_2 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 1,
      { 21, 605, 595 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace184_0_1 -> 41 in the cache */
   /* replace184_0_2 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 607, 41, 595 },
      -1,
   } },
   /* replace184_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 608, 21 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', 'c'))), ('fmul', 'b', 'c')) => ('flrp', 'a', 'b', 'c') */
   /* search185_0_0 -> 21 in the cache */
   /* search185_0_1_0 -> 607 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 607, 611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 612 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search185_1_1 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 614, 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 613, 615 },
      -1,
   } },

   /* replace185_0 -> 21 in the cache */
   /* replace185_1 -> 614 in the cache */
   /* replace185_2 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 614, 610 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', ('fsat', ('fadd', 1.0, ('fneg', 'c')))), ('fmul', 'b', ('fsat', 'c'))) => ('flrp', 'a', 'b', ('fsat', 'c')) */
   /* search186_0_0 -> 21 in the cache */
   /* search186_0_1_0_0 -> 607 in the cache */
   /* search186_0_1_0_1_0 -> 610 in the cache */
   /* search186_0_1_0_1 -> 611 in the cache */
   /* search186_0_1_0 -> 612 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 612 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 618 },
      -1,
   } },
   /* search186_1_0 -> 614 in the cache */
   /* search186_1_1_0 -> 610 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 610 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 614, 620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 619, 621 },
      -1,
   } },

   /* replace186_0 -> 21 in the cache */
   /* replace186_1 -> 614 in the cache */
   /* replace186_2_0 -> 610 in the cache */
   /* replace186_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 21, 614, 620 },
      -1,
   } },

   /* ('~fadd@64', 'a', ('fmul', 'c', ('fadd', 'b', ('fneg', 'a')))) => ('flrp', 'a', 'b', 'c') */
   /* search187_0 -> 21 in the cache */
   /* search187_1_0 -> 610 in the cache */
   /* search187_1_1_0 -> 614 in the cache */
   /* search187_1_1_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 614, 624 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 610, 625 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 21, 626 },
      -1,
   } },

   /* replace187_0 -> 21 in the cache */
   /* replace187_1 -> 614 in the cache */
   /* replace187_2 -> 610 in the cache */
   /* replace187 -> 617 in the cache */

   /* ('~fadd@64', ('fmul', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1')))), ('fmul', 'b', ('b2f', 'c'))) => ('bcsel', 'c', 'b', 'a') */
   /* search188_0_0 -> 21 in the cache */
   /* search188_0_1_0 -> 607 in the cache */
   /* search188_0_1_1_0_0 -> 487 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 487 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 628 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 607, 629 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 21, 630 },
      -1,
   } },
   /* search188_1_0 -> 614 in the cache */
   /* search188_1_1_0 -> 487 in the cache */
   /* search188_1_1 -> 628 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 614, 628 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 631, 632 },
      -1,
   } },

   /* replace188_0 -> 487 in the cache */
   /* replace188_1 -> 614 in the cache */
   /* replace188_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 487, 614, 21 },
      -1,
   } },

   /* ('~fadd@64', 'a', ('fmul', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')))) => ('bcsel', 'c', 'b', 'a') */
   /* search189_0 -> 21 in the cache */
   /* search189_1_0_0 -> 487 in the cache */
   /* search189_1_0 -> 628 in the cache */
   /* search189_1_1_0 -> 614 in the cache */
   /* search189_1_1_1_0 -> 21 in the cache */
   /* search189_1_1_1 -> 624 in the cache */
   /* search189_1_1 -> 625 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 628, 625 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 21, 635 },
      -1,
   } },

   /* replace189_0 -> 487 in the cache */
   /* replace189_1 -> 614 in the cache */
   /* replace189_2 -> 21 in the cache */
   /* replace189 -> 634 in the cache */

   /* ('~ffma@64', 'a', ('fadd', 1.0, ('fneg', ('b2f', 'c@1'))), ('fmul', 'b', ('b2f', 'c@1'))) => ('bcsel', 'c', 'b', 'a') */
   /* search190_0 -> 21 in the cache */
   /* search190_1_0 -> 607 in the cache */
   /* search190_1_1_0_0 -> 487 in the cache */
   /* search190_1_1_0 -> 628 in the cache */
   /* search190_1_1 -> 629 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 607, 629 },
      -1,
   } },
   /* search190_2_0 -> 614 in the cache */
   /* search190_2_1_0 -> 487 in the cache */
   /* search190_2_1 -> 628 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 614, 628 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 21, 637, 638 },
      -1,
   } },

   /* replace190_0 -> 487 in the cache */
   /* replace190_1 -> 614 in the cache */
   /* replace190_2 -> 21 in the cache */
   /* replace190 -> 634 in the cache */

   /* ('~ffma@64', 'b', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), ('b2f', 'c@1'), 'a')) => ('bcsel', 'c', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search191_1_0 -> 487 in the cache */
   /* search191_1 -> 628 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 641 },
      -1,
   } },
   /* search191_2_1_0 -> 487 in the cache */
   /* search191_2_1 -> 628 in the cache */
   /* search191_2_2 -> 641 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 642, 628, 641 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 640, 628, 643 },
      -1,
   } },

   /* replace191_0 -> 487 in the cache */
   /* replace191_1 -> 640 in the cache */
   /* replace191_2 -> 641 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 487, 640, 641 },
      -1,
   } },

   /* ('~ffma@64', ('b2f', 'c@1'), ('fadd', 'b', ('fneg', 'a')), 'a') => ('bcsel', 'c', 'b', 'a') */
   /* search192_0_0 -> 506 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 506 },
      -1,
   } },
   /* search192_1_0 -> 41 in the cache */
   /* search192_1_1_0 -> 641 in the cache */
   /* search192_1_1 -> 642 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 41, 642 },
      -1,
   } },
   /* search192_2 -> 641 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 646, 647, 641 },
      -1,
   } },

   /* replace192_0 -> 506 in the cache */
   /* replace192_1 -> 41 in the cache */
   /* replace192_2 -> 641 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 41, 641 },
      -1,
   } },

   /* ('~ffma@64', ('b2f', 'c@1'), ('ffma', ('fneg', 'a'), 'b', 'd'), ('fmul', 'a', 'b')) => ('bcsel', 'c', 'd', ('fmul', 'a', 'b')) */
   /* search193_0_0 -> 506 in the cache */
   /* search193_0 -> 646 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 650 },
      -1,
   } },
   /* search193_1_1 -> 614 in the cache */
   /* search193_1_2 -> 601 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 651, 614, 601 },
      -1,
   } },
   /* search193_2_0 -> 650 in the cache */
   /* search193_2_1 -> 614 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 650, 614 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 646, 652, 653 },
      -1,
   } },

   /* replace193_0 -> 506 in the cache */
   /* replace193_1 -> 601 in the cache */
   /* replace193_2_0 -> 650 in the cache */
   /* replace193_2_1 -> 614 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 650, 614 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 506, 601, 655 },
      -1,
   } },

   /* ('~fadd@64', 1.0, ('fneg', ('fmul', ('fadd', 1.0, ('fneg', 'a')), ('fadd', 1.0, ('fneg', 'b'))))) => ('flrp', 'b', 1.0, 'a') */
   /* search194_0 -> 607 in the cache */
   /* search194_1_0_0_0 -> 607 in the cache */
   /* search194_1_0_0_1_0 -> 21 in the cache */
   /* search194_1_0_0_1 -> 624 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 607, 624 },
      -1,
   } },
   /* search194_1_0_1_0 -> 607 in the cache */
   /* search194_1_0_1_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 607, 658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 3,
      { 657, 659 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 660 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 607, 661 },
      -1,
   } },

   /* replace194_0 -> 41 in the cache */
   /* replace194_1 -> 607 in the cache */
   /* replace194_2 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 41, 607, 21 },
      -1,
   } },

   /* ('~flrp', ('fmul(is_used_once)', 'a', 'b'), ('fmul(is_used_once)', 'a', 'c'), 'd') => ('fmul', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search195_0_0 -> 151 in the cache */
   /* search195_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 151, 152 },
      0,
   } },
   /* search195_1_0 -> 151 in the cache */
   /* search195_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 151, 154 },
      0,
   } },
   /* search195_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 664, 665, 155 },
      -1,
   } },

   /* replace195_0_0 -> 152 in the cache */
   /* replace195_0_1 -> 154 in the cache */
   /* replace195_0_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 152, 154, 155 },
      -1,
   } },
   /* replace195_1 -> 151 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 667, 151 },
      -1,
   } },

   /* ('~flrp', 'a', 0.0, 'c') => ('fadd', ('fmul', ('fneg', 'a'), 'c'), 'a') */
   /* search196_0 -> 0 in the cache */
   /* search196_1 -> 245 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 0, 245, 669 },
      -1,
   } },

   /* replace196_0_0_0 -> 0 in the cache */
   /* replace196_0_0 -> 201 in the cache */
   /* replace196_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 201, 669 },
      -1,
   } },
   /* replace196_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 671, 0 },
      -1,
   } },

   /* ('ftrunc', 'a') => ('bcsel', ('flt', 'a', 0.0), ('fneg', ('ffloor', ('fabs', 'a'))), ('ffloor', ('fabs', 'a'))) */
   /* search197_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace197_0_0 -> 32 in the cache */
   /* replace197_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 113 },
      -1,
   } },
   /* replace197_1_0_0_0 -> 32 in the cache */
   /* replace197_1_0_0 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 103 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 675 },
      -1,
   } },
   /* replace197_2_0_0 -> 32 in the cache */
   /* replace197_2_0 -> 103 in the cache */
   /* replace197_2 -> 675 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 674, 676, 675 },
      -1,
   } },

   /* ('ffloor@16', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search198_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace198_0 -> 11 in the cache */
   /* replace198_1_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 11, 679 },
      -1,
   } },

   /* ('ffloor@32', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search199_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace199_0 -> 16 in the cache */
   /* replace199_1_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 16, 682 },
      -1,
   } },

   /* ('ffloor@64', 'a') => ('fsub', 'a', ('ffract', 'a')) */
   /* search200_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace200_0 -> 21 in the cache */
   /* replace200_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 21, 685 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search201_0 -> 11 in the cache */
   /* search201_1_0_0 -> 11 in the cache */
   /* search201_1_0 -> 679 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 679 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 11, 687 },
      -1,
   } },

   /* replace201_0 -> 11 in the cache */
   /* replace201 -> 678 in the cache */

   /* ('fadd@32', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search202_0 -> 16 in the cache */
   /* search202_1_0_0 -> 16 in the cache */
   /* search202_1_0 -> 682 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 682 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 16, 689 },
      -1,
   } },

   /* replace202_0 -> 16 in the cache */
   /* replace202 -> 681 in the cache */

   /* ('fadd@64', 'a', ('fneg', ('ffract', 'a'))) => ('ffloor', 'a') */
   /* search203_0 -> 21 in the cache */
   /* search203_1_0_0 -> 21 in the cache */
   /* search203_1_0 -> 685 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 685 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 21, 691 },
      -1,
   } },

   /* replace203_0 -> 21 in the cache */
   /* replace203 -> 684 in the cache */

   /* ('ffract@16', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search204_0 -> 11 in the cache */
   /* search204 -> 679 in the cache */

   /* replace204_0 -> 11 in the cache */
   /* replace204_1_0 -> 11 in the cache */
   /* replace204_1 -> 678 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 11, 678 },
      -1,
   } },

   /* ('ffract@32', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search205_0 -> 16 in the cache */
   /* search205 -> 682 in the cache */

   /* replace205_0 -> 16 in the cache */
   /* replace205_1_0 -> 16 in the cache */
   /* replace205_1 -> 681 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 16, 681 },
      -1,
   } },

   /* ('ffract@64', 'a') => ('fsub', 'a', ('ffloor', 'a')) */
   /* search206_0 -> 21 in the cache */
   /* search206 -> 685 in the cache */

   /* replace206_0 -> 21 in the cache */
   /* replace206_1_0 -> 21 in the cache */
   /* replace206_1 -> 684 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 21, 684 },
      -1,
   } },

   /* ('fceil', 'a') => ('fneg', ('ffloor', ('fneg', 'a'))) */
   /* search207_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace207_0_0_0 -> 32 in the cache */
   /* replace207_0_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 98 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 697 },
      -1,
   } },

   /* ('ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search208_0 -> 11 in the cache */
   /* search208_1 -> 247 in the cache */
   /* search208_2 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 247, 454 },
      -1,
   } },

   /* replace208_0_0 -> 11 in the cache */
   /* replace208_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 11, 247 },
      -1,
   } },
   /* replace208_1 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 700, 454 },
      -1,
   } },

   /* ('ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search209_0 -> 16 in the cache */
   /* search209_1 -> 36 in the cache */
   /* search209_2 -> 137 in the cache */
   /* search209 -> 223 in the cache */

   /* replace209_0_0 -> 16 in the cache */
   /* replace209_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace209_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 702, 137 },
      -1,
   } },

   /* ('ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search210_0 -> 21 in the cache */
   /* search210_1 -> 41 in the cache */
   /* search210_2 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 21, 41, 595 },
      -1,
   } },

   /* replace210_0_0 -> 21 in the cache */
   /* replace210_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace210_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 705, 595 },
      -1,
   } },

   /* ('ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   /* search211_0 -> 16 in the cache */
   /* search211_1 -> 36 in the cache */
   /* search211_2 -> 137 in the cache */
   /* search211 -> 377 in the cache */

   /* replace211_0_0 -> 16 in the cache */
   /* replace211_0_1 -> 36 in the cache */
   /* replace211_0 -> 136 in the cache */
   /* replace211_1 -> 137 in the cache */
   /* replace211 -> 277 in the cache */

   /* ('~ffma@16', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search212_0 -> 11 in the cache */
   /* search212_1 -> 247 in the cache */
   /* search212_2 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 11, 247, 454 },
      -1,
   } },

   /* replace212_0_0 -> 11 in the cache */
   /* replace212_0_1 -> 247 in the cache */
   /* replace212_0 -> 700 in the cache */
   /* replace212_1 -> 454 in the cache */
   /* replace212 -> 701 in the cache */

   /* ('~ffma@32', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search213_0 -> 16 in the cache */
   /* search213_1 -> 36 in the cache */
   /* search213_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace213_0_0 -> 16 in the cache */
   /* replace213_0_1 -> 36 in the cache */
   /* replace213_0 -> 702 in the cache */
   /* replace213_1 -> 137 in the cache */
   /* replace213 -> 703 in the cache */

   /* ('~ffma@64', 'a', 'b', 'c') => ('fadd', ('fmul', 'a', 'b'), 'c') */
   /* search214_0 -> 21 in the cache */
   /* search214_1 -> 41 in the cache */
   /* search214_2 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 21, 41, 595 },
      -1,
   } },

   /* replace214_0_0 -> 21 in the cache */
   /* replace214_0_1 -> 41 in the cache */
   /* replace214_0 -> 705 in the cache */
   /* replace214_1 -> 595 in the cache */
   /* replace214 -> 706 in the cache */

   /* ('~ffmaz', 'a', 'b', 'c') => ('fadd', ('fmulz', 'a', 'b'), 'c') */
   /* search215_0 -> 16 in the cache */
   /* search215_1 -> 36 in the cache */
   /* search215_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace215_0_0 -> 16 in the cache */
   /* replace215_0_1 -> 36 in the cache */
   /* replace215_0 -> 136 in the cache */
   /* replace215_1 -> 137 in the cache */
   /* replace215 -> 277 in the cache */

   /* ('~fmul', ('fadd', ('iand', ('ineg', ('b2i', 'a@bool')), ('fmul', 'b', 'c')), '#d'), '#e') => ('bcsel', 'a', ('fmul', ('fadd', ('fmul', 'b', 'c'), 'd'), 'e'), ('fmul', 'd', 'e')) */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 712 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 714, 715 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_iand,
      2, 2,
      { 713, 716 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 3,
      { 717, 718 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 4,
      { 719, 720 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace216_1_0_0_0 -> 714 in the cache */
   /* replace216_1_0_0_1 -> 715 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 714, 715 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 723, 724 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 725, 726 },
      -1,
   } },
   /* replace216_2_0 -> 724 in the cache */
   /* replace216_2_1 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 724, 726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 722, 727, 728 },
      -1,
   } },

   /* ('fdph', 'a', 'b') => ('fdot4', ('vec4', 'a.x', 'a.y', 'a.z', 1.0), 'b') */
   /* search217_0 -> 0 in the cache */
   /* search217_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   /* replace217_0_3 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 731, 732, 733, 259 },
      -1,
   } },
   /* replace217_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 734, 3 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 'b', 'c', 1.0), 'd') => ('fdph', ('vec3', 'a', 'b', 'c'), 'd') */
   /* search218_0_0 -> 151 in the cache */
   /* search218_0_1 -> 152 in the cache */
   /* search218_0_2 -> 154 in the cache */
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 151, 152, 154, 736 },
      -1,
   } },
   /* search218_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 737, 155 },
      -1,
   } },

   /* replace218_0_0 -> 151 in the cache */
   /* replace218_0_1 -> 152 in the cache */
   /* replace218_0_2 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 151, 152, 154 },
      -1,
   } },
   /* replace218_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 739, 155 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 0.0, 0.0, 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search219_0_0 -> 0 in the cache */
   /* search219_0_1 -> 245 in the cache */
   /* search219_0_2 -> 245 in the cache */
   /* search219_0_3 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 0, 245, 245, 245 },
      -1,
   } },
   /* search219_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 741, 3 },
      -1,
   } },

   /* replace219_0 -> 0 in the cache */
   /* replace219_1 -> 3 in the cache */
   /* replace219 -> 253 in the cache */

   /* ('fdot4', ('vec4', 'a', 'b', 0.0, 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c') */
   /* search220_0_0 -> 128 in the cache */
   /* search220_0_1 -> 129 in the cache */
   { .constant = {
      { nir_search_value_constant, -3 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search220_0_3 -> 743 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 128, 129, 743, 743 },
      -1,
   } },
   /* search220_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 744, 131 },
      -1,
   } },

   /* replace220_0_0 -> 128 in the cache */
   /* replace220_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 128, 129 },
      -1,
   } },
   /* replace220_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 746, 131 },
      -1,
   } },

   /* ('fdot4', ('vec4', 'a', 'b', 'c', 0.0), 'd') => ('fdot3', ('vec3', 'a', 'b', 'c'), 'd') */
   /* search221_0_0 -> 151 in the cache */
   /* search221_0_1 -> 152 in the cache */
   /* search221_0_2 -> 154 in the cache */
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 151, 152, 154, 748 },
      -1,
   } },
   /* search221_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 749, 155 },
      -1,
   } },

   /* replace221_0_0 -> 151 in the cache */
   /* replace221_0_1 -> 152 in the cache */
   /* replace221_0_2 -> 154 in the cache */
   /* replace221_0 -> 739 in the cache */
   /* replace221_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 739, 155 },
      -1,
   } },

   /* ('fdot3', ('vec3', 'a', 0.0, 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search222_0_0 -> 0 in the cache */
   /* search222_0_1 -> 245 in the cache */
   /* search222_0_2 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 0, 245, 245 },
      -1,
   } },
   /* search222_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 752, 3 },
      -1,
   } },

   /* replace222_0 -> 0 in the cache */
   /* replace222_1 -> 3 in the cache */
   /* replace222 -> 253 in the cache */

   /* ('fdot3', ('vec3', 'a', 'b', 0.0), 'c') => ('fdot2', ('vec2', 'a', 'b'), 'c') */
   /* search223_0_0 -> 128 in the cache */
   /* search223_0_1 -> 129 in the cache */
   /* search223_0_2 -> 743 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 128, 129, 743 },
      -1,
   } },
   /* search223_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 754, 131 },
      -1,
   } },

   /* replace223_0_0 -> 128 in the cache */
   /* replace223_0_1 -> 129 in the cache */
   /* replace223_0 -> 746 in the cache */
   /* replace223_1 -> 131 in the cache */
   /* replace223 -> 747 in the cache */

   /* ('fdot2', ('vec2', 'a', 0.0), 'b') => ('fmul', 'a', 'b') */
   /* search224_0_0 -> 0 in the cache */
   /* search224_0_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 0, 245 },
      -1,
   } },
   /* search224_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 756, 3 },
      -1,
   } },

   /* replace224_0 -> 0 in the cache */
   /* replace224_1 -> 3 in the cache */
   /* replace224 -> 253 in the cache */

   /* ('fdot2', 'a', 1.0) => ('fadd', 'a.x', 'a.y') */
   /* search225_0 -> 32 in the cache */
   /* search225_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 32, 230 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      -1, 0,
      { 759, 760 },
      -1,
   } },

   /* ('fdot2', 'a', 'b') => ('fsum2', ('fmul', 'a', 'b')) */
   /* search226_0 -> 0 in the cache */
   /* search226_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace226_0_0 -> 0 in the cache */
   /* replace226_0_1 -> 3 in the cache */
   /* replace226_0 -> 253 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum2,
      -1, 1,
      { 253 },
      -1,
   } },

   /* ('fdot3', 'a', 'b') => ('fsum3', ('fmul', 'a', 'b')) */
   /* search227_0 -> 0 in the cache */
   /* search227_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace227_0_0 -> 0 in the cache */
   /* replace227_0_1 -> 3 in the cache */
   /* replace227_0 -> 253 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum3,
      -1, 1,
      { 253 },
      -1,
   } },

   /* ('fdot4', 'a', 'b') => ('fsum4', ('fmul', 'a', 'b')) */
   /* search228_0 -> 0 in the cache */
   /* search228_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace228_0_0 -> 0 in the cache */
   /* replace228_0_1 -> 3 in the cache */
   /* replace228_0 -> 253 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsum4,
      -1, 1,
      { 253 },
      -1,
   } },

   /* ('fsum2', 'a') => ('fadd', 'a.x', 'a.y') */
   /* search229_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsum2,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace229_0 -> 759 in the cache */
   /* replace229_1 -> 760 in the cache */
   /* replace229 -> 761 in the cache */

   /* ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a'))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      8,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 769 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 770 },
      0,
   } },
   /* search230_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 771, 230 },
      -1,
   } },

   /* replace230_0_0 -> 230 in the cache */
   /* replace230_0_1_0 -> 32 in the cache */
   /* replace230_0_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 230, 98 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 773 },
      -1,
   } },

   /* ('ishl', ('iadd', ('imul', 'a', '#b'), '#c'), '#d') => ('iadd', ('imul', 'a', ('ishl', 'b', 'd')), ('ishl', 'c', 'd')) */
   /* search231_0_0_0 -> 128 in the cache */
   /* search231_0_0_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 128, 272 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 775, 776 },
      -1,
   } },
   /* search231_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 777, 300 },
      -1,
   } },

   /* replace231_0_0 -> 128 in the cache */
   /* replace231_0_1_0 -> 129 in the cache */
   /* replace231_0_1_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 129, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 128, 779 },
      -1,
   } },
   /* replace231_1_0 -> 131 in the cache */
   /* replace231_1_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 131, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 780, 781 },
      -1,
   } },

   /* ('ishl', ('imul', 'a', '#b'), '#c') => ('imul', 'a', ('ishl', 'b', 'c')) */
   /* search232_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 783 },
      -1,
   } },
   /* search232_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 784, 298 },
      -1,
   } },

   /* replace232_0 -> 0 in the cache */
   /* replace232_1_0 -> 3 in the cache */
   /* replace232_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3, 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 786 },
      -1,
   } },

   /* ('ishl@8', ('ishl@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ishl@8', 'a', ('iadd', 'b', 'c')), 0) */
   /* search233_0_0 -> 6 in the cache */
   /* search233_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search233_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 788, 298 },
      -1,
   } },

   /* replace233_0_0_0_0 -> 36 in the cache */
   /* replace233_0_0_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 9 },
      -1,
   } },
   /* replace233_0_0_1_0 -> 137 in the cache */
   /* replace233_0_0_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 137, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 790, 791 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 792, 793 },
      -1,
   } },
   /* replace233_1_0 -> 6 in the cache */
   /* replace233_1_1_0 -> 36 in the cache */
   /* replace233_1_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 36, 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 6, 795 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 794, 796, 797 },
      -1,
   } },

   /* ('ushr@8', ('ushr@8', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 8), ('ushr@8', 'a', ('iadd', 'b', 'c')), 0) */
   /* search234_0_0 -> 6 in the cache */
   /* search234_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search234_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 799, 298 },
      -1,
   } },

   /* replace234_0_0_0_0 -> 36 in the cache */
   /* replace234_0_0_0_1 -> 9 in the cache */
   /* replace234_0_0_0 -> 790 in the cache */
   /* replace234_0_0_1_0 -> 137 in the cache */
   /* replace234_0_0_1_1 -> 9 in the cache */
   /* replace234_0_0_1 -> 791 in the cache */
   /* replace234_0_0 -> 792 in the cache */
   /* replace234_0_1 -> 793 in the cache */
   /* replace234_0 -> 794 in the cache */
   /* replace234_1_0 -> 6 in the cache */
   /* replace234_1_1_0 -> 36 in the cache */
   /* replace234_1_1_1 -> 137 in the cache */
   /* replace234_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 795 },
      -1,
   } },
   /* replace234_2 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 794, 801, 797 },
      -1,
   } },

   /* ('ishr@8', ('ishr@8', 'a', '#b'), '#c') => ('ishr@8', 'a', ('imin', ('iadd', ('iand', 'b', 7), ('iand', 'c', 7)), 7)) */
   /* search235_0_0 -> 6 in the cache */
   /* search235_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 33 },
      -1,
   } },
   /* search235_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 803, 298 },
      -1,
   } },

   /* replace235_0 -> 6 in the cache */
   /* replace235_1_0_0_0 -> 36 in the cache */
   /* replace235_1_0_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 9 },
      -1,
   } },
   /* replace235_1_0_1_0 -> 137 in the cache */
   /* replace235_1_0_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 137, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 805, 806 },
      -1,
   } },
   /* replace235_1_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 807, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 6, 808 },
      -1,
   } },

   /* ('ishl@16', ('ishl@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ishl@16', 'a', ('iadd', 'b', 'c')), 0) */
   /* search236_0_0 -> 11 in the cache */
   /* search236_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search236_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 810, 298 },
      -1,
   } },

   /* replace236_0_0_0_0 -> 36 in the cache */
   /* replace236_0_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 14 },
      -1,
   } },
   /* replace236_0_0_1_0 -> 137 in the cache */
   /* replace236_0_0_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 137, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 812, 813 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 814, 815 },
      -1,
   } },
   /* replace236_1_0 -> 11 in the cache */
   /* replace236_1_1_0 -> 36 in the cache */
   /* replace236_1_1_1 -> 137 in the cache */
   /* replace236_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 795 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 816, 817, 818 },
      -1,
   } },

   /* ('ushr@16', ('ushr@16', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 16), ('ushr@16', 'a', ('iadd', 'b', 'c')), 0) */
   /* search237_0_0 -> 11 in the cache */
   /* search237_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search237_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 820, 298 },
      -1,
   } },

   /* replace237_0_0_0_0 -> 36 in the cache */
   /* replace237_0_0_0_1 -> 14 in the cache */
   /* replace237_0_0_0 -> 812 in the cache */
   /* replace237_0_0_1_0 -> 137 in the cache */
   /* replace237_0_0_1_1 -> 14 in the cache */
   /* replace237_0_0_1 -> 813 in the cache */
   /* replace237_0_0 -> 814 in the cache */
   /* replace237_0_1 -> 815 in the cache */
   /* replace237_0 -> 816 in the cache */
   /* replace237_1_0 -> 11 in the cache */
   /* replace237_1_1_0 -> 36 in the cache */
   /* replace237_1_1_1 -> 137 in the cache */
   /* replace237_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 795 },
      -1,
   } },
   /* replace237_2 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 816, 822, 818 },
      -1,
   } },

   /* ('ishr@16', ('ishr@16', 'a', '#b'), '#c') => ('ishr@16', 'a', ('imin', ('iadd', ('iand', 'b', 15), ('iand', 'c', 15)), 15)) */
   /* search238_0_0 -> 11 in the cache */
   /* search238_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 33 },
      -1,
   } },
   /* search238_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 824, 298 },
      -1,
   } },

   /* replace238_0 -> 11 in the cache */
   /* replace238_1_0_0_0 -> 36 in the cache */
   /* replace238_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 14 },
      -1,
   } },
   /* replace238_1_0_1_0 -> 137 in the cache */
   /* replace238_1_0_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 137, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 826, 827 },
      -1,
   } },
   /* replace238_1_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 828, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 11, 829 },
      -1,
   } },

   /* ('ishl@32', ('ishl@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ishl@32', 'a', ('iadd', 'b', 'c')), 0) */
   /* search239_0_0 -> 16 in the cache */
   /* search239_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search239_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 831, 298 },
      -1,
   } },

   /* replace239_0_0_0_0 -> 36 in the cache */
   /* replace239_0_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 19 },
      -1,
   } },
   /* replace239_0_0_1_0 -> 137 in the cache */
   /* replace239_0_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 137, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 833, 834 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x20 /* 32 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 835, 836 },
      -1,
   } },
   /* replace239_1_0 -> 16 in the cache */
   /* replace239_1_1_0 -> 36 in the cache */
   /* replace239_1_1_1 -> 137 in the cache */
   /* replace239_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 795 },
      -1,
   } },
   /* replace239_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 837, 838, 124 },
      -1,
   } },

   /* ('ushr@32', ('ushr@32', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 32), ('ushr@32', 'a', ('iadd', 'b', 'c')), 0) */
   /* search240_0_0 -> 16 in the cache */
   /* search240_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search240_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 840, 298 },
      -1,
   } },

   /* replace240_0_0_0_0 -> 36 in the cache */
   /* replace240_0_0_0_1 -> 19 in the cache */
   /* replace240_0_0_0 -> 833 in the cache */
   /* replace240_0_0_1_0 -> 137 in the cache */
   /* replace240_0_0_1_1 -> 19 in the cache */
   /* replace240_0_0_1 -> 834 in the cache */
   /* replace240_0_0 -> 835 in the cache */
   /* replace240_0_1 -> 836 in the cache */
   /* replace240_0 -> 837 in the cache */
   /* replace240_1_0 -> 16 in the cache */
   /* replace240_1_1_0 -> 36 in the cache */
   /* replace240_1_1_1 -> 137 in the cache */
   /* replace240_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 795 },
      -1,
   } },
   /* replace240_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 837, 842, 124 },
      -1,
   } },

   /* ('ishr@32', ('ishr@32', 'a', '#b'), '#c') => ('ishr@32', 'a', ('imin', ('iadd', ('iand', 'b', 31), ('iand', 'c', 31)), 31)) */
   /* search241_0_0 -> 16 in the cache */
   /* search241_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 33 },
      -1,
   } },
   /* search241_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 844, 298 },
      -1,
   } },

   /* replace241_0 -> 16 in the cache */
   /* replace241_1_0_0_0 -> 36 in the cache */
   /* replace241_1_0_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 19 },
      -1,
   } },
   /* replace241_1_0_1_0 -> 137 in the cache */
   /* replace241_1_0_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 137, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 846, 847 },
      -1,
   } },
   /* replace241_1_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 848, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 16, 849 },
      -1,
   } },

   /* ('ishl@64', ('ishl@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ishl@64', 'a', ('iadd', 'b', 'c')), 0) */
   /* search242_0_0 -> 21 in the cache */
   /* search242_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search242_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 851, 298 },
      -1,
   } },

   /* replace242_0_0_0_0 -> 36 in the cache */
   /* replace242_0_0_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 36, 24 },
      -1,
   } },
   /* replace242_0_0_1_0 -> 137 in the cache */
   /* replace242_0_0_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 137, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 853, 854 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x40 /* 64 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 3,
      { 855, 856 },
      -1,
   } },
   /* replace242_1_0 -> 21 in the cache */
   /* replace242_1_1_0 -> 36 in the cache */
   /* replace242_1_1_1 -> 137 in the cache */
   /* replace242_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 21, 795 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 857, 858, 859 },
      -1,
   } },

   /* ('ushr@64', ('ushr@64', 'a', '#b'), '#c') => ('bcsel', ('ult', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 64), ('ushr@64', 'a', ('iadd', 'b', 'c')), 0) */
   /* search243_0_0 -> 21 in the cache */
   /* search243_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search243_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 861, 298 },
      -1,
   } },

   /* replace243_0_0_0_0 -> 36 in the cache */
   /* replace243_0_0_0_1 -> 24 in the cache */
   /* replace243_0_0_0 -> 853 in the cache */
   /* replace243_0_0_1_0 -> 137 in the cache */
   /* replace243_0_0_1_1 -> 24 in the cache */
   /* replace243_0_0_1 -> 854 in the cache */
   /* replace243_0_0 -> 855 in the cache */
   /* replace243_0_1 -> 856 in the cache */
   /* replace243_0 -> 857 in the cache */
   /* replace243_1_0 -> 21 in the cache */
   /* replace243_1_1_0 -> 36 in the cache */
   /* replace243_1_1_1 -> 137 in the cache */
   /* replace243_1_1 -> 795 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 795 },
      -1,
   } },
   /* replace243_2 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 857, 863, 859 },
      -1,
   } },

   /* ('ishr@64', ('ishr@64', 'a', '#b'), '#c') => ('ishr@64', 'a', ('imin', ('iadd', ('iand', 'b', 63), ('iand', 'c', 63)), 63)) */
   /* search244_0_0 -> 21 in the cache */
   /* search244_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 33 },
      -1,
   } },
   /* search244_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 865, 298 },
      -1,
   } },

   /* replace244_0 -> 21 in the cache */
   /* replace244_1_0_0_0 -> 36 in the cache */
   /* replace244_1_0_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 24 },
      -1,
   } },
   /* replace244_1_0_1_0 -> 137 in the cache */
   /* replace244_1_0_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 137, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 3,
      { 867, 868 },
      -1,
   } },
   /* replace244_1_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 869, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 4,
      { 21, 870 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b')) */
   /* search245_0_0 -> 6 in the cache */
   /* search245_0_1 -> 33 in the cache */
   /* search245_0 -> 799 in the cache */
   /* search245_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 799, 36 },
      -1,
   } },

   /* replace245_0 -> 6 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   /* replace245_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 873, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 874 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@8', '#b'), 'b') => ('iand', 'a', ('ishl', 255, 'b')) */
   /* search246_0_0 -> 6 in the cache */
   /* search246_0_1 -> 33 in the cache */
   /* search246_0 -> 803 in the cache */
   /* search246_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 803, 36 },
      -1,
   } },

   /* replace246_0 -> 6 in the cache */
   /* replace246_1_0 -> 873 in the cache */
   /* replace246_1_1 -> 36 in the cache */
   /* replace246_1 -> 874 in the cache */
   /* replace246 -> 875 in the cache */

   /* ('ushr', ('ishl', 'a@8', '#b'), 'b') => ('iand', 'a', ('ushr', 255, 'b')) */
   /* search247_0_0 -> 6 in the cache */
   /* search247_0_1 -> 33 in the cache */
   /* search247_0 -> 788 in the cache */
   /* search247_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 788, 36 },
      -1,
   } },

   /* replace247_0 -> 6 in the cache */
   /* replace247_1_0 -> 873 in the cache */
   /* replace247_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 873, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 878 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b')) */
   /* search248_0_0 -> 11 in the cache */
   /* search248_0_1 -> 33 in the cache */
   /* search248_0 -> 820 in the cache */
   /* search248_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 820, 36 },
      -1,
   } },

   /* replace248_0 -> 11 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* replace248_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 881, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 882 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@16', '#b'), 'b') => ('iand', 'a', ('ishl', 65535, 'b')) */
   /* search249_0_0 -> 11 in the cache */
   /* search249_0_1 -> 33 in the cache */
   /* search249_0 -> 824 in the cache */
   /* search249_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 824, 36 },
      -1,
   } },

   /* replace249_0 -> 11 in the cache */
   /* replace249_1_0 -> 881 in the cache */
   /* replace249_1_1 -> 36 in the cache */
   /* replace249_1 -> 882 in the cache */
   /* replace249 -> 883 in the cache */

   /* ('ushr', ('ishl', 'a@16', '#b'), 'b') => ('iand', 'a', ('ushr', 65535, 'b')) */
   /* search250_0_0 -> 11 in the cache */
   /* search250_0_1 -> 33 in the cache */
   /* search250_0 -> 810 in the cache */
   /* search250_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 810, 36 },
      -1,
   } },

   /* replace250_0 -> 11 in the cache */
   /* replace250_1_0 -> 881 in the cache */
   /* replace250_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 881, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 886 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b')) */
   /* search251_0_0 -> 16 in the cache */
   /* search251_0_1 -> 33 in the cache */
   /* search251_0 -> 840 in the cache */
   /* search251_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 840, 36 },
      -1,
   } },

   /* replace251_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   /* replace251_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 889, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 890 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@32', '#b'), 'b') => ('iand', 'a', ('ishl', 4294967295, 'b')) */
   /* search252_0_0 -> 16 in the cache */
   /* search252_0_1 -> 33 in the cache */
   /* search252_0 -> 844 in the cache */
   /* search252_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 844, 36 },
      -1,
   } },

   /* replace252_0 -> 16 in the cache */
   /* replace252_1_0 -> 889 in the cache */
   /* replace252_1_1 -> 36 in the cache */
   /* replace252_1 -> 890 in the cache */
   /* replace252 -> 891 in the cache */

   /* ('ushr', ('ishl', 'a@32', '#b'), 'b') => ('iand', 'a', ('ushr', 4294967295, 'b')) */
   /* search253_0_0 -> 16 in the cache */
   /* search253_0_1 -> 33 in the cache */
   /* search253_0 -> 831 in the cache */
   /* search253_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 831, 36 },
      -1,
   } },

   /* replace253_0 -> 16 in the cache */
   /* replace253_1_0 -> 889 in the cache */
   /* replace253_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 889, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 894 },
      -1,
   } },

   /* ('ishl', ('ushr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b')) */
   /* search254_0_0 -> 21 in the cache */
   /* search254_0_1 -> 33 in the cache */
   /* search254_0 -> 861 in the cache */
   /* search254_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 861, 36 },
      -1,
   } },

   /* replace254_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffffffffffff /* 18446744073709551615 */ },
   } },
   /* replace254_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 897, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 898 },
      -1,
   } },

   /* ('ishl', ('ishr', 'a@64', '#b'), 'b') => ('iand', 'a', ('ishl', 18446744073709551615, 'b')) */
   /* search255_0_0 -> 21 in the cache */
   /* search255_0_1 -> 33 in the cache */
   /* search255_0 -> 865 in the cache */
   /* search255_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 865, 36 },
      -1,
   } },

   /* replace255_0 -> 21 in the cache */
   /* replace255_1_0 -> 897 in the cache */
   /* replace255_1_1 -> 36 in the cache */
   /* replace255_1 -> 898 in the cache */
   /* replace255 -> 899 in the cache */

   /* ('ushr', ('ishl', 'a@64', '#b'), 'b') => ('iand', 'a', ('ushr', 18446744073709551615, 'b')) */
   /* search256_0_0 -> 21 in the cache */
   /* search256_0_1 -> 33 in the cache */
   /* search256_0 -> 851 in the cache */
   /* search256_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 851, 36 },
      -1,
   } },

   /* replace256_0 -> 21 in the cache */
   /* replace256_1_0 -> 897 in the cache */
   /* replace256_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 897, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 902 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', '#b(is_first_5_bits_uge_2)'), -4) => ('ishl', 'a', 'b') */
   /* search257_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      9,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 904 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x4 /* -4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 905, 906 },
      -1,
   } },

   /* replace257_0 -> 16 in the cache */
   /* replace257_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('iand', ('imul', 'a', '#b(is_unsigned_multiple_of_4)'), -4) => ('imul', 'a', 'b') */
   /* search258_0_0 -> 0 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      10,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 0, 909 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { -0x4 /* -4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 910, 911 },
      -1,
   } },

   /* replace258_0 -> 0 in the cache */
   /* replace258_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_2)'), 4294967294) => ('iadd', ('iand', 'a', 4294967294), 'b') */
   /* search259_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      11,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 914 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffffe /* 4294967294 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 915, 916 },
      -1,
   } },

   /* replace259_0_0 -> 16 in the cache */
   /* replace259_0_1 -> 916 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 916 },
      -1,
   } },
   /* replace259_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 918, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_4)'), 4294967292) => ('iadd', ('iand', 'a', 4294967292), 'b') */
   /* search260_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      10,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 920 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffffc /* 4294967292 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 921, 922 },
      -1,
   } },

   /* replace260_0_0 -> 16 in the cache */
   /* replace260_0_1 -> 922 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 922 },
      -1,
   } },
   /* replace260_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 924, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_8)'), 4294967288) => ('iadd', ('iand', 'a', 4294967288), 'b') */
   /* search261_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      12,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 926 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffff8 /* 4294967288 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 927, 928 },
      -1,
   } },

   /* replace261_0_0 -> 16 in the cache */
   /* replace261_0_1 -> 928 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 928 },
      -1,
   } },
   /* replace261_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 930, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_16)'), 4294967280) => ('iadd', ('iand', 'a', 4294967280), 'b') */
   /* search262_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      13,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 932 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfffffff0 /* 4294967280 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 933, 934 },
      -1,
   } },

   /* replace262_0_0 -> 16 in the cache */
   /* replace262_0_1 -> 934 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 934 },
      -1,
   } },
   /* replace262_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 936, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_32)'), 4294967264) => ('iadd', ('iand', 'a', 4294967264), 'b') */
   /* search263_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      14,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 938 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffe0 /* 4294967264 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 939, 940 },
      -1,
   } },

   /* replace263_0_0 -> 16 in the cache */
   /* replace263_0_1 -> 940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 940 },
      -1,
   } },
   /* replace263_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 942, 36 },
      -1,
   } },

   /* ('iand@32', ('iadd@32', 'a', '#b(is_unsigned_multiple_of_64)'), 4294967232) => ('iadd', ('iand', 'a', 4294967232), 'b') */
   /* search264_0_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      15,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 944 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffc0 /* 4294967232 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 945, 946 },
      -1,
   } },

   /* replace264_0_0 -> 16 in the cache */
   /* replace264_0_1 -> 946 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 946 },
      -1,
   } },
   /* replace264_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 948, 36 },
      -1,
   } },

   /* ('ishl@32', ('iand', 'a@32', 2147483647), 1) => ('ishl', 'a', 1) */
   /* search265_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fffffff /* 2147483647 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 950 },
      -1,
   } },
   /* search265_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 951, 381 },
      -1,
   } },

   /* replace265_0 -> 16 in the cache */
   /* replace265_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 381 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 1), 4294967294) => ('ishl', 'a', 1) */
   /* search266_0_0 -> 16 in the cache */
   /* search266_0_1 -> 381 in the cache */
   /* search266_0 -> 953 in the cache */
   /* search266_1 -> 916 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 953, 916 },
      -1,
   } },

   /* replace266_0 -> 16 in the cache */
   /* replace266_1 -> 381 in the cache */
   /* replace266 -> 953 in the cache */

   /* ('iand', ('ushr', 'a@32', 1), 2147483647) => ('ushr', 'a', 1) */
   /* search267_0_0 -> 16 in the cache */
   /* search267_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 381 },
      -1,
   } },
   /* search267_1 -> 950 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 955, 950 },
      -1,
   } },

   /* replace267_0 -> 16 in the cache */
   /* replace267_1 -> 381 in the cache */
   /* replace267 -> 955 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 1073741823), 2) => ('ishl', 'a', 2) */
   /* search268_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3fffffff /* 1073741823 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 957 },
      -1,
   } },
   /* search268_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 958, 386 },
      -1,
   } },

   /* replace268_0 -> 16 in the cache */
   /* replace268_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 386 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 2), 4294967292) => ('ishl', 'a', 2) */
   /* search269_0_0 -> 16 in the cache */
   /* search269_0_1 -> 386 in the cache */
   /* search269_0 -> 960 in the cache */
   /* search269_1 -> 922 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 960, 922 },
      -1,
   } },

   /* replace269_0 -> 16 in the cache */
   /* replace269_1 -> 386 in the cache */
   /* replace269 -> 960 in the cache */

   /* ('iand', ('ushr', 'a@32', 2), 1073741823) => ('ushr', 'a', 2) */
   /* search270_0_0 -> 16 in the cache */
   /* search270_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 386 },
      -1,
   } },
   /* search270_1 -> 957 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 962, 957 },
      -1,
   } },

   /* replace270_0 -> 16 in the cache */
   /* replace270_1 -> 386 in the cache */
   /* replace270 -> 962 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 65535), 16) => ('ishl', 'a', 16) */
   /* search271_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 964 },
      -1,
   } },
   /* search271_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 965, 815 },
      -1,
   } },

   /* replace271_0 -> 16 in the cache */
   /* replace271_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 815 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 16), 4294901760) => ('ishl', 'a', 16) */
   /* search272_0_0 -> 16 in the cache */
   /* search272_0_1 -> 815 in the cache */
   /* search272_0 -> 967 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffff0000 /* 4294901760 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 967, 968 },
      -1,
   } },

   /* replace272_0 -> 16 in the cache */
   /* replace272_1 -> 815 in the cache */
   /* replace272 -> 967 in the cache */

   /* ('iand', ('ushr', 'a@32', 16), 65535) => ('ushr', 'a', 16) */
   /* search273_0_0 -> 16 in the cache */
   /* search273_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 815 },
      -1,
   } },
   /* search273_1 -> 964 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 970, 964 },
      -1,
   } },

   /* replace273_0 -> 16 in the cache */
   /* replace273_1 -> 815 in the cache */
   /* replace273 -> 970 in the cache */

   /* ('ishl@32', ('iand', 'a@32', 255), 24) => ('ishl', 'a', 24) */
   /* search274_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 972 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x18 /* 24 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 973, 974 },
      -1,
   } },

   /* replace274_0 -> 16 in the cache */
   /* replace274_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 974 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 24), 4278190080) => ('ishl', 'a', 24) */
   /* search275_0_0 -> 16 in the cache */
   /* search275_0_1 -> 974 in the cache */
   /* search275_0 -> 976 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff000000 /* 4278190080 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 976, 977 },
      -1,
   } },

   /* replace275_0 -> 16 in the cache */
   /* replace275_1 -> 974 in the cache */
   /* replace275 -> 976 in the cache */

   /* ('iand', ('ushr', 'a@32', 24), 255) => ('ushr', 'a', 24) */
   /* search276_0_0 -> 16 in the cache */
   /* search276_0_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 974 },
      -1,
   } },
   /* search276_1 -> 972 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 979, 972 },
      -1,
   } },

   /* replace276_0 -> 16 in the cache */
   /* replace276_1 -> 974 in the cache */
   /* replace276 -> 979 in the cache */

   /* ('ishl', ('iadd', 'a', '#b'), '#c') => ('iadd', ('ishl', 'a', 'c'), ('ishl', 'b', 'c')) */
   /* search277_0_0 -> 0 in the cache */
   /* search277_0_1 -> 783 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 783 },
      -1,
   } },
   /* search277_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 981, 298 },
      -1,
   } },

   /* replace277_0_0 -> 0 in the cache */
   /* replace277_0_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 0, 137 },
      -1,
   } },
   /* replace277_1_0 -> 3 in the cache */
   /* replace277_1_1 -> 137 in the cache */
   /* replace277_1 -> 786 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 983, 786 },
      -1,
   } },

   /* ('imul', ('iadd(is_used_once)', 'a', '#b'), '#c') => ('iadd', ('imul', 'a', 'c'), ('imul', 'b', 'c')) */
   /* search278_0_0 -> 128 in the cache */
   /* search278_0_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 128, 272 },
      0,
   } },
   /* search278_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 985, 776 },
      -1,
   } },

   /* replace278_0_0 -> 128 in the cache */
   /* replace278_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace278_1_0 -> 129 in the cache */
   /* replace278_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 987, 988 },
      -1,
   } },

   /* ('imul', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('imul', ('iadd', 'a', 'c'), 'd'), ('imul', 'b', 'd')) */
   /* search279_0_0_0 -> 151 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 151, 990 },
      0,
   } },
   /* search279_0_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 2,
      { 991, 154 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 3,
      { 992, 993 },
      -1,
   } },

   /* replace279_0_0_0 -> 151 in the cache */
   /* replace279_0_0_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 151, 154 },
      -1,
   } },
   /* replace279_0_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 2,
      { 995, 155 },
      -1,
   } },
   /* replace279_1_0 -> 152 in the cache */
   /* replace279_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      3, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 996, 997 },
      -1,
   } },

   /* ('ishl', ('iadd(is_used_once)', ('iadd(is_used_once)', 'a', '#b'), 'c'), '#d') => ('iadd', ('ishl', ('iadd', 'a', 'c'), 'd'), ('ishl', 'b', 'd')) */
   /* search280_0_0_0 -> 128 in the cache */
   /* search280_0_0_1 -> 272 in the cache */
   /* search280_0_0 -> 985 in the cache */
   /* search280_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 985, 131 },
      0,
   } },
   /* search280_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 999, 300 },
      -1,
   } },

   /* replace280_0_0_0 -> 128 in the cache */
   /* replace280_0_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace280_0_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 1001, 165 },
      -1,
   } },
   /* replace280_1_0 -> 129 in the cache */
   /* replace280_1_1 -> 165 in the cache */
   /* replace280_1 -> 779 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1002, 779 },
      -1,
   } },

   /* ('inot', ('flt(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('fge', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1004, 1005 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1006 },
      -1,
   } },

   /* replace281_0 -> 0 in the cache */
   /* replace281_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('fge(is_used_once)', 'a(is_a_number)', 'b(is_a_number)')) => ('flt', 'a', 'b') */
   /* search282_0_0 -> 1004 in the cache */
   /* search282_0_1 -> 1005 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1004, 1005 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1009 },
      -1,
   } },

   /* replace282_0 -> 0 in the cache */
   /* replace282_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('feq(is_used_once)', 'a', 'b')) => ('fneu', 'a', 'b') */
   /* search283_0_0 -> 0 in the cache */
   /* search283_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1012 },
      -1,
   } },

   /* replace283_0 -> 0 in the cache */
   /* replace283_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('fneu(is_used_once)', 'a', 'b')) => ('feq', 'a', 'b') */
   /* search284_0_0 -> 0 in the cache */
   /* search284_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1015 },
      -1,
   } },

   /* replace284_0 -> 0 in the cache */
   /* replace284_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ilt(is_used_once)', 'a', 'b')) => ('ige', 'a', 'b') */
   /* search285_0_0 -> 0 in the cache */
   /* search285_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1018 },
      -1,
   } },

   /* replace285_0 -> 0 in the cache */
   /* replace285_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ult(is_used_once)', 'a', 'b')) => ('uge', 'a', 'b') */
   /* search286_0_0 -> 0 in the cache */
   /* search286_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1021 },
      -1,
   } },

   /* replace286_0 -> 0 in the cache */
   /* replace286_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ige(is_used_once)', 'a', 'b')) => ('ilt', 'a', 'b') */
   /* search287_0_0 -> 0 in the cache */
   /* search287_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1024 },
      -1,
   } },

   /* replace287_0 -> 0 in the cache */
   /* replace287_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('uge(is_used_once)', 'a', 'b')) => ('ult', 'a', 'b') */
   /* search288_0_0 -> 0 in the cache */
   /* search288_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1027 },
      -1,
   } },

   /* replace288_0 -> 0 in the cache */
   /* replace288_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ieq(is_used_once)', 'a', 'b')) => ('ine', 'a', 'b') */
   /* search289_0_0 -> 0 in the cache */
   /* search289_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1030 },
      -1,
   } },

   /* replace289_0 -> 0 in the cache */
   /* replace289_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('inot', ('ine(is_used_once)', 'a', 'b')) => ('ieq', 'a', 'b') */
   /* search290_0_0 -> 0 in the cache */
   /* search290_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1033 },
      -1,
   } },

   /* replace290_0 -> 0 in the cache */
   /* replace290_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'b'), ('fneu', 'a', 'b')) => False */
   /* search291_0_0 -> 0 in the cache */
   /* search291_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search291_1_0 -> 0 in the cache */
   /* search291_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1036, 1037 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_bool, { NIR_FALSE /* False */ },
   } },

   /* ('iand', ('flt', 'a', 'b'), ('flt', 'b', 'a')) => False */
   /* search292_0_0 -> 0 in the cache */
   /* search292_0_1 -> 3 in the cache */
   /* search292_0 -> 1011 in the cache */
   /* search292_1_0 -> 3 in the cache */
   /* search292_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1011, 1040 },
      -1,
   } },

   /* replace292 -> 1039 in the cache */

   /* ('iand', ('ieq', 'a', 'b'), ('ine', 'a', 'b')) => False */
   /* search293_0_0 -> 0 in the cache */
   /* search293_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search293_1_0 -> 0 in the cache */
   /* search293_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1042, 1043 },
      -1,
   } },

   /* replace293 -> 1039 in the cache */

   /* ('iand', ('ilt', 'a', 'b'), ('ilt', 'b', 'a')) => False */
   /* search294_0_0 -> 0 in the cache */
   /* search294_0_1 -> 3 in the cache */
   /* search294_0 -> 1026 in the cache */
   /* search294_1_0 -> 3 in the cache */
   /* search294_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1026, 1045 },
      -1,
   } },

   /* replace294 -> 1039 in the cache */

   /* ('iand', ('ult', 'a', 'b'), ('ult', 'b', 'a')) => False */
   /* search295_0_0 -> 0 in the cache */
   /* search295_0_1 -> 3 in the cache */
   /* search295_0 -> 1029 in the cache */
   /* search295_1_0 -> 3 in the cache */
   /* search295_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1029, 1047 },
      -1,
   } },

   /* replace295 -> 1039 in the cache */

   /* ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a') */
   /* search296_0_0 -> 0 in the cache */
   /* search296_0 -> 201 in the cache */
   /* search296_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 201, 1049 },
      -1,
   } },

   /* replace296_0 -> 3 in the cache */
   /* replace296_1 -> 0 in the cache */
   /* replace296 -> 1040 in the cache */

   /* ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a') */
   /* search297_0_0 -> 0 in the cache */
   /* search297_0 -> 201 in the cache */
   /* search297_1_0 -> 3 in the cache */
   /* search297_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 201, 1049 },
      -1,
   } },

   /* replace297_0 -> 3 in the cache */
   /* replace297_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a') */
   /* search298_0_0 -> 0 in the cache */
   /* search298_0 -> 201 in the cache */
   /* search298_1_0 -> 3 in the cache */
   /* search298_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 201, 1049 },
      -1,
   } },

   /* replace298_0 -> 3 in the cache */
   /* replace298_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3, 0 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a') */
   /* search299_0_0 -> 0 in the cache */
   /* search299_0 -> 201 in the cache */
   /* search299_1_0 -> 3 in the cache */
   /* search299_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 201, 1049 },
      -1,
   } },

   /* replace299_0 -> 3 in the cache */
   /* replace299_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3, 0 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a') */
   /* search300_0_0 -> 32 in the cache */
   /* search300_0 -> 98 in the cache */
   /* search300_1 -> 236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 98, 236 },
      -1,
   } },

   /* replace300_0 -> 230 in the cache */
   /* replace300_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 230, 32 },
      -1,
   } },

   /* ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0) */
   /* search301_0 -> 236 in the cache */
   /* search301_1_0 -> 32 in the cache */
   /* search301_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 236, 98 },
      -1,
   } },

   /* replace301_0 -> 32 in the cache */
   /* replace301_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 230 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a') */
   /* search302_0_0 -> 32 in the cache */
   /* search302_0 -> 98 in the cache */
   /* search302_1 -> 236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 98, 236 },
      -1,
   } },

   /* replace302_0 -> 230 in the cache */
   /* replace302_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 230, 32 },
      -1,
   } },

   /* ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0) */
   /* search303_0 -> 236 in the cache */
   /* search303_1_0 -> 32 in the cache */
   /* search303_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 236, 98 },
      -1,
   } },

   /* replace303_0 -> 32 in the cache */
   /* replace303_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 230 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a') */
   /* search304_0_0 -> 32 in the cache */
   /* search304_0 -> 98 in the cache */
   /* search304_1 -> 236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 98, 236 },
      -1,
   } },

   /* replace304_0 -> 230 in the cache */
   /* replace304_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 230, 32 },
      -1,
   } },

   /* ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0) */
   /* search305_0 -> 236 in the cache */
   /* search305_1_0 -> 32 in the cache */
   /* search305_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 236, 98 },
      -1,
   } },

   /* replace305_0 -> 32 in the cache */
   /* replace305_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 230 },
      -1,
   } },

   /* ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      true,
      nir_type_invalid,
      17,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 1070 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1069, 1071 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace306_1 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1073, 1070 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b') */
   /* search307_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      17,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1075, 1076 },
      -1,
   } },

   /* replace307_0 -> 0 in the cache */
   /* replace307_1 -> 3 in the cache */
   /* replace307 -> 1008 in the cache */

   /* ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b') */
   /* search308_0_0 -> 0 in the cache */
   /* search308_0 -> 1075 in the cache */
   /* search308_1 -> 1076 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1075, 1076 },
      -1,
   } },

   /* replace308_0 -> 0 in the cache */
   /* replace308_1 -> 3 in the cache */
   /* replace308 -> 1017 in the cache */

   /* ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b') */
   /* search309_0_0 -> 0 in the cache */
   /* search309_0 -> 1075 in the cache */
   /* search309_1 -> 1076 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1075, 1076 },
      -1,
   } },

   /* replace309_0 -> 0 in the cache */
   /* replace309_1 -> 3 in the cache */
   /* replace309 -> 1014 in the cache */

   /* ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0) */
   /* search310_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 32 },
      0,
   } },
   /* search310_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1080, 230 },
      -1,
   } },

   /* replace310_0 -> 32 in the cache */
   /* replace310_1 -> 230 in the cache */
   /* replace310 -> 1064 in the cache */

   /* ('flt', 0.0, ('fsat(is_used_once)', 'a')) => ('flt', 0.0, 'a') */
   /* search311_0 -> 113 in the cache */
   /* search311_1_0 -> 32 in the cache */
   /* search311_1 -> 1080 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 113, 1080 },
      -1,
   } },

   /* replace311_0 -> 113 in the cache */
   /* replace311_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 113, 32 },
      -1,
   } },

   /* ('fge', 0.0, ('b2f', 'a@1')) => ('inot', 'a') */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search312_1_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 722 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1084, 1085 },
      -1,
   } },

   /* replace312_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('fge', ('fneg', ('b2f', 'a@1')), 0.0) => ('inot', 'a') */
   /* search313_0_0_0 -> 722 in the cache */
   /* search313_0_0 -> 1085 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1085 },
      -1,
   } },
   /* search313_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1088, 1084 },
      -1,
   } },

   /* replace313_0 -> 722 in the cache */
   /* replace313 -> 1087 in the cache */

   /* ('fneu', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b') */
   /* search314_0_0_0 -> 722 in the cache */
   /* search314_0_0 -> 1085 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1085, 1091 },
      -1,
   } },
   /* search314_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1092, 1084 },
      -1,
   } },

   /* replace314_0 -> 722 in the cache */
   /* replace314_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 722, 1090 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('ior', 'a', 'b') */
   /* search315_0_0 -> 722 in the cache */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* search315_0_2_0 -> 1090 in the cache */
   /* search315_0_2 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1095, 1091 },
      -1,
   } },
   /* search315_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1096, 1084 },
      -1,
   } },

   /* replace315_0 -> 722 in the cache */
   /* replace315_1 -> 1090 in the cache */
   /* replace315 -> 1094 in the cache */

   /* ('fneu', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('ior', 'a', 'b') */
   /* search316_0_0 -> 722 in the cache */
   /* search316_0 -> 1085 in the cache */
   /* search316_1_0_0 -> 1090 in the cache */
   /* search316_1_0 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1091 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1085, 1098 },
      -1,
   } },

   /* replace316_0 -> 722 in the cache */
   /* replace316_1 -> 1090 in the cache */
   /* replace316 -> 1094 in the cache */

   /* ('fneu', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('iand', 'a', 'b') */
   /* search317_0_0_0 -> 722 in the cache */
   /* search317_0_0 -> 1085 in the cache */
   /* search317_0_1_0 -> 1090 in the cache */
   /* search317_0_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 1085, 1091 },
      -1,
   } },
   /* search317_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1100, 1084 },
      -1,
   } },

   /* replace317_0 -> 722 in the cache */
   /* replace317_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 722, 1090 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('iand', 'a', 'b') */
   /* search318_0_0 -> 722 in the cache */
   /* search318_0_1_0 -> 1090 in the cache */
   /* search318_0_1 -> 1091 in the cache */
   /* search318_0_2 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1091, 1084 },
      -1,
   } },
   /* search318_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1103, 1084 },
      -1,
   } },

   /* replace318_0 -> 722 in the cache */
   /* replace318_1 -> 1090 in the cache */
   /* replace318 -> 1102 in the cache */

   /* ('fneu', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ixor', 'a', 'b') */
   /* search319_0_0_0 -> 722 in the cache */
   /* search319_0_0 -> 1085 in the cache */
   /* search319_0_1_0_0 -> 1090 in the cache */
   /* search319_0_1_0 -> 1091 in the cache */
   /* search319_0_1 -> 1098 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1085, 1098 },
      -1,
   } },
   /* search319_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1105, 1084 },
      -1,
   } },

   /* replace319_0 -> 722 in the cache */
   /* replace319_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 722, 1090 },
      -1,
   } },

   /* ('fneu', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ixor', 'a', 'b') */
   /* search320_0_0 -> 722 in the cache */
   /* search320_0 -> 1085 in the cache */
   /* search320_1_0 -> 1090 in the cache */
   /* search320_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1085, 1091 },
      -1,
   } },

   /* replace320_0 -> 722 in the cache */
   /* replace320_1 -> 1090 in the cache */
   /* replace320 -> 1107 in the cache */

   /* ('fneu', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ixor', 'a', 'b') */
   /* search321_0_0_0 -> 722 in the cache */
   /* search321_0_0 -> 1085 in the cache */
   /* search321_0 -> 1088 in the cache */
   /* search321_1_0_0 -> 1090 in the cache */
   /* search321_1_0 -> 1091 in the cache */
   /* search321_1 -> 1098 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1088, 1098 },
      -1,
   } },

   /* replace321_0 -> 722 in the cache */
   /* replace321_1 -> 1090 in the cache */
   /* replace321 -> 1107 in the cache */

   /* ('feq', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search322_0_0_0 -> 722 in the cache */
   /* search322_0_0 -> 1085 in the cache */
   /* search322_0_1_0 -> 1090 in the cache */
   /* search322_0_1 -> 1091 in the cache */
   /* search322_0 -> 1092 in the cache */
   /* search322_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1092, 1084 },
      -1,
   } },

   /* replace322_0_0 -> 722 in the cache */
   /* replace322_0_1 -> 1090 in the cache */
   /* replace322_0 -> 1094 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1094 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', 1.0, ('b2f', 'b@1')), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search323_0_0 -> 722 in the cache */
   /* search323_0_1 -> 1095 in the cache */
   /* search323_0_2_0 -> 1090 in the cache */
   /* search323_0_2 -> 1091 in the cache */
   /* search323_0 -> 1096 in the cache */
   /* search323_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1096, 1084 },
      -1,
   } },

   /* replace323_0_0 -> 722 in the cache */
   /* replace323_0_1 -> 1090 in the cache */
   /* replace323_0 -> 1094 in the cache */
   /* replace323 -> 1111 in the cache */

   /* ('feq', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b')) */
   /* search324_0_0 -> 722 in the cache */
   /* search324_0 -> 1085 in the cache */
   /* search324_1_0_0 -> 1090 in the cache */
   /* search324_1_0 -> 1091 in the cache */
   /* search324_1 -> 1098 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1085, 1098 },
      -1,
   } },

   /* replace324_0_0 -> 722 in the cache */
   /* replace324_0_1 -> 1090 in the cache */
   /* replace324_0 -> 1094 in the cache */
   /* replace324 -> 1111 in the cache */

   /* ('feq', ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')), 0.0) => ('inot', ('iand', 'a', 'b')) */
   /* search325_0_0_0 -> 722 in the cache */
   /* search325_0_0 -> 1085 in the cache */
   /* search325_0_1_0 -> 1090 in the cache */
   /* search325_0_1 -> 1091 in the cache */
   /* search325_0 -> 1100 in the cache */
   /* search325_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1100, 1084 },
      -1,
   } },

   /* replace325_0_0 -> 722 in the cache */
   /* replace325_0_1 -> 1090 in the cache */
   /* replace325_0 -> 1102 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 1102 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', ('b2f', 'b@1'), 0.0), 0.0) => ('inot', ('iand', 'a', 'b')) */
   /* search326_0_0 -> 722 in the cache */
   /* search326_0_1_0 -> 1090 in the cache */
   /* search326_0_1 -> 1091 in the cache */
   /* search326_0_2 -> 1084 in the cache */
   /* search326_0 -> 1103 in the cache */
   /* search326_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1103, 1084 },
      -1,
   } },

   /* replace326_0_0 -> 722 in the cache */
   /* replace326_0_1 -> 1090 in the cache */
   /* replace326_0 -> 1102 in the cache */
   /* replace326 -> 1115 in the cache */

   /* ('feq', ('fadd', ('b2f', 'a@1'), ('fneg', ('b2f', 'b@1'))), 0.0) => ('ieq', 'a', 'b') */
   /* search327_0_0_0 -> 722 in the cache */
   /* search327_0_0 -> 1085 in the cache */
   /* search327_0_1_0_0 -> 1090 in the cache */
   /* search327_0_1_0 -> 1091 in the cache */
   /* search327_0_1 -> 1098 in the cache */
   /* search327_0 -> 1105 in the cache */
   /* search327_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1105, 1084 },
      -1,
   } },

   /* replace327_0 -> 722 in the cache */
   /* replace327_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 722, 1090 },
      -1,
   } },

   /* ('feq', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('ieq', 'a', 'b') */
   /* search328_0_0 -> 722 in the cache */
   /* search328_0 -> 1085 in the cache */
   /* search328_1_0 -> 1090 in the cache */
   /* search328_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1085, 1091 },
      -1,
   } },

   /* replace328_0 -> 722 in the cache */
   /* replace328_1 -> 1090 in the cache */
   /* replace328 -> 1118 in the cache */

   /* ('feq', ('fneg', ('b2f', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('ieq', 'a', 'b') */
   /* search329_0_0_0 -> 722 in the cache */
   /* search329_0_0 -> 1085 in the cache */
   /* search329_0 -> 1088 in the cache */
   /* search329_1_0_0 -> 1090 in the cache */
   /* search329_1_0 -> 1091 in the cache */
   /* search329_1 -> 1098 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1088, 1098 },
      -1,
   } },

   /* replace329_0 -> 722 in the cache */
   /* replace329_1 -> 1090 in the cache */
   /* replace329 -> 1118 in the cache */

   /* ('flt', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('ior', 'a', 'b') */
   /* search330_0_0_0_0 -> 722 in the cache */
   /* search330_0_0_0 -> 1085 in the cache */
   /* search330_0_0_1_0 -> 1090 in the cache */
   /* search330_0_0_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1085, 1091 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1121 },
      -1,
   } },
   /* search330_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1122, 1084 },
      -1,
   } },

   /* replace330_0 -> 722 in the cache */
   /* replace330_1 -> 1090 in the cache */
   /* replace330 -> 1094 in the cache */

   /* ('flt', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('ior', 'a', 'b') */
   /* search331_0 -> 1084 in the cache */
   /* search331_1_0_0 -> 722 in the cache */
   /* search331_1_0 -> 1085 in the cache */
   /* search331_1_1_0 -> 1090 in the cache */
   /* search331_1_1 -> 1091 in the cache */
   /* search331_1 -> 1121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1084, 1121 },
      -1,
   } },

   /* replace331_0 -> 722 in the cache */
   /* replace331_1 -> 1090 in the cache */
   /* replace331 -> 1094 in the cache */

   /* ('fge', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))), 0.0) => ('inot', ('ior', 'a', 'b')) */
   /* search332_0_0_0_0 -> 722 in the cache */
   /* search332_0_0_0 -> 1085 in the cache */
   /* search332_0_0_1_0 -> 1090 in the cache */
   /* search332_0_0_1 -> 1091 in the cache */
   /* search332_0_0 -> 1121 in the cache */
   /* search332_0 -> 1122 in the cache */
   /* search332_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1122, 1084 },
      -1,
   } },

   /* replace332_0_0 -> 722 in the cache */
   /* replace332_0_1 -> 1090 in the cache */
   /* replace332_0 -> 1094 in the cache */
   /* replace332 -> 1111 in the cache */

   /* ('fge', 0.0, ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('inot', ('ior', 'a', 'b')) */
   /* search333_0 -> 1084 in the cache */
   /* search333_1_0_0 -> 722 in the cache */
   /* search333_1_0 -> 1085 in the cache */
   /* search333_1_1_0 -> 1090 in the cache */
   /* search333_1_1 -> 1091 in the cache */
   /* search333_1 -> 1121 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1084, 1121 },
      -1,
   } },

   /* replace333_0_0 -> 722 in the cache */
   /* replace333_0_1 -> 1090 in the cache */
   /* replace333_0 -> 1094 in the cache */
   /* replace333 -> 1111 in the cache */

   /* ('flt', 'a', ('fneg', 'a')) => ('flt', 'a', 0.0) */
   /* search334_0 -> 32 in the cache */
   /* search334_1_0 -> 32 in the cache */
   /* search334_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 98 },
      -1,
   } },

   /* replace334_0 -> 32 in the cache */
   /* replace334_1 -> 113 in the cache */
   /* replace334 -> 674 in the cache */

   /* ('fge', 'a', ('fneg', 'a')) => ('fge', 'a', 0.0) */
   /* search335_0 -> 32 in the cache */
   /* search335_1_0 -> 32 in the cache */
   /* search335_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 98 },
      -1,
   } },

   /* replace335_0 -> 32 in the cache */
   /* replace335_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 113 },
      -1,
   } },

   /* ('flt', ('fmin', 'c', ('fneg', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1')))), 0.0) => ('ior', ('flt', 'c', 0.0), ('ior', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1131 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1133 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1132, 1134 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1130, 1136 },
      -1,
   } },
   /* search336_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 2,
      { 1137, 113 },
      -1,
   } },

   /* replace336_0_0 -> 1130 in the cache */
   /* replace336_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1130, 113 },
      -1,
   } },
   /* replace336_1_0 -> 1131 in the cache */
   /* replace336_1_1 -> 1133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1131, 1133 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1139, 1140 },
      -1,
   } },

   /* ('~flt', ('fadd', 'a', 'b'), 'a') => ('flt', 'b', 0.0) */
   /* search337_0_0 -> 0 in the cache */
   /* search337_0_1 -> 3 in the cache */
   /* search337_0 -> 261 in the cache */
   /* search337_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 261, 0 },
      -1,
   } },

   /* replace337_0 -> 3 in the cache */
   /* replace337_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 3, 245 },
      -1,
   } },

   /* ('~fge', ('fadd', 'a', 'b'), 'a') => ('fge', 'b', 0.0) */
   /* search338_0_0 -> 0 in the cache */
   /* search338_0_1 -> 3 in the cache */
   /* search338_0 -> 261 in the cache */
   /* search338_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 261, 0 },
      -1,
   } },

   /* replace338_0 -> 3 in the cache */
   /* replace338_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 3, 245 },
      -1,
   } },

   /* ('~feq', ('fadd', 'a', 'b'), 'a') => ('feq', 'b', 0.0) */
   /* search339_0_0 -> 0 in the cache */
   /* search339_0_1 -> 3 in the cache */
   /* search339_0 -> 202 in the cache */
   /* search339_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 202, 0 },
      -1,
   } },

   /* replace339_0 -> 3 in the cache */
   /* replace339_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 3, 245 },
      -1,
   } },

   /* ('~fneu', ('fadd', 'a', 'b'), 'a') => ('fneu', 'b', 0.0) */
   /* search340_0_0 -> 0 in the cache */
   /* search340_0_1 -> 3 in the cache */
   /* search340_0 -> 202 in the cache */
   /* search340_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 202, 0 },
      -1,
   } },

   /* replace340_0 -> 3 in the cache */
   /* replace340_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 3, 245 },
      -1,
   } },

   /* ('~flt', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('flt', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search341_0_0 -> 128 in the cache */
   /* search341_0_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 128, 272 },
      0,
   } },
   /* search341_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1150, 776 },
      -1,
   } },

   /* replace341_0 -> 128 in the cache */
   /* replace341_1_0 -> 131 in the cache */
   /* replace341_1_1_0 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 131, 1152 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 128, 1153 },
      -1,
   } },

   /* ('~flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('flt', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search342_0_0_0 -> 128 in the cache */
   /* search342_0_0_1 -> 272 in the cache */
   /* search342_0_0 -> 1150 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1150 },
      0,
   } },
   /* search342_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1155, 776 },
      -1,
   } },

   /* replace342_0_0_0 -> 131 in the cache */
   /* replace342_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 131, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1157 },
      -1,
   } },
   /* replace342_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1158, 128 },
      -1,
   } },

   /* ('~fge', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fge', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search343_0_0 -> 128 in the cache */
   /* search343_0_1 -> 272 in the cache */
   /* search343_0 -> 1150 in the cache */
   /* search343_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1150, 776 },
      -1,
   } },

   /* replace343_0 -> 128 in the cache */
   /* replace343_1_0 -> 131 in the cache */
   /* replace343_1_1_0 -> 129 in the cache */
   /* replace343_1_1 -> 1152 in the cache */
   /* replace343_1 -> 1153 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 128, 1153 },
      -1,
   } },

   /* ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fge', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search344_0_0_0 -> 128 in the cache */
   /* search344_0_0_1 -> 272 in the cache */
   /* search344_0_0 -> 1150 in the cache */
   /* search344_0 -> 1155 in the cache */
   /* search344_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1155, 776 },
      -1,
   } },

   /* replace344_0_0_0 -> 131 in the cache */
   /* replace344_0_0_1 -> 129 in the cache */
   /* replace344_0_0 -> 1157 in the cache */
   /* replace344_0 -> 1158 in the cache */
   /* replace344_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1158, 128 },
      -1,
   } },

   /* ('~feq', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('feq', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search345_0_0 -> 128 in the cache */
   /* search345_0_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 128, 272 },
      0,
   } },
   /* search345_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1164, 776 },
      -1,
   } },

   /* replace345_0 -> 128 in the cache */
   /* replace345_1_0 -> 131 in the cache */
   /* replace345_1_1_0 -> 129 in the cache */
   /* replace345_1_1 -> 1152 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 131, 1152 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 128, 1166 },
      -1,
   } },

   /* ('~feq', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('feq', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search346_0_0_0 -> 128 in the cache */
   /* search346_0_0_1 -> 272 in the cache */
   /* search346_0_0 -> 1164 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1164 },
      0,
   } },
   /* search346_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1168, 776 },
      -1,
   } },

   /* replace346_0_0_0 -> 131 in the cache */
   /* replace346_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 131, 129 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1170 },
      -1,
   } },
   /* replace346_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 1171, 128 },
      -1,
   } },

   /* ('~fneu', ('fadd(is_used_once)', 'a', '#b'), '#c') => ('fneu', 'a', ('fadd', 'c', ('fneg', 'b'))) */
   /* search347_0_0 -> 128 in the cache */
   /* search347_0_1 -> 272 in the cache */
   /* search347_0 -> 1164 in the cache */
   /* search347_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1164, 776 },
      -1,
   } },

   /* replace347_0 -> 128 in the cache */
   /* replace347_1_0 -> 131 in the cache */
   /* replace347_1_1_0 -> 129 in the cache */
   /* replace347_1_1 -> 1152 in the cache */
   /* replace347_1 -> 1166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 128, 1166 },
      -1,
   } },

   /* ('~fneu', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', '#b')), '#c') => ('fneu', ('fneg', ('fadd', 'c', 'b')), 'a') */
   /* search348_0_0_0 -> 128 in the cache */
   /* search348_0_0_1 -> 272 in the cache */
   /* search348_0_0 -> 1164 in the cache */
   /* search348_0 -> 1168 in the cache */
   /* search348_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1168, 776 },
      -1,
   } },

   /* replace348_0_0_0 -> 131 in the cache */
   /* replace348_0_0_1 -> 129 in the cache */
   /* replace348_0_0 -> 1170 in the cache */
   /* replace348_0 -> 1171 in the cache */
   /* replace348_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 1171, 128 },
      -1,
   } },

   /* ('ieq', ('iadd', 'a', 'b'), 'a') => ('ieq', 'b', 0) */
   /* search349_0_0 -> 0 in the cache */
   /* search349_0_1 -> 3 in the cache */
   /* search349_0 -> 197 in the cache */
   /* search349_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 197, 0 },
      -1,
   } },

   /* replace349_0 -> 3 in the cache */
   /* replace349_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3, 86 },
      -1,
   } },

   /* ('ine', ('iadd', 'a', 'b'), 'a') => ('ine', 'b', 0) */
   /* search350_0_0 -> 0 in the cache */
   /* search350_0_1 -> 3 in the cache */
   /* search350_0 -> 197 in the cache */
   /* search350_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 197, 0 },
      -1,
   } },

   /* replace350_0 -> 3 in the cache */
   /* replace350_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3, 86 },
      -1,
   } },

   /* ('feq', ('b2f', 'a@1'), 0.0) => ('inot', 'a') */
   /* search351_0_0 -> 722 in the cache */
   /* search351_0 -> 1085 in the cache */
   /* search351_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1085, 1084 },
      -1,
   } },

   /* replace351_0 -> 722 in the cache */
   /* replace351 -> 1087 in the cache */

   /* ('fneu', ('b2f', 'a@1'), 0.0) => a */
   /* search352_0_0 -> 722 in the cache */
   /* search352_0 -> 1085 in the cache */
   /* search352_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1085, 1084 },
      -1,
   } },

   /* replace352 -> 722 in the cache */

   /* ('ieq', ('b2i', 'a@1'), 0) => ('inot', 'a') */
   /* search353_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 722 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1183, 1184 },
      -1,
   } },

   /* replace353_0 -> 722 in the cache */
   /* replace353 -> 1087 in the cache */

   /* ('ine', ('b2i', 'a@1'), 0) => a */
   /* search354_0_0 -> 722 in the cache */
   /* search354_0 -> 1183 in the cache */
   /* search354_1 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1183, 1184 },
      -1,
   } },

   /* replace354 -> 722 in the cache */

   /* ('fneu', ('u2f', 'a'), 0.0) => ('ine', 'a', 0) */
   /* search355_0_0 -> 32 in the cache */
   /* search355_0 -> 104 in the cache */
   /* search355_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 104, 1084 },
      -1,
   } },

   /* replace355_0 -> 32 in the cache */
   /* replace355_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('feq', ('u2f', 'a'), 0.0) => ('ieq', 'a', 0) */
   /* search356_0_0 -> 32 in the cache */
   /* search356_0 -> 104 in the cache */
   /* search356_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 104, 1084 },
      -1,
   } },

   /* replace356_0 -> 32 in the cache */
   /* replace356_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('fge', ('u2f', 'a'), 0.0) => True */
   /* search357_0_0 -> 32 in the cache */
   /* search357_0 -> 104 in the cache */
   /* search357_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 104, 1084 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_bool, { NIR_TRUE /* True */ },
   } },

   /* ('fge', 0.0, ('u2f', 'a')) => ('uge', 0, 'a') */
   /* search358_0 -> 1084 in the cache */
   /* search358_1_0 -> 32 in the cache */
   /* search358_1 -> 104 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1084, 104 },
      -1,
   } },

   /* replace358_0 -> 61 in the cache */
   /* replace358_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('flt', ('u2f', 'a'), 0.0) => False */
   /* search359_0_0 -> 32 in the cache */
   /* search359_0 -> 104 in the cache */
   /* search359_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 104, 1084 },
      -1,
   } },

   /* replace359 -> 1039 in the cache */

   /* ('flt', 0.0, ('u2f', 'a')) => ('ult', 0, 'a') */
   /* search360_0 -> 1084 in the cache */
   /* search360_1_0 -> 32 in the cache */
   /* search360_1 -> 104 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1084, 104 },
      -1,
   } },

   /* replace360_0 -> 61 in the cache */
   /* replace360_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('fneu', ('i2f', 'a'), 0.0) => ('ine', 'a', 0) */
   /* search361_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search361_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1198, 1084 },
      -1,
   } },

   /* replace361_0 -> 32 in the cache */
   /* replace361_1 -> 61 in the cache */
   /* replace361 -> 1188 in the cache */

   /* ('feq', ('i2f', 'a'), 0.0) => ('ieq', 'a', 0) */
   /* search362_0_0 -> 32 in the cache */
   /* search362_0 -> 1198 in the cache */
   /* search362_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1198, 1084 },
      -1,
   } },

   /* replace362_0 -> 32 in the cache */
   /* replace362_1 -> 61 in the cache */
   /* replace362 -> 1190 in the cache */

   /* ('fge', ('i2f', 'a'), 0.0) => ('ige', 'a', 0) */
   /* search363_0_0 -> 32 in the cache */
   /* search363_0 -> 1198 in the cache */
   /* search363_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1198, 1084 },
      -1,
   } },

   /* replace363_0 -> 32 in the cache */
   /* replace363_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('fge', 0.0, ('i2f', 'a')) => ('ige', 0, 'a') */
   /* search364_0 -> 1084 in the cache */
   /* search364_1_0 -> 32 in the cache */
   /* search364_1 -> 1198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1084, 1198 },
      -1,
   } },

   /* replace364_0 -> 61 in the cache */
   /* replace364_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('flt', ('i2f', 'a'), 0.0) => ('ilt', 'a', 0) */
   /* search365_0_0 -> 32 in the cache */
   /* search365_0 -> 1198 in the cache */
   /* search365_1 -> 1084 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1198, 1084 },
      -1,
   } },

   /* replace365_0 -> 32 in the cache */
   /* replace365_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('flt', 0.0, ('i2f', 'a')) => ('ilt', 0, 'a') */
   /* search366_0 -> 1084 in the cache */
   /* search366_1_0 -> 32 in the cache */
   /* search366_1 -> 1198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1084, 1198 },
      -1,
   } },

   /* replace366_0 -> 61 in the cache */
   /* replace366_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 61, 32 },
      -1,
   } },

   /* ('~flt', 0.0, ('fabs', 'a')) => ('fneu', 'a', 0.0) */
   /* search367_0 -> 113 in the cache */
   /* search367_1_0 -> 32 in the cache */
   /* search367_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 113, 103 },
      -1,
   } },

   /* replace367_0 -> 32 in the cache */
   /* replace367_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 32, 113 },
      -1,
   } },

   /* ('~flt', ('fneg', ('fabs', 'a')), 0.0) => ('fneu', 'a', 0.0) */
   /* search368_0_0_0 -> 32 in the cache */
   /* search368_0_0 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 103 },
      -1,
   } },
   /* search368_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1211, 113 },
      -1,
   } },

   /* replace368_0 -> 32 in the cache */
   /* replace368_1 -> 113 in the cache */
   /* replace368 -> 1210 in the cache */

   /* ('fge', 0.0, ('fabs', 'a')) => ('feq', 'a', 0.0) */
   /* search369_0 -> 113 in the cache */
   /* search369_1_0 -> 32 in the cache */
   /* search369_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 113, 103 },
      -1,
   } },

   /* replace369_0 -> 32 in the cache */
   /* replace369_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 113 },
      -1,
   } },

   /* ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0) */
   /* search370_0_0_0 -> 32 in the cache */
   /* search370_0_0 -> 103 in the cache */
   /* search370_0 -> 1211 in the cache */
   /* search370_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1211, 113 },
      -1,
   } },

   /* replace370_0 -> 32 in the cache */
   /* replace370_1 -> 113 in the cache */
   /* replace370 -> 1214 in the cache */

   /* ('iand', ('fge', 'a', 0.0), ('fge', 1.0, 'a')) => ('feq', 'a', ('fsat', 'a')) */
   /* search371_0_0 -> 32 in the cache */
   /* search371_0_1 -> 113 in the cache */
   /* search371_0 -> 1129 in the cache */
   /* search371_1_0 -> 230 in the cache */
   /* search371_1_1 -> 32 in the cache */
   /* search371_1 -> 1062 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1129, 1062 },
      -1,
   } },

   /* replace371_0 -> 32 in the cache */
   /* replace371_1_0 -> 32 in the cache */
   /* replace371_1 -> 206 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 32, 206 },
      -1,
   } },

   /* ('fmax', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('ior', 'a', 'b')) */
   /* search372_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 722 },
      0,
   } },
   /* search372_1_0 -> 1090 in the cache */
   /* search372_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1218, 1091 },
      -1,
   } },

   /* replace372_0_0 -> 722 in the cache */
   /* replace372_0_1 -> 1090 in the cache */
   /* replace372_0 -> 1094 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1094 },
      -1,
   } },

   /* ('fmax', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('iand', 'a', 'b'))) */
   /* search373_0_0_0 -> 722 in the cache */
   /* search373_0_0 -> 1218 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1218 },
      0,
   } },
   /* search373_1_0_0 -> 1090 in the cache */
   /* search373_1_0 -> 1091 in the cache */
   /* search373_1 -> 1098 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1221, 1098 },
      -1,
   } },

   /* replace373_0_0_0 -> 722 in the cache */
   /* replace373_0_0_1 -> 1090 in the cache */
   /* replace373_0_0 -> 1102 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1102 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1223 },
      -1,
   } },

   /* ('fmin', ('b2f(is_used_once)', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b')) */
   /* search374_0_0 -> 722 in the cache */
   /* search374_0 -> 1218 in the cache */
   /* search374_1_0 -> 1090 in the cache */
   /* search374_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1218, 1091 },
      -1,
   } },

   /* replace374_0_0 -> 722 in the cache */
   /* replace374_0_1 -> 1090 in the cache */
   /* replace374_0 -> 1102 in the cache */
   /* replace374 -> 1223 in the cache */

   /* ('fmin', ('fneg(is_used_once)', ('b2f(is_used_once)', 'a@1')), ('fneg', ('b2f', 'b@1'))) => ('fneg', ('b2f', ('ior', 'a', 'b'))) */
   /* search375_0_0_0 -> 722 in the cache */
   /* search375_0_0 -> 1218 in the cache */
   /* search375_0 -> 1221 in the cache */
   /* search375_1_0_0 -> 1090 in the cache */
   /* search375_1_0 -> 1091 in the cache */
   /* search375_1 -> 1098 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1221, 1098 },
      -1,
   } },

   /* replace375_0_0_0 -> 722 in the cache */
   /* replace375_0_0_1 -> 1090 in the cache */
   /* replace375_0_0 -> 1094 in the cache */
   /* replace375_0 -> 1220 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1220 },
      -1,
   } },

   /* ('fmin', ('b2f', 'a@1'), '#b') => ('bcsel', 'a', ('fmin', 'b', 1.0), ('fmin', 'b', 0.0)) */
   /* search376_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search376_1 -> 783 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1228, 783 },
      -1,
   } },

   /* replace376_0 -> 722 in the cache */
   /* replace376_1_0 -> 3 in the cache */
   /* replace376_1_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 3, 259 },
      -1,
   } },
   /* replace376_2_0 -> 3 in the cache */
   /* replace376_2_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 3, 245 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 1230, 1231 },
      -1,
   } },

   /* ('flt', ('fadd(is_used_once)', 'a', ('fneg', 'b')), 0.0) => ('flt', 'a', 'b') */
   /* search377_0_0 -> 0 in the cache */
   /* search377_0_1_0 -> 3 in the cache */
   /* search377_0_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1049 },
      0,
   } },
   /* search377_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1233, 245 },
      -1,
   } },

   /* replace377_0 -> 0 in the cache */
   /* replace377_1 -> 3 in the cache */
   /* replace377 -> 1011 in the cache */

   /* ('fge', ('fneg', ('fabs', 'a')), 0.0) => ('feq', 'a', 0.0) */
   /* search378_0_0_0 -> 32 in the cache */
   /* search378_0_0 -> 103 in the cache */
   /* search378_0 -> 1211 in the cache */
   /* search378_1 -> 113 in the cache */
   /* search378 -> 1215 in the cache */

   /* replace378_0 -> 32 in the cache */
   /* replace378_1 -> 113 in the cache */
   /* replace378 -> 1214 in the cache */

   /* ('~bcsel', ('flt', 'b', 'a'), 'b', 'a') => ('fmin', 'a', 'b') */
   /* search379_0_0 -> 1073 in the cache */
   /* search379_0_1 -> 1070 in the cache */
   /* search379_0 -> 1074 in the cache */
   /* search379_1 -> 1073 in the cache */
   /* search379_2 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1074, 1073, 1070 },
      -1,
   } },

   /* replace379_0 -> 1070 in the cache */
   /* replace379_1 -> 1073 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1070, 1073 },
      -1,
   } },

   /* ('~bcsel', ('flt', 'a', 'b'), 'b', 'a') => ('fmax', 'a', 'b') */
   /* search380_0_0 -> 0 in the cache */
   /* search380_0_1 -> 3 in the cache */
   /* search380_0 -> 1011 in the cache */
   /* search380_1 -> 3 in the cache */
   /* search380_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1011, 3, 0 },
      -1,
   } },

   /* replace380_0 -> 0 in the cache */
   /* replace380_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('~bcsel', ('fge', 'a', 'b'), 'b', 'a') => ('fmin', 'a', 'b') */
   /* search381_0_0 -> 0 in the cache */
   /* search381_0_1 -> 3 in the cache */
   /* search381_0 -> 1008 in the cache */
   /* search381_1 -> 3 in the cache */
   /* search381_2 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1008, 3, 0 },
      -1,
   } },

   /* replace381_0 -> 0 in the cache */
   /* replace381_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('~bcsel', ('fge', 'b', 'a'), 'b', 'a') => ('fmax', 'a', 'b') */
   /* search382_0_0 -> 1073 in the cache */
   /* search382_0_1 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1073, 1070 },
      -1,
   } },
   /* search382_1 -> 1073 in the cache */
   /* search382_2 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1241, 1073, 1070 },
      -1,
   } },

   /* replace382_0 -> 1070 in the cache */
   /* replace382_1 -> 1073 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1070, 1073 },
      -1,
   } },

   /* ('bcsel', ('i2b', 'a'), 'b', 'c') => ('bcsel', ('ine', 'a', 0), 'b', 'c') */
   /* search383_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_search_op_i2b,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search383_1 -> 129 in the cache */
   /* search383_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1244, 129, 131 },
      -1,
   } },

   /* replace383_0_0 -> 32 in the cache */
   /* replace383_0_1 -> 61 in the cache */
   /* replace383_0 -> 1188 in the cache */
   /* replace383_1 -> 129 in the cache */
   /* replace383_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1188, 129, 131 },
      -1,
   } },

   /* ('bcsel', ('inot', 'a'), 'b', 'c') => ('bcsel', 'a', 'c', 'b') */
   /* search384_0_0 -> 722 in the cache */
   /* search384_0 -> 1087 in the cache */
   /* search384_1 -> 129 in the cache */
   /* search384_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1087, 129, 131 },
      -1,
   } },

   /* replace384_0 -> 722 in the cache */
   /* replace384_1 -> 131 in the cache */
   /* replace384_2 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 131, 129 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel', 'a', 'b', 'c'), 'd') => ('bcsel', 'a', 'b', 'd') */
   /* search385_0 -> 722 in the cache */
   /* search385_1_0 -> 722 in the cache */
   /* search385_1_1 -> 152 in the cache */
   /* search385_1_2 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 152, 154 },
      -1,
   } },
   /* search385_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1249, 155 },
      -1,
   } },

   /* replace385_0 -> 722 in the cache */
   /* replace385_1 -> 152 in the cache */
   /* replace385_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 152, 155 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', ('bcsel', 'a', 'c', 'd')) => ('bcsel', 'a', 'b', 'd') */
   /* search386_0 -> 722 in the cache */
   /* search386_1 -> 152 in the cache */
   /* search386_2_0 -> 722 in the cache */
   /* search386_2_1 -> 154 in the cache */
   /* search386_2_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 152, 1252 },
      -1,
   } },

   /* replace386_0 -> 722 in the cache */
   /* replace386_1 -> 152 in the cache */
   /* replace386_2 -> 155 in the cache */
   /* replace386 -> 1251 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search387_0 -> 722 in the cache */
   /* search387_1_0 -> 1090 in the cache */
   /* search387_1_1 -> 715 in the cache */
   /* search387_1_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 715, 724 },
      -1,
   } },
   /* search387_2_0 -> 1090 in the cache */
   /* search387_2_1 -> 715 in the cache */
   /* search387_2_2 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 715, 726 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1254, 1255 },
      -1,
   } },

   /* replace387_0 -> 1090 in the cache */
   /* replace387_1 -> 715 in the cache */
   /* replace387_2_0 -> 722 in the cache */
   /* replace387_2_1 -> 724 in the cache */
   /* replace387_2_2 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 724, 726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 715, 1257 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'c', 'e')) => ('bcsel', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search388_0 -> 722 in the cache */
   /* search388_1_0 -> 1090 in the cache */
   /* search388_1_1 -> 715 in the cache */
   /* search388_1_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 715, 724 },
      0,
   } },
   /* search388_2_0 -> 1090 in the cache */
   /* search388_2_1 -> 715 in the cache */
   /* search388_2_2 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 715, 726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1259, 1260 },
      -1,
   } },

   /* replace388_0 -> 1090 in the cache */
   /* replace388_1 -> 715 in the cache */
   /* replace388_2_0 -> 722 in the cache */
   /* replace388_2_1 -> 724 in the cache */
   /* replace388_2_2 -> 726 in the cache */
   /* replace388_2 -> 1257 in the cache */
   /* replace388 -> 1258 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), ('bcsel(is_used_once)', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search389_0 -> 722 in the cache */
   /* search389_1_0 -> 1090 in the cache */
   /* search389_1_1 -> 154 in the cache */
   /* search389_1_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 154, 155 },
      -1,
   } },
   /* search389_2_0 -> 1090 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search389_2_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 1263, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1262, 1264 },
      -1,
   } },

   /* replace389_0 -> 1090 in the cache */
   /* replace389_1_0 -> 722 in the cache */
   /* replace389_1_1 -> 154 in the cache */
   /* replace389_1_2 -> 1263 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 154, 1263 },
      -1,
   } },
   /* replace389_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 1266, 155 },
      -1,
   } },

   /* ('bcsel', 'a', ('bcsel(is_used_once)', 'b', 'c', 'd'), ('bcsel', 'b', 'e', 'd')) => ('bcsel', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search390_0 -> 722 in the cache */
   /* search390_1_0 -> 1090 in the cache */
   /* search390_1_1 -> 154 in the cache */
   /* search390_1_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 154, 155 },
      0,
   } },
   /* search390_2_0 -> 1090 in the cache */
   /* search390_2_1 -> 1263 in the cache */
   /* search390_2_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1090, 1263, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1268, 1269 },
      -1,
   } },

   /* replace390_0 -> 1090 in the cache */
   /* replace390_1_0 -> 722 in the cache */
   /* replace390_1_1 -> 154 in the cache */
   /* replace390_1_2 -> 1263 in the cache */
   /* replace390_1 -> 1266 in the cache */
   /* replace390_2 -> 155 in the cache */
   /* replace390 -> 1267 in the cache */

   /* ('bcsel', 'a', True, 'b') => ('ior', 'a', 'b') */
   /* search391_0 -> 722 in the cache */
   /* search391_1 -> 1192 in the cache */
   /* search391_2 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1192, 1090 },
      -1,
   } },

   /* replace391_0 -> 722 in the cache */
   /* replace391_1 -> 1090 in the cache */
   /* replace391 -> 1094 in the cache */

   /* ('bcsel', 'a', 'a', 'b') => ('ior', 'a', 'b') */
   /* search392_0 -> 722 in the cache */
   /* search392_1 -> 722 in the cache */
   /* search392_2 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 722, 1090 },
      -1,
   } },

   /* replace392_0 -> 722 in the cache */
   /* replace392_1 -> 1090 in the cache */
   /* replace392 -> 1094 in the cache */

   /* ('bcsel', 'a', 'b', False) => ('iand', 'a', 'b') */
   /* search393_0 -> 722 in the cache */
   /* search393_1 -> 1090 in the cache */
   /* search393_2 -> 1039 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1090, 1039 },
      -1,
   } },

   /* replace393_0 -> 722 in the cache */
   /* replace393_1 -> 1090 in the cache */
   /* replace393 -> 1102 in the cache */

   /* ('bcsel', 'a', 'b', 'a') => ('iand', 'a', 'b') */
   /* search394_0 -> 722 in the cache */
   /* search394_1 -> 1090 in the cache */
   /* search394_2 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1090, 722 },
      -1,
   } },

   /* replace394_0 -> 722 in the cache */
   /* replace394_1 -> 1090 in the cache */
   /* replace394 -> 1102 in the cache */

   /* ('~fmin', 'a', 'a') => a */
   /* search395_0 -> 32 in the cache */
   /* search395_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace395 -> 32 in the cache */

   /* ('~fmax', 'a', 'a') => a */
   /* search396_0 -> 32 in the cache */
   /* search396_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace396 -> 32 in the cache */

   /* ('imin', 'a', 'a') => a */
   /* search397_0 -> 32 in the cache */
   /* search397_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace397 -> 32 in the cache */

   /* ('imax', 'a', 'a') => a */
   /* search398_0 -> 32 in the cache */
   /* search398_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace398 -> 32 in the cache */

   /* ('umin', 'a', 'a') => a */
   /* search399_0 -> 32 in the cache */
   /* search399_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace399 -> 32 in the cache */

   /* ('umin', 'a', 0) => 0 */
   /* search400_0 -> 32 in the cache */
   /* search400_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace400 -> 61 in the cache */

   /* ('umin', 'a', -1) => a */
   /* search401_0 -> 32 in the cache */
   /* search401_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace401 -> 32 in the cache */

   /* ('umax', 'a', 'a') => a */
   /* search402_0 -> 32 in the cache */
   /* search402_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace402 -> 32 in the cache */

   /* ('umax', 'a', 0) => a */
   /* search403_0 -> 32 in the cache */
   /* search403_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace403 -> 32 in the cache */

   /* ('umax', 'a', -1) => -1 */
   /* search404_0 -> 32 in the cache */
   /* search404_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace404 -> 63 in the cache */

   /* ('fmax', ('fmax', 'a', 'b'), 'b') => ('fmax', 'a', 'b') */
   /* search405_0_0 -> 0 in the cache */
   /* search405_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search405_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1285, 3 },
      -1,
   } },

   /* replace405_0 -> 0 in the cache */
   /* replace405_1 -> 3 in the cache */
   /* replace405 -> 1238 in the cache */

   /* ('umax', ('umax', 'a', 'b'), 'b') => ('umax', 'a', 'b') */
   /* search406_0_0 -> 0 in the cache */
   /* search406_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search406_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 2,
      { 1287, 3 },
      -1,
   } },

   /* replace406_0 -> 0 in the cache */
   /* replace406_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('imax', ('imax', 'a', 'b'), 'b') => ('imax', 'a', 'b') */
   /* search407_0_0 -> 0 in the cache */
   /* search407_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search407_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 1290, 3 },
      -1,
   } },

   /* replace407_0 -> 0 in the cache */
   /* replace407_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('fmin', ('fmin', 'a', 'b'), 'b') => ('fmin', 'a', 'b') */
   /* search408_0_0 -> 0 in the cache */
   /* search408_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search408_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1293, 3 },
      -1,
   } },

   /* replace408_0 -> 0 in the cache */
   /* replace408_1 -> 3 in the cache */
   /* replace408 -> 1240 in the cache */

   /* ('umin', ('umin', 'a', 'b'), 'b') => ('umin', 'a', 'b') */
   /* search409_0_0 -> 0 in the cache */
   /* search409_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search409_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1295, 3 },
      -1,
   } },

   /* replace409_0 -> 0 in the cache */
   /* replace409_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('imin', ('imin', 'a', 'b'), 'b') => ('imin', 'a', 'b') */
   /* search410_0_0 -> 0 in the cache */
   /* search410_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search410_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1298, 3 },
      -1,
   } },

   /* replace410_0 -> 0 in the cache */
   /* replace410_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('fmax', ('fmax', ('fmax', 'a', 'b'), 'c'), 'a') => ('fmax', ('fmax', 'a', 'b'), 'c') */
   /* search411_0_0_0 -> 128 in the cache */
   /* search411_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 128, 129 },
      -1,
   } },
   /* search411_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 2,
      { 1301, 131 },
      -1,
   } },
   /* search411_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 1302, 128 },
      -1,
   } },

   /* replace411_0_0 -> 128 in the cache */
   /* replace411_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* replace411_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1304, 131 },
      -1,
   } },

   /* ('umax', ('umax', ('umax', 'a', 'b'), 'c'), 'a') => ('umax', ('umax', 'a', 'b'), 'c') */
   /* search412_0_0_0 -> 128 in the cache */
   /* search412_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      2, 1,
      { 128, 129 },
      -1,
   } },
   /* search412_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 2,
      { 1306, 131 },
      -1,
   } },
   /* search412_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 3,
      { 1307, 128 },
      -1,
   } },

   /* replace412_0_0 -> 128 in the cache */
   /* replace412_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* replace412_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 2,
      { 1309, 131 },
      -1,
   } },

   /* ('imax', ('imax', ('imax', 'a', 'b'), 'c'), 'a') => ('imax', ('imax', 'a', 'b'), 'c') */
   /* search413_0_0_0 -> 128 in the cache */
   /* search413_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      2, 1,
      { 128, 129 },
      -1,
   } },
   /* search413_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 2,
      { 1311, 131 },
      -1,
   } },
   /* search413_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 3,
      { 1312, 128 },
      -1,
   } },

   /* replace413_0_0 -> 128 in the cache */
   /* replace413_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* replace413_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 1314, 131 },
      -1,
   } },

   /* ('fmin', ('fmin', ('fmin', 'a', 'b'), 'c'), 'a') => ('fmin', ('fmin', 'a', 'b'), 'c') */
   /* search414_0_0_0 -> 128 in the cache */
   /* search414_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      2, 1,
      { 128, 129 },
      -1,
   } },
   /* search414_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 1316, 131 },
      -1,
   } },
   /* search414_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 1317, 128 },
      -1,
   } },

   /* replace414_0_0 -> 128 in the cache */
   /* replace414_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* replace414_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1319, 131 },
      -1,
   } },

   /* ('umin', ('umin', ('umin', 'a', 'b'), 'c'), 'a') => ('umin', ('umin', 'a', 'b'), 'c') */
   /* search415_0_0_0 -> 128 in the cache */
   /* search415_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      2, 1,
      { 128, 129 },
      -1,
   } },
   /* search415_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      1, 2,
      { 1321, 131 },
      -1,
   } },
   /* search415_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 3,
      { 1322, 128 },
      -1,
   } },

   /* replace415_0_0 -> 128 in the cache */
   /* replace415_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* replace415_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1324, 131 },
      -1,
   } },

   /* ('imin', ('imin', ('imin', 'a', 'b'), 'c'), 'a') => ('imin', ('imin', 'a', 'b'), 'c') */
   /* search416_0_0_0 -> 128 in the cache */
   /* search416_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      2, 1,
      { 128, 129 },
      -1,
   } },
   /* search416_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      1, 2,
      { 1326, 131 },
      -1,
   } },
   /* search416_1 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 3,
      { 1327, 128 },
      -1,
   } },

   /* replace416_0_0 -> 128 in the cache */
   /* replace416_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 128, 129 },
      -1,
   } },
   /* replace416_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1329, 131 },
      -1,
   } },

   /* ('ieq', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ieq', 'a', 'b') */
   /* search417_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search417_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1331, 1332 },
      -1,
   } },

   /* replace417_0 -> 722 in the cache */
   /* replace417_1 -> 1090 in the cache */
   /* replace417 -> 1118 in the cache */

   /* ('ine', ('b2i8', 'a@1'), ('b2i8', 'b@1')) => ('ine', 'a', 'b') */
   /* search418_0_0 -> 722 in the cache */
   /* search418_0 -> 1331 in the cache */
   /* search418_1_0 -> 1090 in the cache */
   /* search418_1 -> 1332 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1331, 1332 },
      -1,
   } },

   /* replace418_0 -> 722 in the cache */
   /* replace418_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 722, 1090 },
      -1,
   } },

   /* ('ieq', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ieq', 'a', 'b') */
   /* search419_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search419_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1336, 1337 },
      -1,
   } },

   /* replace419_0 -> 722 in the cache */
   /* replace419_1 -> 1090 in the cache */
   /* replace419 -> 1118 in the cache */

   /* ('ine', ('b2i16', 'a@1'), ('b2i16', 'b@1')) => ('ine', 'a', 'b') */
   /* search420_0_0 -> 722 in the cache */
   /* search420_0 -> 1336 in the cache */
   /* search420_1_0 -> 1090 in the cache */
   /* search420_1 -> 1337 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1336, 1337 },
      -1,
   } },

   /* replace420_0 -> 722 in the cache */
   /* replace420_1 -> 1090 in the cache */
   /* replace420 -> 1335 in the cache */

   /* ('ieq', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ieq', 'a', 'b') */
   /* search421_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search421_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1340, 1341 },
      -1,
   } },

   /* replace421_0 -> 722 in the cache */
   /* replace421_1 -> 1090 in the cache */
   /* replace421 -> 1118 in the cache */

   /* ('ine', ('b2i32', 'a@1'), ('b2i32', 'b@1')) => ('ine', 'a', 'b') */
   /* search422_0_0 -> 722 in the cache */
   /* search422_0 -> 1340 in the cache */
   /* search422_1_0 -> 1090 in the cache */
   /* search422_1 -> 1341 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1340, 1341 },
      -1,
   } },

   /* replace422_0 -> 722 in the cache */
   /* replace422_1 -> 1090 in the cache */
   /* replace422 -> 1335 in the cache */

   /* ('ieq', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ieq', 'a', 'b') */
   /* search423_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search423_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1344, 1345 },
      -1,
   } },

   /* replace423_0 -> 722 in the cache */
   /* replace423_1 -> 1090 in the cache */
   /* replace423 -> 1118 in the cache */

   /* ('ine', ('b2i64', 'a@1'), ('b2i64', 'b@1')) => ('ine', 'a', 'b') */
   /* search424_0_0 -> 722 in the cache */
   /* search424_0 -> 1344 in the cache */
   /* search424_1_0 -> 1090 in the cache */
   /* search424_1 -> 1345 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1344, 1345 },
      -1,
   } },

   /* replace424_0 -> 722 in the cache */
   /* replace424_1 -> 1090 in the cache */
   /* replace424 -> 1335 in the cache */

   /* ('feq', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ieq', 'a', 'b') */
   /* search425_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search425_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1348, 1349 },
      -1,
   } },

   /* replace425_0 -> 722 in the cache */
   /* replace425_1 -> 1090 in the cache */
   /* replace425 -> 1118 in the cache */

   /* ('fneu', ('b2f16', 'a@1'), ('b2f16', 'b@1')) => ('ine', 'a', 'b') */
   /* search426_0_0 -> 722 in the cache */
   /* search426_0 -> 1348 in the cache */
   /* search426_1_0 -> 1090 in the cache */
   /* search426_1 -> 1349 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1348, 1349 },
      -1,
   } },

   /* replace426_0 -> 722 in the cache */
   /* replace426_1 -> 1090 in the cache */
   /* replace426 -> 1335 in the cache */

   /* ('feq', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ieq', 'a', 'b') */
   /* search427_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search427_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1352, 1353 },
      -1,
   } },

   /* replace427_0 -> 722 in the cache */
   /* replace427_1 -> 1090 in the cache */
   /* replace427 -> 1118 in the cache */

   /* ('fneu', ('b2f32', 'a@1'), ('b2f32', 'b@1')) => ('ine', 'a', 'b') */
   /* search428_0_0 -> 722 in the cache */
   /* search428_0 -> 1352 in the cache */
   /* search428_1_0 -> 1090 in the cache */
   /* search428_1 -> 1353 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1352, 1353 },
      -1,
   } },

   /* replace428_0 -> 722 in the cache */
   /* replace428_1 -> 1090 in the cache */
   /* replace428 -> 1335 in the cache */

   /* ('feq', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ieq', 'a', 'b') */
   /* search429_0_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 722 },
      -1,
   } },
   /* search429_1_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1356, 1357 },
      -1,
   } },

   /* replace429_0 -> 722 in the cache */
   /* replace429_1 -> 1090 in the cache */
   /* replace429 -> 1118 in the cache */

   /* ('fneu', ('b2f64', 'a@1'), ('b2f64', 'b@1')) => ('ine', 'a', 'b') */
   /* search430_0_0 -> 722 in the cache */
   /* search430_0 -> 1356 in the cache */
   /* search430_1_0 -> 1090 in the cache */
   /* search430_1 -> 1357 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1356, 1357 },
      -1,
   } },

   /* replace430_0 -> 722 in the cache */
   /* replace430_1 -> 1090 in the cache */
   /* replace430 -> 1335 in the cache */

   /* ('iand@8', 'a', ('inot', ('ishr', 'a', 7))) => ('imax', 'a', 0) */
   /* search431_0 -> 6 in the cache */
   /* search431_1_0_0 -> 6 in the cache */
   /* search431_1_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 9 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1360 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 6, 1361 },
      -1,
   } },

   /* replace431_0 -> 6 in the cache */
   /* replace431_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 6, 797 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@8', 128), 0) => ('ige', 'a', 0) */
   /* search432_0_0 -> 6 in the cache */
   /* search432_0_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 6, 7 },
      -1,
   } },
   /* search432_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1364, 797 },
      -1,
   } },

   /* replace432_0 -> 6 in the cache */
   /* replace432_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 797 },
      -1,
   } },

   /* ('ine', ('iand', 'a@8', 128), 128) => ('ige', 'a', 0) */
   /* search433_0_0 -> 6 in the cache */
   /* search433_0_1 -> 7 in the cache */
   /* search433_0 -> 1364 in the cache */
   /* search433_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1364, 7 },
      -1,
   } },

   /* replace433_0 -> 6 in the cache */
   /* replace433_1 -> 797 in the cache */
   /* replace433 -> 1366 in the cache */

   /* ('ine', ('iand', 'a@8', 128), 0) => ('ilt', 'a', 0) */
   /* search434_0_0 -> 6 in the cache */
   /* search434_0_1 -> 7 in the cache */
   /* search434_0 -> 1364 in the cache */
   /* search434_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1364, 797 },
      -1,
   } },

   /* replace434_0 -> 6 in the cache */
   /* replace434_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 797 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@8', 128), 128) => ('ilt', 'a', 0) */
   /* search435_0_0 -> 6 in the cache */
   /* search435_0_1 -> 7 in the cache */
   /* search435_0 -> 1364 in the cache */
   /* search435_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1364, 7 },
      -1,
   } },

   /* replace435_0 -> 6 in the cache */
   /* replace435_1 -> 797 in the cache */
   /* replace435 -> 1369 in the cache */

   /* ('ine', ('ushr', 'a@8', 7), 0) => ('ilt', 'a', 0) */
   /* search436_0_0 -> 6 in the cache */
   /* search436_0_1 -> 9 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 9 },
      -1,
   } },
   /* search436_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1371, 797 },
      -1,
   } },

   /* replace436_0 -> 6 in the cache */
   /* replace436_1 -> 797 in the cache */
   /* replace436 -> 1369 in the cache */

   /* ('ieq', ('ushr', 'a@8', 7), 0) => ('ige', 'a', 0) */
   /* search437_0_0 -> 6 in the cache */
   /* search437_0_1 -> 9 in the cache */
   /* search437_0 -> 1371 in the cache */
   /* search437_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1371, 797 },
      -1,
   } },

   /* replace437_0 -> 6 in the cache */
   /* replace437_1 -> 797 in the cache */
   /* replace437 -> 1366 in the cache */

   /* ('ieq', ('ushr', 'a@8', 7), 1) => ('ilt', 'a', 0) */
   /* search438_0_0 -> 6 in the cache */
   /* search438_0_1 -> 9 in the cache */
   /* search438_0 -> 1371 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1371, 1374 },
      -1,
   } },

   /* replace438_0 -> 6 in the cache */
   /* replace438_1 -> 797 in the cache */
   /* replace438 -> 1369 in the cache */

   /* ('ine', ('ushr', 'a@8', 7), 1) => ('ige', 'a', 0) */
   /* search439_0_0 -> 6 in the cache */
   /* search439_0_1 -> 9 in the cache */
   /* search439_0 -> 1371 in the cache */
   /* search439_1 -> 1374 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1371, 1374 },
      -1,
   } },

   /* replace439_0 -> 6 in the cache */
   /* replace439_1 -> 797 in the cache */
   /* replace439 -> 1366 in the cache */

   /* ('ine', ('ishr', 'a@8', 7), 0) => ('ilt', 'a', 0) */
   /* search440_0_0 -> 6 in the cache */
   /* search440_0_1 -> 9 in the cache */
   /* search440_0 -> 1360 in the cache */
   /* search440_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1360, 797 },
      -1,
   } },

   /* replace440_0 -> 6 in the cache */
   /* replace440_1 -> 797 in the cache */
   /* replace440 -> 1369 in the cache */

   /* ('ieq', ('ishr', 'a@8', 7), 0) => ('ige', 'a', 0) */
   /* search441_0_0 -> 6 in the cache */
   /* search441_0_1 -> 9 in the cache */
   /* search441_0 -> 1360 in the cache */
   /* search441_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1360, 797 },
      -1,
   } },

   /* replace441_0 -> 6 in the cache */
   /* replace441_1 -> 797 in the cache */
   /* replace441 -> 1366 in the cache */

   /* ('ieq', ('ishr', 'a@8', 7), -1) => ('ilt', 'a', 0) */
   /* search442_0_0 -> 6 in the cache */
   /* search442_0_1 -> 9 in the cache */
   /* search442_0 -> 1360 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1360, 1379 },
      -1,
   } },

   /* replace442_0 -> 6 in the cache */
   /* replace442_1 -> 797 in the cache */
   /* replace442 -> 1369 in the cache */

   /* ('ine', ('ishr', 'a@8', 7), -1) => ('ige', 'a', 0) */
   /* search443_0_0 -> 6 in the cache */
   /* search443_0_1 -> 9 in the cache */
   /* search443_0 -> 1360 in the cache */
   /* search443_1 -> 1379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1360, 1379 },
      -1,
   } },

   /* replace443_0 -> 6 in the cache */
   /* replace443_1 -> 797 in the cache */
   /* replace443 -> 1366 in the cache */

   /* ('iand@16', 'a', ('inot', ('ishr', 'a', 15))) => ('imax', 'a', 0) */
   /* search444_0 -> 11 in the cache */
   /* search444_1_0_0 -> 11 in the cache */
   /* search444_1_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 14 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1382 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 1383 },
      -1,
   } },

   /* replace444_0 -> 11 in the cache */
   /* replace444_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 11, 818 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@16', 32768), 0) => ('ige', 'a', 0) */
   /* search445_0_0 -> 11 in the cache */
   /* search445_0_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 11, 12 },
      -1,
   } },
   /* search445_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1386, 818 },
      -1,
   } },

   /* replace445_0 -> 11 in the cache */
   /* replace445_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 818 },
      -1,
   } },

   /* ('ine', ('iand', 'a@16', 32768), 32768) => ('ige', 'a', 0) */
   /* search446_0_0 -> 11 in the cache */
   /* search446_0_1 -> 12 in the cache */
   /* search446_0 -> 1386 in the cache */
   /* search446_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1386, 12 },
      -1,
   } },

   /* replace446_0 -> 11 in the cache */
   /* replace446_1 -> 818 in the cache */
   /* replace446 -> 1388 in the cache */

   /* ('ine', ('iand', 'a@16', 32768), 0) => ('ilt', 'a', 0) */
   /* search447_0_0 -> 11 in the cache */
   /* search447_0_1 -> 12 in the cache */
   /* search447_0 -> 1386 in the cache */
   /* search447_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1386, 818 },
      -1,
   } },

   /* replace447_0 -> 11 in the cache */
   /* replace447_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 818 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@16', 32768), 32768) => ('ilt', 'a', 0) */
   /* search448_0_0 -> 11 in the cache */
   /* search448_0_1 -> 12 in the cache */
   /* search448_0 -> 1386 in the cache */
   /* search448_1 -> 12 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1386, 12 },
      -1,
   } },

   /* replace448_0 -> 11 in the cache */
   /* replace448_1 -> 818 in the cache */
   /* replace448 -> 1391 in the cache */

   /* ('ine', ('ushr', 'a@16', 15), 0) => ('ilt', 'a', 0) */
   /* search449_0_0 -> 11 in the cache */
   /* search449_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 14 },
      -1,
   } },
   /* search449_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1393, 818 },
      -1,
   } },

   /* replace449_0 -> 11 in the cache */
   /* replace449_1 -> 818 in the cache */
   /* replace449 -> 1391 in the cache */

   /* ('ieq', ('ushr', 'a@16', 15), 0) => ('ige', 'a', 0) */
   /* search450_0_0 -> 11 in the cache */
   /* search450_0_1 -> 14 in the cache */
   /* search450_0 -> 1393 in the cache */
   /* search450_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1393, 818 },
      -1,
   } },

   /* replace450_0 -> 11 in the cache */
   /* replace450_1 -> 818 in the cache */
   /* replace450 -> 1388 in the cache */

   /* ('ieq', ('ushr', 'a@16', 15), 1) => ('ilt', 'a', 0) */
   /* search451_0_0 -> 11 in the cache */
   /* search451_0_1 -> 14 in the cache */
   /* search451_0 -> 1393 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1393, 1396 },
      -1,
   } },

   /* replace451_0 -> 11 in the cache */
   /* replace451_1 -> 818 in the cache */
   /* replace451 -> 1391 in the cache */

   /* ('ine', ('ushr', 'a@16', 15), 1) => ('ige', 'a', 0) */
   /* search452_0_0 -> 11 in the cache */
   /* search452_0_1 -> 14 in the cache */
   /* search452_0 -> 1393 in the cache */
   /* search452_1 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1393, 1396 },
      -1,
   } },

   /* replace452_0 -> 11 in the cache */
   /* replace452_1 -> 818 in the cache */
   /* replace452 -> 1388 in the cache */

   /* ('ine', ('ishr', 'a@16', 15), 0) => ('ilt', 'a', 0) */
   /* search453_0_0 -> 11 in the cache */
   /* search453_0_1 -> 14 in the cache */
   /* search453_0 -> 1382 in the cache */
   /* search453_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1382, 818 },
      -1,
   } },

   /* replace453_0 -> 11 in the cache */
   /* replace453_1 -> 818 in the cache */
   /* replace453 -> 1391 in the cache */

   /* ('ieq', ('ishr', 'a@16', 15), 0) => ('ige', 'a', 0) */
   /* search454_0_0 -> 11 in the cache */
   /* search454_0_1 -> 14 in the cache */
   /* search454_0 -> 1382 in the cache */
   /* search454_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1382, 818 },
      -1,
   } },

   /* replace454_0 -> 11 in the cache */
   /* replace454_1 -> 818 in the cache */
   /* replace454 -> 1388 in the cache */

   /* ('ieq', ('ishr', 'a@16', 15), -1) => ('ilt', 'a', 0) */
   /* search455_0_0 -> 11 in the cache */
   /* search455_0_1 -> 14 in the cache */
   /* search455_0 -> 1382 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1382, 1401 },
      -1,
   } },

   /* replace455_0 -> 11 in the cache */
   /* replace455_1 -> 818 in the cache */
   /* replace455 -> 1391 in the cache */

   /* ('ine', ('ishr', 'a@16', 15), -1) => ('ige', 'a', 0) */
   /* search456_0_0 -> 11 in the cache */
   /* search456_0_1 -> 14 in the cache */
   /* search456_0 -> 1382 in the cache */
   /* search456_1 -> 1401 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1382, 1401 },
      -1,
   } },

   /* replace456_0 -> 11 in the cache */
   /* replace456_1 -> 818 in the cache */
   /* replace456 -> 1388 in the cache */

   /* ('iand@32', 'a', ('inot', ('ishr', 'a', 31))) => ('imax', 'a', 0) */
   /* search457_0 -> 16 in the cache */
   /* search457_1_0_0 -> 16 in the cache */
   /* search457_1_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1404 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 1405 },
      -1,
   } },

   /* replace457_0 -> 16 in the cache */
   /* replace457_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 16, 124 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@32', 2147483648), 0) => ('ige', 'a', 0) */
   /* search458_0_0 -> 16 in the cache */
   /* search458_0_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 17 },
      -1,
   } },
   /* search458_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1408, 124 },
      -1,
   } },

   /* replace458_0 -> 16 in the cache */
   /* replace458_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 124 },
      -1,
   } },

   /* ('ine', ('iand', 'a@32', 2147483648), 2147483648) => ('ige', 'a', 0) */
   /* search459_0_0 -> 16 in the cache */
   /* search459_0_1 -> 17 in the cache */
   /* search459_0 -> 1408 in the cache */
   /* search459_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1408, 17 },
      -1,
   } },

   /* replace459_0 -> 16 in the cache */
   /* replace459_1 -> 124 in the cache */
   /* replace459 -> 1410 in the cache */

   /* ('ine', ('iand', 'a@32', 2147483648), 0) => ('ilt', 'a', 0) */
   /* search460_0_0 -> 16 in the cache */
   /* search460_0_1 -> 17 in the cache */
   /* search460_0 -> 1408 in the cache */
   /* search460_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1408, 124 },
      -1,
   } },

   /* replace460_0 -> 16 in the cache */
   /* replace460_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 124 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@32', 2147483648), 2147483648) => ('ilt', 'a', 0) */
   /* search461_0_0 -> 16 in the cache */
   /* search461_0_1 -> 17 in the cache */
   /* search461_0 -> 1408 in the cache */
   /* search461_1 -> 17 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1408, 17 },
      -1,
   } },

   /* replace461_0 -> 16 in the cache */
   /* replace461_1 -> 124 in the cache */
   /* replace461 -> 1413 in the cache */

   /* ('ine', ('ushr', 'a@32', 31), 0) => ('ilt', 'a', 0) */
   /* search462_0_0 -> 16 in the cache */
   /* search462_0_1 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 19 },
      -1,
   } },
   /* search462_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1415, 124 },
      -1,
   } },

   /* replace462_0 -> 16 in the cache */
   /* replace462_1 -> 124 in the cache */
   /* replace462 -> 1413 in the cache */

   /* ('ieq', ('ushr', 'a@32', 31), 0) => ('ige', 'a', 0) */
   /* search463_0_0 -> 16 in the cache */
   /* search463_0_1 -> 19 in the cache */
   /* search463_0 -> 1415 in the cache */
   /* search463_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1415, 124 },
      -1,
   } },

   /* replace463_0 -> 16 in the cache */
   /* replace463_1 -> 124 in the cache */
   /* replace463 -> 1410 in the cache */

   /* ('ieq', ('ushr', 'a@32', 31), 1) => ('ilt', 'a', 0) */
   /* search464_0_0 -> 16 in the cache */
   /* search464_0_1 -> 19 in the cache */
   /* search464_0 -> 1415 in the cache */
   /* search464_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1415, 381 },
      -1,
   } },

   /* replace464_0 -> 16 in the cache */
   /* replace464_1 -> 124 in the cache */
   /* replace464 -> 1413 in the cache */

   /* ('ine', ('ushr', 'a@32', 31), 1) => ('ige', 'a', 0) */
   /* search465_0_0 -> 16 in the cache */
   /* search465_0_1 -> 19 in the cache */
   /* search465_0 -> 1415 in the cache */
   /* search465_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1415, 381 },
      -1,
   } },

   /* replace465_0 -> 16 in the cache */
   /* replace465_1 -> 124 in the cache */
   /* replace465 -> 1410 in the cache */

   /* ('ine', ('ishr', 'a@32', 31), 0) => ('ilt', 'a', 0) */
   /* search466_0_0 -> 16 in the cache */
   /* search466_0_1 -> 19 in the cache */
   /* search466_0 -> 1404 in the cache */
   /* search466_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1404, 124 },
      -1,
   } },

   /* replace466_0 -> 16 in the cache */
   /* replace466_1 -> 124 in the cache */
   /* replace466 -> 1413 in the cache */

   /* ('ieq', ('ishr', 'a@32', 31), 0) => ('ige', 'a', 0) */
   /* search467_0_0 -> 16 in the cache */
   /* search467_0_1 -> 19 in the cache */
   /* search467_0 -> 1404 in the cache */
   /* search467_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1404, 124 },
      -1,
   } },

   /* replace467_0 -> 16 in the cache */
   /* replace467_1 -> 124 in the cache */
   /* replace467 -> 1410 in the cache */

   /* ('ieq', ('ishr', 'a@32', 31), -1) => ('ilt', 'a', 0) */
   /* search468_0_0 -> 16 in the cache */
   /* search468_0_1 -> 19 in the cache */
   /* search468_0 -> 1404 in the cache */
   /* search468_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1404, 126 },
      -1,
   } },

   /* replace468_0 -> 16 in the cache */
   /* replace468_1 -> 124 in the cache */
   /* replace468 -> 1413 in the cache */

   /* ('ine', ('ishr', 'a@32', 31), -1) => ('ige', 'a', 0) */
   /* search469_0_0 -> 16 in the cache */
   /* search469_0_1 -> 19 in the cache */
   /* search469_0 -> 1404 in the cache */
   /* search469_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1404, 126 },
      -1,
   } },

   /* replace469_0 -> 16 in the cache */
   /* replace469_1 -> 124 in the cache */
   /* replace469 -> 1410 in the cache */

   /* ('iand@64', 'a', ('inot', ('ishr', 'a', 63))) => ('imax', 'a', 0) */
   /* search470_0 -> 21 in the cache */
   /* search470_1_0_0 -> 21 in the cache */
   /* search470_1_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 24 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 1424 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 1425 },
      -1,
   } },

   /* replace470_0 -> 21 in the cache */
   /* replace470_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 21, 859 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@64', 9223372036854775808), 0) => ('ige', 'a', 0) */
   /* search471_0_0 -> 21 in the cache */
   /* search471_0_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 21, 22 },
      -1,
   } },
   /* search471_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1428, 859 },
      -1,
   } },

   /* replace471_0 -> 21 in the cache */
   /* replace471_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 859 },
      -1,
   } },

   /* ('ine', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ige', 'a', 0) */
   /* search472_0_0 -> 21 in the cache */
   /* search472_0_1 -> 22 in the cache */
   /* search472_0 -> 1428 in the cache */
   /* search472_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1428, 22 },
      -1,
   } },

   /* replace472_0 -> 21 in the cache */
   /* replace472_1 -> 859 in the cache */
   /* replace472 -> 1430 in the cache */

   /* ('ine', ('iand', 'a@64', 9223372036854775808), 0) => ('ilt', 'a', 0) */
   /* search473_0_0 -> 21 in the cache */
   /* search473_0_1 -> 22 in the cache */
   /* search473_0 -> 1428 in the cache */
   /* search473_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 1428, 859 },
      -1,
   } },

   /* replace473_0 -> 21 in the cache */
   /* replace473_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 859 },
      -1,
   } },

   /* ('ieq', ('iand', 'a@64', 9223372036854775808), 9223372036854775808) => ('ilt', 'a', 0) */
   /* search474_0_0 -> 21 in the cache */
   /* search474_0_1 -> 22 in the cache */
   /* search474_0 -> 1428 in the cache */
   /* search474_1 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 1428, 22 },
      -1,
   } },

   /* replace474_0 -> 21 in the cache */
   /* replace474_1 -> 859 in the cache */
   /* replace474 -> 1433 in the cache */

   /* ('ine', ('ushr', 'a@64', 63), 0) => ('ilt', 'a', 0) */
   /* search475_0_0 -> 21 in the cache */
   /* search475_0_1 -> 24 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 24 },
      -1,
   } },
   /* search475_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1435, 859 },
      -1,
   } },

   /* replace475_0 -> 21 in the cache */
   /* replace475_1 -> 859 in the cache */
   /* replace475 -> 1433 in the cache */

   /* ('ieq', ('ushr', 'a@64', 63), 0) => ('ige', 'a', 0) */
   /* search476_0_0 -> 21 in the cache */
   /* search476_0_1 -> 24 in the cache */
   /* search476_0 -> 1435 in the cache */
   /* search476_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1435, 859 },
      -1,
   } },

   /* replace476_0 -> 21 in the cache */
   /* replace476_1 -> 859 in the cache */
   /* replace476 -> 1430 in the cache */

   /* ('ieq', ('ushr', 'a@64', 63), 1) => ('ilt', 'a', 0) */
   /* search477_0_0 -> 21 in the cache */
   /* search477_0_1 -> 24 in the cache */
   /* search477_0 -> 1435 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1435, 1438 },
      -1,
   } },

   /* replace477_0 -> 21 in the cache */
   /* replace477_1 -> 859 in the cache */
   /* replace477 -> 1433 in the cache */

   /* ('ine', ('ushr', 'a@64', 63), 1) => ('ige', 'a', 0) */
   /* search478_0_0 -> 21 in the cache */
   /* search478_0_1 -> 24 in the cache */
   /* search478_0 -> 1435 in the cache */
   /* search478_1 -> 1438 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1435, 1438 },
      -1,
   } },

   /* replace478_0 -> 21 in the cache */
   /* replace478_1 -> 859 in the cache */
   /* replace478 -> 1430 in the cache */

   /* ('ine', ('ishr', 'a@64', 63), 0) => ('ilt', 'a', 0) */
   /* search479_0_0 -> 21 in the cache */
   /* search479_0_1 -> 24 in the cache */
   /* search479_0 -> 1424 in the cache */
   /* search479_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1424, 859 },
      -1,
   } },

   /* replace479_0 -> 21 in the cache */
   /* replace479_1 -> 859 in the cache */
   /* replace479 -> 1433 in the cache */

   /* ('ieq', ('ishr', 'a@64', 63), 0) => ('ige', 'a', 0) */
   /* search480_0_0 -> 21 in the cache */
   /* search480_0_1 -> 24 in the cache */
   /* search480_0 -> 1424 in the cache */
   /* search480_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1424, 859 },
      -1,
   } },

   /* replace480_0 -> 21 in the cache */
   /* replace480_1 -> 859 in the cache */
   /* replace480 -> 1430 in the cache */

   /* ('ieq', ('ishr', 'a@64', 63), -1) => ('ilt', 'a', 0) */
   /* search481_0_0 -> 21 in the cache */
   /* search481_0_1 -> 24 in the cache */
   /* search481_0 -> 1424 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1424, 1443 },
      -1,
   } },

   /* replace481_0 -> 21 in the cache */
   /* replace481_1 -> 859 in the cache */
   /* replace481 -> 1433 in the cache */

   /* ('ine', ('ishr', 'a@64', 63), -1) => ('ige', 'a', 0) */
   /* search482_0_0 -> 21 in the cache */
   /* search482_0_1 -> 24 in the cache */
   /* search482_0 -> 1424 in the cache */
   /* search482_1 -> 1443 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1424, 1443 },
      -1,
   } },

   /* replace482_0 -> 21 in the cache */
   /* replace482_1 -> 859 in the cache */
   /* replace482 -> 1430 in the cache */

   /* ('fmin', 'a', ('fneg', 'a')) => ('fneg', ('fabs', 'a')) */
   /* search483_0 -> 32 in the cache */
   /* search483_1_0 -> 32 in the cache */
   /* search483_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 98 },
      -1,
   } },

   /* replace483_0_0 -> 32 in the cache */
   /* replace483_0 -> 103 in the cache */
   /* replace483 -> 1211 in the cache */

   /* ('imin', 'a', ('ineg', 'a')) => ('ineg', ('iabs', 'a')) */
   /* search484_0 -> 32 in the cache */
   /* search484_1_0 -> 32 in the cache */
   /* search484_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 100 },
      -1,
   } },

   /* replace484_0_0 -> 32 in the cache */
   /* replace484_0 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 106 },
      -1,
   } },

   /* ('fmin', 'a', ('fneg', ('fabs', 'a'))) => ('fneg', ('fabs', 'a')) */
   /* search485_0 -> 32 in the cache */
   /* search485_1_0_0 -> 32 in the cache */
   /* search485_1_0 -> 103 in the cache */
   /* search485_1 -> 1211 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 1211 },
      -1,
   } },

   /* replace485_0_0 -> 32 in the cache */
   /* replace485_0 -> 103 in the cache */
   /* replace485 -> 1211 in the cache */

   /* ('imin', 'a', ('ineg', ('iabs', 'a'))) => ('ineg', ('iabs', 'a')) */
   /* search486_0 -> 32 in the cache */
   /* search486_1_0_0 -> 32 in the cache */
   /* search486_1_0 -> 106 in the cache */
   /* search486_1 -> 1448 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 1448 },
      -1,
   } },

   /* replace486_0_0 -> 32 in the cache */
   /* replace486_0 -> 106 in the cache */
   /* replace486 -> 1448 in the cache */

   /* ('~fmin', 'a', ('fabs', 'a')) => a */
   /* search487_0 -> 32 in the cache */
   /* search487_1_0 -> 32 in the cache */
   /* search487_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 32, 103 },
      -1,
   } },

   /* replace487 -> 32 in the cache */

   /* ('imin', 'a', ('iabs', 'a')) => a */
   /* search488_0 -> 32 in the cache */
   /* search488_1_0 -> 32 in the cache */
   /* search488_1 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 32, 106 },
      -1,
   } },

   /* replace488 -> 32 in the cache */

   /* ('~fmax', 'a', ('fneg', ('fabs', 'a'))) => a */
   /* search489_0 -> 32 in the cache */
   /* search489_1_0_0 -> 32 in the cache */
   /* search489_1_0 -> 103 in the cache */
   /* search489_1 -> 1211 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 1211 },
      -1,
   } },

   /* replace489 -> 32 in the cache */

   /* ('imax', 'a', ('ineg', ('iabs', 'a'))) => a */
   /* search490_0 -> 32 in the cache */
   /* search490_1_0_0 -> 32 in the cache */
   /* search490_1_0 -> 106 in the cache */
   /* search490_1 -> 1448 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 1448 },
      -1,
   } },

   /* replace490 -> 32 in the cache */

   /* ('fmax', 'a', ('fabs', 'a')) => ('fabs', 'a') */
   /* search491_0 -> 32 in the cache */
   /* search491_1_0 -> 32 in the cache */
   /* search491_1 -> 103 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 103 },
      -1,
   } },

   /* replace491_0 -> 32 in the cache */
   /* replace491 -> 103 in the cache */

   /* ('imax', 'a', ('iabs', 'a')) => ('iabs', 'a') */
   /* search492_0 -> 32 in the cache */
   /* search492_1_0 -> 32 in the cache */
   /* search492_1 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 106 },
      -1,
   } },

   /* replace492_0 -> 32 in the cache */
   /* replace492 -> 106 in the cache */

   /* ('fmax', 'a', ('fneg', 'a')) => ('fabs', 'a') */
   /* search493_0 -> 32 in the cache */
   /* search493_1_0 -> 32 in the cache */
   /* search493_1 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 98 },
      -1,
   } },

   /* replace493_0 -> 32 in the cache */
   /* replace493 -> 103 in the cache */

   /* ('imax', 'a', ('ineg', 'a')) => ('iabs', 'a') */
   /* search494_0 -> 32 in the cache */
   /* search494_1_0 -> 32 in the cache */
   /* search494_1 -> 100 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 32, 100 },
      -1,
   } },

   /* replace494_0 -> 32 in the cache */
   /* replace494 -> 106 in the cache */

   /* ('~fmax', ('fabs', 'a'), 0.0) => ('fabs', 'a') */
   /* search495_0_0 -> 32 in the cache */
   /* search495_0 -> 103 in the cache */
   /* search495_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 103, 113 },
      -1,
   } },

   /* replace495_0 -> 32 in the cache */
   /* replace495 -> 103 in the cache */

   /* ('fmin', ('fmax', 'a', 0.0), 1.0) => ('fsat', 'a') */
   /* search496_0_0 -> 32 in the cache */
   /* search496_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 32, 113 },
      -1,
   } },
   /* search496_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1460, 230 },
      -1,
   } },

   /* replace496_0 -> 32 in the cache */
   /* replace496 -> 206 in the cache */

   /* ('~fmax', ('fmin', 'a', 1.0), 0.0) => ('fsat', 'a') */
   /* search497_0_0 -> 32 in the cache */
   /* search497_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 32, 230 },
      -1,
   } },
   /* search497_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1462, 113 },
      -1,
   } },

   /* replace497_0 -> 32 in the cache */
   /* replace497 -> 206 in the cache */

   /* ('~fmin', ('fmax', 'a', -1.0), 0.0) => ('fneg', ('fsat', ('fneg', 'a'))) */
   /* search498_0_0 -> 32 in the cache */
   /* search498_0_1 -> 236 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 32, 236 },
      -1,
   } },
   /* search498_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1464, 113 },
      -1,
   } },

   /* replace498_0_0_0 -> 32 in the cache */
   /* replace498_0_0 -> 98 in the cache */
   /* replace498_0 -> 207 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 207 },
      -1,
   } },

   /* ('~fmax', ('fmin', 'a', 0.0), -1.0) => ('fneg', ('fsat', ('fneg', 'a'))) */
   /* search499_0_0 -> 32 in the cache */
   /* search499_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 32, 113 },
      -1,
   } },
   /* search499_1 -> 236 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1467, 236 },
      -1,
   } },

   /* replace499_0_0_0 -> 32 in the cache */
   /* replace499_0_0 -> 98 in the cache */
   /* replace499_0 -> 207 in the cache */
   /* replace499 -> 1466 in the cache */

   /* ('fsat', ('fsign', 'a')) => ('b2f', ('!flt', 0.0, 'a')) */
   /* search500_0_0 -> 32 in the cache */
   /* search500_0 -> 239 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 239 },
      -1,
   } },

   /* replace500_0_0 -> 113 in the cache */
   /* replace500_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 113, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1470 },
      -1,
   } },

   /* ('fsat', ('b2f', 'a')) => ('b2f', 'a') */
   /* search501_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 1472 },
      -1,
   } },

   /* replace501_0 -> 32 in the cache */
   /* replace501 -> 1472 in the cache */

   /* ('fsat', 'a') => ('fmin', ('fmax', 'a', 0.0), 1.0) */
   /* search502_0 -> 32 in the cache */
   /* search502 -> 206 in the cache */

   /* replace502_0_0 -> 32 in the cache */
   /* replace502_0_1 -> 113 in the cache */
   /* replace502_0 -> 1460 in the cache */
   /* replace502_1 -> 230 in the cache */
   /* replace502 -> 1461 in the cache */

   /* ('fsat', ('fsat', 'a')) => ('fsat', 'a') */
   /* search503_0_0 -> 32 in the cache */
   /* search503_0 -> 206 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 206 },
      -1,
   } },

   /* replace503_0 -> 32 in the cache */
   /* replace503 -> 206 in the cache */

   /* ('fsat', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fsat', ('fadd', ('fneg', 'a'), ('fneg', 'b'))) */
   /* search504_0_0_0 -> 0 in the cache */
   /* search504_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1475 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1476 },
      -1,
   } },

   /* replace504_0_0_0 -> 0 in the cache */
   /* replace504_0_0 -> 201 in the cache */
   /* replace504_0_1_0 -> 3 in the cache */
   /* replace504_0_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 201, 1049 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1478 },
      -1,
   } },

   /* ('fsat', ('fneg(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fneg', 'a'), 'b')) */
   /* search505_0_0_0 -> 0 in the cache */
   /* search505_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1480 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1481 },
      -1,
   } },

   /* replace505_0_0_0 -> 0 in the cache */
   /* replace505_0_0 -> 201 in the cache */
   /* replace505_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 201, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1483 },
      -1,
   } },

   /* ('fsat', ('fneg(is_used_once)', ('fmulz(is_used_once)', 'a', 'b'))) => ('fsat', ('fmulz', ('fneg', 'a'), 'b')) */
   /* search506_0_0_0 -> 16 in the cache */
   /* search506_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 16, 36 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1485 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1486 },
      -1,
   } },

   /* replace506_0_0_0 -> 16 in the cache */
   /* replace506_0_0 -> 269 in the cache */
   /* replace506_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 269, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1488 },
      -1,
   } },

   /* ('fsat', ('fabs(is_used_once)', ('fmul(is_used_once)', 'a', 'b'))) => ('fsat', ('fmul', ('fabs', 'a'), ('fabs', 'b'))) */
   /* search507_0_0_0 -> 0 in the cache */
   /* search507_0_0_1 -> 3 in the cache */
   /* search507_0_0 -> 1480 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 1480 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1490 },
      -1,
   } },

   /* replace507_0_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace507_0_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1492, 1493 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1494 },
      -1,
   } },

   /* ('fmin', ('fmax', ('fmin', ('fmax', 'a', 'b'), 'c'), 'b'), 'c') => ('fmin', ('fmax', 'a', 'b'), 'c') */
   /* search508_0_0_0_0 -> 128 in the cache */
   /* search508_0_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      3, 1,
      { 128, 129 },
      -1,
   } },
   /* search508_0_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      2, 2,
      { 1496, 131 },
      -1,
   } },
   /* search508_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 3,
      { 1497, 129 },
      -1,
   } },
   /* search508_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 4,
      { 1498, 131 },
      -1,
   } },

   /* replace508_0_0 -> 128 in the cache */
   /* replace508_0_1 -> 129 in the cache */
   /* replace508_0 -> 1304 in the cache */
   /* replace508_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1304, 131 },
      -1,
   } },

   /* ('imin', ('imax', ('imin', ('imax', 'a', 'b'), 'c'), 'b'), 'c') => ('imin', ('imax', 'a', 'b'), 'c') */
   /* search509_0_0_0_0 -> 128 in the cache */
   /* search509_0_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      3, 1,
      { 128, 129 },
      -1,
   } },
   /* search509_0_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      2, 2,
      { 1501, 131 },
      -1,
   } },
   /* search509_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      1, 3,
      { 1502, 129 },
      -1,
   } },
   /* search509_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 4,
      { 1503, 131 },
      -1,
   } },

   /* replace509_0_0 -> 128 in the cache */
   /* replace509_0_1 -> 129 in the cache */
   /* replace509_0 -> 1314 in the cache */
   /* replace509_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1314, 131 },
      -1,
   } },

   /* ('umin', ('umax', ('umin', ('umax', 'a', 'b'), 'c'), 'b'), 'c') => ('umin', ('umax', 'a', 'b'), 'c') */
   /* search510_0_0_0_0 -> 128 in the cache */
   /* search510_0_0_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      3, 1,
      { 128, 129 },
      -1,
   } },
   /* search510_0_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      2, 2,
      { 1506, 131 },
      -1,
   } },
   /* search510_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      1, 3,
      { 1507, 129 },
      -1,
   } },
   /* search510_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 4,
      { 1508, 131 },
      -1,
   } },

   /* replace510_0_0 -> 128 in the cache */
   /* replace510_0_1 -> 129 in the cache */
   /* replace510_0 -> 1309 in the cache */
   /* replace510_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 2,
      { 1309, 131 },
      -1,
   } },

   /* ('fmax', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmax', 'a', 'b')) */
   /* search511_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      18,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1511, 1512 },
      -1,
   } },

   /* replace511_0_0 -> 0 in the cache */
   /* replace511_0_1 -> 3 in the cache */
   /* replace511_0 -> 1238 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1238 },
      -1,
   } },

   /* ('~fmin', ('fsat', 'a'), '#b(is_zero_to_one)') => ('fsat', ('fmin', 'a', 'b')) */
   /* search512_0_0 -> 0 in the cache */
   /* search512_0 -> 1511 in the cache */
   /* search512_1 -> 1512 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1511, 1512 },
      -1,
   } },

   /* replace512_0_0 -> 0 in the cache */
   /* replace512_0_1 -> 3 in the cache */
   /* replace512_0 -> 1240 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1240 },
      -1,
   } },

   /* ('fmax', ('fneg', ('fmin', 'b', 'a')), 'b') => ('fmax', ('fabs', 'b'), ('fneg', 'a')) */
   /* search513_0_0_0 -> 1073 in the cache */
   /* search513_0_0_1 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 1073, 1070 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1517 },
      -1,
   } },
   /* search513_1 -> 1073 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1518, 1073 },
      -1,
   } },

   /* replace513_0_0 -> 1073 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 1073 },
      -1,
   } },
   /* replace513_1_0 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1070 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1520, 1521 },
      -1,
   } },

   /* ('fmin', ('fneg', ('fmax', 'b', 'a')), 'b') => ('fmin', ('fneg', ('fabs', 'b')), ('fneg', 'a')) */
   /* search514_0_0_0 -> 1073 in the cache */
   /* search514_0_0_1 -> 1070 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 1073, 1070 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 1523 },
      -1,
   } },
   /* search514_1 -> 1073 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 1524, 1073 },
      -1,
   } },

   /* replace514_0_0_0 -> 1073 in the cache */
   /* replace514_0_0 -> 1520 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1520 },
      -1,
   } },
   /* replace514_1_0 -> 1070 in the cache */
   /* replace514_1 -> 1521 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1526, 1521 },
      -1,
   } },

   /* ('fmax', ('fadd(is_used_once)', ('fneg', 'a(is_not_negative)'), '#b(is_zero_to_one)'), 0.0) => ('fsat', ('fadd', ('fneg', 'a'), 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1528 },
      -1,
   } },
   /* search515_0_1 -> 1512 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 1529, 1512 },
      0,
   } },
   /* search515_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 1530, 245 },
      -1,
   } },

   /* replace515_0_0_0 -> 0 in the cache */
   /* replace515_0_0 -> 201 in the cache */
   /* replace515_0_1 -> 3 in the cache */
   /* replace515_0 -> 266 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 266 },
      -1,
   } },

   /* ('extract_u8', ('imin', ('imax', 'a', 0), 255), 0) => ('imin', ('imax', 'a', 0), 255) */
   /* search516_0_0_0 -> 32 in the cache */
   /* search516_0_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 32, 61 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 1533, 1534 },
      -1,
   } },
   /* search516_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 2,
      { 1535, 61 },
      -1,
   } },

   /* replace516_0_0 -> 32 in the cache */
   /* replace516_0_1 -> 61 in the cache */
   /* replace516_0 -> 1533 in the cache */
   /* replace516_1 -> 1534 in the cache */
   /* replace516 -> 1535 in the cache */

   /* ('ior', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('!fmax', 'b', 'c')) */
   /* search517_0_0 -> 128 in the cache */
   /* search517_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 129 },
      0,
   } },
   /* search517_1_0 -> 128 in the cache */
   /* search517_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1537, 1538 },
      -1,
   } },

   /* replace517_0 -> 128 in the cache */
   /* replace517_1_0 -> 129 in the cache */
   /* replace517_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      true,
      false,
      nir_op_fmax,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 128, 1540 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c') */
   /* search518_0_0 -> 0 in the cache */
   /* search518_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 669 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search518_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1543, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1542, 1544 },
      -1,
   } },

   /* replace518_0_0 -> 0 in the cache */
   /* replace518_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace518_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1546, 669 },
      -1,
   } },

   /* ('ior', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('!fmin', 'b', 'c')) */
   /* search519_0_0 -> 128 in the cache */
   /* search519_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 128, 129 },
      0,
   } },
   /* search519_1_0 -> 128 in the cache */
   /* search519_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1548, 1549 },
      -1,
   } },

   /* replace519_0 -> 128 in the cache */
   /* replace519_1_0 -> 129 in the cache */
   /* replace519_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      true,
      false,
      nir_op_fmin,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 128, 1551 },
      -1,
   } },

   /* ('ior', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c') */
   /* search520_0_0 -> 0 in the cache */
   /* search520_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 669 },
      0,
   } },
   /* search520_1_0 -> 1543 in the cache */
   /* search520_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1543, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1553, 1554 },
      -1,
   } },

   /* replace520_0_0 -> 0 in the cache */
   /* replace520_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace520_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1556, 669 },
      -1,
   } },

   /* ('ior', ('flt', 'a', '#b'), ('flt', 'a', '#c')) => ('flt', 'a', ('!fmax', 'b', 'c')) */
   /* search521_0_0 -> 128 in the cache */
   /* search521_0_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 272 },
      -1,
   } },
   /* search521_1_0 -> 128 in the cache */
   /* search521_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1558, 1559 },
      -1,
   } },

   /* replace521_0 -> 128 in the cache */
   /* replace521_1_0 -> 129 in the cache */
   /* replace521_1_1 -> 131 in the cache */
   /* replace521_1 -> 1540 in the cache */
   /* replace521 -> 1541 in the cache */

   /* ('ior', ('flt', '#a', 'c'), ('flt', '#b', 'c')) => ('flt', ('!fmin', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search522_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1561, 669 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search522_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1563, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1562, 1564 },
      -1,
   } },

   /* replace522_0_0 -> 0 in the cache */
   /* replace522_0_1 -> 1543 in the cache */
   /* replace522_0 -> 1546 in the cache */
   /* replace522_1 -> 669 in the cache */
   /* replace522 -> 1547 in the cache */

   /* ('ior', ('fge', 'a', '#b'), ('fge', 'a', '#c')) => ('fge', 'a', ('!fmin', 'b', 'c')) */
   /* search523_0_0 -> 128 in the cache */
   /* search523_0_1 -> 272 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 128, 272 },
      -1,
   } },
   /* search523_1_0 -> 128 in the cache */
   /* search523_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 128, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1566, 1567 },
      -1,
   } },

   /* replace523_0 -> 128 in the cache */
   /* replace523_1_0 -> 129 in the cache */
   /* replace523_1_1 -> 131 in the cache */
   /* replace523_1 -> 1551 in the cache */
   /* replace523 -> 1552 in the cache */

   /* ('ior', ('fge', '#a', 'c'), ('fge', '#b', 'c')) => ('fge', ('!fmax', 'a', 'b'), 'c') */
   /* search524_0_0 -> 1561 in the cache */
   /* search524_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1561, 669 },
      -1,
   } },
   /* search524_1_0 -> 1563 in the cache */
   /* search524_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1563, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1569, 1570 },
      -1,
   } },

   /* replace524_0_0 -> 0 in the cache */
   /* replace524_0_1 -> 1543 in the cache */
   /* replace524_0 -> 1556 in the cache */
   /* replace524_1 -> 669 in the cache */
   /* replace524 -> 1557 in the cache */

   /* ('~iand', ('flt(is_used_once)', 'a', 'b'), ('flt', 'a', 'c')) => ('flt', 'a', ('fmin', 'b', 'c')) */
   /* search525_0_0 -> 128 in the cache */
   /* search525_0_1 -> 129 in the cache */
   /* search525_0 -> 1537 in the cache */
   /* search525_1_0 -> 128 in the cache */
   /* search525_1_1 -> 131 in the cache */
   /* search525_1 -> 1538 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1537, 1538 },
      -1,
   } },

   /* replace525_0 -> 128 in the cache */
   /* replace525_1_0 -> 129 in the cache */
   /* replace525_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 128, 1573 },
      -1,
   } },

   /* ('~iand', ('flt(is_used_once)', 'a', 'c'), ('flt', 'b', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c') */
   /* search526_0_0 -> 0 in the cache */
   /* search526_0_1 -> 669 in the cache */
   /* search526_0 -> 1542 in the cache */
   /* search526_1_0 -> 1543 in the cache */
   /* search526_1_1 -> 669 in the cache */
   /* search526_1 -> 1544 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1542, 1544 },
      -1,
   } },

   /* replace526_0_0 -> 0 in the cache */
   /* replace526_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace526_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1576, 669 },
      -1,
   } },

   /* ('~iand', ('fge(is_used_once)', 'a', 'b'), ('fge', 'a', 'c')) => ('fge', 'a', ('fmax', 'b', 'c')) */
   /* search527_0_0 -> 128 in the cache */
   /* search527_0_1 -> 129 in the cache */
   /* search527_0 -> 1548 in the cache */
   /* search527_1_0 -> 128 in the cache */
   /* search527_1_1 -> 131 in the cache */
   /* search527_1 -> 1549 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1548, 1549 },
      -1,
   } },

   /* replace527_0 -> 128 in the cache */
   /* replace527_1_0 -> 129 in the cache */
   /* replace527_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 128, 1579 },
      -1,
   } },

   /* ('~iand', ('fge(is_used_once)', 'a', 'c'), ('fge', 'b', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c') */
   /* search528_0_0 -> 0 in the cache */
   /* search528_0_1 -> 669 in the cache */
   /* search528_0 -> 1553 in the cache */
   /* search528_1_0 -> 1543 in the cache */
   /* search528_1_1 -> 669 in the cache */
   /* search528_1 -> 1554 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1553, 1554 },
      -1,
   } },

   /* replace528_0_0 -> 0 in the cache */
   /* replace528_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace528_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1582, 669 },
      -1,
   } },

   /* ('iand', ('flt', 'a', '#b(is_a_number)'), ('flt', 'a', '#c(is_a_number)')) => ('flt', 'a', ('fmin', 'b', 'c')) */
   /* search529_0_0 -> 128 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 1584 },
      -1,
   } },
   /* search529_1_0 -> 128 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 128, 1586 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1585, 1587 },
      -1,
   } },

   /* replace529_0 -> 128 in the cache */
   /* replace529_1_0 -> 129 in the cache */
   /* replace529_1_1 -> 131 in the cache */
   /* replace529_1 -> 1573 in the cache */
   /* replace529 -> 1574 in the cache */

   /* ('iand', ('flt', '#a(is_a_number)', 'c'), ('flt', '#b(is_a_number)', 'c')) => ('flt', ('fmax', 'a', 'b'), 'c') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search530_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1589, 669 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      2, /* b */
      true,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search530_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1591, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1590, 1592 },
      -1,
   } },

   /* replace530_0_0 -> 0 in the cache */
   /* replace530_0_1 -> 1543 in the cache */
   /* replace530_0 -> 1576 in the cache */
   /* replace530_1 -> 669 in the cache */
   /* replace530 -> 1577 in the cache */

   /* ('iand', ('fge', 'a', '#b(is_a_number)'), ('fge', 'a', '#c(is_a_number)')) => ('fge', 'a', ('fmax', 'b', 'c')) */
   /* search531_0_0 -> 128 in the cache */
   /* search531_0_1 -> 1584 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 128, 1584 },
      -1,
   } },
   /* search531_1_0 -> 128 in the cache */
   /* search531_1_1 -> 1586 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 128, 1586 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1594, 1595 },
      -1,
   } },

   /* replace531_0 -> 128 in the cache */
   /* replace531_1_0 -> 129 in the cache */
   /* replace531_1_1 -> 131 in the cache */
   /* replace531_1 -> 1579 in the cache */
   /* replace531 -> 1580 in the cache */

   /* ('iand', ('fge', '#a(is_a_number)', 'c'), ('fge', '#b(is_a_number)', 'c')) => ('fge', ('fmin', 'a', 'b'), 'c') */
   /* search532_0_0 -> 1589 in the cache */
   /* search532_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1589, 669 },
      -1,
   } },
   /* search532_1_0 -> 1591 in the cache */
   /* search532_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1591, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1597, 1598 },
      -1,
   } },

   /* replace532_0_0 -> 0 in the cache */
   /* replace532_0_1 -> 1543 in the cache */
   /* replace532_0 -> 1582 in the cache */
   /* replace532_1 -> 669 in the cache */
   /* replace532 -> 1583 in the cache */

   /* ('ior', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imax', 'b', 'c')) */
   /* search533_0_0 -> 128 in the cache */
   /* search533_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 128, 129 },
      0,
   } },
   /* search533_1_0 -> 128 in the cache */
   /* search533_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1600, 1601 },
      -1,
   } },

   /* replace533_0 -> 128 in the cache */
   /* replace533_1_0 -> 129 in the cache */
   /* replace533_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 128, 1603 },
      -1,
   } },

   /* ('ior', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imin', 'a', 'b'), 'c') */
   /* search534_0_0 -> 0 in the cache */
   /* search534_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 0, 669 },
      0,
   } },
   /* search534_1_0 -> 1543 in the cache */
   /* search534_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1543, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1605, 1606 },
      -1,
   } },

   /* replace534_0_0 -> 0 in the cache */
   /* replace534_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace534_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1608, 669 },
      -1,
   } },

   /* ('ior', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imin', 'b', 'c')) */
   /* search535_0_0 -> 128 in the cache */
   /* search535_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 128, 129 },
      0,
   } },
   /* search535_1_0 -> 128 in the cache */
   /* search535_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1610, 1611 },
      -1,
   } },

   /* replace535_0 -> 128 in the cache */
   /* replace535_1_0 -> 129 in the cache */
   /* replace535_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 128, 1613 },
      -1,
   } },

   /* ('ior', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imax', 'a', 'b'), 'c') */
   /* search536_0_0 -> 0 in the cache */
   /* search536_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 0, 669 },
      0,
   } },
   /* search536_1_0 -> 1543 in the cache */
   /* search536_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1543, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1615, 1616 },
      -1,
   } },

   /* replace536_0_0 -> 0 in the cache */
   /* replace536_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace536_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1618, 669 },
      -1,
   } },

   /* ('ior', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umax', 'b', 'c')) */
   /* search537_0_0 -> 128 in the cache */
   /* search537_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 128, 129 },
      0,
   } },
   /* search537_1_0 -> 128 in the cache */
   /* search537_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1620, 1621 },
      -1,
   } },

   /* replace537_0 -> 128 in the cache */
   /* replace537_1_0 -> 129 in the cache */
   /* replace537_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 128, 1623 },
      -1,
   } },

   /* ('ior', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umin', 'a', 'b'), 'c') */
   /* search538_0_0 -> 0 in the cache */
   /* search538_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 0, 669 },
      0,
   } },
   /* search538_1_0 -> 1543 in the cache */
   /* search538_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1543, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1625, 1626 },
      -1,
   } },

   /* replace538_0_0 -> 0 in the cache */
   /* replace538_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace538_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1628, 669 },
      -1,
   } },

   /* ('ior', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umin', 'b', 'c')) */
   /* search539_0_0 -> 128 in the cache */
   /* search539_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 128, 129 },
      0,
   } },
   /* search539_1_0 -> 128 in the cache */
   /* search539_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 128, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1630, 1631 },
      -1,
   } },

   /* replace539_0 -> 128 in the cache */
   /* replace539_1_0 -> 129 in the cache */
   /* replace539_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 128, 1633 },
      -1,
   } },

   /* ('ior', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umax', 'a', 'b'), 'c') */
   /* search540_0_0 -> 0 in the cache */
   /* search540_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 0, 669 },
      0,
   } },
   /* search540_1_0 -> 1543 in the cache */
   /* search540_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1543, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1635, 1636 },
      -1,
   } },

   /* replace540_0_0 -> 0 in the cache */
   /* replace540_0_1 -> 1543 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 0, 1543 },
      -1,
   } },
   /* replace540_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1638, 669 },
      -1,
   } },

   /* ('iand', ('ilt(is_used_once)', 'a', 'b'), ('ilt', 'a', 'c')) => ('ilt', 'a', ('imin', 'b', 'c')) */
   /* search541_0_0 -> 128 in the cache */
   /* search541_0_1 -> 129 in the cache */
   /* search541_0 -> 1600 in the cache */
   /* search541_1_0 -> 128 in the cache */
   /* search541_1_1 -> 131 in the cache */
   /* search541_1 -> 1601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1600, 1601 },
      -1,
   } },

   /* replace541_0 -> 128 in the cache */
   /* replace541_1_0 -> 129 in the cache */
   /* replace541_1_1 -> 131 in the cache */
   /* replace541_1 -> 1613 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 128, 1613 },
      -1,
   } },

   /* ('iand', ('ilt(is_used_once)', 'a', 'c'), ('ilt', 'b', 'c')) => ('ilt', ('imax', 'a', 'b'), 'c') */
   /* search542_0_0 -> 0 in the cache */
   /* search542_0_1 -> 669 in the cache */
   /* search542_0 -> 1605 in the cache */
   /* search542_1_0 -> 1543 in the cache */
   /* search542_1_1 -> 669 in the cache */
   /* search542_1 -> 1606 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1605, 1606 },
      -1,
   } },

   /* replace542_0_0 -> 0 in the cache */
   /* replace542_0_1 -> 1543 in the cache */
   /* replace542_0 -> 1618 in the cache */
   /* replace542_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1618, 669 },
      -1,
   } },

   /* ('iand', ('ige(is_used_once)', 'a', 'b'), ('ige', 'a', 'c')) => ('ige', 'a', ('imax', 'b', 'c')) */
   /* search543_0_0 -> 128 in the cache */
   /* search543_0_1 -> 129 in the cache */
   /* search543_0 -> 1610 in the cache */
   /* search543_1_0 -> 128 in the cache */
   /* search543_1_1 -> 131 in the cache */
   /* search543_1 -> 1611 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1610, 1611 },
      -1,
   } },

   /* replace543_0 -> 128 in the cache */
   /* replace543_1_0 -> 129 in the cache */
   /* replace543_1_1 -> 131 in the cache */
   /* replace543_1 -> 1603 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 128, 1603 },
      -1,
   } },

   /* ('iand', ('ige(is_used_once)', 'a', 'c'), ('ige', 'b', 'c')) => ('ige', ('imin', 'a', 'b'), 'c') */
   /* search544_0_0 -> 0 in the cache */
   /* search544_0_1 -> 669 in the cache */
   /* search544_0 -> 1615 in the cache */
   /* search544_1_0 -> 1543 in the cache */
   /* search544_1_1 -> 669 in the cache */
   /* search544_1 -> 1616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1615, 1616 },
      -1,
   } },

   /* replace544_0_0 -> 0 in the cache */
   /* replace544_0_1 -> 1543 in the cache */
   /* replace544_0 -> 1608 in the cache */
   /* replace544_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1608, 669 },
      -1,
   } },

   /* ('iand', ('ult(is_used_once)', 'a', 'b'), ('ult', 'a', 'c')) => ('ult', 'a', ('umin', 'b', 'c')) */
   /* search545_0_0 -> 128 in the cache */
   /* search545_0_1 -> 129 in the cache */
   /* search545_0 -> 1620 in the cache */
   /* search545_1_0 -> 128 in the cache */
   /* search545_1_1 -> 131 in the cache */
   /* search545_1 -> 1621 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1620, 1621 },
      -1,
   } },

   /* replace545_0 -> 128 in the cache */
   /* replace545_1_0 -> 129 in the cache */
   /* replace545_1_1 -> 131 in the cache */
   /* replace545_1 -> 1633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 128, 1633 },
      -1,
   } },

   /* ('iand', ('ult(is_used_once)', 'a', 'c'), ('ult', 'b', 'c')) => ('ult', ('umax', 'a', 'b'), 'c') */
   /* search546_0_0 -> 0 in the cache */
   /* search546_0_1 -> 669 in the cache */
   /* search546_0 -> 1625 in the cache */
   /* search546_1_0 -> 1543 in the cache */
   /* search546_1_1 -> 669 in the cache */
   /* search546_1 -> 1626 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1625, 1626 },
      -1,
   } },

   /* replace546_0_0 -> 0 in the cache */
   /* replace546_0_1 -> 1543 in the cache */
   /* replace546_0 -> 1638 in the cache */
   /* replace546_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1638, 669 },
      -1,
   } },

   /* ('iand', ('uge(is_used_once)', 'a', 'b'), ('uge', 'a', 'c')) => ('uge', 'a', ('umax', 'b', 'c')) */
   /* search547_0_0 -> 128 in the cache */
   /* search547_0_1 -> 129 in the cache */
   /* search547_0 -> 1630 in the cache */
   /* search547_1_0 -> 128 in the cache */
   /* search547_1_1 -> 131 in the cache */
   /* search547_1 -> 1631 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1630, 1631 },
      -1,
   } },

   /* replace547_0 -> 128 in the cache */
   /* replace547_1_0 -> 129 in the cache */
   /* replace547_1_1 -> 131 in the cache */
   /* replace547_1 -> 1623 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 128, 1623 },
      -1,
   } },

   /* ('iand', ('uge(is_used_once)', 'a', 'c'), ('uge', 'b', 'c')) => ('uge', ('umin', 'a', 'b'), 'c') */
   /* search548_0_0 -> 0 in the cache */
   /* search548_0_1 -> 669 in the cache */
   /* search548_0 -> 1635 in the cache */
   /* search548_1_0 -> 1543 in the cache */
   /* search548_1_1 -> 669 in the cache */
   /* search548_1 -> 1636 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1635, 1636 },
      -1,
   } },

   /* replace548_0_0 -> 0 in the cache */
   /* replace548_0_1 -> 1543 in the cache */
   /* replace548_0 -> 1628 in the cache */
   /* replace548_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1628, 669 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'c'), 'd'), ('flt', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd') */
   /* search549_0_0_0 -> 0 in the cache */
   /* search549_0_0_1 -> 669 in the cache */
   /* search549_0_0 -> 1542 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      2, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1542, 1656 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      3, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search549_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1658, 669 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1657, 1659 },
      -1,
   } },

   /* replace549_0_0_0 -> 0 in the cache */
   /* replace549_0_0_1 -> 1658 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      true,
      false,
      nir_op_fmin,
      1, 1,
      { 0, 1658 },
      -1,
   } },
   /* replace549_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1661, 669 },
      -1,
   } },
   /* replace549_1 -> 1656 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1662, 1656 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt', 'a', 'c'), 'd'), ('flt(is_used_once)', 'b', 'c')) => ('ior', ('flt', ('!fmin', 'a', 'b'), 'c'), 'd') */
   /* search550_0_0_0 -> 0 in the cache */
   /* search550_0_0_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 669 },
      -1,
   } },
   /* search550_0_1 -> 1656 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1664, 1656 },
      0,
   } },
   /* search550_1_0 -> 1658 in the cache */
   /* search550_1_1 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1658, 669 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1665, 1666 },
      -1,
   } },

   /* replace550_0_0_0 -> 0 in the cache */
   /* replace550_0_0_1 -> 1658 in the cache */
   /* replace550_0_0 -> 1661 in the cache */
   /* replace550_0_1 -> 669 in the cache */
   /* replace550_0 -> 1662 in the cache */
   /* replace550_1 -> 1656 in the cache */
   /* replace550 -> 1663 in the cache */

   /* ('ior', ('ior(is_used_once)', ('flt(is_used_once)', 'a', 'b'), 'd'), ('flt', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd') */
   /* search551_0_0_0 -> 151 in the cache */
   /* search551_0_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 151, 152 },
      0,
   } },
   /* search551_0_1 -> 1656 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1668, 1656 },
      0,
   } },
   /* search551_1_0 -> 151 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 151, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1669, 1671 },
      -1,
   } },

   /* replace551_0_0 -> 151 in the cache */
   /* replace551_0_1_0 -> 152 in the cache */
   /* replace551_0_1_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      true,
      false,
      nir_op_fmax,
      1, 1,
      { 152, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 151, 1673 },
      -1,
   } },
   /* replace551_1 -> 1656 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1674, 1656 },
      -1,
   } },

   /* ('ior', ('ior(is_used_once)', ('flt', 'a', 'b'), 'd'), ('flt(is_used_once)', 'a', 'c')) => ('ior', ('flt', 'a', ('!fmax', 'b', 'c')), 'd') */
   /* search552_0_0_0 -> 151 in the cache */
   /* search552_0_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 151, 152 },
      -1,
   } },
   /* search552_0_1 -> 1656 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 1676, 1656 },
      0,
   } },
   /* search552_1_0 -> 151 in the cache */
   /* search552_1_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 151, 1670 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 1677, 1678 },
      -1,
   } },

   /* replace552_0_0 -> 151 in the cache */
   /* replace552_0_1_0 -> 152 in the cache */
   /* replace552_0_1_1 -> 1670 in the cache */
   /* replace552_0_1 -> 1673 in the cache */
   /* replace552_0 -> 1674 in the cache */
   /* replace552_1 -> 1656 in the cache */
   /* replace552 -> 1675 in the cache */

   /* ('ior', ('flt', 'a(is_a_number)', 'b(is_a_number)'), ('flt', 'b', 'a')) => ('fneu', 'a', 'b') */
   /* search553_0_0 -> 1004 in the cache */
   /* search553_0_1 -> 1005 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1004, 1005 },
      -1,
   } },
   /* search553_1_0 -> 3 in the cache */
   /* search553_1_1 -> 0 in the cache */
   /* search553_1 -> 1040 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1680, 1040 },
      -1,
   } },

   /* replace553_0 -> 0 in the cache */
   /* replace553_1 -> 3 in the cache */
   /* replace553 -> 1014 in the cache */

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search554_0_0 -> 115 in the cache */
   /* search554_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 115, 11 },
      0,
   } },
   /* search554_1_0 -> 247 in the cache */
   /* search554_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 247, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1682, 1683 },
      -1,
   } },

   /* replace554_0 -> 115 in the cache */
   /* replace554_1_0 -> 11 in the cache */
   /* replace554_1_1_0 -> 247 in the cache */
   /* replace554_1_1 -> 519 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 11, 519 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 115, 1685 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search555_0_0 -> 115 in the cache */
   /* search555_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 115, 11 },
      -1,
   } },
   /* search555_1_0 -> 247 in the cache */
   /* search555_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 247, 115 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1687, 1688 },
      -1,
   } },

   /* replace555_0 -> 115 in the cache */
   /* replace555_1_0 -> 11 in the cache */
   /* replace555_1_1_0 -> 247 in the cache */
   /* replace555_1_1 -> 519 in the cache */
   /* replace555_1 -> 1685 in the cache */
   /* replace555 -> 1686 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search556_0_0 -> 115 in the cache */
   /* search556_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 115, 11 },
      0,
   } },
   /* search556_1_0 -> 247 in the cache */
   /* search556_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 247, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1690, 1691 },
      -1,
   } },

   /* replace556_0 -> 115 in the cache */
   /* replace556_1_0 -> 11 in the cache */
   /* replace556_1_1_0 -> 247 in the cache */
   /* replace556_1_1 -> 519 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 11, 519 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 115, 1693 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search557_0_0 -> 115 in the cache */
   /* search557_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 115, 11 },
      -1,
   } },
   /* search557_1_0 -> 247 in the cache */
   /* search557_1_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 247, 115 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1695, 1696 },
      -1,
   } },

   /* replace557_0 -> 115 in the cache */
   /* replace557_1_0 -> 11 in the cache */
   /* replace557_1_1_0 -> 247 in the cache */
   /* replace557_1_1 -> 519 in the cache */
   /* replace557_1 -> 1693 in the cache */
   /* replace557 -> 1694 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@16'), ('flt', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search558_0_0 -> 115 in the cache */
   /* search558_0_1 -> 11 in the cache */
   /* search558_0 -> 1682 in the cache */
   /* search558_1_0 -> 247 in the cache */
   /* search558_1_1 -> 115 in the cache */
   /* search558_1 -> 1683 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1682, 1683 },
      -1,
   } },

   /* replace558_0 -> 115 in the cache */
   /* replace558_1_0 -> 11 in the cache */
   /* replace558_1_1_0 -> 247 in the cache */
   /* replace558_1_1 -> 519 in the cache */
   /* replace558_1 -> 1693 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 115, 1693 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@16'), ('flt(is_used_once)', 'b@16', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search559_0_0 -> 115 in the cache */
   /* search559_0_1 -> 11 in the cache */
   /* search559_0 -> 1687 in the cache */
   /* search559_1_0 -> 247 in the cache */
   /* search559_1_1 -> 115 in the cache */
   /* search559_1 -> 1688 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1687, 1688 },
      -1,
   } },

   /* replace559_0 -> 115 in the cache */
   /* replace559_1_0 -> 11 in the cache */
   /* replace559_1_1_0 -> 247 in the cache */
   /* replace559_1_1 -> 519 in the cache */
   /* replace559_1 -> 1693 in the cache */
   /* replace559 -> 1699 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@16'), ('fge', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search560_0_0 -> 115 in the cache */
   /* search560_0_1 -> 11 in the cache */
   /* search560_0 -> 1690 in the cache */
   /* search560_1_0 -> 247 in the cache */
   /* search560_1_1 -> 115 in the cache */
   /* search560_1 -> 1691 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1690, 1691 },
      -1,
   } },

   /* replace560_0 -> 115 in the cache */
   /* replace560_1_0 -> 11 in the cache */
   /* replace560_1_1_0 -> 247 in the cache */
   /* replace560_1_1 -> 519 in the cache */
   /* replace560_1 -> 1685 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 115, 1685 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@16'), ('fge(is_used_once)', 'b@16', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search561_0_0 -> 115 in the cache */
   /* search561_0_1 -> 11 in the cache */
   /* search561_0 -> 1695 in the cache */
   /* search561_1_0 -> 247 in the cache */
   /* search561_1_1 -> 115 in the cache */
   /* search561_1 -> 1696 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1695, 1696 },
      -1,
   } },

   /* replace561_0 -> 115 in the cache */
   /* replace561_1_0 -> 11 in the cache */
   /* replace561_1_1_0 -> 247 in the cache */
   /* replace561_1_1 -> 519 in the cache */
   /* replace561_1 -> 1685 in the cache */
   /* replace561 -> 1702 in the cache */

   /* ('bcsel@16', ('feq', 'a', 0.0), 1.0, ('i2f16', ('iadd', ('b2i16', ('flt', 0.0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0.0)))))) => ('i2f16', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search562_0_0 -> 11 in the cache */
   /* search562_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 11, 115 },
      -1,
   } },
   /* search562_1 -> 466 in the cache */
   /* search562_2_0_0_0_0 -> 115 in the cache */
   /* search562_2_0_0_0_1 -> 11 in the cache */
   /* search562_2_0_0_0 -> 1687 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1687 },
      -1,
   } },
   /* search562_2_0_1_0_0_0 -> 11 in the cache */
   /* search562_2_0_1_0_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1706 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1707 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1705, 1708 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 1,
      { 1709 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1704, 466, 1710 },
      -1,
   } },

   /* replace562_0_0_0_0 -> 11 in the cache */
   /* replace562_0_0_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 11, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1712 },
      -1,
   } },
   /* replace562_0_1_0_0_0 -> 11 in the cache */
   /* replace562_0_1_0_0_1 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 115 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1714 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1715 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1713, 1716 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 1,
      { 1717 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@16'), ('b2f', 'c@16')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search563_0 -> 722 in the cache */
   /* search563_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 247 },
      0,
   } },
   /* search563_2_0 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 454 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1719, 1720 },
      -1,
   } },

   /* replace563_0_0 -> 722 in the cache */
   /* replace563_0_1 -> 247 in the cache */
   /* replace563_0_2 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 247, 454 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1722 },
      -1,
   } },

   /* ('~i2f16', ('f2i', 'a@16')) => ('ftrunc', 'a') */
   /* search564_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 1724 },
      -1,
   } },

   /* replace564_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('~f2i16', ('i2f', 'a@16')) => a */
   /* search565_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1727 },
      -1,
   } },

   /* replace565 -> 11 in the cache */

   /* ('~f2i16', ('u2f', 'a@16')) => a */
   /* search566_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1729 },
      -1,
   } },

   /* replace566 -> 11 in the cache */

   /* ('~f2u16', ('i2f', 'a@16')) => a */
   /* search567_0_0 -> 11 in the cache */
   /* search567_0 -> 1727 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1727 },
      -1,
   } },

   /* replace567 -> 11 in the cache */

   /* ('~f2u16', ('u2f', 'a@16')) => a */
   /* search568_0_0 -> 11 in the cache */
   /* search568_0 -> 1729 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1729 },
      -1,
   } },

   /* replace568 -> 11 in the cache */

   /* ('fadd', ('b2f16', ('flt', 0.0, 'a@16')), ('fneg', ('b2f16', ('flt', 'a@16', 0.0)))) => ('fsign', 'a') */
   /* search569_0_0_0 -> 115 in the cache */
   /* search569_0_0_1 -> 11 in the cache */
   /* search569_0_0 -> 1687 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1687 },
      -1,
   } },
   /* search569_1_0_0_0 -> 11 in the cache */
   /* search569_1_0_0_1 -> 115 in the cache */
   /* search569_1_0_0 -> 1706 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 1706 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1734 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1733, 1735 },
      -1,
   } },

   /* replace569_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('iadd', ('b2i16', ('flt', 0, 'a@16')), ('ineg', ('b2i16', ('flt', 'a@16', 0)))) => ('f2i16', ('fsign', 'a')) */
   /* search570_0_0_0 -> 818 in the cache */
   /* search570_0_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 818, 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1738 },
      -1,
   } },
   /* search570_1_0_0_0 -> 11 in the cache */
   /* search570_1_0_0_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 818 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 1740 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1739, 1742 },
      -1,
   } },

   /* replace570_0_0 -> 11 in the cache */
   /* replace570_0 -> 1737 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1737 },
      -1,
   } },

   /* ('f2f16', ('f2f32', 'a@16')) => a */
   /* search571_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1745 },
      -1,
   } },

   /* replace571 -> 11 in the cache */

   /* ('i2i16', ('i2i32', 'a@16')) => a */
   /* search572_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1747 },
      -1,
   } },

   /* replace572 -> 11 in the cache */

   /* ('u2u16', ('u2u32', 'a@16')) => a */
   /* search573_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1749 },
      -1,
   } },

   /* replace573 -> 11 in the cache */

   /* ('f2f16', ('b2f32', 'a@1')) => ('b2f16', 'a') */
   /* search574_0_0 -> 722 in the cache */
   /* search574_0 -> 1352 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1352 },
      -1,
   } },

   /* replace574_0 -> 722 in the cache */
   /* replace574 -> 1348 in the cache */

   /* ('i2i16', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search575_0_0 -> 722 in the cache */
   /* search575_0 -> 1340 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1340 },
      -1,
   } },

   /* replace575_0 -> 722 in the cache */
   /* replace575 -> 1336 in the cache */

   /* ('u2u16', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search576_0_0 -> 722 in the cache */
   /* search576_0 -> 1340 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1340 },
      -1,
   } },

   /* replace576_0 -> 722 in the cache */
   /* replace576 -> 1336 in the cache */

   /* ('f2u32', ('f2f32', 'a@16')) => ('f2u32', 'a') */
   /* search577_0_0 -> 11 in the cache */
   /* search577_0 -> 1745 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1745 },
      -1,
   } },

   /* replace577_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2i32', ('f2f32', 'a@16')) => ('f2i32', 'a') */
   /* search578_0_0 -> 11 in the cache */
   /* search578_0 -> 1745 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1745 },
      -1,
   } },

   /* replace578_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('u2f32', 'a')) => ('u2f16', 'a') */
   /* search579_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1758 },
      -1,
   } },

   /* replace579_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('f2f16', ('i2f32', 'a')) => ('i2f16', 'a') */
   /* search580_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1761 },
      -1,
   } },

   /* replace580_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('u2f32', ('u2u32', 'a@16')) => ('u2f32', 'a') */
   /* search581_0_0 -> 11 in the cache */
   /* search581_0 -> 1749 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 1749 },
      -1,
   } },

   /* replace581_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2f32', ('i2i32', 'a@16')) => ('i2f32', 'a') */
   /* search582_0_0 -> 11 in the cache */
   /* search582_0 -> 1747 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 1747 },
      -1,
   } },

   /* replace582_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('f2f64', 'a@16')) => a */
   /* search583_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1768 },
      -1,
   } },

   /* replace583 -> 11 in the cache */

   /* ('i2i16', ('i2i64', 'a@16')) => a */
   /* search584_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1770 },
      -1,
   } },

   /* replace584 -> 11 in the cache */

   /* ('u2u16', ('u2u64', 'a@16')) => a */
   /* search585_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1772 },
      -1,
   } },

   /* replace585 -> 11 in the cache */

   /* ('f2f16', ('b2f64', 'a@1')) => ('b2f16', 'a') */
   /* search586_0_0 -> 722 in the cache */
   /* search586_0 -> 1356 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1356 },
      -1,
   } },

   /* replace586_0 -> 722 in the cache */
   /* replace586 -> 1348 in the cache */

   /* ('i2i16', ('b2i64', 'a@1')) => ('b2i16', 'a') */
   /* search587_0_0 -> 722 in the cache */
   /* search587_0 -> 1344 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1344 },
      -1,
   } },

   /* replace587_0 -> 722 in the cache */
   /* replace587 -> 1336 in the cache */

   /* ('u2u16', ('b2i64', 'a@1')) => ('b2i16', 'a') */
   /* search588_0_0 -> 722 in the cache */
   /* search588_0 -> 1344 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1344 },
      -1,
   } },

   /* replace588_0 -> 722 in the cache */
   /* replace588 -> 1336 in the cache */

   /* ('f2u64', ('f2f64', 'a@16')) => ('f2u64', 'a') */
   /* search589_0_0 -> 11 in the cache */
   /* search589_0 -> 1768 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1768 },
      -1,
   } },

   /* replace589_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2i64', ('f2f64', 'a@16')) => ('f2i64', 'a') */
   /* search590_0_0 -> 11 in the cache */
   /* search590_0 -> 1768 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1768 },
      -1,
   } },

   /* replace590_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2f16', ('u2f64', 'a')) => ('u2f16', 'a') */
   /* search591_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1781 },
      -1,
   } },

   /* replace591_0 -> 32 in the cache */
   /* replace591 -> 1760 in the cache */

   /* ('f2f16', ('i2f64', 'a')) => ('i2f16', 'a') */
   /* search592_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 1783 },
      -1,
   } },

   /* replace592_0 -> 32 in the cache */
   /* replace592 -> 1763 in the cache */

   /* ('u2f64', ('u2u64', 'a@16')) => ('u2f64', 'a') */
   /* search593_0_0 -> 11 in the cache */
   /* search593_0 -> 1772 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 1772 },
      -1,
   } },

   /* replace593_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2f64', ('i2i64', 'a@16')) => ('i2f64', 'a') */
   /* search594_0_0 -> 11 in the cache */
   /* search594_0 -> 1770 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 1770 },
      -1,
   } },

   /* replace594_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search595_0_0 -> 117 in the cache */
   /* search595_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 117, 16 },
      0,
   } },
   /* search595_1_0 -> 36 in the cache */
   /* search595_1_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 36, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1789, 1790 },
      -1,
   } },

   /* replace595_0 -> 117 in the cache */
   /* replace595_1_0 -> 16 in the cache */
   /* replace595_1_1_0 -> 36 in the cache */
   /* replace595_1_1 -> 585 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 16, 585 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 117, 1792 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search596_0_0 -> 117 in the cache */
   /* search596_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 117, 16 },
      -1,
   } },
   /* search596_1_0 -> 36 in the cache */
   /* search596_1_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 36, 117 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1794, 1795 },
      -1,
   } },

   /* replace596_0 -> 117 in the cache */
   /* replace596_1_0 -> 16 in the cache */
   /* replace596_1_1_0 -> 36 in the cache */
   /* replace596_1_1 -> 585 in the cache */
   /* replace596_1 -> 1792 in the cache */
   /* replace596 -> 1793 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search597_0_0 -> 117 in the cache */
   /* search597_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 117, 16 },
      0,
   } },
   /* search597_1_0 -> 36 in the cache */
   /* search597_1_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 36, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1797, 1798 },
      -1,
   } },

   /* replace597_0 -> 117 in the cache */
   /* replace597_1_0 -> 16 in the cache */
   /* replace597_1_1_0 -> 36 in the cache */
   /* replace597_1_1 -> 585 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 16, 585 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 117, 1800 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search598_0_0 -> 117 in the cache */
   /* search598_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 117, 16 },
      -1,
   } },
   /* search598_1_0 -> 36 in the cache */
   /* search598_1_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 36, 117 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1802, 1803 },
      -1,
   } },

   /* replace598_0 -> 117 in the cache */
   /* replace598_1_0 -> 16 in the cache */
   /* replace598_1_1_0 -> 36 in the cache */
   /* replace598_1_1 -> 585 in the cache */
   /* replace598_1 -> 1800 in the cache */
   /* replace598 -> 1801 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@32'), ('flt', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search599_0_0 -> 117 in the cache */
   /* search599_0_1 -> 16 in the cache */
   /* search599_0 -> 1789 in the cache */
   /* search599_1_0 -> 36 in the cache */
   /* search599_1_1 -> 117 in the cache */
   /* search599_1 -> 1790 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1789, 1790 },
      -1,
   } },

   /* replace599_0 -> 117 in the cache */
   /* replace599_1_0 -> 16 in the cache */
   /* replace599_1_1_0 -> 36 in the cache */
   /* replace599_1_1 -> 585 in the cache */
   /* replace599_1 -> 1800 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 117, 1800 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@32'), ('flt(is_used_once)', 'b@32', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search600_0_0 -> 117 in the cache */
   /* search600_0_1 -> 16 in the cache */
   /* search600_0 -> 1794 in the cache */
   /* search600_1_0 -> 36 in the cache */
   /* search600_1_1 -> 117 in the cache */
   /* search600_1 -> 1795 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1794, 1795 },
      -1,
   } },

   /* replace600_0 -> 117 in the cache */
   /* replace600_1_0 -> 16 in the cache */
   /* replace600_1_1_0 -> 36 in the cache */
   /* replace600_1_1 -> 585 in the cache */
   /* replace600_1 -> 1800 in the cache */
   /* replace600 -> 1806 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@32'), ('fge', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search601_0_0 -> 117 in the cache */
   /* search601_0_1 -> 16 in the cache */
   /* search601_0 -> 1797 in the cache */
   /* search601_1_0 -> 36 in the cache */
   /* search601_1_1 -> 117 in the cache */
   /* search601_1 -> 1798 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1797, 1798 },
      -1,
   } },

   /* replace601_0 -> 117 in the cache */
   /* replace601_1_0 -> 16 in the cache */
   /* replace601_1_1_0 -> 36 in the cache */
   /* replace601_1_1 -> 585 in the cache */
   /* replace601_1 -> 1792 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 117, 1792 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@32'), ('fge(is_used_once)', 'b@32', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search602_0_0 -> 117 in the cache */
   /* search602_0_1 -> 16 in the cache */
   /* search602_0 -> 1802 in the cache */
   /* search602_1_0 -> 36 in the cache */
   /* search602_1_1 -> 117 in the cache */
   /* search602_1 -> 1803 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1802, 1803 },
      -1,
   } },

   /* replace602_0 -> 117 in the cache */
   /* replace602_1_0 -> 16 in the cache */
   /* replace602_1_1_0 -> 36 in the cache */
   /* replace602_1_1 -> 585 in the cache */
   /* replace602_1 -> 1792 in the cache */
   /* replace602 -> 1809 in the cache */

   /* ('bcsel@32', ('feq', 'a', 0.0), 1.0, ('i2f32', ('iadd', ('b2i32', ('flt', 0.0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0.0)))))) => ('i2f32', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search603_0_0 -> 16 in the cache */
   /* search603_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 117 },
      -1,
   } },
   /* search603_1 -> 232 in the cache */
   /* search603_2_0_0_0_0 -> 117 in the cache */
   /* search603_2_0_0_0_1 -> 16 in the cache */
   /* search603_2_0_0_0 -> 1794 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1794 },
      -1,
   } },
   /* search603_2_0_1_0_0_0 -> 16 in the cache */
   /* search603_2_0_1_0_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1813 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1814 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1812, 1815 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 1,
      { 1816 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1811, 232, 1817 },
      -1,
   } },

   /* replace603_0_0_0_0 -> 16 in the cache */
   /* replace603_0_0_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1819 },
      -1,
   } },
   /* replace603_0_1_0_0_0 -> 16 in the cache */
   /* replace603_0_1_0_0_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 117 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1821 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1822 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1820, 1823 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 1,
      { 1824 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@32'), ('b2f', 'c@32')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search604_0 -> 722 in the cache */
   /* search604_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 36 },
      0,
   } },
   /* search604_2_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1826, 1827 },
      -1,
   } },

   /* replace604_0_0 -> 722 in the cache */
   /* replace604_0_1 -> 36 in the cache */
   /* replace604_0_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 36, 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1829 },
      -1,
   } },

   /* ('~i2f32', ('f2i', 'a@32')) => ('ftrunc', 'a') */
   /* search605_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 1831 },
      -1,
   } },

   /* replace605_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('~f2i32', ('i2f', 'a@32')) => a */
   /* search606_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1834 },
      -1,
   } },

   /* replace606 -> 16 in the cache */

   /* ('~f2i32', ('u2f', 'a@32')) => a */
   /* search607_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1836 },
      -1,
   } },

   /* replace607 -> 16 in the cache */

   /* ('~f2u32', ('i2f', 'a@32')) => a */
   /* search608_0_0 -> 16 in the cache */
   /* search608_0 -> 1834 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1834 },
      -1,
   } },

   /* replace608 -> 16 in the cache */

   /* ('~f2u32', ('u2f', 'a@32')) => a */
   /* search609_0_0 -> 16 in the cache */
   /* search609_0 -> 1836 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 1836 },
      -1,
   } },

   /* replace609 -> 16 in the cache */

   /* ('fadd', ('b2f32', ('flt', 0.0, 'a@32')), ('fneg', ('b2f32', ('flt', 'a@32', 0.0)))) => ('fsign', 'a') */
   /* search610_0_0_0 -> 117 in the cache */
   /* search610_0_0_1 -> 16 in the cache */
   /* search610_0_0 -> 1794 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1794 },
      -1,
   } },
   /* search610_1_0_0_0 -> 16 in the cache */
   /* search610_1_0_0_1 -> 117 in the cache */
   /* search610_1_0_0 -> 1813 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 1813 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1841 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1840, 1842 },
      -1,
   } },

   /* replace610_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('iadd', ('b2i32', ('flt', 0, 'a@32')), ('ineg', ('b2i32', ('flt', 'a@32', 0)))) => ('f2i32', ('fsign', 'a')) */
   /* search611_0_0_0 -> 124 in the cache */
   /* search611_0_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1845 },
      -1,
   } },
   /* search611_1_0_0_0 -> 16 in the cache */
   /* search611_1_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1847 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1848 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1846, 1849 },
      -1,
   } },

   /* replace611_0_0 -> 16 in the cache */
   /* replace611_0 -> 1844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 1844 },
      -1,
   } },

   /* ('f2f32', ('f2f64', 'a@32')) => a */
   /* search612_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1852 },
      -1,
   } },

   /* replace612 -> 16 in the cache */

   /* ('i2i32', ('i2i64', 'a@32')) => a */
   /* search613_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1854 },
      -1,
   } },

   /* replace613 -> 16 in the cache */

   /* ('u2u32', ('u2u64', 'a@32')) => a */
   /* search614_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1856 },
      -1,
   } },

   /* replace614 -> 16 in the cache */

   /* ('f2f32', ('b2f64', 'a@1')) => ('b2f32', 'a') */
   /* search615_0_0 -> 722 in the cache */
   /* search615_0 -> 1356 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1356 },
      -1,
   } },

   /* replace615_0 -> 722 in the cache */
   /* replace615 -> 1352 in the cache */

   /* ('i2i32', ('b2i64', 'a@1')) => ('b2i32', 'a') */
   /* search616_0_0 -> 722 in the cache */
   /* search616_0 -> 1344 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1344 },
      -1,
   } },

   /* replace616_0 -> 722 in the cache */
   /* replace616 -> 1340 in the cache */

   /* ('u2u32', ('b2i64', 'a@1')) => ('b2i32', 'a') */
   /* search617_0_0 -> 722 in the cache */
   /* search617_0 -> 1344 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1344 },
      -1,
   } },

   /* replace617_0 -> 722 in the cache */
   /* replace617 -> 1340 in the cache */

   /* ('f2u64', ('f2f64', 'a@32')) => ('f2u64', 'a') */
   /* search618_0_0 -> 16 in the cache */
   /* search618_0 -> 1852 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1852 },
      -1,
   } },

   /* replace618_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2i64', ('f2f64', 'a@32')) => ('f2i64', 'a') */
   /* search619_0_0 -> 16 in the cache */
   /* search619_0 -> 1852 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1852 },
      -1,
   } },

   /* replace619_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2f32', ('u2f64', 'a')) => ('u2f32', 'a') */
   /* search620_0_0 -> 32 in the cache */
   /* search620_0 -> 1781 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1781 },
      -1,
   } },

   /* replace620_0 -> 32 in the cache */
   /* replace620 -> 1758 in the cache */

   /* ('f2f32', ('i2f64', 'a')) => ('i2f32', 'a') */
   /* search621_0_0 -> 32 in the cache */
   /* search621_0 -> 1783 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1783 },
      -1,
   } },

   /* replace621_0 -> 32 in the cache */
   /* replace621 -> 1761 in the cache */

   /* ('u2f64', ('u2u64', 'a@32')) => ('u2f64', 'a') */
   /* search622_0_0 -> 16 in the cache */
   /* search622_0 -> 1856 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 1856 },
      -1,
   } },

   /* replace622_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2f64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2f64', ('i2i64', 'a@32')) => ('i2f64', 'a') */
   /* search623_0_0 -> 16 in the cache */
   /* search623_0 -> 1854 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 1854 },
      -1,
   } },

   /* replace623_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('ior', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   /* search624_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1871, 21 },
      0,
   } },
   /* search624_1_0 -> 41 in the cache */
   /* search624_1_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 41, 1871 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1872, 1873 },
      -1,
   } },

   /* replace624_0 -> 1871 in the cache */
   /* replace624_1_0 -> 21 in the cache */
   /* replace624_1_1_0 -> 41 in the cache */
   /* replace624_1_1 -> 658 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 21, 658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1871, 1875 },
      -1,
   } },

   /* ('ior', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search625_0_0 -> 1871 in the cache */
   /* search625_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1871, 21 },
      -1,
   } },
   /* search625_1_0 -> 41 in the cache */
   /* search625_1_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 41, 1871 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1877, 1878 },
      -1,
   } },

   /* replace625_0 -> 1871 in the cache */
   /* replace625_1_0 -> 21 in the cache */
   /* replace625_1_1_0 -> 41 in the cache */
   /* replace625_1_1 -> 658 in the cache */
   /* replace625_1 -> 1875 in the cache */
   /* replace625 -> 1876 in the cache */

   /* ('ior', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search626_0_0 -> 1871 in the cache */
   /* search626_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1871, 21 },
      0,
   } },
   /* search626_1_0 -> 41 in the cache */
   /* search626_1_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 41, 1871 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1880, 1881 },
      -1,
   } },

   /* replace626_0 -> 1871 in the cache */
   /* replace626_1_0 -> 21 in the cache */
   /* replace626_1_1_0 -> 41 in the cache */
   /* replace626_1_1 -> 658 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 21, 658 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1871, 1883 },
      -1,
   } },

   /* ('ior', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search627_0_0 -> 1871 in the cache */
   /* search627_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1871, 21 },
      -1,
   } },
   /* search627_1_0 -> 41 in the cache */
   /* search627_1_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 41, 1871 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1885, 1886 },
      -1,
   } },

   /* replace627_0 -> 1871 in the cache */
   /* replace627_1_0 -> 21 in the cache */
   /* replace627_1_1_0 -> 41 in the cache */
   /* replace627_1_1 -> 658 in the cache */
   /* replace627_1 -> 1883 in the cache */
   /* replace627 -> 1884 in the cache */

   /* ('~iand', ('flt(is_used_once)', 0.0, 'a@64'), ('flt', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search628_0_0 -> 1871 in the cache */
   /* search628_0_1 -> 21 in the cache */
   /* search628_0 -> 1872 in the cache */
   /* search628_1_0 -> 41 in the cache */
   /* search628_1_1 -> 1871 in the cache */
   /* search628_1 -> 1873 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1872, 1873 },
      -1,
   } },

   /* replace628_0 -> 1871 in the cache */
   /* replace628_1_0 -> 21 in the cache */
   /* replace628_1_1_0 -> 41 in the cache */
   /* replace628_1_1 -> 658 in the cache */
   /* replace628_1 -> 1883 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1871, 1883 },
      -1,
   } },

   /* ('~iand', ('flt', 0.0, 'a@64'), ('flt(is_used_once)', 'b@64', 0.0)) => ('flt', 0.0, ('fmin', 'a', ('fneg', 'b'))) */
   /* search629_0_0 -> 1871 in the cache */
   /* search629_0_1 -> 21 in the cache */
   /* search629_0 -> 1877 in the cache */
   /* search629_1_0 -> 41 in the cache */
   /* search629_1_1 -> 1871 in the cache */
   /* search629_1 -> 1878 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1877, 1878 },
      -1,
   } },

   /* replace629_0 -> 1871 in the cache */
   /* replace629_1_0 -> 21 in the cache */
   /* replace629_1_1_0 -> 41 in the cache */
   /* replace629_1_1 -> 658 in the cache */
   /* replace629_1 -> 1883 in the cache */
   /* replace629 -> 1889 in the cache */

   /* ('~iand', ('fge(is_used_once)', 0.0, 'a@64'), ('fge', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search630_0_0 -> 1871 in the cache */
   /* search630_0_1 -> 21 in the cache */
   /* search630_0 -> 1880 in the cache */
   /* search630_1_0 -> 41 in the cache */
   /* search630_1_1 -> 1871 in the cache */
   /* search630_1 -> 1881 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1880, 1881 },
      -1,
   } },

   /* replace630_0 -> 1871 in the cache */
   /* replace630_1_0 -> 21 in the cache */
   /* replace630_1_1_0 -> 41 in the cache */
   /* replace630_1_1 -> 658 in the cache */
   /* replace630_1 -> 1875 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1871, 1875 },
      -1,
   } },

   /* ('~iand', ('fge', 0.0, 'a@64'), ('fge(is_used_once)', 'b@64', 0.0)) => ('fge', 0.0, ('fmax', 'a', ('fneg', 'b'))) */
   /* search631_0_0 -> 1871 in the cache */
   /* search631_0_1 -> 21 in the cache */
   /* search631_0 -> 1885 in the cache */
   /* search631_1_0 -> 41 in the cache */
   /* search631_1_1 -> 1871 in the cache */
   /* search631_1 -> 1886 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1885, 1886 },
      -1,
   } },

   /* replace631_0 -> 1871 in the cache */
   /* replace631_1_0 -> 21 in the cache */
   /* replace631_1_1_0 -> 41 in the cache */
   /* replace631_1_1 -> 658 in the cache */
   /* replace631_1 -> 1875 in the cache */
   /* replace631 -> 1892 in the cache */

   /* ('bcsel@64', ('feq', 'a', 0.0), 1.0, ('i2f64', ('iadd', ('b2i64', ('flt', 0.0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0.0)))))) => ('i2f64', ('iadd', ('b2i32', ('!fge', 'a', 0.0)), ('ineg', ('b2i32', ('!flt', 'a', 0.0))))) */
   /* search632_0_0 -> 21 in the cache */
   /* search632_0_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 21, 1871 },
      -1,
   } },
   /* search632_1 -> 607 in the cache */
   /* search632_2_0_0_0_0 -> 1871 in the cache */
   /* search632_2_0_0_0_1 -> 21 in the cache */
   /* search632_2_0_0_0 -> 1877 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1877 },
      -1,
   } },
   /* search632_2_0_1_0_0_0 -> 21 in the cache */
   /* search632_2_0_1_0_0_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 1871 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1896 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1897 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 1895, 1898 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 1,
      { 1899 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 1894, 607, 1900 },
      -1,
   } },

   /* replace632_0_0_0_0 -> 21 in the cache */
   /* replace632_0_0_0_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 21, 1871 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1902 },
      -1,
   } },
   /* replace632_0_1_0_0_0 -> 21 in the cache */
   /* replace632_0_1_0_0_1 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 1871 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 1904 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1905 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1903, 1906 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2f64,
      -1, 1,
      { 1907 },
      -1,
   } },

   /* ('bcsel', 'a', ('b2f(is_used_once)', 'b@64'), ('b2f', 'c@64')) => ('b2f', ('bcsel', 'a', 'b', 'c')) */
   /* search633_0 -> 722 in the cache */
   /* search633_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 41 },
      0,
   } },
   /* search633_2_0 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1909, 1910 },
      -1,
   } },

   /* replace633_0_0 -> 722 in the cache */
   /* replace633_0_1 -> 41 in the cache */
   /* replace633_0_2 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 41, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1912 },
      -1,
   } },

   /* ('~i2f64', ('f2i', 'a@64')) => ('ftrunc', 'a') */
   /* search634_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_i2f64,
      -1, 0,
      { 1914 },
      -1,
   } },

   /* replace634_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('~f2i64', ('i2f', 'a@64')) => a */
   /* search635_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1917 },
      -1,
   } },

   /* replace635 -> 21 in the cache */

   /* ('~f2i64', ('u2f', 'a@64')) => a */
   /* search636_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1919 },
      -1,
   } },

   /* replace636 -> 21 in the cache */

   /* ('~f2u64', ('i2f', 'a@64')) => a */
   /* search637_0_0 -> 21 in the cache */
   /* search637_0 -> 1917 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1917 },
      -1,
   } },

   /* replace637 -> 21 in the cache */

   /* ('~f2u64', ('u2f', 'a@64')) => a */
   /* search638_0_0 -> 21 in the cache */
   /* search638_0 -> 1919 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_f2u64,
      -1, 0,
      { 1919 },
      -1,
   } },

   /* replace638 -> 21 in the cache */

   /* ('fadd', ('b2f64', ('flt', 0.0, 'a@64')), ('fneg', ('b2f64', ('flt', 'a@64', 0.0)))) => ('fsign', 'a') */
   /* search639_0_0_0 -> 1871 in the cache */
   /* search639_0_0_1 -> 21 in the cache */
   /* search639_0_0 -> 1877 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1877 },
      -1,
   } },
   /* search639_1_0_0_0 -> 21 in the cache */
   /* search639_1_0_0_1 -> 1871 in the cache */
   /* search639_1_0_0 -> 1896 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 1896 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1924 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1923, 1925 },
      -1,
   } },

   /* replace639_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('iadd', ('b2i64', ('flt', 0, 'a@64')), ('ineg', ('b2i64', ('flt', 'a@64', 0)))) => ('f2i64', ('fsign', 'a')) */
   /* search640_0_0_0 -> 859 in the cache */
   /* search640_0_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 859, 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1928 },
      -1,
   } },
   /* search640_1_0_0_0 -> 21 in the cache */
   /* search640_1_0_0_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 21, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 1930 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1931 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1929, 1932 },
      -1,
   } },

   /* replace640_0_0 -> 21 in the cache */
   /* replace640_0 -> 1927 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2i64,
      -1, 0,
      { 1927 },
      -1,
   } },

   /* ('f2fmp', ('u2f32', 'a@32')) => ('u2fmp', 'a') */
   /* search641_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1935 },
      -1,
   } },

   /* replace641_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2fmp', ('i2f32', 'a@32')) => ('i2fmp', 'a') */
   /* search642_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2f32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1938 },
      -1,
   } },

   /* replace642_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2u16', ('f2fmp', 'a@32')) => ('f2u16', 'a') */
   /* search643_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 1941 },
      -1,
   } },

   /* replace643_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('f2i16', ('f2fmp', 'a@32')) => ('f2i16', 'a') */
   /* search644_0_0 -> 16 in the cache */
   /* search644_0 -> 1941 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 1941 },
      -1,
   } },

   /* replace644_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2imp', ('f2u32', 'a@32')) => ('f2ump', 'a') */
   /* search645_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1946 },
      -1,
   } },

   /* replace645_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2imp', ('f2i32', 'a@32')) => ('f2imp', 'a') */
   /* search646_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1949 },
      -1,
   } },

   /* replace646_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('u2f16', ('i2imp', 'a@32')) => ('u2f16', 'a') */
   /* search647_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 1952 },
      -1,
   } },

   /* replace647_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2f16', ('i2imp', 'a@32')) => ('i2f16', 'a') */
   /* search648_0_0 -> 16 in the cache */
   /* search648_0 -> 1952 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 1952 },
      -1,
   } },

   /* replace648_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('iand', ('u2u16', ('u2u8', 'a@16')), '#b') => ('iand', 'a', ('iand', 'b', 255)) */
   /* search649_0_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1957 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1958, 1959 },
      -1,
   } },

   /* replace649_0 -> 11 in the cache */
   /* replace649_1_0 -> 247 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 247, 1961 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 11, 1962 },
      -1,
   } },

   /* ('u2u16', ('u2u8(is_used_once)', ('iand', 'a@16', '#b'))) => ('iand', 'a', ('iand', 'b', 255)) */
   /* search650_0_0_0 -> 11 in the cache */
   /* search650_0_0_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 1959 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 1964 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 1,
      { 1965 },
      -1,
   } },

   /* replace650_0 -> 11 in the cache */
   /* replace650_1_0 -> 247 in the cache */
   /* replace650_1_1 -> 1961 in the cache */
   /* replace650_1 -> 1962 in the cache */
   /* replace650 -> 1963 in the cache */

   /* ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('iand', 'a', 'b')) */
   /* search651_0_0_0_0 -> 11 in the cache */
   /* search651_0_0_0 -> 1957 in the cache */
   /* search651_0_0 -> 1958 in the cache */
   /* search651_0_1_0_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1967 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1958, 1968 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 1969 },
      -1,
   } },

   /* replace651_0_0 -> 11 in the cache */
   /* replace651_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 1971 },
      -1,
   } },

   /* ('u2u8', ('iand', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('iand', 'a', 'b')) */
   /* search652_0_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1973 },
      -1,
   } },
   /* search652_0_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1975 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1974, 1976 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 1977 },
      -1,
   } },

   /* replace652_0_0 -> 16 in the cache */
   /* replace652_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 1979 },
      -1,
   } },

   /* ('iand', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('iand', 'a', 'c'), 'b') */
   /* search653_0_0 -> 0 in the cache */
   /* search653_0_1 -> 783 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 783 },
      -1,
   } },
   /* search653_1_0 -> 142 in the cache */
   /* search653_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 142, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1981, 1982 },
      -1,
   } },

   /* replace653_0_0 -> 0 in the cache */
   /* replace653_0_1 -> 142 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 142 },
      -1,
   } },
   /* replace653_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 1984, 3 },
      -1,
   } },

   /* ('iand', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('iand', 'a', 'c'), 'b') */
   /* search654_0_0 -> 0 in the cache */
   /* search654_0_1 -> 783 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 783 },
      -1,
   } },
   /* search654_1_0 -> 142 in the cache */
   /* search654_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 142, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1986, 1987 },
      -1,
   } },

   /* replace654_0_0 -> 0 in the cache */
   /* replace654_0_1 -> 142 in the cache */
   /* replace654_0 -> 1984 in the cache */
   /* replace654_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 1984, 3 },
      -1,
   } },

   /* ('iand', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('iand', 'a', 'c'), 'b') */
   /* search655_0_0 -> 0 in the cache */
   /* search655_0_1 -> 783 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 783 },
      -1,
   } },
   /* search655_1_0 -> 142 in the cache */
   /* search655_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 142, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1990, 1991 },
      -1,
   } },

   /* replace655_0_0 -> 0 in the cache */
   /* replace655_0_1 -> 142 in the cache */
   /* replace655_0 -> 1984 in the cache */
   /* replace655_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 1984, 3 },
      -1,
   } },

   /* ('iand', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('iand', 'a', 'c'), 'b') */
   /* search656_0_0 -> 0 in the cache */
   /* search656_0_1 -> 783 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 783 },
      -1,
   } },
   /* search656_1_0 -> 142 in the cache */
   /* search656_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 142, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1994, 1995 },
      -1,
   } },

   /* replace656_0_0 -> 0 in the cache */
   /* replace656_0_1 -> 142 in the cache */
   /* replace656_0 -> 1984 in the cache */
   /* replace656_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 1984, 3 },
      -1,
   } },

   /* ('iand', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('iand', 'a', 'c'), 'b') */
   /* search657_0_0 -> 128 in the cache */
   /* search657_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 128, 33 },
      0,
   } },
   /* search657_1_0 -> 131 in the cache */
   /* search657_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 131, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1998, 1999 },
      -1,
   } },

   /* replace657_0_0 -> 128 in the cache */
   /* replace657_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 128, 131 },
      -1,
   } },
   /* replace657_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2001, 36 },
      -1,
   } },

   /* ('iand', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('iand', 'a', 'c'), 'b') */
   /* search658_0_0 -> 128 in the cache */
   /* search658_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 128, 33 },
      0,
   } },
   /* search658_1_0 -> 131 in the cache */
   /* search658_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 131, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2003, 2004 },
      -1,
   } },

   /* replace658_0_0 -> 128 in the cache */
   /* replace658_0_1 -> 131 in the cache */
   /* replace658_0 -> 2001 in the cache */
   /* replace658_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2001, 36 },
      -1,
   } },

   /* ('iand', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('iand', 'a', 'c'), 'b') */
   /* search659_0_0 -> 128 in the cache */
   /* search659_0_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 128, 33 },
      0,
   } },
   /* search659_1_0 -> 131 in the cache */
   /* search659_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 131, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2007, 2008 },
      -1,
   } },

   /* replace659_0_0 -> 128 in the cache */
   /* replace659_0_1 -> 131 in the cache */
   /* replace659_0 -> 2001 in the cache */
   /* replace659_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2001, 36 },
      -1,
   } },

   /* ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ior', 'a', 'b')) */
   /* search660_0_0_0_0 -> 11 in the cache */
   /* search660_0_0_0 -> 1957 in the cache */
   /* search660_0_0 -> 1958 in the cache */
   /* search660_0_1_0_0 -> 247 in the cache */
   /* search660_0_1_0 -> 1967 in the cache */
   /* search660_0_1 -> 1968 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1958, 1968 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2011 },
      -1,
   } },

   /* replace660_0_0 -> 11 in the cache */
   /* replace660_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 11, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2013 },
      -1,
   } },

   /* ('u2u8', ('ior', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ior', 'a', 'b')) */
   /* search661_0_0_0_0 -> 16 in the cache */
   /* search661_0_0_0 -> 1973 in the cache */
   /* search661_0_0 -> 1974 in the cache */
   /* search661_0_1_0_0 -> 36 in the cache */
   /* search661_0_1_0 -> 1975 in the cache */
   /* search661_0_1 -> 1976 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1974, 1976 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2015 },
      -1,
   } },

   /* replace661_0_0 -> 16 in the cache */
   /* replace661_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2017 },
      -1,
   } },

   /* ('ior', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ior', 'a', 'c'), 'b') */
   /* search662_0_0 -> 0 in the cache */
   /* search662_0_1 -> 783 in the cache */
   /* search662_0 -> 1981 in the cache */
   /* search662_1_0 -> 142 in the cache */
   /* search662_1_1 -> 3 in the cache */
   /* search662_1 -> 1982 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1981, 1982 },
      -1,
   } },

   /* replace662_0_0 -> 0 in the cache */
   /* replace662_0_1 -> 142 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 0, 142 },
      -1,
   } },
   /* replace662_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2020, 3 },
      -1,
   } },

   /* ('ior', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ior', 'a', 'c'), 'b') */
   /* search663_0_0 -> 0 in the cache */
   /* search663_0_1 -> 783 in the cache */
   /* search663_0 -> 1986 in the cache */
   /* search663_1_0 -> 142 in the cache */
   /* search663_1_1 -> 3 in the cache */
   /* search663_1 -> 1987 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1986, 1987 },
      -1,
   } },

   /* replace663_0_0 -> 0 in the cache */
   /* replace663_0_1 -> 142 in the cache */
   /* replace663_0 -> 2020 in the cache */
   /* replace663_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2020, 3 },
      -1,
   } },

   /* ('ior', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ior', 'a', 'c'), 'b') */
   /* search664_0_0 -> 0 in the cache */
   /* search664_0_1 -> 783 in the cache */
   /* search664_0 -> 1990 in the cache */
   /* search664_1_0 -> 142 in the cache */
   /* search664_1_1 -> 3 in the cache */
   /* search664_1 -> 1991 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1990, 1991 },
      -1,
   } },

   /* replace664_0_0 -> 0 in the cache */
   /* replace664_0_1 -> 142 in the cache */
   /* replace664_0 -> 2020 in the cache */
   /* replace664_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2020, 3 },
      -1,
   } },

   /* ('ior', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ior', 'a', 'c'), 'b') */
   /* search665_0_0 -> 0 in the cache */
   /* search665_0_1 -> 783 in the cache */
   /* search665_0 -> 1994 in the cache */
   /* search665_1_0 -> 142 in the cache */
   /* search665_1_1 -> 3 in the cache */
   /* search665_1 -> 1995 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1994, 1995 },
      -1,
   } },

   /* replace665_0_0 -> 0 in the cache */
   /* replace665_0_1 -> 142 in the cache */
   /* replace665_0 -> 2020 in the cache */
   /* replace665_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2020, 3 },
      -1,
   } },

   /* ('ior', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ior', 'a', 'c'), 'b') */
   /* search666_0_0 -> 128 in the cache */
   /* search666_0_1 -> 33 in the cache */
   /* search666_0 -> 1998 in the cache */
   /* search666_1_0 -> 131 in the cache */
   /* search666_1_1 -> 36 in the cache */
   /* search666_1 -> 1999 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1998, 1999 },
      -1,
   } },

   /* replace666_0_0 -> 128 in the cache */
   /* replace666_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 128, 131 },
      -1,
   } },
   /* replace666_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2029, 36 },
      -1,
   } },

   /* ('ior', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ior', 'a', 'c'), 'b') */
   /* search667_0_0 -> 128 in the cache */
   /* search667_0_1 -> 33 in the cache */
   /* search667_0 -> 2003 in the cache */
   /* search667_1_0 -> 131 in the cache */
   /* search667_1_1 -> 36 in the cache */
   /* search667_1 -> 2004 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2003, 2004 },
      -1,
   } },

   /* replace667_0_0 -> 128 in the cache */
   /* replace667_0_1 -> 131 in the cache */
   /* replace667_0 -> 2029 in the cache */
   /* replace667_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2029, 36 },
      -1,
   } },

   /* ('ior', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ior', 'a', 'c'), 'b') */
   /* search668_0_0 -> 128 in the cache */
   /* search668_0_1 -> 33 in the cache */
   /* search668_0 -> 2007 in the cache */
   /* search668_1_0 -> 131 in the cache */
   /* search668_1_1 -> 36 in the cache */
   /* search668_1 -> 2008 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2007, 2008 },
      -1,
   } },

   /* replace668_0_0 -> 128 in the cache */
   /* replace668_0_1 -> 131 in the cache */
   /* replace668_0 -> 2029 in the cache */
   /* replace668_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2029, 36 },
      -1,
   } },

   /* ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@16')), ('u2u16', ('u2u8', 'b@16')))) => ('u2u8', ('ixor', 'a', 'b')) */
   /* search669_0_0_0_0 -> 11 in the cache */
   /* search669_0_0_0 -> 1957 in the cache */
   /* search669_0_0 -> 1958 in the cache */
   /* search669_0_1_0_0 -> 247 in the cache */
   /* search669_0_1_0 -> 1967 in the cache */
   /* search669_0_1 -> 1968 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1958, 1968 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2035 },
      -1,
   } },

   /* replace669_0_0 -> 11 in the cache */
   /* replace669_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 11, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2037 },
      -1,
   } },

   /* ('u2u8', ('ixor', ('u2u16', ('u2u8', 'a@32')), ('u2u16', ('u2u8', 'b@32')))) => ('u2u8', ('ixor', 'a', 'b')) */
   /* search670_0_0_0_0 -> 16 in the cache */
   /* search670_0_0_0 -> 1973 in the cache */
   /* search670_0_0 -> 1974 in the cache */
   /* search670_0_1_0_0 -> 36 in the cache */
   /* search670_0_1_0 -> 1975 in the cache */
   /* search670_0_1 -> 1976 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1974, 1976 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2039 },
      -1,
   } },

   /* replace670_0_0 -> 16 in the cache */
   /* replace670_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 1,
      { 2041 },
      -1,
   } },

   /* ('ixor', ('extract_i8', 'a', '#b'), ('extract_i8', 'c', 'b')) => ('extract_i8', ('ixor', 'a', 'c'), 'b') */
   /* search671_0_0 -> 0 in the cache */
   /* search671_0_1 -> 783 in the cache */
   /* search671_0 -> 1981 in the cache */
   /* search671_1_0 -> 142 in the cache */
   /* search671_1_1 -> 3 in the cache */
   /* search671_1 -> 1982 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1981, 1982 },
      -1,
   } },

   /* replace671_0_0 -> 0 in the cache */
   /* replace671_0_1 -> 142 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 0, 142 },
      -1,
   } },
   /* replace671_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 1,
      { 2044, 3 },
      -1,
   } },

   /* ('ixor', ('extract_u8', 'a', '#b'), ('extract_u8', 'c', 'b')) => ('extract_u8', ('ixor', 'a', 'c'), 'b') */
   /* search672_0_0 -> 0 in the cache */
   /* search672_0_1 -> 783 in the cache */
   /* search672_0 -> 1986 in the cache */
   /* search672_1_0 -> 142 in the cache */
   /* search672_1_1 -> 3 in the cache */
   /* search672_1 -> 1987 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1986, 1987 },
      -1,
   } },

   /* replace672_0_0 -> 0 in the cache */
   /* replace672_0_1 -> 142 in the cache */
   /* replace672_0 -> 2044 in the cache */
   /* replace672_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 1,
      { 2044, 3 },
      -1,
   } },

   /* ('ixor', ('extract_i16', 'a', '#b'), ('extract_i16', 'c', 'b')) => ('extract_i16', ('ixor', 'a', 'c'), 'b') */
   /* search673_0_0 -> 0 in the cache */
   /* search673_0_1 -> 783 in the cache */
   /* search673_0 -> 1990 in the cache */
   /* search673_1_0 -> 142 in the cache */
   /* search673_1_1 -> 3 in the cache */
   /* search673_1 -> 1991 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1990, 1991 },
      -1,
   } },

   /* replace673_0_0 -> 0 in the cache */
   /* replace673_0_1 -> 142 in the cache */
   /* replace673_0 -> 2044 in the cache */
   /* replace673_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 1,
      { 2044, 3 },
      -1,
   } },

   /* ('ixor', ('extract_u16', 'a', '#b'), ('extract_u16', 'c', 'b')) => ('extract_u16', ('ixor', 'a', 'c'), 'b') */
   /* search674_0_0 -> 0 in the cache */
   /* search674_0_1 -> 783 in the cache */
   /* search674_0 -> 1994 in the cache */
   /* search674_1_0 -> 142 in the cache */
   /* search674_1_1 -> 3 in the cache */
   /* search674_1 -> 1995 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1994, 1995 },
      -1,
   } },

   /* replace674_0_0 -> 0 in the cache */
   /* replace674_0_1 -> 142 in the cache */
   /* replace674_0 -> 2044 in the cache */
   /* replace674_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 1,
      { 2044, 3 },
      -1,
   } },

   /* ('ixor', ('ushr(is_used_once)', 'a', '#b'), ('ushr', 'c', 'b')) => ('ushr', ('ixor', 'a', 'c'), 'b') */
   /* search675_0_0 -> 128 in the cache */
   /* search675_0_1 -> 33 in the cache */
   /* search675_0 -> 1998 in the cache */
   /* search675_1_0 -> 131 in the cache */
   /* search675_1_1 -> 36 in the cache */
   /* search675_1 -> 1999 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1998, 1999 },
      -1,
   } },

   /* replace675_0_0 -> 128 in the cache */
   /* replace675_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 128, 131 },
      -1,
   } },
   /* replace675_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2053, 36 },
      -1,
   } },

   /* ('ixor', ('ishr(is_used_once)', 'a', '#b'), ('ishr', 'c', 'b')) => ('ishr', ('ixor', 'a', 'c'), 'b') */
   /* search676_0_0 -> 128 in the cache */
   /* search676_0_1 -> 33 in the cache */
   /* search676_0 -> 2003 in the cache */
   /* search676_1_0 -> 131 in the cache */
   /* search676_1_1 -> 36 in the cache */
   /* search676_1 -> 2004 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2003, 2004 },
      -1,
   } },

   /* replace676_0_0 -> 128 in the cache */
   /* replace676_0_1 -> 131 in the cache */
   /* replace676_0 -> 2053 in the cache */
   /* replace676_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2053, 36 },
      -1,
   } },

   /* ('ixor', ('ishl(is_used_once)', 'a', '#b'), ('ishl', 'c', 'b')) => ('ishl', ('ixor', 'a', 'c'), 'b') */
   /* search677_0_0 -> 128 in the cache */
   /* search677_0_1 -> 33 in the cache */
   /* search677_0 -> 2007 in the cache */
   /* search677_1_0 -> 131 in the cache */
   /* search677_1_1 -> 36 in the cache */
   /* search677_1 -> 2008 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 2007, 2008 },
      -1,
   } },

   /* replace677_0_0 -> 128 in the cache */
   /* replace677_0_1 -> 131 in the cache */
   /* replace677_0 -> 2053 in the cache */
   /* replace677_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2053, 36 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search678_0_0 -> 6 in the cache */
   /* search678_0_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 797 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search678_1_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 2060, 797 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2059, 2061 },
      -1,
   } },

   /* replace678_0_0 -> 6 in the cache */
   /* replace678_0_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 6, 2060 },
      -1,
   } },
   /* replace678_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2063, 797 },
      -1,
   } },

   /* ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search679_0_0 -> 6 in the cache */
   /* search679_0_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 797 },
      -1,
   } },
   /* search679_1_0 -> 2060 in the cache */
   /* search679_1_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 2060, 797 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2065, 2066 },
      -1,
   } },

   /* replace679_0_0 -> 6 in the cache */
   /* replace679_0_1 -> 2060 in the cache */
   /* replace679_0 -> 2063 in the cache */
   /* replace679_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2063, 797 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search680_0_0 -> 6 in the cache */
   /* search680_0_1 -> 797 in the cache */
   /* search680_0 -> 2059 in the cache */
   /* search680_1_0 -> 2060 in the cache */
   /* search680_1_1 -> 797 in the cache */
   /* search680_1 -> 2061 in the cache */
   /* search680 -> 2062 in the cache */

   /* replace680_0_0 -> 6 in the cache */
   /* replace680_0_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 6, 2060 },
      -1,
   } },
   /* replace680_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2069, 797 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search681_0_0 -> 6 in the cache */
   /* search681_0_1 -> 797 in the cache */
   /* search681_0 -> 2059 in the cache */
   /* search681_1_0 -> 2060 in the cache */
   /* search681_1_1 -> 797 in the cache */
   /* search681_1 -> 2061 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2059, 2061 },
      -1,
   } },

   /* replace681_0_0 -> 6 in the cache */
   /* replace681_0_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 6, 2060 },
      -1,
   } },
   /* replace681_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2072, 797 },
      -1,
   } },

   /* ('iand', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search682_0_0 -> 6 in the cache */
   /* search682_0_1 -> 797 in the cache */
   /* search682_0 -> 2065 in the cache */
   /* search682_1_0 -> 2060 in the cache */
   /* search682_1_1 -> 797 in the cache */
   /* search682_1 -> 2066 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2065, 2066 },
      -1,
   } },

   /* replace682_0_0 -> 6 in the cache */
   /* replace682_0_1 -> 2060 in the cache */
   /* replace682_0 -> 2072 in the cache */
   /* replace682_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2072, 797 },
      -1,
   } },

   /* ('ior', ('ine', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search683_0_0 -> 6 in the cache */
   /* search683_0_1 -> 797 in the cache */
   /* search683_0 -> 2065 in the cache */
   /* search683_1_0 -> 2060 in the cache */
   /* search683_1_1 -> 797 in the cache */
   /* search683_1 -> 2066 in the cache */
   /* search683 -> 2067 in the cache */

   /* replace683_0_0 -> 6 in the cache */
   /* replace683_0_1 -> 2060 in the cache */
   /* replace683_0 -> 2069 in the cache */
   /* replace683_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2069, 797 },
      -1,
   } },

   /* ('ineg', ('b2i8', 'a@8')) => a */
   /* search684_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2i8,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2077 },
      -1,
   } },

   /* replace684 -> 6 in the cache */

   /* ('ishl', 'a@8', ('iand', 7, 'b')) => ('ishl', 'a', 'b') */
   /* search685_0 -> 6 in the cache */
   /* search685_1_0 -> 9 in the cache */
   /* search685_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 9, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 6, 2079 },
      -1,
   } },

   /* replace685_0 -> 6 in the cache */
   /* replace685_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ishr', 'a@8', ('iand', 7, 'b')) => ('ishr', 'a', 'b') */
   /* search686_0 -> 6 in the cache */
   /* search686_1_0 -> 9 in the cache */
   /* search686_1_1 -> 36 in the cache */
   /* search686_1 -> 2079 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 6, 2079 },
      -1,
   } },

   /* replace686_0 -> 6 in the cache */
   /* replace686_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ushr', 'a@8', ('iand', 7, 'b')) => ('ushr', 'a', 'b') */
   /* search687_0 -> 6 in the cache */
   /* search687_1_0 -> 9 in the cache */
   /* search687_1_1 -> 36 in the cache */
   /* search687_1 -> 2079 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 2079 },
      -1,
   } },

   /* replace687_0 -> 6 in the cache */
   /* replace687_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 36 },
      -1,
   } },

   /* ('ushr', 'a@8', ('ishl(is_used_once)', ('iand', 'b', 1), 2)) => ('ushr', 'a', ('ishl', 'b', 2)) */
   /* search688_0 -> 6 in the cache */
   /* search688_1_0_0 -> 36 in the cache */
   /* search688_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 36, 381 },
      -1,
   } },
   /* search688_1_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2086, 386 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 6, 2087 },
      -1,
   } },

   /* replace688_0 -> 6 in the cache */
   /* replace688_1_0 -> 36 in the cache */
   /* replace688_1_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 6, 2089 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search689_0_0 -> 11 in the cache */
   /* search689_0_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 818 },
      -1,
   } },
   /* search689_1_0 -> 247 in the cache */
   /* search689_1_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 247, 818 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2091, 2092 },
      -1,
   } },

   /* replace689_0_0 -> 11 in the cache */
   /* replace689_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 11, 247 },
      -1,
   } },
   /* replace689_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2094, 818 },
      -1,
   } },

   /* ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search690_0_0 -> 11 in the cache */
   /* search690_0_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 818 },
      -1,
   } },
   /* search690_1_0 -> 247 in the cache */
   /* search690_1_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 247, 818 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2096, 2097 },
      -1,
   } },

   /* replace690_0_0 -> 11 in the cache */
   /* replace690_0_1 -> 247 in the cache */
   /* replace690_0 -> 2094 in the cache */
   /* replace690_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2094, 818 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search691_0_0 -> 11 in the cache */
   /* search691_0_1 -> 818 in the cache */
   /* search691_0 -> 2091 in the cache */
   /* search691_1_0 -> 247 in the cache */
   /* search691_1_1 -> 818 in the cache */
   /* search691_1 -> 2092 in the cache */
   /* search691 -> 2093 in the cache */

   /* replace691_0_0 -> 11 in the cache */
   /* replace691_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 11, 247 },
      -1,
   } },
   /* replace691_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2100, 818 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search692_0_0 -> 11 in the cache */
   /* search692_0_1 -> 818 in the cache */
   /* search692_0 -> 2091 in the cache */
   /* search692_1_0 -> 247 in the cache */
   /* search692_1_1 -> 818 in the cache */
   /* search692_1 -> 2092 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2091, 2092 },
      -1,
   } },

   /* replace692_0_0 -> 11 in the cache */
   /* replace692_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 11, 247 },
      -1,
   } },
   /* replace692_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2103, 818 },
      -1,
   } },

   /* ('iand', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search693_0_0 -> 11 in the cache */
   /* search693_0_1 -> 818 in the cache */
   /* search693_0 -> 2096 in the cache */
   /* search693_1_0 -> 247 in the cache */
   /* search693_1_1 -> 818 in the cache */
   /* search693_1 -> 2097 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2096, 2097 },
      -1,
   } },

   /* replace693_0_0 -> 11 in the cache */
   /* replace693_0_1 -> 247 in the cache */
   /* replace693_0 -> 2103 in the cache */
   /* replace693_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2103, 818 },
      -1,
   } },

   /* ('ior', ('ine', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search694_0_0 -> 11 in the cache */
   /* search694_0_1 -> 818 in the cache */
   /* search694_0 -> 2096 in the cache */
   /* search694_1_0 -> 247 in the cache */
   /* search694_1_1 -> 818 in the cache */
   /* search694_1 -> 2097 in the cache */
   /* search694 -> 2098 in the cache */

   /* replace694_0_0 -> 11 in the cache */
   /* replace694_0_1 -> 247 in the cache */
   /* replace694_0 -> 2100 in the cache */
   /* replace694_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2100, 818 },
      -1,
   } },

   /* ('ineg', ('b2i16', 'a@16')) => a */
   /* search695_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2i16,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2108 },
      -1,
   } },

   /* replace695 -> 11 in the cache */

   /* ('ishl', 'a@16', ('iand', 15, 'b')) => ('ishl', 'a', 'b') */
   /* search696_0 -> 11 in the cache */
   /* search696_1_0 -> 14 in the cache */
   /* search696_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 14, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 2110 },
      -1,
   } },

   /* replace696_0 -> 11 in the cache */
   /* replace696_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ishr', 'a@16', ('iand', 15, 'b')) => ('ishr', 'a', 'b') */
   /* search697_0 -> 11 in the cache */
   /* search697_1_0 -> 14 in the cache */
   /* search697_1_1 -> 36 in the cache */
   /* search697_1 -> 2110 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 11, 2110 },
      -1,
   } },

   /* replace697_0 -> 11 in the cache */
   /* replace697_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ushr', 'a@16', ('iand', 15, 'b')) => ('ushr', 'a', 'b') */
   /* search698_0 -> 11 in the cache */
   /* search698_1_0 -> 14 in the cache */
   /* search698_1_1 -> 36 in the cache */
   /* search698_1 -> 2110 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2110 },
      -1,
   } },

   /* replace698_0 -> 11 in the cache */
   /* replace698_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ushr', 'a@16', ('ishl(is_used_once)', ('iand', 'b', 1), 3)) => ('ushr', 'a', ('ishl', 'b', 3)) */
   /* search699_0 -> 11 in the cache */
   /* search699_1_0_0 -> 36 in the cache */
   /* search699_1_0_1 -> 381 in the cache */
   /* search699_1_0 -> 2086 in the cache */
   /* search699_1_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2086, 390 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2117 },
      -1,
   } },

   /* replace699_0 -> 11 in the cache */
   /* replace699_1_0 -> 36 in the cache */
   /* replace699_1_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 390 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 2119 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search700_0_0 -> 16 in the cache */
   /* search700_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 124 },
      -1,
   } },
   /* search700_1_0 -> 36 in the cache */
   /* search700_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 36, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2121, 2122 },
      -1,
   } },

   /* replace700_0_0 -> 16 in the cache */
   /* replace700_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace700_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2124, 124 },
      -1,
   } },

   /* ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search701_0_0 -> 16 in the cache */
   /* search701_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 124 },
      -1,
   } },
   /* search701_1_0 -> 36 in the cache */
   /* search701_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 36, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2126, 2127 },
      -1,
   } },

   /* replace701_0_0 -> 16 in the cache */
   /* replace701_0_1 -> 36 in the cache */
   /* replace701_0 -> 2124 in the cache */
   /* replace701_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2124, 124 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search702_0_0 -> 16 in the cache */
   /* search702_0_1 -> 124 in the cache */
   /* search702_0 -> 2121 in the cache */
   /* search702_1_0 -> 36 in the cache */
   /* search702_1_1 -> 124 in the cache */
   /* search702_1 -> 2122 in the cache */
   /* search702 -> 2123 in the cache */

   /* replace702_0_0 -> 16 in the cache */
   /* replace702_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace702_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2130, 124 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search703_0_0 -> 16 in the cache */
   /* search703_0_1 -> 124 in the cache */
   /* search703_0 -> 2121 in the cache */
   /* search703_1_0 -> 36 in the cache */
   /* search703_1_1 -> 124 in the cache */
   /* search703_1 -> 2122 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2121, 2122 },
      -1,
   } },

   /* replace703_0_0 -> 16 in the cache */
   /* replace703_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace703_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2133, 124 },
      -1,
   } },

   /* ('iand', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search704_0_0 -> 16 in the cache */
   /* search704_0_1 -> 124 in the cache */
   /* search704_0 -> 2126 in the cache */
   /* search704_1_0 -> 36 in the cache */
   /* search704_1_1 -> 124 in the cache */
   /* search704_1 -> 2127 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2126, 2127 },
      -1,
   } },

   /* replace704_0_0 -> 16 in the cache */
   /* replace704_0_1 -> 36 in the cache */
   /* replace704_0 -> 2133 in the cache */
   /* replace704_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2133, 124 },
      -1,
   } },

   /* ('ior', ('ine', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search705_0_0 -> 16 in the cache */
   /* search705_0_1 -> 124 in the cache */
   /* search705_0 -> 2126 in the cache */
   /* search705_1_0 -> 36 in the cache */
   /* search705_1_1 -> 124 in the cache */
   /* search705_1 -> 2127 in the cache */
   /* search705 -> 2128 in the cache */

   /* replace705_0_0 -> 16 in the cache */
   /* replace705_0_1 -> 36 in the cache */
   /* replace705_0 -> 2130 in the cache */
   /* replace705_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2130, 124 },
      -1,
   } },

   /* ('ineg', ('b2i32', 'a@32')) => a */
   /* search706_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2138 },
      -1,
   } },

   /* replace706 -> 16 in the cache */

   /* ('ishl', 'a@32', ('iand', 31, 'b')) => ('ishl', 'a', 'b') */
   /* search707_0 -> 16 in the cache */
   /* search707_1_0 -> 19 in the cache */
   /* search707_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 2140 },
      -1,
   } },

   /* replace707_0 -> 16 in the cache */
   /* replace707_1 -> 36 in the cache */
   /* replace707 -> 908 in the cache */

   /* ('ishr', 'a@32', ('iand', 31, 'b')) => ('ishr', 'a', 'b') */
   /* search708_0 -> 16 in the cache */
   /* search708_1_0 -> 19 in the cache */
   /* search708_1_1 -> 36 in the cache */
   /* search708_1 -> 2140 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 16, 2140 },
      -1,
   } },

   /* replace708_0 -> 16 in the cache */
   /* replace708_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ushr', 'a@32', ('iand', 31, 'b')) => ('ushr', 'a', 'b') */
   /* search709_0 -> 16 in the cache */
   /* search709_1_0 -> 19 in the cache */
   /* search709_1_1 -> 36 in the cache */
   /* search709_1 -> 2140 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2140 },
      -1,
   } },

   /* replace709_0 -> 16 in the cache */
   /* replace709_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ushr', 'a@32', ('ishl(is_used_once)', ('iand', 'b', 1), 4)) => ('ushr', 'a', ('ishl', 'b', 4)) */
   /* search710_0 -> 16 in the cache */
   /* search710_1_0_0 -> 36 in the cache */
   /* search710_1_0_1 -> 381 in the cache */
   /* search710_1_0 -> 2086 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x4 /* 4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2086, 2146 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2147 },
      -1,
   } },

   /* replace710_0 -> 16 in the cache */
   /* replace710_1_0 -> 36 in the cache */
   /* replace710_1_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 2146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 2149 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('ior', 'a', 'b'), 0) */
   /* search711_0_0 -> 21 in the cache */
   /* search711_0_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 21, 859 },
      -1,
   } },
   /* search711_1_0 -> 41 in the cache */
   /* search711_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 41, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2151, 2152 },
      -1,
   } },

   /* replace711_0_0 -> 21 in the cache */
   /* replace711_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace711_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2154, 859 },
      -1,
   } },

   /* ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('ior', 'a', 'b'), 0) */
   /* search712_0_0 -> 21 in the cache */
   /* search712_0_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 21, 859 },
      -1,
   } },
   /* search712_1_0 -> 41 in the cache */
   /* search712_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 41, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2156, 2157 },
      -1,
   } },

   /* replace712_0_0 -> 21 in the cache */
   /* replace712_0_1 -> 41 in the cache */
   /* replace712_0 -> 2154 in the cache */
   /* replace712_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2154, 859 },
      -1,
   } },

   /* ('iand', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umax', 'a', 'b'), 0) */
   /* search713_0_0 -> 21 in the cache */
   /* search713_0_1 -> 859 in the cache */
   /* search713_0 -> 2151 in the cache */
   /* search713_1_0 -> 41 in the cache */
   /* search713_1_1 -> 859 in the cache */
   /* search713_1 -> 2152 in the cache */
   /* search713 -> 2153 in the cache */

   /* replace713_0_0 -> 21 in the cache */
   /* replace713_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace713_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2160, 859 },
      -1,
   } },

   /* ('ior', ('ieq', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search714_0_0 -> 21 in the cache */
   /* search714_0_1 -> 859 in the cache */
   /* search714_0 -> 2151 in the cache */
   /* search714_1_0 -> 41 in the cache */
   /* search714_1_1 -> 859 in the cache */
   /* search714_1 -> 2152 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2151, 2152 },
      -1,
   } },

   /* replace714_0_0 -> 21 in the cache */
   /* replace714_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace714_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 2163, 859 },
      -1,
   } },

   /* ('iand', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search715_0_0 -> 21 in the cache */
   /* search715_0_1 -> 859 in the cache */
   /* search715_0 -> 2156 in the cache */
   /* search715_1_0 -> 41 in the cache */
   /* search715_1_1 -> 859 in the cache */
   /* search715_1 -> 2157 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2156, 2157 },
      -1,
   } },

   /* replace715_0_0 -> 21 in the cache */
   /* replace715_0_1 -> 41 in the cache */
   /* replace715_0 -> 2163 in the cache */
   /* replace715_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2163, 859 },
      -1,
   } },

   /* ('ior', ('ine', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umax', 'a', 'b'), 0) */
   /* search716_0_0 -> 21 in the cache */
   /* search716_0_1 -> 859 in the cache */
   /* search716_0 -> 2156 in the cache */
   /* search716_1_0 -> 41 in the cache */
   /* search716_1_1 -> 859 in the cache */
   /* search716_1 -> 2157 in the cache */
   /* search716 -> 2158 in the cache */

   /* replace716_0_0 -> 21 in the cache */
   /* replace716_0_1 -> 41 in the cache */
   /* replace716_0 -> 2160 in the cache */
   /* replace716_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 2160, 859 },
      -1,
   } },

   /* ('ineg', ('b2i64', 'a@64')) => a */
   /* search717_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2i64,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2168 },
      -1,
   } },

   /* replace717 -> 21 in the cache */

   /* ('ishl', 'a@64', ('iand', 63, 'b')) => ('ishl', 'a', 'b') */
   /* search718_0 -> 21 in the cache */
   /* search718_1_0 -> 24 in the cache */
   /* search718_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 24, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 21, 2170 },
      -1,
   } },

   /* replace718_0 -> 21 in the cache */
   /* replace718_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ishr', 'a@64', ('iand', 63, 'b')) => ('ishr', 'a', 'b') */
   /* search719_0 -> 21 in the cache */
   /* search719_1_0 -> 24 in the cache */
   /* search719_1_1 -> 36 in the cache */
   /* search719_1 -> 2170 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 21, 2170 },
      -1,
   } },

   /* replace719_0 -> 21 in the cache */
   /* replace719_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ushr', 'a@64', ('iand', 63, 'b')) => ('ushr', 'a', 'b') */
   /* search720_0 -> 21 in the cache */
   /* search720_1_0 -> 24 in the cache */
   /* search720_1_1 -> 36 in the cache */
   /* search720_1 -> 2170 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 2170 },
      -1,
   } },

   /* replace720_0 -> 21 in the cache */
   /* replace720_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 36 },
      -1,
   } },

   /* ('ushr', 'a@64', ('ishl(is_used_once)', ('iand', 'b', 1), 5)) => ('ushr', 'a', ('ishl', 'b', 5)) */
   /* search721_0 -> 21 in the cache */
   /* search721_1_0_0 -> 36 in the cache */
   /* search721_1_0_1 -> 381 in the cache */
   /* search721_1_0 -> 2086 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x5 /* 5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 2086, 2177 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 21, 2178 },
      -1,
   } },

   /* replace721_0 -> 21 in the cache */
   /* replace721_1_0 -> 36 in the cache */
   /* replace721_1_1 -> 2177 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 36, 2177 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 2180 },
      -1,
   } },

   /* ('ior', ('ieq', 'a', 0), ('ieq', 'a', 1)) => ('uge', 1, 'a') */
   /* search722_0_0 -> 32 in the cache */
   /* search722_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 61 },
      -1,
   } },
   /* search722_1_0 -> 32 in the cache */
   /* search722_1_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 2182, 2183 },
      -1,
   } },

   /* replace722_0 -> 35 in the cache */
   /* replace722_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 35, 32 },
      -1,
   } },

   /* ('ior', ('uge', 1, 'a'), ('ieq', 'a', 2)) => ('uge', 2, 'a') */
   /* search723_0_0 -> 35 in the cache */
   /* search723_0_1 -> 32 in the cache */
   /* search723_0 -> 2185 in the cache */
   /* search723_1_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 2186 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2185, 2187 },
      -1,
   } },

   /* replace723_0 -> 2186 in the cache */
   /* replace723_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2186, 32 },
      -1,
   } },

   /* ('ior', ('uge', 2, 'a'), ('ieq', 'a', 3)) => ('uge', 3, 'a') */
   /* search724_0_0 -> 2186 in the cache */
   /* search724_0_1 -> 32 in the cache */
   /* search724_0 -> 2189 in the cache */
   /* search724_1_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 32, 2190 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2189, 2191 },
      -1,
   } },

   /* replace724_0 -> 2190 in the cache */
   /* replace724_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 2190, 32 },
      -1,
   } },

   /* ('ior', 'a', ('ieq', 'a', False)) => True */
   /* search725_0 -> 722 in the cache */
   /* search725_1_0 -> 722 in the cache */
   /* search725_1_1 -> 1039 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 722, 1039 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 722, 2194 },
      -1,
   } },

   /* replace725 -> 1192 in the cache */

   /* ('ior', 'a', ('inot', 'a')) => -1 */
   /* search726_0 -> 32 in the cache */
   /* search726_1_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 2196 },
      -1,
   } },

   /* replace726 -> 63 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ine', 'a', 'b') */
   /* search727_0_0_0 -> 722 in the cache */
   /* search727_0_0 -> 1183 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1183 },
      -1,
   } },
   /* search727_1_0_0 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1090 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2199 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace727_0 -> 722 in the cache */
   /* replace727_1 -> 1090 in the cache */
   /* replace727 -> 1335 in the cache */

   /* ('b2i', ('ine', 'a@1', 'b@1')) => ('b2i', ('ixor', 'a', 'b')) */
   /* search728_0_0 -> 722 in the cache */
   /* search728_0_1 -> 1090 in the cache */
   /* search728_0 -> 1335 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1335 },
      -1,
   } },

   /* replace728_0_0 -> 722 in the cache */
   /* replace728_0_1 -> 1090 in the cache */
   /* replace728_0 -> 1107 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1107 },
      -1,
   } },

   /* ('iand', ('inot', ('iand', ('ior', ('ieq', 'a', 0), 'b'), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', 'b', 'c')), ('ilt', 'a', 0)) */
   /* search729_0_0_0_0_0 -> 32 in the cache */
   /* search729_0_0_0_0_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 32, 61 },
      -1,
   } },
   /* search729_0_0_0_1 -> 1090 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 2,
      { 2204, 1090 },
      -1,
   } },
   /* search729_0_0_1 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 3,
      { 2205, 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 2206 },
      -1,
   } },
   /* search729_1_0 -> 32 in the cache */
   /* search729_1_1 -> 61 in the cache */
   /* search729_1 -> 1206 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2207, 1206 },
      -1,
   } },

   /* replace729_0_0_0 -> 1090 in the cache */
   /* replace729_0_0_1 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1090, 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2209 },
      -1,
   } },
   /* replace729_1_0 -> 32 in the cache */
   /* replace729_1_1 -> 61 in the cache */
   /* replace729_1 -> 1206 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2210, 1206 },
      -1,
   } },

   /* ('iand', ('inot', ('iand', ('ieq', ('umin', 'a', 'b'), 0), 'c')), ('ilt', 'a', 0)) => ('iand', ('inot', ('iand', ('ieq', 'b', 0), 'c')), ('ilt', 'a', 0)) */
   /* search730_0_0_0_0_0 -> 0 in the cache */
   /* search730_0_0_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_umin,
      3, 1,
      { 0, 3 },
      -1,
   } },
   /* search730_0_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 2,
      { 2212, 86 },
      -1,
   } },
   /* search730_0_0_1 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 3,
      { 2213, 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 2214 },
      -1,
   } },
   /* search730_1_0 -> 0 in the cache */
   /* search730_1_1 -> 86 in the cache */
   /* search730_1 -> 90 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2215, 90 },
      -1,
   } },

   /* replace730_0_0_0_0 -> 3 in the cache */
   /* replace730_0_0_0_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 3, 86 },
      -1,
   } },
   /* replace730_0_0_1 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 2,
      { 2217, 449 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 2,
      { 2218 },
      -1,
   } },
   /* replace730_1_0 -> 0 in the cache */
   /* replace730_1_1 -> 86 in the cache */
   /* replace730_1 -> 90 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2219, 90 },
      -1,
   } },

   /* ('flt', 'a', ('fmax', 'b', 'a')) => ('flt', 'a', 'b') */
   /* search731_0 -> 32 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search731_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 32, 2222 },
      -1,
   } },

   /* replace731_0 -> 32 in the cache */
   /* replace731_1 -> 2221 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 2221 },
      -1,
   } },

   /* ('flt', ('fmin', 'a', 'b'), 'a') => ('flt', 'b', 'a') */
   /* search732_0_0 -> 0 in the cache */
   /* search732_0_1 -> 3 in the cache */
   /* search732_0 -> 1240 in the cache */
   /* search732_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1240, 0 },
      -1,
   } },

   /* replace732_0 -> 3 in the cache */
   /* replace732_1 -> 0 in the cache */
   /* replace732 -> 1040 in the cache */

   /* ('~fge', 'a', ('fmin', 'b', 'a')) => True */
   /* search733_0 -> 32 in the cache */
   /* search733_1_0 -> 2221 in the cache */
   /* search733_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 32, 2226 },
      -1,
   } },

   /* replace733 -> 1192 in the cache */

   /* ('~fge', ('fmax', 'a', 'b'), 'a') => True */
   /* search734_0_0 -> 0 in the cache */
   /* search734_0_1 -> 3 in the cache */
   /* search734_0 -> 1238 in the cache */
   /* search734_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1238, 0 },
      -1,
   } },

   /* replace734 -> 1192 in the cache */

   /* ('flt', 'a', ('fmin', 'b', 'a')) => False */
   /* search735_0 -> 32 in the cache */
   /* search735_1_0 -> 2221 in the cache */
   /* search735_1_1 -> 32 in the cache */
   /* search735_1 -> 2226 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 32, 2226 },
      -1,
   } },

   /* replace735 -> 1039 in the cache */

   /* ('flt', ('fmax', 'a', 'b'), 'a') => False */
   /* search736_0_0 -> 0 in the cache */
   /* search736_0_1 -> 3 in the cache */
   /* search736_0 -> 1238 in the cache */
   /* search736_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 1238, 0 },
      -1,
   } },

   /* replace736 -> 1039 in the cache */

   /* ('~fge', 'a', ('fmax', 'b', 'a')) => ('fge', 'a', 'b') */
   /* search737_0 -> 32 in the cache */
   /* search737_1_0 -> 2221 in the cache */
   /* search737_1_1 -> 32 in the cache */
   /* search737_1 -> 2222 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 32, 2222 },
      -1,
   } },

   /* replace737_0 -> 32 in the cache */
   /* replace737_1 -> 2221 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 32, 2221 },
      -1,
   } },

   /* ('~fge', ('fmin', 'a', 'b'), 'a') => ('fge', 'b', 'a') */
   /* search738_0_0 -> 0 in the cache */
   /* search738_0_1 -> 3 in the cache */
   /* search738_0 -> 1240 in the cache */
   /* search738_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 1240, 0 },
      -1,
   } },

   /* replace738_0 -> 3 in the cache */
   /* replace738_1 -> 0 in the cache */
   /* replace738 -> 1052 in the cache */

   /* ('ilt', 'a', ('imax', 'b', 'a')) => ('ilt', 'a', 'b') */
   /* search739_0 -> 32 in the cache */
   /* search739_1_0 -> 2221 in the cache */
   /* search739_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 32, 2234 },
      -1,
   } },

   /* replace739_0 -> 32 in the cache */
   /* replace739_1 -> 2221 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 2221 },
      -1,
   } },

   /* ('ilt', ('imin', 'a', 'b'), 'a') => ('ilt', 'b', 'a') */
   /* search740_0_0 -> 0 in the cache */
   /* search740_0_1 -> 3 in the cache */
   /* search740_0 -> 1300 in the cache */
   /* search740_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1300, 0 },
      -1,
   } },

   /* replace740_0 -> 3 in the cache */
   /* replace740_1 -> 0 in the cache */
   /* replace740 -> 1045 in the cache */

   /* ('ige', 'a', ('imin', 'b', 'a')) => True */
   /* search741_0 -> 32 in the cache */
   /* search741_1_0 -> 2221 in the cache */
   /* search741_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 32, 2238 },
      -1,
   } },

   /* replace741 -> 1192 in the cache */

   /* ('ige', ('imax', 'a', 'b'), 'a') => True */
   /* search742_0_0 -> 0 in the cache */
   /* search742_0_1 -> 3 in the cache */
   /* search742_0 -> 1292 in the cache */
   /* search742_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1292, 0 },
      -1,
   } },

   /* replace742 -> 1192 in the cache */

   /* ('ult', 'a', ('umax', 'b', 'a')) => ('ult', 'a', 'b') */
   /* search743_0 -> 32 in the cache */
   /* search743_1_0 -> 2221 in the cache */
   /* search743_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2241 },
      -1,
   } },

   /* replace743_0 -> 32 in the cache */
   /* replace743_1 -> 2221 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 2221 },
      -1,
   } },

   /* ('ult', ('umin', 'a', 'b'), 'a') => ('ult', 'b', 'a') */
   /* search744_0_0 -> 0 in the cache */
   /* search744_0_1 -> 3 in the cache */
   /* search744_0 -> 1297 in the cache */
   /* search744_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1297, 0 },
      -1,
   } },

   /* replace744_0 -> 3 in the cache */
   /* replace744_1 -> 0 in the cache */
   /* replace744 -> 1047 in the cache */

   /* ('uge', 'a', ('umin', 'b', 'a')) => True */
   /* search745_0 -> 32 in the cache */
   /* search745_1_0 -> 2221 in the cache */
   /* search745_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2245 },
      -1,
   } },

   /* replace745 -> 1192 in the cache */

   /* ('uge', ('umax', 'a', 'b'), 'a') => True */
   /* search746_0_0 -> 0 in the cache */
   /* search746_0_1 -> 3 in the cache */
   /* search746_0 -> 1289 in the cache */
   /* search746_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1289, 0 },
      -1,
   } },

   /* replace746 -> 1192 in the cache */

   /* ('ilt', 'a', ('imin', 'b', 'a')) => False */
   /* search747_0 -> 32 in the cache */
   /* search747_1_0 -> 2221 in the cache */
   /* search747_1_1 -> 32 in the cache */
   /* search747_1 -> 2238 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 32, 2238 },
      -1,
   } },

   /* replace747 -> 1039 in the cache */

   /* ('ilt', ('imax', 'a', 'b'), 'a') => False */
   /* search748_0_0 -> 0 in the cache */
   /* search748_0_1 -> 3 in the cache */
   /* search748_0 -> 1292 in the cache */
   /* search748_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 1292, 0 },
      -1,
   } },

   /* replace748 -> 1039 in the cache */

   /* ('ige', 'a', ('imax', 'b', 'a')) => ('ige', 'a', 'b') */
   /* search749_0 -> 32 in the cache */
   /* search749_1_0 -> 2221 in the cache */
   /* search749_1_1 -> 32 in the cache */
   /* search749_1 -> 2234 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 32, 2234 },
      -1,
   } },

   /* replace749_0 -> 32 in the cache */
   /* replace749_1 -> 2221 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 2221 },
      -1,
   } },

   /* ('ige', ('imin', 'a', 'b'), 'a') => ('ige', 'b', 'a') */
   /* search750_0_0 -> 0 in the cache */
   /* search750_0_1 -> 3 in the cache */
   /* search750_0 -> 1300 in the cache */
   /* search750_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 1300, 0 },
      -1,
   } },

   /* replace750_0 -> 3 in the cache */
   /* replace750_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('ult', 'a', ('umin', 'b', 'a')) => False */
   /* search751_0 -> 32 in the cache */
   /* search751_1_0 -> 2221 in the cache */
   /* search751_1_1 -> 32 in the cache */
   /* search751_1 -> 2245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2245 },
      -1,
   } },

   /* replace751 -> 1039 in the cache */

   /* ('ult', ('umax', 'a', 'b'), 'a') => False */
   /* search752_0_0 -> 0 in the cache */
   /* search752_0_1 -> 3 in the cache */
   /* search752_0 -> 1289 in the cache */
   /* search752_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 1289, 0 },
      -1,
   } },

   /* replace752 -> 1039 in the cache */

   /* ('uge', 'a', ('umax', 'b', 'a')) => ('uge', 'a', 'b') */
   /* search753_0 -> 32 in the cache */
   /* search753_1_0 -> 2221 in the cache */
   /* search753_1_1 -> 32 in the cache */
   /* search753_1 -> 2241 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2241 },
      -1,
   } },

   /* replace753_0 -> 32 in the cache */
   /* replace753_1 -> 2221 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 2221 },
      -1,
   } },

   /* ('uge', ('umin', 'a', 'b'), 'a') => ('uge', 'b', 'a') */
   /* search754_0_0 -> 0 in the cache */
   /* search754_0_1 -> 3 in the cache */
   /* search754_0 -> 1297 in the cache */
   /* search754_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 1297, 0 },
      -1,
   } },

   /* replace754_0 -> 3 in the cache */
   /* replace754_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3, 0 },
      -1,
   } },

   /* ('ult', 'a', ('iand', 'b', 'a')) => False */
   /* search755_0 -> 32 in the cache */
   /* search755_1_0 -> 2221 in the cache */
   /* search755_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 32, 2260 },
      -1,
   } },

   /* replace755 -> 1039 in the cache */

   /* ('ult', ('ior', 'a', 'b'), 'a') => False */
   /* search756_0_0 -> 0 in the cache */
   /* search756_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 0, 3 },
      -1,
   } },
   /* search756_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2262, 0 },
      -1,
   } },

   /* replace756 -> 1039 in the cache */

   /* ('uge', 'a', ('iand', 'b', 'a')) => True */
   /* search757_0 -> 32 in the cache */
   /* search757_1_0 -> 2221 in the cache */
   /* search757_1_1 -> 32 in the cache */
   /* search757_1 -> 2260 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 32, 2260 },
      -1,
   } },

   /* replace757 -> 1192 in the cache */

   /* ('uge', ('ior', 'a', 'b'), 'a') => True */
   /* search758_0_0 -> 0 in the cache */
   /* search758_0_1 -> 3 in the cache */
   /* search758_0 -> 2262 in the cache */
   /* search758_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2262, 0 },
      -1,
   } },

   /* replace758 -> 1192 in the cache */

   /* ('ilt', '#a', ('imax', '#b', 'c')) => ('ior', ('ilt', 'a', 'b'), ('ilt', 'a', 'c')) */
   /* search759_0 -> 271 in the cache */
   /* search759_1_0 -> 272 in the cache */
   /* search759_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 272, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 271, 2266 },
      -1,
   } },

   /* replace759_0_0 -> 128 in the cache */
   /* replace759_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 128, 129 },
      -1,
   } },
   /* replace759_1_0 -> 128 in the cache */
   /* replace759_1_1 -> 131 in the cache */
   /* replace759_1 -> 1601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2268, 1601 },
      -1,
   } },

   /* ('ilt', ('imin', '#a', 'b'), '#c') => ('ior', ('ilt', 'a', 'c'), ('ilt', 'b', 'c')) */
   /* search760_0_0 -> 271 in the cache */
   /* search760_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 271, 129 },
      -1,
   } },
   /* search760_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 2270, 776 },
      -1,
   } },

   /* replace760_0_0 -> 128 in the cache */
   /* replace760_0_1 -> 131 in the cache */
   /* replace760_0 -> 1601 in the cache */
   /* replace760_1_0 -> 129 in the cache */
   /* replace760_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1601, 2272 },
      -1,
   } },

   /* ('ige', '#a', ('imin', '#b', 'c')) => ('ior', ('ige', 'a', 'b'), ('ige', 'a', 'c')) */
   /* search761_0 -> 271 in the cache */
   /* search761_1_0 -> 272 in the cache */
   /* search761_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 272, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 271, 2274 },
      -1,
   } },

   /* replace761_0_0 -> 128 in the cache */
   /* replace761_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 128, 129 },
      -1,
   } },
   /* replace761_1_0 -> 128 in the cache */
   /* replace761_1_1 -> 131 in the cache */
   /* replace761_1 -> 1611 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2276, 1611 },
      -1,
   } },

   /* ('ige', ('imax', '#a', 'b'), '#c') => ('ior', ('ige', 'a', 'c'), ('ige', 'b', 'c')) */
   /* search762_0_0 -> 271 in the cache */
   /* search762_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 271, 129 },
      -1,
   } },
   /* search762_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2278, 776 },
      -1,
   } },

   /* replace762_0_0 -> 128 in the cache */
   /* replace762_0_1 -> 131 in the cache */
   /* replace762_0 -> 1611 in the cache */
   /* replace762_1_0 -> 129 in the cache */
   /* replace762_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1611, 2280 },
      -1,
   } },

   /* ('ult', '#a', ('umax', '#b', 'c')) => ('ior', ('ult', 'a', 'b'), ('ult', 'a', 'c')) */
   /* search763_0 -> 271 in the cache */
   /* search763_1_0 -> 272 in the cache */
   /* search763_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 272, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 271, 2282 },
      -1,
   } },

   /* replace763_0_0 -> 128 in the cache */
   /* replace763_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 128, 129 },
      -1,
   } },
   /* replace763_1_0 -> 128 in the cache */
   /* replace763_1_1 -> 131 in the cache */
   /* replace763_1 -> 1621 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2284, 1621 },
      -1,
   } },

   /* ('ult', ('umin', '#a', 'b'), '#c') => ('ior', ('ult', 'a', 'c'), ('ult', 'b', 'c')) */
   /* search764_0_0 -> 271 in the cache */
   /* search764_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 271, 129 },
      -1,
   } },
   /* search764_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2286, 776 },
      -1,
   } },

   /* replace764_0_0 -> 128 in the cache */
   /* replace764_0_1 -> 131 in the cache */
   /* replace764_0 -> 1621 in the cache */
   /* replace764_1_0 -> 129 in the cache */
   /* replace764_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1621, 2288 },
      -1,
   } },

   /* ('uge', '#a', ('umin', '#b', 'c')) => ('ior', ('uge', 'a', 'b'), ('uge', 'a', 'c')) */
   /* search765_0 -> 271 in the cache */
   /* search765_1_0 -> 272 in the cache */
   /* search765_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 272, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 271, 2290 },
      -1,
   } },

   /* replace765_0_0 -> 128 in the cache */
   /* replace765_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 128, 129 },
      -1,
   } },
   /* replace765_1_0 -> 128 in the cache */
   /* replace765_1_1 -> 131 in the cache */
   /* replace765_1 -> 1631 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2292, 1631 },
      -1,
   } },

   /* ('uge', ('umax', '#a', 'b'), '#c') => ('ior', ('uge', 'a', 'c'), ('uge', 'b', 'c')) */
   /* search766_0_0 -> 271 in the cache */
   /* search766_0_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 271, 129 },
      -1,
   } },
   /* search766_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2294, 776 },
      -1,
   } },

   /* replace766_0_0 -> 128 in the cache */
   /* replace766_0_1 -> 131 in the cache */
   /* replace766_0 -> 1631 in the cache */
   /* replace766_1_0 -> 129 in the cache */
   /* replace766_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 129, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1631, 2296 },
      -1,
   } },

   /* ('ilt', '#a', ('imin', '#b', 'c')) => ('iand', ('ilt', 'a', 'b'), ('ilt', 'a', 'c')) */
   /* search767_0 -> 271 in the cache */
   /* search767_1_0 -> 272 in the cache */
   /* search767_1_1 -> 131 in the cache */
   /* search767_1 -> 2274 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 271, 2274 },
      -1,
   } },

   /* replace767_0_0 -> 128 in the cache */
   /* replace767_0_1 -> 129 in the cache */
   /* replace767_0 -> 2268 in the cache */
   /* replace767_1_0 -> 128 in the cache */
   /* replace767_1_1 -> 131 in the cache */
   /* replace767_1 -> 1601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2268, 1601 },
      -1,
   } },

   /* ('ilt', ('imax', '#a', 'b'), '#c') => ('iand', ('ilt', 'a', 'c'), ('ilt', 'b', 'c')) */
   /* search768_0_0 -> 271 in the cache */
   /* search768_0_1 -> 129 in the cache */
   /* search768_0 -> 2278 in the cache */
   /* search768_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 2278, 776 },
      -1,
   } },

   /* replace768_0_0 -> 128 in the cache */
   /* replace768_0_1 -> 131 in the cache */
   /* replace768_0 -> 1601 in the cache */
   /* replace768_1_0 -> 129 in the cache */
   /* replace768_1_1 -> 131 in the cache */
   /* replace768_1 -> 2272 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1601, 2272 },
      -1,
   } },

   /* ('ige', '#a', ('imax', '#b', 'c')) => ('iand', ('ige', 'a', 'b'), ('ige', 'a', 'c')) */
   /* search769_0 -> 271 in the cache */
   /* search769_1_0 -> 272 in the cache */
   /* search769_1_1 -> 131 in the cache */
   /* search769_1 -> 2266 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 271, 2266 },
      -1,
   } },

   /* replace769_0_0 -> 128 in the cache */
   /* replace769_0_1 -> 129 in the cache */
   /* replace769_0 -> 2276 in the cache */
   /* replace769_1_0 -> 128 in the cache */
   /* replace769_1_1 -> 131 in the cache */
   /* replace769_1 -> 1611 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2276, 1611 },
      -1,
   } },

   /* ('ige', ('imin', '#a', 'b'), '#c') => ('iand', ('ige', 'a', 'c'), ('ige', 'b', 'c')) */
   /* search770_0_0 -> 271 in the cache */
   /* search770_0_1 -> 129 in the cache */
   /* search770_0 -> 2270 in the cache */
   /* search770_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 2270, 776 },
      -1,
   } },

   /* replace770_0_0 -> 128 in the cache */
   /* replace770_0_1 -> 131 in the cache */
   /* replace770_0 -> 1611 in the cache */
   /* replace770_1_0 -> 129 in the cache */
   /* replace770_1_1 -> 131 in the cache */
   /* replace770_1 -> 2280 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1611, 2280 },
      -1,
   } },

   /* ('ult', '#a', ('umin', '#b', 'c')) => ('iand', ('ult', 'a', 'b'), ('ult', 'a', 'c')) */
   /* search771_0 -> 271 in the cache */
   /* search771_1_0 -> 272 in the cache */
   /* search771_1_1 -> 131 in the cache */
   /* search771_1 -> 2290 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 271, 2290 },
      -1,
   } },

   /* replace771_0_0 -> 128 in the cache */
   /* replace771_0_1 -> 129 in the cache */
   /* replace771_0 -> 2284 in the cache */
   /* replace771_1_0 -> 128 in the cache */
   /* replace771_1_1 -> 131 in the cache */
   /* replace771_1 -> 1621 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2284, 1621 },
      -1,
   } },

   /* ('ult', ('umax', '#a', 'b'), '#c') => ('iand', ('ult', 'a', 'c'), ('ult', 'b', 'c')) */
   /* search772_0_0 -> 271 in the cache */
   /* search772_0_1 -> 129 in the cache */
   /* search772_0 -> 2294 in the cache */
   /* search772_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 2294, 776 },
      -1,
   } },

   /* replace772_0_0 -> 128 in the cache */
   /* replace772_0_1 -> 131 in the cache */
   /* replace772_0 -> 1621 in the cache */
   /* replace772_1_0 -> 129 in the cache */
   /* replace772_1_1 -> 131 in the cache */
   /* replace772_1 -> 2288 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1621, 2288 },
      -1,
   } },

   /* ('uge', '#a', ('umax', '#b', 'c')) => ('iand', ('uge', 'a', 'b'), ('uge', 'a', 'c')) */
   /* search773_0 -> 271 in the cache */
   /* search773_1_0 -> 272 in the cache */
   /* search773_1_1 -> 131 in the cache */
   /* search773_1 -> 2282 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 271, 2282 },
      -1,
   } },

   /* replace773_0_0 -> 128 in the cache */
   /* replace773_0_1 -> 129 in the cache */
   /* replace773_0 -> 2292 in the cache */
   /* replace773_1_0 -> 128 in the cache */
   /* replace773_1_1 -> 131 in the cache */
   /* replace773_1 -> 1631 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2292, 1631 },
      -1,
   } },

   /* ('uge', ('umin', '#a', 'b'), '#c') => ('iand', ('uge', 'a', 'c'), ('uge', 'b', 'c')) */
   /* search774_0_0 -> 271 in the cache */
   /* search774_0_1 -> 129 in the cache */
   /* search774_0 -> 2286 in the cache */
   /* search774_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 1,
      { 2286, 776 },
      -1,
   } },

   /* replace774_0_0 -> 128 in the cache */
   /* replace774_0_1 -> 131 in the cache */
   /* replace774_0 -> 1631 in the cache */
   /* replace774_1_0 -> 129 in the cache */
   /* replace774_1_1 -> 131 in the cache */
   /* replace774_1 -> 2296 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1631, 2296 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a', 0), ('ineg', ('ishr', 'a', 'b')), ('ishr', 'a', 'b')) => ('iabs', ('ishr', 'a', 'b')) */
   /* search775_0_0 -> 32 in the cache */
   /* search775_0_1 -> 61 in the cache */
   /* search775_0 -> 1206 in the cache */
   /* search775_1_0_0 -> 32 in the cache */
   /* search775_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 32, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2314 },
      -1,
   } },
   /* search775_2_0 -> 32 in the cache */
   /* search775_2_1 -> 36 in the cache */
   /* search775_2 -> 2314 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1206, 2315, 2314 },
      -1,
   } },

   /* replace775_0_0 -> 32 in the cache */
   /* replace775_0_1 -> 36 in the cache */
   /* replace775_0 -> 2314 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2314 },
      -1,
   } },

   /* ('iabs', ('ishr', ('iabs', 'a'), 'b')) => ('ishr', ('iabs', 'a'), 'b') */
   /* search776_0_0_0 -> 32 in the cache */
   /* search776_0_0 -> 106 in the cache */
   /* search776_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 106, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2318 },
      -1,
   } },

   /* replace776_0_0 -> 32 in the cache */
   /* replace776_0 -> 106 in the cache */
   /* replace776_1 -> 36 in the cache */
   /* replace776 -> 2318 in the cache */

   /* ('fabs', ('slt', 'a', 'b')) => ('slt', 'a', 'b') */
   /* search777_0_0 -> 0 in the cache */
   /* search777_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2320 },
      -1,
   } },

   /* replace777_0 -> 0 in the cache */
   /* replace777_1 -> 3 in the cache */
   /* replace777 -> 2320 in the cache */

   /* ('fabs', ('sge', 'a', 'b')) => ('sge', 'a', 'b') */
   /* search778_0_0 -> 0 in the cache */
   /* search778_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2322 },
      -1,
   } },

   /* replace778_0 -> 0 in the cache */
   /* replace778_1 -> 3 in the cache */
   /* replace778 -> 2322 in the cache */

   /* ('fabs', ('seq', 'a', 'b')) => ('seq', 'a', 'b') */
   /* search779_0_0 -> 0 in the cache */
   /* search779_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 2324 },
      -1,
   } },

   /* replace779_0 -> 0 in the cache */
   /* replace779_1 -> 3 in the cache */
   /* replace779 -> 2324 in the cache */

   /* ('fabs', ('sne', 'a', 'b')) => ('sne', 'a', 'b') */
   /* search780_0_0 -> 0 in the cache */
   /* search780_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 2326 },
      -1,
   } },

   /* replace780_0 -> 0 in the cache */
   /* replace780_1 -> 3 in the cache */
   /* replace780 -> 2326 in the cache */

   /* ('slt', 'a', 'b') => ('b2f', ('flt', 'a', 'b')) */
   /* search781_0 -> 0 in the cache */
   /* search781_1 -> 3 in the cache */
   /* search781 -> 2320 in the cache */

   /* replace781_0_0 -> 0 in the cache */
   /* replace781_0_1 -> 3 in the cache */
   /* replace781_0 -> 1011 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1011 },
      -1,
   } },

   /* ('sge', 'a', 'b') => ('b2f', ('fge', 'a', 'b')) */
   /* search782_0 -> 0 in the cache */
   /* search782_1 -> 3 in the cache */
   /* search782 -> 2322 in the cache */

   /* replace782_0_0 -> 0 in the cache */
   /* replace782_0_1 -> 3 in the cache */
   /* replace782_0 -> 1008 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1008 },
      -1,
   } },

   /* ('seq', 'a', 'b') => ('b2f', ('feq', 'a', 'b')) */
   /* search783_0 -> 0 in the cache */
   /* search783_1 -> 3 in the cache */
   /* search783 -> 2324 in the cache */

   /* replace783_0_0 -> 0 in the cache */
   /* replace783_0_1 -> 3 in the cache */
   /* replace783_0 -> 1017 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1017 },
      -1,
   } },

   /* ('sne', 'a', 'b') => ('b2f', ('fneu', 'a', 'b')) */
   /* search784_0 -> 0 in the cache */
   /* search784_1 -> 3 in the cache */
   /* search784 -> 2326 in the cache */

   /* replace784_0_0 -> 0 in the cache */
   /* replace784_0_1 -> 3 in the cache */
   /* replace784_0 -> 1014 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 1,
      { 1014 },
      -1,
   } },

   /* ('seq', ('seq', 'a', 'b'), 1.0) => ('seq', 'a', 'b') */
   /* search785_0_0 -> 0 in the cache */
   /* search785_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search785_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2332, 259 },
      -1,
   } },

   /* replace785_0 -> 0 in the cache */
   /* replace785_1 -> 3 in the cache */
   /* replace785 -> 2324 in the cache */

   /* ('seq', ('sne', 'a', 'b'), 1.0) => ('sne', 'a', 'b') */
   /* search786_0_0 -> 0 in the cache */
   /* search786_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search786_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2334, 259 },
      -1,
   } },

   /* replace786_0 -> 0 in the cache */
   /* replace786_1 -> 3 in the cache */
   /* replace786 -> 2326 in the cache */

   /* ('seq', ('slt', 'a', 'b'), 1.0) => ('slt', 'a', 'b') */
   /* search787_0_0 -> 0 in the cache */
   /* search787_0_1 -> 3 in the cache */
   /* search787_0 -> 2320 in the cache */
   /* search787_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2320, 259 },
      -1,
   } },

   /* replace787_0 -> 0 in the cache */
   /* replace787_1 -> 3 in the cache */
   /* replace787 -> 2320 in the cache */

   /* ('seq', ('sge', 'a', 'b'), 1.0) => ('sge', 'a', 'b') */
   /* search788_0_0 -> 0 in the cache */
   /* search788_0_1 -> 3 in the cache */
   /* search788_0 -> 2322 in the cache */
   /* search788_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2322, 259 },
      -1,
   } },

   /* replace788_0 -> 0 in the cache */
   /* replace788_1 -> 3 in the cache */
   /* replace788 -> 2322 in the cache */

   /* ('sne', ('seq', 'a', 'b'), 0.0) => ('seq', 'a', 'b') */
   /* search789_0_0 -> 0 in the cache */
   /* search789_0_1 -> 3 in the cache */
   /* search789_0 -> 2332 in the cache */
   /* search789_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2332, 245 },
      -1,
   } },

   /* replace789_0 -> 0 in the cache */
   /* replace789_1 -> 3 in the cache */
   /* replace789 -> 2324 in the cache */

   /* ('sne', ('sne', 'a', 'b'), 0.0) => ('sne', 'a', 'b') */
   /* search790_0_0 -> 0 in the cache */
   /* search790_0_1 -> 3 in the cache */
   /* search790_0 -> 2334 in the cache */
   /* search790_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2334, 245 },
      -1,
   } },

   /* replace790_0 -> 0 in the cache */
   /* replace790_1 -> 3 in the cache */
   /* replace790 -> 2326 in the cache */

   /* ('sne', ('slt', 'a', 'b'), 0.0) => ('slt', 'a', 'b') */
   /* search791_0_0 -> 0 in the cache */
   /* search791_0_1 -> 3 in the cache */
   /* search791_0 -> 2320 in the cache */
   /* search791_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2320, 245 },
      -1,
   } },

   /* replace791_0 -> 0 in the cache */
   /* replace791_1 -> 3 in the cache */
   /* replace791 -> 2320 in the cache */

   /* ('sne', ('sge', 'a', 'b'), 0.0) => ('sge', 'a', 'b') */
   /* search792_0_0 -> 0 in the cache */
   /* search792_0_1 -> 3 in the cache */
   /* search792_0 -> 2322 in the cache */
   /* search792_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2322, 245 },
      -1,
   } },

   /* replace792_0 -> 0 in the cache */
   /* replace792_1 -> 3 in the cache */
   /* replace792 -> 2322 in the cache */

   /* ('seq', ('seq', 'a', 'b'), 0.0) => ('sne', 'a', 'b') */
   /* search793_0_0 -> 0 in the cache */
   /* search793_0_1 -> 3 in the cache */
   /* search793_0 -> 2332 in the cache */
   /* search793_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2332, 245 },
      -1,
   } },

   /* replace793_0 -> 0 in the cache */
   /* replace793_1 -> 3 in the cache */
   /* replace793 -> 2326 in the cache */

   /* ('seq', ('sne', 'a', 'b'), 0.0) => ('seq', 'a', 'b') */
   /* search794_0_0 -> 0 in the cache */
   /* search794_0_1 -> 3 in the cache */
   /* search794_0 -> 2334 in the cache */
   /* search794_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2334, 245 },
      -1,
   } },

   /* replace794_0 -> 0 in the cache */
   /* replace794_1 -> 3 in the cache */
   /* replace794 -> 2324 in the cache */

   /* ('seq', ('slt', 'a', 'b'), 0.0) => ('sge', 'a', 'b') */
   /* search795_0_0 -> 0 in the cache */
   /* search795_0_1 -> 3 in the cache */
   /* search795_0 -> 2320 in the cache */
   /* search795_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2320, 245 },
      -1,
   } },

   /* replace795_0 -> 0 in the cache */
   /* replace795_1 -> 3 in the cache */
   /* replace795 -> 2322 in the cache */

   /* ('seq', ('sge', 'a', 'b'), 0.0) => ('slt', 'a', 'b') */
   /* search796_0_0 -> 0 in the cache */
   /* search796_0_1 -> 3 in the cache */
   /* search796_0 -> 2322 in the cache */
   /* search796_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2322, 245 },
      -1,
   } },

   /* replace796_0 -> 0 in the cache */
   /* replace796_1 -> 3 in the cache */
   /* replace796 -> 2320 in the cache */

   /* ('sne', ('seq', 'a', 'b'), 1.0) => ('sne', 'a', 'b') */
   /* search797_0_0 -> 0 in the cache */
   /* search797_0_1 -> 3 in the cache */
   /* search797_0 -> 2332 in the cache */
   /* search797_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2332, 259 },
      -1,
   } },

   /* replace797_0 -> 0 in the cache */
   /* replace797_1 -> 3 in the cache */
   /* replace797 -> 2326 in the cache */

   /* ('sne', ('sne', 'a', 'b'), 1.0) => ('seq', 'a', 'b') */
   /* search798_0_0 -> 0 in the cache */
   /* search798_0_1 -> 3 in the cache */
   /* search798_0 -> 2334 in the cache */
   /* search798_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 2,
      { 2334, 259 },
      -1,
   } },

   /* replace798_0 -> 0 in the cache */
   /* replace798_1 -> 3 in the cache */
   /* replace798 -> 2324 in the cache */

   /* ('sne', ('slt', 'a', 'b'), 1.0) => ('sge', 'a', 'b') */
   /* search799_0_0 -> 0 in the cache */
   /* search799_0_1 -> 3 in the cache */
   /* search799_0 -> 2320 in the cache */
   /* search799_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2320, 259 },
      -1,
   } },

   /* replace799_0 -> 0 in the cache */
   /* replace799_1 -> 3 in the cache */
   /* replace799 -> 2322 in the cache */

   /* ('sne', ('sge', 'a', 'b'), 1.0) => ('slt', 'a', 'b') */
   /* search800_0_0 -> 0 in the cache */
   /* search800_0_1 -> 3 in the cache */
   /* search800_0 -> 2322 in the cache */
   /* search800_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2322, 259 },
      -1,
   } },

   /* replace800_0 -> 0 in the cache */
   /* replace800_1 -> 3 in the cache */
   /* replace800 -> 2320 in the cache */

   /* ('fall_equal2', 'a', 'b') => ('fmin', ('seq', 'a.x', 'b.x'), ('seq', 'a.y', 'b.y')) */
   /* search801_0 -> 16 in the cache */
   /* search801_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal2,
      0, 1,
      { 16, 36 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 1,
      { 2351, 2352 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      1, 1,
      { 2354, 2355 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      -1, 2,
      { 2353, 2356 },
      -1,
   } },

   /* ('fall_equal3', 'a', 'b') => ('seq', ('fany_nequal3', 'a', 'b'), 0.0) */
   /* search802_0 -> 16 in the cache */
   /* search802_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal3,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace802_0_0 -> 16 in the cache */
   /* replace802_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal3,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace802_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2359, 117 },
      -1,
   } },

   /* ('fall_equal4', 'a', 'b') => ('seq', ('fany_nequal4', 'a', 'b'), 0.0) */
   /* search803_0 -> 16 in the cache */
   /* search803_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fall_equal4,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace803_0_0 -> 16 in the cache */
   /* replace803_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal4,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace803_1 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_seq,
      0, 2,
      { 2362, 117 },
      -1,
   } },

   /* ('fany_nequal2', 'a', 'b') => ('fmax', ('sne', 'a.x', 'b.x'), ('sne', 'a.y', 'b.y')) */
   /* search804_0 -> 16 in the cache */
   /* search804_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal2,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace804_0_0 -> 2351 in the cache */
   /* replace804_0_1 -> 2352 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 2351, 2352 },
      -1,
   } },
   /* replace804_1_0 -> 2354 in the cache */
   /* replace804_1_1 -> 2355 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 2354, 2355 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      -1, 2,
      { 2365, 2366 },
      -1,
   } },

   /* ('fany_nequal3', 'a', 'b') => ('fsat', ('fdot3', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search805_0 -> 16 in the cache */
   /* search805_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal3,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace805_0_0_0 -> 16 in the cache */
   /* replace805_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      0, 1,
      { 16, 36 },
      -1,
   } },
   /* replace805_0_1_0 -> 16 in the cache */
   /* replace805_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sne,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot3,
      -1, 2,
      { 2369, 2370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2371 },
      -1,
   } },

   /* ('fany_nequal4', 'a', 'b') => ('fsat', ('fdot4', ('sne', 'a', 'b'), ('sne', 'a', 'b'))) */
   /* search806_0 -> 16 in the cache */
   /* search806_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fany_nequal4,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace806_0_0_0 -> 16 in the cache */
   /* replace806_0_0_1 -> 36 in the cache */
   /* replace806_0_0 -> 2369 in the cache */
   /* replace806_0_1_0 -> 16 in the cache */
   /* replace806_0_1_1 -> 36 in the cache */
   /* replace806_0_1 -> 2370 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdot4,
      -1, 2,
      { 2369, 2370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 2,
      { 2374 },
      -1,
   } },

   /* ('ball_iequal2', 'a', 'b') => ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')) */
   /* search807_0 -> 0 in the cache */
   /* search807_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace807_0_0 -> 731 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace807_1_0 -> 732 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {1},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2378, 2380 },
      -1,
   } },

   /* ('ball_iequal3', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('ieq', 'a.z', 'b.z')) */
   /* search808_0 -> 0 in the cache */
   /* search808_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace808_0_0_0 -> 731 in the cache */
   /* replace808_0_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace808_0_1_0 -> 732 in the cache */
   /* replace808_0_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2383, 2384 },
      -1,
   } },
   /* replace808_1_0 -> 733 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {2},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 733, 2386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2385, 2387 },
      -1,
   } },

   /* ('ball_iequal4', 'a', 'b') => ('iand', ('iand', ('ieq', 'a.x', 'b.x'), ('ieq', 'a.y', 'b.y')), ('iand', ('ieq', 'a.z', 'b.z'), ('ieq', 'a.w', 'b.w'))) */
   /* search809_0 -> 0 in the cache */
   /* search809_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace809_0_0_0 -> 731 in the cache */
   /* replace809_0_0_1 -> 2377 in the cache */
   /* replace809_0_0 -> 2378 in the cache */
   /* replace809_0_1_0 -> 732 in the cache */
   /* replace809_0_1_1 -> 2379 in the cache */
   /* replace809_0_1 -> 2380 in the cache */
   /* replace809_0 -> 2381 in the cache */
   /* replace809_1_0_0 -> 733 in the cache */
   /* replace809_1_0_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 733, 2386 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {3},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      3, 1,
      { 2391, 2392 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2390, 2393 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2381, 2394 },
      -1,
   } },

   /* ('bany_inequal2', 'a', 'b') => ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')) */
   /* search810_0 -> 0 in the cache */
   /* search810_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace810_0_0 -> 731 in the cache */
   /* replace810_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace810_1_0 -> 732 in the cache */
   /* replace810_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2397, 2398 },
      -1,
   } },

   /* ('bany_inequal3', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ine', 'a.z', 'b.z')) */
   /* search811_0 -> 0 in the cache */
   /* search811_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace811_0_0_0 -> 731 in the cache */
   /* replace811_0_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace811_0_1_0 -> 732 in the cache */
   /* replace811_0_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2401, 2402 },
      -1,
   } },
   /* replace811_1_0 -> 733 in the cache */
   /* replace811_1_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      3, 1,
      { 733, 2386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 4,
      { 2403, 2404 },
      -1,
   } },

   /* ('bany_inequal4', 'a', 'b') => ('ior', ('ior', ('ine', 'a.x', 'b.x'), ('ine', 'a.y', 'b.y')), ('ior', ('ine', 'a.z', 'b.z'), ('ine', 'a.w', 'b.w'))) */
   /* search812_0 -> 0 in the cache */
   /* search812_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace812_0_0_0 -> 731 in the cache */
   /* replace812_0_0_1 -> 2377 in the cache */
   /* replace812_0_0 -> 2397 in the cache */
   /* replace812_0_1_0 -> 732 in the cache */
   /* replace812_0_1_1 -> 2379 in the cache */
   /* replace812_0_1 -> 2398 in the cache */
   /* replace812_0 -> 2399 in the cache */
   /* replace812_1_0_0 -> 733 in the cache */
   /* replace812_1_0_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 733, 2386 },
      -1,
   } },
   /* replace812_1_1_0 -> 2391 in the cache */
   /* replace812_1_1_1 -> 2392 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      3, 1,
      { 2391, 2392 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2407, 2408 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2399, 2409 },
      -1,
   } },

   /* ('ball_fequal2', 'a', 'b') => ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')) */
   /* search813_0 -> 0 in the cache */
   /* search813_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace813_0_0 -> 731 in the cache */
   /* replace813_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace813_1_0 -> 732 in the cache */
   /* replace813_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2412, 2413 },
      -1,
   } },

   /* ('ball_fequal3', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('feq', 'a.z', 'b.z')) */
   /* search814_0 -> 0 in the cache */
   /* search814_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace814_0_0_0 -> 731 in the cache */
   /* replace814_0_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace814_0_1_0 -> 732 in the cache */
   /* replace814_0_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2416, 2417 },
      -1,
   } },
   /* replace814_1_0 -> 733 in the cache */
   /* replace814_1_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      3, 1,
      { 733, 2386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 4,
      { 2418, 2419 },
      -1,
   } },

   /* ('ball_fequal4', 'a', 'b') => ('iand', ('iand', ('feq', 'a.x', 'b.x'), ('feq', 'a.y', 'b.y')), ('iand', ('feq', 'a.z', 'b.z'), ('feq', 'a.w', 'b.w'))) */
   /* search815_0 -> 0 in the cache */
   /* search815_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace815_0_0_0 -> 731 in the cache */
   /* replace815_0_0_1 -> 2377 in the cache */
   /* replace815_0_0 -> 2412 in the cache */
   /* replace815_0_1_0 -> 732 in the cache */
   /* replace815_0_1_1 -> 2379 in the cache */
   /* replace815_0_1 -> 2413 in the cache */
   /* replace815_0 -> 2414 in the cache */
   /* replace815_1_0_0 -> 733 in the cache */
   /* replace815_1_0_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 733, 2386 },
      -1,
   } },
   /* replace815_1_1_0 -> 2391 in the cache */
   /* replace815_1_1_1 -> 2392 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      3, 1,
      { 2391, 2392 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 2,
      { 2422, 2423 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 4,
      { 2414, 2424 },
      -1,
   } },

   /* ('bany_fnequal2', 'a', 'b') => ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')) */
   /* search816_0 -> 0 in the cache */
   /* search816_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal2,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace816_0_0 -> 731 in the cache */
   /* replace816_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace816_1_0 -> 732 in the cache */
   /* replace816_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2427, 2428 },
      -1,
   } },

   /* ('bany_fnequal3', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('fneu', 'a.z', 'b.z')) */
   /* search817_0 -> 0 in the cache */
   /* search817_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal3,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace817_0_0_0 -> 731 in the cache */
   /* replace817_0_0_1 -> 2377 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 731, 2377 },
      -1,
   } },
   /* replace817_0_1_0 -> 732 in the cache */
   /* replace817_0_1_1 -> 2379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 732, 2379 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2431, 2432 },
      -1,
   } },
   /* replace817_1_0 -> 733 in the cache */
   /* replace817_1_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      3, 1,
      { 733, 2386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 4,
      { 2433, 2434 },
      -1,
   } },

   /* ('bany_fnequal4', 'a', 'b') => ('ior', ('ior', ('fneu', 'a.x', 'b.x'), ('fneu', 'a.y', 'b.y')), ('ior', ('fneu', 'a.z', 'b.z'), ('fneu', 'a.w', 'b.w'))) */
   /* search818_0 -> 0 in the cache */
   /* search818_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal4,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace818_0_0_0 -> 731 in the cache */
   /* replace818_0_0_1 -> 2377 in the cache */
   /* replace818_0_0 -> 2427 in the cache */
   /* replace818_0_1_0 -> 732 in the cache */
   /* replace818_0_1_1 -> 2379 in the cache */
   /* replace818_0_1 -> 2428 in the cache */
   /* replace818_0 -> 2429 in the cache */
   /* replace818_1_0_0 -> 733 in the cache */
   /* replace818_1_0_1 -> 2386 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 733, 2386 },
      -1,
   } },
   /* replace818_1_1_0 -> 2391 in the cache */
   /* replace818_1_1_1 -> 2392 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      3, 1,
      { 2391, 2392 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 2,
      { 2437, 2438 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 4,
      { 2429, 2439 },
      -1,
   } },

   /* ('feq', ('seq', 'a', 'b'), 1.0) => ('feq', 'a', 'b') */
   /* search819_0_0 -> 0 in the cache */
   /* search819_0_1 -> 3 in the cache */
   /* search819_0 -> 2332 in the cache */
   /* search819_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2332, 259 },
      -1,
   } },

   /* replace819_0 -> 0 in the cache */
   /* replace819_1 -> 3 in the cache */
   /* replace819 -> 1017 in the cache */

   /* ('feq', ('sne', 'a', 'b'), 1.0) => ('fneu', 'a', 'b') */
   /* search820_0_0 -> 0 in the cache */
   /* search820_0_1 -> 3 in the cache */
   /* search820_0 -> 2334 in the cache */
   /* search820_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2334, 259 },
      -1,
   } },

   /* replace820_0 -> 0 in the cache */
   /* replace820_1 -> 3 in the cache */
   /* replace820 -> 1014 in the cache */

   /* ('feq', ('slt', 'a', 'b'), 1.0) => ('flt', 'a', 'b') */
   /* search821_0_0 -> 0 in the cache */
   /* search821_0_1 -> 3 in the cache */
   /* search821_0 -> 2320 in the cache */
   /* search821_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2320, 259 },
      -1,
   } },

   /* replace821_0 -> 0 in the cache */
   /* replace821_1 -> 3 in the cache */
   /* replace821 -> 1011 in the cache */

   /* ('feq', ('sge', 'a', 'b'), 1.0) => ('fge', 'a', 'b') */
   /* search822_0_0 -> 0 in the cache */
   /* search822_0_1 -> 3 in the cache */
   /* search822_0 -> 2322 in the cache */
   /* search822_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2322, 259 },
      -1,
   } },

   /* replace822_0 -> 0 in the cache */
   /* replace822_1 -> 3 in the cache */
   /* replace822 -> 1008 in the cache */

   /* ('fneu', ('seq', 'a', 'b'), 0.0) => ('feq', 'a', 'b') */
   /* search823_0_0 -> 0 in the cache */
   /* search823_0_1 -> 3 in the cache */
   /* search823_0 -> 2332 in the cache */
   /* search823_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2332, 245 },
      -1,
   } },

   /* replace823_0 -> 0 in the cache */
   /* replace823_1 -> 3 in the cache */
   /* replace823 -> 1017 in the cache */

   /* ('fneu', ('sne', 'a', 'b'), 0.0) => ('fneu', 'a', 'b') */
   /* search824_0_0 -> 0 in the cache */
   /* search824_0_1 -> 3 in the cache */
   /* search824_0 -> 2334 in the cache */
   /* search824_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2334, 245 },
      -1,
   } },

   /* replace824_0 -> 0 in the cache */
   /* replace824_1 -> 3 in the cache */
   /* replace824 -> 1014 in the cache */

   /* ('fneu', ('slt', 'a', 'b'), 0.0) => ('flt', 'a', 'b') */
   /* search825_0_0 -> 0 in the cache */
   /* search825_0_1 -> 3 in the cache */
   /* search825_0 -> 2320 in the cache */
   /* search825_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2320, 245 },
      -1,
   } },

   /* replace825_0 -> 0 in the cache */
   /* replace825_1 -> 3 in the cache */
   /* replace825 -> 1011 in the cache */

   /* ('fneu', ('sge', 'a', 'b'), 0.0) => ('fge', 'a', 'b') */
   /* search826_0_0 -> 0 in the cache */
   /* search826_0_1 -> 3 in the cache */
   /* search826_0 -> 2322 in the cache */
   /* search826_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2322, 245 },
      -1,
   } },

   /* replace826_0 -> 0 in the cache */
   /* replace826_1 -> 3 in the cache */
   /* replace826 -> 1008 in the cache */

   /* ('feq', ('seq', 'a', 'b'), 0.0) => ('fneu', 'a', 'b') */
   /* search827_0_0 -> 0 in the cache */
   /* search827_0_1 -> 3 in the cache */
   /* search827_0 -> 2332 in the cache */
   /* search827_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2332, 245 },
      -1,
   } },

   /* replace827_0 -> 0 in the cache */
   /* replace827_1 -> 3 in the cache */
   /* replace827 -> 1014 in the cache */

   /* ('feq', ('sne', 'a', 'b'), 0.0) => ('feq', 'a', 'b') */
   /* search828_0_0 -> 0 in the cache */
   /* search828_0_1 -> 3 in the cache */
   /* search828_0 -> 2334 in the cache */
   /* search828_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 2334, 245 },
      -1,
   } },

   /* replace828_0 -> 0 in the cache */
   /* replace828_1 -> 3 in the cache */
   /* replace828 -> 1017 in the cache */

   /* ('feq', ('slt', 'a', 'b'), 0.0) => ('fge', 'a', 'b') */
   /* search829_0_0 -> 0 in the cache */
   /* search829_0_1 -> 3 in the cache */
   /* search829_0 -> 2320 in the cache */
   /* search829_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2320, 245 },
      -1,
   } },

   /* replace829_0 -> 0 in the cache */
   /* replace829_1 -> 3 in the cache */
   /* replace829 -> 1008 in the cache */

   /* ('feq', ('sge', 'a', 'b'), 0.0) => ('flt', 'a', 'b') */
   /* search830_0_0 -> 0 in the cache */
   /* search830_0_1 -> 3 in the cache */
   /* search830_0 -> 2322 in the cache */
   /* search830_1 -> 245 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2322, 245 },
      -1,
   } },

   /* replace830_0 -> 0 in the cache */
   /* replace830_1 -> 3 in the cache */
   /* replace830 -> 1011 in the cache */

   /* ('fneu', ('seq', 'a', 'b'), 1.0) => ('fneu', 'a', 'b') */
   /* search831_0_0 -> 0 in the cache */
   /* search831_0_1 -> 3 in the cache */
   /* search831_0 -> 2332 in the cache */
   /* search831_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2332, 259 },
      -1,
   } },

   /* replace831_0 -> 0 in the cache */
   /* replace831_1 -> 3 in the cache */
   /* replace831 -> 1014 in the cache */

   /* ('fneu', ('sne', 'a', 'b'), 1.0) => ('feq', 'a', 'b') */
   /* search832_0_0 -> 0 in the cache */
   /* search832_0_1 -> 3 in the cache */
   /* search832_0 -> 2334 in the cache */
   /* search832_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 2334, 259 },
      -1,
   } },

   /* replace832_0 -> 0 in the cache */
   /* replace832_1 -> 3 in the cache */
   /* replace832 -> 1017 in the cache */

   /* ('fneu', ('slt', 'a', 'b'), 1.0) => ('fge', 'a', 'b') */
   /* search833_0_0 -> 0 in the cache */
   /* search833_0_1 -> 3 in the cache */
   /* search833_0 -> 2320 in the cache */
   /* search833_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2320, 259 },
      -1,
   } },

   /* replace833_0 -> 0 in the cache */
   /* replace833_1 -> 3 in the cache */
   /* replace833 -> 1008 in the cache */

   /* ('fneu', ('sge', 'a', 'b'), 1.0) => ('flt', 'a', 'b') */
   /* search834_0_0 -> 0 in the cache */
   /* search834_0_1 -> 3 in the cache */
   /* search834_0 -> 2322 in the cache */
   /* search834_1 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2322, 259 },
      -1,
   } },

   /* replace834_0 -> 0 in the cache */
   /* replace834_1 -> 3 in the cache */
   /* replace834 -> 1011 in the cache */

   /* ('fneu', ('fneg', 'a'), 'a') => ('fneu', 'a', 0.0) */
   /* search835_0_0 -> 32 in the cache */
   /* search835_0 -> 98 in the cache */
   /* search835_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 98, 32 },
      -1,
   } },

   /* replace835_0 -> 32 in the cache */
   /* replace835_1 -> 113 in the cache */
   /* replace835 -> 1210 in the cache */

   /* ('feq', ('fneg', 'a'), 'a') => ('feq', 'a', 0.0) */
   /* search836_0_0 -> 32 in the cache */
   /* search836_0 -> 98 in the cache */
   /* search836_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 98, 32 },
      -1,
   } },

   /* replace836_0 -> 32 in the cache */
   /* replace836_1 -> 113 in the cache */
   /* replace836 -> 1214 in the cache */

   /* ('imul', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search837_0_0 -> 722 in the cache */
   /* search837_0 -> 1183 in the cache */
   /* search837_1_0 -> 1090 in the cache */
   /* search837_1 -> 2199 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 1183, 2199 },
      -1,
   } },

   /* replace837_0_0 -> 722 in the cache */
   /* replace837_0_1 -> 1090 in the cache */
   /* replace837_0 -> 1102 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1102 },
      -1,
   } },

   /* ('iand', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('iand', 'a', 'b')) */
   /* search838_0_0 -> 722 in the cache */
   /* search838_0 -> 1183 in the cache */
   /* search838_1_0 -> 1090 in the cache */
   /* search838_1 -> 2199 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1183, 2199 },
      -1,
   } },

   /* replace838_0_0 -> 722 in the cache */
   /* replace838_0_1 -> 1090 in the cache */
   /* replace838_0 -> 1102 in the cache */
   /* replace838 -> 2460 in the cache */

   /* ('ior', ('b2i', 'a@1'), ('b2i', 'b@1')) => ('b2i', ('ior', 'a', 'b')) */
   /* search839_0_0 -> 722 in the cache */
   /* search839_0 -> 1183 in the cache */
   /* search839_1_0 -> 1090 in the cache */
   /* search839_1 -> 2199 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1183, 2199 },
      -1,
   } },

   /* replace839_0_0 -> 722 in the cache */
   /* replace839_0_1 -> 1090 in the cache */
   /* replace839_0 -> 1094 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 1094 },
      -1,
   } },

   /* ('fmul', ('b2f', 'a@1'), ('b2f', 'b@1')) => ('b2f', ('iand', 'a', 'b')) */
   /* search840_0_0 -> 722 in the cache */
   /* search840_0 -> 1085 in the cache */
   /* search840_1_0 -> 1090 in the cache */
   /* search840_1 -> 1091 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1085, 1091 },
      -1,
   } },

   /* replace840_0_0 -> 722 in the cache */
   /* replace840_0_1 -> 1090 in the cache */
   /* replace840_0 -> 1102 in the cache */
   /* replace840 -> 1223 in the cache */

   /* ('fsat', ('fadd', ('b2f', 'a@1'), ('b2f', 'b@1'))) => ('b2f', ('ior', 'a', 'b')) */
   /* search841_0_0_0 -> 722 in the cache */
   /* search841_0_0 -> 1085 in the cache */
   /* search841_0_1_0 -> 1090 in the cache */
   /* search841_0_1 -> 1091 in the cache */
   /* search841_0 -> 1121 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 1121 },
      -1,
   } },

   /* replace841_0_0 -> 722 in the cache */
   /* replace841_0_1 -> 1090 in the cache */
   /* replace841_0 -> 1094 in the cache */
   /* replace841 -> 1220 in the cache */

   /* ('iand', 'a@bool16', 1.0) => ('b2f', 'a') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search842_1 -> 466 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2466, 466 },
      -1,
   } },

   /* replace842_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('iand', 'a@bool32', 1.0) => ('b2f', 'a') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search843_1 -> 232 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2469, 232 },
      -1,
   } },

   /* replace843_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('flt', ('fneg', ('b2f', 'a@1')), 0) => a */
   /* search844_0_0_0 -> 722 in the cache */
   /* search844_0_0 -> 1085 in the cache */
   /* search844_0 -> 1088 in the cache */
   /* search844_1 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1088, 1184 },
      -1,
   } },

   /* replace844 -> 722 in the cache */

   /* ('ilt', 'a', 'a') => False */
   /* search845_0 -> 32 in the cache */
   /* search845_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace845 -> 1039 in the cache */

   /* ('ige', 'a', 'a') => True */
   /* search846_0 -> 32 in the cache */
   /* search846_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace846 -> 1192 in the cache */

   /* ('ieq', 'a', 'a') => True */
   /* search847_0 -> 32 in the cache */
   /* search847_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace847 -> 1192 in the cache */

   /* ('ine', 'a', 'a') => False */
   /* search848_0 -> 32 in the cache */
   /* search848_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace848 -> 1039 in the cache */

   /* ('ult', 'a', 'a') => False */
   /* search849_0 -> 32 in the cache */
   /* search849_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace849 -> 1039 in the cache */

   /* ('uge', 'a', 'a') => True */
   /* search850_0 -> 32 in the cache */
   /* search850_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace850 -> 1192 in the cache */

   /* ('flt', 'a', 'a') => False */
   /* search851_0 -> 32 in the cache */
   /* search851_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace851 -> 1039 in the cache */

   /* ('fge', 'a(is_a_number)', 'a') => True */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      16,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search852_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2480, 32 },
      -1,
   } },

   /* replace852 -> 1192 in the cache */

   /* ('feq', 'a(is_a_number)', 'a') => True */
   /* search853_0 -> 2480 in the cache */
   /* search853_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 2480, 32 },
      -1,
   } },

   /* replace853 -> 1192 in the cache */

   /* ('fneu', 'a(is_a_number)', 'a') => False */
   /* search854_0 -> 2480 in the cache */
   /* search854_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 2480, 32 },
      -1,
   } },

   /* replace854 -> 1039 in the cache */

   /* ('iand', 'a', 'a') => a */
   /* search855_0 -> 32 in the cache */
   /* search855_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace855 -> 32 in the cache */

   /* ('iand', 'a', -1) => a */
   /* search856_0 -> 32 in the cache */
   /* search856_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace856 -> 32 in the cache */

   /* ('iand', 'a', 0) => 0 */
   /* search857_0 -> 32 in the cache */
   /* search857_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace857 -> 61 in the cache */

   /* ('ior', 'a', 'a') => a */
   /* search858_0 -> 32 in the cache */
   /* search858_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace858 -> 32 in the cache */

   /* ('ior', 'a', 0) => a */
   /* search859_0 -> 32 in the cache */
   /* search859_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace859 -> 32 in the cache */

   /* ('ior', 'a', True) => True */
   /* search860_0 -> 722 in the cache */
   /* search860_1 -> 1192 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 722, 1192 },
      -1,
   } },

   /* replace860 -> 1192 in the cache */

   /* ('ixor', 'a', 'a') => 0 */
   /* search861_0 -> 32 in the cache */
   /* search861_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      -1, 0,
      { 32, 32 },
      -1,
   } },

   /* replace861 -> 61 in the cache */

   /* ('ixor', 'a', 0) => a */
   /* search862_0 -> 32 in the cache */
   /* search862_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* replace862 -> 32 in the cache */

   /* ('ixor', 'a', ('ixor', 'a', 'b')) => b */
   /* search863_0 -> 0 in the cache */
   /* search863_1_0 -> 0 in the cache */
   /* search863_1_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 0, 2492 },
      -1,
   } },

   /* replace863 -> 3 in the cache */

   /* ('ixor', 'a', -1) => ('inot', 'a') */
   /* search864_0 -> 32 in the cache */
   /* search864_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 32, 63 },
      -1,
   } },

   /* replace864_0 -> 32 in the cache */
   /* replace864 -> 2196 in the cache */

   /* ('inot', ('inot', 'a')) => a */
   /* search865_0_0 -> 32 in the cache */
   /* search865_0 -> 2196 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 2196 },
      -1,
   } },

   /* replace865 -> 32 in the cache */

   /* ('ior', ('iand', 'a', 'b'), 'b') => b */
   /* search866_0_0 -> 0 in the cache */
   /* search866_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* search866_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2496, 3 },
      -1,
   } },

   /* replace866 -> 3 in the cache */

   /* ('ior', ('ior', 'a', 'b'), 'b') => ('ior', 'a', 'b') */
   /* search867_0_0 -> 0 in the cache */
   /* search867_0_1 -> 3 in the cache */
   /* search867_0 -> 84 in the cache */
   /* search867_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 84, 3 },
      -1,
   } },

   /* replace867_0 -> 0 in the cache */
   /* replace867_1 -> 3 in the cache */
   /* replace867 -> 2262 in the cache */

   /* ('iand', ('ior', 'a', 'b'), 'b') => b */
   /* search868_0_0 -> 0 in the cache */
   /* search868_0_1 -> 3 in the cache */
   /* search868_0 -> 84 in the cache */
   /* search868_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 84, 3 },
      -1,
   } },

   /* replace868 -> 3 in the cache */

   /* ('iand', ('iand', 'a', 'b'), 'b') => ('iand', 'a', 'b') */
   /* search869_0_0 -> 0 in the cache */
   /* search869_0_1 -> 3 in the cache */
   /* search869_0 -> 2496 in the cache */
   /* search869_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 2496, 3 },
      -1,
   } },

   /* replace869_0 -> 0 in the cache */
   /* replace869_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('inot', 'a'), ('inot', 'b')) => ('inot', ('ior', 'a', 'b')) */
   /* search870_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 0 },
      -1,
   } },
   /* search870_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2502, 2503 },
      -1,
   } },

   /* replace870_0_0 -> 0 in the cache */
   /* replace870_0_1 -> 3 in the cache */
   /* replace870_0 -> 2262 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2262 },
      -1,
   } },

   /* ('ior', ('inot', 'a'), ('inot', 'b')) => ('inot', ('iand', 'a', 'b')) */
   /* search871_0_0 -> 0 in the cache */
   /* search871_0 -> 2502 in the cache */
   /* search871_1_0 -> 3 in the cache */
   /* search871_1 -> 2503 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2502, 2503 },
      -1,
   } },

   /* replace871_0_0 -> 0 in the cache */
   /* replace871_0_1 -> 3 in the cache */
   /* replace871_0 -> 2501 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 2501 },
      -1,
   } },

   /* ('ishl', 0, 'a') => 0 */
   /* search872_0 -> 1184 in the cache */
   /* search872_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 1184, 16 },
      -1,
   } },

   /* replace872 -> 1184 in the cache */

   /* ('ishl', 'a', 0) => a */
   /* search873_0 -> 32 in the cache */
   /* search873_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 32, 124 },
      -1,
   } },

   /* replace873 -> 32 in the cache */

   /* ('ishr', 0, 'a') => 0 */
   /* search874_0 -> 1184 in the cache */
   /* search874_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 1184, 16 },
      -1,
   } },

   /* replace874 -> 1184 in the cache */

   /* ('ishr', -1, 'a') => -1 */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   /* search875_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 2511, 16 },
      -1,
   } },

   /* replace875 -> 2511 in the cache */

   /* ('ishr', 'a', 0) => a */
   /* search876_0 -> 32 in the cache */
   /* search876_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 32, 124 },
      -1,
   } },

   /* replace876 -> 32 in the cache */

   /* ('ushr', 0, 'a') => 0 */
   /* search877_0 -> 1184 in the cache */
   /* search877_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 1184, 16 },
      -1,
   } },

   /* replace877 -> 1184 in the cache */

   /* ('ushr', 'a', 0) => a */
   /* search878_0 -> 32 in the cache */
   /* search878_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 124 },
      -1,
   } },

   /* replace878 -> 32 in the cache */

   /* ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('urol', 'a', 'b') */
   /* search879_0_0 -> 11 in the cache */
   /* search879_0_1 -> 36 in the cache */
   /* search879_0 -> 2112 in the cache */
   /* search879_1_0 -> 11 in the cache */
   /* search879_1_1_0 -> 815 in the cache */
   /* search879_1_1_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 815, 2516 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 11, 2517 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2112, 2518 },
      -1,
   } },

   /* replace879_0 -> 11 in the cache */
   /* replace879_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ior', ('ishl@16', 'a', 'b'), ('ushr@16', 'a', ('isub', 16, 'b'))) => ('urol', 'a', 'b') */
   /* search880_0_0 -> 11 in the cache */
   /* search880_0_1 -> 36 in the cache */
   /* search880_0 -> 2112 in the cache */
   /* search880_1_0 -> 11 in the cache */
   /* search880_1_1_0 -> 815 in the cache */
   /* search880_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 815, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 2521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2112, 2522 },
      -1,
   } },

   /* replace880_0 -> 11 in the cache */
   /* replace880_1 -> 36 in the cache */
   /* replace880 -> 2520 in the cache */

   /* ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('urol', 'a', 'b') */
   /* search881_0_0 -> 16 in the cache */
   /* search881_0_1 -> 36 in the cache */
   /* search881_0 -> 908 in the cache */
   /* search881_1_0 -> 16 in the cache */
   /* search881_1_1_0 -> 836 in the cache */
   /* search881_1_1_1_0 -> 36 in the cache */
   /* search881_1_1_1 -> 2516 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 836, 2516 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 2524 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 908, 2525 },
      -1,
   } },

   /* replace881_0 -> 16 in the cache */
   /* replace881_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_urol,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('ishl@32', 'a', 'b'), ('ushr@32', 'a', ('isub', 32, 'b'))) => ('urol', 'a', 'b') */
   /* search882_0_0 -> 16 in the cache */
   /* search882_0_1 -> 36 in the cache */
   /* search882_0 -> 908 in the cache */
   /* search882_1_0 -> 16 in the cache */
   /* search882_1_1_0 -> 836 in the cache */
   /* search882_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 836, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 2528 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 908, 2529 },
      -1,
   } },

   /* replace882_0 -> 16 in the cache */
   /* replace882_1 -> 36 in the cache */
   /* replace882 -> 2527 in the cache */

   /* ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('iadd', 16, ('ineg', 'b')))) => ('uror', 'a', 'b') */
   /* search883_0_0 -> 11 in the cache */
   /* search883_0_1 -> 36 in the cache */
   /* search883_0 -> 2116 in the cache */
   /* search883_1_0 -> 11 in the cache */
   /* search883_1_1_0 -> 815 in the cache */
   /* search883_1_1_1_0 -> 36 in the cache */
   /* search883_1_1_1 -> 2516 in the cache */
   /* search883_1_1 -> 2517 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 11, 2517 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2116, 2531 },
      -1,
   } },

   /* replace883_0 -> 11 in the cache */
   /* replace883_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 11, 36 },
      -1,
   } },

   /* ('ior', ('ushr@16', 'a', 'b'), ('ishl@16', 'a', ('isub', 16, 'b'))) => ('uror', 'a', 'b') */
   /* search884_0_0 -> 11 in the cache */
   /* search884_0_1 -> 36 in the cache */
   /* search884_0 -> 2116 in the cache */
   /* search884_1_0 -> 11 in the cache */
   /* search884_1_1_0 -> 815 in the cache */
   /* search884_1_1_1 -> 36 in the cache */
   /* search884_1_1 -> 2521 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 2521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2116, 2534 },
      -1,
   } },

   /* replace884_0 -> 11 in the cache */
   /* replace884_1 -> 36 in the cache */
   /* replace884 -> 2533 in the cache */

   /* ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('iadd', 32, ('ineg', 'b')))) => ('uror', 'a', 'b') */
   /* search885_0_0 -> 16 in the cache */
   /* search885_0_1 -> 36 in the cache */
   /* search885_0 -> 2145 in the cache */
   /* search885_1_0 -> 16 in the cache */
   /* search885_1_1_0 -> 836 in the cache */
   /* search885_1_1_1_0 -> 36 in the cache */
   /* search885_1_1_1 -> 2516 in the cache */
   /* search885_1_1 -> 2524 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 2524 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 2145, 2536 },
      -1,
   } },

   /* replace885_0 -> 16 in the cache */
   /* replace885_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_uror,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('ushr@32', 'a', 'b'), ('ishl@32', 'a', ('isub', 32, 'b'))) => ('uror', 'a', 'b') */
   /* search886_0_0 -> 16 in the cache */
   /* search886_0_1 -> 36 in the cache */
   /* search886_0 -> 2145 in the cache */
   /* search886_1_0 -> 16 in the cache */
   /* search886_1_1_0 -> 836 in the cache */
   /* search886_1_1_1 -> 36 in the cache */
   /* search886_1_1 -> 2528 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 2528 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2145, 2539 },
      -1,
   } },

   /* replace886_0 -> 16 in the cache */
   /* replace886_1 -> 36 in the cache */
   /* replace886 -> 2538 in the cache */

   /* ('urol@16', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 16, 'b'))) */
   /* search887_0 -> 11 in the cache */
   /* search887_1 -> 36 in the cache */
   /* search887 -> 2520 in the cache */

   /* replace887_0_0 -> 11 in the cache */
   /* replace887_0_1 -> 36 in the cache */
   /* replace887_0 -> 2112 in the cache */
   /* replace887_1_0 -> 11 in the cache */
   /* replace887_1_1_0 -> 815 in the cache */
   /* replace887_1_1_1 -> 36 in the cache */
   /* replace887_1_1 -> 2521 in the cache */
   /* replace887_1 -> 2522 in the cache */
   /* replace887 -> 2523 in the cache */

   /* ('urol@32', 'a', 'b') => ('ior', ('ishl', 'a', 'b'), ('ushr', 'a', ('isub', 32, 'b'))) */
   /* search888_0 -> 16 in the cache */
   /* search888_1 -> 36 in the cache */
   /* search888 -> 2527 in the cache */

   /* replace888_0_0 -> 16 in the cache */
   /* replace888_0_1 -> 36 in the cache */
   /* replace888_0 -> 908 in the cache */
   /* replace888_1_0 -> 16 in the cache */
   /* replace888_1_1_0 -> 836 in the cache */
   /* replace888_1_1_1 -> 36 in the cache */
   /* replace888_1_1 -> 2528 in the cache */
   /* replace888_1 -> 2529 in the cache */
   /* replace888 -> 2530 in the cache */

   /* ('uror@16', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 16, 'b'))) */
   /* search889_0 -> 11 in the cache */
   /* search889_1 -> 36 in the cache */
   /* search889 -> 2533 in the cache */

   /* replace889_0_0 -> 11 in the cache */
   /* replace889_0_1 -> 36 in the cache */
   /* replace889_0 -> 2116 in the cache */
   /* replace889_1_0 -> 11 in the cache */
   /* replace889_1_1_0 -> 815 in the cache */
   /* replace889_1_1_1 -> 36 in the cache */
   /* replace889_1_1 -> 2521 in the cache */
   /* replace889_1 -> 2534 in the cache */
   /* replace889 -> 2535 in the cache */

   /* ('uror@32', 'a', 'b') => ('ior', ('ushr', 'a', 'b'), ('ishl', 'a', ('isub', 32, 'b'))) */
   /* search890_0 -> 16 in the cache */
   /* search890_1 -> 36 in the cache */
   /* search890 -> 2538 in the cache */

   /* replace890_0_0 -> 16 in the cache */
   /* replace890_0_1 -> 36 in the cache */
   /* replace890_0 -> 2145 in the cache */
   /* replace890_1_0 -> 16 in the cache */
   /* replace890_1_1_0 -> 836 in the cache */
   /* replace890_1_1_1 -> 36 in the cache */
   /* replace890_1_1 -> 2528 in the cache */
   /* replace890_1 -> 2539 in the cache */
   /* replace890 -> 2540 in the cache */

   /* ('~fexp2', ('flog2', 'a')) => a */
   /* search891_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 2541 },
      -1,
   } },

   /* replace891 -> 32 in the cache */

   /* ('~flog2', ('fexp2', 'a')) => a */
   /* search892_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2543 },
      -1,
   } },

   /* replace892 -> 32 in the cache */

   /* ('fpow', 'a', 'b') => ('fexp2', ('fmul', ('flog2', 'a'), 'b')) */
   /* search893_0 -> 0 in the cache */
   /* search893_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace893_0_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace893_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2546, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2547 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 'b')) => ('fpow', 'a', 'b') */
   /* search894_0_0_0 -> 0 in the cache */
   /* search894_0_0 -> 2546 in the cache */
   /* search894_0_1 -> 3 in the cache */
   /* search894_0 -> 2547 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2547 },
      -1,
   } },

   /* replace894_0 -> 0 in the cache */
   /* replace894_1 -> 3 in the cache */
   /* replace894 -> 2545 in the cache */

   /* ('~fexp2', ('fadd', ('fmul', ('flog2', 'a'), 'b'), ('fmul', ('flog2', 'c'), 'd'))) => ('~fmul', ('fpow', 'a', 'b'), ('fpow', 'c', 'd')) */
   /* search895_0_0_0_0 -> 151 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 151 },
      -1,
   } },
   /* search895_0_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 2550, 152 },
      -1,
   } },
   /* search895_0_1_0_0 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 154 },
      -1,
   } },
   /* search895_0_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 2552, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 2551, 2553 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 3,
      { 2554 },
      -1,
   } },

   /* replace895_0_0 -> 151 in the cache */
   /* replace895_0_1 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 151, 152 },
      -1,
   } },
   /* replace895_1_0 -> 154 in the cache */
   /* replace895_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2556, 2557 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 0.5)) => ('fsqrt', 'a') */
   /* search896_0_0_0 -> 32 in the cache */
   /* search896_0_0 -> 2541 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe0000000000000 /* 0.5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2541, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2560 },
      -1,
   } },

   /* replace896_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 2.0)) => ('fmul', 'a', 'a') */
   /* search897_0_0_0 -> 32 in the cache */
   /* search897_0_0 -> 2541 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x4000000000000000 /* 2.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2541, 2563 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2564 },
      -1,
   } },

   /* replace897_0 -> 32 in the cache */
   /* replace897_1 -> 32 in the cache */
   /* replace897 -> 240 in the cache */

   /* ('~fexp2', ('fmul', ('flog2', 'a'), 4.0)) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a')) */
   /* search898_0_0_0 -> 32 in the cache */
   /* search898_0_0 -> 2541 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x4010000000000000 /* 4.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2541, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2567 },
      -1,
   } },

   /* replace898_0_0 -> 32 in the cache */
   /* replace898_0_1 -> 32 in the cache */
   /* replace898_0 -> 240 in the cache */
   /* replace898_1_0 -> 32 in the cache */
   /* replace898_1_1 -> 32 in the cache */
   /* replace898_1 -> 240 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 240, 240 },
      -1,
   } },

   /* ('~fpow', 'a', 1.0) => a */
   /* search899_0 -> 32 in the cache */
   /* search899_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 230 },
      -1,
   } },

   /* replace899 -> 32 in the cache */

   /* ('~fpow', 'a', 2.0) => ('fmul', 'a', 'a') */
   /* search900_0 -> 32 in the cache */
   /* search900_1 -> 2563 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 2563 },
      -1,
   } },

   /* replace900_0 -> 32 in the cache */
   /* replace900_1 -> 32 in the cache */
   /* replace900 -> 240 in the cache */

   /* ('~fpow', 'a', 4.0) => ('fmul', ('fmul', 'a', 'a'), ('fmul', 'a', 'a')) */
   /* search901_0 -> 32 in the cache */
   /* search901_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 2566 },
      -1,
   } },

   /* replace901_0_0 -> 32 in the cache */
   /* replace901_0_1 -> 32 in the cache */
   /* replace901_0 -> 240 in the cache */
   /* replace901_1_0 -> 32 in the cache */
   /* replace901_1_1 -> 32 in the cache */
   /* replace901_1 -> 240 in the cache */
   /* replace901 -> 2569 in the cache */

   /* ('~fpow', 2.0, 'a') => ('fexp2', 'a') */
   /* search902_0 -> 2563 in the cache */
   /* search902_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 2563, 32 },
      -1,
   } },

   /* replace902_0 -> 32 in the cache */
   /* replace902 -> 2543 in the cache */

   /* ('~fpow', ('fpow', 'a', 2.2), 0.454545) => a */
   /* search903_0_0 -> 32 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x400199999999999a /* 2.2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 32, 2574 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fdd1743e963dc48 /* 0.454545 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 2575, 2576 },
      -1,
   } },

   /* replace903 -> 32 in the cache */

   /* ('~fpow', ('fabs', ('fpow', 'a', 2.2)), 0.454545) => ('fabs', 'a') */
   /* search904_0_0_0 -> 32 in the cache */
   /* search904_0_0_1 -> 2574 in the cache */
   /* search904_0_0 -> 2575 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2575 },
      -1,
   } },
   /* search904_1 -> 2576 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 2578, 2576 },
      -1,
   } },

   /* replace904_0 -> 32 in the cache */
   /* replace904 -> 103 in the cache */

   /* ('~fsqrt', ('fexp2', 'a')) => ('fexp2', ('fmul', 0.5, 'a')) */
   /* search905_0_0 -> 32 in the cache */
   /* search905_0 -> 2543 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 2543 },
      -1,
   } },

   /* replace905_0_0 -> 2559 in the cache */
   /* replace905_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2559, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2581 },
      -1,
   } },

   /* ('~frcp', ('fexp2', 'a')) => ('fexp2', ('fneg', 'a')) */
   /* search906_0_0 -> 32 in the cache */
   /* search906_0 -> 2543 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2543 },
      -1,
   } },

   /* replace906_0_0 -> 32 in the cache */
   /* replace906_0 -> 98 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 98 },
      -1,
   } },

   /* ('~frsq', ('fexp2', 'a')) => ('fexp2', ('fmul', -0.5, 'a')) */
   /* search907_0_0 -> 32 in the cache */
   /* search907_0 -> 2543 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 2543 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbfe0000000000000 /* -0.5 */ },
   } },
   /* replace907_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2586, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 2587 },
      -1,
   } },

   /* ('~flog2', ('fsqrt', 'a')) => ('fmul', 0.5, ('flog2', 'a')) */
   /* search908_0_0 -> 32 in the cache */
   /* search908_0 -> 2562 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2562 },
      -1,
   } },

   /* replace908_0 -> 2559 in the cache */
   /* replace908_1_0 -> 32 in the cache */
   /* replace908_1 -> 2541 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2559, 2541 },
      -1,
   } },

   /* ('~flog2', ('frcp', 'a')) => ('fneg', ('flog2', 'a')) */
   /* search909_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2591 },
      -1,
   } },

   /* replace909_0_0 -> 32 in the cache */
   /* replace909_0 -> 2541 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2541 },
      -1,
   } },

   /* ('~flog2', ('frsq', 'a')) => ('fmul', -0.5, ('flog2', 'a')) */
   /* search910_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2594 },
      -1,
   } },

   /* replace910_0 -> 2586 in the cache */
   /* replace910_1_0 -> 32 in the cache */
   /* replace910_1 -> 2541 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2586, 2541 },
      -1,
   } },

   /* ('~flog2', ('fpow', 'a', 'b')) => ('fmul', 'b', ('flog2', 'a')) */
   /* search911_0_0 -> 0 in the cache */
   /* search911_0_1 -> 3 in the cache */
   /* search911_0 -> 2545 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 2545 },
      -1,
   } },

   /* replace911_0 -> 3 in the cache */
   /* replace911_1_0 -> 0 in the cache */
   /* replace911_1 -> 2546 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 2546 },
      -1,
   } },

   /* ('~fmul', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('fadd', 'a', 'b')) */
   /* search912_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 0 },
      0,
   } },
   /* search912_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 3 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 2599, 2600 },
      -1,
   } },

   /* replace912_0_0 -> 0 in the cache */
   /* replace912_0_1 -> 3 in the cache */
   /* replace912_0 -> 261 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 1,
      { 261 },
      -1,
   } },

   /* ('bcsel', ('flt', 'a', 0.0), 0.0, ('fsqrt', 'a')) => ('fsqrt', ('fmax', 'a', 0.0)) */
   /* search913_0_0 -> 32 in the cache */
   /* search913_0_1 -> 113 in the cache */
   /* search913_0 -> 674 in the cache */
   /* search913_1 -> 113 in the cache */
   /* search913_2_0 -> 32 in the cache */
   /* search913_2 -> 2562 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 674, 113, 2562 },
      -1,
   } },

   /* replace913_0_0 -> 32 in the cache */
   /* replace913_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 32, 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 1,
      { 2604 },
      -1,
   } },

   /* ('~fmul', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a') */
   /* search914_0_0 -> 32 in the cache */
   /* search914_0 -> 2562 in the cache */
   /* search914_1_0 -> 32 in the cache */
   /* search914_1 -> 2562 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 2562, 2562 },
      -1,
   } },

   /* replace914_0 -> 32 in the cache */
   /* replace914 -> 103 in the cache */

   /* ('~fmulz', ('fsqrt', 'a'), ('fsqrt', 'a')) => ('fabs', 'a') */
   /* search915_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search915_1_0 -> 16 in the cache */
   /* search915_1 -> 2607 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      -1, 0,
      { 2607, 2607 },
      -1,
   } },

   /* replace915_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('~fdiv', 1.0, 'a') => ('frcp', 'a') */
   /* search916_0 -> 230 in the cache */
   /* search916_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 230, 32 },
      -1,
   } },

   /* replace916_0 -> 32 in the cache */
   /* replace916 -> 2591 in the cache */

   /* ('fdiv', 'a', 'b') => ('fmul', 'a', ('frcp', 'b')) */
   /* search917_0 -> 0 in the cache */
   /* search917_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace917_0 -> 0 in the cache */
   /* replace917_1_0 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2612 },
      -1,
   } },

   /* ('~frcp', ('frcp', 'a')) => a */
   /* search918_0_0 -> 32 in the cache */
   /* search918_0 -> 2591 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2591 },
      -1,
   } },

   /* replace918 -> 32 in the cache */

   /* ('~frcp', ('fsqrt', 'a')) => ('frsq', 'a') */
   /* search919_0_0 -> 32 in the cache */
   /* search919_0 -> 2562 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2562 },
      -1,
   } },

   /* replace919_0 -> 32 in the cache */
   /* replace919 -> 2594 in the cache */

   /* ('fsqrt', 'a') => ('frcp', ('frsq', 'a')) */
   /* search920_0 -> 32 in the cache */
   /* search920 -> 2562 in the cache */

   /* replace920_0_0 -> 32 in the cache */
   /* replace920_0 -> 2594 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2594 },
      -1,
   } },

   /* ('~frcp', ('frsq', 'a')) => ('fsqrt', 'a') */
   /* search921_0_0 -> 32 in the cache */
   /* search921_0 -> 2594 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 2594 },
      -1,
   } },

   /* replace921_0 -> 32 in the cache */
   /* replace921 -> 2562 in the cache */

   /* ('fsin', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.5))), 1.0)))), 4.0)) */
   /* search922_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace922_0_0_0_0_0_0 -> 2563 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fc45f306dc9c883 /* 0.15915494309189535 */ },
   } },
   /* replace922_0_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      5, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_0_0_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      4, 2,
      { 2620, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2621 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 3,
      { 2563, 2622 },
      -1,
   } },
   /* replace922_0_0_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2623, 230 },
      -1,
   } },
   /* replace922_0_0_0_1_0_0_0 -> 2563 in the cache */
   /* replace922_0_0_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      9, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_0_0_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      8, 2,
      { 2625, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2626 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      7, 3,
      { 2563, 2627 },
      -1,
   } },
   /* replace922_0_0_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2628, 230 },
      -1,
   } },
   /* replace922_0_0_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace922_0_0_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      12, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_0_0_1_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      11, 2,
      { 2630, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2631 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      10, 3,
      { 2563, 2632 },
      -1,
   } },
   /* replace922_0_0_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2633, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2634 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      6, 7,
      { 2629, 2635 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2624, 2636 },
      -1,
   } },
   /* replace922_0_0_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 11,
      { 2637, 2566 },
      -1,
   } },
   /* replace922_0_1_0_0_0_0_0 -> 2563 in the cache */
   /* replace922_0_1_0_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_1_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      16, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_1_0_0_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      15, 2,
      { 2639, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2640 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      14, 3,
      { 2563, 2641 },
      -1,
   } },
   /* replace922_0_1_0_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2642, 230 },
      -1,
   } },
   /* replace922_0_1_0_0_1_0_0_0 -> 2563 in the cache */
   /* replace922_0_1_0_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_1_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      20, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_1_0_0_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      19, 2,
      { 2644, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2645 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      18, 3,
      { 2563, 2646 },
      -1,
   } },
   /* replace922_0_1_0_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2647, 230 },
      -1,
   } },
   /* replace922_0_1_0_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace922_0_1_0_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_1_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      23, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_1_0_0_1_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      22, 2,
      { 2649, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2650 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      21, 3,
      { 2563, 2651 },
      -1,
   } },
   /* replace922_0_1_0_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2652, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2653 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      17, 7,
      { 2648, 2654 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2643, 2655 },
      -1,
   } },
   /* replace922_0_1_0_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      13, 11,
      { 2656, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 11,
      { 2657 },
      -1,
   } },
   /* replace922_0_2_0_0_0_0_0 -> 2563 in the cache */
   /* replace922_0_2_0_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_2_0_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      27, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_2_0_0_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      26, 2,
      { 2659, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2660 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      25, 3,
      { 2563, 2661 },
      -1,
   } },
   /* replace922_0_2_0_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2662, 230 },
      -1,
   } },
   /* replace922_0_2_0_0_1_0_0_0 -> 2563 in the cache */
   /* replace922_0_2_0_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_2_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      31, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_2_0_0_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      30, 2,
      { 2664, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2665 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      29, 3,
      { 2563, 2666 },
      -1,
   } },
   /* replace922_0_2_0_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2667, 230 },
      -1,
   } },
   /* replace922_0_2_0_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace922_0_2_0_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_0_2_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      34, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_0_2_0_0_1_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      33, 2,
      { 2669, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      32, 3,
      { 2563, 2671 },
      -1,
   } },
   /* replace922_0_2_0_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2672, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2673 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      28, 7,
      { 2668, 2674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2663, 2675 },
      -1,
   } },
   /* replace922_0_2_0_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      24, 11,
      { 2676, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 11,
      { 2677 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 34,
      { 2638, 2658, 2678 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fcccccccccccccd /* 0.225 */ },
   } },
   /* replace922_2_0_0_0_0 -> 2563 in the cache */
   /* replace922_2_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_2_0_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      38, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_2_0_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      37, 2,
      { 2681, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2682 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      36, 3,
      { 2563, 2683 },
      -1,
   } },
   /* replace922_2_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2684, 230 },
      -1,
   } },
   /* replace922_2_0_1_0_0_0 -> 2563 in the cache */
   /* replace922_2_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_2_0_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      42, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_2_0_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      41, 2,
      { 2686, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2687 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      40, 3,
      { 2563, 2688 },
      -1,
   } },
   /* replace922_2_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2689, 230 },
      -1,
   } },
   /* replace922_2_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace922_2_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace922_2_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      45, 1,
      { 2619, 32 },
      -1,
   } },
   /* replace922_2_0_1_1_0_0_1_0_1 -> 2559 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      44, 2,
      { 2691, 2559 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2692 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      43, 3,
      { 2563, 2693 },
      -1,
   } },
   /* replace922_2_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2694, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2695 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      39, 7,
      { 2690, 2696 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2685, 2697 },
      -1,
   } },
   /* replace922_2_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      35, 11,
      { 2698, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 46,
      { 2679, 2680, 2699 },
      -1,
   } },

   /* ('fcos', 'a') => ('ffma', ('ffma', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0), ('fabs', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)), ('fneg', ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0))), 0.225, ('fmul', ('fsub', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fmul', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0), ('fabs', ('fsub', ('fmul', 2.0, ('ffract', ('fadd', ('fmul', 0.15915494309189535, 'a'), 0.75))), 1.0)))), 4.0)) */
   /* search923_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace923_0_0_0_0_0_0 -> 2563 in the cache */
   /* replace923_0_0_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_0_0_0_0_1_0_0 -> 2620 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe8000000000000 /* 0.75 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      4, 2,
      { 2620, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2703 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 3,
      { 2563, 2704 },
      -1,
   } },
   /* replace923_0_0_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2705, 230 },
      -1,
   } },
   /* replace923_0_0_0_1_0_0_0 -> 2563 in the cache */
   /* replace923_0_0_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_0_0_1_0_0_1_0_0 -> 2625 in the cache */
   /* replace923_0_0_0_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      8, 2,
      { 2625, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2707 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      7, 3,
      { 2563, 2708 },
      -1,
   } },
   /* replace923_0_0_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2709, 230 },
      -1,
   } },
   /* replace923_0_0_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace923_0_0_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_0_0_1_1_0_0_1_0_0 -> 2630 in the cache */
   /* replace923_0_0_0_1_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      11, 2,
      { 2630, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2711 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      10, 3,
      { 2563, 2712 },
      -1,
   } },
   /* replace923_0_0_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2713, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2714 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      6, 7,
      { 2710, 2715 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2706, 2716 },
      -1,
   } },
   /* replace923_0_0_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 11,
      { 2717, 2566 },
      -1,
   } },
   /* replace923_0_1_0_0_0_0_0 -> 2563 in the cache */
   /* replace923_0_1_0_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_1_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_1_0_0_0_0_1_0_0 -> 2639 in the cache */
   /* replace923_0_1_0_0_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      15, 2,
      { 2639, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2719 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      14, 3,
      { 2563, 2720 },
      -1,
   } },
   /* replace923_0_1_0_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2721, 230 },
      -1,
   } },
   /* replace923_0_1_0_0_1_0_0_0 -> 2563 in the cache */
   /* replace923_0_1_0_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_1_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_1_0_0_1_0_0_1_0_0 -> 2644 in the cache */
   /* replace923_0_1_0_0_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      19, 2,
      { 2644, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2723 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      18, 3,
      { 2563, 2724 },
      -1,
   } },
   /* replace923_0_1_0_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2725, 230 },
      -1,
   } },
   /* replace923_0_1_0_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace923_0_1_0_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_1_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_1_0_0_1_1_0_0_1_0_0 -> 2649 in the cache */
   /* replace923_0_1_0_0_1_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      22, 2,
      { 2649, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2727 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      21, 3,
      { 2563, 2728 },
      -1,
   } },
   /* replace923_0_1_0_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2729, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2730 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      17, 7,
      { 2726, 2731 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2722, 2732 },
      -1,
   } },
   /* replace923_0_1_0_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      13, 11,
      { 2733, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 11,
      { 2734 },
      -1,
   } },
   /* replace923_0_2_0_0_0_0_0 -> 2563 in the cache */
   /* replace923_0_2_0_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_2_0_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_2_0_0_0_0_1_0_0 -> 2659 in the cache */
   /* replace923_0_2_0_0_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      26, 2,
      { 2659, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2736 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      25, 3,
      { 2563, 2737 },
      -1,
   } },
   /* replace923_0_2_0_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2738, 230 },
      -1,
   } },
   /* replace923_0_2_0_0_1_0_0_0 -> 2563 in the cache */
   /* replace923_0_2_0_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_2_0_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_2_0_0_1_0_0_1_0_0 -> 2664 in the cache */
   /* replace923_0_2_0_0_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      30, 2,
      { 2664, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2740 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      29, 3,
      { 2563, 2741 },
      -1,
   } },
   /* replace923_0_2_0_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2742, 230 },
      -1,
   } },
   /* replace923_0_2_0_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace923_0_2_0_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_0_2_0_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_0_2_0_0_1_1_0_0_1_0_0 -> 2669 in the cache */
   /* replace923_0_2_0_0_1_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      33, 2,
      { 2669, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2744 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      32, 3,
      { 2563, 2745 },
      -1,
   } },
   /* replace923_0_2_0_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2746, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2747 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      28, 7,
      { 2743, 2748 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2739, 2749 },
      -1,
   } },
   /* replace923_0_2_0_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      24, 11,
      { 2750, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 11,
      { 2751 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 34,
      { 2718, 2735, 2752 },
      -1,
   } },
   /* replace923_1 -> 2680 in the cache */
   /* replace923_2_0_0_0_0 -> 2563 in the cache */
   /* replace923_2_0_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_2_0_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_2_0_0_0_1_0_0 -> 2681 in the cache */
   /* replace923_2_0_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      37, 2,
      { 2681, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2754 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      36, 3,
      { 2563, 2755 },
      -1,
   } },
   /* replace923_2_0_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2756, 230 },
      -1,
   } },
   /* replace923_2_0_1_0_0_0 -> 2563 in the cache */
   /* replace923_2_0_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_2_0_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_2_0_1_0_0_1_0_0 -> 2686 in the cache */
   /* replace923_2_0_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      41, 2,
      { 2686, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2758 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      40, 3,
      { 2563, 2759 },
      -1,
   } },
   /* replace923_2_0_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2760, 230 },
      -1,
   } },
   /* replace923_2_0_1_1_0_0_0 -> 2563 in the cache */
   /* replace923_2_0_1_1_0_0_1_0_0_0 -> 2619 in the cache */
   /* replace923_2_0_1_1_0_0_1_0_0_1 -> 32 in the cache */
   /* replace923_2_0_1_1_0_0_1_0_0 -> 2691 in the cache */
   /* replace923_2_0_1_1_0_0_1_0_1 -> 2702 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      44, 2,
      { 2691, 2702 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 2,
      { 2762 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      43, 3,
      { 2563, 2763 },
      -1,
   } },
   /* replace923_2_0_1_1_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 3,
      { 2764, 230 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 3,
      { 2765 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      39, 7,
      { 2761, 2766 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 10,
      { 2757, 2767 },
      -1,
   } },
   /* replace923_2_1 -> 2566 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      35, 11,
      { 2768, 2566 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 46,
      { 2753, 2680, 2769 },
      -1,
   } },

   /* ('i2b16(is_used_by_if)', 'a') => ('ine16', 'a', 0) */
   /* search924_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2b16,
      -1, 0,
      { 32 },
      2,
   } },

   /* replace924_0 -> 32 in the cache */
   /* replace924_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ine16,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('i2b32(is_used_by_if)', 'a') => ('ine32', 'a', 0) */
   /* search925_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2b32,
      -1, 0,
      { 32 },
      2,
   } },

   /* replace925_0 -> 32 in the cache */
   /* replace925_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ine32,
      0, 1,
      { 32, 61 },
      -1,
   } },

   /* ('i2b1(is_used_by_if)', 'a') => ('ine', 'a', 0) */
   /* search926_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2b1,
      -1, 0,
      { 32 },
      2,
   } },

   /* replace926_0 -> 32 in the cache */
   /* replace926_1 -> 61 in the cache */
   /* replace926 -> 1188 in the cache */

   /* ('ieq', 'a', True) => a */
   /* search927_0 -> 722 in the cache */
   /* search927_1 -> 1192 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 722, 1192 },
      -1,
   } },

   /* replace927 -> 722 in the cache */

   /* ('ine(is_not_used_by_if)', 'a', True) => ('inot', 'a') */
   /* search928_0 -> 722 in the cache */
   /* search928_1 -> 1192 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 722, 1192 },
      3,
   } },

   /* replace928_0 -> 722 in the cache */
   /* replace928 -> 1087 in the cache */

   /* ('ine', 'a', False) => a */
   /* search929_0 -> 722 in the cache */
   /* search929_1 -> 1039 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 722, 1039 },
      -1,
   } },

   /* replace929 -> 722 in the cache */

   /* ('ieq(is_not_used_by_if)', 'a', False) => ('inot', 'a') */
   /* search930_0 -> 722 in the cache */
   /* search930_1 -> 1039 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 722, 1039 },
      3,
   } },

   /* replace930_0 -> 722 in the cache */
   /* replace930 -> 1087 in the cache */

   /* ('bcsel', 'a', True, False) => a */
   /* search931_0 -> 722 in the cache */
   /* search931_1 -> 1192 in the cache */
   /* search931_2 -> 1039 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1192, 1039 },
      -1,
   } },

   /* replace931 -> 722 in the cache */

   /* ('bcsel', 'a', False, True) => ('inot', 'a') */
   /* search932_0 -> 722 in the cache */
   /* search932_1 -> 1039 in the cache */
   /* search932_2 -> 1192 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1039, 1192 },
      -1,
   } },

   /* replace932_0 -> 722 in the cache */
   /* replace932 -> 1087 in the cache */

   /* ('bcsel', True, 'b', 'c') => b */
   /* search933_0 -> 1192 in the cache */
   /* search933_1 -> 1073 in the cache */
   /* search933_2 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1192, 1073, 669 },
      -1,
   } },

   /* replace933 -> 1073 in the cache */

   /* ('bcsel', False, 'b', 'c') => c */
   /* search934_0 -> 1039 in the cache */
   /* search934_1 -> 1073 in the cache */
   /* search934_2 -> 669 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1039, 1073, 669 },
      -1,
   } },

   /* replace934 -> 669 in the cache */

   /* ('bcsel@16', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search935_0 -> 722 in the cache */
   /* search935_1 -> 466 in the cache */
   /* search935_2 -> 115 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 466, 115 },
      -1,
   } },

   /* replace935_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('bcsel@16', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search936_0 -> 722 in the cache */
   /* search936_1 -> 115 in the cache */
   /* search936_2 -> 466 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 115, 466 },
      -1,
   } },

   /* replace936_0_0 -> 722 in the cache */
   /* replace936_0 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1087 },
      -1,
   } },

   /* ('bcsel@16', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search937_0 -> 722 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2788, 2789 },
      -1,
   } },

   /* replace937_0_0 -> 722 in the cache */
   /* replace937_0 -> 2785 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2785 },
      -1,
   } },

   /* ('bcsel@16', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search938_0 -> 722 in the cache */
   /* search938_1 -> 2789 in the cache */
   /* search938_2 -> 2788 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2789, 2788 },
      -1,
   } },

   /* replace938_0_0_0 -> 722 in the cache */
   /* replace938_0_0 -> 1087 in the cache */
   /* replace938_0 -> 2787 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2787 },
      -1,
   } },

   /* ('bcsel@32', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search939_0 -> 722 in the cache */
   /* search939_1 -> 232 in the cache */
   /* search939_2 -> 117 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 232, 117 },
      -1,
   } },

   /* replace939_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('bcsel@32', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search940_0 -> 722 in the cache */
   /* search940_1 -> 117 in the cache */
   /* search940_2 -> 232 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 117, 232 },
      -1,
   } },

   /* replace940_0_0 -> 722 in the cache */
   /* replace940_0 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1087 },
      -1,
   } },

   /* ('bcsel@32', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search941_0 -> 722 in the cache */
   /* search941_1 -> 267 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 267, 2798 },
      -1,
   } },

   /* replace941_0_0 -> 722 in the cache */
   /* replace941_0 -> 2795 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2795 },
      -1,
   } },

   /* ('bcsel@32', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search942_0 -> 722 in the cache */
   /* search942_1 -> 2798 in the cache */
   /* search942_2 -> 267 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2798, 267 },
      -1,
   } },

   /* replace942_0_0_0 -> 722 in the cache */
   /* replace942_0_0 -> 1087 in the cache */
   /* replace942_0 -> 2797 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2797 },
      -1,
   } },

   /* ('bcsel@64', 'a', 1.0, 0.0) => ('b2f', 'a') */
   /* search943_0 -> 722 in the cache */
   /* search943_1 -> 607 in the cache */
   /* search943_2 -> 1871 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 607, 1871 },
      -1,
   } },

   /* replace943_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('bcsel@64', 'a', 0.0, 1.0) => ('b2f', ('inot', 'a')) */
   /* search944_0 -> 722 in the cache */
   /* search944_1 -> 1871 in the cache */
   /* search944_2 -> 607 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1871, 607 },
      -1,
   } },

   /* replace944_0_0 -> 722 in the cache */
   /* replace944_0 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 1087 },
      -1,
   } },

   /* ('bcsel@64', 'a', -1.0, -0.0) => ('fneg', ('b2f', 'a')) */
   /* search945_0 -> 722 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x8000000000000000 /* -0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2807, 2808 },
      -1,
   } },

   /* replace945_0_0 -> 722 in the cache */
   /* replace945_0 -> 2804 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2804 },
      -1,
   } },

   /* ('bcsel@64', 'a', -0.0, -1.0) => ('fneg', ('b2f', ('inot', 'a'))) */
   /* search946_0 -> 722 in the cache */
   /* search946_1 -> 2808 in the cache */
   /* search946_2 -> 2807 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2808, 2807 },
      -1,
   } },

   /* replace946_0_0_0 -> 722 in the cache */
   /* replace946_0_0 -> 1087 in the cache */
   /* replace946_0 -> 2806 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2806 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', 'b') => b */
   /* search947_0 -> 722 in the cache */
   /* search947_1 -> 3 in the cache */
   /* search947_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 3, 3 },
      -1,
   } },

   /* replace947 -> 3 in the cache */

   /* ('~fcsel', 'a', 'b', 'b') => b */
   /* search948_0 -> 16 in the cache */
   /* search948_1 -> 36 in the cache */
   /* search948_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 16, 36, 36 },
      -1,
   } },

   /* replace948 -> 36 in the cache */

   /* ('bcsel', 'a', -1, 0) => ('ineg', ('b2i', 'a@1')) */
   /* search949_0 -> 722 in the cache */
   /* search949_1 -> 2511 in the cache */
   /* search949_2 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2511, 1184 },
      -1,
   } },

   /* replace949_0_0 -> 722 in the cache */
   /* replace949_0 -> 1183 in the cache */
   /* replace949 -> 2198 in the cache */

   /* ('bcsel', 'a', 0, -1) => ('ineg', ('b2i', ('inot', 'a'))) */
   /* search950_0 -> 722 in the cache */
   /* search950_1 -> 1184 in the cache */
   /* search950_2 -> 2511 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1184, 2511 },
      -1,
   } },

   /* replace950_0_0_0 -> 722 in the cache */
   /* replace950_0_0 -> 1087 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1087 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2817 },
      -1,
   } },

   /* ('bcsel', 'a', 1, 0) => ('b2i', 'a@1') */
   /* search951_0 -> 722 in the cache */
   { .constant = {
      { nir_search_value_constant, 0 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   /* search951_2 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 2819, 1184 },
      -1,
   } },

   /* replace951_0 -> 722 in the cache */
   /* replace951 -> 1183 in the cache */

   /* ('bcsel', 'a', 0, 1) => ('b2i', ('inot', 'a')) */
   /* search952_0 -> 722 in the cache */
   /* search952_1 -> 1184 in the cache */
   /* search952_2 -> 2819 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1184, 2819 },
      -1,
   } },

   /* replace952_0_0 -> 722 in the cache */
   /* replace952_0 -> 1087 in the cache */
   /* replace952 -> 2817 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search953_0_0_0 -> 722 in the cache */
   /* search953_0_0 -> 1183 in the cache */
   /* search953_0 -> 2198 in the cache */
   /* search953_1_0_0 -> 1090 in the cache */
   /* search953_1_0 -> 2199 in the cache */
   /* search953_1 -> 2200 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace953_0_0_0 -> 722 in the cache */
   /* replace953_0_0_1 -> 1090 in the cache */
   /* replace953_0_0 -> 1102 in the cache */
   /* replace953_0 -> 2460 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2460 },
      -1,
   } },

   /* ('ior', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search954_0_0_0 -> 722 in the cache */
   /* search954_0_0 -> 1183 in the cache */
   /* search954_0 -> 2198 in the cache */
   /* search954_1_0_0 -> 1090 in the cache */
   /* search954_1_0 -> 2199 in the cache */
   /* search954_1 -> 2200 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace954_0_0_0 -> 722 in the cache */
   /* replace954_0_0_1 -> 1090 in the cache */
   /* replace954_0_0 -> 1094 in the cache */
   /* replace954_0 -> 2463 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 2463 },
      -1,
   } },

   /* ('ieq', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a') */
   /* search955_0_0_0 -> 722 in the cache */
   /* search955_0_0 -> 1183 in the cache */
   /* search955_0 -> 2198 in the cache */
   /* search955_1 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2198, 1184 },
      -1,
   } },

   /* replace955_0 -> 722 in the cache */
   /* replace955 -> 1087 in the cache */

   /* ('ieq', ('ineg', ('b2i', 'a@1')), -1) => a */
   /* search956_0_0_0 -> 722 in the cache */
   /* search956_0_0 -> 1183 in the cache */
   /* search956_0 -> 2198 in the cache */
   /* search956_1 -> 2511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2198, 2511 },
      -1,
   } },

   /* replace956 -> 722 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), 0) => a */
   /* search957_0_0_0 -> 722 in the cache */
   /* search957_0_0 -> 1183 in the cache */
   /* search957_0 -> 2198 in the cache */
   /* search957_1 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2198, 1184 },
      -1,
   } },

   /* replace957 -> 722 in the cache */

   /* ('ine', ('ineg', ('b2i', 'a@1')), -1) => ('inot', 'a') */
   /* search958_0_0_0 -> 722 in the cache */
   /* search958_0_0 -> 1183 in the cache */
   /* search958_0 -> 2198 in the cache */
   /* search958_1 -> 2511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2198, 2511 },
      -1,
   } },

   /* replace958_0 -> 722 in the cache */
   /* replace958 -> 1087 in the cache */

   /* ('ige', ('ineg', ('b2i', 'a@1')), 0) => ('inot', 'a') */
   /* search959_0_0_0 -> 722 in the cache */
   /* search959_0_0 -> 1183 in the cache */
   /* search959_0 -> 2198 in the cache */
   /* search959_1 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2198, 1184 },
      -1,
   } },

   /* replace959_0 -> 722 in the cache */
   /* replace959 -> 1087 in the cache */

   /* ('ilt', ('ineg', ('b2i', 'a@1')), 0) => a */
   /* search960_0_0_0 -> 722 in the cache */
   /* search960_0_0 -> 1183 in the cache */
   /* search960_0 -> 2198 in the cache */
   /* search960_1 -> 1184 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2198, 1184 },
      -1,
   } },

   /* replace960 -> 722 in the cache */

   /* ('ult', 0, ('ineg', ('b2i', 'a@1'))) => a */
   /* search961_0 -> 1184 in the cache */
   /* search961_1_0_0 -> 722 in the cache */
   /* search961_1_0 -> 1183 in the cache */
   /* search961_1 -> 2198 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1184, 2198 },
      -1,
   } },

   /* replace961 -> 722 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a')), 1.0) => ('b2f', 'a') */
   /* search962_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 2833 },
      -1,
   } },
   /* search962_1 -> 1095 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2834, 1095 },
      -1,
   } },

   /* replace962_0 -> 32 in the cache */
   /* replace962 -> 1472 in the cache */

   /* ('iand', ('ineg', ('b2i', 'a')), 1) => ('b2i', 'a') */
   /* search963_0_0_0 -> 32 in the cache */
   /* search963_0_0 -> 2833 in the cache */
   /* search963_0 -> 2834 in the cache */
   /* search963_1 -> 2819 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 2834, 2819 },
      -1,
   } },

   /* replace963_0 -> 32 in the cache */
   /* replace963 -> 2833 in the cache */

   /* ('imax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search964_0_0_0 -> 722 in the cache */
   /* search964_0_0 -> 1183 in the cache */
   /* search964_0 -> 2198 in the cache */
   /* search964_1_0_0 -> 1090 in the cache */
   /* search964_1_0 -> 2199 in the cache */
   /* search964_1 -> 2200 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace964_0_0_0 -> 722 in the cache */
   /* replace964_0_0_1 -> 1090 in the cache */
   /* replace964_0_0 -> 1102 in the cache */
   /* replace964_0 -> 2460 in the cache */
   /* replace964 -> 2823 in the cache */

   /* ('imin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search965_0_0_0 -> 722 in the cache */
   /* search965_0_0 -> 1183 in the cache */
   /* search965_0 -> 2198 in the cache */
   /* search965_1_0_0 -> 1090 in the cache */
   /* search965_1_0 -> 2199 in the cache */
   /* search965_1 -> 2200 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace965_0_0_0 -> 722 in the cache */
   /* replace965_0_0_1 -> 1090 in the cache */
   /* replace965_0_0 -> 1094 in the cache */
   /* replace965_0 -> 2463 in the cache */
   /* replace965 -> 2825 in the cache */

   /* ('umax', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('ior', 'a', 'b'))) */
   /* search966_0_0_0 -> 722 in the cache */
   /* search966_0_0 -> 1183 in the cache */
   /* search966_0 -> 2198 in the cache */
   /* search966_1_0_0 -> 1090 in the cache */
   /* search966_1_0 -> 2199 in the cache */
   /* search966_1 -> 2200 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace966_0_0_0 -> 722 in the cache */
   /* replace966_0_0_1 -> 1090 in the cache */
   /* replace966_0_0 -> 1094 in the cache */
   /* replace966_0 -> 2463 in the cache */
   /* replace966 -> 2825 in the cache */

   /* ('umin', ('ineg', ('b2i', 'a@1')), ('ineg', ('b2i', 'b@1'))) => ('ineg', ('b2i', ('iand', 'a', 'b'))) */
   /* search967_0_0_0 -> 722 in the cache */
   /* search967_0_0 -> 1183 in the cache */
   /* search967_0 -> 2198 in the cache */
   /* search967_1_0_0 -> 1090 in the cache */
   /* search967_1_0 -> 2199 in the cache */
   /* search967_1 -> 2200 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 2198, 2200 },
      -1,
   } },

   /* replace967_0_0_0 -> 722 in the cache */
   /* replace967_0_0_1 -> 1090 in the cache */
   /* replace967_0_0 -> 1102 in the cache */
   /* replace967_0 -> 2460 in the cache */
   /* replace967 -> 2823 in the cache */

   /* ('i2b16', ('b2i', 'a@16')) => a */
   /* search968_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2b16,
      -1, 0,
      { 2841 },
      -1,
   } },

   /* replace968 -> 11 in the cache */

   /* ('i2b32', ('b2i', 'a@32')) => a */
   /* search969_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2b32,
      -1, 0,
      { 2843 },
      -1,
   } },

   /* replace969 -> 16 in the cache */

   /* ('f2i', ('ftrunc', 'a')) => ('f2i', 'a') */
   /* search970_0_0 -> 32 in the cache */
   /* search970_0 -> 673 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 673 },
      -1,
   } },

   /* replace970_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('f2u', ('ftrunc', 'a')) => ('f2u', 'a') */
   /* search971_0_0 -> 32 in the cache */
   /* search971_0 -> 673 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 673 },
      -1,
   } },

   /* replace971_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2b', ('ineg', 'a')) => ('i2b', 'a') */
   /* search972_0_0 -> 32 in the cache */
   /* search972_0 -> 100 in the cache */
   /* search972 -> 111 in the cache */

   /* replace972_0 -> 32 in the cache */
   /* replace972 -> 112 in the cache */

   /* ('i2b', ('iabs', 'a')) => ('i2b', 'a') */
   /* search973_0_0 -> 32 in the cache */
   /* search973_0 -> 106 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2b,
      -1, 0,
      { 106 },
      -1,
   } },

   /* replace973_0 -> 32 in the cache */
   /* replace973 -> 112 in the cache */

   /* ('inot', ('f2b1', 'a')) => ('feq', 'a', 0.0) */
   /* search974_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_f2b1,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 2850 },
      -1,
   } },

   /* replace974_0 -> 32 in the cache */
   /* replace974_1 -> 113 in the cache */
   /* replace974 -> 1214 in the cache */

   /* ('f2fmp', ('f2f32', 'a@16')) => a */
   /* search975_0_0 -> 11 in the cache */
   /* search975_0 -> 1745 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1745 },
      -1,
   } },

   /* replace975 -> 11 in the cache */

   /* ('i2imp', ('i2i32', 'a@16')) => a */
   /* search976_0_0 -> 11 in the cache */
   /* search976_0 -> 1747 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1747 },
      -1,
   } },

   /* replace976 -> 11 in the cache */

   /* ('i2imp', ('u2u32', 'a@16')) => a */
   /* search977_0_0 -> 11 in the cache */
   /* search977_0 -> 1749 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1749 },
      -1,
   } },

   /* replace977 -> 11 in the cache */

   /* ('f2imp', ('f2f32', 'a@16')) => ('f2i16', 'a') */
   /* search978_0_0 -> 11 in the cache */
   /* search978_0 -> 1745 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 1745 },
      -1,
   } },

   /* replace978_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2ump', ('f2f32', 'a@16')) => ('f2u16', 'a') */
   /* search979_0_0 -> 11 in the cache */
   /* search979_0 -> 1745 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 1745 },
      -1,
   } },

   /* replace979_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('i2fmp', ('i2i32', 'a@16')) => ('i2f16', 'a') */
   /* search980_0_0 -> 11 in the cache */
   /* search980_0 -> 1747 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 1747 },
      -1,
   } },

   /* replace980_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('u2fmp', ('u2u32', 'a@16')) => ('u2f16', 'a') */
   /* search981_0_0 -> 11 in the cache */
   /* search981_0 -> 1749 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 1749 },
      -1,
   } },

   /* replace981_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* ('f2fmp', ('b2f32', 'a@1')) => ('b2f16', 'a') */
   /* search982_0_0 -> 722 in the cache */
   /* search982_0 -> 1352 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 1352 },
      -1,
   } },

   /* replace982_0 -> 722 in the cache */
   /* replace982 -> 1348 in the cache */

   /* ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search983_0_0 -> 722 in the cache */
   /* search983_0 -> 1340 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 1340 },
      -1,
   } },

   /* replace983_0 -> 722 in the cache */
   /* replace983 -> 1336 in the cache */

   /* ('i2imp', ('b2i32', 'a@1')) => ('b2i16', 'a') */
   /* search984_0_0 -> 722 in the cache */
   /* search984_0 -> 1340 in the cache */
   /* search984 -> 2864 in the cache */

   /* replace984_0 -> 722 in the cache */
   /* replace984 -> 1336 in the cache */

   /* ('f2imp', ('b2f32', 'a@1')) => ('b2i16', 'a') */
   /* search985_0_0 -> 722 in the cache */
   /* search985_0 -> 1352 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 1352 },
      -1,
   } },

   /* replace985_0 -> 722 in the cache */
   /* replace985 -> 1336 in the cache */

   /* ('f2ump', ('b2f32', 'a@1')) => ('b2i16', 'a') */
   /* search986_0_0 -> 722 in the cache */
   /* search986_0 -> 1352 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 1352 },
      -1,
   } },

   /* replace986_0 -> 722 in the cache */
   /* replace986 -> 1336 in the cache */

   /* ('i2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a') */
   /* search987_0_0 -> 722 in the cache */
   /* search987_0 -> 1340 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 1340 },
      -1,
   } },

   /* replace987_0 -> 722 in the cache */
   /* replace987 -> 1348 in the cache */

   /* ('u2fmp', ('b2i32', 'a@1')) => ('b2f16', 'a') */
   /* search988_0_0 -> 722 in the cache */
   /* search988_0 -> 1340 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 1340 },
      -1,
   } },

   /* replace988_0 -> 722 in the cache */
   /* replace988 -> 1348 in the cache */

   /* ('f2f32', ('f2fmp', 'a@32')) => a */
   /* search989_0_0 -> 16 in the cache */
   /* search989_0 -> 1941 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1941 },
      -1,
   } },

   /* replace989 -> 16 in the cache */

   /* ('i2i32', ('i2imp', 'a@32')) => a */
   /* search990_0_0 -> 16 in the cache */
   /* search990_0 -> 1952 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1952 },
      -1,
   } },

   /* replace990 -> 16 in the cache */

   /* ('u2u32', ('i2imp', 'a@32')) => a */
   /* search991_0_0 -> 16 in the cache */
   /* search991_0 -> 1952 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1952 },
      -1,
   } },

   /* replace991 -> 16 in the cache */

   /* ('i2i32', ('f2imp', 'a@32')) => ('f2i32', 'a') */
   /* search992_0_0 -> 16 in the cache */
   /* search992_0 -> 1951 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1951 },
      -1,
   } },

   /* replace992_0 -> 16 in the cache */
   /* replace992 -> 1949 in the cache */

   /* ('u2u32', ('f2ump', 'a@32')) => ('f2u32', 'a') */
   /* search993_0_0 -> 16 in the cache */
   /* search993_0 -> 1948 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1948 },
      -1,
   } },

   /* replace993_0 -> 16 in the cache */
   /* replace993 -> 1946 in the cache */

   /* ('f2f32', ('i2fmp', 'a@32')) => ('i2f32', 'a') */
   /* search994_0_0 -> 16 in the cache */
   /* search994_0 -> 1940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1940 },
      -1,
   } },

   /* replace994_0 -> 16 in the cache */
   /* replace994 -> 1938 in the cache */

   /* ('f2f32', ('u2fmp', 'a@32')) => ('u2f32', 'a') */
   /* search995_0_0 -> 16 in the cache */
   /* search995_0 -> 1937 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1937 },
      -1,
   } },

   /* replace995_0 -> 16 in the cache */
   /* replace995 -> 1935 in the cache */

   /* ('~f2f32', ('f2f64', 'a@32')) => a */
   /* search996_0_0 -> 16 in the cache */
   /* search996_0 -> 1852 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 1852 },
      -1,
   } },

   /* replace996 -> 16 in the cache */

   /* ('ffloor', 'a(is_integral)') => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      20,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 2877 },
      -1,
   } },

   /* replace997 -> 32 in the cache */

   /* ('fceil', 'a(is_integral)') => a */
   /* search998_0 -> 2877 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 2877 },
      -1,
   } },

   /* replace998 -> 32 in the cache */

   /* ('ftrunc', 'a(is_integral)') => a */
   /* search999_0 -> 2877 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 2877 },
      -1,
   } },

   /* replace999 -> 32 in the cache */

   /* ('fround_even', 'a(is_integral)') => a */
   /* search1000_0 -> 2877 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 2877 },
      -1,
   } },

   /* replace1000 -> 32 in the cache */

   /* ('~ffract', 'a(is_integral)') => 0.0 */
   /* search1001_0 -> 2877 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 2877 },
      -1,
   } },

   /* replace1001 -> 113 in the cache */

   /* ('fabs', 'a(is_not_negative)') => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2883 },
      -1,
   } },

   /* replace1002 -> 32 in the cache */

   /* ('iabs', 'a(is_not_negative)') => a */
   /* search1003_0 -> 2883 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 2883 },
      -1,
   } },

   /* replace1003 -> 32 in the cache */

   /* ('fsat', 'a(is_not_positive)') => 0.0 */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 2886 },
      -1,
   } },

   /* replace1004 -> 113 in the cache */

   /* ('~fmin', 'a(is_not_negative)', 1.0) => ('fsat', 'a') */
   /* search1005_0 -> 2883 in the cache */
   /* search1005_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 2883, 230 },
      -1,
   } },

   /* replace1005_0 -> 32 in the cache */
   /* replace1005 -> 206 in the cache */

   /* ('flt', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => False */
   /* search1006_0_0_0_0 -> 32 in the cache */
   /* search1006_0_0_0 -> 206 in the cache */
   /* search1006_0_0_1_0_0 -> 32 in the cache */
   /* search1006_0_0_1_0 -> 206 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 206 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 206, 2889 },
      -1,
   } },
   /* search1006_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 2890, 230 },
      -1,
   } },
   /* search1006_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 2,
      { 2891, 113 },
      -1,
   } },

   /* replace1006 -> 1039 in the cache */

   /* ('flt', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => False */
   /* search1007_0_0_0_0_0 -> 32 in the cache */
   /* search1007_0_0_0_0 -> 206 in the cache */
   /* search1007_0_0_0_1_0 -> 32 in the cache */
   /* search1007_0_0_0_1 -> 206 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 206, 206 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2893 },
      -1,
   } },
   /* search1007_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 2894, 230 },
      -1,
   } },
   /* search1007_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 2895, 113 },
      -1,
   } },

   /* replace1007 -> 1039 in the cache */

   /* ('fmax', ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fmul', ('fsat', 'a'), ('fneg', ('fsat', 'a'))), 1.0) */
   /* search1008_0_0_0_0 -> 32 in the cache */
   /* search1008_0_0_0 -> 206 in the cache */
   /* search1008_0_0_1_0_0 -> 32 in the cache */
   /* search1008_0_0_1_0 -> 206 in the cache */
   /* search1008_0_0_1 -> 2889 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 206, 2889 },
      -1,
   } },
   /* search1008_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 2897, 230 },
      -1,
   } },
   /* search1008_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 2898, 113 },
      -1,
   } },

   /* replace1008_0_0_0 -> 32 in the cache */
   /* replace1008_0_0 -> 206 in the cache */
   /* replace1008_0_1_0_0 -> 32 in the cache */
   /* replace1008_0_1_0 -> 206 in the cache */
   /* replace1008_0_1 -> 2889 in the cache */
   /* replace1008_0 -> 2890 in the cache */
   /* replace1008_1 -> 230 in the cache */
   /* replace1008 -> 2891 in the cache */

   /* ('fmax', ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0), 0.0) => ('fadd', ('fneg', ('fmul', ('fsat', 'a'), ('fsat', 'a'))), 1.0) */
   /* search1009_0_0_0_0_0 -> 32 in the cache */
   /* search1009_0_0_0_0 -> 206 in the cache */
   /* search1009_0_0_0_1_0 -> 32 in the cache */
   /* search1009_0_0_0_1 -> 206 in the cache */
   /* search1009_0_0_0 -> 2893 in the cache */
   /* search1009_0_0 -> 2894 in the cache */
   /* search1009_0_1 -> 230 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 2894, 230 },
      -1,
   } },
   /* search1009_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 2,
      { 2900, 113 },
      -1,
   } },

   /* replace1009_0_0_0_0 -> 32 in the cache */
   /* replace1009_0_0_0 -> 206 in the cache */
   /* replace1009_0_0_1_0 -> 32 in the cache */
   /* replace1009_0_0_1 -> 206 in the cache */
   /* replace1009_0_0 -> 2893 in the cache */
   /* replace1009_0 -> 2894 in the cache */
   /* replace1009_1 -> 230 in the cache */
   /* replace1009 -> 2895 in the cache */

   /* ('fneu', 'a(is_not_zero)', 0.0) => True */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      22,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1010_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 2902, 113 },
      -1,
   } },

   /* replace1010 -> 1192 in the cache */

   /* ('feq', 'a(is_not_zero)', 0.0) => False */
   /* search1011_0 -> 2902 in the cache */
   /* search1011_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 2902, 113 },
      -1,
   } },

   /* replace1011 -> 1039 in the cache */

   /* ('fge', 'a(is_a_number_not_negative)', 'b(is_a_number_not_positive)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      23,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      24,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2905, 2906 },
      -1,
   } },

   /* replace1012 -> 1192 in the cache */

   /* ('fge', 'a(is_not_positive)', 'b(is_gt_zero)') => False */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      25,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2908, 2909 },
      -1,
   } },

   /* replace1013 -> 1039 in the cache */

   /* ('fge', 'a(is_lt_zero)', 'b(is_not_negative)') => False */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      26,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      19,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2911, 2912 },
      -1,
   } },

   /* replace1014 -> 1039 in the cache */

   /* ('flt', 'a(is_not_negative)', 'b(is_not_positive)') => False */
   /* search1015_0 -> 1528 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      21,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1528, 2914 },
      -1,
   } },

   /* replace1015 -> 1039 in the cache */

   /* ('flt', 'a(is_a_number_not_positive)', 'b(is_a_number_gt_zero)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      24,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      27,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 2916, 2917 },
      -1,
   } },

   /* replace1016 -> 1192 in the cache */

   /* ('flt', 'a(is_a_number_lt_zero)', 'b(is_a_number_not_negative)') => True */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      28,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      23,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 2919, 2920 },
      -1,
   } },

   /* replace1017 -> 1192 in the cache */

   /* ('ine', 'a(is_not_zero)', 0) => True */
   /* search1018_0 -> 2902 in the cache */
   /* search1018_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 2902, 61 },
      -1,
   } },

   /* replace1018 -> 1192 in the cache */

   /* ('ieq', 'a(is_not_zero)', 0) => False */
   /* search1019_0 -> 2902 in the cache */
   /* search1019_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 2902, 61 },
      -1,
   } },

   /* replace1019 -> 1039 in the cache */

   /* ('ige', 'a(is_not_negative)', 'b(is_not_positive)') => True */
   /* search1020_0 -> 1528 in the cache */
   /* search1020_1 -> 2914 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1528, 2914 },
      -1,
   } },

   /* replace1020 -> 1192 in the cache */

   /* ('ige', 'a(is_not_positive)', 'b(is_gt_zero)') => False */
   /* search1021_0 -> 2908 in the cache */
   /* search1021_1 -> 2909 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2908, 2909 },
      -1,
   } },

   /* replace1021 -> 1039 in the cache */

   /* ('ige', 'a(is_lt_zero)', 'b(is_not_negative)') => False */
   /* search1022_0 -> 2911 in the cache */
   /* search1022_1 -> 2912 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2911, 2912 },
      -1,
   } },

   /* replace1022 -> 1039 in the cache */

   /* ('ilt', 'a(is_not_negative)', 'b(is_not_positive)') => False */
   /* search1023_0 -> 1528 in the cache */
   /* search1023_1 -> 2914 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1528, 2914 },
      -1,
   } },

   /* replace1023 -> 1039 in the cache */

   /* ('ilt', 'a(is_not_positive)', 'b(is_gt_zero)') => True */
   /* search1024_0 -> 2908 in the cache */
   /* search1024_1 -> 2909 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2908, 2909 },
      -1,
   } },

   /* replace1024 -> 1192 in the cache */

   /* ('ilt', 'a(is_lt_zero)', 'b(is_not_negative)') => True */
   /* search1025_0 -> 2911 in the cache */
   /* search1025_1 -> 2912 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2911, 2912 },
      -1,
   } },

   /* replace1025 -> 1192 in the cache */

   /* ('ult', 0, 'a(is_gt_zero)') => True */
   /* search1026_0 -> 61 in the cache */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      25,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 61, 2930 },
      -1,
   } },

   /* replace1026 -> 1192 in the cache */

   /* ('ult', 'a', 0) => False */
   /* search1027_0 -> 32 in the cache */
   /* search1027_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* replace1027 -> 1039 in the cache */

   /* ('unpack_64_2x32_split_x', ('pack_64_2x32_split', 'a', 'b')) => a */
   /* search1028_0_0 -> 16 in the cache */
   /* search1028_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1028 -> 16 in the cache */

   /* ('unpack_64_2x32_split_y', ('pack_64_2x32_split', 'a', 'b')) => b */
   /* search1029_0_0 -> 16 in the cache */
   /* search1029_0_1 -> 36 in the cache */
   /* search1029_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1029 -> 36 in the cache */

   /* ('unpack_64_2x32_split_x', ('pack_64_2x32', 'a')) => a.x */
   /* search1030_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 2936 },
      -1,
   } },

   /* replace1030 -> 2351 in the cache */

   /* ('unpack_64_2x32_split_y', ('pack_64_2x32', 'a')) => a.y */
   /* search1031_0_0 -> 16 in the cache */
   /* search1031_0 -> 2936 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 2936 },
      -1,
   } },

   /* replace1031 -> 2354 in the cache */

   /* ('unpack_64_2x32', ('pack_64_2x32_split', 'a', 'b')) => ('vec2', 'a', 'b') */
   /* search1032_0_0 -> 16 in the cache */
   /* search1032_0_1 -> 36 in the cache */
   /* search1032_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1032_0 -> 16 in the cache */
   /* replace1032_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('unpack_64_2x32', ('pack_64_2x32', 'a')) => a */
   /* search1033_0_0 -> 16 in the cache */
   /* search1033_0 -> 2936 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 2936 },
      -1,
   } },

   /* replace1033 -> 16 in the cache */

   /* ('unpack_double_2x32_dxil', ('pack_double_2x32_dxil', 'a')) => a */
   /* search1034_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_double_2x32_dxil,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_double_2x32_dxil,
      -1, 0,
      { 2942 },
      -1,
   } },

   /* replace1034 -> 16 in the cache */

   /* ('pack_64_2x32_split', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')) => a */
   /* search1035_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 0,
      { 21 },
      -1,
   } },
   /* search1035_1_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 0,
      { 2944, 2945 },
      -1,
   } },

   /* replace1035 -> 21 in the cache */

   /* ('pack_64_2x32', ('vec2', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a'))) => a */
   /* search1036_0_0_0 -> 21 in the cache */
   /* search1036_0_0 -> 2944 in the cache */
   /* search1036_0_1_0 -> 21 in the cache */
   /* search1036_0_1 -> 2945 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 2944, 2945 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 2947 },
      -1,
   } },

   /* replace1036 -> 21 in the cache */

   /* ('pack_64_2x32', ('unpack_64_2x32', 'a')) => a */
   /* search1037_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32,
      -1, 0,
      { 2949 },
      -1,
   } },

   /* replace1037 -> 21 in the cache */

   /* ('pack_double_2x32_dxil', ('unpack_double_2x32_dxil', 'a')) => a */
   /* search1038_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_double_2x32_dxil,
      -1, 0,
      { 21 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_double_2x32_dxil,
      -1, 0,
      { 2951 },
      -1,
   } },

   /* replace1038 -> 21 in the cache */

   /* ('iand', ('ieq', ('unpack_32_2x16_split_x', 'a'), '#b'), ('ieq', ('unpack_32_2x16_split_y', 'a'), '#c')) => ('ieq', 'a', ('pack_32_2x16_split', 'b', 'c')) */
   /* search1039_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1039_0_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 2953, 1959 },
      -1,
   } },
   /* search1039_1_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_y,
      -1, 0,
      { 16 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 2955, 2956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 2954, 2957 },
      -1,
   } },

   /* replace1039_0 -> 16 in the cache */
   /* replace1039_1_0 -> 247 in the cache */
   /* replace1039_1_1 -> 454 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 247, 454 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 2959 },
      -1,
   } },

   /* ('ushr', 'a@16', 8) => ('extract_u8', 'a', 1) */
   /* search1040_0 -> 11 in the cache */
   /* search1040_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 11, 793 },
      -1,
   } },

   /* replace1040_0 -> 11 in the cache */
   /* replace1040_1 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 1396 },
      -1,
   } },

   /* ('ushr', 'a@32', 24) => ('extract_u8', 'a', 3) */
   /* search1041_0 -> 16 in the cache */
   /* search1041_1 -> 974 in the cache */
   /* search1041 -> 979 in the cache */

   /* replace1041_0 -> 16 in the cache */
   /* replace1041_1 -> 390 in the cache */
   /* replace1041 -> 407 in the cache */

   /* ('ushr', 'a@64', 56) => ('extract_u8', 'a', 7) */
   /* search1042_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x38 /* 56 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 2963 },
      -1,
   } },

   /* replace1042_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7 /* 7 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 2965 },
      -1,
   } },

   /* ('ishr', 'a@16', 8) => ('extract_i8', 'a', 1) */
   /* search1043_0 -> 11 in the cache */
   /* search1043_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 11, 793 },
      -1,
   } },

   /* replace1043_0 -> 11 in the cache */
   /* replace1043_1 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 1396 },
      -1,
   } },

   /* ('ishr', 'a@32', 24) => ('extract_i8', 'a', 3) */
   /* search1044_0 -> 16 in the cache */
   /* search1044_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 974 },
      -1,
   } },

   /* replace1044_0 -> 16 in the cache */
   /* replace1044_1 -> 390 in the cache */
   /* replace1044 -> 391 in the cache */

   /* ('ishr', 'a@64', 56) => ('extract_i8', 'a', 7) */
   /* search1045_0 -> 21 in the cache */
   /* search1045_1 -> 2963 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 2963 },
      -1,
   } },

   /* replace1045_0 -> 21 in the cache */
   /* replace1045_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 2965 },
      -1,
   } },

   /* ('iand', 255, 'a') => ('extract_u8', 'a', 0) */
   /* search1046_0 -> 1534 in the cache */
   /* search1046_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1534, 32 },
      -1,
   } },

   /* replace1046_0 -> 32 in the cache */
   /* replace1046_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('u2u8', ('extract_u16', 'a', 1)) => ('u2u8', ('extract_u8', 'a', 2)) */
   /* search1047_0_0 -> 32 in the cache */
   /* search1047_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2974 },
      -1,
   } },

   /* replace1047_0_0 -> 32 in the cache */
   /* replace1047_0_1 -> 2186 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 2186 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2976 },
      -1,
   } },

   /* ('u2u8', ('ushr', 'a', 8)) => ('u2u8', ('extract_u8', 'a', 1)) */
   /* search1048_0_0 -> 32 in the cache */
   /* search1048_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 32, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2978 },
      -1,
   } },

   /* replace1048_0_0 -> 32 in the cache */
   /* replace1048_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 32, 35 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2980 },
      -1,
   } },

   /* ('i2i16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('i2i16', ('extract_i8', 'a', 'b')) */
   /* search1049_0_0_0 -> 0 in the cache */
   /* search1049_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2982 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2983 },
      -1,
   } },

   /* replace1049_0_0 -> 0 in the cache */
   /* replace1049_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2985 },
      -1,
   } },

   /* ('u2u16', ('u2u8', ('extract_u8', 'a', 'b'))) => ('u2u16', ('extract_u8', 'a', 'b')) */
   /* search1050_0_0_0 -> 0 in the cache */
   /* search1050_0_0_1 -> 3 in the cache */
   /* search1050_0_0 -> 2982 in the cache */
   /* search1050_0 -> 2983 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2983 },
      -1,
   } },

   /* replace1050_0_0 -> 0 in the cache */
   /* replace1050_0_1 -> 3 in the cache */
   /* replace1050_0 -> 2982 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2982 },
      -1,
   } },

   /* ('ubfe', 'a', 0, 8) => ('extract_u8', 'a', 0) */
   /* search1051_0 -> 16 in the cache */
   /* search1051_1 -> 124 in the cache */
   /* search1051_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 124, 793 },
      -1,
   } },

   /* replace1051_0 -> 16 in the cache */
   /* replace1051_1 -> 124 in the cache */
   /* replace1051 -> 397 in the cache */

   /* ('ubfe', 'a', 8, 8) => ('extract_u8', 'a', 1) */
   /* search1052_0 -> 16 in the cache */
   /* search1052_1 -> 793 in the cache */
   /* search1052_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 793, 793 },
      -1,
   } },

   /* replace1052_0 -> 16 in the cache */
   /* replace1052_1 -> 381 in the cache */
   /* replace1052 -> 400 in the cache */

   /* ('ubfe', 'a', 16, 8) => ('extract_u8', 'a', 2) */
   /* search1053_0 -> 16 in the cache */
   /* search1053_1 -> 815 in the cache */
   /* search1053_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 815, 793 },
      -1,
   } },

   /* replace1053_0 -> 16 in the cache */
   /* replace1053_1 -> 386 in the cache */
   /* replace1053 -> 404 in the cache */

   /* ('ubfe', 'a', 24, 8) => ('extract_u8', 'a', 3) */
   /* search1054_0 -> 16 in the cache */
   /* search1054_1 -> 974 in the cache */
   /* search1054_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 974, 793 },
      -1,
   } },

   /* replace1054_0 -> 16 in the cache */
   /* replace1054_1 -> 390 in the cache */
   /* replace1054 -> 407 in the cache */

   /* ('ibfe', 'a', 0, 8) => ('extract_i8', 'a', 0) */
   /* search1055_0 -> 16 in the cache */
   /* search1055_1 -> 124 in the cache */
   /* search1055_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 124, 793 },
      -1,
   } },

   /* replace1055_0 -> 16 in the cache */
   /* replace1055_1 -> 124 in the cache */
   /* replace1055 -> 378 in the cache */

   /* ('ibfe', 'a', 8, 8) => ('extract_i8', 'a', 1) */
   /* search1056_0 -> 16 in the cache */
   /* search1056_1 -> 793 in the cache */
   /* search1056_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 793, 793 },
      -1,
   } },

   /* replace1056_0 -> 16 in the cache */
   /* replace1056_1 -> 381 in the cache */
   /* replace1056 -> 382 in the cache */

   /* ('ibfe', 'a', 16, 8) => ('extract_i8', 'a', 2) */
   /* search1057_0 -> 16 in the cache */
   /* search1057_1 -> 815 in the cache */
   /* search1057_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 815, 793 },
      -1,
   } },

   /* replace1057_0 -> 16 in the cache */
   /* replace1057_1 -> 386 in the cache */
   /* replace1057 -> 387 in the cache */

   /* ('ibfe', 'a', 24, 8) => ('extract_i8', 'a', 3) */
   /* search1058_0 -> 16 in the cache */
   /* search1058_1 -> 974 in the cache */
   /* search1058_2 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 974, 793 },
      -1,
   } },

   /* replace1058_0 -> 16 in the cache */
   /* replace1058_1 -> 390 in the cache */
   /* replace1058 -> 391 in the cache */

   /* ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search1059_0_0 -> 0 in the cache */
   /* search1059_0_1 -> 3 in the cache */
   /* search1059_0 -> 2985 in the cache */
   /* search1059_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2985, 86 },
      -1,
   } },

   /* replace1059_0 -> 0 in the cache */
   /* replace1059_1 -> 3 in the cache */
   /* replace1059 -> 2982 in the cache */

   /* ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search1060_0_0 -> 0 in the cache */
   /* search1060_0_1 -> 3 in the cache */
   /* search1060_0 -> 2982 in the cache */
   /* search1060_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2982, 86 },
      -1,
   } },

   /* replace1060_0 -> 0 in the cache */
   /* replace1060_1 -> 3 in the cache */
   /* replace1060 -> 2982 in the cache */

   /* ('ushr', ('ishl', 'a@32', 16), 16) => ('extract_u16', 'a', 0) */
   /* search1061_0_0 -> 16 in the cache */
   /* search1061_0_1 -> 815 in the cache */
   /* search1061_0 -> 967 in the cache */
   /* search1061_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 967, 815 },
      -1,
   } },

   /* replace1061_0 -> 16 in the cache */
   /* replace1061_1 -> 124 in the cache */
   /* replace1061 -> 429 in the cache */

   /* ('ushr', 'a@32', 16) => ('extract_u16', 'a', 1) */
   /* search1062_0 -> 16 in the cache */
   /* search1062_1 -> 815 in the cache */
   /* search1062 -> 970 in the cache */

   /* replace1062_0 -> 16 in the cache */
   /* replace1062_1 -> 381 in the cache */
   /* replace1062 -> 432 in the cache */

   /* ('ishr', ('ishl', 'a@32', 16), 16) => ('extract_i16', 'a', 0) */
   /* search1063_0_0 -> 16 in the cache */
   /* search1063_0_1 -> 815 in the cache */
   /* search1063_0 -> 967 in the cache */
   /* search1063_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 967, 815 },
      -1,
   } },

   /* replace1063_0 -> 16 in the cache */
   /* replace1063_1 -> 124 in the cache */
   /* replace1063 -> 421 in the cache */

   /* ('ishr', 'a@32', 16) => ('extract_i16', 'a', 1) */
   /* search1064_0 -> 16 in the cache */
   /* search1064_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 815 },
      -1,
   } },

   /* replace1064_0 -> 16 in the cache */
   /* replace1064_1 -> 381 in the cache */
   /* replace1064 -> 424 in the cache */

   /* ('iand', 65535, 'a') => ('extract_u16', 'a', 0) */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* search1065_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3002, 32 },
      -1,
   } },

   /* replace1065_0 -> 32 in the cache */
   /* replace1065_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 32, 61 },
      -1,
   } },

   /* ('ubfe', 'a', 0, 16) => ('extract_u16', 'a', 0) */
   /* search1066_0 -> 16 in the cache */
   /* search1066_1 -> 124 in the cache */
   /* search1066_2 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 124, 815 },
      -1,
   } },

   /* replace1066_0 -> 16 in the cache */
   /* replace1066_1 -> 124 in the cache */
   /* replace1066 -> 429 in the cache */

   /* ('ubfe', 'a', 16, 16) => ('extract_u16', 'a', 1) */
   /* search1067_0 -> 16 in the cache */
   /* search1067_1 -> 815 in the cache */
   /* search1067_2 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 815, 815 },
      -1,
   } },

   /* replace1067_0 -> 16 in the cache */
   /* replace1067_1 -> 381 in the cache */
   /* replace1067 -> 432 in the cache */

   /* ('ibfe', 'a', 0, 16) => ('extract_i16', 'a', 0) */
   /* search1068_0 -> 16 in the cache */
   /* search1068_1 -> 124 in the cache */
   /* search1068_2 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 124, 815 },
      -1,
   } },

   /* replace1068_0 -> 16 in the cache */
   /* replace1068_1 -> 124 in the cache */
   /* replace1068 -> 421 in the cache */

   /* ('ibfe', 'a', 16, 16) => ('extract_i16', 'a', 1) */
   /* search1069_0 -> 16 in the cache */
   /* search1069_1 -> 815 in the cache */
   /* search1069_2 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 815, 815 },
      -1,
   } },

   /* replace1069_0 -> 16 in the cache */
   /* replace1069_1 -> 381 in the cache */
   /* replace1069 -> 424 in the cache */

   /* ('ior', ('ishl', ('u2u32', 'a@8'), 24), ('ior', ('ishl', ('u2u32', 'b@8'), 16), ('ior', ('ishl', ('u2u32', 'c@8'), 8), ('u2u32', 'd@8')))) => ('pack_32_4x8', ('vec4', 'd', 'c', 'b', 'a')) */
   /* search1070_0_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 6 },
      -1,
   } },
   /* search1070_0_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3009, 974 },
      -1,
   } },
   /* search1070_1_0_0_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2060 },
      -1,
   } },
   /* search1070_1_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3011, 815 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3013 },
      -1,
   } },
   /* search1070_1_1_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3014, 793 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3016 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 3015, 3017 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 2,
      { 3012, 3018 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3010, 3019 },
      -1,
   } },

   /* replace1070_0_0 -> 3016 in the cache */
   /* replace1070_0_1 -> 3013 in the cache */
   /* replace1070_0_2 -> 2060 in the cache */
   /* replace1070_0_3 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3016, 3013, 2060, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_4x8,
      -1, 0,
      { 3021 },
      -1,
   } },

   /* ('extract_u16', ('extract_i16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b') */
   /* search1071_0_0 -> 0 in the cache */
   /* search1071_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 0, 3 },
      -1,
   } },
   /* search1071_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3023, 86 },
      -1,
   } },

   /* replace1071_0 -> 0 in the cache */
   /* replace1071_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('extract_u16', ('extract_u16', 'a', 'b'), 0) => ('extract_u16', 'a', 'b') */
   /* search1072_0_0 -> 0 in the cache */
   /* search1072_0_1 -> 3 in the cache */
   /* search1072_0 -> 3025 in the cache */
   /* search1072_1 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3025, 86 },
      -1,
   } },

   /* replace1072_0 -> 0 in the cache */
   /* replace1072_1 -> 3 in the cache */
   /* replace1072 -> 3025 in the cache */

   /* ('pack_64_2x32_split', 'a', 'b') => ('ior', ('u2u64', 'a'), ('ishl', ('u2u64', 'b'), 32)) */
   /* search1073_0 -> 16 in the cache */
   /* search1073_1 -> 36 in the cache */
   /* search1073 -> 2933 in the cache */

   /* replace1073_0_0 -> 16 in the cache */
   /* replace1073_0 -> 1856 in the cache */
   /* replace1073_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 36 },
      -1,
   } },
   /* replace1073_1_1 -> 836 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3027, 836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1856, 3028 },
      -1,
   } },

   /* ('pack_32_2x16_split', 'a', 'b') => ('ior', ('u2u32', 'a'), ('ishl', ('u2u32', 'b'), 16)) */
   /* search1074_0 -> 11 in the cache */
   /* search1074_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_2x16_split,
      -1, 0,
      { 11, 247 },
      -1,
   } },

   /* replace1074_0_0 -> 11 in the cache */
   /* replace1074_0 -> 1749 in the cache */
   /* replace1074_1_0_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 247 },
      -1,
   } },
   /* replace1074_1_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3031, 815 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1749, 3032 },
      -1,
   } },

   /* ('unpack_64_2x32_split_x', 'a') => ('u2u32', 'a') */
   /* search1075_0 -> 21 in the cache */
   /* search1075 -> 2944 in the cache */

   /* replace1075_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 21 },
      -1,
   } },

   /* ('unpack_64_2x32_split_y', 'a') => ('u2u32', ('ushr', 'a', 32)) */
   /* search1076_0 -> 21 in the cache */
   /* search1076 -> 2945 in the cache */

   /* replace1076_0_0 -> 21 in the cache */
   /* replace1076_0_1 -> 836 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 836 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3035 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', 'a') => ('u2u16', 'a') */
   /* search1077_0 -> 16 in the cache */
   /* search1077 -> 2953 in the cache */

   /* replace1077_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_y', 'a') => ('u2u16', ('ushr', 'a', 16)) */
   /* search1078_0 -> 16 in the cache */
   /* search1078 -> 2955 in the cache */

   /* replace1078_0_0 -> 16 in the cache */
   /* replace1078_0_1 -> 815 in the cache */
   /* replace1078_0 -> 970 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 970 },
      -1,
   } },

   /* ('unpack_half_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_half_2x16_split_x', 'a') */
   /* search1079_0_0 -> 16 in the cache */
   /* search1079_0_1 -> 964 in the cache */
   /* search1079_0 -> 965 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 1,
      { 965 },
      -1,
   } },

   /* replace1079_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', ('iand', 'a', 65535)) => ('unpack_32_2x16_split_x', 'a') */
   /* search1080_0_0 -> 16 in the cache */
   /* search1080_0_1 -> 964 in the cache */
   /* search1080_0 -> 965 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 1,
      { 965 },
      -1,
   } },

   /* replace1080_0 -> 16 in the cache */
   /* replace1080 -> 2953 in the cache */

   /* ('unpack_64_2x32_split_x', ('iand', 'a', 4294967295)) => ('unpack_64_2x32_split_x', 'a') */
   /* search1081_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 3042 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_x,
      -1, 1,
      { 3043 },
      -1,
   } },

   /* replace1081_0 -> 21 in the cache */
   /* replace1081 -> 2944 in the cache */

   /* ('unpack_half_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1082_0_0 -> 16 in the cache */
   /* search1082_0_1 -> 968 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 968 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 1,
      { 3045 },
      -1,
   } },

   /* replace1082_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_y,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('unpack_32_2x16_split_y', ('iand', 'a', 4294901760)) => ('unpack_32_2x16_split_y', 'a') */
   /* search1083_0_0 -> 16 in the cache */
   /* search1083_0_1 -> 968 in the cache */
   /* search1083_0 -> 3045 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_y,
      -1, 1,
      { 3045 },
      -1,
   } },

   /* replace1083_0 -> 16 in the cache */
   /* replace1083 -> 2955 in the cache */

   /* ('unpack_64_2x32_split_y', ('iand', 'a', 18446744069414584320)) => ('unpack_64_2x32_split_y', 'a') */
   /* search1084_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffffffff00000000 /* 18446744069414584320 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 3049 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 1,
      { 3050 },
      -1,
   } },

   /* replace1084_0 -> 21 in the cache */
   /* replace1084 -> 2945 in the cache */

   /* ('unpack_half_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_half_2x16_split_x', 'a') */
   /* search1085_0_0 -> 16 in the cache */
   /* search1085_0_1 -> 124 in the cache */
   /* search1085_0 -> 429 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 429 },
      -1,
   } },

   /* replace1085_0 -> 16 in the cache */
   /* replace1085 -> 3040 in the cache */

   /* ('unpack_half_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1086_0_0 -> 16 in the cache */
   /* search1086_0_1 -> 381 in the cache */
   /* search1086_0 -> 432 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 432 },
      -1,
   } },

   /* replace1086_0 -> 16 in the cache */
   /* replace1086 -> 3047 in the cache */

   /* ('unpack_half_2x16_split_x', ('ushr', 'a', 16)) => ('unpack_half_2x16_split_y', 'a') */
   /* search1087_0_0 -> 16 in the cache */
   /* search1087_0_1 -> 815 in the cache */
   /* search1087_0 -> 970 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_half_2x16_split_x,
      -1, 0,
      { 970 },
      -1,
   } },

   /* replace1087_0 -> 16 in the cache */
   /* replace1087 -> 3047 in the cache */

   /* ('unpack_32_2x16_split_x', ('extract_u16', 'a', 0)) => ('unpack_32_2x16_split_x', 'a') */
   /* search1088_0_0 -> 16 in the cache */
   /* search1088_0_1 -> 124 in the cache */
   /* search1088_0 -> 429 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 429 },
      -1,
   } },

   /* replace1088_0 -> 16 in the cache */
   /* replace1088 -> 2953 in the cache */

   /* ('unpack_32_2x16_split_x', ('extract_u16', 'a', 1)) => ('unpack_32_2x16_split_y', 'a') */
   /* search1089_0_0 -> 16 in the cache */
   /* search1089_0_1 -> 381 in the cache */
   /* search1089_0 -> 432 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_unpack_32_2x16_split_x,
      -1, 0,
      { 432 },
      -1,
   } },

   /* replace1089_0 -> 16 in the cache */
   /* replace1089 -> 2955 in the cache */

   /* ('ishl', ('pack_half_2x16', ('vec2', 'a', 0)), 16) => ('pack_half_2x16', ('vec2', 0, 'a')) */
   /* search1090_0_0_0 -> 16 in the cache */
   /* search1090_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3057 },
      -1,
   } },
   /* search1090_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3058, 815 },
      -1,
   } },

   /* replace1090_0_0 -> 124 in the cache */
   /* replace1090_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3060 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16', ('vec2', 0, 'a')), 16) => ('pack_half_2x16', ('vec2', 'a', 0)) */
   /* search1091_0_0_0 -> 124 in the cache */
   /* search1091_0_0_1 -> 16 in the cache */
   /* search1091_0_0 -> 3060 in the cache */
   /* search1091_0 -> 3061 in the cache */
   /* search1091_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3061, 815 },
      -1,
   } },

   /* replace1091_0_0 -> 16 in the cache */
   /* replace1091_0_1 -> 124 in the cache */
   /* replace1091_0 -> 3057 in the cache */
   /* replace1091 -> 3058 in the cache */

   /* ('iadd', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b')) */
   /* search1092_0_0_0 -> 16 in the cache */
   /* search1092_0_0_1 -> 124 in the cache */
   /* search1092_0_0 -> 3057 in the cache */
   /* search1092_0 -> 3058 in the cache */
   /* search1092_1_0_0 -> 124 in the cache */
   /* search1092_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 124, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 3063 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3058, 3064 },
      -1,
   } },

   /* replace1092_0_0 -> 16 in the cache */
   /* replace1092_0_1 -> 36 in the cache */
   /* replace1092_0 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16,
      -1, 0,
      { 2940 },
      -1,
   } },

   /* ('ior', ('pack_half_2x16', ('vec2', 'a', 0)), ('pack_half_2x16', ('vec2', 0, 'b'))) => ('pack_half_2x16', ('vec2', 'a', 'b')) */
   /* search1093_0_0_0 -> 16 in the cache */
   /* search1093_0_0_1 -> 124 in the cache */
   /* search1093_0_0 -> 3057 in the cache */
   /* search1093_0 -> 3058 in the cache */
   /* search1093_1_0_0 -> 124 in the cache */
   /* search1093_1_0_1 -> 36 in the cache */
   /* search1093_1_0 -> 3063 in the cache */
   /* search1093_1 -> 3064 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3058, 3064 },
      -1,
   } },

   /* replace1093_0_0 -> 16 in the cache */
   /* replace1093_0_1 -> 36 in the cache */
   /* replace1093_0 -> 2940 in the cache */
   /* replace1093 -> 3066 in the cache */

   /* ('ishl', ('pack_half_2x16_split', 'a', 0), 16) => ('pack_half_2x16_split', 0, 'a') */
   /* search1094_0_0 -> 16 in the cache */
   /* search1094_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* search1094_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3068, 815 },
      -1,
   } },

   /* replace1094_0 -> 124 in the cache */
   /* replace1094_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 124, 16 },
      -1,
   } },

   /* ('ushr', ('pack_half_2x16_split', 0, 'a'), 16) => ('pack_half_2x16_split', 'a', 0) */
   /* search1095_0_0 -> 124 in the cache */
   /* search1095_0_1 -> 16 in the cache */
   /* search1095_0 -> 3070 in the cache */
   /* search1095_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3070, 815 },
      -1,
   } },

   /* replace1095_0 -> 16 in the cache */
   /* replace1095_1 -> 124 in the cache */
   /* replace1095 -> 3068 in the cache */

   /* ('extract_u16', ('pack_half_2x16_split', 0, 'a'), 1) => ('pack_half_2x16_split', 'a', 0) */
   /* search1096_0_0 -> 124 in the cache */
   /* search1096_0_1 -> 16 in the cache */
   /* search1096_0 -> 3070 in the cache */
   /* search1096_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3070, 381 },
      -1,
   } },

   /* replace1096_0 -> 16 in the cache */
   /* replace1096_1 -> 124 in the cache */
   /* replace1096 -> 3068 in the cache */

   /* ('iadd', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b') */
   /* search1097_0_0 -> 16 in the cache */
   /* search1097_0_1 -> 124 in the cache */
   /* search1097_0 -> 3068 in the cache */
   /* search1097_1_0 -> 124 in the cache */
   /* search1097_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 124, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3068, 3073 },
      -1,
   } },

   /* replace1097_0 -> 16 in the cache */
   /* replace1097_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_half_2x16_split,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* ('ior', ('pack_half_2x16_split', 'a', 0), ('pack_half_2x16_split', 0, 'b')) => ('pack_half_2x16_split', 'a', 'b') */
   /* search1098_0_0 -> 16 in the cache */
   /* search1098_0_1 -> 124 in the cache */
   /* search1098_0 -> 3068 in the cache */
   /* search1098_1_0 -> 124 in the cache */
   /* search1098_1_1 -> 36 in the cache */
   /* search1098_1 -> 3073 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3068, 3073 },
      -1,
   } },

   /* replace1098_0 -> 16 in the cache */
   /* replace1098_1 -> 36 in the cache */
   /* replace1098 -> 3075 in the cache */

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('i2i', 'a') */
   /* search1099_0_0 -> 6 in the cache */
   /* search1099_0_1 -> 2060 in the cache */
   /* search1099_0_2 -> 3013 in the cache */
   /* search1099_0_3 -> 3016 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_32_4x8_split,
      -1, 0,
      { 6, 2060, 3013, 3016 },
      -1,
   } },
   /* search1099_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3077, 124 },
      -1,
   } },

   /* replace1099_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('i2i', 'b') */
   /* search1100_0_0 -> 6 in the cache */
   /* search1100_0_1 -> 2060 in the cache */
   /* search1100_0_2 -> 3013 in the cache */
   /* search1100_0_3 -> 3016 in the cache */
   /* search1100_0 -> 3077 in the cache */
   /* search1100_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3077, 381 },
      -1,
   } },

   /* replace1100_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 2060 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('i2i', 'c') */
   /* search1101_0_0 -> 6 in the cache */
   /* search1101_0_1 -> 2060 in the cache */
   /* search1101_0_2 -> 3013 in the cache */
   /* search1101_0_3 -> 3016 in the cache */
   /* search1101_0 -> 3077 in the cache */
   /* search1101_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3077, 386 },
      -1,
   } },

   /* replace1101_0 -> 3013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3013 },
      -1,
   } },

   /* ('extract_i8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('i2i', 'd') */
   /* search1102_0_0 -> 6 in the cache */
   /* search1102_0_1 -> 2060 in the cache */
   /* search1102_0_2 -> 3013 in the cache */
   /* search1102_0_3 -> 3016 in the cache */
   /* search1102_0 -> 3077 in the cache */
   /* search1102_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3077, 390 },
      -1,
   } },

   /* replace1102_0 -> 3016 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 3016 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 0) => ('u2u', 'a') */
   /* search1103_0_0 -> 6 in the cache */
   /* search1103_0_1 -> 2060 in the cache */
   /* search1103_0_2 -> 3013 in the cache */
   /* search1103_0_3 -> 3016 in the cache */
   /* search1103_0 -> 3077 in the cache */
   /* search1103_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3077, 124 },
      -1,
   } },

   /* replace1103_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 1) => ('u2u', 'b') */
   /* search1104_0_0 -> 6 in the cache */
   /* search1104_0_1 -> 2060 in the cache */
   /* search1104_0_2 -> 3013 in the cache */
   /* search1104_0_3 -> 3016 in the cache */
   /* search1104_0 -> 3077 in the cache */
   /* search1104_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3077, 381 },
      -1,
   } },

   /* replace1104_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 2060 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 2) => ('u2u', 'c') */
   /* search1105_0_0 -> 6 in the cache */
   /* search1105_0_1 -> 2060 in the cache */
   /* search1105_0_2 -> 3013 in the cache */
   /* search1105_0_3 -> 3016 in the cache */
   /* search1105_0 -> 3077 in the cache */
   /* search1105_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3077, 386 },
      -1,
   } },

   /* replace1105_0 -> 3013 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3013 },
      -1,
   } },

   /* ('extract_u8', ('pack_32_4x8_split', 'a', 'b', 'c', 'd'), 3) => ('u2u', 'd') */
   /* search1106_0_0 -> 6 in the cache */
   /* search1106_0_1 -> 2060 in the cache */
   /* search1106_0_2 -> 3013 in the cache */
   /* search1106_0_3 -> 3016 in the cache */
   /* search1106_0 -> 3077 in the cache */
   /* search1106_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3077, 390 },
      -1,
   } },

   /* replace1106_0 -> 3016 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 3016 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@16', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1107_0_0 -> 11 in the cache */
   /* search1107_0_1 -> 793 in the cache */
   /* search1107_0 -> 2961 in the cache */
   /* search1107_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2961, 818 },
      -1,
   } },

   /* replace1107_0 -> 11 in the cache */
   /* replace1107_1 -> 1396 in the cache */
   /* replace1107 -> 2962 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1108_0_0 -> 16 in the cache */
   /* search1108_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 16, 793 },
      -1,
   } },
   /* search1108_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3095, 124 },
      -1,
   } },

   /* replace1108_0 -> 16 in the cache */
   /* replace1108_1 -> 381 in the cache */
   /* replace1108 -> 400 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1109_0_0 -> 16 in the cache */
   /* search1109_0_1 -> 815 in the cache */
   /* search1109_0 -> 970 in the cache */
   /* search1109_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 970, 124 },
      -1,
   } },

   /* replace1109_0 -> 16 in the cache */
   /* replace1109_1 -> 386 in the cache */
   /* replace1109 -> 404 in the cache */

   /* ('extract_u8', ('ushr', 'a@32', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1110_0_0 -> 16 in the cache */
   /* search1110_0_1 -> 974 in the cache */
   /* search1110_0 -> 979 in the cache */
   /* search1110_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 979, 124 },
      -1,
   } },

   /* replace1110_0 -> 16 in the cache */
   /* replace1110_1 -> 390 in the cache */
   /* replace1110 -> 407 in the cache */

   /* ('extract_u8', ('ushr', 'a@64', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1111_0_0 -> 21 in the cache */
   /* search1111_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 793 },
      -1,
   } },
   /* search1111_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3099, 859 },
      -1,
   } },

   /* replace1111_0 -> 21 in the cache */
   /* replace1111_1 -> 1438 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 1438 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1112_0_0 -> 21 in the cache */
   /* search1112_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 815 },
      -1,
   } },
   /* search1112_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3102, 859 },
      -1,
   } },

   /* replace1112_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3104 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1113_0_0 -> 21 in the cache */
   /* search1113_0_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 974 },
      -1,
   } },
   /* search1113_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3106, 859 },
      -1,
   } },

   /* replace1113_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3108 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 32), 0) => ('extract_u8', 'a', 4) */
   /* search1114_0_0 -> 21 in the cache */
   /* search1114_0_1 -> 836 in the cache */
   /* search1114_0 -> 3035 in the cache */
   /* search1114_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3035, 859 },
      -1,
   } },

   /* replace1114_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x4 /* 4 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3111 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 40), 0) => ('extract_u8', 'a', 5) */
   /* search1115_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x28 /* 40 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3113 },
      -1,
   } },
   /* search1115_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3114, 859 },
      -1,
   } },

   /* replace1115_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x5 /* 5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3116 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 48), 0) => ('extract_u8', 'a', 6) */
   /* search1116_0_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x30 /* 48 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 21, 3118 },
      -1,
   } },
   /* search1116_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3119, 859 },
      -1,
   } },

   /* replace1116_0 -> 21 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x6 /* 6 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 3121 },
      -1,
   } },

   /* ('extract_u8', ('ushr', 'a@64', 56), 0) => ('extract_u8', 'a', 7) */
   /* search1117_0_0 -> 21 in the cache */
   /* search1117_0_1 -> 2963 in the cache */
   /* search1117_0 -> 2964 in the cache */
   /* search1117_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2964, 859 },
      -1,
   } },

   /* replace1117_0 -> 21 in the cache */
   /* replace1117_1 -> 2965 in the cache */
   /* replace1117 -> 2966 in the cache */

   /* ('extract_u8', ('ishr', 'a@16', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1118_0_0 -> 11 in the cache */
   /* search1118_0_1 -> 793 in the cache */
   /* search1118_0 -> 2967 in the cache */
   /* search1118_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2967, 818 },
      -1,
   } },

   /* replace1118_0 -> 11 in the cache */
   /* replace1118_1 -> 1396 in the cache */
   /* replace1118 -> 2962 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1119_0_0 -> 16 in the cache */
   /* search1119_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 16, 793 },
      -1,
   } },
   /* search1119_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3125, 124 },
      -1,
   } },

   /* replace1119_0 -> 16 in the cache */
   /* replace1119_1 -> 381 in the cache */
   /* replace1119 -> 400 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1120_0_0 -> 16 in the cache */
   /* search1120_0_1 -> 815 in the cache */
   /* search1120_0 -> 3001 in the cache */
   /* search1120_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3001, 124 },
      -1,
   } },

   /* replace1120_0 -> 16 in the cache */
   /* replace1120_1 -> 386 in the cache */
   /* replace1120 -> 404 in the cache */

   /* ('extract_u8', ('ishr', 'a@32', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1121_0_0 -> 16 in the cache */
   /* search1121_0_1 -> 974 in the cache */
   /* search1121_0 -> 2969 in the cache */
   /* search1121_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2969, 124 },
      -1,
   } },

   /* replace1121_0 -> 16 in the cache */
   /* replace1121_1 -> 390 in the cache */
   /* replace1121 -> 407 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 8), 0) => ('extract_u8', 'a', 1) */
   /* search1122_0_0 -> 21 in the cache */
   /* search1122_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 793 },
      -1,
   } },
   /* search1122_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3129, 859 },
      -1,
   } },

   /* replace1122_0 -> 21 in the cache */
   /* replace1122_1 -> 1438 in the cache */
   /* replace1122 -> 3101 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 16), 0) => ('extract_u8', 'a', 2) */
   /* search1123_0_0 -> 21 in the cache */
   /* search1123_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 815 },
      -1,
   } },
   /* search1123_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3131, 859 },
      -1,
   } },

   /* replace1123_0 -> 21 in the cache */
   /* replace1123_1 -> 3104 in the cache */
   /* replace1123 -> 3105 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 24), 0) => ('extract_u8', 'a', 3) */
   /* search1124_0_0 -> 21 in the cache */
   /* search1124_0_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 974 },
      -1,
   } },
   /* search1124_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3133, 859 },
      -1,
   } },

   /* replace1124_0 -> 21 in the cache */
   /* replace1124_1 -> 3108 in the cache */
   /* replace1124 -> 3109 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 32), 0) => ('extract_u8', 'a', 4) */
   /* search1125_0_0 -> 21 in the cache */
   /* search1125_0_1 -> 836 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 836 },
      -1,
   } },
   /* search1125_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3135, 859 },
      -1,
   } },

   /* replace1125_0 -> 21 in the cache */
   /* replace1125_1 -> 3111 in the cache */
   /* replace1125 -> 3112 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 40), 0) => ('extract_u8', 'a', 5) */
   /* search1126_0_0 -> 21 in the cache */
   /* search1126_0_1 -> 3113 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3113 },
      -1,
   } },
   /* search1126_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3137, 859 },
      -1,
   } },

   /* replace1126_0 -> 21 in the cache */
   /* replace1126_1 -> 3116 in the cache */
   /* replace1126 -> 3117 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 48), 0) => ('extract_u8', 'a', 6) */
   /* search1127_0_0 -> 21 in the cache */
   /* search1127_0_1 -> 3118 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 21, 3118 },
      -1,
   } },
   /* search1127_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3139, 859 },
      -1,
   } },

   /* replace1127_0 -> 21 in the cache */
   /* replace1127_1 -> 3121 in the cache */
   /* replace1127 -> 3122 in the cache */

   /* ('extract_u8', ('ishr', 'a@64', 56), 0) => ('extract_u8', 'a', 7) */
   /* search1128_0_0 -> 21 in the cache */
   /* search1128_0_1 -> 2963 in the cache */
   /* search1128_0 -> 2970 in the cache */
   /* search1128_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2970, 859 },
      -1,
   } },

   /* replace1128_0 -> 21 in the cache */
   /* replace1128_1 -> 2965 in the cache */
   /* replace1128 -> 2966 in the cache */

   /* ('extract_u8', ('extract_u16', 'a', 1), 0) => ('extract_u8', 'a', 2) */
   /* search1129_0_0 -> 32 in the cache */
   /* search1129_0_1 -> 35 in the cache */
   /* search1129_0 -> 2974 in the cache */
   /* search1129_1 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 2974, 61 },
      -1,
   } },

   /* replace1129_0 -> 32 in the cache */
   /* replace1129_1 -> 2186 in the cache */
   /* replace1129 -> 2976 in the cache */

   /* ('extract_u8', ('ishl', 'a@16', 8), 1) => ('extract_u8', 'a', 0) */
   /* search1130_0_0 -> 11 in the cache */
   /* search1130_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 793 },
      -1,
   } },
   /* search1130_1 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3143, 1396 },
      -1,
   } },

   /* replace1130_0 -> 11 in the cache */
   /* replace1130_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 818 },
      -1,
   } },

   /* ('extract_u8', ('ishl', 'a@32', 8), 3) => ('extract_u8', 'a', 2) */
   /* search1131_0_0 -> 16 in the cache */
   /* search1131_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 793 },
      -1,
   } },
   /* search1131_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3146, 390 },
      -1,
   } },

   /* replace1131_0 -> 16 in the cache */
   /* replace1131_1 -> 386 in the cache */
   /* replace1131 -> 404 in the cache */

   /* ('extract_u8', ('ishl', 'a@32', 16), 3) => ('extract_u8', 'a', 1) */
   /* search1132_0_0 -> 16 in the cache */
   /* search1132_0_1 -> 815 in the cache */
   /* search1132_0 -> 967 in the cache */
   /* search1132_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 967, 390 },
      -1,
   } },

   /* replace1132_0 -> 16 in the cache */
   /* replace1132_1 -> 381 in the cache */
   /* replace1132 -> 400 in the cache */

   /* ('extract_u8', ('ishl', 'a@32', 24), 3) => ('extract_u8', 'a', 0) */
   /* search1133_0_0 -> 16 in the cache */
   /* search1133_0_1 -> 974 in the cache */
   /* search1133_0 -> 976 in the cache */
   /* search1133_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 976, 390 },
      -1,
   } },

   /* replace1133_0 -> 16 in the cache */
   /* replace1133_1 -> 124 in the cache */
   /* replace1133 -> 397 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 8), 7) => ('extract_u8', 'a', 6) */
   /* search1134_0_0 -> 21 in the cache */
   /* search1134_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 793 },
      -1,
   } },
   /* search1134_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3150, 2965 },
      -1,
   } },

   /* replace1134_0 -> 21 in the cache */
   /* replace1134_1 -> 3121 in the cache */
   /* replace1134 -> 3122 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 16), 7) => ('extract_u8', 'a', 5) */
   /* search1135_0_0 -> 21 in the cache */
   /* search1135_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 815 },
      -1,
   } },
   /* search1135_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3152, 2965 },
      -1,
   } },

   /* replace1135_0 -> 21 in the cache */
   /* replace1135_1 -> 3116 in the cache */
   /* replace1135 -> 3117 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 24), 7) => ('extract_u8', 'a', 4) */
   /* search1136_0_0 -> 21 in the cache */
   /* search1136_0_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 974 },
      -1,
   } },
   /* search1136_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3154, 2965 },
      -1,
   } },

   /* replace1136_0 -> 21 in the cache */
   /* replace1136_1 -> 3111 in the cache */
   /* replace1136 -> 3112 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 32), 7) => ('extract_u8', 'a', 3) */
   /* search1137_0_0 -> 21 in the cache */
   /* search1137_0_1 -> 836 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 836 },
      -1,
   } },
   /* search1137_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3156, 2965 },
      -1,
   } },

   /* replace1137_0 -> 21 in the cache */
   /* replace1137_1 -> 3108 in the cache */
   /* replace1137 -> 3109 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 40), 7) => ('extract_u8', 'a', 2) */
   /* search1138_0_0 -> 21 in the cache */
   /* search1138_0_1 -> 3113 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3113 },
      -1,
   } },
   /* search1138_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3158, 2965 },
      -1,
   } },

   /* replace1138_0 -> 21 in the cache */
   /* replace1138_1 -> 3104 in the cache */
   /* replace1138 -> 3105 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 48), 7) => ('extract_u8', 'a', 1) */
   /* search1139_0_0 -> 21 in the cache */
   /* search1139_0_1 -> 3118 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 3118 },
      -1,
   } },
   /* search1139_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3160, 2965 },
      -1,
   } },

   /* replace1139_0 -> 21 in the cache */
   /* replace1139_1 -> 1438 in the cache */
   /* replace1139 -> 3101 in the cache */

   /* ('extract_u8', ('ishl', 'a@64', 56), 7) => ('extract_u8', 'a', 0) */
   /* search1140_0_0 -> 21 in the cache */
   /* search1140_0_1 -> 2963 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 2963 },
      -1,
   } },
   /* search1140_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3162, 2965 },
      -1,
   } },

   /* replace1140_0 -> 21 in the cache */
   /* replace1140_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 21, 859 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@16', 8), 1) => ('extract_i8', 'a', 0) */
   /* search1141_0_0 -> 11 in the cache */
   /* search1141_0_1 -> 793 in the cache */
   /* search1141_0 -> 3143 in the cache */
   /* search1141_1 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3143, 1396 },
      -1,
   } },

   /* replace1141_0 -> 11 in the cache */
   /* replace1141_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 818 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@32', 8), 3) => ('extract_i8', 'a', 2) */
   /* search1142_0_0 -> 16 in the cache */
   /* search1142_0_1 -> 793 in the cache */
   /* search1142_0 -> 3146 in the cache */
   /* search1142_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3146, 390 },
      -1,
   } },

   /* replace1142_0 -> 16 in the cache */
   /* replace1142_1 -> 386 in the cache */
   /* replace1142 -> 387 in the cache */

   /* ('extract_i8', ('ishl', 'a@32', 16), 3) => ('extract_i8', 'a', 1) */
   /* search1143_0_0 -> 16 in the cache */
   /* search1143_0_1 -> 815 in the cache */
   /* search1143_0 -> 967 in the cache */
   /* search1143_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 967, 390 },
      -1,
   } },

   /* replace1143_0 -> 16 in the cache */
   /* replace1143_1 -> 381 in the cache */
   /* replace1143 -> 382 in the cache */

   /* ('extract_i8', ('ishl', 'a@32', 24), 3) => ('extract_i8', 'a', 0) */
   /* search1144_0_0 -> 16 in the cache */
   /* search1144_0_1 -> 974 in the cache */
   /* search1144_0 -> 976 in the cache */
   /* search1144_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 976, 390 },
      -1,
   } },

   /* replace1144_0 -> 16 in the cache */
   /* replace1144_1 -> 124 in the cache */
   /* replace1144 -> 378 in the cache */

   /* ('extract_i8', ('ishl', 'a@64', 8), 7) => ('extract_i8', 'a', 6) */
   /* search1145_0_0 -> 21 in the cache */
   /* search1145_0_1 -> 793 in the cache */
   /* search1145_0 -> 3150 in the cache */
   /* search1145_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3150, 2965 },
      -1,
   } },

   /* replace1145_0 -> 21 in the cache */
   /* replace1145_1 -> 3121 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3121 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 16), 7) => ('extract_i8', 'a', 5) */
   /* search1146_0_0 -> 21 in the cache */
   /* search1146_0_1 -> 815 in the cache */
   /* search1146_0 -> 3152 in the cache */
   /* search1146_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3152, 2965 },
      -1,
   } },

   /* replace1146_0 -> 21 in the cache */
   /* replace1146_1 -> 3116 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3116 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 24), 7) => ('extract_i8', 'a', 4) */
   /* search1147_0_0 -> 21 in the cache */
   /* search1147_0_1 -> 974 in the cache */
   /* search1147_0 -> 3154 in the cache */
   /* search1147_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3154, 2965 },
      -1,
   } },

   /* replace1147_0 -> 21 in the cache */
   /* replace1147_1 -> 3111 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3111 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 32), 7) => ('extract_i8', 'a', 3) */
   /* search1148_0_0 -> 21 in the cache */
   /* search1148_0_1 -> 836 in the cache */
   /* search1148_0 -> 3156 in the cache */
   /* search1148_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3156, 2965 },
      -1,
   } },

   /* replace1148_0 -> 21 in the cache */
   /* replace1148_1 -> 3108 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3108 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 40), 7) => ('extract_i8', 'a', 2) */
   /* search1149_0_0 -> 21 in the cache */
   /* search1149_0_1 -> 3113 in the cache */
   /* search1149_0 -> 3158 in the cache */
   /* search1149_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3158, 2965 },
      -1,
   } },

   /* replace1149_0 -> 21 in the cache */
   /* replace1149_1 -> 3104 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 3104 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 48), 7) => ('extract_i8', 'a', 1) */
   /* search1150_0_0 -> 21 in the cache */
   /* search1150_0_1 -> 3118 in the cache */
   /* search1150_0 -> 3160 in the cache */
   /* search1150_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3160, 2965 },
      -1,
   } },

   /* replace1150_0 -> 21 in the cache */
   /* replace1150_1 -> 1438 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 1438 },
      -1,
   } },

   /* ('extract_i8', ('ishl', 'a@64', 56), 7) => ('extract_i8', 'a', 0) */
   /* search1151_0_0 -> 21 in the cache */
   /* search1151_0_1 -> 2963 in the cache */
   /* search1151_0 -> 3162 in the cache */
   /* search1151_1 -> 2965 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3162, 2965 },
      -1,
   } },

   /* replace1151_0 -> 21 in the cache */
   /* replace1151_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 21, 859 },
      -1,
   } },

   /* ('ussub_4x8_vc4', 'a', 0) => a */
   /* search1152_0 -> 16 in the cache */
   /* search1152_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ussub_4x8_vc4,
      -1, 0,
      { 16, 124 },
      -1,
   } },

   /* replace1152 -> 16 in the cache */

   /* ('ussub_4x8_vc4', 'a', -1) => 0 */
   /* search1153_0 -> 16 in the cache */
   /* search1153_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ussub_4x8_vc4,
      -1, 0,
      { 16, 126 },
      -1,
   } },

   /* replace1153 -> 124 in the cache */

   /* ('fsub', 'a', 'b') => ('fadd', 'a', ('fneg', 'b')) */
   /* search1154_0 -> 0 in the cache */
   /* search1154_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1154_0 -> 0 in the cache */
   /* replace1154_1_0 -> 3 in the cache */
   /* replace1154_1 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1049 },
      -1,
   } },

   /* ('isub', 'a', 'b') => ('iadd', 'a', ('ineg', 'b')) */
   /* search1155_0 -> 0 in the cache */
   /* search1155_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1155_0 -> 0 in the cache */
   /* replace1155_1_0 -> 3 in the cache */
   /* replace1155_1 -> 93 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 93 },
      -1,
   } },

   /* ('uabs_usub', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b')) */
   /* search1156_0 -> 0 in the cache */
   /* search1156_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uabs_usub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1156_0_0 -> 0 in the cache */
   /* replace1156_0_1 -> 3 in the cache */
   /* replace1156_0 -> 1029 in the cache */
   /* replace1156_1_0_0 -> 0 in the cache */
   /* replace1156_1_0_1 -> 3 in the cache */
   /* replace1156_1_0 -> 3188 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3188 },
      -1,
   } },
   /* replace1156_2_0 -> 0 in the cache */
   /* replace1156_2_1 -> 3 in the cache */
   /* replace1156_2 -> 3188 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1029, 3191, 3188 },
      -1,
   } },

   /* ('uabs_isub', 'a', 'b') => ('bcsel', ('ilt', 'a', 'b'), ('ineg', ('isub', 'a', 'b')), ('isub', 'a', 'b')) */
   /* search1157_0 -> 0 in the cache */
   /* search1157_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uabs_isub,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1157_0_0 -> 0 in the cache */
   /* replace1157_0_1 -> 3 in the cache */
   /* replace1157_0 -> 1026 in the cache */
   /* replace1157_1_0_0 -> 0 in the cache */
   /* replace1157_1_0_1 -> 3 in the cache */
   /* replace1157_1_0 -> 3188 in the cache */
   /* replace1157_1 -> 3191 in the cache */
   /* replace1157_2_0 -> 0 in the cache */
   /* replace1157_2_1 -> 3 in the cache */
   /* replace1157_2 -> 3188 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1026, 3191, 3188 },
      -1,
   } },

   /* ('fmul(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmul', 'a', 'b')) */
   /* search1158_0_0 -> 0 in the cache */
   /* search1158_0 -> 201 in the cache */
   /* search1158_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 201, 3 },
      4,
   } },

   /* replace1158_0_0 -> 0 in the cache */
   /* replace1158_0_1 -> 3 in the cache */
   /* replace1158_0 -> 253 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 253 },
      -1,
   } },

   /* ('fmulz(is_used_by_non_fsat)', ('fneg', 'a'), 'b') => ('fneg', ('fmulz', 'a', 'b')) */
   /* search1159_0_0 -> 16 in the cache */
   /* search1159_0 -> 269 in the cache */
   /* search1159_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 269, 36 },
      4,
   } },

   /* replace1159_0_0 -> 16 in the cache */
   /* replace1159_0_1 -> 36 in the cache */
   /* replace1159_0 -> 258 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 258 },
      -1,
   } },

   /* ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   /* search1160_0_0 -> 128 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 128 },
      -1,
   } },
   /* search1160_1_0 -> 129 in the cache */
   /* search1160_1 -> 1152 in the cache */
   /* search1160_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 3199, 1152, 131 },
      -1,
   } },

   /* replace1160_0 -> 128 in the cache */
   /* replace1160_1 -> 129 in the cache */
   /* replace1160_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 128, 129, 131 },
      -1,
   } },

   /* ('ffmaz', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search1161_0_0 -> 16 in the cache */
   /* search1161_0 -> 269 in the cache */
   /* search1161_1_0 -> 36 in the cache */
   /* search1161_1 -> 585 in the cache */
   /* search1161_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 269, 585, 137 },
      -1,
   } },

   /* replace1161_0 -> 16 in the cache */
   /* replace1161_1 -> 36 in the cache */
   /* replace1161_2 -> 137 in the cache */
   /* replace1161 -> 377 in the cache */

   /* ('imul', ('ineg', 'a'), 'b') => ('ineg', ('imul', 'a', 'b')) */
   /* search1162_0_0 -> 0 in the cache */
   /* search1162_0 -> 196 in the cache */
   /* search1162_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 196, 3 },
      -1,
   } },

   /* replace1162_0_0 -> 0 in the cache */
   /* replace1162_0_1 -> 3 in the cache */
   /* replace1162_0 -> 913 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 913 },
      -1,
   } },

   /* ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmul', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3205, 3206 },
      0,
   } },
   /* search1163_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3207, 776 },
      0,
   } },

   /* replace1163_0_0 -> 128 in the cache */
   /* replace1163_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace1163_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3209, 129 },
      -1,
   } },

   /* ('~fmulz(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fmulz', ('fmulz', 'a', 'c'), 'b') */
   /* search1164_0_0 -> 317 in the cache */
   /* search1164_0_1 -> 327 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 317, 327 },
      0,
   } },
   /* search1164_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3211, 298 },
      0,
   } },

   /* replace1164_0_0 -> 16 in the cache */
   /* replace1164_0_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 16, 137 },
      -1,
   } },
   /* replace1164_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3213, 36 },
      -1,
   } },

   /* ('~fmul(is_used_once)', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)') => ('fmulz', ('fmul', 'a', 'c'), 'b') */
   /* search1165_0_0 -> 317 in the cache */
   /* search1165_0_1 -> 327 in the cache */
   /* search1165_0 -> 3211 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3211, 3215 },
      0,
   } },

   /* replace1165_0_0 -> 16 in the cache */
   /* replace1165_0_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 16, 137 },
      -1,
   } },
   /* replace1165_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 3217, 36 },
      -1,
   } },

   /* ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1166_0_0 -> 3205 in the cache */
   /* search1166_0_1 -> 3206 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 3205, 3206 },
      0,
   } },
   /* search1166_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 3219, 776 },
      0,
   } },

   /* replace1166_0_0 -> 128 in the cache */
   /* replace1166_0_1 -> 131 in the cache */
   /* replace1166_0 -> 987 in the cache */
   /* replace1166_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 987, 129 },
      -1,
   } },

   /* ('~ffma', ('fmul(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3222, 3223 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1167_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3224, 3225, 155 },
      -1,
   } },

   /* replace1167_0_0 -> 151 in the cache */
   /* replace1167_0_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 151, 154 },
      -1,
   } },
   /* replace1167_1 -> 152 in the cache */
   /* replace1167_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3227, 152, 155 },
      -1,
   } },

   /* ('~ffmaz', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c', 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd') */
   /* search1168_0_0 -> 317 in the cache */
   /* search1168_0_1 -> 327 in the cache */
   /* search1168_0 -> 3211 in the cache */
   /* search1168_1 -> 298 in the cache */
   /* search1168_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3211, 298, 165 },
      -1,
   } },

   /* replace1168_0_0 -> 16 in the cache */
   /* replace1168_0_1 -> 137 in the cache */
   /* replace1168_0 -> 3213 in the cache */
   /* replace1168_1 -> 36 in the cache */
   /* replace1168_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3213, 36, 165 },
      -1,
   } },

   /* ('~ffma', ('fmulz(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c(is_finite_not_zero)', 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd') */
   /* search1169_0_0 -> 317 in the cache */
   /* search1169_0_1 -> 327 in the cache */
   /* search1169_0 -> 3211 in the cache */
   /* search1169_1 -> 3215 in the cache */
   /* search1169_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3211, 3215, 165 },
      -1,
   } },

   /* replace1169_0_0 -> 16 in the cache */
   /* replace1169_0_1 -> 137 in the cache */
   /* replace1169_0 -> 3217 in the cache */
   /* replace1169_1 -> 36 in the cache */
   /* replace1169_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3217, 36, 165 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_fmul)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1170_0_0 -> 3205 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      29,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3205, 3233 },
      0,
   } },
   /* search1170_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3234, 776 },
      0,
   } },

   /* replace1170_0_0 -> 128 in the cache */
   /* replace1170_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace1170_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3236, 129 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1171_0_0 -> 3205 in the cache */
   /* search1171_0_1 -> 3206 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3205, 3206 },
      0,
   } },
   /* search1171_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3238, 776 },
      0,
   } },

   /* replace1171_0_0 -> 128 in the cache */
   /* replace1171_0_1 -> 131 in the cache */
   /* replace1171_0 -> 3236 in the cache */
   /* replace1171_1 -> 129 in the cache */
   /* replace1171 -> 3237 in the cache */

   /* ('~fadd(is_used_once)', ('ffma(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffma', 'a', 'b', 'd'), 'c') */
   /* search1172_0_0 -> 3222 in the cache */
   /* search1172_0_1 -> 152 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 3222, 152, 3240 },
      0,
   } },
   /* search1172_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3241, 993 },
      0,
   } },

   /* replace1172_0_0 -> 151 in the cache */
   /* replace1172_0_1 -> 152 in the cache */
   /* replace1172_0_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 151, 152, 155 },
      -1,
   } },
   /* replace1172_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3243, 154 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('ffmaz(is_used_once)', 'a(is_not_const)', 'b', 'c(is_not_const)'), '#d') => ('fadd', ('ffmaz', 'a', 'b', 'd'), 'c') */
   /* search1173_0_0 -> 317 in the cache */
   /* search1173_0_1 -> 36 in the cache */
   /* search1173_0_2 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 317, 36, 336 },
      0,
   } },
   /* search1173_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3245, 300 },
      0,
   } },

   /* replace1173_0_0 -> 16 in the cache */
   /* replace1173_0_1 -> 36 in the cache */
   /* replace1173_0_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 16, 36, 165 },
      -1,
   } },
   /* replace1173_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3247, 137 },
      -1,
   } },

   /* ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), '#c') => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1174_0_0 -> 3205 in the cache */
   /* search1174_0_1 -> 3206 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 3205, 3206 },
      0,
   } },
   /* search1174_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 3249, 776 },
      0,
   } },

   /* replace1174_0_0 -> 128 in the cache */
   /* replace1174_0_1 -> 131 in the cache */
   /* replace1174_0 -> 1001 in the cache */
   /* replace1174_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 1001, 129 },
      -1,
   } },

   /* ('~fmul', '#a', ('fmul', 'b(is_not_const)', '#c')) => ('fmul', ('fmul', 'a', 'c'), 'b') */
   /* search1175_0 -> 271 in the cache */
   /* search1175_1_0 -> 3206 in the cache */
   /* search1175_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 271, 3252 },
      -1,
   } },

   /* replace1175_0_0 -> 128 in the cache */
   /* replace1175_0_1 -> 131 in the cache */
   /* replace1175_0 -> 3209 in the cache */
   /* replace1175_1 -> 129 in the cache */
   /* replace1175 -> 3210 in the cache */

   /* ('~fmulz', '#a', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmulz', 'a', 'c'), 'b') */
   /* search1176_0 -> 275 in the cache */
   /* search1176_1_0 -> 327 in the cache */
   /* search1176_1_1 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 327, 298 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 2,
      { 275, 3254 },
      -1,
   } },

   /* replace1176_0_0 -> 16 in the cache */
   /* replace1176_0_1 -> 137 in the cache */
   /* replace1176_0 -> 3213 in the cache */
   /* replace1176_1 -> 36 in the cache */
   /* replace1176 -> 3214 in the cache */

   /* ('~fmul', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c')) => ('fmulz', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1177_1_0 -> 327 in the cache */
   /* search1177_1_1 -> 298 in the cache */
   /* search1177_1 -> 3254 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 3256, 3254 },
      -1,
   } },

   /* replace1177_0_0 -> 16 in the cache */
   /* replace1177_0_1 -> 137 in the cache */
   /* replace1177_0 -> 3217 in the cache */
   /* replace1177_1 -> 36 in the cache */
   /* replace1177 -> 3218 in the cache */

   /* ('~ffma', '#a', ('fmul', 'b(is_not_const)', '#c'), 'd') => ('ffma', ('fmul', 'a', 'c'), 'b', 'd') */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1178_1_0 -> 3223 in the cache */
   /* search1178_1_1 -> 3225 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 3223, 3225 },
      -1,
   } },
   /* search1178_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3258, 3259, 155 },
      -1,
   } },

   /* replace1178_0_0 -> 151 in the cache */
   /* replace1178_0_1 -> 154 in the cache */
   /* replace1178_0 -> 3227 in the cache */
   /* replace1178_1 -> 152 in the cache */
   /* replace1178_2 -> 155 in the cache */
   /* replace1178 -> 3228 in the cache */

   /* ('~ffmaz', '#a', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmulz', 'a', 'c'), 'b', 'd') */
   /* search1179_0 -> 275 in the cache */
   /* search1179_1_0 -> 327 in the cache */
   /* search1179_1_1 -> 298 in the cache */
   /* search1179_1 -> 3254 in the cache */
   /* search1179_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 275, 3254, 165 },
      -1,
   } },

   /* replace1179_0_0 -> 16 in the cache */
   /* replace1179_0_1 -> 137 in the cache */
   /* replace1179_0 -> 3213 in the cache */
   /* replace1179_1 -> 36 in the cache */
   /* replace1179_2 -> 165 in the cache */
   /* replace1179 -> 3230 in the cache */

   /* ('~ffmaz', '#a(is_finite_not_zero)', ('fmulz', 'b(is_not_const)', '#c'), 'd') => ('ffmaz', ('fmul', 'a', 'c'), 'b', 'd') */
   /* search1180_0 -> 3256 in the cache */
   /* search1180_1_0 -> 327 in the cache */
   /* search1180_1_1 -> 298 in the cache */
   /* search1180_1 -> 3254 in the cache */
   /* search1180_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 3256, 3254, 165 },
      -1,
   } },

   /* replace1180_0_0 -> 16 in the cache */
   /* replace1180_0_1 -> 137 in the cache */
   /* replace1180_0 -> 3217 in the cache */
   /* replace1180_1 -> 36 in the cache */
   /* replace1180_2 -> 165 in the cache */
   /* replace1180 -> 3232 in the cache */

   /* ('imul', '#a', ('imul', 'b(is_not_const)', '#c')) => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1181_0 -> 271 in the cache */
   /* search1181_1_0 -> 3206 in the cache */
   /* search1181_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 271, 3263 },
      -1,
   } },

   /* replace1181_0_0 -> 128 in the cache */
   /* replace1181_0_1 -> 131 in the cache */
   /* replace1181_0 -> 987 in the cache */
   /* replace1181_1 -> 129 in the cache */
   /* replace1181 -> 3221 in the cache */

   /* ('~fadd', '#a', ('fadd', 'b(is_not_const)', '#c')) => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1182_0 -> 271 in the cache */
   /* search1182_1_0 -> 3206 in the cache */
   /* search1182_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 271, 3265 },
      -1,
   } },

   /* replace1182_0_0 -> 128 in the cache */
   /* replace1182_0_1 -> 131 in the cache */
   /* replace1182_0 -> 3236 in the cache */
   /* replace1182_1 -> 129 in the cache */
   /* replace1182 -> 3237 in the cache */

   /* ('~fadd', '#a', ('fneg', ('fadd', 'b(is_not_const)', '#c'))) => ('fadd', ('fadd', 'a', ('fneg', 'c')), ('fneg', 'b')) */
   /* search1183_0 -> 271 in the cache */
   /* search1183_1_0_0 -> 3206 in the cache */
   /* search1183_1_0_1 -> 776 in the cache */
   /* search1183_1_0 -> 3265 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3265 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 271, 3267 },
      -1,
   } },

   /* replace1183_0_0 -> 128 in the cache */
   /* replace1183_0_1_0 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 128, 3269 },
      -1,
   } },
   /* replace1183_1_0 -> 129 in the cache */
   /* replace1183_1 -> 1152 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3270, 1152 },
      -1,
   } },

   /* ('~fadd', '#a', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffma', 'b', 'c', ('fadd', 'a', 'd')) */
   /* search1184_0 -> 3258 in the cache */
   /* search1184_1_0 -> 3223 in the cache */
   /* search1184_1_1 -> 3240 in the cache */
   /* search1184_1_2 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 3223, 3240, 993 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3258, 3272 },
      -1,
   } },

   /* replace1184_0 -> 152 in the cache */
   /* replace1184_1 -> 154 in the cache */
   /* replace1184_2_0 -> 151 in the cache */
   /* replace1184_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 151, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 152, 154, 3274 },
      -1,
   } },

   /* ('~fadd', '#a', ('fneg', ('ffma', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffma', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd'))) */
   /* search1185_0 -> 3258 in the cache */
   /* search1185_1_0_0 -> 3223 in the cache */
   /* search1185_1_0_1 -> 3240 in the cache */
   /* search1185_1_0_2 -> 993 in the cache */
   /* search1185_1_0 -> 3272 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3272 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3258, 3276 },
      -1,
   } },

   /* replace1185_0_0 -> 152 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 152 },
      -1,
   } },
   /* replace1185_1 -> 154 in the cache */
   /* replace1185_2_0 -> 151 in the cache */
   /* replace1185_2_1_0 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 151, 3279 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 3278, 154, 3280 },
      -1,
   } },

   /* ('~fadd', '#a', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d')) => ('ffmaz', 'b', 'c', ('fadd', 'a', 'd')) */
   /* search1186_0 -> 275 in the cache */
   /* search1186_1_0 -> 327 in the cache */
   /* search1186_1_1 -> 336 in the cache */
   /* search1186_1_2 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 327, 336, 300 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 275, 3282 },
      -1,
   } },

   /* replace1186_0 -> 36 in the cache */
   /* replace1186_1 -> 137 in the cache */
   /* replace1186_2_0 -> 16 in the cache */
   /* replace1186_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 36, 137, 3284 },
      -1,
   } },

   /* ('~fadd', '#a', ('fneg', ('ffmaz', 'b(is_not_const)', 'c(is_not_const)', '#d'))) => ('ffmaz', ('fneg', 'b'), 'c', ('fadd', 'a', ('fneg', 'd'))) */
   /* search1187_0 -> 275 in the cache */
   /* search1187_1_0_0 -> 327 in the cache */
   /* search1187_1_0_1 -> 336 in the cache */
   /* search1187_1_0_2 -> 300 in the cache */
   /* search1187_1_0 -> 3282 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 3282 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 275, 3286 },
      -1,
   } },

   /* replace1187_0_0 -> 36 in the cache */
   /* replace1187_0 -> 585 in the cache */
   /* replace1187_1 -> 137 in the cache */
   /* replace1187_2_0 -> 16 in the cache */
   /* replace1187_2_1_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 16, 3288 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 585, 137, 3289 },
      -1,
   } },

   /* ('iadd', '#a', ('iadd', 'b(is_not_const)', '#c')) => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1188_0 -> 271 in the cache */
   /* search1188_1_0 -> 3206 in the cache */
   /* search1188_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 271, 3291 },
      -1,
   } },

   /* replace1188_0_0 -> 128 in the cache */
   /* replace1188_0_1 -> 131 in the cache */
   /* replace1188_0 -> 1001 in the cache */
   /* replace1188_1 -> 129 in the cache */
   /* replace1188 -> 3251 in the cache */

   /* ('iand', '#a', ('iand', 'b(is_not_const)', '#c')) => ('iand', ('iand', 'a', 'c'), 'b') */
   /* search1189_0 -> 271 in the cache */
   /* search1189_1_0 -> 3206 in the cache */
   /* search1189_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 271, 3293 },
      -1,
   } },

   /* replace1189_0_0 -> 128 in the cache */
   /* replace1189_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace1189_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3295, 129 },
      -1,
   } },

   /* ('ior', '#a', ('ior', 'b(is_not_const)', '#c')) => ('ior', ('ior', 'a', 'c'), 'b') */
   /* search1190_0 -> 271 in the cache */
   /* search1190_1_0 -> 3206 in the cache */
   /* search1190_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 271, 3297 },
      -1,
   } },

   /* replace1190_0_0 -> 128 in the cache */
   /* replace1190_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace1190_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 3299, 129 },
      -1,
   } },

   /* ('ixor', '#a', ('ixor', 'b(is_not_const)', '#c')) => ('ixor', ('ixor', 'a', 'c'), 'b') */
   /* search1191_0 -> 271 in the cache */
   /* search1191_1_0 -> 3206 in the cache */
   /* search1191_1_1 -> 776 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 3206, 776 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 271, 3301 },
      -1,
   } },

   /* replace1191_0_0 -> 128 in the cache */
   /* replace1191_0_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 128, 131 },
      -1,
   } },
   /* replace1191_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 2,
      { 3303, 129 },
      -1,
   } },

   /* ('~fadd', ('fadd(is_used_once)', 'a(is_fmul)', 'b(is_fmul)'), 'c(is_not_fmul)') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      29,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1192_0_1 -> 3233 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 3305, 3233 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      8,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 3306, 3307 },
      -1,
   } },

   /* replace1192_0_0 -> 128 in the cache */
   /* replace1192_0_1 -> 131 in the cache */
   /* replace1192_0 -> 3236 in the cache */
   /* replace1192_1 -> 129 in the cache */
   /* replace1192 -> 3237 in the cache */

   /* ('idiv', ('imul(no_signed_wrap)', 'a', 'b'), 'b') => a */
   /* search1193_0_0 -> 0 in the cache */
   /* search1193_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 0, 3 },
      5,
   } },
   /* search1193_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_idiv,
      -1, 1,
      { 3309, 3 },
      -1,
   } },

   /* replace1193 -> 0 in the cache */

   /* ('bcsel', ('ige', ('find_lsb', 'a'), 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a') */
   /* search1194_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1194_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3311, 124 },
      -1,
   } },
   /* search1194_1_0 -> 32 in the cache */
   /* search1194_1 -> 3311 in the cache */
   /* search1194_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3312, 3311, 126 },
      -1,
   } },

   /* replace1194_0 -> 32 in the cache */
   /* replace1194 -> 3311 in the cache */

   /* ('bcsel', ('ige', ('ifind_msb', 'a'), 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1195_0_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 16 },
      -1,
   } },
   /* search1195_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3314, 124 },
      -1,
   } },
   /* search1195_1_0 -> 16 in the cache */
   /* search1195_1 -> 3314 in the cache */
   /* search1195_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3315, 3314, 126 },
      -1,
   } },

   /* replace1195_0 -> 16 in the cache */
   /* replace1195 -> 3314 in the cache */

   /* ('bcsel', ('ige', ('ufind_msb', 'a'), 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a') */
   /* search1196_0_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1196_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3317, 124 },
      -1,
   } },
   /* search1196_1_0 -> 32 in the cache */
   /* search1196_1 -> 3317 in the cache */
   /* search1196_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3318, 3317, 126 },
      -1,
   } },

   /* replace1196_0 -> 32 in the cache */
   /* replace1196 -> 3317 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('find_lsb', 'a'), -1) => ('find_lsb', 'a') */
   /* search1197_0_0 -> 32 in the cache */
   /* search1197_0_1 -> 61 in the cache */
   /* search1197_0 -> 1188 in the cache */
   /* search1197_1_0 -> 32 in the cache */
   /* search1197_1 -> 3311 in the cache */
   /* search1197_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1188, 3311, 126 },
      -1,
   } },

   /* replace1197_0 -> 32 in the cache */
   /* replace1197 -> 3311 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1198_0_0 -> 16 in the cache */
   /* search1198_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 124 },
      -1,
   } },
   /* search1198_1_0 -> 16 in the cache */
   /* search1198_1 -> 3314 in the cache */
   /* search1198_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3321, 3314, 126 },
      -1,
   } },

   /* replace1198_0 -> 16 in the cache */
   /* replace1198 -> 3314 in the cache */

   /* ('bcsel', ('ine', 'a', 0), ('ufind_msb', 'a'), -1) => ('ufind_msb', 'a') */
   /* search1199_0_0 -> 32 in the cache */
   /* search1199_0_1 -> 61 in the cache */
   /* search1199_0 -> 1188 in the cache */
   /* search1199_1_0 -> 32 in the cache */
   /* search1199_1 -> 3317 in the cache */
   /* search1199_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1188, 3317, 126 },
      -1,
   } },

   /* replace1199_0 -> 32 in the cache */
   /* replace1199 -> 3317 in the cache */

   /* ('bcsel', ('ine', 'a', -1), ('ifind_msb', 'a'), -1) => ('ifind_msb', 'a') */
   /* search1200_0_0 -> 16 in the cache */
   /* search1200_0_1 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 126 },
      -1,
   } },
   /* search1200_1_0 -> 16 in the cache */
   /* search1200_1 -> 3314 in the cache */
   /* search1200_2 -> 126 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3324, 3314, 126 },
      -1,
   } },

   /* replace1200_0 -> 16 in the cache */
   /* replace1200 -> 3314 in the cache */

   /* ('~fmul', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b') */
   /* search1201_0_0 -> 506 in the cache */
   /* search1201_0_1 -> 264 in the cache */
   /* search1201_0_2 -> 259 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 264, 259 },
      0,
   } },
   /* search1201_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3326, 3 },
      -1,
   } },

   /* replace1201_0 -> 506 in the cache */
   /* replace1201_1_0 -> 3 in the cache */
   /* replace1201_1 -> 1049 in the cache */
   /* replace1201_2 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 1049, 3 },
      -1,
   } },

   /* ('~fmul', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b')) */
   /* search1202_0_0 -> 506 in the cache */
   /* search1202_0_1 -> 259 in the cache */
   /* search1202_0_2 -> 264 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 259, 264 },
      0,
   } },
   /* search1202_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3329, 3 },
      -1,
   } },

   /* replace1202_0 -> 506 in the cache */
   /* replace1202_1 -> 3 in the cache */
   /* replace1202_2_0 -> 3 in the cache */
   /* replace1202_2 -> 1049 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 3, 1049 },
      -1,
   } },

   /* ('~fmulz', ('bcsel(is_used_once)', 'c', -1.0, 1.0), 'b') => ('bcsel', 'c', ('fneg', 'b'), 'b') */
   /* search1203_0_0 -> 506 in the cache */
   /* search1203_0_1 -> 267 in the cache */
   /* search1203_0_2 -> 232 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 267, 232 },
      0,
   } },
   /* search1203_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 3332, 36 },
      -1,
   } },

   /* replace1203_0 -> 506 in the cache */
   /* replace1203_1_0 -> 36 in the cache */
   /* replace1203_1 -> 585 in the cache */
   /* replace1203_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 585, 36 },
      -1,
   } },

   /* ('~fmulz', ('bcsel(is_used_once)', 'c', 1.0, -1.0), 'b') => ('bcsel', 'c', 'b', ('fneg', 'b')) */
   /* search1204_0_0 -> 506 in the cache */
   /* search1204_0_1 -> 232 in the cache */
   /* search1204_0_2 -> 267 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 232, 267 },
      0,
   } },
   /* search1204_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 3335, 36 },
      -1,
   } },

   /* replace1204_0 -> 506 in the cache */
   /* replace1204_1 -> 36 in the cache */
   /* replace1204_2_0 -> 36 in the cache */
   /* replace1204_2 -> 585 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 36, 585 },
      -1,
   } },

   /* ('~bcsel', ('flt', 'a', 0.0), ('fneg', 'a'), 'a') => ('fabs', 'a') */
   /* search1205_0_0 -> 32 in the cache */
   /* search1205_0_1 -> 113 in the cache */
   /* search1205_0 -> 674 in the cache */
   /* search1205_1_0 -> 32 in the cache */
   /* search1205_1 -> 98 in the cache */
   /* search1205_2 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 674, 98, 32 },
      -1,
   } },

   /* replace1205_0 -> 32 in the cache */
   /* replace1205 -> 103 in the cache */

   /* ('bcsel', 'a', ('bcsel', 'b', 'c', 'd'), 'd') => ('bcsel', ('iand', 'a', 'b'), 'c', 'd') */
   /* search1206_0 -> 722 in the cache */
   /* search1206_1_0 -> 1090 in the cache */
   /* search1206_1_1 -> 154 in the cache */
   /* search1206_1_2 -> 155 in the cache */
   /* search1206_1 -> 1262 in the cache */
   /* search1206_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 1262, 155 },
      -1,
   } },

   /* replace1206_0_0 -> 722 in the cache */
   /* replace1206_0_1 -> 1090 in the cache */
   /* replace1206_0 -> 1102 in the cache */
   /* replace1206_1 -> 154 in the cache */
   /* replace1206_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 1102, 154, 155 },
      -1,
   } },

   /* ('bcsel', 'a', 'b', ('bcsel', 'c', 'b', 'd')) => ('bcsel', ('ior', 'a', 'c'), 'b', 'd') */
   /* search1207_0 -> 722 in the cache */
   /* search1207_1 -> 152 in the cache */
   /* search1207_2_0 -> 449 in the cache */
   /* search1207_2_1 -> 152 in the cache */
   /* search1207_2_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 449, 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 152, 3341 },
      -1,
   } },

   /* replace1207_0_0 -> 722 in the cache */
   /* replace1207_0_1 -> 449 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 722, 449 },
      -1,
   } },
   /* replace1207_1 -> 152 in the cache */
   /* replace1207_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3343, 152, 155 },
      -1,
   } },

   /* ('fmod', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ffloor', ('fdiv', 'a', 'b')))) */
   /* search1208_0 -> 0 in the cache */
   /* search1208_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1208_0 -> 0 in the cache */
   /* replace1208_1_0 -> 3 in the cache */
   /* replace1208_1_1_0_0 -> 0 in the cache */
   /* replace1208_1_1_0_1 -> 3 in the cache */
   /* replace1208_1_1_0 -> 2611 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 2611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3346 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 1,
      { 0, 3347 },
      -1,
   } },

   /* ('frem', 'a', 'b') => ('fsub', 'a', ('fmul', 'b', ('ftrunc', ('fdiv', 'a', 'b')))) */
   /* search1209_0 -> 0 in the cache */
   /* search1209_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1209_0 -> 0 in the cache */
   /* replace1209_1_0 -> 3 in the cache */
   /* replace1209_1_1_0_0 -> 0 in the cache */
   /* replace1209_1_1_0_1 -> 3 in the cache */
   /* replace1209_1_1_0 -> 2611 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ftrunc,
      -1, 0,
      { 2611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3, 3350 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 1,
      { 0, 3351 },
      -1,
   } },

   /* ('uadd_carry', 'a', 'b') => ('b2i', ('ult', ('iadd', 'a', 'b'), 'a')) */
   /* search1210_0 -> 0 in the cache */
   /* search1210_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uadd_carry,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1210_0_0_0 -> 0 in the cache */
   /* replace1210_0_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1210_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3354, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 1,
      { 3355 },
      -1,
   } },

   /* ('usub_borrow', 'a', 'b') => ('b2i', ('ult', 'a', 'b')) */
   /* search1211_0 -> 0 in the cache */
   /* search1211_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_usub_borrow,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1211_0_0 -> 0 in the cache */
   /* replace1211_0_1 -> 3 in the cache */
   /* replace1211_0 -> 1029 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 1029 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bfi', ('bfm', 'bits', 'offset'), 'insert', 'base')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* base */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* insert */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_insert,
      -1, 0,
      { 3359, 3360, 3361, 3362 },
      -1,
   } },

   /* replace1212_0_0 -> 19 in the cache */
   /* replace1212_0_1 -> 3362 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 19, 3362 },
      -1,
   } },
   /* replace1212_1 -> 3360 in the cache */
   /* replace1212_2_0_0 -> 3362 in the cache */
   /* replace1212_2_0_1 -> 3361 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 3362, 3361 },
      -1,
   } },
   /* replace1212_2_1 -> 3360 in the cache */
   /* replace1212_2_2 -> 3359 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfi,
      -1, 0,
      { 3365, 3360, 3359 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3364, 3360, 3366 },
      -1,
   } },

   /* ('ihadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1213_0 -> 0 in the cache */
   /* search1213_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ihadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1213_0_0 -> 0 in the cache */
   /* replace1213_0_1 -> 3 in the cache */
   /* replace1213_0 -> 2496 in the cache */
   /* replace1213_1_0_0 -> 0 in the cache */
   /* replace1213_1_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ixor,
      2, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1213_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3369, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 2496, 3370 },
      -1,
   } },

   /* ('uhadd', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1214_0 -> 0 in the cache */
   /* search1214_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1214_0_0 -> 0 in the cache */
   /* replace1214_0_1 -> 3 in the cache */
   /* replace1214_0 -> 2496 in the cache */
   /* replace1214_1_0_0 -> 0 in the cache */
   /* replace1214_1_0_1 -> 3 in the cache */
   /* replace1214_1_0 -> 3369 in the cache */
   /* replace1214_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3369, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 2496, 3373 },
      -1,
   } },

   /* ('irhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1215_0 -> 0 in the cache */
   /* search1215_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_irhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1215_0_0 -> 0 in the cache */
   /* replace1215_0_1 -> 3 in the cache */
   /* replace1215_0 -> 2262 in the cache */
   /* replace1215_1_0_0 -> 0 in the cache */
   /* replace1215_1_0_1 -> 3 in the cache */
   /* replace1215_1_0 -> 2492 in the cache */
   /* replace1215_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 2492, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 2262, 3376 },
      -1,
   } },

   /* ('urhadd', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1216_0 -> 0 in the cache */
   /* search1216_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_urhadd,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1216_0_0 -> 0 in the cache */
   /* replace1216_0_1 -> 3 in the cache */
   /* replace1216_0 -> 2262 in the cache */
   /* replace1216_1_0_0 -> 0 in the cache */
   /* replace1216_1_0_1 -> 3 in the cache */
   /* replace1216_1_0 -> 2492 in the cache */
   /* replace1216_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 2492, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 2262, 3379 },
      -1,
   } },

   /* ('ihadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1217_0 -> 21 in the cache */
   /* search1217_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ihadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1217_0_0 -> 21 in the cache */
   /* replace1217_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1217_1_0_0 -> 21 in the cache */
   /* replace1217_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      2, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1217_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3383, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 3382, 3384 },
      -1,
   } },

   /* ('uhadd@64', 'a', 'b') => ('iadd', ('iand', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1218_0 -> 21 in the cache */
   /* search1218_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1218_0_0 -> 21 in the cache */
   /* replace1218_0_1 -> 41 in the cache */
   /* replace1218_0 -> 3382 in the cache */
   /* replace1218_1_0_0 -> 21 in the cache */
   /* replace1218_1_0_1 -> 41 in the cache */
   /* replace1218_1_0 -> 3383 in the cache */
   /* replace1218_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3383, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 3382, 3387 },
      -1,
   } },

   /* ('irhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ishr', ('ixor', 'a', 'b'), 1)) */
   /* search1219_0 -> 21 in the cache */
   /* search1219_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_irhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1219_0_0 -> 21 in the cache */
   /* replace1219_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1219_1_0_0 -> 21 in the cache */
   /* replace1219_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1219_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3391, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 3390, 3392 },
      -1,
   } },

   /* ('urhadd@64', 'a', 'b') => ('isub', ('ior', 'a', 'b'), ('ushr', ('ixor', 'a', 'b'), 1)) */
   /* search1220_0 -> 21 in the cache */
   /* search1220_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_urhadd,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1220_0_0 -> 21 in the cache */
   /* replace1220_0_1 -> 41 in the cache */
   /* replace1220_0 -> 3390 in the cache */
   /* replace1220_1_0_0 -> 21 in the cache */
   /* replace1220_1_0_1 -> 41 in the cache */
   /* replace1220_1_0 -> 3391 in the cache */
   /* replace1220_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 3391, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 2,
      { 3390, 3395 },
      -1,
   } },

   /* ('imul_32x16', 'a', 'b') => ('imul', 'a', ('extract_i16', 'b', 0)) */
   /* search1221_0 -> 16 in the cache */
   /* search1221_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul_32x16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1221_0 -> 16 in the cache */
   /* replace1221_1_0 -> 36 in the cache */
   /* replace1221_1_1 -> 124 in the cache */
   /* replace1221_1 -> 422 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 422 },
      -1,
   } },

   /* ('umul_32x16', 'a', 'b') => ('imul', 'a', ('extract_u16', 'b', 0)) */
   /* search1222_0 -> 16 in the cache */
   /* search1222_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_32x16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1222_0 -> 16 in the cache */
   /* replace1222_1_0 -> 36 in the cache */
   /* replace1222_1_1 -> 124 in the cache */
   /* replace1222_1 -> 430 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 16, 430 },
      -1,
   } },

   /* ('uadd_sat@64', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b')) */
   /* search1223_0 -> 21 in the cache */
   /* search1223_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1223_0_0_0 -> 21 in the cache */
   /* replace1223_0_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1223_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 3402, 21 },
      -1,
   } },
   /* replace1223_1 -> 1443 in the cache */
   /* replace1223_2_0 -> 21 in the cache */
   /* replace1223_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3403, 1443, 3404 },
      -1,
   } },

   /* ('uadd_sat', 'a', 'b') => ('bcsel', ('ult', ('iadd', 'a', 'b'), 'a'), -1, ('iadd', 'a', 'b')) */
   /* search1224_0 -> 0 in the cache */
   /* search1224_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_uadd_sat,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1224_0_0_0 -> 0 in the cache */
   /* replace1224_0_0_1 -> 3 in the cache */
   /* replace1224_0_0 -> 3354 in the cache */
   /* replace1224_0_1 -> 0 in the cache */
   /* replace1224_0 -> 3355 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   /* replace1224_2_0 -> 0 in the cache */
   /* replace1224_2_1 -> 3 in the cache */
   /* replace1224_2 -> 197 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3355, 3407, 197 },
      -1,
   } },

   /* ('usub_sat', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b')) */
   /* search1225_0 -> 0 in the cache */
   /* search1225_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* replace1225_0_0 -> 0 in the cache */
   /* replace1225_0_1 -> 3 in the cache */
   /* replace1225_0 -> 1029 in the cache */
   /* replace1225_1 -> 86 in the cache */
   /* replace1225_2_0 -> 0 in the cache */
   /* replace1225_2_1 -> 3 in the cache */
   /* replace1225_2 -> 3188 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1029, 86, 3188 },
      -1,
   } },

   /* ('usub_sat@64', 'a', 'b') => ('bcsel', ('ult', 'a', 'b'), 0, ('isub', 'a', 'b')) */
   /* search1226_0 -> 21 in the cache */
   /* search1226_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_usub_sat,
      -1, 0,
      { 21, 41 },
      -1,
   } },

   /* replace1226_0_0 -> 21 in the cache */
   /* replace1226_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   /* replace1226_1 -> 859 in the cache */
   /* replace1226_2_0 -> 21 in the cache */
   /* replace1226_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3412, 859, 3413 },
      -1,
   } },

   /* ('iadd_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ilt', 'b', 0)), ('ige', ('iadd', 'a', 'b'), 0)), ('iadd', 'a', 'b'), 9223372036854775807)) */
   /* search1227_0 -> 21 in the cache */
   /* search1227_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 21, 41 },
      -1,
   } },

   /* replace1227_0_0_0_0 -> 21 in the cache */
   /* replace1227_0_0_0_1 -> 859 in the cache */
   /* replace1227_0_0_0 -> 1433 in the cache */
   /* replace1227_0_0_1_0 -> 41 in the cache */
   /* replace1227_0_0_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 41, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1433, 3416 },
      -1,
   } },
   /* replace1227_0_1_0_0 -> 21 in the cache */
   /* replace1227_0_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1227_0_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3418, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3417, 3419 },
      -1,
   } },
   /* replace1227_1 -> 22 in the cache */
   /* replace1227_2_0_0_0_0 -> 21 in the cache */
   /* replace1227_2_0_0_0_1 -> 859 in the cache */
   /* replace1227_2_0_0_0 -> 1433 in the cache */
   /* replace1227_2_0_0_1_0 -> 41 in the cache */
   /* replace1227_2_0_0_1_1 -> 859 in the cache */
   /* replace1227_2_0_0_1 -> 3416 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      4, 1,
      { 1433, 3416 },
      -1,
   } },
   /* replace1227_2_0_1_0_0 -> 21 in the cache */
   /* replace1227_2_0_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 1,
      { 21, 41 },
      -1,
   } },
   /* replace1227_2_0_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3422, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      3, 3,
      { 3421, 3423 },
      -1,
   } },
   /* replace1227_2_1_0 -> 21 in the cache */
   /* replace1227_2_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      6, 1,
      { 21, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fffffffffffffff /* 9223372036854775807 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3424, 3425, 3426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 7,
      { 3420, 22, 3427 },
      -1,
   } },

   /* ('isub_sat@64', 'a', 'b') => ('bcsel', ('iand', ('iand', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), 9223372036854775808, ('bcsel', ('ior', ('ior', ('ilt', 'a', 0), ('ige', 'b', 0)), ('ige', ('isub', 'a', 'b'), 0)), ('isub', 'a', 'b'), 9223372036854775807)) */
   /* search1228_0 -> 21 in the cache */
   /* search1228_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 21, 41 },
      -1,
   } },

   /* replace1228_0_0_0_0 -> 21 in the cache */
   /* replace1228_0_0_0_1 -> 859 in the cache */
   /* replace1228_0_0_0 -> 1433 in the cache */
   /* replace1228_0_0_1_0 -> 41 in the cache */
   /* replace1228_0_0_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 41, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 1433, 3430 },
      -1,
   } },
   /* replace1228_0_1_0_0 -> 21 in the cache */
   /* replace1228_0_1_0_1 -> 41 in the cache */
   /* replace1228_0_1_0 -> 3413 in the cache */
   /* replace1228_0_1_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3413, 859 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3431, 3432 },
      -1,
   } },
   /* replace1228_1 -> 22 in the cache */
   /* replace1228_2_0_0_0_0 -> 21 in the cache */
   /* replace1228_2_0_0_0_1 -> 859 in the cache */
   /* replace1228_2_0_0_0 -> 1433 in the cache */
   /* replace1228_2_0_0_1_0 -> 41 in the cache */
   /* replace1228_2_0_0_1_1 -> 859 in the cache */
   /* replace1228_2_0_0_1 -> 3430 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      3, 1,
      { 1433, 3430 },
      -1,
   } },
   /* replace1228_2_0_1_0_0 -> 21 in the cache */
   /* replace1228_2_0_1_0_1 -> 41 in the cache */
   /* replace1228_2_0_1_0 -> 3413 in the cache */
   /* replace1228_2_0_1_1 -> 859 in the cache */
   /* replace1228_2_0_1 -> 3432 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 2,
      { 3434, 3432 },
      -1,
   } },
   /* replace1228_2_1_0 -> 21 in the cache */
   /* replace1228_2_1_1 -> 41 in the cache */
   /* replace1228_2_1 -> 3413 in the cache */
   /* replace1228_2_2 -> 3426 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3435, 3413, 3426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3433, 22, 3436 },
      -1,
   } },

   /* ('ilt', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1229_0_0 -> 21 in the cache */
   /* search1229_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search1229_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3438, 859 },
      -1,
   } },

   /* replace1229_0_0_0 -> 21 in the cache */
   /* replace1229_0_0 -> 2945 in the cache */
   /* replace1229_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_64_2x32_split_y,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 2945, 3440 },
      -1,
   } },
   /* replace1229_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3441, 124 },
      -1,
   } },

   /* ('ilt', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ilt', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1230_0_0 -> 21 in the cache */
   /* search1230_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 21, 41 },
      0,
   } },
   /* search1230_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3443, 859 },
      -1,
   } },

   /* replace1230_0_0_0 -> 21 in the cache */
   /* replace1230_0_0 -> 2945 in the cache */
   /* replace1230_0_1_0 -> 41 in the cache */
   /* replace1230_0_1 -> 3440 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 2945, 3440 },
      -1,
   } },
   /* replace1230_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3445, 124 },
      -1,
   } },

   /* ('ige', ('imax(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imax', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1231_0_0 -> 21 in the cache */
   /* search1231_0_1 -> 41 in the cache */
   /* search1231_0 -> 3438 in the cache */
   /* search1231_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3438, 859 },
      -1,
   } },

   /* replace1231_0_0_0 -> 21 in the cache */
   /* replace1231_0_0 -> 2945 in the cache */
   /* replace1231_0_1_0 -> 41 in the cache */
   /* replace1231_0_1 -> 3440 in the cache */
   /* replace1231_0 -> 3441 in the cache */
   /* replace1231_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3441, 124 },
      -1,
   } },

   /* ('ige', ('imin(is_used_once)', 'a@64', 'b@64'), 0) => ('ige', ('imin', ('unpack_64_2x32_split_y', 'a'), ('unpack_64_2x32_split_y', 'b')), 0) */
   /* search1232_0_0 -> 21 in the cache */
   /* search1232_0_1 -> 41 in the cache */
   /* search1232_0 -> 3443 in the cache */
   /* search1232_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3443, 859 },
      -1,
   } },

   /* replace1232_0_0_0 -> 21 in the cache */
   /* replace1232_0_0 -> 2945 in the cache */
   /* replace1232_0_1_0 -> 41 in the cache */
   /* replace1232_0_1 -> 3440 in the cache */
   /* replace1232_0 -> 3445 in the cache */
   /* replace1232_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 1,
      { 3445, 124 },
      -1,
   } },

   /* ('ilt', 'a@64', 0) => ('ilt', ('unpack_64_2x32_split_y', 'a'), 0) */
   /* search1233_0 -> 21 in the cache */
   /* search1233_1 -> 859 in the cache */
   /* search1233 -> 1433 in the cache */

   /* replace1233_0_0 -> 21 in the cache */
   /* replace1233_0 -> 2945 in the cache */
   /* replace1233_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2945, 124 },
      -1,
   } },

   /* ('ige', 'a@64', 0) => ('ige', ('unpack_64_2x32_split_y', 'a'), 0) */
   /* search1234_0 -> 21 in the cache */
   /* search1234_1 -> 859 in the cache */
   /* search1234 -> 1430 in the cache */

   /* replace1234_0_0 -> 21 in the cache */
   /* replace1234_0 -> 2945 in the cache */
   /* replace1234_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2945, 124 },
      -1,
   } },

   /* ('ine', 'a@64', 0) => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1235_0 -> 21 in the cache */
   /* search1235_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 21, 859 },
      -1,
   } },

   /* replace1235_0_0_0 -> 21 in the cache */
   /* replace1235_0_0 -> 2944 in the cache */
   /* replace1235_0_1_0 -> 21 in the cache */
   /* replace1235_0_1 -> 2945 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 2944, 2945 },
      -1,
   } },
   /* replace1235_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3454, 124 },
      -1,
   } },

   /* ('ieq', 'a@64', 0) => ('ieq', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1236_0 -> 21 in the cache */
   /* search1236_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 21, 859 },
      -1,
   } },

   /* replace1236_0_0_0 -> 21 in the cache */
   /* replace1236_0_0 -> 2944 in the cache */
   /* replace1236_0_1_0 -> 21 in the cache */
   /* replace1236_0_1 -> 2945 in the cache */
   /* replace1236_0 -> 3454 in the cache */
   /* replace1236_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 3454, 124 },
      -1,
   } },

   /* ('ult', 0, 'a@64') => ('ine', ('ior', ('unpack_64_2x32_split_x', 'a'), ('unpack_64_2x32_split_y', 'a')), 0) */
   /* search1237_0 -> 859 in the cache */
   /* search1237_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 859, 21 },
      -1,
   } },

   /* replace1237_0_0_0 -> 21 in the cache */
   /* replace1237_0_0 -> 2944 in the cache */
   /* replace1237_0_1_0 -> 21 in the cache */
   /* replace1237_0_1 -> 2945 in the cache */
   /* replace1237_0 -> 3454 in the cache */
   /* replace1237_1 -> 124 in the cache */
   /* replace1237 -> 3455 in the cache */

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('ior', ('iand', 'base', ('inot', ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset'))), ('iand', ('ishl', 'insert', 'offset'), ('ishl', ('isub', ('ishl', 1, 'bits'), 1), 'offset')))) */
   /* search1238_0 -> 3359 in the cache */
   /* search1238_1 -> 3360 in the cache */
   /* search1238_2 -> 3361 in the cache */
   /* search1238_3 -> 3362 in the cache */
   /* search1238 -> 3363 in the cache */

   /* replace1238_0_0 -> 19 in the cache */
   /* replace1238_0_1 -> 3362 in the cache */
   /* replace1238_0 -> 3364 in the cache */
   /* replace1238_1 -> 3360 in the cache */
   /* replace1238_2_0_0 -> 3359 in the cache */
   /* replace1238_2_0_1_0_0_0_0 -> 381 in the cache */
   /* replace1238_2_0_1_0_0_0_1 -> 3362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 381, 3362 },
      -1,
   } },
   /* replace1238_2_0_1_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3459, 381 },
      -1,
   } },
   /* replace1238_2_0_1_0_1 -> 3361 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3460, 3361 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3461 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 3359, 3462 },
      -1,
   } },
   /* replace1238_2_1_0_0 -> 3360 in the cache */
   /* replace1238_2_1_0_1 -> 3361 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3360, 3361 },
      -1,
   } },
   /* replace1238_2_1_1_0_0_0 -> 381 in the cache */
   /* replace1238_2_1_1_0_0_1 -> 3362 in the cache */
   /* replace1238_2_1_1_0_0 -> 3459 in the cache */
   /* replace1238_2_1_1_0_1 -> 381 in the cache */
   /* replace1238_2_1_1_0 -> 3460 in the cache */
   /* replace1238_2_1_1_1 -> 3361 in the cache */
   /* replace1238_2_1_1 -> 3461 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 3464, 3461 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3463, 3465 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 3,
      { 3364, 3360, 3466 },
      -1,
   } },

   /* ('bitfield_insert', 'base', 'insert', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'insert', ('bitfield_select', ('bfm', 'bits', 'offset'), ('ishl', 'insert', 'offset'), 'base')) */
   /* search1239_0 -> 3359 in the cache */
   /* search1239_1 -> 3360 in the cache */
   /* search1239_2 -> 3361 in the cache */
   /* search1239_3 -> 3362 in the cache */
   /* search1239 -> 3363 in the cache */

   /* replace1239_0_0 -> 19 in the cache */
   /* replace1239_0_1 -> 3362 in the cache */
   /* replace1239_0 -> 3364 in the cache */
   /* replace1239_1 -> 3360 in the cache */
   /* replace1239_2_0_0 -> 3362 in the cache */
   /* replace1239_2_0_1 -> 3361 in the cache */
   /* replace1239_2_0 -> 3365 in the cache */
   /* replace1239_2_1_0 -> 3360 in the cache */
   /* replace1239_2_1_1 -> 3361 in the cache */
   /* replace1239_2_1 -> 3464 in the cache */
   /* replace1239_2_2 -> 3359 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 3365, 3464, 3359 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3364, 3360, 3468 },
      -1,
   } },

   /* ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ibfe', 'value', 'offset', 'bits')) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* offset */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibitfield_extract,
      -1, 0,
      { 3470, 3471, 3472 },
      -1,
   } },

   /* replace1240_0_0 -> 19 in the cache */
   /* replace1240_0_1 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 19, 3472 },
      -1,
   } },
   /* replace1240_1 -> 3470 in the cache */
   /* replace1240_2_0 -> 3470 in the cache */
   /* replace1240_2_1 -> 3471 in the cache */
   /* replace1240_2_2 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 3470, 3471, 3472 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3474, 3470, 3475 },
      -1,
   } },

   /* ('ubitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ult', 31, 'bits'), 'value', ('ubfe', 'value', 'offset', 'bits')) */
   /* search1241_0 -> 3470 in the cache */
   /* search1241_1 -> 3471 in the cache */
   /* search1241_2 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubitfield_extract,
      -1, 0,
      { 3470, 3471, 3472 },
      -1,
   } },

   /* replace1241_0_0 -> 19 in the cache */
   /* replace1241_0_1 -> 3472 in the cache */
   /* replace1241_0 -> 3474 in the cache */
   /* replace1241_1 -> 3470 in the cache */
   /* replace1241_2_0 -> 3470 in the cache */
   /* replace1241_2_1 -> 3471 in the cache */
   /* replace1241_2_2 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 3470, 3471, 3472 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3474, 3470, 3478 },
      -1,
   } },

   /* ('bitfield_select', 'a', 'b', 0) => ('iand', 'a', 'b') */
   /* search1242_0 -> 0 in the cache */
   /* search1242_1 -> 3 in the cache */
   /* search1242_2 -> 86 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 0, 3, 86 },
      -1,
   } },

   /* replace1242_0 -> 0 in the cache */
   /* replace1242_1 -> 3 in the cache */
   /* replace1242 -> 2501 in the cache */

   /* ('bitfield_select', 'a', ('iand', 'a', 'b'), 'c') => ('bitfield_select', 'a', 'b', 'c') */
   /* search1243_0 -> 128 in the cache */
   /* search1243_1_0 -> 128 in the cache */
   /* search1243_1_1 -> 129 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 128, 129 },
      -1,
   } },
   /* search1243_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 1,
      { 128, 3481, 131 },
      -1,
   } },

   /* replace1243_0 -> 128 in the cache */
   /* replace1243_1 -> 129 in the cache */
   /* replace1243_2 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bitfield_select,
      -1, 0,
      { 128, 129, 131 },
      -1,
   } },

   /* ('ubfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ubfe', 'value', 'offset', 'bits') */
   /* search1244_0 -> 3470 in the cache */
   /* search1244_1 -> 3471 in the cache */
   /* search1244_2_0 -> 19 in the cache */
   /* search1244_2_1 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 3472 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 1,
      { 3470, 3471, 3484 },
      -1,
   } },

   /* replace1244_0 -> 3470 in the cache */
   /* replace1244_1 -> 3471 in the cache */
   /* replace1244_2 -> 3472 in the cache */
   /* replace1244 -> 3478 in the cache */

   /* ('ubfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ubfe', 'value', 'offset', 'bits') */
   /* search1245_0 -> 3470 in the cache */
   /* search1245_1_0 -> 19 in the cache */
   /* search1245_1_1 -> 3471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 3471 },
      -1,
   } },
   /* search1245_2 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 1,
      { 3470, 3486, 3472 },
      -1,
   } },

   /* replace1245_0 -> 3470 in the cache */
   /* replace1245_1 -> 3471 in the cache */
   /* replace1245_2 -> 3472 in the cache */
   /* replace1245 -> 3478 in the cache */

   /* ('ibfe', 'value', 'offset', ('iand', 31, 'bits')) => ('ibfe', 'value', 'offset', 'bits') */
   /* search1246_0 -> 3470 in the cache */
   /* search1246_1 -> 3471 in the cache */
   /* search1246_2_0 -> 19 in the cache */
   /* search1246_2_1 -> 3472 in the cache */
   /* search1246_2 -> 3484 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 1,
      { 3470, 3471, 3484 },
      -1,
   } },

   /* replace1246_0 -> 3470 in the cache */
   /* replace1246_1 -> 3471 in the cache */
   /* replace1246_2 -> 3472 in the cache */
   /* replace1246 -> 3475 in the cache */

   /* ('ibfe', 'value', ('iand', 31, 'offset'), 'bits') => ('ibfe', 'value', 'offset', 'bits') */
   /* search1247_0 -> 3470 in the cache */
   /* search1247_1_0 -> 19 in the cache */
   /* search1247_1_1 -> 3471 in the cache */
   /* search1247_1 -> 3486 in the cache */
   /* search1247_2 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 1,
      { 3470, 3486, 3472 },
      -1,
   } },

   /* replace1247_0 -> 3470 in the cache */
   /* replace1247_1 -> 3471 in the cache */
   /* replace1247_2 -> 3472 in the cache */
   /* replace1247 -> 3475 in the cache */

   /* ('bfm', 'bits', ('iand', 31, 'offset')) => ('bfm', 'bits', 'offset') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* bits */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1248_1_0 -> 19 in the cache */
   /* search1248_1_1 -> 3471 in the cache */
   /* search1248_1 -> 3486 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 1,
      { 3490, 3486 },
      -1,
   } },

   /* replace1248_0 -> 3490 in the cache */
   /* replace1248_1 -> 3471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 0,
      { 3490, 3471 },
      -1,
   } },

   /* ('bfm', ('iand', 31, 'bits'), 'offset') => ('bfm', 'bits', 'offset') */
   /* search1249_0_0 -> 19 in the cache */
   /* search1249_0_1 -> 3490 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 19, 3490 },
      -1,
   } },
   /* search1249_1 -> 3471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bfm,
      -1, 1,
      { 3493, 3471 },
      -1,
   } },

   /* replace1249_0 -> 3490 in the cache */
   /* replace1249_1 -> 3471 in the cache */
   /* replace1249 -> 3492 in the cache */

   /* ('ubfe', 'a', 'b', 0) => 0 */
   /* search1250_0 -> 16 in the cache */
   /* search1250_1 -> 36 in the cache */
   /* search1250_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 36, 124 },
      -1,
   } },

   /* replace1250 -> 124 in the cache */

   /* ('ibfe', 'a', 'b', 0) => 0 */
   /* search1251_0 -> 16 in the cache */
   /* search1251_1 -> 36 in the cache */
   /* search1251_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 36, 124 },
      -1,
   } },

   /* replace1251 -> 124 in the cache */

   /* ('ubfe', 'a', 0, '#b') => ('iand', 'a', ('ushr', 4294967295, ('ineg', 'b'))) */
   /* search1252_0 -> 16 in the cache */
   /* search1252_1 -> 124 in the cache */
   /* search1252_2 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 124, 33 },
      -1,
   } },

   /* replace1252_0 -> 16 in the cache */
   /* replace1252_1_0 -> 889 in the cache */
   /* replace1252_1_1_0 -> 36 in the cache */
   /* replace1252_1_1 -> 2516 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 889, 2516 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 16, 3498 },
      -1,
   } },

   /* ('b2i32', ('i2b', ('ubfe', 'a', 'b', 1))) => ('ubfe', 'a', 'b', 1) */
   /* search1253_0_0_0 -> 16 in the cache */
   /* search1253_0_0_1 -> 36 in the cache */
   /* search1253_0_0_2 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 36, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2b,
      -1, 0,
      { 3500 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 3501 },
      -1,
   } },

   /* replace1253_0 -> 16 in the cache */
   /* replace1253_1 -> 36 in the cache */
   /* replace1253_2 -> 381 in the cache */
   /* replace1253 -> 3500 in the cache */

   /* ('b2i32', ('i2b', ('ibfe', 'a', 'b', 1))) => ('ubfe', 'a', 'b', 1) */
   /* search1254_0_0_0 -> 16 in the cache */
   /* search1254_0_0_1 -> 36 in the cache */
   /* search1254_0_0_2 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 36, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2b,
      -1, 0,
      { 3503 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2i32,
      -1, 0,
      { 3504 },
      -1,
   } },

   /* replace1254_0 -> 16 in the cache */
   /* replace1254_1 -> 36 in the cache */
   /* replace1254_2 -> 381 in the cache */
   /* replace1254 -> 3500 in the cache */

   /* ('ine', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1255_0_0 -> 16 in the cache */
   /* search1255_0_1 -> 33 in the cache */
   /* search1255_0_2 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 16, 33, 298 },
      0,
   } },
   /* search1255_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3506, 124 },
      -1,
   } },

   /* replace1255_0_0 -> 16 in the cache */
   /* replace1255_0_1_0_0 -> 889 in the cache */
   /* replace1255_0_1_0_1_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 889, 3508 },
      -1,
   } },
   /* replace1255_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3509, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 3510 },
      -1,
   } },
   /* replace1255_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 3511, 124 },
      -1,
   } },

   /* ('ieq', ('ibfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1256_0_0 -> 16 in the cache */
   /* search1256_0_1 -> 33 in the cache */
   /* search1256_0_2 -> 298 in the cache */
   /* search1256_0 -> 3506 in the cache */
   /* search1256_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3506, 124 },
      -1,
   } },

   /* replace1256_0_0 -> 16 in the cache */
   /* replace1256_0_1_0_0 -> 889 in the cache */
   /* replace1256_0_1_0_1_0 -> 137 in the cache */
   /* replace1256_0_1_0_1 -> 3508 in the cache */
   /* replace1256_0_1_0 -> 3509 in the cache */
   /* replace1256_0_1_1 -> 36 in the cache */
   /* replace1256_0_1 -> 3510 in the cache */
   /* replace1256_0 -> 3511 in the cache */
   /* replace1256_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 3511, 124 },
      -1,
   } },

   /* ('ine', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ine', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1257_0_0 -> 16 in the cache */
   /* search1257_0_1 -> 33 in the cache */
   /* search1257_0_2 -> 298 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 16, 33, 298 },
      0,
   } },
   /* search1257_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3515, 124 },
      -1,
   } },

   /* replace1257_0_0 -> 16 in the cache */
   /* replace1257_0_1_0_0 -> 889 in the cache */
   /* replace1257_0_1_0_1_0 -> 137 in the cache */
   /* replace1257_0_1_0_1 -> 3508 in the cache */
   /* replace1257_0_1_0 -> 3509 in the cache */
   /* replace1257_0_1_1 -> 36 in the cache */
   /* replace1257_0_1 -> 3510 in the cache */
   /* replace1257_0 -> 3511 in the cache */
   /* replace1257_1 -> 124 in the cache */
   /* replace1257 -> 3512 in the cache */

   /* ('ieq', ('ubfe(is_used_once)', 'a', '#b', '#c'), 0) => ('ieq', ('iand', 'a', ('ishl', ('ushr', 4294967295, ('ineg', 'c')), 'b')), 0) */
   /* search1258_0_0 -> 16 in the cache */
   /* search1258_0_1 -> 33 in the cache */
   /* search1258_0_2 -> 298 in the cache */
   /* search1258_0 -> 3515 in the cache */
   /* search1258_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3515, 124 },
      -1,
   } },

   /* replace1258_0_0 -> 16 in the cache */
   /* replace1258_0_1_0_0 -> 889 in the cache */
   /* replace1258_0_1_0_1_0 -> 137 in the cache */
   /* replace1258_0_1_0_1 -> 3508 in the cache */
   /* replace1258_0_1_0 -> 3509 in the cache */
   /* replace1258_0_1_1 -> 36 in the cache */
   /* replace1258_0_1 -> 3510 in the cache */
   /* replace1258_0 -> 3511 in the cache */
   /* replace1258_1 -> 124 in the cache */
   /* replace1258 -> 3514 in the cache */

   /* ('ibitfield_extract', 'value', 'offset', 'bits') => ('bcsel', ('ieq', 0, 'bits'), 0, ('ishr', ('ishl', 'value', ('isub', ('isub', 32, 'bits'), 'offset')), ('isub', 32, 'bits'))) */
   /* search1259_0 -> 3470 in the cache */
   /* search1259_1 -> 3471 in the cache */
   /* search1259_2 -> 3472 in the cache */
   /* search1259 -> 3473 in the cache */

   /* replace1259_0_0 -> 124 in the cache */
   /* replace1259_0_1 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 124, 3472 },
      -1,
   } },
   /* replace1259_1 -> 124 in the cache */
   /* replace1259_2_0_0 -> 3470 in the cache */
   /* replace1259_2_0_1_0_0 -> 836 in the cache */
   /* replace1259_2_0_1_0_1 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 836, 3472 },
      -1,
   } },
   /* replace1259_2_0_1_1 -> 3471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3519, 3471 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3470, 3520 },
      -1,
   } },
   /* replace1259_2_1_0 -> 836 in the cache */
   /* replace1259_2_1_1 -> 3472 in the cache */
   /* replace1259_2_1 -> 3519 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 3521, 3519 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3518, 124, 3522 },
      -1,
   } },

   /* ('ubitfield_extract', 'value', 'offset', 'bits') => ('iand', ('ushr', 'value', 'offset'), ('bcsel', ('ieq', 'bits', 32), 4294967295, ('isub', ('ishl', 1, 'bits'), 1))) */
   /* search1260_0 -> 3470 in the cache */
   /* search1260_1 -> 3471 in the cache */
   /* search1260_2 -> 3472 in the cache */
   /* search1260 -> 3477 in the cache */

   /* replace1260_0_0 -> 3470 in the cache */
   /* replace1260_0_1 -> 3471 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 3470, 3471 },
      -1,
   } },
   /* replace1260_1_0_0 -> 3472 in the cache */
   /* replace1260_1_0_1 -> 836 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 3472, 836 },
      -1,
   } },
   /* replace1260_1_1 -> 889 in the cache */
   /* replace1260_1_2_0_0 -> 381 in the cache */
   /* replace1260_1_2_0_1 -> 3472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 381, 3472 },
      -1,
   } },
   /* replace1260_1_2_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 3526, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 3525, 889, 3527 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3524, 3528 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('ufind_msb', ('bcsel', ('ilt', 'value', 0), ('inot', 'value'), 'value')) */
   /* search1261_0 -> 3470 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb,
      -1, 0,
      { 3470 },
      -1,
   } },

   /* replace1261_0_0_0 -> 3470 in the cache */
   /* replace1261_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3470, 124 },
      -1,
   } },
   /* replace1261_0_1_0 -> 3470 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 3470 },
      -1,
   } },
   /* replace1261_0_2 -> 3470 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3531, 3532, 3470 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 3533 },
      -1,
   } },

   /* ('ifind_msb', 'value') => ('bcsel', ('ige', ('ifind_msb_rev', 'value'), 0), ('isub', 31, ('ifind_msb_rev', 'value')), ('ifind_msb_rev', 'value')) */
   /* search1262_0 -> 3470 in the cache */
   /* search1262 -> 3530 in the cache */

   /* replace1262_0_0_0 -> 3470 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ifind_msb_rev,
      -1, 0,
      { 3470 },
      -1,
   } },
   /* replace1262_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3535, 124 },
      -1,
   } },
   /* replace1262_1_0 -> 19 in the cache */
   /* replace1262_1_1_0 -> 3470 in the cache */
   /* replace1262_1_1 -> 3535 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 3535 },
      -1,
   } },
   /* replace1262_2_0 -> 3470 in the cache */
   /* replace1262_2 -> 3535 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3536, 3537, 3535 },
      -1,
   } },

   /* ('ufind_msb', 'value') => ('bcsel', ('ige', ('ufind_msb_rev', 'value'), 0), ('isub', 31, ('ufind_msb_rev', 'value')), ('ufind_msb_rev', 'value')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* value */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 0,
      { 3539 },
      -1,
   } },

   /* replace1263_0_0_0 -> 3539 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb_rev,
      -1, 0,
      { 3539 },
      -1,
   } },
   /* replace1263_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3541, 124 },
      -1,
   } },
   /* replace1263_1_0 -> 19 in the cache */
   /* replace1263_1_1_0 -> 3539 in the cache */
   /* replace1263_1_1 -> 3541 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 19, 3541 },
      -1,
   } },
   /* replace1263_2_0 -> 3539 in the cache */
   /* replace1263_2 -> 3541 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3542, 3543, 3541 },
      -1,
   } },

   /* ('find_lsb', 'value') => ('ufind_msb', ('iand', 'value', ('ineg', 'value'))) */
   /* search1264_0 -> 3539 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3539 },
      -1,
   } },

   /* replace1264_0_0 -> 3539 in the cache */
   /* replace1264_0_1_0 -> 3539 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3539 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3539, 3546 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ufind_msb,
      -1, 1,
      { 3547 },
      -1,
   } },

   /* ('extract_i8', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 3, 'b'), 8)), 24) */
   /* search1265_0 -> 16 in the cache */
   /* search1265_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1265_0_0 -> 16 in the cache */
   /* replace1265_0_1_0_0 -> 390 in the cache */
   /* replace1265_0_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 390, 36 },
      -1,
   } },
   /* replace1265_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3550, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 3551 },
      -1,
   } },
   /* replace1265_1 -> 974 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3552, 974 },
      -1,
   } },

   /* ('extract_u8', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 8)), 255) */
   /* search1266_0 -> 16 in the cache */
   /* search1266_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1266_0_0 -> 16 in the cache */
   /* replace1266_0_1_0 -> 36 in the cache */
   /* replace1266_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 36, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 3555 },
      -1,
   } },
   /* replace1266_1 -> 972 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3556, 972 },
      -1,
   } },

   /* ('extract_i16', 'a', 'b@32') => ('ishr', ('ishl', 'a', ('imul', ('isub', 1, 'b'), 16)), 16) */
   /* search1267_0 -> 16 in the cache */
   /* search1267_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1267_0_0 -> 16 in the cache */
   /* replace1267_0_1_0_0 -> 381 in the cache */
   /* replace1267_0_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 381, 36 },
      -1,
   } },
   /* replace1267_0_1_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3559, 815 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 16, 3560 },
      -1,
   } },
   /* replace1267_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 3561, 815 },
      -1,
   } },

   /* ('extract_u16', 'a', 'b@32') => ('iand', ('ushr', 'a', ('imul', 'b', 16)), 65535) */
   /* search1268_0 -> 16 in the cache */
   /* search1268_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1268_0_0 -> 16 in the cache */
   /* replace1268_0_1_0 -> 36 in the cache */
   /* replace1268_0_1_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 36, 815 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 16, 3564 },
      -1,
   } },
   /* replace1268_1 -> 964 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3565, 964 },
      -1,
   } },

   /* ('pack_unorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 65535.0)))) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* v */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_2x16,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1269_0_0_0_0_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 3567 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x40efffe000000000 /* 65535.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3569, 3570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 1,
      { 3571 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 1,
      { 3572 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec2_to_uint,
      -1, 1,
      { 3573 },
      -1,
   } },

   /* ('pack_unorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2u32', ('fround_even', ('fmul', ('fsat', 'v'), 255.0)))) */
   /* search1270_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_unorm_4x8,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1270_0_0_0_0_0 -> 3567 in the cache */
   /* replace1270_0_0_0_0 -> 3569 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x406fe00000000000 /* 255.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 3569, 3576 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 1,
      { 3577 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2u32,
      -1, 1,
      { 3578 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec4_to_uint,
      -1, 1,
      { 3579 },
      -1,
   } },

   /* ('pack_snorm_2x16', 'v') => ('pack_uvec2_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 32767.0)))) */
   /* search1271_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_2x16,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1271_0_0_0_0_0 -> 232 in the cache */
   /* replace1271_0_0_0_0_1_0 -> 267 in the cache */
   /* replace1271_0_0_0_0_1_1 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      2, 1,
      { 267, 3567 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 2,
      { 232, 3582 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x40dfffc000000000 /* 32767.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 3583, 3584 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 3,
      { 3585 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 3,
      { 3586 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec2_to_uint,
      -1, 3,
      { 3587 },
      -1,
   } },

   /* ('pack_snorm_4x8', 'v') => ('pack_uvec4_to_uint', ('f2i32', ('fround_even', ('fmul', ('fmin', 1.0, ('fmax', -1.0, 'v')), 127.0)))) */
   /* search1272_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_snorm_4x8,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1272_0_0_0_0_0 -> 232 in the cache */
   /* replace1272_0_0_0_0_1_0 -> 267 in the cache */
   /* replace1272_0_0_0_0_1_1 -> 3567 in the cache */
   /* replace1272_0_0_0_0_1 -> 3582 in the cache */
   /* replace1272_0_0_0_0 -> 3583 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x405fc00000000000 /* 127.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 3,
      { 3583, 3590 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 3,
      { 3591 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2i32,
      -1, 3,
      { 3592 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_pack_uvec4_to_uint,
      -1, 3,
      { 3593 },
      -1,
   } },

   /* ('unpack_unorm_2x16', 'v') => ('fdiv', ('u2f32', ('vec2', ('extract_u16', 'v', 0), ('extract_u16', 'v', 1))), 65535.0) */
   /* search1273_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_2x16,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1273_0_0_0_0 -> 3567 in the cache */
   /* replace1273_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3567, 124 },
      -1,
   } },
   /* replace1273_0_0_1_0 -> 3567 in the cache */
   /* replace1273_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 3567, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3596, 3597 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 3598 },
      -1,
   } },
   /* replace1273_1 -> 3570 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3599, 3570 },
      -1,
   } },

   /* ('unpack_unorm_4x8', 'v') => ('fdiv', ('u2f32', ('vec4', ('extract_u8', 'v', 0), ('extract_u8', 'v', 1), ('extract_u8', 'v', 2), ('extract_u8', 'v', 3))), 255.0) */
   /* search1274_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_unorm_4x8,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1274_0_0_0_0 -> 3567 in the cache */
   /* replace1274_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3567, 124 },
      -1,
   } },
   /* replace1274_0_0_1_0 -> 3567 in the cache */
   /* replace1274_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3567, 381 },
      -1,
   } },
   /* replace1274_0_0_2_0 -> 3567 in the cache */
   /* replace1274_0_0_2_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3567, 386 },
      -1,
   } },
   /* replace1274_0_0_3_0 -> 3567 in the cache */
   /* replace1274_0_0_3_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 3567, 390 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3602, 3603, 3604, 3605 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2f32,
      -1, 0,
      { 3606 },
      -1,
   } },
   /* replace1274_1 -> 3576 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3607, 3576 },
      -1,
   } },

   /* ('unpack_snorm_2x16', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec2', ('extract_i16', 'v', 0), ('extract_i16', 'v', 1))), 32767.0))) */
   /* search1275_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_2x16,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1275_0 -> 232 in the cache */
   /* replace1275_1_0 -> 267 in the cache */
   /* replace1275_1_1_0_0_0_0 -> 3567 in the cache */
   /* replace1275_1_1_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 3567, 124 },
      -1,
   } },
   /* replace1275_1_1_0_0_1_0 -> 3567 in the cache */
   /* replace1275_1_1_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 3567, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3610, 3611 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 3612 },
      -1,
   } },
   /* replace1275_1_1_1 -> 3584 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3613, 3584 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 267, 3614 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 232, 3615 },
      -1,
   } },

   /* ('unpack_snorm_4x8', 'v') => ('fmin', 1.0, ('fmax', -1.0, ('fdiv', ('i2f', ('vec4', ('extract_i8', 'v', 0), ('extract_i8', 'v', 1), ('extract_i8', 'v', 2), ('extract_i8', 'v', 3))), 127.0))) */
   /* search1276_0 -> 3567 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_unpack_snorm_4x8,
      -1, 0,
      { 3567 },
      -1,
   } },

   /* replace1276_0 -> 232 in the cache */
   /* replace1276_1_0 -> 267 in the cache */
   /* replace1276_1_1_0_0_0_0 -> 3567 in the cache */
   /* replace1276_1_1_0_0_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3567, 124 },
      -1,
   } },
   /* replace1276_1_1_0_0_1_0 -> 3567 in the cache */
   /* replace1276_1_1_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3567, 381 },
      -1,
   } },
   /* replace1276_1_1_0_0_2_0 -> 3567 in the cache */
   /* replace1276_1_1_0_0_2_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3567, 386 },
      -1,
   } },
   /* replace1276_1_1_0_0_3_0 -> 3567 in the cache */
   /* replace1276_1_1_0_0_3_1 -> 390 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 3567, 390 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3618, 3619, 3620, 3621 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 3622 },
      -1,
   } },
   /* replace1276_1_1_1 -> 3590 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 3623, 3590 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 267, 3624 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 2,
      { 232, 3625 },
      -1,
   } },

   /* ('pack_half_2x16_split', 'a@32', 'b@32') => ('ior', ('ishl', ('u2u32', ('f2f16', 'b')), 16), ('u2u32', ('f2f16', 'a'))) */
   /* search1277_0 -> 16 in the cache */
   /* search1277_1 -> 36 in the cache */
   /* search1277 -> 3075 in the cache */

   /* replace1277_0_0_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3627 },
      -1,
   } },
   /* replace1277_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 3628, 815 },
      -1,
   } },
   /* replace1277_1_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3630 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3629, 3631 },
      -1,
   } },

   /* ('unpack_half_2x16_split_x', 'a@32') => ('f2f32', ('u2u16', 'a')) */
   /* search1278_0 -> 16 in the cache */
   /* search1278 -> 3040 in the cache */

   /* replace1278_0_0 -> 16 in the cache */
   /* replace1278_0 -> 3037 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 3037 },
      -1,
   } },

   /* ('unpack_half_2x16_split_y', 'a@32') => ('f2f32', ('u2u16', ('ushr', 'a', 16))) */
   /* search1279_0 -> 16 in the cache */
   /* search1279 -> 3047 in the cache */

   /* replace1279_0_0_0 -> 16 in the cache */
   /* replace1279_0_0_1 -> 815 in the cache */
   /* replace1279_0_0 -> 970 in the cache */
   /* replace1279_0 -> 3038 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 3038 },
      -1,
   } },

   /* ('pack_32_2x16_split', 'a@16', 'b@16') => ('ior', ('ishl', ('u2u32', 'b'), 16), ('u2u32', 'a')) */
   /* search1280_0 -> 11 in the cache */
   /* search1280_1 -> 247 in the cache */
   /* search1280 -> 3030 in the cache */

   /* replace1280_0_0_0 -> 247 in the cache */
   /* replace1280_0_0 -> 3031 in the cache */
   /* replace1280_0_1 -> 815 in the cache */
   /* replace1280_0 -> 3032 in the cache */
   /* replace1280_1_0 -> 11 in the cache */
   /* replace1280_1 -> 1749 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3032, 1749 },
      -1,
   } },

   /* ('unpack_32_2x16_split_x', 'a@32') => ('u2u16', 'a') */
   /* search1281_0 -> 16 in the cache */
   /* search1281 -> 2953 in the cache */

   /* replace1281_0 -> 16 in the cache */
   /* replace1281 -> 3037 in the cache */

   /* ('unpack_32_2x16_split_y', 'a@32') => ('u2u16', ('ushr', 'a', 16)) */
   /* search1282_0 -> 16 in the cache */
   /* search1282 -> 2955 in the cache */

   /* replace1282_0_0 -> 16 in the cache */
   /* replace1282_0_1 -> 815 in the cache */
   /* replace1282_0 -> 970 in the cache */
   /* replace1282 -> 3038 in the cache */

   /* ('isign', 'a') => ('imin', ('imax', 'a', -1), 1) */
   /* search1283_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1283_0_0 -> 32 in the cache */
   /* replace1283_0_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 32, 63 },
      -1,
   } },
   /* replace1283_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      0, 2,
      { 3637, 35 },
      -1,
   } },

   /* ('imin', ('imax', 'a', -1), 1) => ('isign', 'a') */
   /* search1284_0_0 -> 32 in the cache */
   /* search1284_0_1 -> 63 in the cache */
   /* search1284_0 -> 3637 in the cache */
   /* search1284_1 -> 35 in the cache */
   /* search1284 -> 3638 in the cache */

   /* replace1284_0 -> 32 in the cache */
   /* replace1284 -> 3636 in the cache */

   /* ('imax', ('imin', 'a', 1), -1) => ('isign', 'a') */
   /* search1285_0_0 -> 32 in the cache */
   /* search1285_0_1 -> 35 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 32, 35 },
      -1,
   } },
   /* search1285_1 -> 63 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 2,
      { 3639, 63 },
      -1,
   } },

   /* replace1285_0 -> 32 in the cache */
   /* replace1285 -> 3636 in the cache */

   /* ('fsign', 'a') => ('fsub', ('b2f', ('!flt', 0.0, 'a')), ('b2f', ('!flt', 'a', 0.0))) */
   /* search1286_0 -> 32 in the cache */
   /* search1286 -> 239 in the cache */

   /* replace1286_0_0_0 -> 113 in the cache */
   /* replace1286_0_0_1 -> 32 in the cache */
   /* replace1286_0_0 -> 1470 in the cache */
   /* replace1286_0 -> 1471 in the cache */
   /* replace1286_1_0_0 -> 32 in the cache */
   /* replace1286_1_0_1 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 32, 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 3641 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 1471, 3642 },
      -1,
   } },

   /* ('amul', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1287_0 -> 0 in the cache */
   /* search1287_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_amul,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* replace1287_0 -> 0 in the cache */
   /* replace1287_1 -> 3 in the cache */
   /* replace1287 -> 913 in the cache */

   /* ('umul24', 'a', 'b') => ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)) */
   /* search1288_0 -> 16 in the cache */
   /* search1288_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1288_0_0 -> 16 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffff /* 16777215 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 16, 3646 },
      -1,
   } },
   /* replace1288_1_0 -> 36 in the cache */
   /* replace1288_1_1 -> 3646 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 36, 3646 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 3,
      { 3647, 3648 },
      -1,
   } },

   /* ('umad24', 'a', 'b', 'c') => ('iadd', ('imul', ('iand', 'a', 16777215), ('iand', 'b', 16777215)), 'c') */
   /* search1289_0 -> 16 in the cache */
   /* search1289_1 -> 36 in the cache */
   /* search1289_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umad24,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace1289_0_0_0 -> 16 in the cache */
   /* replace1289_0_0_1 -> 3646 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 16, 3646 },
      -1,
   } },
   /* replace1289_0_1_0 -> 36 in the cache */
   /* replace1289_0_1_1 -> 3646 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 1,
      { 36, 3646 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 3,
      { 3651, 3652 },
      -1,
   } },
   /* replace1289_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 4,
      { 3653, 137 },
      -1,
   } },

   /* ('imul24_relaxed', 'a', 'b') => ('imul24', 'a', 'b') */
   /* search1290_0 -> 16 in the cache */
   /* search1290_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24_relaxed,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1290_0 -> 16 in the cache */
   /* replace1290_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* ('imul24_relaxed', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1291_0 -> 16 in the cache */
   /* search1291_1 -> 36 in the cache */
   /* search1291 -> 3655 in the cache */

   /* replace1291_0 -> 16 in the cache */
   /* replace1291_1 -> 36 in the cache */
   /* replace1291 -> 49 in the cache */

   /* ('umad24_relaxed', 'a', 'b', 'c') => ('umad24', 'a', 'b', 'c') */
   /* search1292_0 -> 16 in the cache */
   /* search1292_1 -> 36 in the cache */
   /* search1292_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umad24_relaxed,
      0, 1,
      { 16, 36, 137 },
      -1,
   } },

   /* replace1292_0 -> 16 in the cache */
   /* replace1292_1 -> 36 in the cache */
   /* replace1292_2 -> 137 in the cache */
   /* replace1292 -> 3650 in the cache */

   /* ('umad24_relaxed', 'a', 'b', 'c') => ('iadd', ('umul24_relaxed', 'a', 'b'), 'c') */
   /* search1293_0 -> 16 in the cache */
   /* search1293_1 -> 36 in the cache */
   /* search1293_2 -> 137 in the cache */
   /* search1293 -> 3657 in the cache */

   /* replace1293_0_0 -> 16 in the cache */
   /* replace1293_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24_relaxed,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1293_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 3658, 137 },
      -1,
   } },

   /* ('umul24_relaxed', 'a', 'b') => ('umul24', 'a', 'b') */
   /* search1294_0 -> 16 in the cache */
   /* search1294_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul24_relaxed,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1294_0 -> 16 in the cache */
   /* replace1294_1 -> 36 in the cache */
   /* replace1294 -> 3645 in the cache */

   /* ('umul24_relaxed', 'a', 'b') => ('imul', 'a', 'b') */
   /* search1295_0 -> 16 in the cache */
   /* search1295_1 -> 36 in the cache */
   /* search1295 -> 3660 in the cache */

   /* replace1295_0 -> 16 in the cache */
   /* replace1295_1 -> 36 in the cache */
   /* replace1295 -> 49 in the cache */

   /* ('imad24_ir3', 'a', 'b', 0) => ('imul24', 'a', 'b') */
   /* search1296_0 -> 16 in the cache */
   /* search1296_1 -> 36 in the cache */
   /* search1296_2 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 36, 124 },
      -1,
   } },

   /* replace1296_0 -> 16 in the cache */
   /* replace1296_1 -> 36 in the cache */
   /* replace1296 -> 3656 in the cache */

   /* ('imad24_ir3', 'a', 0, 'c') => c */
   /* search1297_0 -> 16 in the cache */
   /* search1297_1 -> 124 in the cache */
   /* search1297_2 -> 540 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 124, 540 },
      -1,
   } },

   /* replace1297 -> 540 in the cache */

   /* ('imad24_ir3', 'a', 1, 'c') => ('iadd', 'a', 'c') */
   /* search1298_0 -> 16 in the cache */
   /* search1298_1 -> 381 in the cache */
   /* search1298_2 -> 540 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 16, 381, 540 },
      -1,
   } },

   /* replace1298_0 -> 16 in the cache */
   /* replace1298_1 -> 540 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 540 },
      -1,
   } },

   /* ('imad24_ir3', '#a', '#b', 'c') => ('iadd', ('imul', 'a', 'b'), 'c') */
   /* search1299_0 -> 275 in the cache */
   /* search1299_1 -> 33 in the cache */
   /* search1299_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imad24_ir3,
      0, 1,
      { 275, 33, 137 },
      -1,
   } },

   /* replace1299_0_0 -> 16 in the cache */
   /* replace1299_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1299_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 3666, 137 },
      -1,
   } },

   /* ('imul24', 'a', '#b@32(is_pos_power_of_two)') => ('ishl', 'a', ('find_lsb', 'b')) */
   /* search1300_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 3668 },
      -1,
   } },

   /* replace1300_0 -> 16 in the cache */
   /* replace1300_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 3670 },
      -1,
   } },

   /* ('imul24', 'a', '#b@32(is_neg_power_of_two)') => ('ineg', ('ishl', 'a', ('find_lsb', ('iabs', 'b')))) */
   /* search1301_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 3672 },
      -1,
   } },

   /* replace1301_0_0 -> 16 in the cache */
   /* replace1301_0_1_0_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_find_lsb,
      -1, 0,
      { 3674 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 16, 3675 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 3676 },
      -1,
   } },

   /* ('imul24', 'a', 0) => 0 */
   /* search1302_0 -> 16 in the cache */
   /* search1302_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul24,
      0, 1,
      { 16, 124 },
      -1,
   } },

   /* replace1302 -> 124 in the cache */

   /* ('fcsel', ('slt', 0, 'a'), 'b', 'c') => ('fcsel_gt', 'a', 'b', 'c') */
   /* search1303_0_0 -> 124 in the cache */
   /* search1303_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   /* search1303_1 -> 36 in the cache */
   /* search1303_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 3679, 36, 137 },
      -1,
   } },

   /* replace1303_0 -> 16 in the cache */
   /* replace1303_1 -> 36 in the cache */
   /* replace1303_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_gt,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },

   /* ('fcsel', ('slt', 'a', 0), 'b', 'c') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c') */
   /* search1304_0_0 -> 16 in the cache */
   /* search1304_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_slt,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* search1304_1 -> 36 in the cache */
   /* search1304_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 3682, 36, 137 },
      -1,
   } },

   /* replace1304_0_0 -> 16 in the cache */
   /* replace1304_0 -> 269 in the cache */
   /* replace1304_1 -> 36 in the cache */
   /* replace1304_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_gt,
      -1, 0,
      { 269, 36, 137 },
      -1,
   } },

   /* ('fcsel', ('sge', 'a', 0), 'b', 'c') => ('fcsel_ge', 'a', 'b', 'c') */
   /* search1305_0_0 -> 16 in the cache */
   /* search1305_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* search1305_1 -> 36 in the cache */
   /* search1305_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 3685, 36, 137 },
      -1,
   } },

   /* replace1305_0 -> 16 in the cache */
   /* replace1305_1 -> 36 in the cache */
   /* replace1305_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_ge,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },

   /* ('fcsel', ('sge', 0, 'a'), 'b', 'c') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c') */
   /* search1306_0_0 -> 124 in the cache */
   /* search1306_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_sge,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   /* search1306_1 -> 36 in the cache */
   /* search1306_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel,
      -1, 0,
      { 3688, 36, 137 },
      -1,
   } },

   /* replace1306_0_0 -> 16 in the cache */
   /* replace1306_0 -> 269 in the cache */
   /* replace1306_1 -> 36 in the cache */
   /* replace1306_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcsel_ge,
      -1, 0,
      { 269, 36, 137 },
      -1,
   } },

   /* ('bcsel', ('ilt', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'b', 'c') */
   /* search1307_0_0 -> 124 in the cache */
   /* search1307_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   /* search1307_1 -> 36 in the cache */
   /* search1307_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3691, 36, 137 },
      -1,
   } },

   /* replace1307_0 -> 16 in the cache */
   /* replace1307_1 -> 36 in the cache */
   /* replace1307_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_gt,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },

   /* ('bcsel', ('ilt', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'c', 'b') */
   /* search1308_0_0 -> 16 in the cache */
   /* search1308_0_1 -> 124 in the cache */
   /* search1308_0 -> 1413 in the cache */
   /* search1308_1 -> 36 in the cache */
   /* search1308_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1413, 36, 137 },
      -1,
   } },

   /* replace1308_0 -> 16 in the cache */
   /* replace1308_1 -> 137 in the cache */
   /* replace1308_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_ge,
      -1, 0,
      { 16, 137, 36 },
      -1,
   } },

   /* ('bcsel', ('ige', 'a@32', 0), 'b@32', 'c@32') => ('i32csel_ge', 'a', 'b', 'c') */
   /* search1309_0_0 -> 16 in the cache */
   /* search1309_0_1 -> 124 in the cache */
   /* search1309_0 -> 1410 in the cache */
   /* search1309_1 -> 36 in the cache */
   /* search1309_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1410, 36, 137 },
      -1,
   } },

   /* replace1309_0 -> 16 in the cache */
   /* replace1309_1 -> 36 in the cache */
   /* replace1309_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_ge,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },

   /* ('bcsel', ('ige', 0, 'a@32'), 'b@32', 'c@32') => ('i32csel_gt', 'a', 'c', 'b') */
   /* search1310_0_0 -> 124 in the cache */
   /* search1310_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   /* search1310_1 -> 36 in the cache */
   /* search1310_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3698, 36, 137 },
      -1,
   } },

   /* replace1310_0 -> 16 in the cache */
   /* replace1310_1 -> 137 in the cache */
   /* replace1310_2 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i32csel_gt,
      -1, 0,
      { 16, 137, 36 },
      -1,
   } },

   /* ('bcsel', ('flt', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_gt', 'a', 'b', 'c') */
   /* search1311_0_0 -> 124 in the cache */
   /* search1311_0_1 -> 16 in the cache */
   /* search1311_0 -> 1845 in the cache */
   /* search1311_1 -> 36 in the cache */
   /* search1311_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1845, 36, 137 },
      -1,
   } },

   /* replace1311_0 -> 16 in the cache */
   /* replace1311_1 -> 36 in the cache */
   /* replace1311_2 -> 137 in the cache */
   /* replace1311 -> 3681 in the cache */

   /* ('bcsel', ('flt', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_gt', ('fneg', 'a'), 'b', 'c') */
   /* search1312_0_0 -> 16 in the cache */
   /* search1312_0_1 -> 124 in the cache */
   /* search1312_0 -> 1847 in the cache */
   /* search1312_1 -> 36 in the cache */
   /* search1312_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1847, 36, 137 },
      -1,
   } },

   /* replace1312_0_0 -> 16 in the cache */
   /* replace1312_0 -> 269 in the cache */
   /* replace1312_1 -> 36 in the cache */
   /* replace1312_2 -> 137 in the cache */
   /* replace1312 -> 3684 in the cache */

   /* ('bcsel', ('fge', 'a@32', 0), 'b@32', 'c@32') => ('fcsel_ge', 'a', 'b', 'c') */
   /* search1313_0_0 -> 16 in the cache */
   /* search1313_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* search1313_1 -> 36 in the cache */
   /* search1313_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3703, 36, 137 },
      -1,
   } },

   /* replace1313_0 -> 16 in the cache */
   /* replace1313_1 -> 36 in the cache */
   /* replace1313_2 -> 137 in the cache */
   /* replace1313 -> 3687 in the cache */

   /* ('bcsel', ('fge', 0, 'a@32'), 'b@32', 'c@32') => ('fcsel_ge', ('fneg', 'a'), 'b', 'c') */
   /* search1314_0_0 -> 124 in the cache */
   /* search1314_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 124, 16 },
      -1,
   } },
   /* search1314_1 -> 36 in the cache */
   /* search1314_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3705, 36, 137 },
      -1,
   } },

   /* replace1314_0_0 -> 16 in the cache */
   /* replace1314_0 -> 269 in the cache */
   /* replace1314_1 -> 36 in the cache */
   /* replace1314_2 -> 137 in the cache */
   /* replace1314 -> 3690 in the cache */

   /* ('iadd_sat@8', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 127, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 128, ('iadd', 'a', 'b'))) */
   /* search1315_0 -> 6 in the cache */
   /* search1315_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 6, 2060 },
      -1,
   } },

   /* replace1315_0_0 -> 2060 in the cache */
   /* replace1315_0_1 -> 1374 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 2060, 1374 },
      -1,
   } },
   /* replace1315_1_0_0_0 -> 6 in the cache */
   /* replace1315_1_0_0_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 6, 2060 },
      -1,
   } },
   /* replace1315_1_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3709, 6 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1315_1_2_0 -> 6 in the cache */
   /* replace1315_1_2_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 6, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3710, 3711, 3712 },
      -1,
   } },
   /* replace1315_2_0_0 -> 6 in the cache */
   /* replace1315_2_0_1_0 -> 6 in the cache */
   /* replace1315_2_0_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 6, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 6, 3714 },
      -1,
   } },
   /* replace1315_2_1 -> 7 in the cache */
   /* replace1315_2_2_0 -> 6 in the cache */
   /* replace1315_2_2_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 6, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3715, 7, 3716 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3708, 3713, 3717 },
      -1,
   } },

   /* ('isub_sat@8', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 127, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 128, ('isub', 'a', 'b'))) */
   /* search1316_0 -> 6 in the cache */
   /* search1316_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 6, 2060 },
      -1,
   } },

   /* replace1316_0_0 -> 2060 in the cache */
   /* replace1316_0_1 -> 797 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 2060, 797 },
      -1,
   } },
   /* replace1316_1_0_0_0 -> 6 in the cache */
   /* replace1316_1_0_0_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 6, 2060 },
      -1,
   } },
   /* replace1316_1_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3721, 6 },
      -1,
   } },
   /* replace1316_1_1 -> 3711 in the cache */
   /* replace1316_1_2_0 -> 6 in the cache */
   /* replace1316_1_2_1 -> 2060 in the cache */
   /* replace1316_1_2 -> 3721 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3722, 3711, 3721 },
      -1,
   } },
   /* replace1316_2_0_0 -> 6 in the cache */
   /* replace1316_2_0_1_0 -> 6 in the cache */
   /* replace1316_2_0_1_1 -> 2060 in the cache */
   /* replace1316_2_0_1 -> 3721 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 3721 },
      -1,
   } },
   /* replace1316_2_1 -> 7 in the cache */
   /* replace1316_2_2_0 -> 6 in the cache */
   /* replace1316_2_2_1 -> 2060 in the cache */
   /* replace1316_2_2 -> 3721 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3724, 7, 3721 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3720, 3723, 3725 },
      -1,
   } },

   /* ('iadd_sat@16', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 32767, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 32768, ('iadd', 'a', 'b'))) */
   /* search1317_0 -> 11 in the cache */
   /* search1317_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 11, 247 },
      -1,
   } },

   /* replace1317_0_0 -> 247 in the cache */
   /* replace1317_0_1 -> 1396 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 247, 1396 },
      -1,
   } },
   /* replace1317_1_0_0_0 -> 11 in the cache */
   /* replace1317_1_0_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 11, 247 },
      -1,
   } },
   /* replace1317_1_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3729, 11 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1317_1_2_0 -> 11 in the cache */
   /* replace1317_1_2_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 11, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3730, 3731, 3732 },
      -1,
   } },
   /* replace1317_2_0_0 -> 11 in the cache */
   /* replace1317_2_0_1_0 -> 11 in the cache */
   /* replace1317_2_0_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 11, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 11, 3734 },
      -1,
   } },
   /* replace1317_2_1 -> 12 in the cache */
   /* replace1317_2_2_0 -> 11 in the cache */
   /* replace1317_2_2_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 11, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3735, 12, 3736 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3728, 3733, 3737 },
      -1,
   } },

   /* ('isub_sat@16', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 32767, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 32768, ('isub', 'a', 'b'))) */
   /* search1318_0 -> 11 in the cache */
   /* search1318_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 11, 247 },
      -1,
   } },

   /* replace1318_0_0 -> 247 in the cache */
   /* replace1318_0_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 247, 818 },
      -1,
   } },
   /* replace1318_1_0_0_0 -> 11 in the cache */
   /* replace1318_1_0_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 11, 247 },
      -1,
   } },
   /* replace1318_1_0_1 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3741, 11 },
      -1,
   } },
   /* replace1318_1_1 -> 3731 in the cache */
   /* replace1318_1_2_0 -> 11 in the cache */
   /* replace1318_1_2_1 -> 247 in the cache */
   /* replace1318_1_2 -> 3741 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3742, 3731, 3741 },
      -1,
   } },
   /* replace1318_2_0_0 -> 11 in the cache */
   /* replace1318_2_0_1_0 -> 11 in the cache */
   /* replace1318_2_0_1_1 -> 247 in the cache */
   /* replace1318_2_0_1 -> 3741 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 3741 },
      -1,
   } },
   /* replace1318_2_1 -> 12 in the cache */
   /* replace1318_2_2_0 -> 11 in the cache */
   /* replace1318_2_2_1 -> 247 in the cache */
   /* replace1318_2_2 -> 3741 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3744, 12, 3741 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3740, 3743, 3745 },
      -1,
   } },

   /* ('iadd_sat@32', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 2147483647, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 2147483648, ('iadd', 'a', 'b'))) */
   /* search1319_0 -> 16 in the cache */
   /* search1319_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd_sat,
      0, 1,
      { 16, 36 },
      -1,
   } },

   /* replace1319_0_0 -> 36 in the cache */
   /* replace1319_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 36, 381 },
      -1,
   } },
   /* replace1319_1_0_0_0 -> 16 in the cache */
   /* replace1319_1_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 16, 36 },
      -1,
   } },
   /* replace1319_1_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3749, 16 },
      -1,
   } },
   /* replace1319_1_1 -> 950 in the cache */
   /* replace1319_1_2_0 -> 16 in the cache */
   /* replace1319_1_2_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3750, 950, 3751 },
      -1,
   } },
   /* replace1319_2_0_0 -> 16 in the cache */
   /* replace1319_2_0_1_0 -> 16 in the cache */
   /* replace1319_2_0_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 16, 3753 },
      -1,
   } },
   /* replace1319_2_1 -> 17 in the cache */
   /* replace1319_2_2_0 -> 16 in the cache */
   /* replace1319_2_2_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 16, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3754, 17, 3755 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3748, 3752, 3756 },
      -1,
   } },

   /* ('isub_sat@32', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 2147483647, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 2147483648, ('isub', 'a', 'b'))) */
   /* search1320_0 -> 16 in the cache */
   /* search1320_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub_sat,
      -1, 0,
      { 16, 36 },
      -1,
   } },

   /* replace1320_0_0 -> 36 in the cache */
   /* replace1320_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 36, 124 },
      -1,
   } },
   /* replace1320_1_0_0_0 -> 16 in the cache */
   /* replace1320_1_0_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 16, 36 },
      -1,
   } },
   /* replace1320_1_0_1 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3760, 16 },
      -1,
   } },
   /* replace1320_1_1 -> 950 in the cache */
   /* replace1320_1_2_0 -> 16 in the cache */
   /* replace1320_1_2_1 -> 36 in the cache */
   /* replace1320_1_2 -> 3760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3761, 950, 3760 },
      -1,
   } },
   /* replace1320_2_0_0 -> 16 in the cache */
   /* replace1320_2_0_1_0 -> 16 in the cache */
   /* replace1320_2_0_1_1 -> 36 in the cache */
   /* replace1320_2_0_1 -> 3760 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 3760 },
      -1,
   } },
   /* replace1320_2_1 -> 17 in the cache */
   /* replace1320_2_2_0 -> 16 in the cache */
   /* replace1320_2_2_1 -> 36 in the cache */
   /* replace1320_2_2 -> 3760 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3763, 17, 3760 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3759, 3762, 3764 },
      -1,
   } },

   /* ('iadd_sat@64', 'a', 'b') => ('bcsel', ('ige', 'b', 1), ('bcsel', ('ilt', ('iadd', 'a', 'b'), 'a'), 9223372036854775807, ('iadd', 'a', 'b')), ('bcsel', ('ilt', 'a', ('iadd', 'a', 'b')), 9223372036854775808, ('iadd', 'a', 'b'))) */
   /* search1321_0 -> 21 in the cache */
   /* search1321_1 -> 41 in the cache */
   /* search1321 -> 3415 in the cache */

   /* replace1321_0_0 -> 41 in the cache */
   /* replace1321_0_1 -> 1438 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 41, 1438 },
      -1,
   } },
   /* replace1321_1_0_0_0 -> 21 in the cache */
   /* replace1321_1_0_0_1 -> 41 in the cache */
   /* replace1321_1_0_0 -> 3402 in the cache */
   /* replace1321_1_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 3402, 21 },
      -1,
   } },
   /* replace1321_1_1 -> 3426 in the cache */
   /* replace1321_1_2_0 -> 21 in the cache */
   /* replace1321_1_2_1 -> 41 in the cache */
   /* replace1321_1_2 -> 3404 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3767, 3426, 3404 },
      -1,
   } },
   /* replace1321_2_0_0 -> 21 in the cache */
   /* replace1321_2_0_1_0 -> 21 in the cache */
   /* replace1321_2_0_1_1 -> 41 in the cache */
   /* replace1321_2_0_1 -> 3418 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 1,
      { 21, 3418 },
      -1,
   } },
   /* replace1321_2_1 -> 22 in the cache */
   /* replace1321_2_2_0 -> 21 in the cache */
   /* replace1321_2_2_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      3, 1,
      { 21, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 3769, 22, 3770 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 3766, 3768, 3771 },
      -1,
   } },

   /* ('isub_sat@64', 'a', 'b') => ('bcsel', ('ilt', 'b', 0), ('bcsel', ('ilt', ('isub', 'a', 'b'), 'a'), 9223372036854775807, ('isub', 'a', 'b')), ('bcsel', ('ilt', 'a', ('isub', 'a', 'b')), 9223372036854775808, ('isub', 'a', 'b'))) */
   /* search1322_0 -> 21 in the cache */
   /* search1322_1 -> 41 in the cache */
   /* search1322 -> 3429 in the cache */

   /* replace1322_0_0 -> 41 in the cache */
   /* replace1322_0_1 -> 859 in the cache */
   /* replace1322_0 -> 3416 in the cache */
   /* replace1322_1_0_0_0 -> 21 in the cache */
   /* replace1322_1_0_0_1 -> 41 in the cache */
   /* replace1322_1_0_0 -> 3413 in the cache */
   /* replace1322_1_0_1 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3413, 21 },
      -1,
   } },
   /* replace1322_1_1 -> 3426 in the cache */
   /* replace1322_1_2_0 -> 21 in the cache */
   /* replace1322_1_2_1 -> 41 in the cache */
   /* replace1322_1_2 -> 3413 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3773, 3426, 3413 },
      -1,
   } },
   /* replace1322_2_0_0 -> 21 in the cache */
   /* replace1322_2_0_1_0 -> 21 in the cache */
   /* replace1322_2_0_1_1 -> 41 in the cache */
   /* replace1322_2_0_1 -> 3413 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 3413 },
      -1,
   } },
   /* replace1322_2_1 -> 22 in the cache */
   /* replace1322_2_2_0 -> 21 in the cache */
   /* replace1322_2_2_1 -> 41 in the cache */
   /* replace1322_2_2 -> 3413 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3775, 22, 3413 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 3416, 3774, 3776 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('fneu', 'c', 'd')) */
   /* search1323_0_0_0 -> 0 in the cache */
   /* search1323_0_0_1 -> 3 in the cache */
   /* search1323_0_0 -> 1036 in the cache */
   /* search1323_0_1_0 -> 154 in the cache */
   /* search1323_0_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      2, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1036, 3778 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 3779 },
      -1,
   } },

   /* replace1323_0_0 -> 0 in the cache */
   /* replace1323_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 0, 3 },
      -1,
   } },
   /* replace1323_1_0 -> 154 in the cache */
   /* replace1323_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      2, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3781, 3782 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('feq', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('fneu', 'c', 'd')) */
   /* search1324_0_0_0 -> 0 in the cache */
   /* search1324_0_0_1 -> 3 in the cache */
   /* search1324_0_0 -> 1036 in the cache */
   /* search1324_0_1_0 -> 154 in the cache */
   /* search1324_0_1_1 -> 155 in the cache */
   /* search1324_0_1 -> 3778 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1036, 3778 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 3784 },
      -1,
   } },

   /* replace1324_0_0 -> 0 in the cache */
   /* replace1324_0_1 -> 3 in the cache */
   /* replace1324_0 -> 3781 in the cache */
   /* replace1324_1_0 -> 154 in the cache */
   /* replace1324_1_1 -> 155 in the cache */
   /* replace1324_1 -> 3782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3781, 3782 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('fneu', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1325_0_0_0 -> 0 in the cache */
   /* search1325_0_0_1 -> 3 in the cache */
   /* search1325_0_0 -> 1036 in the cache */
   /* search1325_0_1_0 -> 154 in the cache */
   /* search1325_0_1_1 -> 155 in the cache */
   /* search1325_0_1 -> 3782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1036, 3782 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 3787 },
      -1,
   } },

   /* replace1325_0_0 -> 0 in the cache */
   /* replace1325_0_1 -> 3 in the cache */
   /* replace1325_0 -> 3781 in the cache */
   /* replace1325_1_0 -> 154 in the cache */
   /* replace1325_1_1 -> 155 in the cache */
   /* replace1325_1 -> 3778 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3781, 3778 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('feq', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('fneu', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1326_0_0_0 -> 0 in the cache */
   /* search1326_0_0_1 -> 3 in the cache */
   /* search1326_0_0 -> 1036 in the cache */
   /* search1326_0_1_0 -> 154 in the cache */
   /* search1326_0_1_1 -> 155 in the cache */
   /* search1326_0_1 -> 3782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1036, 3782 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 3790 },
      -1,
   } },

   /* replace1326_0_0 -> 0 in the cache */
   /* replace1326_0_1 -> 3 in the cache */
   /* replace1326_0 -> 3781 in the cache */
   /* replace1326_1_0 -> 154 in the cache */
   /* replace1326_1_1 -> 155 in the cache */
   /* replace1326_1 -> 3778 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3781, 3778 },
      -1,
   } },

   /* ('inot', ('ior(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('iand', ('feq', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1327_0_0_0 -> 0 in the cache */
   /* search1327_0_0_1 -> 3 in the cache */
   /* search1327_0_0 -> 3781 in the cache */
   /* search1327_0_1_0 -> 154 in the cache */
   /* search1327_0_1_1 -> 155 in the cache */
   /* search1327_0_1 -> 3782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3781, 3782 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 3793 },
      -1,
   } },

   /* replace1327_0_0 -> 0 in the cache */
   /* replace1327_0_1 -> 3 in the cache */
   /* replace1327_0 -> 1036 in the cache */
   /* replace1327_1_0 -> 154 in the cache */
   /* replace1327_1_1 -> 155 in the cache */
   /* replace1327_1 -> 3778 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 1036, 3778 },
      -1,
   } },

   /* ('inot', ('iand(is_used_once)', ('fneu', 'a', 'b'), ('fneu', 'c', 'd'))) => ('ior', ('feq', 'a', 'b'), ('feq', 'c', 'd')) */
   /* search1328_0_0_0 -> 0 in the cache */
   /* search1328_0_0_1 -> 3 in the cache */
   /* search1328_0_0 -> 3781 in the cache */
   /* search1328_0_1_0 -> 154 in the cache */
   /* search1328_0_1_1 -> 155 in the cache */
   /* search1328_0_1 -> 3782 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3781, 3782 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 3,
      { 3796 },
      -1,
   } },

   /* replace1328_0_0 -> 0 in the cache */
   /* replace1328_0_1 -> 3 in the cache */
   /* replace1328_0 -> 1036 in the cache */
   /* replace1328_1_0 -> 154 in the cache */
   /* replace1328_1_1 -> 155 in the cache */
   /* replace1328_1 -> 3778 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 1036, 3778 },
      -1,
   } },

   /* ('f2b1', ('b2f', 'a@1')) => a */
   /* search1329_0_0 -> 722 in the cache */
   /* search1329_0 -> 1085 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_f2b1,
      -1, 0,
      { 1085 },
      -1,
   } },

   /* replace1329 -> 722 in the cache */

   /* ('i2b1', ('b2i', 'a@1')) => a */
   /* search1330_0_0 -> 722 in the cache */
   /* search1330_0 -> 1183 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2b1,
      -1, 0,
      { 1183 },
      -1,
   } },

   /* replace1330 -> 722 in the cache */

   /* ('f2b8', ('b2f', 'a@8')) => a */
   /* search1331_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_f2b8,
      -1, 0,
      { 3801 },
      -1,
   } },

   /* replace1331 -> 6 in the cache */

   /* ('i2b8', ('b2i', 'a@8')) => a */
   /* search1332_0_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2i,
      -1, 0,
      { 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2b8,
      -1, 0,
      { 3803 },
      -1,
   } },

   /* replace1332 -> 6 in the cache */

   /* ('f2b16', ('b2f', 'a@16')) => a */
   /* search1333_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 11 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2b16,
      -1, 0,
      { 3805 },
      -1,
   } },

   /* replace1333 -> 11 in the cache */

   /* ('i2b16', ('b2i', 'a@16')) => a */
   /* search1334_0_0 -> 11 in the cache */
   /* search1334_0 -> 2841 in the cache */
   /* search1334 -> 2842 in the cache */

   /* replace1334 -> 11 in the cache */

   /* ('f2b32', ('b2f', 'a@32')) => a */
   /* search1335_0_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_b2f,
      -1, 0,
      { 16 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2b32,
      -1, 0,
      { 3807 },
      -1,
   } },

   /* replace1335 -> 16 in the cache */

   /* ('i2b32', ('b2i', 'a@32')) => a */
   /* search1336_0_0 -> 16 in the cache */
   /* search1336_0 -> 2843 in the cache */
   /* search1336 -> 2844 in the cache */

   /* replace1336 -> 16 in the cache */

   /* ('f2f', ('b2f', 'a')) => ('b2f', 'a') */
   /* search1337_0_0 -> 32 in the cache */
   /* search1337_0 -> 1472 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2f,
      -1, 0,
      { 1472 },
      -1,
   } },

   /* replace1337_0 -> 32 in the cache */
   /* replace1337 -> 1472 in the cache */

   /* ('f2u', ('b2f', 'a')) => ('b2i', 'a') */
   /* search1338_0_0 -> 32 in the cache */
   /* search1338_0 -> 1472 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2u,
      -1, 0,
      { 1472 },
      -1,
   } },

   /* replace1338_0 -> 32 in the cache */
   /* replace1338 -> 2833 in the cache */

   /* ('f2i', ('b2f', 'a')) => ('b2i', 'a') */
   /* search1339_0_0 -> 32 in the cache */
   /* search1339_0 -> 1472 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_f2i,
      -1, 0,
      { 1472 },
      -1,
   } },

   /* replace1339_0 -> 32 in the cache */
   /* replace1339 -> 2833 in the cache */

   /* ('u2f', ('b2i', 'a')) => ('b2f', 'a') */
   /* search1340_0_0 -> 32 in the cache */
   /* search1340_0 -> 2833 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2f,
      -1, 0,
      { 2833 },
      -1,
   } },

   /* replace1340_0 -> 32 in the cache */
   /* replace1340 -> 1472 in the cache */

   /* ('u2u', ('b2i', 'a')) => ('b2i', 'a') */
   /* search1341_0_0 -> 32 in the cache */
   /* search1341_0 -> 2833 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_u2u,
      -1, 0,
      { 2833 },
      -1,
   } },

   /* replace1341_0 -> 32 in the cache */
   /* replace1341 -> 2833 in the cache */

   /* ('i2f', ('b2i', 'a')) => ('b2f', 'a') */
   /* search1342_0_0 -> 32 in the cache */
   /* search1342_0 -> 2833 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2f,
      -1, 0,
      { 2833 },
      -1,
   } },

   /* replace1342_0 -> 32 in the cache */
   /* replace1342 -> 1472 in the cache */

   /* ('i2i', ('b2i', 'a')) => ('b2i', 'a') */
   /* search1343_0_0 -> 32 in the cache */
   /* search1343_0 -> 2833 in the cache */
   { .expression = {
      { nir_search_value_expression, 0 },
      false,
      false,
      false,
      nir_search_op_i2i,
      -1, 0,
      { 2833 },
      -1,
   } },

   /* replace1343_0 -> 32 in the cache */
   /* replace1343 -> 2833 in the cache */

   /* ('i2i1', 'a@1') => a */
   /* search1344_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 722 },
      -1,
   } },

   /* replace1344 -> 722 in the cache */

   /* ('i2i8', 'a@8') => a */
   /* search1345_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1345 -> 6 in the cache */

   /* ('i2i16', 'a@16') => a */
   /* search1346_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1346 -> 11 in the cache */

   /* ('i2i32', 'a@32') => a */
   /* search1347_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1347 -> 16 in the cache */

   /* ('i2i64', 'a@64') => a */
   /* search1348_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1348 -> 21 in the cache */

   /* ('u2u1', 'a@1') => a */
   /* search1349_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 722 },
      -1,
   } },

   /* replace1349 -> 722 in the cache */

   /* ('u2u8', 'a@8') => a */
   /* search1350_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1350 -> 6 in the cache */

   /* ('u2u16', 'a@16') => a */
   /* search1351_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1351 -> 11 in the cache */

   /* ('u2u32', 'a@32') => a */
   /* search1352_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1352 -> 16 in the cache */

   /* ('u2u64', 'a@64') => a */
   /* search1353_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1353 -> 21 in the cache */

   /* ('f2f16', 'a@16') => a */
   /* search1354_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1354 -> 11 in the cache */

   /* ('f2f32', 'a@32') => a */
   /* search1355_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1355 -> 16 in the cache */

   /* ('f2f64', 'a@64') => a */
   /* search1356_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 21 },
      -1,
   } },

   /* replace1356 -> 21 in the cache */

   /* ('b2b1', 'a@1') => a */
   /* search1357_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_b2b1,
      -1, 0,
      { 722 },
      -1,
   } },

   /* replace1357 -> 722 in the cache */

   /* ('b2b8', 'a@8') => a */
   /* search1358_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_b2b8,
      -1, 0,
      { 6 },
      -1,
   } },

   /* replace1358 -> 6 in the cache */

   /* ('b2b16', 'a@16') => a */
   /* search1359_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2b16,
      -1, 0,
      { 11 },
      -1,
   } },

   /* replace1359 -> 11 in the cache */

   /* ('b2b32', 'a@32') => a */
   /* search1360_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2b32,
      -1, 0,
      { 16 },
      -1,
   } },

   /* replace1360 -> 16 in the cache */

   /* ('i2i1', ('i2i8', 'a')) => ('i2i1', 'a') */
   /* search1361_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3833 },
      -1,
   } },

   /* replace1361_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('i2i1', ('u2u8', 'a')) => ('u2u1', 'a') */
   /* search1362_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3836 },
      -1,
   } },

   /* replace1362_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 32 },
      -1,
   } },

   /* ('u2u1', ('i2i8', 'a')) => ('i2i1', 'a') */
   /* search1363_0_0 -> 32 in the cache */
   /* search1363_0 -> 3833 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3833 },
      -1,
   } },

   /* replace1363_0 -> 32 in the cache */
   /* replace1363 -> 3835 in the cache */

   /* ('u2u1', ('u2u8', 'a')) => ('u2u1', 'a') */
   /* search1364_0_0 -> 32 in the cache */
   /* search1364_0 -> 3836 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3836 },
      -1,
   } },

   /* replace1364_0 -> 32 in the cache */
   /* replace1364 -> 3838 in the cache */

   /* ('i2i1', ('i2i16', 'a')) => ('i2i1', 'a') */
   /* search1365_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3841 },
      -1,
   } },

   /* replace1365_0 -> 32 in the cache */
   /* replace1365 -> 3835 in the cache */

   /* ('i2i1', ('u2u16', 'a')) => ('u2u1', 'a') */
   /* search1366_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3843 },
      -1,
   } },

   /* replace1366_0 -> 32 in the cache */
   /* replace1366 -> 3838 in the cache */

   /* ('u2u1', ('i2i16', 'a')) => ('i2i1', 'a') */
   /* search1367_0_0 -> 32 in the cache */
   /* search1367_0 -> 3841 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3841 },
      -1,
   } },

   /* replace1367_0 -> 32 in the cache */
   /* replace1367 -> 3835 in the cache */

   /* ('u2u1', ('u2u16', 'a')) => ('u2u1', 'a') */
   /* search1368_0_0 -> 32 in the cache */
   /* search1368_0 -> 3843 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3843 },
      -1,
   } },

   /* replace1368_0 -> 32 in the cache */
   /* replace1368 -> 3838 in the cache */

   /* ('i2i1', ('i2i32', 'a')) => ('i2i1', 'a') */
   /* search1369_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3847 },
      -1,
   } },

   /* replace1369_0 -> 32 in the cache */
   /* replace1369 -> 3835 in the cache */

   /* ('i2i1', ('u2u32', 'a')) => ('u2u1', 'a') */
   /* search1370_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3849 },
      -1,
   } },

   /* replace1370_0 -> 32 in the cache */
   /* replace1370 -> 3838 in the cache */

   /* ('u2u1', ('i2i32', 'a')) => ('i2i1', 'a') */
   /* search1371_0_0 -> 32 in the cache */
   /* search1371_0 -> 3847 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3847 },
      -1,
   } },

   /* replace1371_0 -> 32 in the cache */
   /* replace1371 -> 3835 in the cache */

   /* ('u2u1', ('u2u32', 'a')) => ('u2u1', 'a') */
   /* search1372_0_0 -> 32 in the cache */
   /* search1372_0 -> 3849 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3849 },
      -1,
   } },

   /* replace1372_0 -> 32 in the cache */
   /* replace1372 -> 3838 in the cache */

   /* ('i2i1', ('i2i64', 'a')) => ('i2i1', 'a') */
   /* search1373_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1373_0 -> 32 in the cache */
   /* replace1373 -> 3835 in the cache */

   /* ('i2i1', ('u2u64', 'a')) => ('u2u1', 'a') */
   /* search1374_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_i2i1,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1374_0 -> 32 in the cache */
   /* replace1374 -> 3838 in the cache */

   /* ('u2u1', ('i2i64', 'a')) => ('i2i1', 'a') */
   /* search1375_0_0 -> 32 in the cache */
   /* search1375_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1375_0 -> 32 in the cache */
   /* replace1375 -> 3835 in the cache */

   /* ('u2u1', ('u2u64', 'a')) => ('u2u1', 'a') */
   /* search1376_0_0 -> 32 in the cache */
   /* search1376_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_u2u1,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1376_0 -> 32 in the cache */
   /* replace1376 -> 3838 in the cache */

   /* ('i2i8', ('i2i1', 'a@1')) => ('i2i8', 'a') */
   /* search1377_0_0 -> 722 in the cache */
   /* search1377_0 -> 3816 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3816 },
      -1,
   } },

   /* replace1377_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('u2u8', ('u2u1', 'a@1')) => ('u2u8', 'a') */
   /* search1378_0_0 -> 722 in the cache */
   /* search1378_0 -> 3821 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3821 },
      -1,
   } },

   /* replace1378_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('i2i8', ('i2i16', 'a')) => ('i2i8', 'a') */
   /* search1379_0_0 -> 32 in the cache */
   /* search1379_0 -> 3841 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3841 },
      -1,
   } },

   /* replace1379_0 -> 32 in the cache */
   /* replace1379 -> 3833 in the cache */

   /* ('i2i8', ('u2u16', 'a')) => ('u2u8', 'a') */
   /* search1380_0_0 -> 32 in the cache */
   /* search1380_0 -> 3843 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3843 },
      -1,
   } },

   /* replace1380_0 -> 32 in the cache */
   /* replace1380 -> 3836 in the cache */

   /* ('u2u8', ('i2i16', 'a')) => ('i2i8', 'a') */
   /* search1381_0_0 -> 32 in the cache */
   /* search1381_0 -> 3841 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3841 },
      -1,
   } },

   /* replace1381_0 -> 32 in the cache */
   /* replace1381 -> 3833 in the cache */

   /* ('u2u8', ('u2u16', 'a')) => ('u2u8', 'a') */
   /* search1382_0_0 -> 32 in the cache */
   /* search1382_0 -> 3843 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3843 },
      -1,
   } },

   /* replace1382_0 -> 32 in the cache */
   /* replace1382 -> 3836 in the cache */

   /* ('i2i8', ('i2i32', 'a')) => ('i2i8', 'a') */
   /* search1383_0_0 -> 32 in the cache */
   /* search1383_0 -> 3847 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3847 },
      -1,
   } },

   /* replace1383_0 -> 32 in the cache */
   /* replace1383 -> 3833 in the cache */

   /* ('i2i8', ('u2u32', 'a')) => ('u2u8', 'a') */
   /* search1384_0_0 -> 32 in the cache */
   /* search1384_0 -> 3849 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3849 },
      -1,
   } },

   /* replace1384_0 -> 32 in the cache */
   /* replace1384 -> 3836 in the cache */

   /* ('u2u8', ('i2i32', 'a')) => ('i2i8', 'a') */
   /* search1385_0_0 -> 32 in the cache */
   /* search1385_0 -> 3847 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3847 },
      -1,
   } },

   /* replace1385_0 -> 32 in the cache */
   /* replace1385 -> 3833 in the cache */

   /* ('u2u8', ('u2u32', 'a')) => ('u2u8', 'a') */
   /* search1386_0_0 -> 32 in the cache */
   /* search1386_0 -> 3849 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3849 },
      -1,
   } },

   /* replace1386_0 -> 32 in the cache */
   /* replace1386 -> 3836 in the cache */

   /* ('i2i8', ('i2i64', 'a')) => ('i2i8', 'a') */
   /* search1387_0_0 -> 32 in the cache */
   /* search1387_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1387_0 -> 32 in the cache */
   /* replace1387 -> 3833 in the cache */

   /* ('i2i8', ('u2u64', 'a')) => ('u2u8', 'a') */
   /* search1388_0_0 -> 32 in the cache */
   /* search1388_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1388_0 -> 32 in the cache */
   /* replace1388 -> 3836 in the cache */

   /* ('u2u8', ('i2i64', 'a')) => ('i2i8', 'a') */
   /* search1389_0_0 -> 32 in the cache */
   /* search1389_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1389_0 -> 32 in the cache */
   /* replace1389 -> 3833 in the cache */

   /* ('u2u8', ('u2u64', 'a')) => ('u2u8', 'a') */
   /* search1390_0_0 -> 32 in the cache */
   /* search1390_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1390_0 -> 32 in the cache */
   /* replace1390 -> 3836 in the cache */

   /* ('i2i16', ('i2i1', 'a@1')) => ('i2i16', 'a') */
   /* search1391_0_0 -> 722 in the cache */
   /* search1391_0 -> 3816 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3816 },
      -1,
   } },

   /* replace1391_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('u2u16', ('u2u1', 'a@1')) => ('u2u16', 'a') */
   /* search1392_0_0 -> 722 in the cache */
   /* search1392_0 -> 3821 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3821 },
      -1,
   } },

   /* replace1392_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('i2i16', ('i2i8', 'a@1')) => ('i2i16', 'a') */
   /* search1393_0_0 -> 722 in the cache */
   /* search1393_0 -> 3860 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3860 },
      -1,
   } },

   /* replace1393_0 -> 722 in the cache */
   /* replace1393 -> 3876 in the cache */

   /* ('u2u16', ('u2u8', 'a@1')) => ('u2u16', 'a') */
   /* search1394_0_0 -> 722 in the cache */
   /* search1394_0 -> 3862 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3862 },
      -1,
   } },

   /* replace1394_0 -> 722 in the cache */
   /* replace1394 -> 3878 in the cache */

   /* ('i2i16', ('i2i8', 'a@8')) => ('i2i16', 'a') */
   /* search1395_0_0 -> 6 in the cache */
   /* search1395_0 -> 3817 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3817 },
      -1,
   } },

   /* replace1395_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2u16', ('u2u8', 'a@8')) => ('u2u16', 'a') */
   /* search1396_0_0 -> 6 in the cache */
   /* search1396_0 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3822 },
      -1,
   } },

   /* replace1396_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2i16', ('i2i32', 'a')) => ('i2i16', 'a') */
   /* search1397_0_0 -> 32 in the cache */
   /* search1397_0 -> 3847 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3847 },
      -1,
   } },

   /* replace1397_0 -> 32 in the cache */
   /* replace1397 -> 3841 in the cache */

   /* ('i2i16', ('u2u32', 'a')) => ('u2u16', 'a') */
   /* search1398_0_0 -> 32 in the cache */
   /* search1398_0 -> 3849 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3849 },
      -1,
   } },

   /* replace1398_0 -> 32 in the cache */
   /* replace1398 -> 3843 in the cache */

   /* ('u2u16', ('i2i32', 'a')) => ('i2i16', 'a') */
   /* search1399_0_0 -> 32 in the cache */
   /* search1399_0 -> 3847 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3847 },
      -1,
   } },

   /* replace1399_0 -> 32 in the cache */
   /* replace1399 -> 3841 in the cache */

   /* ('u2u16', ('u2u32', 'a')) => ('u2u16', 'a') */
   /* search1400_0_0 -> 32 in the cache */
   /* search1400_0 -> 3849 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3849 },
      -1,
   } },

   /* replace1400_0 -> 32 in the cache */
   /* replace1400 -> 3843 in the cache */

   /* ('i2i16', ('i2i64', 'a')) => ('i2i16', 'a') */
   /* search1401_0_0 -> 32 in the cache */
   /* search1401_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1401_0 -> 32 in the cache */
   /* replace1401 -> 3841 in the cache */

   /* ('i2i16', ('u2u64', 'a')) => ('u2u16', 'a') */
   /* search1402_0_0 -> 32 in the cache */
   /* search1402_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1402_0 -> 32 in the cache */
   /* replace1402 -> 3843 in the cache */

   /* ('u2u16', ('i2i64', 'a')) => ('i2i16', 'a') */
   /* search1403_0_0 -> 32 in the cache */
   /* search1403_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1403_0 -> 32 in the cache */
   /* replace1403 -> 3841 in the cache */

   /* ('u2u16', ('u2u64', 'a')) => ('u2u16', 'a') */
   /* search1404_0_0 -> 32 in the cache */
   /* search1404_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1404_0 -> 32 in the cache */
   /* replace1404 -> 3843 in the cache */

   /* ('i2i32', ('i2i1', 'a@1')) => ('i2i32', 'a') */
   /* search1405_0_0 -> 722 in the cache */
   /* search1405_0 -> 3816 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3816 },
      -1,
   } },

   /* replace1405_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('u2u32', ('u2u1', 'a@1')) => ('u2u32', 'a') */
   /* search1406_0_0 -> 722 in the cache */
   /* search1406_0 -> 3821 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3821 },
      -1,
   } },

   /* replace1406_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('i2i32', ('i2i8', 'a@1')) => ('i2i32', 'a') */
   /* search1407_0_0 -> 722 in the cache */
   /* search1407_0 -> 3860 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3860 },
      -1,
   } },

   /* replace1407_0 -> 722 in the cache */
   /* replace1407 -> 3894 in the cache */

   /* ('u2u32', ('u2u8', 'a@1')) => ('u2u32', 'a') */
   /* search1408_0_0 -> 722 in the cache */
   /* search1408_0 -> 3862 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3862 },
      -1,
   } },

   /* replace1408_0 -> 722 in the cache */
   /* replace1408 -> 3896 in the cache */

   /* ('i2i32', ('i2i8', 'a@8')) => ('i2i32', 'a') */
   /* search1409_0_0 -> 6 in the cache */
   /* search1409_0 -> 3817 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3817 },
      -1,
   } },

   /* replace1409_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2u32', ('u2u8', 'a@8')) => ('u2u32', 'a') */
   /* search1410_0_0 -> 6 in the cache */
   /* search1410_0 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3822 },
      -1,
   } },

   /* replace1410_0 -> 6 in the cache */
   /* replace1410 -> 3009 in the cache */

   /* ('i2i32', ('i2i16', 'a@1')) => ('i2i32', 'a') */
   /* search1411_0_0 -> 722 in the cache */
   /* search1411_0 -> 3876 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3876 },
      -1,
   } },

   /* replace1411_0 -> 722 in the cache */
   /* replace1411 -> 3894 in the cache */

   /* ('u2u32', ('u2u16', 'a@1')) => ('u2u32', 'a') */
   /* search1412_0_0 -> 722 in the cache */
   /* search1412_0 -> 3878 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3878 },
      -1,
   } },

   /* replace1412_0 -> 722 in the cache */
   /* replace1412 -> 3896 in the cache */

   /* ('i2i32', ('i2i16', 'a@8')) => ('i2i32', 'a') */
   /* search1413_0_0 -> 6 in the cache */
   /* search1413_0 -> 3882 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3882 },
      -1,
   } },

   /* replace1413_0 -> 6 in the cache */
   /* replace1413 -> 3900 in the cache */

   /* ('u2u32', ('u2u16', 'a@8')) => ('u2u32', 'a') */
   /* search1414_0_0 -> 6 in the cache */
   /* search1414_0 -> 3884 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3884 },
      -1,
   } },

   /* replace1414_0 -> 6 in the cache */
   /* replace1414 -> 3009 in the cache */

   /* ('i2i32', ('i2i16', 'a@16')) => ('i2i32', 'a') */
   /* search1415_0_0 -> 11 in the cache */
   /* search1415_0 -> 3818 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3818 },
      -1,
   } },

   /* replace1415_0 -> 11 in the cache */
   /* replace1415 -> 1747 in the cache */

   /* ('u2u32', ('u2u16', 'a@16')) => ('u2u32', 'a') */
   /* search1416_0_0 -> 11 in the cache */
   /* search1416_0 -> 3823 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3823 },
      -1,
   } },

   /* replace1416_0 -> 11 in the cache */
   /* replace1416 -> 1749 in the cache */

   /* ('i2i32', ('i2i64', 'a')) => ('i2i32', 'a') */
   /* search1417_0_0 -> 32 in the cache */
   /* search1417_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1417_0 -> 32 in the cache */
   /* replace1417 -> 3847 in the cache */

   /* ('i2i32', ('u2u64', 'a')) => ('u2u32', 'a') */
   /* search1418_0_0 -> 32 in the cache */
   /* search1418_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1418_0 -> 32 in the cache */
   /* replace1418 -> 3849 in the cache */

   /* ('u2u32', ('i2i64', 'a')) => ('i2i32', 'a') */
   /* search1419_0_0 -> 32 in the cache */
   /* search1419_0 -> 3853 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3853 },
      -1,
   } },

   /* replace1419_0 -> 32 in the cache */
   /* replace1419 -> 3847 in the cache */

   /* ('u2u32', ('u2u64', 'a')) => ('u2u32', 'a') */
   /* search1420_0_0 -> 32 in the cache */
   /* search1420_0 -> 3855 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3855 },
      -1,
   } },

   /* replace1420_0 -> 32 in the cache */
   /* replace1420 -> 3849 in the cache */

   /* ('i2i64', ('i2i1', 'a@1')) => ('i2i64', 'a') */
   /* search1421_0_0 -> 722 in the cache */
   /* search1421_0 -> 3816 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3816 },
      -1,
   } },

   /* replace1421_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('u2u64', ('u2u1', 'a@1')) => ('u2u64', 'a') */
   /* search1422_0_0 -> 722 in the cache */
   /* search1422_0 -> 3821 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3821 },
      -1,
   } },

   /* replace1422_0 -> 722 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 722 },
      -1,
   } },

   /* ('i2i64', ('i2i8', 'a@1')) => ('i2i64', 'a') */
   /* search1423_0_0 -> 722 in the cache */
   /* search1423_0 -> 3860 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3860 },
      -1,
   } },

   /* replace1423_0 -> 722 in the cache */
   /* replace1423 -> 3913 in the cache */

   /* ('u2u64', ('u2u8', 'a@1')) => ('u2u64', 'a') */
   /* search1424_0_0 -> 722 in the cache */
   /* search1424_0 -> 3862 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3862 },
      -1,
   } },

   /* replace1424_0 -> 722 in the cache */
   /* replace1424 -> 3915 in the cache */

   /* ('i2i64', ('i2i8', 'a@8')) => ('i2i64', 'a') */
   /* search1425_0_0 -> 6 in the cache */
   /* search1425_0 -> 3817 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3817 },
      -1,
   } },

   /* replace1425_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('u2u64', ('u2u8', 'a@8')) => ('u2u64', 'a') */
   /* search1426_0_0 -> 6 in the cache */
   /* search1426_0 -> 3822 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3822 },
      -1,
   } },

   /* replace1426_0 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 6 },
      -1,
   } },

   /* ('i2i64', ('i2i16', 'a@1')) => ('i2i64', 'a') */
   /* search1427_0_0 -> 722 in the cache */
   /* search1427_0 -> 3876 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3876 },
      -1,
   } },

   /* replace1427_0 -> 722 in the cache */
   /* replace1427 -> 3913 in the cache */

   /* ('u2u64', ('u2u16', 'a@1')) => ('u2u64', 'a') */
   /* search1428_0_0 -> 722 in the cache */
   /* search1428_0 -> 3878 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3878 },
      -1,
   } },

   /* replace1428_0 -> 722 in the cache */
   /* replace1428 -> 3915 in the cache */

   /* ('i2i64', ('i2i16', 'a@8')) => ('i2i64', 'a') */
   /* search1429_0_0 -> 6 in the cache */
   /* search1429_0 -> 3882 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3882 },
      -1,
   } },

   /* replace1429_0 -> 6 in the cache */
   /* replace1429 -> 3919 in the cache */

   /* ('u2u64', ('u2u16', 'a@8')) => ('u2u64', 'a') */
   /* search1430_0_0 -> 6 in the cache */
   /* search1430_0 -> 3884 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3884 },
      -1,
   } },

   /* replace1430_0 -> 6 in the cache */
   /* replace1430 -> 3921 in the cache */

   /* ('i2i64', ('i2i16', 'a@16')) => ('i2i64', 'a') */
   /* search1431_0_0 -> 11 in the cache */
   /* search1431_0 -> 3818 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3818 },
      -1,
   } },

   /* replace1431_0 -> 11 in the cache */
   /* replace1431 -> 1770 in the cache */

   /* ('u2u64', ('u2u16', 'a@16')) => ('u2u64', 'a') */
   /* search1432_0_0 -> 11 in the cache */
   /* search1432_0 -> 3823 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3823 },
      -1,
   } },

   /* replace1432_0 -> 11 in the cache */
   /* replace1432 -> 1772 in the cache */

   /* ('i2i64', ('i2i32', 'a@1')) => ('i2i64', 'a') */
   /* search1433_0_0 -> 722 in the cache */
   /* search1433_0 -> 3894 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3894 },
      -1,
   } },

   /* replace1433_0 -> 722 in the cache */
   /* replace1433 -> 3913 in the cache */

   /* ('u2u64', ('u2u32', 'a@1')) => ('u2u64', 'a') */
   /* search1434_0_0 -> 722 in the cache */
   /* search1434_0 -> 3896 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3896 },
      -1,
   } },

   /* replace1434_0 -> 722 in the cache */
   /* replace1434 -> 3915 in the cache */

   /* ('i2i64', ('i2i32', 'a@8')) => ('i2i64', 'a') */
   /* search1435_0_0 -> 6 in the cache */
   /* search1435_0 -> 3900 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3900 },
      -1,
   } },

   /* replace1435_0 -> 6 in the cache */
   /* replace1435 -> 3919 in the cache */

   /* ('u2u64', ('u2u32', 'a@8')) => ('u2u64', 'a') */
   /* search1436_0_0 -> 6 in the cache */
   /* search1436_0 -> 3009 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3009 },
      -1,
   } },

   /* replace1436_0 -> 6 in the cache */
   /* replace1436 -> 3921 in the cache */

   /* ('i2i64', ('i2i32', 'a@16')) => ('i2i64', 'a') */
   /* search1437_0_0 -> 11 in the cache */
   /* search1437_0 -> 1747 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 1747 },
      -1,
   } },

   /* replace1437_0 -> 11 in the cache */
   /* replace1437 -> 1770 in the cache */

   /* ('u2u64', ('u2u32', 'a@16')) => ('u2u64', 'a') */
   /* search1438_0_0 -> 11 in the cache */
   /* search1438_0 -> 1749 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 1749 },
      -1,
   } },

   /* replace1438_0 -> 11 in the cache */
   /* replace1438 -> 1772 in the cache */

   /* ('i2i64', ('i2i32', 'a@32')) => ('i2i64', 'a') */
   /* search1439_0_0 -> 16 in the cache */
   /* search1439_0 -> 3819 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3819 },
      -1,
   } },

   /* replace1439_0 -> 16 in the cache */
   /* replace1439 -> 1854 in the cache */

   /* ('u2u64', ('u2u32', 'a@32')) => ('u2u64', 'a') */
   /* search1440_0_0 -> 16 in the cache */
   /* search1440_0 -> 3824 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3824 },
      -1,
   } },

   /* replace1440_0 -> 16 in the cache */
   /* replace1440 -> 1856 in the cache */

   /* ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a') */
   /* search1441_0_0 -> 16 in the cache */
   /* search1441_0_1 -> 36 in the cache */
   /* search1441_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1441_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2i8', ('pack_64_2x32_split', 'a', 'b')) => ('i2i8', 'a') */
   /* search1442_0_0 -> 16 in the cache */
   /* search1442_0_1 -> 36 in the cache */
   /* search1442_0 -> 2933 in the cache */
   /* search1442 -> 3936 in the cache */

   /* replace1442_0 -> 16 in the cache */
   /* replace1442 -> 3937 in the cache */

   /* ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a') */
   /* search1443_0_0 -> 16 in the cache */
   /* search1443_0_1 -> 36 in the cache */
   /* search1443_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1443_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 16 },
      -1,
   } },

   /* ('i2i16', ('pack_64_2x32_split', 'a', 'b')) => ('i2i16', 'a') */
   /* search1444_0_0 -> 16 in the cache */
   /* search1444_0_1 -> 36 in the cache */
   /* search1444_0 -> 2933 in the cache */
   /* search1444 -> 3938 in the cache */

   /* replace1444_0 -> 16 in the cache */
   /* replace1444 -> 3939 in the cache */

   /* ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a') */
   /* search1445_0_0 -> 16 in the cache */
   /* search1445_0_1 -> 36 in the cache */
   /* search1445_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1445_0 -> 16 in the cache */
   /* replace1445 -> 3819 in the cache */

   /* ('i2i32', ('pack_64_2x32_split', 'a', 'b')) => ('i2i32', 'a') */
   /* search1446_0_0 -> 16 in the cache */
   /* search1446_0_1 -> 36 in the cache */
   /* search1446_0 -> 2933 in the cache */
   /* search1446 -> 3940 in the cache */

   /* replace1446_0 -> 16 in the cache */
   /* replace1446 -> 3819 in the cache */

   /* ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a') */
   /* search1447_0_0 -> 16 in the cache */
   /* search1447_0_1 -> 36 in the cache */
   /* search1447_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1447_0 -> 16 in the cache */
   /* replace1447 -> 1973 in the cache */

   /* ('u2u8', ('pack_64_2x32_split', 'a', 'b')) => ('u2u8', 'a') */
   /* search1448_0_0 -> 16 in the cache */
   /* search1448_0_1 -> 36 in the cache */
   /* search1448_0 -> 2933 in the cache */
   /* search1448 -> 3941 in the cache */

   /* replace1448_0 -> 16 in the cache */
   /* replace1448 -> 1973 in the cache */

   /* ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a') */
   /* search1449_0_0 -> 16 in the cache */
   /* search1449_0_1 -> 36 in the cache */
   /* search1449_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1449_0 -> 16 in the cache */
   /* replace1449 -> 3037 in the cache */

   /* ('u2u16', ('pack_64_2x32_split', 'a', 'b')) => ('u2u16', 'a') */
   /* search1450_0_0 -> 16 in the cache */
   /* search1450_0_1 -> 36 in the cache */
   /* search1450_0 -> 2933 in the cache */
   /* search1450 -> 3942 in the cache */

   /* replace1450_0 -> 16 in the cache */
   /* replace1450 -> 3037 in the cache */

   /* ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a') */
   /* search1451_0_0 -> 16 in the cache */
   /* search1451_0_1 -> 36 in the cache */
   /* search1451_0 -> 2933 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 2933 },
      -1,
   } },

   /* replace1451_0 -> 16 in the cache */
   /* replace1451 -> 3824 in the cache */

   /* ('u2u32', ('pack_64_2x32_split', 'a', 'b')) => ('u2u32', 'a') */
   /* search1452_0_0 -> 16 in the cache */
   /* search1452_0_1 -> 36 in the cache */
   /* search1452_0 -> 2933 in the cache */
   /* search1452 -> 3943 in the cache */

   /* replace1452_0 -> 16 in the cache */
   /* replace1452 -> 3824 in the cache */

   /* ('ieq', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1453_0_0 -> 6 in the cache */
   /* search1453_0 -> 3882 in the cache */
   /* search1453_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3882, 3944 },
      -1,
   } },

   /* replace1453_0 -> 6 in the cache */
   /* replace1453_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 6, 3946 },
      -1,
   } },

   /* ('ine', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1454_0_0 -> 6 in the cache */
   /* search1454_0 -> 3882 in the cache */
   /* search1454_1_0 -> 2060 in the cache */
   /* search1454_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3882, 3944 },
      -1,
   } },

   /* replace1454_0 -> 6 in the cache */
   /* replace1454_1_0 -> 2060 in the cache */
   /* replace1454_1 -> 3946 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 6, 3946 },
      -1,
   } },

   /* ('ige', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1455_0_0 -> 6 in the cache */
   /* search1455_0 -> 3882 in the cache */
   /* search1455_1_0 -> 2060 in the cache */
   /* search1455_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3882, 3944 },
      -1,
   } },

   /* replace1455_0 -> 6 in the cache */
   /* replace1455_1_0 -> 2060 in the cache */
   /* replace1455_1 -> 3946 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 3946 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'a@8'), ('i2i16', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1456_0_0 -> 6 in the cache */
   /* search1456_0 -> 3882 in the cache */
   /* search1456_1_0 -> 2060 in the cache */
   /* search1456_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3882, 3944 },
      -1,
   } },

   /* replace1456_0 -> 6 in the cache */
   /* replace1456_1_0 -> 2060 in the cache */
   /* replace1456_1 -> 3946 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 3946 },
      -1,
   } },

   /* ('ige', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3954 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3955, 3957 },
      -1,
   } },

   /* replace1457_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* replace1457_1 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3959, 3956 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'b@8'), ('i2i16', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1458_0_0 -> 3954 in the cache */
   /* search1458_0 -> 3955 in the cache */
   /* search1458_1_0 -> 3956 in the cache */
   /* search1458_1 -> 3957 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3955, 3957 },
      -1,
   } },

   /* replace1458_0_0 -> 3954 in the cache */
   /* replace1458_0 -> 3959 in the cache */
   /* replace1458_1 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3959, 3956 },
      -1,
   } },

   /* ('ieq', ('i2i16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i16', ('i2i8', 'b')), 'b')) */
   /* search1459_0_0 -> 6 in the cache */
   /* search1459_0 -> 3882 in the cache */
   /* search1459_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3882, 1959 },
      -1,
   } },

   /* replace1459_0_0 -> 6 in the cache */
   /* replace1459_0_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 3964 },
      -1,
   } },
   /* replace1459_1_0_0_0 -> 247 in the cache */
   /* replace1459_1_0_0 -> 3964 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3964 },
      -1,
   } },
   /* replace1459_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 3966, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 3965, 3967 },
      -1,
   } },

   /* ('ine', ('i2i16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i16', ('i2i8', 'b')), 'b')) */
   /* search1460_0_0 -> 6 in the cache */
   /* search1460_0 -> 3882 in the cache */
   /* search1460_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3882, 1959 },
      -1,
   } },

   /* replace1460_0_0 -> 6 in the cache */
   /* replace1460_0_1_0 -> 247 in the cache */
   /* replace1460_0_1 -> 3964 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 3964 },
      -1,
   } },
   /* replace1460_1_0_0_0 -> 247 in the cache */
   /* replace1460_1_0_0 -> 3964 in the cache */
   /* replace1460_1_0 -> 3966 in the cache */
   /* replace1460_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 3966, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 3970, 3971 },
      -1,
   } },

   /* ('ilt', ('i2i16', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1461_0_0 -> 6 in the cache */
   /* search1461_0 -> 3882 in the cache */
   /* search1461_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3882, 1959 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1461_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3974, 247 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1461_1_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3976, 247 },
      -1,
   } },
   /* replace1461_1_1_0 -> 6 in the cache */
   /* replace1461_1_1_1_0 -> 247 in the cache */
   /* replace1461_1_1_1 -> 3964 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 3964 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3977, 3978 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3975, 3979 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i16', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1462_1_0 -> 2060 in the cache */
   /* search1462_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3981, 3944 },
      -1,
   } },

   /* replace1462_0_0 -> 11 in the cache */
   /* replace1462_0_1 -> 3976 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 3976 },
      -1,
   } },
   /* replace1462_1_0_0 -> 11 in the cache */
   /* replace1462_1_0_1 -> 3974 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 3974 },
      -1,
   } },
   /* replace1462_1_1_0_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 11 },
      -1,
   } },
   /* replace1462_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3985, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3984, 3986 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3983, 3987 },
      -1,
   } },

   /* ('ige', ('i2i16', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1463_0_0 -> 6 in the cache */
   /* search1463_0 -> 3882 in the cache */
   /* search1463_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3882, 1959 },
      -1,
   } },

   /* replace1463_0_0 -> 3976 in the cache */
   /* replace1463_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3976, 247 },
      -1,
   } },
   /* replace1463_1_0_0 -> 3974 in the cache */
   /* replace1463_1_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3974, 247 },
      -1,
   } },
   /* replace1463_1_1_0 -> 6 in the cache */
   /* replace1463_1_1_1_0 -> 247 in the cache */
   /* replace1463_1_1_1 -> 3964 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 3964 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3991, 3992 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3990, 3993 },
      -1,
   } },

   /* ('ige', '#a', ('i2i16', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1464_0 -> 3981 in the cache */
   /* search1464_1_0 -> 2060 in the cache */
   /* search1464_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3981, 3944 },
      -1,
   } },

   /* replace1464_0_0 -> 11 in the cache */
   /* replace1464_0_1 -> 3974 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 3974 },
      -1,
   } },
   /* replace1464_1_0_0 -> 11 in the cache */
   /* replace1464_1_0_1 -> 3976 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 3976 },
      -1,
   } },
   /* replace1464_1_1_0_0 -> 11 in the cache */
   /* replace1464_1_1_0 -> 3985 in the cache */
   /* replace1464_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3985, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 3997, 3998 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 3996, 3999 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1465_0_0 -> 6 in the cache */
   /* search1465_0 -> 3900 in the cache */
   /* search1465_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3900, 4001 },
      -1,
   } },

   /* replace1465_0 -> 6 in the cache */
   /* replace1465_1_0 -> 2060 in the cache */
   /* replace1465_1 -> 3946 in the cache */
   /* replace1465 -> 3947 in the cache */

   /* ('ine', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1466_0_0 -> 6 in the cache */
   /* search1466_0 -> 3900 in the cache */
   /* search1466_1_0 -> 2060 in the cache */
   /* search1466_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3900, 4001 },
      -1,
   } },

   /* replace1466_0 -> 6 in the cache */
   /* replace1466_1_0 -> 2060 in the cache */
   /* replace1466_1 -> 3946 in the cache */
   /* replace1466 -> 3949 in the cache */

   /* ('ige', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1467_0_0 -> 6 in the cache */
   /* search1467_0 -> 3900 in the cache */
   /* search1467_1_0 -> 2060 in the cache */
   /* search1467_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3900, 4001 },
      -1,
   } },

   /* replace1467_0 -> 6 in the cache */
   /* replace1467_1_0 -> 2060 in the cache */
   /* replace1467_1 -> 3946 in the cache */
   /* replace1467 -> 3951 in the cache */

   /* ('ilt', ('i2i32', 'a@8'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1468_0_0 -> 6 in the cache */
   /* search1468_0 -> 3900 in the cache */
   /* search1468_1_0 -> 2060 in the cache */
   /* search1468_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3900, 4001 },
      -1,
   } },

   /* replace1468_0 -> 6 in the cache */
   /* replace1468_1_0 -> 2060 in the cache */
   /* replace1468_1 -> 3946 in the cache */
   /* replace1468 -> 3953 in the cache */

   /* ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   /* search1469_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* search1469_1_0 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4006, 4007 },
      -1,
   } },

   /* replace1469_0_0 -> 3954 in the cache */
   /* replace1469_0 -> 3959 in the cache */
   /* replace1469_1 -> 3956 in the cache */
   /* replace1469 -> 3960 in the cache */

   /* ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1470_0_0 -> 3954 in the cache */
   /* search1470_0 -> 4006 in the cache */
   /* search1470_1_0 -> 3956 in the cache */
   /* search1470_1 -> 4007 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4006, 4007 },
      -1,
   } },

   /* replace1470_0_0 -> 3954 in the cache */
   /* replace1470_0 -> 3959 in the cache */
   /* replace1470_1 -> 3956 in the cache */
   /* replace1470 -> 3962 in the cache */

   /* ('ieq', ('i2i32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i32', ('i2i8', 'b')), 'b')) */
   /* search1471_0_0 -> 6 in the cache */
   /* search1471_0 -> 3900 in the cache */
   /* search1471_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3900, 33 },
      -1,
   } },

   /* replace1471_0_0 -> 6 in the cache */
   /* replace1471_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4011 },
      -1,
   } },
   /* replace1471_1_0_0_0 -> 36 in the cache */
   /* replace1471_1_0_0 -> 4011 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4011 },
      -1,
   } },
   /* replace1471_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4013, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4012, 4014 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i32', ('i2i8', 'b')), 'b')) */
   /* search1472_0_0 -> 6 in the cache */
   /* search1472_0 -> 3900 in the cache */
   /* search1472_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3900, 33 },
      -1,
   } },

   /* replace1472_0_0 -> 6 in the cache */
   /* replace1472_0_1_0 -> 36 in the cache */
   /* replace1472_0_1 -> 4011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4011 },
      -1,
   } },
   /* replace1472_1_0_0_0 -> 36 in the cache */
   /* replace1472_1_0_0 -> 4011 in the cache */
   /* replace1472_1_0 -> 4013 in the cache */
   /* replace1472_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4013, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4017, 4018 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1473_0_0 -> 6 in the cache */
   /* search1473_0 -> 3900 in the cache */
   /* search1473_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3900, 33 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1473_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4021, 36 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1473_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4023, 36 },
      -1,
   } },
   /* replace1473_1_1_0 -> 6 in the cache */
   /* replace1473_1_1_1_0 -> 36 in the cache */
   /* replace1473_1_1_1 -> 4011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4011 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4024, 4025 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4022, 4026 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i32', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   /* search1474_0 -> 275 in the cache */
   /* search1474_1_0 -> 2060 in the cache */
   /* search1474_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 275, 4001 },
      -1,
   } },

   /* replace1474_0_0 -> 16 in the cache */
   /* replace1474_0_1 -> 4023 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4023 },
      -1,
   } },
   /* replace1474_1_0_0 -> 16 in the cache */
   /* replace1474_1_0_1 -> 4021 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4021 },
      -1,
   } },
   /* replace1474_1_1_0_0 -> 16 in the cache */
   /* replace1474_1_1_0 -> 3937 in the cache */
   /* replace1474_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3937, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4030, 4031 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4029, 4032 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1475_0_0 -> 6 in the cache */
   /* search1475_0 -> 3900 in the cache */
   /* search1475_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3900, 33 },
      -1,
   } },

   /* replace1475_0_0 -> 4023 in the cache */
   /* replace1475_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4023, 36 },
      -1,
   } },
   /* replace1475_1_0_0 -> 4021 in the cache */
   /* replace1475_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4021, 36 },
      -1,
   } },
   /* replace1475_1_1_0 -> 6 in the cache */
   /* replace1475_1_1_1_0 -> 36 in the cache */
   /* replace1475_1_1_1 -> 4011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4011 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4036, 4037 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4035, 4038 },
      -1,
   } },

   /* ('ige', '#a', ('i2i32', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1476_0 -> 275 in the cache */
   /* search1476_1_0 -> 2060 in the cache */
   /* search1476_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 275, 4001 },
      -1,
   } },

   /* replace1476_0_0 -> 16 in the cache */
   /* replace1476_0_1 -> 4021 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4021 },
      -1,
   } },
   /* replace1476_1_0_0 -> 16 in the cache */
   /* replace1476_1_0_1 -> 4023 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4023 },
      -1,
   } },
   /* replace1476_1_1_0_0 -> 16 in the cache */
   /* replace1476_1_1_0 -> 3937 in the cache */
   /* replace1476_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3937, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4042, 4043 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4041, 4044 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i8', 'b')) */
   /* search1477_0_0 -> 6 in the cache */
   /* search1477_0 -> 3919 in the cache */
   /* search1477_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3919, 4046 },
      -1,
   } },

   /* replace1477_0 -> 6 in the cache */
   /* replace1477_1_0 -> 2060 in the cache */
   /* replace1477_1 -> 3946 in the cache */
   /* replace1477 -> 3947 in the cache */

   /* ('ine', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i8', 'b')) */
   /* search1478_0_0 -> 6 in the cache */
   /* search1478_0 -> 3919 in the cache */
   /* search1478_1_0 -> 2060 in the cache */
   /* search1478_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3919, 4046 },
      -1,
   } },

   /* replace1478_0 -> 6 in the cache */
   /* replace1478_1_0 -> 2060 in the cache */
   /* replace1478_1 -> 3946 in the cache */
   /* replace1478 -> 3949 in the cache */

   /* ('ige', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i8', 'b')) */
   /* search1479_0_0 -> 6 in the cache */
   /* search1479_0 -> 3919 in the cache */
   /* search1479_1_0 -> 2060 in the cache */
   /* search1479_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3919, 4046 },
      -1,
   } },

   /* replace1479_0 -> 6 in the cache */
   /* replace1479_1_0 -> 2060 in the cache */
   /* replace1479_1 -> 3946 in the cache */
   /* replace1479 -> 3951 in the cache */

   /* ('ilt', ('i2i64', 'a@8'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i8', 'b')) */
   /* search1480_0_0 -> 6 in the cache */
   /* search1480_0 -> 3919 in the cache */
   /* search1480_1_0 -> 2060 in the cache */
   /* search1480_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3919, 4046 },
      -1,
   } },

   /* replace1480_0 -> 6 in the cache */
   /* replace1480_1_0 -> 2060 in the cache */
   /* replace1480_1 -> 3946 in the cache */
   /* replace1480 -> 3953 in the cache */

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ige', ('i2i8', 'b'), 'a') */
   /* search1481_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* search1481_1_0 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 3956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4051, 4052 },
      -1,
   } },

   /* replace1481_0_0 -> 3954 in the cache */
   /* replace1481_0 -> 3959 in the cache */
   /* replace1481_1 -> 3956 in the cache */
   /* replace1481 -> 3960 in the cache */

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@8')) => ('ilt', ('i2i8', 'b'), 'a') */
   /* search1482_0_0 -> 3954 in the cache */
   /* search1482_0 -> 4051 in the cache */
   /* search1482_1_0 -> 3956 in the cache */
   /* search1482_1 -> 4052 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4051, 4052 },
      -1,
   } },

   /* replace1482_0_0 -> 3954 in the cache */
   /* replace1482_0 -> 3959 in the cache */
   /* replace1482_1 -> 3956 in the cache */
   /* replace1482 -> 3962 in the cache */

   /* ('ieq', ('i2i64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('i2i8', 'b')), ('ieq', ('i2i64', ('i2i8', 'b')), 'b')) */
   /* search1483_0_0 -> 6 in the cache */
   /* search1483_0 -> 3919 in the cache */
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3919, 4055 },
      -1,
   } },

   /* replace1483_0_0 -> 6 in the cache */
   /* replace1483_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4057 },
      -1,
   } },
   /* replace1483_1_0_0_0 -> 41 in the cache */
   /* replace1483_1_0_0 -> 4057 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4057 },
      -1,
   } },
   /* replace1483_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4059, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4058, 4060 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('i2i8', 'b')), ('ine', ('i2i64', ('i2i8', 'b')), 'b')) */
   /* search1484_0_0 -> 6 in the cache */
   /* search1484_0 -> 3919 in the cache */
   /* search1484_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3919, 4055 },
      -1,
   } },

   /* replace1484_0_0 -> 6 in the cache */
   /* replace1484_0_1_0 -> 41 in the cache */
   /* replace1484_0_1 -> 4057 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4057 },
      -1,
   } },
   /* replace1484_1_0_0_0 -> 41 in the cache */
   /* replace1484_1_0_0 -> 4057 in the cache */
   /* replace1484_1_0 -> 4059 in the cache */
   /* replace1484_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4059, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4063, 4064 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@8'), '#b') => ('iand', ('ilt', -128, 'b'), ('ior', ('ilt', 127, 'b'), ('ilt', 'a', ('i2i8', 'b')))) */
   /* search1485_0_0 -> 6 in the cache */
   /* search1485_0 -> 3919 in the cache */
   /* search1485_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3919, 4055 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x80 /* -128 */ },
   } },
   /* replace1485_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4067, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7f /* 127 */ },
   } },
   /* replace1485_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4069, 41 },
      -1,
   } },
   /* replace1485_1_1_0 -> 6 in the cache */
   /* replace1485_1_1_1_0 -> 41 in the cache */
   /* replace1485_1_1_1 -> 4057 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 6, 4057 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4070, 4071 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4068, 4072 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@8')) => ('iand', ('ilt', 'a', 127), ('ior', ('ilt', 'a', -128), ('ilt', ('i2i8', 'a'), 'b'))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1486_1_0 -> 2060 in the cache */
   /* search1486_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4074, 4046 },
      -1,
   } },

   /* replace1486_0_0 -> 21 in the cache */
   /* replace1486_0_1 -> 4069 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4069 },
      -1,
   } },
   /* replace1486_1_0_0 -> 21 in the cache */
   /* replace1486_1_0_1 -> 4067 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4067 },
      -1,
   } },
   /* replace1486_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1486_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4078, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4077, 4079 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4076, 4080 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@8'), '#b') => ('iand', ('ige', 127, 'b'), ('ior', ('ige', -128, 'b'), ('ige', 'a', ('i2i8', 'b')))) */
   /* search1487_0_0 -> 6 in the cache */
   /* search1487_0 -> 3919 in the cache */
   /* search1487_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3919, 4055 },
      -1,
   } },

   /* replace1487_0_0 -> 4069 in the cache */
   /* replace1487_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4069, 41 },
      -1,
   } },
   /* replace1487_1_0_0 -> 4067 in the cache */
   /* replace1487_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4067, 41 },
      -1,
   } },
   /* replace1487_1_1_0 -> 6 in the cache */
   /* replace1487_1_1_1_0 -> 41 in the cache */
   /* replace1487_1_1_1 -> 4057 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 6, 4057 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4084, 4085 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4083, 4086 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@8')) => ('iand', ('ige', 'a', -128), ('ior', ('ige', 'a', 127), ('ige', ('i2i8', 'a'), 'b'))) */
   /* search1488_0 -> 4074 in the cache */
   /* search1488_1_0 -> 2060 in the cache */
   /* search1488_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4074, 4046 },
      -1,
   } },

   /* replace1488_0_0 -> 21 in the cache */
   /* replace1488_0_1 -> 4067 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4067 },
      -1,
   } },
   /* replace1488_1_0_0 -> 21 in the cache */
   /* replace1488_1_0_1 -> 4069 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4069 },
      -1,
   } },
   /* replace1488_1_1_0_0 -> 21 in the cache */
   /* replace1488_1_1_0 -> 4078 in the cache */
   /* replace1488_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4078, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4090, 4091 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4089, 4092 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1489_0_0 -> 11 in the cache */
   /* search1489_0 -> 1747 in the cache */
   /* search1489_1_0 -> 2060 in the cache */
   /* search1489_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1747, 4001 },
      -1,
   } },

   /* replace1489_0 -> 11 in the cache */
   /* replace1489_1_0 -> 2060 in the cache */
   /* replace1489_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 3944 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1490_0_0 -> 11 in the cache */
   /* search1490_0 -> 1747 in the cache */
   /* search1490_1_0 -> 2060 in the cache */
   /* search1490_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1747, 4001 },
      -1,
   } },

   /* replace1490_0 -> 11 in the cache */
   /* replace1490_1_0 -> 2060 in the cache */
   /* replace1490_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 3944 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1491_0_0 -> 11 in the cache */
   /* search1491_0 -> 1747 in the cache */
   /* search1491_1_0 -> 2060 in the cache */
   /* search1491_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1747, 4001 },
      -1,
   } },

   /* replace1491_0 -> 11 in the cache */
   /* replace1491_1_0 -> 2060 in the cache */
   /* replace1491_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 3944 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@8')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1492_0_0 -> 11 in the cache */
   /* search1492_0 -> 1747 in the cache */
   /* search1492_1_0 -> 2060 in the cache */
   /* search1492_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1747, 4001 },
      -1,
   } },

   /* replace1492_0 -> 11 in the cache */
   /* replace1492_1_0 -> 2060 in the cache */
   /* replace1492_1 -> 3944 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 3944 },
      -1,
   } },

   /* ('ige', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1493_0_0 -> 3954 in the cache */
   /* search1493_0 -> 4006 in the cache */
   /* search1493_1_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4006, 4102 },
      -1,
   } },

   /* replace1493_0_0 -> 3954 in the cache */
   /* replace1493_0 -> 3955 in the cache */
   /* replace1493_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3955, 511 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'b@8'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1494_0_0 -> 3954 in the cache */
   /* search1494_0 -> 4006 in the cache */
   /* search1494_1_0 -> 511 in the cache */
   /* search1494_1 -> 4102 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4006, 4102 },
      -1,
   } },

   /* replace1494_0_0 -> 3954 in the cache */
   /* replace1494_0 -> 3955 in the cache */
   /* replace1494_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3955, 511 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1495_0_0 -> 11 in the cache */
   /* search1495_0 -> 1747 in the cache */
   /* search1495_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1747, 4107 },
      -1,
   } },

   /* replace1495_0 -> 11 in the cache */
   /* replace1495_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4109 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1496_0_0 -> 11 in the cache */
   /* search1496_0 -> 1747 in the cache */
   /* search1496_1_0 -> 247 in the cache */
   /* search1496_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1747, 4107 },
      -1,
   } },

   /* replace1496_0 -> 11 in the cache */
   /* replace1496_1_0 -> 247 in the cache */
   /* replace1496_1 -> 4109 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4109 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1497_0_0 -> 11 in the cache */
   /* search1497_0 -> 1747 in the cache */
   /* search1497_1_0 -> 247 in the cache */
   /* search1497_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1747, 4107 },
      -1,
   } },

   /* replace1497_0 -> 11 in the cache */
   /* replace1497_1_0 -> 247 in the cache */
   /* replace1497_1 -> 4109 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4109 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), ('i2i32', 'b@16')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1498_0_0 -> 11 in the cache */
   /* search1498_0 -> 1747 in the cache */
   /* search1498_1_0 -> 247 in the cache */
   /* search1498_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1747, 4107 },
      -1,
   } },

   /* replace1498_0 -> 11 in the cache */
   /* replace1498_1_0 -> 247 in the cache */
   /* replace1498_1 -> 4109 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4109 },
      -1,
   } },

   /* ('ige', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1499_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 500 },
      -1,
   } },
   /* search1499_1_0 -> 511 in the cache */
   /* search1499_1 -> 4102 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4117, 4102 },
      -1,
   } },

   /* replace1499_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 500 },
      -1,
   } },
   /* replace1499_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4119, 511 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'b@16'), ('i2i32', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1500_0_0 -> 500 in the cache */
   /* search1500_0 -> 4117 in the cache */
   /* search1500_1_0 -> 511 in the cache */
   /* search1500_1 -> 4102 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4117, 4102 },
      -1,
   } },

   /* replace1500_0_0 -> 500 in the cache */
   /* replace1500_0 -> 4119 in the cache */
   /* replace1500_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4119, 511 },
      -1,
   } },

   /* ('ieq', ('i2i32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i32', ('i2i16', 'b')), 'b')) */
   /* search1501_0_0 -> 11 in the cache */
   /* search1501_0 -> 1747 in the cache */
   /* search1501_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1747, 33 },
      -1,
   } },

   /* replace1501_0_0 -> 11 in the cache */
   /* replace1501_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4124 },
      -1,
   } },
   /* replace1501_1_0_0_0 -> 36 in the cache */
   /* replace1501_1_0_0 -> 4124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4124 },
      -1,
   } },
   /* replace1501_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4126, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4125, 4127 },
      -1,
   } },

   /* ('ine', ('i2i32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i32', ('i2i16', 'b')), 'b')) */
   /* search1502_0_0 -> 11 in the cache */
   /* search1502_0 -> 1747 in the cache */
   /* search1502_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1747, 33 },
      -1,
   } },

   /* replace1502_0_0 -> 11 in the cache */
   /* replace1502_0_1_0 -> 36 in the cache */
   /* replace1502_0_1 -> 4124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4124 },
      -1,
   } },
   /* replace1502_1_0_0_0 -> 36 in the cache */
   /* replace1502_1_0_0 -> 4124 in the cache */
   /* replace1502_1_0 -> 4126 in the cache */
   /* replace1502_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4126, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4130, 4131 },
      -1,
   } },

   /* ('ilt', ('i2i32', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b')))) */
   /* search1503_0_0 -> 11 in the cache */
   /* search1503_0 -> 1747 in the cache */
   /* search1503_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1747, 33 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x8000 /* -32768 */ },
   } },
   /* replace1503_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4134, 36 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1503_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4136, 36 },
      -1,
   } },
   /* replace1503_1_1_0 -> 11 in the cache */
   /* replace1503_1_1_1_0 -> 36 in the cache */
   /* replace1503_1_1_1 -> 4124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4137, 4138 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4135, 4139 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i32', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b'))) */
   /* search1504_0 -> 275 in the cache */
   /* search1504_1_0 -> 247 in the cache */
   /* search1504_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 275, 4107 },
      -1,
   } },

   /* replace1504_0_0 -> 16 in the cache */
   /* replace1504_0_1 -> 4136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4136 },
      -1,
   } },
   /* replace1504_1_0_0 -> 16 in the cache */
   /* replace1504_1_0_1 -> 4134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4134 },
      -1,
   } },
   /* replace1504_1_1_0_0 -> 16 in the cache */
   /* replace1504_1_1_0 -> 3939 in the cache */
   /* replace1504_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 3939, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4143, 4144 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4142, 4145 },
      -1,
   } },

   /* ('ige', ('i2i32', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b')))) */
   /* search1505_0_0 -> 11 in the cache */
   /* search1505_0 -> 1747 in the cache */
   /* search1505_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1747, 33 },
      -1,
   } },

   /* replace1505_0_0 -> 4136 in the cache */
   /* replace1505_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4136, 36 },
      -1,
   } },
   /* replace1505_1_0_0 -> 4134 in the cache */
   /* replace1505_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4134, 36 },
      -1,
   } },
   /* replace1505_1_1_0 -> 11 in the cache */
   /* replace1505_1_1_1_0 -> 36 in the cache */
   /* replace1505_1_1_1 -> 4124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4124 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4149, 4150 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4148, 4151 },
      -1,
   } },

   /* ('ige', '#a', ('i2i32', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b'))) */
   /* search1506_0 -> 275 in the cache */
   /* search1506_1_0 -> 247 in the cache */
   /* search1506_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 275, 4107 },
      -1,
   } },

   /* replace1506_0_0 -> 16 in the cache */
   /* replace1506_0_1 -> 4134 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4134 },
      -1,
   } },
   /* replace1506_1_0_0 -> 16 in the cache */
   /* replace1506_1_0_1 -> 4136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4136 },
      -1,
   } },
   /* replace1506_1_1_0_0 -> 16 in the cache */
   /* replace1506_1_1_0 -> 3939 in the cache */
   /* replace1506_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 3939, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4155, 4156 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4154, 4157 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1507_0_0 -> 11 in the cache */
   /* search1507_0 -> 1770 in the cache */
   /* search1507_1_0 -> 2060 in the cache */
   /* search1507_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1770, 4046 },
      -1,
   } },

   /* replace1507_0 -> 11 in the cache */
   /* replace1507_1_0 -> 2060 in the cache */
   /* replace1507_1 -> 3944 in the cache */
   /* replace1507 -> 4095 in the cache */

   /* ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1508_0_0 -> 11 in the cache */
   /* search1508_0 -> 1770 in the cache */
   /* search1508_1_0 -> 2060 in the cache */
   /* search1508_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1770, 4046 },
      -1,
   } },

   /* replace1508_0 -> 11 in the cache */
   /* replace1508_1_0 -> 2060 in the cache */
   /* replace1508_1 -> 3944 in the cache */
   /* replace1508 -> 4097 in the cache */

   /* ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1509_0_0 -> 11 in the cache */
   /* search1509_0 -> 1770 in the cache */
   /* search1509_1_0 -> 2060 in the cache */
   /* search1509_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1770, 4046 },
      -1,
   } },

   /* replace1509_0 -> 11 in the cache */
   /* replace1509_1_0 -> 2060 in the cache */
   /* replace1509_1 -> 3944 in the cache */
   /* replace1509 -> 4099 in the cache */

   /* ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1510_0_0 -> 11 in the cache */
   /* search1510_0 -> 1770 in the cache */
   /* search1510_1_0 -> 2060 in the cache */
   /* search1510_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1770, 4046 },
      -1,
   } },

   /* replace1510_0 -> 11 in the cache */
   /* replace1510_1_0 -> 2060 in the cache */
   /* replace1510_1 -> 3944 in the cache */
   /* replace1510 -> 4101 in the cache */

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1511_0_0 -> 3954 in the cache */
   /* search1511_0 -> 4051 in the cache */
   /* search1511_1_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4051, 4163 },
      -1,
   } },

   /* replace1511_0_0 -> 3954 in the cache */
   /* replace1511_0 -> 3955 in the cache */
   /* replace1511_1 -> 511 in the cache */
   /* replace1511 -> 4104 in the cache */

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1512_0_0 -> 3954 in the cache */
   /* search1512_0 -> 4051 in the cache */
   /* search1512_1_0 -> 511 in the cache */
   /* search1512_1 -> 4163 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4051, 4163 },
      -1,
   } },

   /* replace1512_0_0 -> 3954 in the cache */
   /* replace1512_0 -> 3955 in the cache */
   /* replace1512_1 -> 511 in the cache */
   /* replace1512 -> 4106 in the cache */

   /* ('ieq', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i16', 'b')) */
   /* search1513_0_0 -> 11 in the cache */
   /* search1513_0 -> 1770 in the cache */
   /* search1513_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1770, 4166 },
      -1,
   } },

   /* replace1513_0 -> 11 in the cache */
   /* replace1513_1_0 -> 247 in the cache */
   /* replace1513_1 -> 4109 in the cache */
   /* replace1513 -> 4110 in the cache */

   /* ('ine', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i16', 'b')) */
   /* search1514_0_0 -> 11 in the cache */
   /* search1514_0 -> 1770 in the cache */
   /* search1514_1_0 -> 247 in the cache */
   /* search1514_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1770, 4166 },
      -1,
   } },

   /* replace1514_0 -> 11 in the cache */
   /* replace1514_1_0 -> 247 in the cache */
   /* replace1514_1 -> 4109 in the cache */
   /* replace1514 -> 4112 in the cache */

   /* ('ige', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i16', 'b')) */
   /* search1515_0_0 -> 11 in the cache */
   /* search1515_0 -> 1770 in the cache */
   /* search1515_1_0 -> 247 in the cache */
   /* search1515_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1770, 4166 },
      -1,
   } },

   /* replace1515_0 -> 11 in the cache */
   /* replace1515_1_0 -> 247 in the cache */
   /* replace1515_1 -> 4109 in the cache */
   /* replace1515 -> 4114 in the cache */

   /* ('ilt', ('i2i64', 'a@16'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i16', 'b')) */
   /* search1516_0_0 -> 11 in the cache */
   /* search1516_0 -> 1770 in the cache */
   /* search1516_1_0 -> 247 in the cache */
   /* search1516_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1770, 4166 },
      -1,
   } },

   /* replace1516_0 -> 11 in the cache */
   /* replace1516_1_0 -> 247 in the cache */
   /* replace1516_1 -> 4109 in the cache */
   /* replace1516 -> 4116 in the cache */

   /* ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ige', ('i2i16', 'b'), 'a') */
   /* search1517_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 500 },
      -1,
   } },
   /* search1517_1_0 -> 511 in the cache */
   /* search1517_1 -> 4163 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4171, 4163 },
      -1,
   } },

   /* replace1517_0_0 -> 500 in the cache */
   /* replace1517_0 -> 4119 in the cache */
   /* replace1517_1 -> 511 in the cache */
   /* replace1517 -> 4120 in the cache */

   /* ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@16')) => ('ilt', ('i2i16', 'b'), 'a') */
   /* search1518_0_0 -> 500 in the cache */
   /* search1518_0 -> 4171 in the cache */
   /* search1518_1_0 -> 511 in the cache */
   /* search1518_1 -> 4163 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4171, 4163 },
      -1,
   } },

   /* replace1518_0_0 -> 500 in the cache */
   /* replace1518_0 -> 4119 in the cache */
   /* replace1518_1 -> 511 in the cache */
   /* replace1518 -> 4122 in the cache */

   /* ('ieq', ('i2i64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('i2i16', 'b')), ('ieq', ('i2i64', ('i2i16', 'b')), 'b')) */
   /* search1519_0_0 -> 11 in the cache */
   /* search1519_0 -> 1770 in the cache */
   /* search1519_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1770, 4055 },
      -1,
   } },

   /* replace1519_0_0 -> 11 in the cache */
   /* replace1519_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4175 },
      -1,
   } },
   /* replace1519_1_0_0_0 -> 41 in the cache */
   /* replace1519_1_0_0 -> 4175 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4175 },
      -1,
   } },
   /* replace1519_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4177, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4176, 4178 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('i2i16', 'b')), ('ine', ('i2i64', ('i2i16', 'b')), 'b')) */
   /* search1520_0_0 -> 11 in the cache */
   /* search1520_0 -> 1770 in the cache */
   /* search1520_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1770, 4055 },
      -1,
   } },

   /* replace1520_0_0 -> 11 in the cache */
   /* replace1520_0_1_0 -> 41 in the cache */
   /* replace1520_0_1 -> 4175 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4175 },
      -1,
   } },
   /* replace1520_1_0_0_0 -> 41 in the cache */
   /* replace1520_1_0_0 -> 4175 in the cache */
   /* replace1520_1_0 -> 4177 in the cache */
   /* replace1520_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4177, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4181, 4182 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@16'), '#b') => ('iand', ('ilt', -32768, 'b'), ('ior', ('ilt', 32767, 'b'), ('ilt', 'a', ('i2i16', 'b')))) */
   /* search1521_0_0 -> 11 in the cache */
   /* search1521_0 -> 1770 in the cache */
   /* search1521_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1770, 4055 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x8000 /* -32768 */ },
   } },
   /* replace1521_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4185, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fff /* 32767 */ },
   } },
   /* replace1521_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4187, 41 },
      -1,
   } },
   /* replace1521_1_1_0 -> 11 in the cache */
   /* replace1521_1_1_1_0 -> 41 in the cache */
   /* replace1521_1_1_1 -> 4175 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 11, 4175 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4188, 4189 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4186, 4190 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@16')) => ('iand', ('ilt', 'a', 32767), ('ior', ('ilt', 'a', -32768), ('ilt', ('i2i16', 'a'), 'b'))) */
   /* search1522_0 -> 4074 in the cache */
   /* search1522_1_0 -> 247 in the cache */
   /* search1522_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4074, 4166 },
      -1,
   } },

   /* replace1522_0_0 -> 21 in the cache */
   /* replace1522_0_1 -> 4187 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4187 },
      -1,
   } },
   /* replace1522_1_0_0 -> 21 in the cache */
   /* replace1522_1_0_1 -> 4185 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4185 },
      -1,
   } },
   /* replace1522_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1522_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4195, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4194, 4196 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4193, 4197 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@16'), '#b') => ('iand', ('ige', 32767, 'b'), ('ior', ('ige', -32768, 'b'), ('ige', 'a', ('i2i16', 'b')))) */
   /* search1523_0_0 -> 11 in the cache */
   /* search1523_0 -> 1770 in the cache */
   /* search1523_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1770, 4055 },
      -1,
   } },

   /* replace1523_0_0 -> 4187 in the cache */
   /* replace1523_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4187, 41 },
      -1,
   } },
   /* replace1523_1_0_0 -> 4185 in the cache */
   /* replace1523_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4185, 41 },
      -1,
   } },
   /* replace1523_1_1_0 -> 11 in the cache */
   /* replace1523_1_1_1_0 -> 41 in the cache */
   /* replace1523_1_1_1 -> 4175 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 11, 4175 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4201, 4202 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4200, 4203 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@16')) => ('iand', ('ige', 'a', -32768), ('ior', ('ige', 'a', 32767), ('ige', ('i2i16', 'a'), 'b'))) */
   /* search1524_0 -> 4074 in the cache */
   /* search1524_1_0 -> 247 in the cache */
   /* search1524_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4074, 4166 },
      -1,
   } },

   /* replace1524_0_0 -> 21 in the cache */
   /* replace1524_0_1 -> 4185 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4185 },
      -1,
   } },
   /* replace1524_1_0_0 -> 21 in the cache */
   /* replace1524_1_0_1 -> 4187 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4187 },
      -1,
   } },
   /* replace1524_1_1_0_0 -> 21 in the cache */
   /* replace1524_1_1_0 -> 4195 in the cache */
   /* replace1524_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4195, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4207, 4208 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4206, 4209 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1525_0_0 -> 16 in the cache */
   /* search1525_0 -> 1854 in the cache */
   /* search1525_1_0 -> 2060 in the cache */
   /* search1525_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1854, 4046 },
      -1,
   } },

   /* replace1525_0 -> 16 in the cache */
   /* replace1525_1_0 -> 2060 in the cache */
   /* replace1525_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4001 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1526_0_0 -> 16 in the cache */
   /* search1526_0 -> 1854 in the cache */
   /* search1526_1_0 -> 2060 in the cache */
   /* search1526_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1854, 4046 },
      -1,
   } },

   /* replace1526_0 -> 16 in the cache */
   /* replace1526_1_0 -> 2060 in the cache */
   /* replace1526_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4001 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1527_0_0 -> 16 in the cache */
   /* search1527_0 -> 1854 in the cache */
   /* search1527_1_0 -> 2060 in the cache */
   /* search1527_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1854, 4046 },
      -1,
   } },

   /* replace1527_0 -> 16 in the cache */
   /* replace1527_1_0 -> 2060 in the cache */
   /* replace1527_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4001 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@8')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1528_0_0 -> 16 in the cache */
   /* search1528_0 -> 1854 in the cache */
   /* search1528_1_0 -> 2060 in the cache */
   /* search1528_1 -> 4046 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1854, 4046 },
      -1,
   } },

   /* replace1528_0 -> 16 in the cache */
   /* replace1528_1_0 -> 2060 in the cache */
   /* replace1528_1 -> 4001 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4001 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1529_0_0 -> 3954 in the cache */
   /* search1529_0 -> 4051 in the cache */
   /* search1529_1_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4051, 4219 },
      -1,
   } },

   /* replace1529_0_0 -> 3954 in the cache */
   /* replace1529_0 -> 4006 in the cache */
   /* replace1529_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4006, 364 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@8'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1530_0_0 -> 3954 in the cache */
   /* search1530_0 -> 4051 in the cache */
   /* search1530_1_0 -> 364 in the cache */
   /* search1530_1 -> 4219 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4051, 4219 },
      -1,
   } },

   /* replace1530_0_0 -> 3954 in the cache */
   /* replace1530_0 -> 4006 in the cache */
   /* replace1530_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4006, 364 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1531_0_0 -> 16 in the cache */
   /* search1531_0 -> 1854 in the cache */
   /* search1531_1_0 -> 247 in the cache */
   /* search1531_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1854, 4166 },
      -1,
   } },

   /* replace1531_0 -> 16 in the cache */
   /* replace1531_1_0 -> 247 in the cache */
   /* replace1531_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4107 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1532_0_0 -> 16 in the cache */
   /* search1532_0 -> 1854 in the cache */
   /* search1532_1_0 -> 247 in the cache */
   /* search1532_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1854, 4166 },
      -1,
   } },

   /* replace1532_0 -> 16 in the cache */
   /* replace1532_1_0 -> 247 in the cache */
   /* replace1532_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4107 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1533_0_0 -> 16 in the cache */
   /* search1533_0 -> 1854 in the cache */
   /* search1533_1_0 -> 247 in the cache */
   /* search1533_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1854, 4166 },
      -1,
   } },

   /* replace1533_0 -> 16 in the cache */
   /* replace1533_1_0 -> 247 in the cache */
   /* replace1533_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4107 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@16')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1534_0_0 -> 16 in the cache */
   /* search1534_0 -> 1854 in the cache */
   /* search1534_1_0 -> 247 in the cache */
   /* search1534_1 -> 4166 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1854, 4166 },
      -1,
   } },

   /* replace1534_0 -> 16 in the cache */
   /* replace1534_1_0 -> 247 in the cache */
   /* replace1534_1 -> 4107 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4107 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1535_0_0 -> 500 in the cache */
   /* search1535_0 -> 4171 in the cache */
   /* search1535_1_0 -> 364 in the cache */
   /* search1535_1 -> 4219 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4171, 4219 },
      -1,
   } },

   /* replace1535_0_0 -> 500 in the cache */
   /* replace1535_0 -> 4117 in the cache */
   /* replace1535_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4117, 364 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@16'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1536_0_0 -> 500 in the cache */
   /* search1536_0 -> 4171 in the cache */
   /* search1536_1_0 -> 364 in the cache */
   /* search1536_1 -> 4219 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4171, 4219 },
      -1,
   } },

   /* replace1536_0_0 -> 500 in the cache */
   /* replace1536_0 -> 4117 in the cache */
   /* replace1536_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4117, 364 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ieq', 'a', ('i2i32', 'b')) */
   /* search1537_0_0 -> 16 in the cache */
   /* search1537_0 -> 1854 in the cache */
   /* search1537_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1854, 4236 },
      -1,
   } },

   /* replace1537_0 -> 16 in the cache */
   /* replace1537_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4238 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ine', 'a', ('i2i32', 'b')) */
   /* search1538_0_0 -> 16 in the cache */
   /* search1538_0 -> 1854 in the cache */
   /* search1538_1_0 -> 36 in the cache */
   /* search1538_1 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1854, 4236 },
      -1,
   } },

   /* replace1538_0 -> 16 in the cache */
   /* replace1538_1_0 -> 36 in the cache */
   /* replace1538_1 -> 4238 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4238 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ige', 'a', ('i2i32', 'b')) */
   /* search1539_0_0 -> 16 in the cache */
   /* search1539_0 -> 1854 in the cache */
   /* search1539_1_0 -> 36 in the cache */
   /* search1539_1 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1854, 4236 },
      -1,
   } },

   /* replace1539_0 -> 16 in the cache */
   /* replace1539_1_0 -> 36 in the cache */
   /* replace1539_1 -> 4238 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4238 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), ('i2i64', 'b@32')) => ('ilt', 'a', ('i2i32', 'b')) */
   /* search1540_0_0 -> 16 in the cache */
   /* search1540_0 -> 1854 in the cache */
   /* search1540_1_0 -> 36 in the cache */
   /* search1540_1 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1854, 4236 },
      -1,
   } },

   /* replace1540_0 -> 16 in the cache */
   /* replace1540_1_0 -> 36 in the cache */
   /* replace1540_1 -> 4238 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4238 },
      -1,
   } },

   /* ('ige', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ige', ('i2i32', 'b'), 'a') */
   /* search1541_0_0 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 362 },
      -1,
   } },
   /* search1541_1_0 -> 364 in the cache */
   /* search1541_1 -> 4219 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4246, 4219 },
      -1,
   } },

   /* replace1541_0_0 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 362 },
      -1,
   } },
   /* replace1541_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4248, 364 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'b@32'), ('i2i64', 'a@32')) => ('ilt', ('i2i32', 'b'), 'a') */
   /* search1542_0_0 -> 362 in the cache */
   /* search1542_0 -> 4246 in the cache */
   /* search1542_1_0 -> 364 in the cache */
   /* search1542_1 -> 4219 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4246, 4219 },
      -1,
   } },

   /* replace1542_0_0 -> 362 in the cache */
   /* replace1542_0 -> 4248 in the cache */
   /* replace1542_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4248, 364 },
      -1,
   } },

   /* ('ieq', ('i2i64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('i2i32', 'b')), ('ieq', ('i2i64', ('i2i32', 'b')), 'b')) */
   /* search1543_0_0 -> 16 in the cache */
   /* search1543_0 -> 1854 in the cache */
   /* search1543_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1854, 4055 },
      -1,
   } },

   /* replace1543_0_0 -> 16 in the cache */
   /* replace1543_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 4253 },
      -1,
   } },
   /* replace1543_1_0_0_0 -> 41 in the cache */
   /* replace1543_1_0_0 -> 4253 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_i2i64,
      -1, 0,
      { 4253 },
      -1,
   } },
   /* replace1543_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4255, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4254, 4256 },
      -1,
   } },

   /* ('ine', ('i2i64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('i2i32', 'b')), ('ine', ('i2i64', ('i2i32', 'b')), 'b')) */
   /* search1544_0_0 -> 16 in the cache */
   /* search1544_0 -> 1854 in the cache */
   /* search1544_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1854, 4055 },
      -1,
   } },

   /* replace1544_0_0 -> 16 in the cache */
   /* replace1544_0_1_0 -> 41 in the cache */
   /* replace1544_0_1 -> 4253 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 4253 },
      -1,
   } },
   /* replace1544_1_0_0_0 -> 41 in the cache */
   /* replace1544_1_0_0 -> 4253 in the cache */
   /* replace1544_1_0 -> 4255 in the cache */
   /* replace1544_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4255, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4259, 4260 },
      -1,
   } },

   /* ('ilt', ('i2i64', 'a@32'), '#b') => ('iand', ('ilt', -2147483648, 'b'), ('ior', ('ilt', 2147483647, 'b'), ('ilt', 'a', ('i2i32', 'b')))) */
   /* search1545_0_0 -> 16 in the cache */
   /* search1545_0 -> 1854 in the cache */
   /* search1545_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 1854, 4055 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { -0x80000000 /* -2147483648 */ },
   } },
   /* replace1545_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4263, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x7fffffff /* 2147483647 */ },
   } },
   /* replace1545_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4265, 41 },
      -1,
   } },
   /* replace1545_1_1_0 -> 16 in the cache */
   /* replace1545_1_1_1_0 -> 41 in the cache */
   /* replace1545_1_1_1 -> 4253 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 16, 4253 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4266, 4267 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4264, 4268 },
      -1,
   } },

   /* ('ilt', '#a', ('i2i64', 'b@32')) => ('iand', ('ilt', 'a', 2147483647), ('ior', ('ilt', 'a', -2147483648), ('ilt', ('i2i32', 'a'), 'b'))) */
   /* search1546_0 -> 4074 in the cache */
   /* search1546_1_0 -> 36 in the cache */
   /* search1546_1 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4074, 4236 },
      -1,
   } },

   /* replace1546_0_0 -> 21 in the cache */
   /* replace1546_0_1 -> 4265 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4265 },
      -1,
   } },
   /* replace1546_1_0_0 -> 21 in the cache */
   /* replace1546_1_0_1 -> 4263 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 21, 4263 },
      -1,
   } },
   /* replace1546_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1546_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 4273, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4272, 4274 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4271, 4275 },
      -1,
   } },

   /* ('ige', ('i2i64', 'a@32'), '#b') => ('iand', ('ige', 2147483647, 'b'), ('ior', ('ige', -2147483648, 'b'), ('ige', 'a', ('i2i32', 'b')))) */
   /* search1547_0_0 -> 16 in the cache */
   /* search1547_0 -> 1854 in the cache */
   /* search1547_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 1854, 4055 },
      -1,
   } },

   /* replace1547_0_0 -> 4265 in the cache */
   /* replace1547_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4265, 41 },
      -1,
   } },
   /* replace1547_1_0_0 -> 4263 in the cache */
   /* replace1547_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4263, 41 },
      -1,
   } },
   /* replace1547_1_1_0 -> 16 in the cache */
   /* replace1547_1_1_1_0 -> 41 in the cache */
   /* replace1547_1_1_1 -> 4253 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 16, 4253 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4279, 4280 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4278, 4281 },
      -1,
   } },

   /* ('ige', '#a', ('i2i64', 'b@32')) => ('iand', ('ige', 'a', -2147483648), ('ior', ('ige', 'a', 2147483647), ('ige', ('i2i32', 'a'), 'b'))) */
   /* search1548_0 -> 4074 in the cache */
   /* search1548_1_0 -> 36 in the cache */
   /* search1548_1 -> 4236 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4074, 4236 },
      -1,
   } },

   /* replace1548_0_0 -> 21 in the cache */
   /* replace1548_0_1 -> 4263 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4263 },
      -1,
   } },
   /* replace1548_1_0_0 -> 21 in the cache */
   /* replace1548_1_0_1 -> 4265 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 21, 4265 },
      -1,
   } },
   /* replace1548_1_1_0_0 -> 21 in the cache */
   /* replace1548_1_1_0 -> 4273 in the cache */
   /* replace1548_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 4273, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4285, 4286 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4284, 4287 },
      -1,
   } },

   /* ('ieq', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1549_0_0 -> 6 in the cache */
   /* search1549_0 -> 3884 in the cache */
   /* search1549_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3884, 4289 },
      -1,
   } },

   /* replace1549_0 -> 6 in the cache */
   /* replace1549_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 6, 4291 },
      -1,
   } },

   /* ('ine', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1550_0_0 -> 6 in the cache */
   /* search1550_0 -> 3884 in the cache */
   /* search1550_1_0 -> 2060 in the cache */
   /* search1550_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3884, 4289 },
      -1,
   } },

   /* replace1550_0 -> 6 in the cache */
   /* replace1550_1_0 -> 2060 in the cache */
   /* replace1550_1 -> 4291 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 6, 4291 },
      -1,
   } },

   /* ('uge', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1551_0_0 -> 6 in the cache */
   /* search1551_0 -> 3884 in the cache */
   /* search1551_1_0 -> 2060 in the cache */
   /* search1551_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3884, 4289 },
      -1,
   } },

   /* replace1551_0 -> 6 in the cache */
   /* replace1551_1_0 -> 2060 in the cache */
   /* replace1551_1 -> 4291 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 4291 },
      -1,
   } },

   /* ('ult', ('u2u16', 'a@8'), ('u2u16', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1552_0_0 -> 6 in the cache */
   /* search1552_0 -> 3884 in the cache */
   /* search1552_1_0 -> 2060 in the cache */
   /* search1552_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3884, 4289 },
      -1,
   } },

   /* replace1552_0 -> 6 in the cache */
   /* replace1552_1_0 -> 2060 in the cache */
   /* replace1552_1 -> 4291 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 4291 },
      -1,
   } },

   /* ('uge', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1553_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* search1553_1_0 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4299, 4300 },
      -1,
   } },

   /* replace1553_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* replace1553_1 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4302, 3956 },
      -1,
   } },

   /* ('ult', ('u2u16', 'b@8'), ('u2u16', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1554_0_0 -> 3954 in the cache */
   /* search1554_0 -> 4299 in the cache */
   /* search1554_1_0 -> 3956 in the cache */
   /* search1554_1 -> 4300 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4299, 4300 },
      -1,
   } },

   /* replace1554_0_0 -> 3954 in the cache */
   /* replace1554_0 -> 4302 in the cache */
   /* replace1554_1 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4302, 3956 },
      -1,
   } },

   /* ('ieq', ('u2u16', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u16', ('u2u8', 'b')), 'b')) */
   /* search1555_0_0 -> 6 in the cache */
   /* search1555_0 -> 3884 in the cache */
   /* search1555_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3884, 1959 },
      -1,
   } },

   /* replace1555_0_0 -> 6 in the cache */
   /* replace1555_0_1_0 -> 247 in the cache */
   /* replace1555_0_1 -> 1967 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 1967 },
      -1,
   } },
   /* replace1555_1_0_0_0 -> 247 in the cache */
   /* replace1555_1_0_0 -> 1967 in the cache */
   /* replace1555_1_0 -> 1968 in the cache */
   /* replace1555_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 1968, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4307, 4308 },
      -1,
   } },

   /* ('ine', ('u2u16', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u16', ('u2u8', 'b')), 'b')) */
   /* search1556_0_0 -> 6 in the cache */
   /* search1556_0 -> 3884 in the cache */
   /* search1556_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3884, 1959 },
      -1,
   } },

   /* replace1556_0_0 -> 6 in the cache */
   /* replace1556_0_1_0 -> 247 in the cache */
   /* replace1556_0_1 -> 1967 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 1967 },
      -1,
   } },
   /* replace1556_1_0_0_0 -> 247 in the cache */
   /* replace1556_1_0_0 -> 1967 in the cache */
   /* replace1556_1_0 -> 1968 in the cache */
   /* replace1556_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 1968, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4311, 4312 },
      -1,
   } },

   /* ('ult', ('u2u16', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1557_0_0 -> 6 in the cache */
   /* search1557_0 -> 3884 in the cache */
   /* search1557_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3884, 1959 },
      -1,
   } },

   /* replace1557_0_0 -> 818 in the cache */
   /* replace1557_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 818, 247 },
      -1,
   } },
   /* replace1557_1_0_0 -> 1961 in the cache */
   /* replace1557_1_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1961, 247 },
      -1,
   } },
   /* replace1557_1_1_0 -> 6 in the cache */
   /* replace1557_1_1_1_0 -> 247 in the cache */
   /* replace1557_1_1_1 -> 1967 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 1967 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4316, 4317 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4315, 4318 },
      -1,
   } },

   /* ('ult', '#a', ('u2u16', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1558_0 -> 3981 in the cache */
   /* search1558_1_0 -> 2060 in the cache */
   /* search1558_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3981, 4289 },
      -1,
   } },

   /* replace1558_0_0 -> 11 in the cache */
   /* replace1558_0_1 -> 1961 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 1961 },
      -1,
   } },
   /* replace1558_1_0_0 -> 11 in the cache */
   /* replace1558_1_0_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 818 },
      -1,
   } },
   /* replace1558_1_1_0_0 -> 11 in the cache */
   /* replace1558_1_1_0 -> 1957 in the cache */
   /* replace1558_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1957, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4322, 4323 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4321, 4324 },
      -1,
   } },

   /* ('uge', ('u2u16', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1559_0_0 -> 6 in the cache */
   /* search1559_0 -> 3884 in the cache */
   /* search1559_1 -> 1959 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3884, 1959 },
      -1,
   } },

   /* replace1559_0_0 -> 1961 in the cache */
   /* replace1559_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1961, 247 },
      -1,
   } },
   /* replace1559_1_0_0 -> 818 in the cache */
   /* replace1559_1_0_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 818, 247 },
      -1,
   } },
   /* replace1559_1_1_0 -> 6 in the cache */
   /* replace1559_1_1_1_0 -> 247 in the cache */
   /* replace1559_1_1_1 -> 1967 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 1967 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4328, 4329 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4327, 4330 },
      -1,
   } },

   /* ('uge', '#a', ('u2u16', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1560_0 -> 3981 in the cache */
   /* search1560_1_0 -> 2060 in the cache */
   /* search1560_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3981, 4289 },
      -1,
   } },

   /* replace1560_0_0 -> 11 in the cache */
   /* replace1560_0_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 818 },
      -1,
   } },
   /* replace1560_1_0_0 -> 11 in the cache */
   /* replace1560_1_0_1 -> 1961 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 1961 },
      -1,
   } },
   /* replace1560_1_1_0_0 -> 11 in the cache */
   /* replace1560_1_1_0 -> 1957 in the cache */
   /* replace1560_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1957, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4334, 4335 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4333, 4336 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1561_0_0 -> 6 in the cache */
   /* search1561_0 -> 3009 in the cache */
   /* search1561_1_0 -> 2060 in the cache */
   /* search1561_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3009, 3011 },
      -1,
   } },

   /* replace1561_0 -> 6 in the cache */
   /* replace1561_1_0 -> 2060 in the cache */
   /* replace1561_1 -> 4291 in the cache */
   /* replace1561 -> 4292 in the cache */

   /* ('ine', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1562_0_0 -> 6 in the cache */
   /* search1562_0 -> 3009 in the cache */
   /* search1562_1_0 -> 2060 in the cache */
   /* search1562_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3009, 3011 },
      -1,
   } },

   /* replace1562_0 -> 6 in the cache */
   /* replace1562_1_0 -> 2060 in the cache */
   /* replace1562_1 -> 4291 in the cache */
   /* replace1562 -> 4294 in the cache */

   /* ('uge', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1563_0_0 -> 6 in the cache */
   /* search1563_0 -> 3009 in the cache */
   /* search1563_1_0 -> 2060 in the cache */
   /* search1563_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3009, 3011 },
      -1,
   } },

   /* replace1563_0 -> 6 in the cache */
   /* replace1563_1_0 -> 2060 in the cache */
   /* replace1563_1 -> 4291 in the cache */
   /* replace1563 -> 4296 in the cache */

   /* ('ult', ('u2u32', 'a@8'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1564_0_0 -> 6 in the cache */
   /* search1564_0 -> 3009 in the cache */
   /* search1564_1_0 -> 2060 in the cache */
   /* search1564_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3009, 3011 },
      -1,
   } },

   /* replace1564_0 -> 6 in the cache */
   /* replace1564_1_0 -> 2060 in the cache */
   /* replace1564_1 -> 4291 in the cache */
   /* replace1564 -> 4298 in the cache */

   /* ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1565_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* search1565_1_0 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4342, 4343 },
      -1,
   } },

   /* replace1565_0_0 -> 3954 in the cache */
   /* replace1565_0 -> 4302 in the cache */
   /* replace1565_1 -> 3956 in the cache */
   /* replace1565 -> 4303 in the cache */

   /* ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1566_0_0 -> 3954 in the cache */
   /* search1566_0 -> 4342 in the cache */
   /* search1566_1_0 -> 3956 in the cache */
   /* search1566_1 -> 4343 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4342, 4343 },
      -1,
   } },

   /* replace1566_0_0 -> 3954 in the cache */
   /* replace1566_0 -> 4302 in the cache */
   /* replace1566_1 -> 3956 in the cache */
   /* replace1566 -> 4305 in the cache */

   /* ('ieq', ('u2u32', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u32', ('u2u8', 'b')), 'b')) */
   /* search1567_0_0 -> 6 in the cache */
   /* search1567_0 -> 3009 in the cache */
   /* search1567_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3009, 33 },
      -1,
   } },

   /* replace1567_0_0 -> 6 in the cache */
   /* replace1567_0_1_0 -> 36 in the cache */
   /* replace1567_0_1 -> 1975 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 1975 },
      -1,
   } },
   /* replace1567_1_0_0_0 -> 36 in the cache */
   /* replace1567_1_0_0 -> 1975 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1975 },
      -1,
   } },
   /* replace1567_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4348, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4347, 4349 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u32', ('u2u8', 'b')), 'b')) */
   /* search1568_0_0 -> 6 in the cache */
   /* search1568_0 -> 3009 in the cache */
   /* search1568_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3009, 33 },
      -1,
   } },

   /* replace1568_0_0 -> 6 in the cache */
   /* replace1568_0_1_0 -> 36 in the cache */
   /* replace1568_0_1 -> 1975 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 1975 },
      -1,
   } },
   /* replace1568_1_0_0_0 -> 36 in the cache */
   /* replace1568_1_0_0 -> 1975 in the cache */
   /* replace1568_1_0 -> 4348 in the cache */
   /* replace1568_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4348, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4352, 4353 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1569_0_0 -> 6 in the cache */
   /* search1569_0 -> 3009 in the cache */
   /* search1569_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3009, 33 },
      -1,
   } },

   /* replace1569_0_0 -> 124 in the cache */
   /* replace1569_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 124, 36 },
      -1,
   } },
   /* replace1569_1_0_0 -> 972 in the cache */
   /* replace1569_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 972, 36 },
      -1,
   } },
   /* replace1569_1_1_0 -> 6 in the cache */
   /* replace1569_1_1_1_0 -> 36 in the cache */
   /* replace1569_1_1_1 -> 1975 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 1975 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4357, 4358 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4356, 4359 },
      -1,
   } },

   /* ('ult', '#a', ('u2u32', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1570_0 -> 275 in the cache */
   /* search1570_1_0 -> 2060 in the cache */
   /* search1570_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 275, 3011 },
      -1,
   } },

   /* replace1570_0_0 -> 16 in the cache */
   /* replace1570_0_1 -> 972 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 972 },
      -1,
   } },
   /* replace1570_1_0_0 -> 16 in the cache */
   /* replace1570_1_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* replace1570_1_1_0_0 -> 16 in the cache */
   /* replace1570_1_1_0 -> 1973 in the cache */
   /* replace1570_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1973, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4363, 4364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4362, 4365 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1571_0_0 -> 6 in the cache */
   /* search1571_0 -> 3009 in the cache */
   /* search1571_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3009, 33 },
      -1,
   } },

   /* replace1571_0_0 -> 972 in the cache */
   /* replace1571_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 972, 36 },
      -1,
   } },
   /* replace1571_1_0_0 -> 124 in the cache */
   /* replace1571_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 124, 36 },
      -1,
   } },
   /* replace1571_1_1_0 -> 6 in the cache */
   /* replace1571_1_1_1_0 -> 36 in the cache */
   /* replace1571_1_1_1 -> 1975 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 1975 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4369, 4370 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4368, 4371 },
      -1,
   } },

   /* ('uge', '#a', ('u2u32', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1572_0 -> 275 in the cache */
   /* search1572_1_0 -> 2060 in the cache */
   /* search1572_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 275, 3011 },
      -1,
   } },

   /* replace1572_0_0 -> 16 in the cache */
   /* replace1572_0_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 124 },
      -1,
   } },
   /* replace1572_1_0_0 -> 16 in the cache */
   /* replace1572_1_0_1 -> 972 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 972 },
      -1,
   } },
   /* replace1572_1_1_0_0 -> 16 in the cache */
   /* replace1572_1_1_0 -> 1973 in the cache */
   /* replace1572_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1973, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4375, 4376 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4374, 4377 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u8', 'b')) */
   /* search1573_0_0 -> 6 in the cache */
   /* search1573_0 -> 3921 in the cache */
   /* search1573_1_0 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3921, 4379 },
      -1,
   } },

   /* replace1573_0 -> 6 in the cache */
   /* replace1573_1_0 -> 2060 in the cache */
   /* replace1573_1 -> 4291 in the cache */
   /* replace1573 -> 4292 in the cache */

   /* ('ine', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u8', 'b')) */
   /* search1574_0_0 -> 6 in the cache */
   /* search1574_0 -> 3921 in the cache */
   /* search1574_1_0 -> 2060 in the cache */
   /* search1574_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3921, 4379 },
      -1,
   } },

   /* replace1574_0 -> 6 in the cache */
   /* replace1574_1_0 -> 2060 in the cache */
   /* replace1574_1 -> 4291 in the cache */
   /* replace1574 -> 4294 in the cache */

   /* ('uge', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u8', 'b')) */
   /* search1575_0_0 -> 6 in the cache */
   /* search1575_0 -> 3921 in the cache */
   /* search1575_1_0 -> 2060 in the cache */
   /* search1575_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3921, 4379 },
      -1,
   } },

   /* replace1575_0 -> 6 in the cache */
   /* replace1575_1_0 -> 2060 in the cache */
   /* replace1575_1 -> 4291 in the cache */
   /* replace1575 -> 4296 in the cache */

   /* ('ult', ('u2u64', 'a@8'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u8', 'b')) */
   /* search1576_0_0 -> 6 in the cache */
   /* search1576_0 -> 3921 in the cache */
   /* search1576_1_0 -> 2060 in the cache */
   /* search1576_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3921, 4379 },
      -1,
   } },

   /* replace1576_0 -> 6 in the cache */
   /* replace1576_1_0 -> 2060 in the cache */
   /* replace1576_1 -> 4291 in the cache */
   /* replace1576 -> 4298 in the cache */

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('uge', ('u2u8', 'b'), 'a') */
   /* search1577_0_0 -> 3954 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3954 },
      -1,
   } },
   /* search1577_1_0 -> 3956 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 3956 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4384, 4385 },
      -1,
   } },

   /* replace1577_0_0 -> 3954 in the cache */
   /* replace1577_0 -> 4302 in the cache */
   /* replace1577_1 -> 3956 in the cache */
   /* replace1577 -> 4303 in the cache */

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@8')) => ('ult', ('u2u8', 'b'), 'a') */
   /* search1578_0_0 -> 3954 in the cache */
   /* search1578_0 -> 4384 in the cache */
   /* search1578_1_0 -> 3956 in the cache */
   /* search1578_1 -> 4385 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4384, 4385 },
      -1,
   } },

   /* replace1578_0_0 -> 3954 in the cache */
   /* replace1578_0 -> 4302 in the cache */
   /* replace1578_1 -> 3956 in the cache */
   /* replace1578 -> 4305 in the cache */

   /* ('ieq', ('u2u64', 'a@8'), '#b') => ('iand', ('ieq', 'a', ('u2u8', 'b')), ('ieq', ('u2u64', ('u2u8', 'b')), 'b')) */
   /* search1579_0_0 -> 6 in the cache */
   /* search1579_0 -> 3921 in the cache */
   /* search1579_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 3921, 4055 },
      -1,
   } },

   /* replace1579_0_0 -> 6 in the cache */
   /* replace1579_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 6, 4389 },
      -1,
   } },
   /* replace1579_1_0_0_0 -> 41 in the cache */
   /* replace1579_1_0_0 -> 4389 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4389 },
      -1,
   } },
   /* replace1579_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4391, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4390, 4392 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@8'), '#b') => ('ior', ('ine', 'a', ('u2u8', 'b')), ('ine', ('u2u64', ('u2u8', 'b')), 'b')) */
   /* search1580_0_0 -> 6 in the cache */
   /* search1580_0 -> 3921 in the cache */
   /* search1580_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 3921, 4055 },
      -1,
   } },

   /* replace1580_0_0 -> 6 in the cache */
   /* replace1580_0_1_0 -> 41 in the cache */
   /* replace1580_0_1 -> 4389 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 6, 4389 },
      -1,
   } },
   /* replace1580_1_0_0_0 -> 41 in the cache */
   /* replace1580_1_0_0 -> 4389 in the cache */
   /* replace1580_1_0 -> 4391 in the cache */
   /* replace1580_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4391, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4395, 4396 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@8'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 255, 'b'), ('ult', 'a', ('u2u8', 'b')))) */
   /* search1581_0_0 -> 6 in the cache */
   /* search1581_0 -> 3921 in the cache */
   /* search1581_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3921, 4055 },
      -1,
   } },

   /* replace1581_0_0 -> 859 in the cache */
   /* replace1581_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 859, 41 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xff /* 255 */ },
   } },
   /* replace1581_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4400, 41 },
      -1,
   } },
   /* replace1581_1_1_0 -> 6 in the cache */
   /* replace1581_1_1_1_0 -> 41 in the cache */
   /* replace1581_1_1_1 -> 4389 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 6, 4389 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4401, 4402 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4399, 4403 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@8')) => ('iand', ('ult', 'a', 255), ('ior', ('ult', 'a', 0), ('ult', ('u2u8', 'a'), 'b'))) */
   /* search1582_0 -> 4074 in the cache */
   /* search1582_1_0 -> 2060 in the cache */
   /* search1582_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4074, 4379 },
      -1,
   } },

   /* replace1582_0_0 -> 21 in the cache */
   /* replace1582_0_1 -> 4400 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 4400 },
      -1,
   } },
   /* replace1582_1_0_0 -> 21 in the cache */
   /* replace1582_1_0_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 859 },
      -1,
   } },
   /* replace1582_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1582_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4408, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4407, 4409 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4406, 4410 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@8'), '#b') => ('iand', ('uge', 255, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u8', 'b')))) */
   /* search1583_0_0 -> 6 in the cache */
   /* search1583_0 -> 3921 in the cache */
   /* search1583_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3921, 4055 },
      -1,
   } },

   /* replace1583_0_0 -> 4400 in the cache */
   /* replace1583_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4400, 41 },
      -1,
   } },
   /* replace1583_1_0_0 -> 859 in the cache */
   /* replace1583_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 859, 41 },
      -1,
   } },
   /* replace1583_1_1_0 -> 6 in the cache */
   /* replace1583_1_1_1_0 -> 41 in the cache */
   /* replace1583_1_1_1 -> 4389 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 6, 4389 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4414, 4415 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4413, 4416 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@8')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 255), ('uge', ('u2u8', 'a'), 'b'))) */
   /* search1584_0 -> 4074 in the cache */
   /* search1584_1_0 -> 2060 in the cache */
   /* search1584_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4074, 4379 },
      -1,
   } },

   /* replace1584_0_0 -> 21 in the cache */
   /* replace1584_0_1 -> 859 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 859 },
      -1,
   } },
   /* replace1584_1_0_0 -> 21 in the cache */
   /* replace1584_1_0_1 -> 4400 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 4400 },
      -1,
   } },
   /* replace1584_1_1_0_0 -> 21 in the cache */
   /* replace1584_1_1_0 -> 4408 in the cache */
   /* replace1584_1_1_1 -> 2060 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4408, 2060 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4420, 4421 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4419, 4422 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1585_0_0 -> 11 in the cache */
   /* search1585_0 -> 1749 in the cache */
   /* search1585_1_0 -> 2060 in the cache */
   /* search1585_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1749, 3011 },
      -1,
   } },

   /* replace1585_0 -> 11 in the cache */
   /* replace1585_1_0 -> 2060 in the cache */
   /* replace1585_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4289 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1586_0_0 -> 11 in the cache */
   /* search1586_0 -> 1749 in the cache */
   /* search1586_1_0 -> 2060 in the cache */
   /* search1586_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1749, 3011 },
      -1,
   } },

   /* replace1586_0 -> 11 in the cache */
   /* replace1586_1_0 -> 2060 in the cache */
   /* replace1586_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4289 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1587_0_0 -> 11 in the cache */
   /* search1587_0 -> 1749 in the cache */
   /* search1587_1_0 -> 2060 in the cache */
   /* search1587_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1749, 3011 },
      -1,
   } },

   /* replace1587_0 -> 11 in the cache */
   /* replace1587_1_0 -> 2060 in the cache */
   /* replace1587_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4289 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@8')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1588_0_0 -> 11 in the cache */
   /* search1588_0 -> 1749 in the cache */
   /* search1588_1_0 -> 2060 in the cache */
   /* search1588_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1749, 3011 },
      -1,
   } },

   /* replace1588_0 -> 11 in the cache */
   /* replace1588_1_0 -> 2060 in the cache */
   /* replace1588_1 -> 4289 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4289 },
      -1,
   } },

   /* ('uge', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1589_0_0 -> 3954 in the cache */
   /* search1589_0 -> 4342 in the cache */
   /* search1589_1_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4342, 4432 },
      -1,
   } },

   /* replace1589_0_0 -> 3954 in the cache */
   /* replace1589_0 -> 4299 in the cache */
   /* replace1589_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4299, 511 },
      -1,
   } },

   /* ('ult', ('u2u32', 'b@8'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1590_0_0 -> 3954 in the cache */
   /* search1590_0 -> 4342 in the cache */
   /* search1590_1_0 -> 511 in the cache */
   /* search1590_1 -> 4432 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4342, 4432 },
      -1,
   } },

   /* replace1590_0_0 -> 3954 in the cache */
   /* replace1590_0 -> 4299 in the cache */
   /* replace1590_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4299, 511 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1591_0_0 -> 11 in the cache */
   /* search1591_0 -> 1749 in the cache */
   /* search1591_1_0 -> 247 in the cache */
   /* search1591_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1749, 3031 },
      -1,
   } },

   /* replace1591_0 -> 11 in the cache */
   /* replace1591_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 11, 4438 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1592_0_0 -> 11 in the cache */
   /* search1592_0 -> 1749 in the cache */
   /* search1592_1_0 -> 247 in the cache */
   /* search1592_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1749, 3031 },
      -1,
   } },

   /* replace1592_0 -> 11 in the cache */
   /* replace1592_1_0 -> 247 in the cache */
   /* replace1592_1 -> 4438 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 11, 4438 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1593_0_0 -> 11 in the cache */
   /* search1593_0 -> 1749 in the cache */
   /* search1593_1_0 -> 247 in the cache */
   /* search1593_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1749, 3031 },
      -1,
   } },

   /* replace1593_0 -> 11 in the cache */
   /* replace1593_1_0 -> 247 in the cache */
   /* replace1593_1 -> 4438 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4438 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), ('u2u32', 'b@16')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1594_0_0 -> 11 in the cache */
   /* search1594_0 -> 1749 in the cache */
   /* search1594_1_0 -> 247 in the cache */
   /* search1594_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1749, 3031 },
      -1,
   } },

   /* replace1594_0 -> 11 in the cache */
   /* replace1594_1_0 -> 247 in the cache */
   /* replace1594_1 -> 4438 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4438 },
      -1,
   } },

   /* ('uge', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1595_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 500 },
      -1,
   } },
   /* search1595_1_0 -> 511 in the cache */
   /* search1595_1 -> 4432 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4446, 4432 },
      -1,
   } },

   /* replace1595_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 500 },
      -1,
   } },
   /* replace1595_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4448, 511 },
      -1,
   } },

   /* ('ult', ('u2u32', 'b@16'), ('u2u32', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1596_0_0 -> 500 in the cache */
   /* search1596_0 -> 4446 in the cache */
   /* search1596_1_0 -> 511 in the cache */
   /* search1596_1 -> 4432 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4446, 4432 },
      -1,
   } },

   /* replace1596_0_0 -> 500 in the cache */
   /* replace1596_0 -> 4448 in the cache */
   /* replace1596_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4448, 511 },
      -1,
   } },

   /* ('ieq', ('u2u32', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u32', ('u2u16', 'b')), 'b')) */
   /* search1597_0_0 -> 11 in the cache */
   /* search1597_0 -> 1749 in the cache */
   /* search1597_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1749, 33 },
      -1,
   } },

   /* replace1597_0_0 -> 11 in the cache */
   /* replace1597_0_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4453 },
      -1,
   } },
   /* replace1597_1_0_0_0 -> 36 in the cache */
   /* replace1597_1_0_0 -> 4453 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4453 },
      -1,
   } },
   /* replace1597_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4455, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4454, 4456 },
      -1,
   } },

   /* ('ine', ('u2u32', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u32', ('u2u16', 'b')), 'b')) */
   /* search1598_0_0 -> 11 in the cache */
   /* search1598_0 -> 1749 in the cache */
   /* search1598_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1749, 33 },
      -1,
   } },

   /* replace1598_0_0 -> 11 in the cache */
   /* replace1598_0_1_0 -> 36 in the cache */
   /* replace1598_0_1 -> 4453 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4453 },
      -1,
   } },
   /* replace1598_1_0_0_0 -> 36 in the cache */
   /* replace1598_1_0_0 -> 4453 in the cache */
   /* replace1598_1_0 -> 4455 in the cache */
   /* replace1598_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4455, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4459, 4460 },
      -1,
   } },

   /* ('ult', ('u2u32', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b')))) */
   /* search1599_0_0 -> 11 in the cache */
   /* search1599_0 -> 1749 in the cache */
   /* search1599_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1749, 33 },
      -1,
   } },

   /* replace1599_0_0 -> 124 in the cache */
   /* replace1599_0_1 -> 36 in the cache */
   /* replace1599_0 -> 4356 in the cache */
   /* replace1599_1_0_0 -> 964 in the cache */
   /* replace1599_1_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 964, 36 },
      -1,
   } },
   /* replace1599_1_1_0 -> 11 in the cache */
   /* replace1599_1_1_1_0 -> 36 in the cache */
   /* replace1599_1_1_1 -> 4453 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4453 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4463, 4464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4356, 4465 },
      -1,
   } },

   /* ('ult', '#a', ('u2u32', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b'))) */
   /* search1600_0 -> 275 in the cache */
   /* search1600_1_0 -> 247 in the cache */
   /* search1600_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 275, 3031 },
      -1,
   } },

   /* replace1600_0_0 -> 16 in the cache */
   /* replace1600_0_1 -> 964 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 964 },
      -1,
   } },
   /* replace1600_1_0_0 -> 16 in the cache */
   /* replace1600_1_0_1 -> 124 in the cache */
   /* replace1600_1_0 -> 4363 in the cache */
   /* replace1600_1_1_0_0 -> 16 in the cache */
   /* replace1600_1_1_0 -> 3037 in the cache */
   /* replace1600_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3037, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4363, 4469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4468, 4470 },
      -1,
   } },

   /* ('uge', ('u2u32', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b')))) */
   /* search1601_0_0 -> 11 in the cache */
   /* search1601_0 -> 1749 in the cache */
   /* search1601_1 -> 33 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1749, 33 },
      -1,
   } },

   /* replace1601_0_0 -> 964 in the cache */
   /* replace1601_0_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 964, 36 },
      -1,
   } },
   /* replace1601_1_0_0 -> 124 in the cache */
   /* replace1601_1_0_1 -> 36 in the cache */
   /* replace1601_1_0 -> 4369 in the cache */
   /* replace1601_1_1_0 -> 11 in the cache */
   /* replace1601_1_1_1_0 -> 36 in the cache */
   /* replace1601_1_1_1 -> 4453 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4453 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4369, 4474 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4473, 4475 },
      -1,
   } },

   /* ('uge', '#a', ('u2u32', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b'))) */
   /* search1602_0 -> 275 in the cache */
   /* search1602_1_0 -> 247 in the cache */
   /* search1602_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 275, 3031 },
      -1,
   } },

   /* replace1602_0_0 -> 16 in the cache */
   /* replace1602_0_1 -> 124 in the cache */
   /* replace1602_0 -> 4374 in the cache */
   /* replace1602_1_0_0 -> 16 in the cache */
   /* replace1602_1_0_1 -> 964 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 964 },
      -1,
   } },
   /* replace1602_1_1_0_0 -> 16 in the cache */
   /* replace1602_1_1_0 -> 3037 in the cache */
   /* replace1602_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3037, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4478, 4479 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4374, 4480 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1603_0_0 -> 11 in the cache */
   /* search1603_0 -> 1772 in the cache */
   /* search1603_1_0 -> 2060 in the cache */
   /* search1603_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1772, 4379 },
      -1,
   } },

   /* replace1603_0 -> 11 in the cache */
   /* replace1603_1_0 -> 2060 in the cache */
   /* replace1603_1 -> 4289 in the cache */
   /* replace1603 -> 4425 in the cache */

   /* ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1604_0_0 -> 11 in the cache */
   /* search1604_0 -> 1772 in the cache */
   /* search1604_1_0 -> 2060 in the cache */
   /* search1604_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1772, 4379 },
      -1,
   } },

   /* replace1604_0 -> 11 in the cache */
   /* replace1604_1_0 -> 2060 in the cache */
   /* replace1604_1 -> 4289 in the cache */
   /* replace1604 -> 4427 in the cache */

   /* ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1605_0_0 -> 11 in the cache */
   /* search1605_0 -> 1772 in the cache */
   /* search1605_1_0 -> 2060 in the cache */
   /* search1605_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1772, 4379 },
      -1,
   } },

   /* replace1605_0 -> 11 in the cache */
   /* replace1605_1_0 -> 2060 in the cache */
   /* replace1605_1 -> 4289 in the cache */
   /* replace1605 -> 4429 in the cache */

   /* ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1606_0_0 -> 11 in the cache */
   /* search1606_0 -> 1772 in the cache */
   /* search1606_1_0 -> 2060 in the cache */
   /* search1606_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1772, 4379 },
      -1,
   } },

   /* replace1606_0 -> 11 in the cache */
   /* replace1606_1_0 -> 2060 in the cache */
   /* replace1606_1 -> 4289 in the cache */
   /* replace1606 -> 4431 in the cache */

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1607_0_0 -> 3954 in the cache */
   /* search1607_0 -> 4384 in the cache */
   /* search1607_1_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4384, 4486 },
      -1,
   } },

   /* replace1607_0_0 -> 3954 in the cache */
   /* replace1607_0 -> 4299 in the cache */
   /* replace1607_1 -> 511 in the cache */
   /* replace1607 -> 4434 in the cache */

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1608_0_0 -> 3954 in the cache */
   /* search1608_0 -> 4384 in the cache */
   /* search1608_1_0 -> 511 in the cache */
   /* search1608_1 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4384, 4486 },
      -1,
   } },

   /* replace1608_0_0 -> 3954 in the cache */
   /* replace1608_0 -> 4299 in the cache */
   /* replace1608_1 -> 511 in the cache */
   /* replace1608 -> 4436 in the cache */

   /* ('ieq', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u16', 'b')) */
   /* search1609_0_0 -> 11 in the cache */
   /* search1609_0 -> 1772 in the cache */
   /* search1609_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1772, 4489 },
      -1,
   } },

   /* replace1609_0 -> 11 in the cache */
   /* replace1609_1_0 -> 247 in the cache */
   /* replace1609_1 -> 4438 in the cache */
   /* replace1609 -> 4439 in the cache */

   /* ('ine', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u16', 'b')) */
   /* search1610_0_0 -> 11 in the cache */
   /* search1610_0 -> 1772 in the cache */
   /* search1610_1_0 -> 247 in the cache */
   /* search1610_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1772, 4489 },
      -1,
   } },

   /* replace1610_0 -> 11 in the cache */
   /* replace1610_1_0 -> 247 in the cache */
   /* replace1610_1 -> 4438 in the cache */
   /* replace1610 -> 4441 in the cache */

   /* ('uge', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u16', 'b')) */
   /* search1611_0_0 -> 11 in the cache */
   /* search1611_0 -> 1772 in the cache */
   /* search1611_1_0 -> 247 in the cache */
   /* search1611_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1772, 4489 },
      -1,
   } },

   /* replace1611_0 -> 11 in the cache */
   /* replace1611_1_0 -> 247 in the cache */
   /* replace1611_1 -> 4438 in the cache */
   /* replace1611 -> 4443 in the cache */

   /* ('ult', ('u2u64', 'a@16'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u16', 'b')) */
   /* search1612_0_0 -> 11 in the cache */
   /* search1612_0 -> 1772 in the cache */
   /* search1612_1_0 -> 247 in the cache */
   /* search1612_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1772, 4489 },
      -1,
   } },

   /* replace1612_0 -> 11 in the cache */
   /* replace1612_1_0 -> 247 in the cache */
   /* replace1612_1 -> 4438 in the cache */
   /* replace1612 -> 4445 in the cache */

   /* ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('uge', ('u2u16', 'b'), 'a') */
   /* search1613_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 500 },
      -1,
   } },
   /* search1613_1_0 -> 511 in the cache */
   /* search1613_1 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4494, 4486 },
      -1,
   } },

   /* replace1613_0_0 -> 500 in the cache */
   /* replace1613_0 -> 4448 in the cache */
   /* replace1613_1 -> 511 in the cache */
   /* replace1613 -> 4449 in the cache */

   /* ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@16')) => ('ult', ('u2u16', 'b'), 'a') */
   /* search1614_0_0 -> 500 in the cache */
   /* search1614_0 -> 4494 in the cache */
   /* search1614_1_0 -> 511 in the cache */
   /* search1614_1 -> 4486 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4494, 4486 },
      -1,
   } },

   /* replace1614_0_0 -> 500 in the cache */
   /* replace1614_0 -> 4448 in the cache */
   /* replace1614_1 -> 511 in the cache */
   /* replace1614 -> 4451 in the cache */

   /* ('ieq', ('u2u64', 'a@16'), '#b') => ('iand', ('ieq', 'a', ('u2u16', 'b')), ('ieq', ('u2u64', ('u2u16', 'b')), 'b')) */
   /* search1615_0_0 -> 11 in the cache */
   /* search1615_0 -> 1772 in the cache */
   /* search1615_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1772, 4055 },
      -1,
   } },

   /* replace1615_0_0 -> 11 in the cache */
   /* replace1615_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 11, 4498 },
      -1,
   } },
   /* replace1615_1_0_0_0 -> 41 in the cache */
   /* replace1615_1_0_0 -> 4498 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4498 },
      -1,
   } },
   /* replace1615_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4500, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4499, 4501 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@16'), '#b') => ('ior', ('ine', 'a', ('u2u16', 'b')), ('ine', ('u2u64', ('u2u16', 'b')), 'b')) */
   /* search1616_0_0 -> 11 in the cache */
   /* search1616_0 -> 1772 in the cache */
   /* search1616_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1772, 4055 },
      -1,
   } },

   /* replace1616_0_0 -> 11 in the cache */
   /* replace1616_0_1_0 -> 41 in the cache */
   /* replace1616_0_1 -> 4498 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 11, 4498 },
      -1,
   } },
   /* replace1616_1_0_0_0 -> 41 in the cache */
   /* replace1616_1_0_0 -> 4498 in the cache */
   /* replace1616_1_0 -> 4500 in the cache */
   /* replace1616_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4500, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4504, 4505 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@16'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 65535, 'b'), ('ult', 'a', ('u2u16', 'b')))) */
   /* search1617_0_0 -> 11 in the cache */
   /* search1617_0 -> 1772 in the cache */
   /* search1617_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1772, 4055 },
      -1,
   } },

   /* replace1617_0_0 -> 859 in the cache */
   /* replace1617_0_1 -> 41 in the cache */
   /* replace1617_0 -> 4399 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0xffff /* 65535 */ },
   } },
   /* replace1617_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4508, 41 },
      -1,
   } },
   /* replace1617_1_1_0 -> 11 in the cache */
   /* replace1617_1_1_1_0 -> 41 in the cache */
   /* replace1617_1_1_1 -> 4498 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 11, 4498 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4509, 4510 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4399, 4511 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@16')) => ('iand', ('ult', 'a', 65535), ('ior', ('ult', 'a', 0), ('ult', ('u2u16', 'a'), 'b'))) */
   /* search1618_0 -> 4074 in the cache */
   /* search1618_1_0 -> 247 in the cache */
   /* search1618_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4074, 4489 },
      -1,
   } },

   /* replace1618_0_0 -> 21 in the cache */
   /* replace1618_0_1 -> 4508 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 4508 },
      -1,
   } },
   /* replace1618_1_0_0 -> 21 in the cache */
   /* replace1618_1_0_1 -> 859 in the cache */
   /* replace1618_1_0 -> 4407 in the cache */
   /* replace1618_1_1_0_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 21 },
      -1,
   } },
   /* replace1618_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4515, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4407, 4516 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4514, 4517 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@16'), '#b') => ('iand', ('uge', 65535, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u16', 'b')))) */
   /* search1619_0_0 -> 11 in the cache */
   /* search1619_0 -> 1772 in the cache */
   /* search1619_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1772, 4055 },
      -1,
   } },

   /* replace1619_0_0 -> 4508 in the cache */
   /* replace1619_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4508, 41 },
      -1,
   } },
   /* replace1619_1_0_0 -> 859 in the cache */
   /* replace1619_1_0_1 -> 41 in the cache */
   /* replace1619_1_0 -> 4414 in the cache */
   /* replace1619_1_1_0 -> 11 in the cache */
   /* replace1619_1_1_1_0 -> 41 in the cache */
   /* replace1619_1_1_1 -> 4498 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 11, 4498 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4414, 4521 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4520, 4522 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@16')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 65535), ('uge', ('u2u16', 'a'), 'b'))) */
   /* search1620_0 -> 4074 in the cache */
   /* search1620_1_0 -> 247 in the cache */
   /* search1620_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4074, 4489 },
      -1,
   } },

   /* replace1620_0_0 -> 21 in the cache */
   /* replace1620_0_1 -> 859 in the cache */
   /* replace1620_0 -> 4419 in the cache */
   /* replace1620_1_0_0 -> 21 in the cache */
   /* replace1620_1_0_1 -> 4508 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 4508 },
      -1,
   } },
   /* replace1620_1_1_0_0 -> 21 in the cache */
   /* replace1620_1_1_0 -> 4515 in the cache */
   /* replace1620_1_1_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4515, 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4525, 4526 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4419, 4527 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1621_0_0 -> 16 in the cache */
   /* search1621_0 -> 1856 in the cache */
   /* search1621_1_0 -> 2060 in the cache */
   /* search1621_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1856, 4379 },
      -1,
   } },

   /* replace1621_0 -> 16 in the cache */
   /* replace1621_1_0 -> 2060 in the cache */
   /* replace1621_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3011 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1622_0_0 -> 16 in the cache */
   /* search1622_0 -> 1856 in the cache */
   /* search1622_1_0 -> 2060 in the cache */
   /* search1622_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1856, 4379 },
      -1,
   } },

   /* replace1622_0 -> 16 in the cache */
   /* replace1622_1_0 -> 2060 in the cache */
   /* replace1622_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 3011 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1623_0_0 -> 16 in the cache */
   /* search1623_0 -> 1856 in the cache */
   /* search1623_1_0 -> 2060 in the cache */
   /* search1623_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1856, 4379 },
      -1,
   } },

   /* replace1623_0 -> 16 in the cache */
   /* replace1623_1_0 -> 2060 in the cache */
   /* replace1623_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 3011 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@8')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1624_0_0 -> 16 in the cache */
   /* search1624_0 -> 1856 in the cache */
   /* search1624_1_0 -> 2060 in the cache */
   /* search1624_1 -> 4379 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1856, 4379 },
      -1,
   } },

   /* replace1624_0 -> 16 in the cache */
   /* replace1624_1_0 -> 2060 in the cache */
   /* replace1624_1 -> 3011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 3011 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1625_0_0 -> 3954 in the cache */
   /* search1625_0 -> 4384 in the cache */
   /* search1625_1_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4384, 4537 },
      -1,
   } },

   /* replace1625_0_0 -> 3954 in the cache */
   /* replace1625_0 -> 4342 in the cache */
   /* replace1625_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4342, 364 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@8'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1626_0_0 -> 3954 in the cache */
   /* search1626_0 -> 4384 in the cache */
   /* search1626_1_0 -> 364 in the cache */
   /* search1626_1 -> 4537 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4384, 4537 },
      -1,
   } },

   /* replace1626_0_0 -> 3954 in the cache */
   /* replace1626_0 -> 4342 in the cache */
   /* replace1626_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4342, 364 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1627_0_0 -> 16 in the cache */
   /* search1627_0 -> 1856 in the cache */
   /* search1627_1_0 -> 247 in the cache */
   /* search1627_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1856, 4489 },
      -1,
   } },

   /* replace1627_0 -> 16 in the cache */
   /* replace1627_1_0 -> 247 in the cache */
   /* replace1627_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 3031 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1628_0_0 -> 16 in the cache */
   /* search1628_0 -> 1856 in the cache */
   /* search1628_1_0 -> 247 in the cache */
   /* search1628_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1856, 4489 },
      -1,
   } },

   /* replace1628_0 -> 16 in the cache */
   /* replace1628_1_0 -> 247 in the cache */
   /* replace1628_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 3031 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1629_0_0 -> 16 in the cache */
   /* search1629_0 -> 1856 in the cache */
   /* search1629_1_0 -> 247 in the cache */
   /* search1629_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1856, 4489 },
      -1,
   } },

   /* replace1629_0 -> 16 in the cache */
   /* replace1629_1_0 -> 247 in the cache */
   /* replace1629_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 3031 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@16')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1630_0_0 -> 16 in the cache */
   /* search1630_0 -> 1856 in the cache */
   /* search1630_1_0 -> 247 in the cache */
   /* search1630_1 -> 4489 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1856, 4489 },
      -1,
   } },

   /* replace1630_0 -> 16 in the cache */
   /* replace1630_1_0 -> 247 in the cache */
   /* replace1630_1 -> 3031 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 3031 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1631_0_0 -> 500 in the cache */
   /* search1631_0 -> 4494 in the cache */
   /* search1631_1_0 -> 364 in the cache */
   /* search1631_1 -> 4537 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4494, 4537 },
      -1,
   } },

   /* replace1631_0_0 -> 500 in the cache */
   /* replace1631_0 -> 4446 in the cache */
   /* replace1631_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4446, 364 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@16'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1632_0_0 -> 500 in the cache */
   /* search1632_0 -> 4494 in the cache */
   /* search1632_1_0 -> 364 in the cache */
   /* search1632_1 -> 4537 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4494, 4537 },
      -1,
   } },

   /* replace1632_0_0 -> 500 in the cache */
   /* replace1632_0 -> 4446 in the cache */
   /* replace1632_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4446, 364 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ieq', 'a', ('u2u32', 'b')) */
   /* search1633_0_0 -> 16 in the cache */
   /* search1633_0 -> 1856 in the cache */
   /* search1633_1_0 -> 36 in the cache */
   /* search1633_1 -> 3027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1856, 3027 },
      -1,
   } },

   /* replace1633_0 -> 16 in the cache */
   /* replace1633_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 16, 4555 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ine', 'a', ('u2u32', 'b')) */
   /* search1634_0_0 -> 16 in the cache */
   /* search1634_0 -> 1856 in the cache */
   /* search1634_1_0 -> 36 in the cache */
   /* search1634_1 -> 3027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1856, 3027 },
      -1,
   } },

   /* replace1634_0 -> 16 in the cache */
   /* replace1634_1_0 -> 36 in the cache */
   /* replace1634_1 -> 4555 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 16, 4555 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('uge', 'a', ('u2u32', 'b')) */
   /* search1635_0_0 -> 16 in the cache */
   /* search1635_0 -> 1856 in the cache */
   /* search1635_1_0 -> 36 in the cache */
   /* search1635_1 -> 3027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1856, 3027 },
      -1,
   } },

   /* replace1635_0 -> 16 in the cache */
   /* replace1635_1_0 -> 36 in the cache */
   /* replace1635_1 -> 4555 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 4555 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), ('u2u64', 'b@32')) => ('ult', 'a', ('u2u32', 'b')) */
   /* search1636_0_0 -> 16 in the cache */
   /* search1636_0 -> 1856 in the cache */
   /* search1636_1_0 -> 36 in the cache */
   /* search1636_1 -> 3027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1856, 3027 },
      -1,
   } },

   /* replace1636_0 -> 16 in the cache */
   /* replace1636_1_0 -> 36 in the cache */
   /* replace1636_1 -> 4555 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 4555 },
      -1,
   } },

   /* ('uge', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('uge', ('u2u32', 'b'), 'a') */
   /* search1637_0_0 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 362 },
      -1,
   } },
   /* search1637_1_0 -> 364 in the cache */
   /* search1637_1 -> 4537 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4563, 4537 },
      -1,
   } },

   /* replace1637_0_0 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 362 },
      -1,
   } },
   /* replace1637_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4565, 364 },
      -1,
   } },

   /* ('ult', ('u2u64', 'b@32'), ('u2u64', 'a@32')) => ('ult', ('u2u32', 'b'), 'a') */
   /* search1638_0_0 -> 362 in the cache */
   /* search1638_0 -> 4563 in the cache */
   /* search1638_1_0 -> 364 in the cache */
   /* search1638_1 -> 4537 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4563, 4537 },
      -1,
   } },

   /* replace1638_0_0 -> 362 in the cache */
   /* replace1638_0 -> 4565 in the cache */
   /* replace1638_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4565, 364 },
      -1,
   } },

   /* ('ieq', ('u2u64', 'a@32'), '#b') => ('iand', ('ieq', 'a', ('u2u32', 'b')), ('ieq', ('u2u64', ('u2u32', 'b')), 'b')) */
   /* search1639_0_0 -> 16 in the cache */
   /* search1639_0 -> 1856 in the cache */
   /* search1639_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 1856, 4055 },
      -1,
   } },

   /* replace1639_0_0 -> 16 in the cache */
   /* replace1639_0_1_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 16, 4570 },
      -1,
   } },
   /* replace1639_1_0_0_0 -> 41 in the cache */
   /* replace1639_1_0_0 -> 4570 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_u2u64,
      -1, 0,
      { 4570 },
      -1,
   } },
   /* replace1639_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 4572, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 4571, 4573 },
      -1,
   } },

   /* ('ine', ('u2u64', 'a@32'), '#b') => ('ior', ('ine', 'a', ('u2u32', 'b')), ('ine', ('u2u64', ('u2u32', 'b')), 'b')) */
   /* search1640_0_0 -> 16 in the cache */
   /* search1640_0 -> 1856 in the cache */
   /* search1640_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 1856, 4055 },
      -1,
   } },

   /* replace1640_0_0 -> 16 in the cache */
   /* replace1640_0_1_0 -> 41 in the cache */
   /* replace1640_0_1 -> 4570 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 16, 4570 },
      -1,
   } },
   /* replace1640_1_0_0_0 -> 41 in the cache */
   /* replace1640_1_0_0 -> 4570 in the cache */
   /* replace1640_1_0 -> 4572 in the cache */
   /* replace1640_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 4572, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 4576, 4577 },
      -1,
   } },

   /* ('ult', ('u2u64', 'a@32'), '#b') => ('iand', ('ult', 0, 'b'), ('ior', ('ult', 4294967295, 'b'), ('ult', 'a', ('u2u32', 'b')))) */
   /* search1641_0_0 -> 16 in the cache */
   /* search1641_0 -> 1856 in the cache */
   /* search1641_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 1856, 4055 },
      -1,
   } },

   /* replace1641_0_0 -> 859 in the cache */
   /* replace1641_0_1 -> 41 in the cache */
   /* replace1641_0 -> 4399 in the cache */
   /* replace1641_1_0_0 -> 3042 in the cache */
   /* replace1641_1_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3042, 41 },
      -1,
   } },
   /* replace1641_1_1_0 -> 16 in the cache */
   /* replace1641_1_1_1_0 -> 41 in the cache */
   /* replace1641_1_1_1 -> 4570 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 16, 4570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4580, 4581 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4399, 4582 },
      -1,
   } },

   /* ('ult', '#a', ('u2u64', 'b@32')) => ('iand', ('ult', 'a', 4294967295), ('ior', ('ult', 'a', 0), ('ult', ('u2u32', 'a'), 'b'))) */
   /* search1642_0 -> 4074 in the cache */
   /* search1642_1_0 -> 36 in the cache */
   /* search1642_1 -> 3027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 4074, 3027 },
      -1,
   } },

   /* replace1642_0_0 -> 21 in the cache */
   /* replace1642_0_1 -> 3042 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 21, 3042 },
      -1,
   } },
   /* replace1642_1_0_0 -> 21 in the cache */
   /* replace1642_1_0_1 -> 859 in the cache */
   /* replace1642_1_0 -> 4407 in the cache */
   /* replace1642_1_1_0_0 -> 21 in the cache */
   /* replace1642_1_1_0 -> 3034 in the cache */
   /* replace1642_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 3034, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4407, 4586 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4585, 4587 },
      -1,
   } },

   /* ('uge', ('u2u64', 'a@32'), '#b') => ('iand', ('uge', 4294967295, 'b'), ('ior', ('uge', 0, 'b'), ('uge', 'a', ('u2u32', 'b')))) */
   /* search1643_0_0 -> 16 in the cache */
   /* search1643_0 -> 1856 in the cache */
   /* search1643_1 -> 4055 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 1856, 4055 },
      -1,
   } },

   /* replace1643_0_0 -> 3042 in the cache */
   /* replace1643_0_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3042, 41 },
      -1,
   } },
   /* replace1643_1_0_0 -> 859 in the cache */
   /* replace1643_1_0_1 -> 41 in the cache */
   /* replace1643_1_0 -> 4414 in the cache */
   /* replace1643_1_1_0 -> 16 in the cache */
   /* replace1643_1_1_1_0 -> 41 in the cache */
   /* replace1643_1_1_1 -> 4570 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 16, 4570 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4414, 4591 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4590, 4592 },
      -1,
   } },

   /* ('uge', '#a', ('u2u64', 'b@32')) => ('iand', ('uge', 'a', 0), ('ior', ('uge', 'a', 4294967295), ('uge', ('u2u32', 'a'), 'b'))) */
   /* search1644_0 -> 4074 in the cache */
   /* search1644_1_0 -> 36 in the cache */
   /* search1644_1 -> 3027 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 4074, 3027 },
      -1,
   } },

   /* replace1644_0_0 -> 21 in the cache */
   /* replace1644_0_1 -> 859 in the cache */
   /* replace1644_0 -> 4419 in the cache */
   /* replace1644_1_0_0 -> 21 in the cache */
   /* replace1644_1_0_1 -> 3042 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 21, 3042 },
      -1,
   } },
   /* replace1644_1_1_0_0 -> 21 in the cache */
   /* replace1644_1_1_0 -> 3034 in the cache */
   /* replace1644_1_1_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 3034, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 4595, 4596 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 4419, 4597 },
      -1,
   } },

   /* ('feq', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('feq', 'a', ('f2f16', 'b')) */
   /* search1645_0_0 -> 11 in the cache */
   /* search1645_0 -> 1745 in the cache */
   /* search1645_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1745, 4599 },
      -1,
   } },

   /* replace1645_0 -> 11 in the cache */
   /* replace1645_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 11, 4601 },
      -1,
   } },

   /* ('fneu', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fneu', 'a', ('f2f16', 'b')) */
   /* search1646_0_0 -> 11 in the cache */
   /* search1646_0 -> 1745 in the cache */
   /* search1646_1_0 -> 247 in the cache */
   /* search1646_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1745, 4599 },
      -1,
   } },

   /* replace1646_0 -> 11 in the cache */
   /* replace1646_1_0 -> 247 in the cache */
   /* replace1646_1 -> 4601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 11, 4601 },
      -1,
   } },

   /* ('fge', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('fge', 'a', ('f2f16', 'b')) */
   /* search1647_0_0 -> 11 in the cache */
   /* search1647_0 -> 1745 in the cache */
   /* search1647_1_0 -> 247 in the cache */
   /* search1647_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1745, 4599 },
      -1,
   } },

   /* replace1647_0 -> 11 in the cache */
   /* replace1647_1_0 -> 247 in the cache */
   /* replace1647_1 -> 4601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 11, 4601 },
      -1,
   } },

   /* ('flt', ('f2f32', 'a@16'), ('f2f32', 'b@16')) => ('flt', 'a', ('f2f16', 'b')) */
   /* search1648_0_0 -> 11 in the cache */
   /* search1648_0 -> 1745 in the cache */
   /* search1648_1_0 -> 247 in the cache */
   /* search1648_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1745, 4599 },
      -1,
   } },

   /* replace1648_0 -> 11 in the cache */
   /* replace1648_1_0 -> 247 in the cache */
   /* replace1648_1 -> 4601 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 11, 4601 },
      -1,
   } },

   /* ('fge', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('fge', ('f2f16', 'b'), 'a') */
   /* search1649_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 500 },
      -1,
   } },
   /* search1649_1_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4609, 4610 },
      -1,
   } },

   /* replace1649_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 500 },
      -1,
   } },
   /* replace1649_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4612, 511 },
      -1,
   } },

   /* ('flt', ('f2f32', 'b@16'), ('f2f32', 'a@16')) => ('flt', ('f2f16', 'b'), 'a') */
   /* search1650_0_0 -> 500 in the cache */
   /* search1650_0 -> 4609 in the cache */
   /* search1650_1_0 -> 511 in the cache */
   /* search1650_1 -> 4610 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4609, 4610 },
      -1,
   } },

   /* replace1650_0_0 -> 500 in the cache */
   /* replace1650_0 -> 4612 in the cache */
   /* replace1650_1 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4612, 511 },
      -1,
   } },

   /* ('feq', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f16', 'b')) */
   /* search1651_0_0 -> 11 in the cache */
   /* search1651_0 -> 1768 in the cache */
   /* search1651_1_0 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 247 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1768, 4616 },
      -1,
   } },

   /* replace1651_0 -> 11 in the cache */
   /* replace1651_1_0 -> 247 in the cache */
   /* replace1651_1 -> 4601 in the cache */
   /* replace1651 -> 4602 in the cache */

   /* ('fneu', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f16', 'b')) */
   /* search1652_0_0 -> 11 in the cache */
   /* search1652_0 -> 1768 in the cache */
   /* search1652_1_0 -> 247 in the cache */
   /* search1652_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1768, 4616 },
      -1,
   } },

   /* replace1652_0 -> 11 in the cache */
   /* replace1652_1_0 -> 247 in the cache */
   /* replace1652_1 -> 4601 in the cache */
   /* replace1652 -> 4604 in the cache */

   /* ('fge', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f16', 'b')) */
   /* search1653_0_0 -> 11 in the cache */
   /* search1653_0 -> 1768 in the cache */
   /* search1653_1_0 -> 247 in the cache */
   /* search1653_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1768, 4616 },
      -1,
   } },

   /* replace1653_0 -> 11 in the cache */
   /* replace1653_1_0 -> 247 in the cache */
   /* replace1653_1 -> 4601 in the cache */
   /* replace1653 -> 4606 in the cache */

   /* ('flt', ('f2f64', 'a@16'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f16', 'b')) */
   /* search1654_0_0 -> 11 in the cache */
   /* search1654_0 -> 1768 in the cache */
   /* search1654_1_0 -> 247 in the cache */
   /* search1654_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1768, 4616 },
      -1,
   } },

   /* replace1654_0 -> 11 in the cache */
   /* replace1654_1_0 -> 247 in the cache */
   /* replace1654_1 -> 4601 in the cache */
   /* replace1654 -> 4608 in the cache */

   /* ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('fge', ('f2f16', 'b'), 'a') */
   /* search1655_0_0 -> 500 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 500 },
      -1,
   } },
   /* search1655_1_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4621, 4622 },
      -1,
   } },

   /* replace1655_0_0 -> 500 in the cache */
   /* replace1655_0 -> 4612 in the cache */
   /* replace1655_1 -> 511 in the cache */
   /* replace1655 -> 4613 in the cache */

   /* ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@16')) => ('flt', ('f2f16', 'b'), 'a') */
   /* search1656_0_0 -> 500 in the cache */
   /* search1656_0 -> 4621 in the cache */
   /* search1656_1_0 -> 511 in the cache */
   /* search1656_1 -> 4622 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4621, 4622 },
      -1,
   } },

   /* replace1656_0_0 -> 500 in the cache */
   /* replace1656_0 -> 4612 in the cache */
   /* replace1656_1 -> 511 in the cache */
   /* replace1656 -> 4615 in the cache */

   /* ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('feq', 'a', ('f2f32', 'b')) */
   /* search1657_0_0 -> 16 in the cache */
   /* search1657_0 -> 1852 in the cache */
   /* search1657_1_0 -> 247 in the cache */
   /* search1657_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1852, 4616 },
      -1,
   } },

   /* replace1657_0 -> 16 in the cache */
   /* replace1657_1_0 -> 247 in the cache */
   /* replace1657_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 4599 },
      -1,
   } },

   /* ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fneu', 'a', ('f2f32', 'b')) */
   /* search1658_0_0 -> 16 in the cache */
   /* search1658_0 -> 1852 in the cache */
   /* search1658_1_0 -> 247 in the cache */
   /* search1658_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1852, 4616 },
      -1,
   } },

   /* replace1658_0 -> 16 in the cache */
   /* replace1658_1_0 -> 247 in the cache */
   /* replace1658_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 16, 4599 },
      -1,
   } },

   /* ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('fge', 'a', ('f2f32', 'b')) */
   /* search1659_0_0 -> 16 in the cache */
   /* search1659_0 -> 1852 in the cache */
   /* search1659_1_0 -> 247 in the cache */
   /* search1659_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1852, 4616 },
      -1,
   } },

   /* replace1659_0 -> 16 in the cache */
   /* replace1659_1_0 -> 247 in the cache */
   /* replace1659_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 4599 },
      -1,
   } },

   /* ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@16')) => ('flt', 'a', ('f2f32', 'b')) */
   /* search1660_0_0 -> 16 in the cache */
   /* search1660_0 -> 1852 in the cache */
   /* search1660_1_0 -> 247 in the cache */
   /* search1660_1 -> 4616 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1852, 4616 },
      -1,
   } },

   /* replace1660_0 -> 16 in the cache */
   /* replace1660_1_0 -> 247 in the cache */
   /* replace1660_1 -> 4599 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 4599 },
      -1,
   } },

   /* ('fge', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a') */
   /* search1661_0_0 -> 500 in the cache */
   /* search1661_0 -> 4621 in the cache */
   /* search1661_1_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4621, 4633 },
      -1,
   } },

   /* replace1661_0_0 -> 500 in the cache */
   /* replace1661_0 -> 4609 in the cache */
   /* replace1661_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4609, 364 },
      -1,
   } },

   /* ('flt', ('f2f64', 'b@16'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a') */
   /* search1662_0_0 -> 500 in the cache */
   /* search1662_0 -> 4621 in the cache */
   /* search1662_1_0 -> 364 in the cache */
   /* search1662_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4621, 4633 },
      -1,
   } },

   /* replace1662_0_0 -> 500 in the cache */
   /* replace1662_0 -> 4609 in the cache */
   /* replace1662_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4609, 364 },
      -1,
   } },

   /* ('feq', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('feq', 'a', ('f2f32', 'b')) */
   /* search1663_0_0 -> 16 in the cache */
   /* search1663_0 -> 1852 in the cache */
   /* search1663_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1852, 4638 },
      -1,
   } },

   /* replace1663_0 -> 16 in the cache */
   /* replace1663_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 16, 4640 },
      -1,
   } },

   /* ('fneu', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fneu', 'a', ('f2f32', 'b')) */
   /* search1664_0_0 -> 16 in the cache */
   /* search1664_0 -> 1852 in the cache */
   /* search1664_1_0 -> 36 in the cache */
   /* search1664_1 -> 4638 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1852, 4638 },
      -1,
   } },

   /* replace1664_0 -> 16 in the cache */
   /* replace1664_1_0 -> 36 in the cache */
   /* replace1664_1 -> 4640 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 16, 4640 },
      -1,
   } },

   /* ('fge', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('fge', 'a', ('f2f32', 'b')) */
   /* search1665_0_0 -> 16 in the cache */
   /* search1665_0 -> 1852 in the cache */
   /* search1665_1_0 -> 36 in the cache */
   /* search1665_1 -> 4638 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1852, 4638 },
      -1,
   } },

   /* replace1665_0 -> 16 in the cache */
   /* replace1665_1_0 -> 36 in the cache */
   /* replace1665_1 -> 4640 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 16, 4640 },
      -1,
   } },

   /* ('flt', ('f2f64', 'a@32'), ('f2f64', 'b@32')) => ('flt', 'a', ('f2f32', 'b')) */
   /* search1666_0_0 -> 16 in the cache */
   /* search1666_0 -> 1852 in the cache */
   /* search1666_1_0 -> 36 in the cache */
   /* search1666_1 -> 4638 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1852, 4638 },
      -1,
   } },

   /* replace1666_0 -> 16 in the cache */
   /* replace1666_1_0 -> 36 in the cache */
   /* replace1666_1 -> 4640 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 16, 4640 },
      -1,
   } },

   /* ('fge', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('fge', ('f2f32', 'b'), 'a') */
   /* search1667_0_0 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_f2f64,
      -1, 0,
      { 362 },
      -1,
   } },
   /* search1667_1_0 -> 364 in the cache */
   /* search1667_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4648, 4633 },
      -1,
   } },

   /* replace1667_0_0 -> 362 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 362 },
      -1,
   } },
   /* replace1667_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 4650, 364 },
      -1,
   } },

   /* ('flt', ('f2f64', 'b@32'), ('f2f64', 'a@32')) => ('flt', ('f2f32', 'b'), 'a') */
   /* search1668_0_0 -> 362 in the cache */
   /* search1668_0 -> 4648 in the cache */
   /* search1668_1_0 -> 364 in the cache */
   /* search1668_1 -> 4633 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4648, 4633 },
      -1,
   } },

   /* replace1668_0_0 -> 362 in the cache */
   /* replace1668_0 -> 4650 in the cache */
   /* replace1668_1 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 4650, 364 },
      -1,
   } },

   /* ('i2i32', ('iand', 'a@64', 4294967295)) => ('u2u32', 'a') */
   /* search1669_0_0 -> 21 in the cache */
   /* search1669_0_1 -> 3042 in the cache */
   /* search1669_0 -> 3043 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 1,
      { 3043 },
      -1,
   } },

   /* replace1669_0 -> 21 in the cache */
   /* replace1669 -> 3034 in the cache */

   /* ('i2i16', ('iand', 'a@32', 65535)) => ('u2u16', 'a') */
   /* search1670_0_0 -> 16 in the cache */
   /* search1670_0_1 -> 964 in the cache */
   /* search1670_0 -> 965 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 1,
      { 965 },
      -1,
   } },

   /* replace1670_0 -> 16 in the cache */
   /* replace1670 -> 3037 in the cache */

   /* ('i2i16', ('iand', 'a@64', 65535)) => ('u2u16', 'a') */
   /* search1671_0_0 -> 21 in the cache */
   /* search1671_0_1 -> 4508 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 4508 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 1,
      { 4656 },
      -1,
   } },

   /* replace1671_0 -> 21 in the cache */
   /* replace1671 -> 4515 in the cache */

   /* ('i2i8', ('iand', 'a@16', 255)) => ('u2u8', 'a') */
   /* search1672_0_0 -> 11 in the cache */
   /* search1672_0_1 -> 1961 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 11, 1961 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 4658 },
      -1,
   } },

   /* replace1672_0 -> 11 in the cache */
   /* replace1672 -> 1957 in the cache */

   /* ('i2i8', ('iand', 'a@32', 255)) => ('u2u8', 'a') */
   /* search1673_0_0 -> 16 in the cache */
   /* search1673_0_1 -> 972 in the cache */
   /* search1673_0 -> 973 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 973 },
      -1,
   } },

   /* replace1673_0 -> 16 in the cache */
   /* replace1673 -> 1973 in the cache */

   /* ('i2i8', ('iand', 'a@64', 255)) => ('u2u8', 'a') */
   /* search1674_0_0 -> 21 in the cache */
   /* search1674_0_1 -> 4400 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 21, 4400 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 1,
      { 4661 },
      -1,
   } },

   /* replace1674_0 -> 21 in the cache */
   /* replace1674 -> 4408 in the cache */

   /* ('u2u16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a */
   /* search1675_0_0 -> 11 in the cache */
   /* search1675_0 -> 1957 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 1957 },
      6,
   } },

   /* replace1675 -> 11 in the cache */

   /* ('extract_u8(only_lower_8_bits_used)', 'a@16', 0) => a */
   /* search1676_0 -> 11 in the cache */
   /* search1676_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 11, 818 },
      6,
   } },

   /* replace1676 -> 11 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1677_0 -> 506 in the cache */
   /* search1677_1_0_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4665 },
      -1,
   } },
   /* search1677_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4666, 473 },
      6,
   } },

   /* replace1677_0 -> 506 in the cache */
   /* replace1677_1 -> 511 in the cache */
   /* replace1677_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 511, 473 },
      -1,
   } },

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1678_0 -> 506 in the cache */
   /* search1678_1_0_0 -> 511 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 511 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 4669 },
      -1,
   } },
   /* search1678_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4670, 473 },
      6,
   } },

   /* replace1678_0 -> 506 in the cache */
   /* replace1678_1 -> 511 in the cache */
   /* replace1678_2 -> 473 in the cache */
   /* replace1678 -> 4668 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1679_0 -> 506 in the cache */
   /* search1679_1_0 -> 511 in the cache */
   /* search1679_1_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 511, 818 },
      -1,
   } },
   /* search1679_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4672, 473 },
      6,
   } },

   /* replace1679_0 -> 506 in the cache */
   /* replace1679_1 -> 511 in the cache */
   /* replace1679_2 -> 473 in the cache */
   /* replace1679 -> 4668 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1680_0_0_0 -> 11 in the cache */
   /* search1680_0_0 -> 1957 in the cache */
   /* search1680_0 -> 1958 in the cache */
   /* search1680_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 1958, 247 },
      6,
   } },

   /* replace1680_0 -> 11 in the cache */
   /* replace1680_1 -> 247 in the cache */
   /* replace1680 -> 3729 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1681_0_0_0 -> 11 in the cache */
   /* search1681_0_0 -> 3985 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 3985 },
      -1,
   } },
   /* search1681_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4675, 247 },
      6,
   } },

   /* replace1681_0 -> 11 in the cache */
   /* replace1681_1 -> 247 in the cache */
   /* replace1681 -> 3729 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1682_0_0 -> 11 in the cache */
   /* search1682_0_1 -> 818 in the cache */
   /* search1682_0 -> 3145 in the cache */
   /* search1682_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3145, 247 },
      6,
   } },

   /* replace1682_0 -> 11 in the cache */
   /* replace1682_1 -> 247 in the cache */
   /* replace1682 -> 3729 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1683_0_0_0 -> 11 in the cache */
   /* search1683_0_0 -> 1957 in the cache */
   /* search1683_0 -> 1958 in the cache */
   /* search1683_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 1958, 247 },
      6,
   } },

   /* replace1683_0 -> 11 in the cache */
   /* replace1683_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 11, 247 },
      -1,
   } },

   /* ('imul(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1684_0_0_0 -> 11 in the cache */
   /* search1684_0_0 -> 3985 in the cache */
   /* search1684_0 -> 4675 in the cache */
   /* search1684_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4675, 247 },
      6,
   } },

   /* replace1684_0 -> 11 in the cache */
   /* replace1684_1 -> 247 in the cache */
   /* replace1684 -> 4679 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('imul', 'a', 'b') */
   /* search1685_0_0 -> 11 in the cache */
   /* search1685_0_1 -> 818 in the cache */
   /* search1685_0 -> 3145 in the cache */
   /* search1685_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3145, 247 },
      6,
   } },

   /* replace1685_0 -> 11 in the cache */
   /* replace1685_1 -> 247 in the cache */
   /* replace1685 -> 4679 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1686_0_0_0 -> 11 in the cache */
   /* search1686_0_0 -> 1957 in the cache */
   /* search1686_0 -> 1958 in the cache */
   /* search1686_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 1958, 247 },
      6,
   } },

   /* replace1686_0 -> 11 in the cache */
   /* replace1686_1 -> 247 in the cache */
   /* replace1686 -> 1971 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1687_0_0_0 -> 11 in the cache */
   /* search1687_0_0 -> 3985 in the cache */
   /* search1687_0 -> 4675 in the cache */
   /* search1687_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4675, 247 },
      6,
   } },

   /* replace1687_0 -> 11 in the cache */
   /* replace1687_1 -> 247 in the cache */
   /* replace1687 -> 1971 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('iand', 'a', 'b') */
   /* search1688_0_0 -> 11 in the cache */
   /* search1688_0_1 -> 818 in the cache */
   /* search1688_0 -> 3145 in the cache */
   /* search1688_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3145, 247 },
      6,
   } },

   /* replace1688_0 -> 11 in the cache */
   /* replace1688_1 -> 247 in the cache */
   /* replace1688 -> 1971 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1689_0_0_0 -> 11 in the cache */
   /* search1689_0_0 -> 1957 in the cache */
   /* search1689_0 -> 1958 in the cache */
   /* search1689_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 1958, 247 },
      6,
   } },

   /* replace1689_0 -> 11 in the cache */
   /* replace1689_1 -> 247 in the cache */
   /* replace1689 -> 2013 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1690_0_0_0 -> 11 in the cache */
   /* search1690_0_0 -> 3985 in the cache */
   /* search1690_0 -> 4675 in the cache */
   /* search1690_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4675, 247 },
      6,
   } },

   /* replace1690_0 -> 11 in the cache */
   /* replace1690_1 -> 247 in the cache */
   /* replace1690 -> 2013 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ior', 'a', 'b') */
   /* search1691_0_0 -> 11 in the cache */
   /* search1691_0_1 -> 818 in the cache */
   /* search1691_0 -> 3145 in the cache */
   /* search1691_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3145, 247 },
      6,
   } },

   /* replace1691_0 -> 11 in the cache */
   /* replace1691_1 -> 247 in the cache */
   /* replace1691 -> 2013 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1692_0_0_0 -> 11 in the cache */
   /* search1692_0_0 -> 1957 in the cache */
   /* search1692_0 -> 1958 in the cache */
   /* search1692_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 1958, 247 },
      6,
   } },

   /* replace1692_0 -> 11 in the cache */
   /* replace1692_1 -> 247 in the cache */
   /* replace1692 -> 2037 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1693_0_0_0 -> 11 in the cache */
   /* search1693_0_0 -> 3985 in the cache */
   /* search1693_0 -> 4675 in the cache */
   /* search1693_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4675, 247 },
      6,
   } },

   /* replace1693_0 -> 11 in the cache */
   /* replace1693_1 -> 247 in the cache */
   /* replace1693 -> 2037 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@16', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1694_0_0 -> 11 in the cache */
   /* search1694_0_1 -> 818 in the cache */
   /* search1694_0 -> 3145 in the cache */
   /* search1694_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 3145, 247 },
      6,
   } },

   /* replace1694_0 -> 11 in the cache */
   /* replace1694_1 -> 247 in the cache */
   /* replace1694 -> 2037 in the cache */

   /* ('i2i16(only_lower_8_bits_used)', ('u2u8', 'a@16')) => a */
   /* search1695_0_0 -> 11 in the cache */
   /* search1695_0 -> 1957 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1957 },
      6,
   } },

   /* replace1695 -> 11 in the cache */

   /* ('extract_i8(only_lower_8_bits_used)', 'a@16', 0) => a */
   /* search1696_0 -> 11 in the cache */
   /* search1696_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 11, 818 },
      6,
   } },

   /* replace1696 -> 11 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('u2u8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1697_0 -> 506 in the cache */
   /* search1697_1_0_0 -> 511 in the cache */
   /* search1697_1_0 -> 4665 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4665 },
      -1,
   } },
   /* search1697_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4693, 473 },
      6,
   } },

   /* replace1697_0 -> 506 in the cache */
   /* replace1697_1 -> 511 in the cache */
   /* replace1697_2 -> 473 in the cache */
   /* replace1697 -> 4668 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i16', ('i2i8', 'a@16')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1698_0 -> 506 in the cache */
   /* search1698_1_0_0 -> 511 in the cache */
   /* search1698_1_0 -> 4669 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 4669 },
      -1,
   } },
   /* search1698_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4695, 473 },
      6,
   } },

   /* replace1698_0 -> 506 in the cache */
   /* replace1698_1 -> 511 in the cache */
   /* replace1698_2 -> 473 in the cache */
   /* replace1698 -> 4668 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@16', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1699_0 -> 506 in the cache */
   /* search1699_1_0 -> 511 in the cache */
   /* search1699_1_1 -> 818 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 511, 818 },
      -1,
   } },
   /* search1699_2 -> 473 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4697, 473 },
      6,
   } },

   /* replace1699_0 -> 506 in the cache */
   /* replace1699_1 -> 511 in the cache */
   /* replace1699_2 -> 473 in the cache */
   /* replace1699 -> 4668 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1700_0_0_0 -> 11 in the cache */
   /* search1700_0_0 -> 1957 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 1957 },
      -1,
   } },
   /* search1700_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4699, 247 },
      6,
   } },

   /* replace1700_0 -> 11 in the cache */
   /* replace1700_1 -> 247 in the cache */
   /* replace1700 -> 3729 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iadd', 'a', 'b') */
   /* search1701_0_0_0 -> 11 in the cache */
   /* search1701_0_0 -> 3985 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 3985 },
      -1,
   } },
   /* search1701_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4701, 247 },
      6,
   } },

   /* replace1701_0 -> 11 in the cache */
   /* replace1701_1 -> 247 in the cache */
   /* replace1701 -> 3729 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1702_0_0 -> 11 in the cache */
   /* search1702_0_1 -> 818 in the cache */
   /* search1702_0 -> 3166 in the cache */
   /* search1702_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 3166, 247 },
      6,
   } },

   /* replace1702_0 -> 11 in the cache */
   /* replace1702_1 -> 247 in the cache */
   /* replace1702 -> 3729 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1703_0_0_0 -> 11 in the cache */
   /* search1703_0_0 -> 1957 in the cache */
   /* search1703_0 -> 4699 in the cache */
   /* search1703_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4699, 247 },
      6,
   } },

   /* replace1703_0 -> 11 in the cache */
   /* replace1703_1 -> 247 in the cache */
   /* replace1703 -> 4679 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('imul', 'a', 'b') */
   /* search1704_0_0_0 -> 11 in the cache */
   /* search1704_0_0 -> 3985 in the cache */
   /* search1704_0 -> 4701 in the cache */
   /* search1704_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4701, 247 },
      6,
   } },

   /* replace1704_0 -> 11 in the cache */
   /* replace1704_1 -> 247 in the cache */
   /* replace1704 -> 4679 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('imul', 'a', 'b') */
   /* search1705_0_0 -> 11 in the cache */
   /* search1705_0_1 -> 818 in the cache */
   /* search1705_0 -> 3166 in the cache */
   /* search1705_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 3166, 247 },
      6,
   } },

   /* replace1705_0 -> 11 in the cache */
   /* replace1705_1 -> 247 in the cache */
   /* replace1705 -> 4679 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1706_0_0_0 -> 11 in the cache */
   /* search1706_0_0 -> 1957 in the cache */
   /* search1706_0 -> 4699 in the cache */
   /* search1706_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4699, 247 },
      6,
   } },

   /* replace1706_0 -> 11 in the cache */
   /* replace1706_1 -> 247 in the cache */
   /* replace1706 -> 1971 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('iand', 'a', 'b') */
   /* search1707_0_0_0 -> 11 in the cache */
   /* search1707_0_0 -> 3985 in the cache */
   /* search1707_0 -> 4701 in the cache */
   /* search1707_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4701, 247 },
      6,
   } },

   /* replace1707_0 -> 11 in the cache */
   /* replace1707_1 -> 247 in the cache */
   /* replace1707 -> 1971 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('iand', 'a', 'b') */
   /* search1708_0_0 -> 11 in the cache */
   /* search1708_0_1 -> 818 in the cache */
   /* search1708_0 -> 3166 in the cache */
   /* search1708_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 3166, 247 },
      6,
   } },

   /* replace1708_0 -> 11 in the cache */
   /* replace1708_1 -> 247 in the cache */
   /* replace1708 -> 1971 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1709_0_0_0 -> 11 in the cache */
   /* search1709_0_0 -> 1957 in the cache */
   /* search1709_0 -> 4699 in the cache */
   /* search1709_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4699, 247 },
      6,
   } },

   /* replace1709_0 -> 11 in the cache */
   /* replace1709_1 -> 247 in the cache */
   /* replace1709 -> 2013 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ior', 'a', 'b') */
   /* search1710_0_0_0 -> 11 in the cache */
   /* search1710_0_0 -> 3985 in the cache */
   /* search1710_0 -> 4701 in the cache */
   /* search1710_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4701, 247 },
      6,
   } },

   /* replace1710_0 -> 11 in the cache */
   /* replace1710_1 -> 247 in the cache */
   /* replace1710 -> 2013 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ior', 'a', 'b') */
   /* search1711_0_0 -> 11 in the cache */
   /* search1711_0_1 -> 818 in the cache */
   /* search1711_0 -> 3166 in the cache */
   /* search1711_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 3166, 247 },
      6,
   } },

   /* replace1711_0 -> 11 in the cache */
   /* replace1711_1 -> 247 in the cache */
   /* replace1711 -> 2013 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i16', ('u2u8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1712_0_0_0 -> 11 in the cache */
   /* search1712_0_0 -> 1957 in the cache */
   /* search1712_0 -> 4699 in the cache */
   /* search1712_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4699, 247 },
      6,
   } },

   /* replace1712_0 -> 11 in the cache */
   /* replace1712_1 -> 247 in the cache */
   /* replace1712 -> 2037 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i16', ('i2i8', 'a@16')), 'b') => ('ixor', 'a', 'b') */
   /* search1713_0_0_0 -> 11 in the cache */
   /* search1713_0_0 -> 3985 in the cache */
   /* search1713_0 -> 4701 in the cache */
   /* search1713_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4701, 247 },
      6,
   } },

   /* replace1713_0 -> 11 in the cache */
   /* replace1713_1 -> 247 in the cache */
   /* replace1713 -> 2037 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@16', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1714_0_0 -> 11 in the cache */
   /* search1714_0_1 -> 818 in the cache */
   /* search1714_0 -> 3166 in the cache */
   /* search1714_1 -> 247 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 3166, 247 },
      6,
   } },

   /* replace1714_0 -> 11 in the cache */
   /* replace1714_1 -> 247 in the cache */
   /* replace1714 -> 2037 in the cache */

   /* ('u2u32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a */
   /* search1715_0_0 -> 16 in the cache */
   /* search1715_0 -> 1973 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1973 },
      6,
   } },

   /* replace1715 -> 16 in the cache */

   /* ('extract_u8(only_lower_8_bits_used)', 'a@32', 0) => a */
   /* search1716_0 -> 16 in the cache */
   /* search1716_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 16, 124 },
      6,
   } },

   /* replace1716 -> 16 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1717_0 -> 506 in the cache */
   /* search1717_1_0_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4718 },
      -1,
   } },
   /* search1717_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4719, 544 },
      6,
   } },

   /* replace1717_0 -> 506 in the cache */
   /* replace1717_1 -> 364 in the cache */
   /* replace1717_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 364, 544 },
      -1,
   } },

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('u2u32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1718_0 -> 506 in the cache */
   /* search1718_1_0_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4722 },
      -1,
   } },
   /* search1718_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4723, 544 },
      6,
   } },

   /* replace1718_0 -> 506 in the cache */
   /* replace1718_1 -> 364 in the cache */
   /* replace1718_2 -> 544 in the cache */
   /* replace1718 -> 4721 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_u8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1719_0 -> 506 in the cache */
   /* search1719_1_0 -> 364 in the cache */
   /* search1719_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 364, 124 },
      -1,
   } },
   /* search1719_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4725, 544 },
      6,
   } },

   /* replace1719_0 -> 506 in the cache */
   /* replace1719_1 -> 364 in the cache */
   /* replace1719_2 -> 544 in the cache */
   /* replace1719 -> 4721 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1720_0_0_0 -> 16 in the cache */
   /* search1720_0_0 -> 1973 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 1973 },
      -1,
   } },
   /* search1720_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4727, 36 },
      6,
   } },

   /* replace1720_0 -> 16 in the cache */
   /* replace1720_1 -> 36 in the cache */
   /* replace1720 -> 3749 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1721_0_0_0 -> 16 in the cache */
   /* search1721_0_0 -> 3937 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3937 },
      -1,
   } },
   /* search1721_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4729, 36 },
      6,
   } },

   /* replace1721_0 -> 16 in the cache */
   /* replace1721_1 -> 36 in the cache */
   /* replace1721 -> 3749 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1722_0_0 -> 16 in the cache */
   /* search1722_0_1 -> 124 in the cache */
   /* search1722_0 -> 397 in the cache */
   /* search1722_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 397, 36 },
      6,
   } },

   /* replace1722_0 -> 16 in the cache */
   /* replace1722_1 -> 36 in the cache */
   /* replace1722 -> 3749 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1723_0_0_0 -> 16 in the cache */
   /* search1723_0_0 -> 1973 in the cache */
   /* search1723_0 -> 4727 in the cache */
   /* search1723_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4727, 36 },
      6,
   } },

   /* replace1723_0 -> 16 in the cache */
   /* replace1723_1 -> 36 in the cache */
   /* replace1723 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1724_0_0_0 -> 16 in the cache */
   /* search1724_0_0 -> 3937 in the cache */
   /* search1724_0 -> 4729 in the cache */
   /* search1724_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4729, 36 },
      6,
   } },

   /* replace1724_0 -> 16 in the cache */
   /* replace1724_1 -> 36 in the cache */
   /* replace1724 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1725_0_0 -> 16 in the cache */
   /* search1725_0_1 -> 124 in the cache */
   /* search1725_0 -> 397 in the cache */
   /* search1725_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 397, 36 },
      6,
   } },

   /* replace1725_0 -> 16 in the cache */
   /* replace1725_1 -> 36 in the cache */
   /* replace1725 -> 49 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1726_0_0_0 -> 16 in the cache */
   /* search1726_0_0 -> 1973 in the cache */
   /* search1726_0 -> 4727 in the cache */
   /* search1726_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4727, 36 },
      6,
   } },

   /* replace1726_0 -> 16 in the cache */
   /* replace1726_1 -> 36 in the cache */
   /* replace1726 -> 1979 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1727_0_0_0 -> 16 in the cache */
   /* search1727_0_0 -> 3937 in the cache */
   /* search1727_0 -> 4729 in the cache */
   /* search1727_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4729, 36 },
      6,
   } },

   /* replace1727_0 -> 16 in the cache */
   /* replace1727_1 -> 36 in the cache */
   /* replace1727 -> 1979 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1728_0_0 -> 16 in the cache */
   /* search1728_0_1 -> 124 in the cache */
   /* search1728_0 -> 397 in the cache */
   /* search1728_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 397, 36 },
      6,
   } },

   /* replace1728_0 -> 16 in the cache */
   /* replace1728_1 -> 36 in the cache */
   /* replace1728 -> 1979 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1729_0_0_0 -> 16 in the cache */
   /* search1729_0_0 -> 1973 in the cache */
   /* search1729_0 -> 4727 in the cache */
   /* search1729_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4727, 36 },
      6,
   } },

   /* replace1729_0 -> 16 in the cache */
   /* replace1729_1 -> 36 in the cache */
   /* replace1729 -> 2017 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1730_0_0_0 -> 16 in the cache */
   /* search1730_0_0 -> 3937 in the cache */
   /* search1730_0 -> 4729 in the cache */
   /* search1730_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4729, 36 },
      6,
   } },

   /* replace1730_0 -> 16 in the cache */
   /* replace1730_1 -> 36 in the cache */
   /* replace1730 -> 2017 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1731_0_0 -> 16 in the cache */
   /* search1731_0_1 -> 124 in the cache */
   /* search1731_0 -> 397 in the cache */
   /* search1731_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 397, 36 },
      6,
   } },

   /* replace1731_0 -> 16 in the cache */
   /* replace1731_1 -> 36 in the cache */
   /* replace1731 -> 2017 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1732_0_0_0 -> 16 in the cache */
   /* search1732_0_0 -> 1973 in the cache */
   /* search1732_0 -> 4727 in the cache */
   /* search1732_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4727, 36 },
      6,
   } },

   /* replace1732_0 -> 16 in the cache */
   /* replace1732_1 -> 36 in the cache */
   /* replace1732 -> 2041 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('u2u32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1733_0_0_0 -> 16 in the cache */
   /* search1733_0_0 -> 3937 in the cache */
   /* search1733_0 -> 4729 in the cache */
   /* search1733_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4729, 36 },
      6,
   } },

   /* replace1733_0 -> 16 in the cache */
   /* replace1733_1 -> 36 in the cache */
   /* replace1733 -> 2041 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_u8', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1734_0_0 -> 16 in the cache */
   /* search1734_0_1 -> 124 in the cache */
   /* search1734_0 -> 397 in the cache */
   /* search1734_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 397, 36 },
      6,
   } },

   /* replace1734_0 -> 16 in the cache */
   /* replace1734_1 -> 36 in the cache */
   /* replace1734 -> 2041 in the cache */

   /* ('i2i32(only_lower_8_bits_used)', ('u2u8', 'a@32')) => a */
   /* search1735_0_0 -> 16 in the cache */
   /* search1735_0 -> 1973 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1973 },
      6,
   } },

   /* replace1735 -> 16 in the cache */

   /* ('extract_i8(only_lower_8_bits_used)', 'a@32', 0) => a */
   /* search1736_0 -> 16 in the cache */
   /* search1736_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 16, 124 },
      6,
   } },

   /* replace1736 -> 16 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('u2u8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1737_0 -> 506 in the cache */
   /* search1737_1_0_0 -> 364 in the cache */
   /* search1737_1_0 -> 4718 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4718 },
      -1,
   } },
   /* search1737_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4746, 544 },
      6,
   } },

   /* replace1737_0 -> 506 in the cache */
   /* replace1737_1 -> 364 in the cache */
   /* replace1737_2 -> 544 in the cache */
   /* replace1737 -> 4721 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('i2i32', ('i2i8', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1738_0 -> 506 in the cache */
   /* search1738_1_0_0 -> 364 in the cache */
   /* search1738_1_0 -> 4722 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4722 },
      -1,
   } },
   /* search1738_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4748, 544 },
      6,
   } },

   /* replace1738_0 -> 506 in the cache */
   /* replace1738_1 -> 364 in the cache */
   /* replace1738_2 -> 544 in the cache */
   /* replace1738 -> 4721 in the cache */

   /* ('bcsel(only_lower_8_bits_used)', 'c', ('extract_i8', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1739_0 -> 506 in the cache */
   /* search1739_1_0 -> 364 in the cache */
   /* search1739_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 364, 124 },
      -1,
   } },
   /* search1739_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4750, 544 },
      6,
   } },

   /* replace1739_0 -> 506 in the cache */
   /* replace1739_1 -> 364 in the cache */
   /* replace1739_2 -> 544 in the cache */
   /* replace1739 -> 4721 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1740_0_0_0 -> 16 in the cache */
   /* search1740_0_0 -> 1973 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 1973 },
      -1,
   } },
   /* search1740_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4752, 36 },
      6,
   } },

   /* replace1740_0 -> 16 in the cache */
   /* replace1740_1 -> 36 in the cache */
   /* replace1740 -> 3749 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1741_0_0_0 -> 16 in the cache */
   /* search1741_0_0 -> 3937 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3937 },
      -1,
   } },
   /* search1741_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4754, 36 },
      6,
   } },

   /* replace1741_0 -> 16 in the cache */
   /* replace1741_1 -> 36 in the cache */
   /* replace1741 -> 3749 in the cache */

   /* ('iadd(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1742_0_0 -> 16 in the cache */
   /* search1742_0_1 -> 124 in the cache */
   /* search1742_0 -> 378 in the cache */
   /* search1742_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 378, 36 },
      6,
   } },

   /* replace1742_0 -> 16 in the cache */
   /* replace1742_1 -> 36 in the cache */
   /* replace1742 -> 3749 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1743_0_0_0 -> 16 in the cache */
   /* search1743_0_0 -> 1973 in the cache */
   /* search1743_0 -> 4752 in the cache */
   /* search1743_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4752, 36 },
      6,
   } },

   /* replace1743_0 -> 16 in the cache */
   /* replace1743_1 -> 36 in the cache */
   /* replace1743 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1744_0_0_0 -> 16 in the cache */
   /* search1744_0_0 -> 3937 in the cache */
   /* search1744_0 -> 4754 in the cache */
   /* search1744_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4754, 36 },
      6,
   } },

   /* replace1744_0 -> 16 in the cache */
   /* replace1744_1 -> 36 in the cache */
   /* replace1744 -> 49 in the cache */

   /* ('imul(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1745_0_0 -> 16 in the cache */
   /* search1745_0_1 -> 124 in the cache */
   /* search1745_0 -> 378 in the cache */
   /* search1745_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 378, 36 },
      6,
   } },

   /* replace1745_0 -> 16 in the cache */
   /* replace1745_1 -> 36 in the cache */
   /* replace1745 -> 49 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1746_0_0_0 -> 16 in the cache */
   /* search1746_0_0 -> 1973 in the cache */
   /* search1746_0 -> 4752 in the cache */
   /* search1746_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4752, 36 },
      6,
   } },

   /* replace1746_0 -> 16 in the cache */
   /* replace1746_1 -> 36 in the cache */
   /* replace1746 -> 1979 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1747_0_0_0 -> 16 in the cache */
   /* search1747_0_0 -> 3937 in the cache */
   /* search1747_0 -> 4754 in the cache */
   /* search1747_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4754, 36 },
      6,
   } },

   /* replace1747_0 -> 16 in the cache */
   /* replace1747_1 -> 36 in the cache */
   /* replace1747 -> 1979 in the cache */

   /* ('iand(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1748_0_0 -> 16 in the cache */
   /* search1748_0_1 -> 124 in the cache */
   /* search1748_0 -> 378 in the cache */
   /* search1748_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 378, 36 },
      6,
   } },

   /* replace1748_0 -> 16 in the cache */
   /* replace1748_1 -> 36 in the cache */
   /* replace1748 -> 1979 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1749_0_0_0 -> 16 in the cache */
   /* search1749_0_0 -> 1973 in the cache */
   /* search1749_0 -> 4752 in the cache */
   /* search1749_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4752, 36 },
      6,
   } },

   /* replace1749_0 -> 16 in the cache */
   /* replace1749_1 -> 36 in the cache */
   /* replace1749 -> 2017 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1750_0_0_0 -> 16 in the cache */
   /* search1750_0_0 -> 3937 in the cache */
   /* search1750_0 -> 4754 in the cache */
   /* search1750_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4754, 36 },
      6,
   } },

   /* replace1750_0 -> 16 in the cache */
   /* replace1750_1 -> 36 in the cache */
   /* replace1750 -> 2017 in the cache */

   /* ('ior(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1751_0_0 -> 16 in the cache */
   /* search1751_0_1 -> 124 in the cache */
   /* search1751_0 -> 378 in the cache */
   /* search1751_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 378, 36 },
      6,
   } },

   /* replace1751_0 -> 16 in the cache */
   /* replace1751_1 -> 36 in the cache */
   /* replace1751 -> 2017 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i32', ('u2u8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1752_0_0_0 -> 16 in the cache */
   /* search1752_0_0 -> 1973 in the cache */
   /* search1752_0 -> 4752 in the cache */
   /* search1752_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4752, 36 },
      6,
   } },

   /* replace1752_0 -> 16 in the cache */
   /* replace1752_1 -> 36 in the cache */
   /* replace1752 -> 2041 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('i2i32', ('i2i8', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1753_0_0_0 -> 16 in the cache */
   /* search1753_0_0 -> 3937 in the cache */
   /* search1753_0 -> 4754 in the cache */
   /* search1753_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4754, 36 },
      6,
   } },

   /* replace1753_0 -> 16 in the cache */
   /* replace1753_1 -> 36 in the cache */
   /* replace1753 -> 2041 in the cache */

   /* ('ixor(only_lower_8_bits_used)', ('extract_i8', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1754_0_0 -> 16 in the cache */
   /* search1754_0_1 -> 124 in the cache */
   /* search1754_0 -> 378 in the cache */
   /* search1754_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 378, 36 },
      6,
   } },

   /* replace1754_0 -> 16 in the cache */
   /* replace1754_1 -> 36 in the cache */
   /* replace1754 -> 2041 in the cache */

   /* ('u2u32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a */
   /* search1755_0_0 -> 16 in the cache */
   /* search1755_0 -> 3037 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3037 },
      7,
   } },

   /* replace1755 -> 16 in the cache */

   /* ('extract_u16(only_lower_16_bits_used)', 'a@32', 0) => a */
   /* search1756_0 -> 16 in the cache */
   /* search1756_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 16, 124 },
      7,
   } },

   /* replace1756 -> 16 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1757_0 -> 506 in the cache */
   /* search1757_1_0_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4771 },
      -1,
   } },
   /* search1757_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4772, 544 },
      7,
   } },

   /* replace1757_0 -> 506 in the cache */
   /* replace1757_1 -> 364 in the cache */
   /* replace1757_2 -> 544 in the cache */
   /* replace1757 -> 4721 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('u2u32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1758_0 -> 506 in the cache */
   /* search1758_1_0_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 364 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 4774 },
      -1,
   } },
   /* search1758_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4775, 544 },
      7,
   } },

   /* replace1758_0 -> 506 in the cache */
   /* replace1758_1 -> 364 in the cache */
   /* replace1758_2 -> 544 in the cache */
   /* replace1758 -> 4721 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('extract_u16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1759_0 -> 506 in the cache */
   /* search1759_1_0 -> 364 in the cache */
   /* search1759_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 364, 124 },
      -1,
   } },
   /* search1759_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4777, 544 },
      7,
   } },

   /* replace1759_0 -> 506 in the cache */
   /* replace1759_1 -> 364 in the cache */
   /* replace1759_2 -> 544 in the cache */
   /* replace1759 -> 4721 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1760_0_0_0 -> 16 in the cache */
   /* search1760_0_0 -> 3037 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3037 },
      -1,
   } },
   /* search1760_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4779, 36 },
      7,
   } },

   /* replace1760_0 -> 16 in the cache */
   /* replace1760_1 -> 36 in the cache */
   /* replace1760 -> 3749 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1761_0_0_0 -> 16 in the cache */
   /* search1761_0_0 -> 3939 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 3939 },
      -1,
   } },
   /* search1761_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4781, 36 },
      7,
   } },

   /* replace1761_0 -> 16 in the cache */
   /* replace1761_1 -> 36 in the cache */
   /* replace1761 -> 3749 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1762_0_0 -> 16 in the cache */
   /* search1762_0_1 -> 124 in the cache */
   /* search1762_0 -> 429 in the cache */
   /* search1762_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 429, 36 },
      7,
   } },

   /* replace1762_0 -> 16 in the cache */
   /* replace1762_1 -> 36 in the cache */
   /* replace1762 -> 3749 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1763_0_0_0 -> 16 in the cache */
   /* search1763_0_0 -> 3037 in the cache */
   /* search1763_0 -> 4779 in the cache */
   /* search1763_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4779, 36 },
      7,
   } },

   /* replace1763_0 -> 16 in the cache */
   /* replace1763_1 -> 36 in the cache */
   /* replace1763 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1764_0_0_0 -> 16 in the cache */
   /* search1764_0_0 -> 3939 in the cache */
   /* search1764_0 -> 4781 in the cache */
   /* search1764_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4781, 36 },
      7,
   } },

   /* replace1764_0 -> 16 in the cache */
   /* replace1764_1 -> 36 in the cache */
   /* replace1764 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1765_0_0 -> 16 in the cache */
   /* search1765_0_1 -> 124 in the cache */
   /* search1765_0 -> 429 in the cache */
   /* search1765_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 429, 36 },
      7,
   } },

   /* replace1765_0 -> 16 in the cache */
   /* replace1765_1 -> 36 in the cache */
   /* replace1765 -> 49 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1766_0_0_0 -> 16 in the cache */
   /* search1766_0_0 -> 3037 in the cache */
   /* search1766_0 -> 4779 in the cache */
   /* search1766_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4779, 36 },
      7,
   } },

   /* replace1766_0 -> 16 in the cache */
   /* replace1766_1 -> 36 in the cache */
   /* replace1766 -> 1979 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1767_0_0_0 -> 16 in the cache */
   /* search1767_0_0 -> 3939 in the cache */
   /* search1767_0 -> 4781 in the cache */
   /* search1767_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4781, 36 },
      7,
   } },

   /* replace1767_0 -> 16 in the cache */
   /* replace1767_1 -> 36 in the cache */
   /* replace1767 -> 1979 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1768_0_0 -> 16 in the cache */
   /* search1768_0_1 -> 124 in the cache */
   /* search1768_0 -> 429 in the cache */
   /* search1768_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 429, 36 },
      7,
   } },

   /* replace1768_0 -> 16 in the cache */
   /* replace1768_1 -> 36 in the cache */
   /* replace1768 -> 1979 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1769_0_0_0 -> 16 in the cache */
   /* search1769_0_0 -> 3037 in the cache */
   /* search1769_0 -> 4779 in the cache */
   /* search1769_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4779, 36 },
      7,
   } },

   /* replace1769_0 -> 16 in the cache */
   /* replace1769_1 -> 36 in the cache */
   /* replace1769 -> 2017 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1770_0_0_0 -> 16 in the cache */
   /* search1770_0_0 -> 3939 in the cache */
   /* search1770_0 -> 4781 in the cache */
   /* search1770_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4781, 36 },
      7,
   } },

   /* replace1770_0 -> 16 in the cache */
   /* replace1770_1 -> 36 in the cache */
   /* replace1770 -> 2017 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1771_0_0 -> 16 in the cache */
   /* search1771_0_1 -> 124 in the cache */
   /* search1771_0 -> 429 in the cache */
   /* search1771_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 429, 36 },
      7,
   } },

   /* replace1771_0 -> 16 in the cache */
   /* replace1771_1 -> 36 in the cache */
   /* replace1771 -> 2017 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('u2u32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1772_0_0_0 -> 16 in the cache */
   /* search1772_0_0 -> 3037 in the cache */
   /* search1772_0 -> 4779 in the cache */
   /* search1772_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4779, 36 },
      7,
   } },

   /* replace1772_0 -> 16 in the cache */
   /* replace1772_1 -> 36 in the cache */
   /* replace1772 -> 2041 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('u2u32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1773_0_0_0 -> 16 in the cache */
   /* search1773_0_0 -> 3939 in the cache */
   /* search1773_0 -> 4781 in the cache */
   /* search1773_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4781, 36 },
      7,
   } },

   /* replace1773_0 -> 16 in the cache */
   /* replace1773_1 -> 36 in the cache */
   /* replace1773 -> 2041 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('extract_u16', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1774_0_0 -> 16 in the cache */
   /* search1774_0_1 -> 124 in the cache */
   /* search1774_0 -> 429 in the cache */
   /* search1774_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 429, 36 },
      7,
   } },

   /* replace1774_0 -> 16 in the cache */
   /* replace1774_1 -> 36 in the cache */
   /* replace1774 -> 2041 in the cache */

   /* ('i2i32(only_lower_16_bits_used)', ('u2u16', 'a@32')) => a */
   /* search1775_0_0 -> 16 in the cache */
   /* search1775_0 -> 3037 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3037 },
      7,
   } },

   /* replace1775 -> 16 in the cache */

   /* ('extract_i16(only_lower_16_bits_used)', 'a@32', 0) => a */
   /* search1776_0 -> 16 in the cache */
   /* search1776_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 16, 124 },
      7,
   } },

   /* replace1776 -> 16 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('u2u16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1777_0 -> 506 in the cache */
   /* search1777_1_0_0 -> 364 in the cache */
   /* search1777_1_0 -> 4771 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4771 },
      -1,
   } },
   /* search1777_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4798, 544 },
      7,
   } },

   /* replace1777_0 -> 506 in the cache */
   /* replace1777_1 -> 364 in the cache */
   /* replace1777_2 -> 544 in the cache */
   /* replace1777 -> 4721 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('i2i32', ('i2i16', 'a@32')), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1778_0 -> 506 in the cache */
   /* search1778_1_0_0 -> 364 in the cache */
   /* search1778_1_0 -> 4774 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 4774 },
      -1,
   } },
   /* search1778_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4800, 544 },
      7,
   } },

   /* replace1778_0 -> 506 in the cache */
   /* replace1778_1 -> 364 in the cache */
   /* replace1778_2 -> 544 in the cache */
   /* replace1778 -> 4721 in the cache */

   /* ('bcsel(only_lower_16_bits_used)', 'c', ('extract_i16', 'a@32', 0), 'b') => ('bcsel', 'c', 'a', 'b') */
   /* search1779_0 -> 506 in the cache */
   /* search1779_1_0 -> 364 in the cache */
   /* search1779_1_1 -> 124 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 364, 124 },
      -1,
   } },
   /* search1779_2 -> 544 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 4802, 544 },
      7,
   } },

   /* replace1779_0 -> 506 in the cache */
   /* replace1779_1 -> 364 in the cache */
   /* replace1779_2 -> 544 in the cache */
   /* replace1779 -> 4721 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1780_0_0_0 -> 16 in the cache */
   /* search1780_0_0 -> 3037 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3037 },
      -1,
   } },
   /* search1780_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4804, 36 },
      7,
   } },

   /* replace1780_0 -> 16 in the cache */
   /* replace1780_1 -> 36 in the cache */
   /* replace1780 -> 3749 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iadd', 'a', 'b') */
   /* search1781_0_0_0 -> 16 in the cache */
   /* search1781_0_0 -> 3939 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 3939 },
      -1,
   } },
   /* search1781_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 4806, 36 },
      7,
   } },

   /* replace1781_0 -> 16 in the cache */
   /* replace1781_1 -> 36 in the cache */
   /* replace1781 -> 3749 in the cache */

   /* ('iadd(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iadd', 'a', 'b') */
   /* search1782_0_0 -> 16 in the cache */
   /* search1782_0_1 -> 124 in the cache */
   /* search1782_0 -> 421 in the cache */
   /* search1782_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 421, 36 },
      7,
   } },

   /* replace1782_0 -> 16 in the cache */
   /* replace1782_1 -> 36 in the cache */
   /* replace1782 -> 3749 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1783_0_0_0 -> 16 in the cache */
   /* search1783_0_0 -> 3037 in the cache */
   /* search1783_0 -> 4804 in the cache */
   /* search1783_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4804, 36 },
      7,
   } },

   /* replace1783_0 -> 16 in the cache */
   /* replace1783_1 -> 36 in the cache */
   /* replace1783 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('imul', 'a', 'b') */
   /* search1784_0_0_0 -> 16 in the cache */
   /* search1784_0_0 -> 3939 in the cache */
   /* search1784_0 -> 4806 in the cache */
   /* search1784_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 4806, 36 },
      7,
   } },

   /* replace1784_0 -> 16 in the cache */
   /* replace1784_1 -> 36 in the cache */
   /* replace1784 -> 49 in the cache */

   /* ('imul(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('imul', 'a', 'b') */
   /* search1785_0_0 -> 16 in the cache */
   /* search1785_0_1 -> 124 in the cache */
   /* search1785_0 -> 421 in the cache */
   /* search1785_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 421, 36 },
      7,
   } },

   /* replace1785_0 -> 16 in the cache */
   /* replace1785_1 -> 36 in the cache */
   /* replace1785 -> 49 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1786_0_0_0 -> 16 in the cache */
   /* search1786_0_0 -> 3037 in the cache */
   /* search1786_0 -> 4804 in the cache */
   /* search1786_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4804, 36 },
      7,
   } },

   /* replace1786_0 -> 16 in the cache */
   /* replace1786_1 -> 36 in the cache */
   /* replace1786 -> 1979 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('iand', 'a', 'b') */
   /* search1787_0_0_0 -> 16 in the cache */
   /* search1787_0_0 -> 3939 in the cache */
   /* search1787_0 -> 4806 in the cache */
   /* search1787_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 4806, 36 },
      7,
   } },

   /* replace1787_0 -> 16 in the cache */
   /* replace1787_1 -> 36 in the cache */
   /* replace1787 -> 1979 in the cache */

   /* ('iand(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('iand', 'a', 'b') */
   /* search1788_0_0 -> 16 in the cache */
   /* search1788_0_1 -> 124 in the cache */
   /* search1788_0 -> 421 in the cache */
   /* search1788_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 421, 36 },
      7,
   } },

   /* replace1788_0 -> 16 in the cache */
   /* replace1788_1 -> 36 in the cache */
   /* replace1788 -> 1979 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1789_0_0_0 -> 16 in the cache */
   /* search1789_0_0 -> 3037 in the cache */
   /* search1789_0 -> 4804 in the cache */
   /* search1789_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4804, 36 },
      7,
   } },

   /* replace1789_0 -> 16 in the cache */
   /* replace1789_1 -> 36 in the cache */
   /* replace1789 -> 2017 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ior', 'a', 'b') */
   /* search1790_0_0_0 -> 16 in the cache */
   /* search1790_0_0 -> 3939 in the cache */
   /* search1790_0 -> 4806 in the cache */
   /* search1790_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 4806, 36 },
      7,
   } },

   /* replace1790_0 -> 16 in the cache */
   /* replace1790_1 -> 36 in the cache */
   /* replace1790 -> 2017 in the cache */

   /* ('ior(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ior', 'a', 'b') */
   /* search1791_0_0 -> 16 in the cache */
   /* search1791_0_1 -> 124 in the cache */
   /* search1791_0 -> 421 in the cache */
   /* search1791_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 421, 36 },
      7,
   } },

   /* replace1791_0 -> 16 in the cache */
   /* replace1791_1 -> 36 in the cache */
   /* replace1791 -> 2017 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('i2i32', ('u2u16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1792_0_0_0 -> 16 in the cache */
   /* search1792_0_0 -> 3037 in the cache */
   /* search1792_0 -> 4804 in the cache */
   /* search1792_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4804, 36 },
      7,
   } },

   /* replace1792_0 -> 16 in the cache */
   /* replace1792_1 -> 36 in the cache */
   /* replace1792 -> 2041 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('i2i32', ('i2i16', 'a@32')), 'b') => ('ixor', 'a', 'b') */
   /* search1793_0_0_0 -> 16 in the cache */
   /* search1793_0_0 -> 3939 in the cache */
   /* search1793_0 -> 4806 in the cache */
   /* search1793_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 4806, 36 },
      7,
   } },

   /* replace1793_0 -> 16 in the cache */
   /* replace1793_1 -> 36 in the cache */
   /* replace1793 -> 2041 in the cache */

   /* ('ixor(only_lower_16_bits_used)', ('extract_i16', 'a@32', 0), 'b') => ('ixor', 'a', 'b') */
   /* search1794_0_0 -> 16 in the cache */
   /* search1794_0_1 -> 124 in the cache */
   /* search1794_0 -> 421 in the cache */
   /* search1794_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 421, 36 },
      7,
   } },

   /* replace1794_0 -> 16 in the cache */
   /* replace1794_1 -> 36 in the cache */
   /* replace1794 -> 2041 in the cache */

   /* ('ldexp@16', 'x', 'exp') => ('fmul', ('fmul', 'x', ('i2i16', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -30), 30), 1), 15), 10))), ('i2i16', ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -30), 30), ('ishr', ('imin', ('imax', 'exp', -30), 30), 1)), 15), 10))) */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* exp */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 4821, 4822 },
      -1,
   } },

   /* replace1795_0_0 -> 4821 in the cache */
   /* replace1795_0_1_0_0_0_0_0_0 -> 4822 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x1e /* -30 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 4822, 4824 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1e /* 30 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 4825, 4826 },
      -1,
   } },
   /* replace1795_0_1_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 4827, 381 },
      -1,
   } },
   /* replace1795_0_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 4828, 14 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xa /* 10 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 4829, 4830 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 3,
      { 4831 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 4821, 4832 },
      -1,
   } },
   /* replace1795_1_0_0_0_0_0_0 -> 4822 in the cache */
   /* replace1795_1_0_0_0_0_0_1 -> 4824 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 4822, 4824 },
      -1,
   } },
   /* replace1795_1_0_0_0_0_1 -> 4826 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 4834, 4826 },
      -1,
   } },
   /* replace1795_1_0_0_0_1_0_0_0 -> 4822 in the cache */
   /* replace1795_1_0_0_0_1_0_0_1 -> 4824 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 4822, 4824 },
      -1,
   } },
   /* replace1795_1_0_0_0_1_0_1 -> 4826 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 4836, 4826 },
      -1,
   } },
   /* replace1795_1_0_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 4837, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 4835, 4838 },
      -1,
   } },
   /* replace1795_1_0_0_1 -> 14 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 4839, 14 },
      -1,
   } },
   /* replace1795_1_0_1 -> 4830 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 4840, 4830 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 5,
      { 4841 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 4833, 4842 },
      -1,
   } },

   /* ('ldexp@32', 'x', 'exp') => ('fmul', ('fmul', 'x', ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -254), 254), 1), 127), 23)), ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -254), 254), ('ishr', ('imin', ('imax', 'exp', -254), 254), 1)), 127), 23)) */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1796_1 -> 4822 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 4844, 4822 },
      -1,
   } },

   /* replace1796_0_0 -> 4844 in the cache */
   /* replace1796_0_1_0_0_0_0_0 -> 4822 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0xfe /* -254 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 4822, 4846 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xfe /* 254 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 4847, 4848 },
      -1,
   } },
   /* replace1796_0_1_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 4849, 381 },
      -1,
   } },
   /* replace1796_0_1_0_1 -> 4023 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 4850, 4023 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x17 /* 23 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 4851, 4852 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 4844, 4853 },
      -1,
   } },
   /* replace1796_1_0_0_0_0_0 -> 4822 in the cache */
   /* replace1796_1_0_0_0_0_1 -> 4846 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 4822, 4846 },
      -1,
   } },
   /* replace1796_1_0_0_0_1 -> 4848 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 4855, 4848 },
      -1,
   } },
   /* replace1796_1_0_0_1_0_0_0 -> 4822 in the cache */
   /* replace1796_1_0_0_1_0_0_1 -> 4846 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 4822, 4846 },
      -1,
   } },
   /* replace1796_1_0_0_1_0_1 -> 4848 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 4857, 4848 },
      -1,
   } },
   /* replace1796_1_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 4858, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 4856, 4859 },
      -1,
   } },
   /* replace1796_1_0_1 -> 4023 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 4860, 4023 },
      -1,
   } },
   /* replace1796_1_1 -> 4852 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 4861, 4852 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 4854, 4862 },
      -1,
   } },

   /* ('ldexp@64', 'x', 'exp') => ('fmul', ('fmul', 'x', ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1), 1023), 20))), ('pack_64_2x32_split', 0, ('ishl', ('iadd', ('isub', ('imin', ('imax', 'exp', -2046), 2046), ('ishr', ('imin', ('imax', 'exp', -2046), 2046), 1)), 1023), 20))) */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* x */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1797_1 -> 4822 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ldexp,
      -1, 0,
      { 4864, 4822 },
      -1,
   } },

   /* replace1797_0_0 -> 4864 in the cache */
   /* replace1797_0_1_0 -> 124 in the cache */
   /* replace1797_0_1_1_0_0_0_0_0 -> 4822 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x7fe /* -2046 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      4, 1,
      { 4822, 4866 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x7fe /* 2046 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      3, 2,
      { 4867, 4868 },
      -1,
   } },
   /* replace1797_0_1_1_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 4869, 381 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3ff /* 1023 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      2, 3,
      { 4870, 4871 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x14 /* 20 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 3,
      { 4872, 4873 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 3,
      { 124, 4874 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 4,
      { 4864, 4875 },
      -1,
   } },
   /* replace1797_1_0 -> 124 in the cache */
   /* replace1797_1_1_0_0_0_0_0 -> 4822 in the cache */
   /* replace1797_1_1_0_0_0_0_1 -> 4866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      7, 1,
      { 4822, 4866 },
      -1,
   } },
   /* replace1797_1_1_0_0_0_1 -> 4868 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      6, 2,
      { 4877, 4868 },
      -1,
   } },
   /* replace1797_1_1_0_0_1_0_0_0 -> 4822 in the cache */
   /* replace1797_1_1_0_0_1_0_0_1 -> 4866 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imax,
      9, 1,
      { 4822, 4866 },
      -1,
   } },
   /* replace1797_1_1_0_0_1_0_1 -> 4868 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imin,
      8, 2,
      { 4879, 4868 },
      -1,
   } },
   /* replace1797_1_1_0_0_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 2,
      { 4880, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 4,
      { 4878, 4881 },
      -1,
   } },
   /* replace1797_1_1_0_1 -> 4871 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      5, 5,
      { 4882, 4871 },
      -1,
   } },
   /* replace1797_1_1_1 -> 4873 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 4883, 4873 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_pack_64_2x32_split,
      -1, 5,
      { 124, 4884 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 10,
      { 4876, 4885 },
      -1,
   } },

   /* ('ior(many-comm-expr)', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 1431655765), 1), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 858993459), 2), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 252645135), 4), ('ushr', ('iand', ('ior', ('ishl', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 16711935), 8), ('ushr', ('iand', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 4278255360), 8)), 4042322160), 4)), 3435973836), 2)), 2863311530), 1)) => ('bitfield_reverse', 'x') */
   /* search1798_0_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 4844, 815 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 4844, 815 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      8, 1,
      { 4887, 4888 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00ff /* 16711935 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      7, 2,
      { 4889, 4890 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_0_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4891, 793 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_0_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      10, 1,
      { 4887, 4888 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00ff00 /* 4278255360 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      9, 2,
      { 4893, 4894 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_0_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4895, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      6, 5,
      { 4892, 4896 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf0f0f0f /* 252645135 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      5, 6,
      { 4897, 4898 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 4899, 2146 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      14, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      13, 2,
      { 4901, 4890 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4902, 793 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      16, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      15, 2,
      { 4904, 4894 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4905, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      12, 5,
      { 4903, 4906 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xf0f0f0f0 /* 4042322160 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      11, 6,
      { 4907, 4908 },
      -1,
   } },
   /* search1798_0_0_0_0_0_0_1_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 4909, 2146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      4, 13,
      { 4900, 4910 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x33333333 /* 858993459 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 14,
      { 4911, 4912 },
      -1,
   } },
   /* search1798_0_0_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 14,
      { 4913, 386 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      22, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      21, 2,
      { 4915, 4890 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4916, 793 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      24, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      23, 2,
      { 4918, 4894 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4919, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      20, 5,
      { 4917, 4920 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_0_1 -> 4898 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      19, 6,
      { 4921, 4898 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 4922, 2146 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      28, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      27, 2,
      { 4924, 4890 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4925, 793 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      30, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      29, 2,
      { 4927, 4894 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4928, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      26, 5,
      { 4926, 4929 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_0_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      25, 6,
      { 4930, 4908 },
      -1,
   } },
   /* search1798_0_0_0_1_0_0_1_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 4931, 2146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      18, 13,
      { 4923, 4932 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xcccccccc /* 3435973836 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      17, 14,
      { 4933, 4934 },
      -1,
   } },
   /* search1798_0_0_0_1_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 14,
      { 4935, 386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 29,
      { 4914, 4936 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x55555555 /* 1431655765 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 30,
      { 4937, 4938 },
      -1,
   } },
   /* search1798_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 30,
      { 4939, 381 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      38, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      37, 2,
      { 4941, 4890 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4942, 793 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      40, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      39, 2,
      { 4944, 4894 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4945, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      36, 5,
      { 4943, 4946 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_0_1 -> 4898 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      35, 6,
      { 4947, 4898 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 4948, 2146 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      44, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      43, 2,
      { 4950, 4890 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4951, 793 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      46, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      45, 2,
      { 4953, 4894 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4954, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      42, 5,
      { 4952, 4955 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_0_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      41, 6,
      { 4956, 4908 },
      -1,
   } },
   /* search1798_1_0_0_0_0_0_1_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 4957, 2146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      34, 13,
      { 4949, 4958 },
      -1,
   } },
   /* search1798_1_0_0_0_0_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      33, 14,
      { 4959, 4912 },
      -1,
   } },
   /* search1798_1_0_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 14,
      { 4960, 386 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      52, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      51, 2,
      { 4962, 4890 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4963, 793 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      54, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      53, 2,
      { 4965, 4894 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4966, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      50, 5,
      { 4964, 4967 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_0_1 -> 4898 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      49, 6,
      { 4968, 4898 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 6,
      { 4969, 2146 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      58, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_0_0_0_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      57, 2,
      { 4971, 4890 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 2,
      { 4972, 793 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      60, 1,
      { 4887, 4888 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_0_1_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      59, 2,
      { 4974, 4894 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_0_1_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 2,
      { 4975, 793 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      56, 5,
      { 4973, 4976 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_0_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      55, 6,
      { 4977, 4908 },
      -1,
   } },
   /* search1798_1_0_0_1_0_0_1_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 6,
      { 4978, 2146 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      48, 13,
      { 4970, 4979 },
      -1,
   } },
   /* search1798_1_0_0_1_0_1 -> 4934 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      47, 14,
      { 4980, 4934 },
      -1,
   } },
   /* search1798_1_0_0_1_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 14,
      { 4981, 386 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      32, 29,
      { 4961, 4982 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xaaaaaaaa /* 2863311530 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      31, 30,
      { 4983, 4984 },
      -1,
   } },
   /* search1798_1_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 30,
      { 4985, 381 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 61,
      { 4940, 4986 },
      -1,
   } },

   /* replace1798_0 -> 4844 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bitfield_reverse,
      -1, 0,
      { 4844 },
      -1,
   } },

   /* ('ior(many-comm-expr)', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 4278255360), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 4042322160), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 3435973836), ('iand', ('ushr', ('ior', ('iand', ('ishl', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 2863311530), ('iand', ('ushr', ('ior', ('ishl', 'x@32', 16), ('ushr', 'x@32', 16)), 1), 1431655765)), 2), 858993459)), 4), 252645135)), 8), 16711935)) => ('bitfield_reverse', 'x') */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_0 -> 4889 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4889, 381 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_0_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      7, 2,
      { 4989, 4984 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_0 -> 4893 in the cache */
   /* search1799_0_0_0_0_0_0_0_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4893, 381 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_0_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      9, 2,
      { 4991, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      6, 5,
      { 4990, 4992 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 4993, 386 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_0_1 -> 4934 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      5, 6,
      { 4994, 4934 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_0 -> 4901 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4901, 381 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_1_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      13, 2,
      { 4996, 4984 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_0 -> 4904 in the cache */
   /* search1799_0_0_0_0_0_0_1_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4904, 381 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_1_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      15, 2,
      { 4998, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      12, 5,
      { 4997, 4999 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_1_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5000, 386 },
      -1,
   } },
   /* search1799_0_0_0_0_0_0_1_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      11, 6,
      { 5001, 4912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      4, 13,
      { 4995, 5002 },
      -1,
   } },
   /* search1799_0_0_0_0_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 13,
      { 5003, 2146 },
      -1,
   } },
   /* search1799_0_0_0_0_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      3, 14,
      { 5004, 4908 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_0 -> 4915 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4915, 381 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_0_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      21, 2,
      { 5006, 4984 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_0 -> 4918 in the cache */
   /* search1799_0_0_0_1_0_0_0_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4918, 381 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_0_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      23, 2,
      { 5008, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      20, 5,
      { 5007, 5009 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5010, 386 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_0_1 -> 4934 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      19, 6,
      { 5011, 4934 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_0 -> 4924 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4924, 381 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_1_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      27, 2,
      { 5013, 4984 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_0 -> 4927 in the cache */
   /* search1799_0_0_0_1_0_0_1_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4927, 381 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_1_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      29, 2,
      { 5015, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      26, 5,
      { 5014, 5016 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_1_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5017, 386 },
      -1,
   } },
   /* search1799_0_0_0_1_0_0_1_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      25, 6,
      { 5018, 4912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      18, 13,
      { 5012, 5019 },
      -1,
   } },
   /* search1799_0_0_0_1_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 13,
      { 5020, 2146 },
      -1,
   } },
   /* search1799_0_0_0_1_1 -> 4898 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      17, 14,
      { 5021, 4898 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      2, 29,
      { 5005, 5022 },
      -1,
   } },
   /* search1799_0_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 29,
      { 5023, 793 },
      -1,
   } },
   /* search1799_0_1 -> 4894 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      1, 30,
      { 5024, 4894 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_0 -> 4941 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4941, 381 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_0_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      37, 2,
      { 5026, 4984 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_0 -> 4944 in the cache */
   /* search1799_1_0_0_0_0_0_0_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4944, 381 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_0_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      39, 2,
      { 5028, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      36, 5,
      { 5027, 5029 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5030, 386 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_0_1 -> 4934 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      35, 6,
      { 5031, 4934 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_0 -> 4950 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4950, 381 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_1_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      43, 2,
      { 5033, 4984 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_0 -> 4953 in the cache */
   /* search1799_1_0_0_0_0_0_1_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4953, 381 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_1_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      45, 2,
      { 5035, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      42, 5,
      { 5034, 5036 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_1_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5037, 386 },
      -1,
   } },
   /* search1799_1_0_0_0_0_0_1_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      41, 6,
      { 5038, 4912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      34, 13,
      { 5032, 5039 },
      -1,
   } },
   /* search1799_1_0_0_0_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 13,
      { 5040, 2146 },
      -1,
   } },
   /* search1799_1_0_0_0_1 -> 4908 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      33, 14,
      { 5041, 4908 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_0 -> 4962 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4962, 381 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_0_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      51, 2,
      { 5043, 4984 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_0 -> 4965 in the cache */
   /* search1799_1_0_0_1_0_0_0_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4965, 381 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_0_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      53, 2,
      { 5045, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      50, 5,
      { 5044, 5046 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_0_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 5,
      { 5047, 386 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_0_1 -> 4934 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      49, 6,
      { 5048, 4934 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_0 -> 4971 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_0_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 4971, 381 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_1_0_0_0_1 -> 4984 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      57, 2,
      { 5050, 4984 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0_0_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0_0_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0_0 -> 4887 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0_1_0 -> 4844 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0_1_1 -> 815 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0_1 -> 4888 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_0 -> 4974 in the cache */
   /* search1799_1_0_0_1_0_0_1_0_0_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 1,
      { 4974, 381 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_1_0_0_1_1 -> 4938 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      59, 2,
      { 5052, 4938 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      56, 5,
      { 5051, 5053 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_1_0_1 -> 386 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 5,
      { 5054, 386 },
      -1,
   } },
   /* search1799_1_0_0_1_0_0_1_1 -> 4912 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      55, 6,
      { 5055, 4912 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      48, 13,
      { 5049, 5056 },
      -1,
   } },
   /* search1799_1_0_0_1_0_1 -> 2146 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 13,
      { 5057, 2146 },
      -1,
   } },
   /* search1799_1_0_0_1_1 -> 4898 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      47, 14,
      { 5058, 4898 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      32, 29,
      { 5042, 5059 },
      -1,
   } },
   /* search1799_1_0_1 -> 793 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 29,
      { 5060, 793 },
      -1,
   } },
   /* search1799_1_1 -> 4890 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      31, 30,
      { 5061, 4890 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ior,
      0, 61,
      { 5025, 5062 },
      -1,
   } },

   /* replace1799_0 -> 4844 in the cache */
   /* replace1799 -> 4988 in the cache */

   /* ('ball_iequal2', ('ieq', 'a', 'b'), -1) => ('ball_iequal2', 'a', 'b') */
   /* search1800_0_0 -> 0 in the cache */
   /* search1800_0_1 -> 3 in the cache */
   /* search1800_0 -> 1042 in the cache */
   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_int, { -0x1 /* -1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 2,
      { 1042, 5064 },
      -1,
   } },

   /* replace1800_0 -> 0 in the cache */
   /* replace1800_1 -> 3 in the cache */
   /* replace1800 -> 2376 in the cache */

   /* ('ball_iequal2', ('feq', 'a', 'b'), -1) => ('ball_fequal2', 'a', 'b') */
   /* search1801_0_0 -> 0 in the cache */
   /* search1801_0_1 -> 3 in the cache */
   /* search1801_0 -> 1036 in the cache */
   /* search1801_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal2,
      0, 2,
      { 1036, 5064 },
      -1,
   } },

   /* replace1801_0 -> 0 in the cache */
   /* replace1801_1 -> 3 in the cache */
   /* replace1801 -> 2411 in the cache */

   /* ('bany_inequal2', ('ine', 'a', 'b'), 0) => ('bany_inequal2', 'a', 'b') */
   /* search1802_0_0 -> 0 in the cache */
   /* search1802_0_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 0, 3 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 2,
      { 5067, 5068 },
      -1,
   } },

   /* replace1802_0 -> 0 in the cache */
   /* replace1802_1 -> 3 in the cache */
   /* replace1802 -> 2396 in the cache */

   /* ('bany_inequal2', ('fneu', 'a', 'b'), 0) => ('bany_fnequal2', 'a', 'b') */
   /* search1803_0_0 -> 0 in the cache */
   /* search1803_0_1 -> 3 in the cache */
   /* search1803_0 -> 3781 in the cache */
   /* search1803_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal2,
      0, 2,
      { 3781, 5068 },
      -1,
   } },

   /* replace1803_0 -> 0 in the cache */
   /* replace1803_1 -> 3 in the cache */
   /* replace1803 -> 2426 in the cache */

   /* ('ball_iequal3', ('ieq', 'a', 'b'), -1) => ('ball_iequal3', 'a', 'b') */
   /* search1804_0_0 -> 0 in the cache */
   /* search1804_0_1 -> 3 in the cache */
   /* search1804_0 -> 1042 in the cache */
   /* search1804_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 2,
      { 1042, 5064 },
      -1,
   } },

   /* replace1804_0 -> 0 in the cache */
   /* replace1804_1 -> 3 in the cache */
   /* replace1804 -> 2382 in the cache */

   /* ('ball_iequal3', ('feq', 'a', 'b'), -1) => ('ball_fequal3', 'a', 'b') */
   /* search1805_0_0 -> 0 in the cache */
   /* search1805_0_1 -> 3 in the cache */
   /* search1805_0 -> 1036 in the cache */
   /* search1805_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal3,
      0, 2,
      { 1036, 5064 },
      -1,
   } },

   /* replace1805_0 -> 0 in the cache */
   /* replace1805_1 -> 3 in the cache */
   /* replace1805 -> 2415 in the cache */

   /* ('bany_inequal3', ('ine', 'a', 'b'), 0) => ('bany_inequal3', 'a', 'b') */
   /* search1806_0_0 -> 0 in the cache */
   /* search1806_0_1 -> 3 in the cache */
   /* search1806_0 -> 5067 in the cache */
   /* search1806_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 2,
      { 5067, 5068 },
      -1,
   } },

   /* replace1806_0 -> 0 in the cache */
   /* replace1806_1 -> 3 in the cache */
   /* replace1806 -> 2400 in the cache */

   /* ('bany_inequal3', ('fneu', 'a', 'b'), 0) => ('bany_fnequal3', 'a', 'b') */
   /* search1807_0_0 -> 0 in the cache */
   /* search1807_0_1 -> 3 in the cache */
   /* search1807_0 -> 3781 in the cache */
   /* search1807_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal3,
      0, 2,
      { 3781, 5068 },
      -1,
   } },

   /* replace1807_0 -> 0 in the cache */
   /* replace1807_1 -> 3 in the cache */
   /* replace1807 -> 2430 in the cache */

   /* ('ball_iequal4', ('ieq', 'a', 'b'), -1) => ('ball_iequal4', 'a', 'b') */
   /* search1808_0_0 -> 0 in the cache */
   /* search1808_0_1 -> 3 in the cache */
   /* search1808_0 -> 1042 in the cache */
   /* search1808_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 2,
      { 1042, 5064 },
      -1,
   } },

   /* replace1808_0 -> 0 in the cache */
   /* replace1808_1 -> 3 in the cache */
   /* replace1808 -> 2389 in the cache */

   /* ('ball_iequal4', ('feq', 'a', 'b'), -1) => ('ball_fequal4', 'a', 'b') */
   /* search1809_0_0 -> 0 in the cache */
   /* search1809_0_1 -> 3 in the cache */
   /* search1809_0 -> 1036 in the cache */
   /* search1809_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal4,
      0, 2,
      { 1036, 5064 },
      -1,
   } },

   /* replace1809_0 -> 0 in the cache */
   /* replace1809_1 -> 3 in the cache */
   /* replace1809 -> 2421 in the cache */

   /* ('bany_inequal4', ('ine', 'a', 'b'), 0) => ('bany_inequal4', 'a', 'b') */
   /* search1810_0_0 -> 0 in the cache */
   /* search1810_0_1 -> 3 in the cache */
   /* search1810_0 -> 5067 in the cache */
   /* search1810_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 2,
      { 5067, 5068 },
      -1,
   } },

   /* replace1810_0 -> 0 in the cache */
   /* replace1810_1 -> 3 in the cache */
   /* replace1810 -> 2406 in the cache */

   /* ('bany_inequal4', ('fneu', 'a', 'b'), 0) => ('bany_fnequal4', 'a', 'b') */
   /* search1811_0_0 -> 0 in the cache */
   /* search1811_0_1 -> 3 in the cache */
   /* search1811_0 -> 3781 in the cache */
   /* search1811_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal4,
      0, 2,
      { 3781, 5068 },
      -1,
   } },

   /* replace1811_0 -> 0 in the cache */
   /* replace1811_1 -> 3 in the cache */
   /* replace1811 -> 2436 in the cache */

   /* ('ball_iequal8', ('ieq', 'a', 'b'), -1) => ('ball_iequal8', 'a', 'b') */
   /* search1812_0_0 -> 0 in the cache */
   /* search1812_0_1 -> 3 in the cache */
   /* search1812_0 -> 1042 in the cache */
   /* search1812_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 2,
      { 1042, 5064 },
      -1,
   } },

   /* replace1812_0 -> 0 in the cache */
   /* replace1812_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ball_iequal8', ('feq', 'a', 'b'), -1) => ('ball_fequal8', 'a', 'b') */
   /* search1813_0_0 -> 0 in the cache */
   /* search1813_0_1 -> 3 in the cache */
   /* search1813_0 -> 1036 in the cache */
   /* search1813_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal8,
      0, 2,
      { 1036, 5064 },
      -1,
   } },

   /* replace1813_0 -> 0 in the cache */
   /* replace1813_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bany_inequal8', ('ine', 'a', 'b'), 0) => ('bany_inequal8', 'a', 'b') */
   /* search1814_0_0 -> 0 in the cache */
   /* search1814_0_1 -> 3 in the cache */
   /* search1814_0 -> 5067 in the cache */
   /* search1814_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 2,
      { 5067, 5068 },
      -1,
   } },

   /* replace1814_0 -> 0 in the cache */
   /* replace1814_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bany_inequal8', ('fneu', 'a', 'b'), 0) => ('bany_fnequal8', 'a', 'b') */
   /* search1815_0_0 -> 0 in the cache */
   /* search1815_0_1 -> 3 in the cache */
   /* search1815_0 -> 3781 in the cache */
   /* search1815_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal8,
      0, 2,
      { 3781, 5068 },
      -1,
   } },

   /* replace1815_0 -> 0 in the cache */
   /* replace1815_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal8,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ball_iequal16', ('ieq', 'a', 'b'), -1) => ('ball_iequal16', 'a', 'b') */
   /* search1816_0_0 -> 0 in the cache */
   /* search1816_0_1 -> 3 in the cache */
   /* search1816_0 -> 1042 in the cache */
   /* search1816_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 2,
      { 1042, 5064 },
      -1,
   } },

   /* replace1816_0 -> 0 in the cache */
   /* replace1816_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('ball_iequal16', ('feq', 'a', 'b'), -1) => ('ball_fequal16', 'a', 'b') */
   /* search1817_0_0 -> 0 in the cache */
   /* search1817_0_1 -> 3 in the cache */
   /* search1817_0 -> 1036 in the cache */
   /* search1817_1 -> 5064 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_iequal16,
      0, 2,
      { 1036, 5064 },
      -1,
   } },

   /* replace1817_0 -> 0 in the cache */
   /* replace1817_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ball_fequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bany_inequal16', ('ine', 'a', 'b'), 0) => ('bany_inequal16', 'a', 'b') */
   /* search1818_0_0 -> 0 in the cache */
   /* search1818_0_1 -> 3 in the cache */
   /* search1818_0 -> 5067 in the cache */
   /* search1818_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 2,
      { 5067, 5068 },
      -1,
   } },

   /* replace1818_0 -> 0 in the cache */
   /* replace1818_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('bany_inequal16', ('fneu', 'a', 'b'), 0) => ('bany_fnequal16', 'a', 'b') */
   /* search1819_0_0 -> 0 in the cache */
   /* search1819_0_1 -> 3 in the cache */
   /* search1819_0 -> 3781 in the cache */
   /* search1819_1 -> 5068 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_inequal16,
      0, 2,
      { 3781, 5068 },
      -1,
   } },

   /* replace1819_0 -> 0 in the cache */
   /* replace1819_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bany_fnequal16,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('flt', 'a', 'b')) => ('!flt', 'a', 'b') */
   /* search1820_0_0 -> 0 in the cache */
   /* search1820_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search1820_1_0 -> 0 in the cache */
   /* search1820_1_1 -> 3 in the cache */
   /* search1820_1 -> 1011 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5095, 1011 },
      -1,
   } },

   /* replace1820_0 -> 0 in the cache */
   /* replace1820_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('flt', 'b', 'a')) => ('!flt', 'b', 'a') */
   /* search1821_0_0 -> 32 in the cache */
   /* search1821_0_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 32, 32 },
      -1,
   } },
   /* search1821_1_0 -> 2221 in the cache */
   /* search1821_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5098, 5099 },
      -1,
   } },

   /* replace1821_0 -> 2221 in the cache */
   /* replace1821_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 2221, 32 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('fge', 'a', 'b')) => ('!fge', 'a', 'b') */
   /* search1822_0_0 -> 0 in the cache */
   /* search1822_0_1 -> 0 in the cache */
   /* search1822_0 -> 5095 in the cache */
   /* search1822_1_0 -> 0 in the cache */
   /* search1822_1_1 -> 3 in the cache */
   /* search1822_1 -> 1008 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5095, 1008 },
      -1,
   } },

   /* replace1822_0 -> 0 in the cache */
   /* replace1822_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('fge', 'b', 'a')) => ('!fge', 'b', 'a') */
   /* search1823_0_0 -> 32 in the cache */
   /* search1823_0_1 -> 32 in the cache */
   /* search1823_0 -> 5098 in the cache */
   /* search1823_1_0 -> 2221 in the cache */
   /* search1823_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 5098, 5104 },
      -1,
   } },

   /* replace1823_0 -> 2221 in the cache */
   /* replace1823_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_fge,
      -1, 0,
      { 2221, 32 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('feq', 'a', 'b')) => ('!feq', 'a', 'b') */
   /* search1824_0_0 -> 0 in the cache */
   /* search1824_0_1 -> 0 in the cache */
   /* search1824_0 -> 5095 in the cache */
   /* search1824_1_0 -> 0 in the cache */
   /* search1824_1_1 -> 3 in the cache */
   /* search1824_1 -> 1036 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5095, 1036 },
      -1,
   } },

   /* replace1824_0 -> 0 in the cache */
   /* replace1824_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_feq,
      0, 1,
      { 0, 3 },
      -1,
   } },

   /* ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'a')) => ('!feq', 'b', 'a') */
   /* search1825_0_0 -> 32 in the cache */
   /* search1825_0_1 -> 32 in the cache */
   /* search1825_0 -> 5098 in the cache */
   /* search1825_1_0 -> 2221 in the cache */
   /* search1825_1_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 2221, 32 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 5098, 5109 },
      -1,
   } },

   /* replace1825_0 -> 2221 in the cache */
   /* replace1825_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_feq,
      0, 1,
      { 2221, 32 },
      -1,
   } },

   /* ('feq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('feq', 'b', 'd'), ('feq', 'c', 'd')) */
   /* search1826_0_0 -> 722 in the cache */
   /* search1826_0_1 -> 990 in the cache */
   /* search1826_0_2 -> 3225 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 990, 3225 },
      -1,
   } },
   /* search1826_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 5112, 993 },
      -1,
   } },

   /* replace1826_0 -> 722 in the cache */
   /* replace1826_1_0 -> 152 in the cache */
   /* replace1826_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1826_2_0 -> 154 in the cache */
   /* replace1826_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5114, 5115 },
      -1,
   } },

   /* ('fneu', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fneu', 'b', 'd'), ('fneu', 'c', 'd')) */
   /* search1827_0_0 -> 722 in the cache */
   /* search1827_0_1 -> 990 in the cache */
   /* search1827_0_2 -> 3225 in the cache */
   /* search1827_0 -> 5112 in the cache */
   /* search1827_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 5112, 993 },
      -1,
   } },

   /* replace1827_0 -> 722 in the cache */
   /* replace1827_1_0 -> 152 in the cache */
   /* replace1827_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1827_2_0 -> 154 in the cache */
   /* replace1827_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5118, 5119 },
      -1,
   } },

   /* ('ieq', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ieq', 'b', 'd'), ('ieq', 'c', 'd')) */
   /* search1828_0_0 -> 722 in the cache */
   /* search1828_0_1 -> 990 in the cache */
   /* search1828_0_2 -> 3225 in the cache */
   /* search1828_0 -> 5112 in the cache */
   /* search1828_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 5112, 993 },
      -1,
   } },

   /* replace1828_0 -> 722 in the cache */
   /* replace1828_1_0 -> 152 in the cache */
   /* replace1828_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1828_2_0 -> 154 in the cache */
   /* replace1828_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5122, 5123 },
      -1,
   } },

   /* ('ine', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ine', 'b', 'd'), ('ine', 'c', 'd')) */
   /* search1829_0_0 -> 722 in the cache */
   /* search1829_0_1 -> 990 in the cache */
   /* search1829_0_2 -> 3225 in the cache */
   /* search1829_0 -> 5112 in the cache */
   /* search1829_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 5112, 993 },
      -1,
   } },

   /* replace1829_0 -> 722 in the cache */
   /* replace1829_1_0 -> 152 in the cache */
   /* replace1829_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1829_2_0 -> 154 in the cache */
   /* replace1829_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5126, 5127 },
      -1,
   } },

   /* ('flt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('flt', 'b', 'd'), ('flt', 'c', 'd')) */
   /* search1830_0_0 -> 722 in the cache */
   /* search1830_0_1 -> 990 in the cache */
   /* search1830_0_2 -> 3225 in the cache */
   /* search1830_0 -> 5112 in the cache */
   /* search1830_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5112, 993 },
      -1,
   } },

   /* replace1830_0 -> 722 in the cache */
   /* replace1830_1_0 -> 152 in the cache */
   /* replace1830_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   /* replace1830_2_0 -> 154 in the cache */
   /* replace1830_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5130, 5131 },
      -1,
   } },

   /* ('flt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('flt', 'd', 'b'), ('flt', 'd', 'c')) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* d */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1831_1_0 -> 1131 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5134, 5135 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5133, 5136 },
      -1,
   } },

   /* replace1831_0 -> 1131 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5138, 5139 },
      -1,
   } },
   /* replace1831_2_0 -> 5138 in the cache */
   /* replace1831_2_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 5138, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5140, 5141 },
      -1,
   } },

   /* ('fge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('fge', 'b', 'd'), ('fge', 'c', 'd')) */
   /* search1832_0_0 -> 722 in the cache */
   /* search1832_0_1 -> 990 in the cache */
   /* search1832_0_2 -> 3225 in the cache */
   /* search1832_0 -> 5112 in the cache */
   /* search1832_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5112, 993 },
      -1,
   } },

   /* replace1832_0 -> 722 in the cache */
   /* replace1832_1_0 -> 152 in the cache */
   /* replace1832_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   /* replace1832_2_0 -> 154 in the cache */
   /* replace1832_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5144, 5145 },
      -1,
   } },

   /* ('fge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('fge', 'd', 'b'), ('fge', 'd', 'c')) */
   /* search1833_0 -> 5133 in the cache */
   /* search1833_1_0 -> 1131 in the cache */
   /* search1833_1_1 -> 5134 in the cache */
   /* search1833_1_2 -> 5135 in the cache */
   /* search1833_1 -> 5136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5133, 5136 },
      -1,
   } },

   /* replace1833_0 -> 1131 in the cache */
   /* replace1833_1_0 -> 5138 in the cache */
   /* replace1833_1_1 -> 5139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5138, 5139 },
      -1,
   } },
   /* replace1833_2_0 -> 5138 in the cache */
   /* replace1833_2_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 5138, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5148, 5149 },
      -1,
   } },

   /* ('ilt', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ilt', 'b', 'd'), ('ilt', 'c', 'd')) */
   /* search1834_0_0 -> 722 in the cache */
   /* search1834_0_1 -> 990 in the cache */
   /* search1834_0_2 -> 3225 in the cache */
   /* search1834_0 -> 5112 in the cache */
   /* search1834_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5112, 993 },
      -1,
   } },

   /* replace1834_0 -> 722 in the cache */
   /* replace1834_1_0 -> 152 in the cache */
   /* replace1834_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   /* replace1834_2_0 -> 154 in the cache */
   /* replace1834_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5152, 5153 },
      -1,
   } },

   /* ('ilt', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ilt', 'd', 'b'), ('ilt', 'd', 'c')) */
   /* search1835_0 -> 5133 in the cache */
   /* search1835_1_0 -> 1131 in the cache */
   /* search1835_1_1 -> 5134 in the cache */
   /* search1835_1_2 -> 5135 in the cache */
   /* search1835_1 -> 5136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5133, 5136 },
      -1,
   } },

   /* replace1835_0 -> 1131 in the cache */
   /* replace1835_1_0 -> 5138 in the cache */
   /* replace1835_1_1 -> 5139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5138, 5139 },
      -1,
   } },
   /* replace1835_2_0 -> 5138 in the cache */
   /* replace1835_2_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ilt,
      -1, 0,
      { 5138, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5156, 5157 },
      -1,
   } },

   /* ('ige', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ige', 'b', 'd'), ('ige', 'c', 'd')) */
   /* search1836_0_0 -> 722 in the cache */
   /* search1836_0_1 -> 990 in the cache */
   /* search1836_0_2 -> 3225 in the cache */
   /* search1836_0 -> 5112 in the cache */
   /* search1836_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5112, 993 },
      -1,
   } },

   /* replace1836_0 -> 722 in the cache */
   /* replace1836_1_0 -> 152 in the cache */
   /* replace1836_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   /* replace1836_2_0 -> 154 in the cache */
   /* replace1836_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5160, 5161 },
      -1,
   } },

   /* ('ige', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ige', 'd', 'b'), ('ige', 'd', 'c')) */
   /* search1837_0 -> 5133 in the cache */
   /* search1837_1_0 -> 1131 in the cache */
   /* search1837_1_1 -> 5134 in the cache */
   /* search1837_1_2 -> 5135 in the cache */
   /* search1837_1 -> 5136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5133, 5136 },
      -1,
   } },

   /* replace1837_0 -> 1131 in the cache */
   /* replace1837_1_0 -> 5138 in the cache */
   /* replace1837_1_1 -> 5139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5138, 5139 },
      -1,
   } },
   /* replace1837_2_0 -> 5138 in the cache */
   /* replace1837_2_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ige,
      -1, 0,
      { 5138, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5164, 5165 },
      -1,
   } },

   /* ('ult', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('ult', 'b', 'd'), ('ult', 'c', 'd')) */
   /* search1838_0_0 -> 722 in the cache */
   /* search1838_0_1 -> 990 in the cache */
   /* search1838_0_2 -> 3225 in the cache */
   /* search1838_0 -> 5112 in the cache */
   /* search1838_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5112, 993 },
      -1,
   } },

   /* replace1838_0 -> 722 in the cache */
   /* replace1838_1_0 -> 152 in the cache */
   /* replace1838_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   /* replace1838_2_0 -> 154 in the cache */
   /* replace1838_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5168, 5169 },
      -1,
   } },

   /* ('ult', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('ult', 'd', 'b'), ('ult', 'd', 'c')) */
   /* search1839_0 -> 5133 in the cache */
   /* search1839_1_0 -> 1131 in the cache */
   /* search1839_1_1 -> 5134 in the cache */
   /* search1839_1_2 -> 5135 in the cache */
   /* search1839_1 -> 5136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5133, 5136 },
      -1,
   } },

   /* replace1839_0 -> 1131 in the cache */
   /* replace1839_1_0 -> 5138 in the cache */
   /* replace1839_1_1 -> 5139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5138, 5139 },
      -1,
   } },
   /* replace1839_2_0 -> 5138 in the cache */
   /* replace1839_2_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 0,
      { 5138, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5172, 5173 },
      -1,
   } },

   /* ('uge', ('bcsel', 'a', '#b', '#c'), '#d') => ('bcsel', 'a', ('uge', 'b', 'd'), ('uge', 'c', 'd')) */
   /* search1840_0_0 -> 722 in the cache */
   /* search1840_0_1 -> 990 in the cache */
   /* search1840_0_2 -> 3225 in the cache */
   /* search1840_0 -> 5112 in the cache */
   /* search1840_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5112, 993 },
      -1,
   } },

   /* replace1840_0 -> 722 in the cache */
   /* replace1840_1_0 -> 152 in the cache */
   /* replace1840_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   /* replace1840_2_0 -> 154 in the cache */
   /* replace1840_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5176, 5177 },
      -1,
   } },

   /* ('uge', '#d', ('bcsel', 'a', '#b', '#c')) => ('bcsel', 'a', ('uge', 'd', 'b'), ('uge', 'd', 'c')) */
   /* search1841_0 -> 5133 in the cache */
   /* search1841_1_0 -> 1131 in the cache */
   /* search1841_1_1 -> 5134 in the cache */
   /* search1841_1_2 -> 5135 in the cache */
   /* search1841_1 -> 5136 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5133, 5136 },
      -1,
   } },

   /* replace1841_0 -> 1131 in the cache */
   /* replace1841_1_0 -> 5138 in the cache */
   /* replace1841_1_1 -> 5139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5138, 5139 },
      -1,
   } },
   /* replace1841_2_0 -> 5138 in the cache */
   /* replace1841_2_1 -> 1670 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_uge,
      -1, 0,
      { 5138, 1670 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 1131, 5180, 5181 },
      -1,
   } },

   /* ('fadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fadd', 'b', 'd'), ('fadd', 'c', 'd')) */
   /* search1842_0_0 -> 722 in the cache */
   /* search1842_0_1 -> 990 in the cache */
   /* search1842_0_2 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 990, 154 },
      0,
   } },
   /* search1842_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 5183, 993 },
      -1,
   } },

   /* replace1842_0 -> 722 in the cache */
   /* replace1842_1_0 -> 152 in the cache */
   /* replace1842_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1842_2_0 -> 154 in the cache */
   /* replace1842_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5185, 5186 },
      -1,
   } },

   /* ('fmul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmul', 'b', 'd'), ('fmul', 'c', 'd')) */
   /* search1843_0_0 -> 722 in the cache */
   /* search1843_0_1 -> 990 in the cache */
   /* search1843_0_2 -> 154 in the cache */
   /* search1843_0 -> 5183 in the cache */
   /* search1843_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 5183, 993 },
      -1,
   } },

   /* replace1843_0 -> 722 in the cache */
   /* replace1843_1_0 -> 152 in the cache */
   /* replace1843_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1843_2_0 -> 154 in the cache */
   /* replace1843_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5189, 5190 },
      -1,
   } },

   /* ('fmulz', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('fmulz', 'b', 'd'), ('fmulz', 'c', 'd')) */
   /* search1844_0_0 -> 722 in the cache */
   /* search1844_0_1 -> 33 in the cache */
   /* search1844_0_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 33, 137 },
      0,
   } },
   /* search1844_1 -> 300 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 5192, 300 },
      -1,
   } },

   /* replace1844_0 -> 722 in the cache */
   /* replace1844_1_0 -> 36 in the cache */
   /* replace1844_1_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 165 },
      -1,
   } },
   /* replace1844_2_0 -> 137 in the cache */
   /* replace1844_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 137, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5194, 5195 },
      -1,
   } },

   /* ('iadd', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('iadd', 'b', 'd'), ('iadd', 'c', 'd')) */
   /* search1845_0_0 -> 722 in the cache */
   /* search1845_0_1 -> 990 in the cache */
   /* search1845_0_2 -> 154 in the cache */
   /* search1845_0 -> 5183 in the cache */
   /* search1845_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5183, 993 },
      -1,
   } },

   /* replace1845_0 -> 722 in the cache */
   /* replace1845_1_0 -> 152 in the cache */
   /* replace1845_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1845_2_0 -> 154 in the cache */
   /* replace1845_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5198, 5199 },
      -1,
   } },

   /* ('imul', ('bcsel(is_used_once)', 'a', '#b', 'c'), '#d') => ('bcsel', 'a', ('imul', 'b', 'd'), ('imul', 'c', 'd')) */
   /* search1846_0_0 -> 722 in the cache */
   /* search1846_0_1 -> 990 in the cache */
   /* search1846_0_2 -> 154 in the cache */
   /* search1846_0 -> 5183 in the cache */
   /* search1846_1 -> 993 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 5183, 993 },
      -1,
   } },

   /* replace1846_0 -> 722 in the cache */
   /* replace1846_1_0 -> 152 in the cache */
   /* replace1846_1_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 152, 155 },
      -1,
   } },
   /* replace1846_2_0 -> 154 in the cache */
   /* replace1846_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5202, 5203 },
      -1,
   } },

   /* ('fddx', 'a') => 0.0 */
   /* search1847_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1847 -> 113 in the cache */

   /* ('fddx_fine', 'a') => 0.0 */
   /* search1848_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1848 -> 113 in the cache */

   /* ('fddx_coarse', 'a') => 0.0 */
   /* search1849_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1849 -> 113 in the cache */

   /* ('fddy', 'a') => 0.0 */
   /* search1850_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1850 -> 113 in the cache */

   /* ('fddy_fine', 'a') => 0.0 */
   /* search1851_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1851 -> 113 in the cache */

   /* ('fddy_coarse', 'a') => 0.0 */
   /* search1852_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 32 },
      -1,
   } },

   /* replace1852 -> 113 in the cache */

   /* ('umul_low', '#a(is_lower_half_zero)', 'b') => 0 */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      true,
      nir_type_invalid,
      30,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1853_1 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umul_low,
      0, 1,
      { 5211, 36 },
      -1,
   } },

   /* replace1853 -> 124 in the cache */

   /* ('imadsh_mix16', '#a@32(is_lower_half_zero)', 'b@32', 'c@32') => c */
   /* search1854_0 -> 5211 in the cache */
   /* search1854_1 -> 36 in the cache */
   /* search1854_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imadsh_mix16,
      -1, 0,
      { 5211, 36, 137 },
      -1,
   } },

   /* replace1854 -> 137 in the cache */

   /* ('imadsh_mix16', 'a@32', '#b@32(is_upper_half_zero)', 'c@32') => c */
   /* search1855_0 -> 16 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      31,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1855_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_imadsh_mix16,
      -1, 0,
      { 16, 5214, 137 },
      -1,
   } },

   /* replace1855 -> 137 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search1856_0 -> 722 in the cache */
   /* search1856_1_0 -> 714 in the cache */
   /* search1856_1_1 -> 715 in the cache */
   /* search1856_1_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 715, 724 },
      0,
   } },
   /* search1856_2_0 -> 714 in the cache */
   /* search1856_2_1 -> 715 in the cache */
   /* search1856_2_2 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 715, 726 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5216, 5217 },
      -1,
   } },

   /* replace1856_0 -> 714 in the cache */
   /* replace1856_1 -> 715 in the cache */
   /* replace1856_2_0 -> 722 in the cache */
   /* replace1856_2_1 -> 724 in the cache */
   /* replace1856_2_2 -> 726 in the cache */
   /* replace1856_2 -> 1257 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 715, 1257 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'c', 'e')) => ('flrp', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search1857_0 -> 722 in the cache */
   /* search1857_1_0 -> 714 in the cache */
   /* search1857_1_1 -> 715 in the cache */
   /* search1857_1_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 715, 724 },
      -1,
   } },
   /* search1857_2_0 -> 714 in the cache */
   /* search1857_2_1 -> 715 in the cache */
   /* search1857_2_2 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 715, 726 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5220, 5221 },
      -1,
   } },

   /* replace1857_0 -> 714 in the cache */
   /* replace1857_1 -> 715 in the cache */
   /* replace1857_2_0 -> 722 in the cache */
   /* replace1857_2_1 -> 724 in the cache */
   /* replace1857_2_2 -> 726 in the cache */
   /* replace1857_2 -> 1257 in the cache */
   /* replace1857 -> 5219 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search1858_0 -> 722 in the cache */
   /* search1858_1_0 -> 714 in the cache */
   /* search1858_1_1 -> 715 in the cache */
   /* search1858_1_2 -> 724 in the cache */
   /* search1858_1 -> 5216 in the cache */
   /* search1858_2_0 -> 714 in the cache */
   /* search1858_2_1 -> 726 in the cache */
   /* search1858_2_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 726, 724 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5216, 5223 },
      -1,
   } },

   /* replace1858_0 -> 714 in the cache */
   /* replace1858_1_0 -> 722 in the cache */
   /* replace1858_1_1 -> 715 in the cache */
   /* replace1858_1_2 -> 726 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 715, 726 },
      -1,
   } },
   /* replace1858_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 5225, 724 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'b', 'e', 'd')) => ('flrp', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search1859_0 -> 722 in the cache */
   /* search1859_1_0 -> 714 in the cache */
   /* search1859_1_1 -> 715 in the cache */
   /* search1859_1_2 -> 724 in the cache */
   /* search1859_1 -> 5220 in the cache */
   /* search1859_2_0 -> 714 in the cache */
   /* search1859_2_1 -> 726 in the cache */
   /* search1859_2_2 -> 724 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 714, 726, 724 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5220, 5227 },
      -1,
   } },

   /* replace1859_0 -> 714 in the cache */
   /* replace1859_1_0 -> 722 in the cache */
   /* replace1859_1_1 -> 715 in the cache */
   /* replace1859_1_2 -> 726 in the cache */
   /* replace1859_1 -> 5225 in the cache */
   /* replace1859_2 -> 724 in the cache */
   /* replace1859 -> 5226 in the cache */

   /* ('bcsel', 'a', ('flrp(is_used_once)', 'b', 'c', 'd'), ('flrp', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd') */
   /* search1860_0 -> 722 in the cache */
   /* search1860_1_0 -> 152 in the cache */
   /* search1860_1_1 -> 154 in the cache */
   /* search1860_1_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 152, 154, 155 },
      0,
   } },
   /* search1860_2_0 -> 1263 in the cache */
   /* search1860_2_1 -> 154 in the cache */
   /* search1860_2_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 1263, 154, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5229, 5230 },
      -1,
   } },

   /* replace1860_0_0 -> 722 in the cache */
   /* replace1860_0_1 -> 152 in the cache */
   /* replace1860_0_2 -> 1263 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 152, 1263 },
      -1,
   } },
   /* replace1860_1 -> 154 in the cache */
   /* replace1860_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 5232, 154, 155 },
      -1,
   } },

   /* ('bcsel', 'a', ('flrp', 'b', 'c', 'd'), ('flrp(is_used_once)', 'e', 'c', 'd')) => ('flrp', ('bcsel', 'a', 'b', 'e'), 'c', 'd') */
   /* search1861_0 -> 722 in the cache */
   /* search1861_1_0 -> 152 in the cache */
   /* search1861_1_1 -> 154 in the cache */
   /* search1861_1_2 -> 155 in the cache */
   /* search1861_1 -> 667 in the cache */
   /* search1861_2_0 -> 1263 in the cache */
   /* search1861_2_1 -> 154 in the cache */
   /* search1861_2_2 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 1263, 154, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 667, 5234 },
      -1,
   } },

   /* replace1861_0_0 -> 722 in the cache */
   /* replace1861_0_1 -> 152 in the cache */
   /* replace1861_0_2 -> 1263 in the cache */
   /* replace1861_0 -> 5232 in the cache */
   /* replace1861_1 -> 154 in the cache */
   /* replace1861_2 -> 155 in the cache */
   /* replace1861 -> 5233 in the cache */

   /* ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c'), ('fmulz', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1862_0 -> 722 in the cache */
   /* search1862_1_0 -> 36 in the cache */
   /* search1862_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 137 },
      0,
   } },
   /* search1862_2_0 -> 36 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 5237 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5236, 5238 },
      -1,
   } },

   /* replace1862_0 -> 36 in the cache */
   /* replace1862_1_0 -> 722 in the cache */
   /* replace1862_1_1 -> 137 in the cache */
   /* replace1862_1_2 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 137, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 5240 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmulz(is_used_once)', 'b', 'c(is_not_const)'), ('fmulz', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1863_0 -> 722 in the cache */
   /* search1863_1_0 -> 36 in the cache */
   /* search1863_1_1 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 336 },
      0,
   } },
   /* search1863_2_0 -> 36 in the cache */
   /* search1863_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5242, 5243 },
      -1,
   } },

   /* replace1863_0 -> 36 in the cache */
   /* replace1863_1_0 -> 722 in the cache */
   /* replace1863_1_1 -> 137 in the cache */
   /* replace1863_1_2 -> 165 in the cache */
   /* replace1863_1 -> 5240 in the cache */
   /* replace1863 -> 5241 in the cache */

   /* ('bcsel', 'a', ('fmulz', 'b', 'c(is_not_const)'), ('fmulz(is_used_once)', 'b', 'd')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1864_0 -> 722 in the cache */
   /* search1864_1_0 -> 36 in the cache */
   /* search1864_1_1 -> 336 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 336 },
      -1,
   } },
   /* search1864_2_0 -> 36 in the cache */
   /* search1864_2_1 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 165 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5245, 5246 },
      -1,
   } },

   /* replace1864_0 -> 36 in the cache */
   /* replace1864_1_0 -> 722 in the cache */
   /* replace1864_1_1 -> 137 in the cache */
   /* replace1864_1_2 -> 165 in the cache */
   /* replace1864_1 -> 5240 in the cache */
   /* replace1864 -> 5241 in the cache */

   /* ('bcsel', 'a', ('fmulz', 'b', 'c'), ('fmulz(is_used_once)', 'b', 'd(is_not_const)')) => ('fmulz', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1865_0 -> 722 in the cache */
   /* search1865_1_0 -> 36 in the cache */
   /* search1865_1_1 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      0, 1,
      { 36, 137 },
      -1,
   } },
   /* search1865_2_0 -> 36 in the cache */
   /* search1865_2_1 -> 5237 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 36, 5237 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5248, 5249 },
      -1,
   } },

   /* replace1865_0 -> 36 in the cache */
   /* replace1865_1_0 -> 722 in the cache */
   /* replace1865_1_1 -> 137 in the cache */
   /* replace1865_1_2 -> 165 in the cache */
   /* replace1865_1 -> 5240 in the cache */
   /* replace1865 -> 5241 in the cache */

   /* ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c'), ('fmul', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1866_0 -> 722 in the cache */
   /* search1866_1_0 -> 152 in the cache */
   /* search1866_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1866_2_0 -> 152 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5251, 5253 },
      -1,
   } },

   /* replace1866_0 -> 152 in the cache */
   /* replace1866_1_0 -> 722 in the cache */
   /* replace1866_1_1 -> 154 in the cache */
   /* replace1866_1_2 -> 155 in the cache */
   /* replace1866_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmul(is_used_once)', 'b', 'c(is_not_const)'), ('fmul', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1867_0 -> 722 in the cache */
   /* search1867_1_0 -> 152 in the cache */
   /* search1867_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1867_2_0 -> 152 in the cache */
   /* search1867_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5256, 5257 },
      -1,
   } },

   /* replace1867_0 -> 152 in the cache */
   /* replace1867_1_0 -> 722 in the cache */
   /* replace1867_1_1 -> 154 in the cache */
   /* replace1867_1_2 -> 155 in the cache */
   /* replace1867_1 -> 1252 in the cache */
   /* replace1867 -> 5255 in the cache */

   /* ('bcsel', 'a', ('fmul', 'b', 'c(is_not_const)'), ('fmul(is_used_once)', 'b', 'd')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1868_0 -> 722 in the cache */
   /* search1868_1_0 -> 152 in the cache */
   /* search1868_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1868_2_0 -> 152 in the cache */
   /* search1868_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5259, 5260 },
      -1,
   } },

   /* replace1868_0 -> 152 in the cache */
   /* replace1868_1_0 -> 722 in the cache */
   /* replace1868_1_1 -> 154 in the cache */
   /* replace1868_1_2 -> 155 in the cache */
   /* replace1868_1 -> 1252 in the cache */
   /* replace1868 -> 5255 in the cache */

   /* ('bcsel', 'a', ('fmul', 'b', 'c'), ('fmul(is_used_once)', 'b', 'd(is_not_const)')) => ('fmul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1869_0 -> 722 in the cache */
   /* search1869_1_0 -> 152 in the cache */
   /* search1869_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1869_2_0 -> 152 in the cache */
   /* search1869_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5262, 5263 },
      -1,
   } },

   /* replace1869_0 -> 152 in the cache */
   /* replace1869_1_0 -> 722 in the cache */
   /* replace1869_1_1 -> 154 in the cache */
   /* replace1869_1_2 -> 155 in the cache */
   /* replace1869_1 -> 1252 in the cache */
   /* replace1869 -> 5255 in the cache */

   /* ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c'), ('iadd', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1870_0 -> 722 in the cache */
   /* search1870_1_0 -> 152 in the cache */
   /* search1870_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1870_2_0 -> 152 in the cache */
   /* search1870_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5265, 5266 },
      -1,
   } },

   /* replace1870_0 -> 152 in the cache */
   /* replace1870_1_0 -> 722 in the cache */
   /* replace1870_1_1 -> 154 in the cache */
   /* replace1870_1_2 -> 155 in the cache */
   /* replace1870_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('iadd(is_used_once)', 'b', 'c(is_not_const)'), ('iadd', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1871_0 -> 722 in the cache */
   /* search1871_1_0 -> 152 in the cache */
   /* search1871_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1871_2_0 -> 152 in the cache */
   /* search1871_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5269, 5270 },
      -1,
   } },

   /* replace1871_0 -> 152 in the cache */
   /* replace1871_1_0 -> 722 in the cache */
   /* replace1871_1_1 -> 154 in the cache */
   /* replace1871_1_2 -> 155 in the cache */
   /* replace1871_1 -> 1252 in the cache */
   /* replace1871 -> 5268 in the cache */

   /* ('bcsel', 'a', ('iadd', 'b', 'c(is_not_const)'), ('iadd(is_used_once)', 'b', 'd')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1872_0 -> 722 in the cache */
   /* search1872_1_0 -> 152 in the cache */
   /* search1872_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1872_2_0 -> 152 in the cache */
   /* search1872_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5272, 5273 },
      -1,
   } },

   /* replace1872_0 -> 152 in the cache */
   /* replace1872_1_0 -> 722 in the cache */
   /* replace1872_1_1 -> 154 in the cache */
   /* replace1872_1_2 -> 155 in the cache */
   /* replace1872_1 -> 1252 in the cache */
   /* replace1872 -> 5268 in the cache */

   /* ('bcsel', 'a', ('iadd', 'b', 'c'), ('iadd(is_used_once)', 'b', 'd(is_not_const)')) => ('iadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1873_0 -> 722 in the cache */
   /* search1873_1_0 -> 152 in the cache */
   /* search1873_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1873_2_0 -> 152 in the cache */
   /* search1873_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5275, 5276 },
      -1,
   } },

   /* replace1873_0 -> 152 in the cache */
   /* replace1873_1_0 -> 722 in the cache */
   /* replace1873_1_1 -> 154 in the cache */
   /* replace1873_1_2 -> 155 in the cache */
   /* replace1873_1 -> 1252 in the cache */
   /* replace1873 -> 5268 in the cache */

   /* ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c'), ('imul', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1874_0 -> 722 in the cache */
   /* search1874_1_0 -> 152 in the cache */
   /* search1874_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1874_2_0 -> 152 in the cache */
   /* search1874_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5278, 5279 },
      -1,
   } },

   /* replace1874_0 -> 152 in the cache */
   /* replace1874_1_0 -> 722 in the cache */
   /* replace1874_1_1 -> 154 in the cache */
   /* replace1874_1_2 -> 155 in the cache */
   /* replace1874_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('imul(is_used_once)', 'b', 'c(is_not_const)'), ('imul', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1875_0 -> 722 in the cache */
   /* search1875_1_0 -> 152 in the cache */
   /* search1875_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1875_2_0 -> 152 in the cache */
   /* search1875_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5282, 5283 },
      -1,
   } },

   /* replace1875_0 -> 152 in the cache */
   /* replace1875_1_0 -> 722 in the cache */
   /* replace1875_1_1 -> 154 in the cache */
   /* replace1875_1_2 -> 155 in the cache */
   /* replace1875_1 -> 1252 in the cache */
   /* replace1875 -> 5281 in the cache */

   /* ('bcsel', 'a', ('imul', 'b', 'c(is_not_const)'), ('imul(is_used_once)', 'b', 'd')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1876_0 -> 722 in the cache */
   /* search1876_1_0 -> 152 in the cache */
   /* search1876_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1876_2_0 -> 152 in the cache */
   /* search1876_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5285, 5286 },
      -1,
   } },

   /* replace1876_0 -> 152 in the cache */
   /* replace1876_1_0 -> 722 in the cache */
   /* replace1876_1_1 -> 154 in the cache */
   /* replace1876_1_2 -> 155 in the cache */
   /* replace1876_1 -> 1252 in the cache */
   /* replace1876 -> 5281 in the cache */

   /* ('bcsel', 'a', ('imul', 'b', 'c'), ('imul(is_used_once)', 'b', 'd(is_not_const)')) => ('imul', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1877_0 -> 722 in the cache */
   /* search1877_1_0 -> 152 in the cache */
   /* search1877_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1877_2_0 -> 152 in the cache */
   /* search1877_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5288, 5289 },
      -1,
   } },

   /* replace1877_0 -> 152 in the cache */
   /* replace1877_1_0 -> 722 in the cache */
   /* replace1877_1_1 -> 154 in the cache */
   /* replace1877_1_2 -> 155 in the cache */
   /* replace1877_1 -> 1252 in the cache */
   /* replace1877 -> 5281 in the cache */

   /* ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c'), ('iand', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1878_0 -> 722 in the cache */
   /* search1878_1_0 -> 152 in the cache */
   /* search1878_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1878_2_0 -> 152 in the cache */
   /* search1878_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5291, 5292 },
      -1,
   } },

   /* replace1878_0 -> 152 in the cache */
   /* replace1878_1_0 -> 722 in the cache */
   /* replace1878_1_1 -> 154 in the cache */
   /* replace1878_1_2 -> 155 in the cache */
   /* replace1878_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('iand(is_used_once)', 'b', 'c(is_not_const)'), ('iand', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1879_0 -> 722 in the cache */
   /* search1879_1_0 -> 152 in the cache */
   /* search1879_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1879_2_0 -> 152 in the cache */
   /* search1879_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5295, 5296 },
      -1,
   } },

   /* replace1879_0 -> 152 in the cache */
   /* replace1879_1_0 -> 722 in the cache */
   /* replace1879_1_1 -> 154 in the cache */
   /* replace1879_1_2 -> 155 in the cache */
   /* replace1879_1 -> 1252 in the cache */
   /* replace1879 -> 5294 in the cache */

   /* ('bcsel', 'a', ('iand', 'b', 'c(is_not_const)'), ('iand(is_used_once)', 'b', 'd')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1880_0 -> 722 in the cache */
   /* search1880_1_0 -> 152 in the cache */
   /* search1880_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1880_2_0 -> 152 in the cache */
   /* search1880_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5298, 5299 },
      -1,
   } },

   /* replace1880_0 -> 152 in the cache */
   /* replace1880_1_0 -> 722 in the cache */
   /* replace1880_1_1 -> 154 in the cache */
   /* replace1880_1_2 -> 155 in the cache */
   /* replace1880_1 -> 1252 in the cache */
   /* replace1880 -> 5294 in the cache */

   /* ('bcsel', 'a', ('iand', 'b', 'c'), ('iand(is_used_once)', 'b', 'd(is_not_const)')) => ('iand', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1881_0 -> 722 in the cache */
   /* search1881_1_0 -> 152 in the cache */
   /* search1881_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1881_2_0 -> 152 in the cache */
   /* search1881_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_iand,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5301, 5302 },
      -1,
   } },

   /* replace1881_0 -> 152 in the cache */
   /* replace1881_1_0 -> 722 in the cache */
   /* replace1881_1_1 -> 154 in the cache */
   /* replace1881_1_2 -> 155 in the cache */
   /* replace1881_1 -> 1252 in the cache */
   /* replace1881 -> 5294 in the cache */

   /* ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c'), ('ior', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1882_0 -> 722 in the cache */
   /* search1882_1_0 -> 152 in the cache */
   /* search1882_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1882_2_0 -> 152 in the cache */
   /* search1882_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5304, 5305 },
      -1,
   } },

   /* replace1882_0 -> 152 in the cache */
   /* replace1882_1_0 -> 722 in the cache */
   /* replace1882_1_1 -> 154 in the cache */
   /* replace1882_1_2 -> 155 in the cache */
   /* replace1882_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('ior(is_used_once)', 'b', 'c(is_not_const)'), ('ior', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1883_0 -> 722 in the cache */
   /* search1883_1_0 -> 152 in the cache */
   /* search1883_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1883_2_0 -> 152 in the cache */
   /* search1883_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5308, 5309 },
      -1,
   } },

   /* replace1883_0 -> 152 in the cache */
   /* replace1883_1_0 -> 722 in the cache */
   /* replace1883_1_1 -> 154 in the cache */
   /* replace1883_1_2 -> 155 in the cache */
   /* replace1883_1 -> 1252 in the cache */
   /* replace1883 -> 5307 in the cache */

   /* ('bcsel', 'a', ('ior', 'b', 'c(is_not_const)'), ('ior(is_used_once)', 'b', 'd')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1884_0 -> 722 in the cache */
   /* search1884_1_0 -> 152 in the cache */
   /* search1884_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1884_2_0 -> 152 in the cache */
   /* search1884_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5311, 5312 },
      -1,
   } },

   /* replace1884_0 -> 152 in the cache */
   /* replace1884_1_0 -> 722 in the cache */
   /* replace1884_1_1 -> 154 in the cache */
   /* replace1884_1_2 -> 155 in the cache */
   /* replace1884_1 -> 1252 in the cache */
   /* replace1884 -> 5307 in the cache */

   /* ('bcsel', 'a', ('ior', 'b', 'c'), ('ior(is_used_once)', 'b', 'd(is_not_const)')) => ('ior', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1885_0 -> 722 in the cache */
   /* search1885_1_0 -> 152 in the cache */
   /* search1885_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1885_2_0 -> 152 in the cache */
   /* search1885_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ior,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5314, 5315 },
      -1,
   } },

   /* replace1885_0 -> 152 in the cache */
   /* replace1885_1_0 -> 722 in the cache */
   /* replace1885_1_1 -> 154 in the cache */
   /* replace1885_1_2 -> 155 in the cache */
   /* replace1885_1 -> 1252 in the cache */
   /* replace1885 -> 5307 in the cache */

   /* ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c'), ('ixor', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1886_0 -> 722 in the cache */
   /* search1886_1_0 -> 152 in the cache */
   /* search1886_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1886_2_0 -> 152 in the cache */
   /* search1886_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5317, 5318 },
      -1,
   } },

   /* replace1886_0 -> 152 in the cache */
   /* replace1886_1_0 -> 722 in the cache */
   /* replace1886_1_1 -> 154 in the cache */
   /* replace1886_1_2 -> 155 in the cache */
   /* replace1886_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('ixor(is_used_once)', 'b', 'c(is_not_const)'), ('ixor', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1887_0 -> 722 in the cache */
   /* search1887_1_0 -> 152 in the cache */
   /* search1887_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1887_2_0 -> 152 in the cache */
   /* search1887_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5321, 5322 },
      -1,
   } },

   /* replace1887_0 -> 152 in the cache */
   /* replace1887_1_0 -> 722 in the cache */
   /* replace1887_1_1 -> 154 in the cache */
   /* replace1887_1_2 -> 155 in the cache */
   /* replace1887_1 -> 1252 in the cache */
   /* replace1887 -> 5320 in the cache */

   /* ('bcsel', 'a', ('ixor', 'b', 'c(is_not_const)'), ('ixor(is_used_once)', 'b', 'd')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1888_0 -> 722 in the cache */
   /* search1888_1_0 -> 152 in the cache */
   /* search1888_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1888_2_0 -> 152 in the cache */
   /* search1888_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5324, 5325 },
      -1,
   } },

   /* replace1888_0 -> 152 in the cache */
   /* replace1888_1_0 -> 722 in the cache */
   /* replace1888_1_1 -> 154 in the cache */
   /* replace1888_1_2 -> 155 in the cache */
   /* replace1888_1 -> 1252 in the cache */
   /* replace1888 -> 5320 in the cache */

   /* ('bcsel', 'a', ('ixor', 'b', 'c'), ('ixor(is_used_once)', 'b', 'd(is_not_const)')) => ('ixor', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1889_0 -> 722 in the cache */
   /* search1889_1_0 -> 152 in the cache */
   /* search1889_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1889_2_0 -> 152 in the cache */
   /* search1889_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_ixor,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5327, 5328 },
      -1,
   } },

   /* replace1889_0 -> 152 in the cache */
   /* replace1889_1_0 -> 722 in the cache */
   /* replace1889_1_1 -> 154 in the cache */
   /* replace1889_1_2 -> 155 in the cache */
   /* replace1889_1 -> 1252 in the cache */
   /* replace1889 -> 5320 in the cache */

   /* ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c'), ('fmin', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1890_0 -> 722 in the cache */
   /* search1890_1_0 -> 152 in the cache */
   /* search1890_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1890_2_0 -> 152 in the cache */
   /* search1890_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5330, 5331 },
      -1,
   } },

   /* replace1890_0 -> 152 in the cache */
   /* replace1890_1_0 -> 722 in the cache */
   /* replace1890_1_1 -> 154 in the cache */
   /* replace1890_1_2 -> 155 in the cache */
   /* replace1890_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmin(is_used_once)', 'b', 'c(is_not_const)'), ('fmin', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1891_0 -> 722 in the cache */
   /* search1891_1_0 -> 152 in the cache */
   /* search1891_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1891_2_0 -> 152 in the cache */
   /* search1891_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5334, 5335 },
      -1,
   } },

   /* replace1891_0 -> 152 in the cache */
   /* replace1891_1_0 -> 722 in the cache */
   /* replace1891_1_1 -> 154 in the cache */
   /* replace1891_1_2 -> 155 in the cache */
   /* replace1891_1 -> 1252 in the cache */
   /* replace1891 -> 5333 in the cache */

   /* ('bcsel', 'a', ('fmin', 'b', 'c(is_not_const)'), ('fmin(is_used_once)', 'b', 'd')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1892_0 -> 722 in the cache */
   /* search1892_1_0 -> 152 in the cache */
   /* search1892_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1892_2_0 -> 152 in the cache */
   /* search1892_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5337, 5338 },
      -1,
   } },

   /* replace1892_0 -> 152 in the cache */
   /* replace1892_1_0 -> 722 in the cache */
   /* replace1892_1_1 -> 154 in the cache */
   /* replace1892_1_2 -> 155 in the cache */
   /* replace1892_1 -> 1252 in the cache */
   /* replace1892 -> 5333 in the cache */

   /* ('bcsel', 'a', ('fmin', 'b', 'c'), ('fmin(is_used_once)', 'b', 'd(is_not_const)')) => ('fmin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1893_0 -> 722 in the cache */
   /* search1893_1_0 -> 152 in the cache */
   /* search1893_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1893_2_0 -> 152 in the cache */
   /* search1893_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5340, 5341 },
      -1,
   } },

   /* replace1893_0 -> 152 in the cache */
   /* replace1893_1_0 -> 722 in the cache */
   /* replace1893_1_1 -> 154 in the cache */
   /* replace1893_1_2 -> 155 in the cache */
   /* replace1893_1 -> 1252 in the cache */
   /* replace1893 -> 5333 in the cache */

   /* ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c'), ('fmax', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1894_0 -> 722 in the cache */
   /* search1894_1_0 -> 152 in the cache */
   /* search1894_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1894_2_0 -> 152 in the cache */
   /* search1894_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5343, 5344 },
      -1,
   } },

   /* replace1894_0 -> 152 in the cache */
   /* replace1894_1_0 -> 722 in the cache */
   /* replace1894_1_1 -> 154 in the cache */
   /* replace1894_1_2 -> 155 in the cache */
   /* replace1894_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('fmax(is_used_once)', 'b', 'c(is_not_const)'), ('fmax', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1895_0 -> 722 in the cache */
   /* search1895_1_0 -> 152 in the cache */
   /* search1895_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1895_2_0 -> 152 in the cache */
   /* search1895_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5347, 5348 },
      -1,
   } },

   /* replace1895_0 -> 152 in the cache */
   /* replace1895_1_0 -> 722 in the cache */
   /* replace1895_1_1 -> 154 in the cache */
   /* replace1895_1_2 -> 155 in the cache */
   /* replace1895_1 -> 1252 in the cache */
   /* replace1895 -> 5346 in the cache */

   /* ('bcsel', 'a', ('fmax', 'b', 'c(is_not_const)'), ('fmax(is_used_once)', 'b', 'd')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1896_0 -> 722 in the cache */
   /* search1896_1_0 -> 152 in the cache */
   /* search1896_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1896_2_0 -> 152 in the cache */
   /* search1896_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5350, 5351 },
      -1,
   } },

   /* replace1896_0 -> 152 in the cache */
   /* replace1896_1_0 -> 722 in the cache */
   /* replace1896_1_1 -> 154 in the cache */
   /* replace1896_1_2 -> 155 in the cache */
   /* replace1896_1 -> 1252 in the cache */
   /* replace1896 -> 5346 in the cache */

   /* ('bcsel', 'a', ('fmax', 'b', 'c'), ('fmax(is_used_once)', 'b', 'd(is_not_const)')) => ('fmax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1897_0 -> 722 in the cache */
   /* search1897_1_0 -> 152 in the cache */
   /* search1897_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1897_2_0 -> 152 in the cache */
   /* search1897_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5353, 5354 },
      -1,
   } },

   /* replace1897_0 -> 152 in the cache */
   /* replace1897_1_0 -> 722 in the cache */
   /* replace1897_1_1 -> 154 in the cache */
   /* replace1897_1_2 -> 155 in the cache */
   /* replace1897_1 -> 1252 in the cache */
   /* replace1897 -> 5346 in the cache */

   /* ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c'), ('imin', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1898_0 -> 722 in the cache */
   /* search1898_1_0 -> 152 in the cache */
   /* search1898_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1898_2_0 -> 152 in the cache */
   /* search1898_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5356, 5357 },
      -1,
   } },

   /* replace1898_0 -> 152 in the cache */
   /* replace1898_1_0 -> 722 in the cache */
   /* replace1898_1_1 -> 154 in the cache */
   /* replace1898_1_2 -> 155 in the cache */
   /* replace1898_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('imin(is_used_once)', 'b', 'c(is_not_const)'), ('imin', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1899_0 -> 722 in the cache */
   /* search1899_1_0 -> 152 in the cache */
   /* search1899_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1899_2_0 -> 152 in the cache */
   /* search1899_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5360, 5361 },
      -1,
   } },

   /* replace1899_0 -> 152 in the cache */
   /* replace1899_1_0 -> 722 in the cache */
   /* replace1899_1_1 -> 154 in the cache */
   /* replace1899_1_2 -> 155 in the cache */
   /* replace1899_1 -> 1252 in the cache */
   /* replace1899 -> 5359 in the cache */

   /* ('bcsel', 'a', ('imin', 'b', 'c(is_not_const)'), ('imin(is_used_once)', 'b', 'd')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1900_0 -> 722 in the cache */
   /* search1900_1_0 -> 152 in the cache */
   /* search1900_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1900_2_0 -> 152 in the cache */
   /* search1900_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5363, 5364 },
      -1,
   } },

   /* replace1900_0 -> 152 in the cache */
   /* replace1900_1_0 -> 722 in the cache */
   /* replace1900_1_1 -> 154 in the cache */
   /* replace1900_1_2 -> 155 in the cache */
   /* replace1900_1 -> 1252 in the cache */
   /* replace1900 -> 5359 in the cache */

   /* ('bcsel', 'a', ('imin', 'b', 'c'), ('imin(is_used_once)', 'b', 'd(is_not_const)')) => ('imin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1901_0 -> 722 in the cache */
   /* search1901_1_0 -> 152 in the cache */
   /* search1901_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1901_2_0 -> 152 in the cache */
   /* search1901_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imin,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5366, 5367 },
      -1,
   } },

   /* replace1901_0 -> 152 in the cache */
   /* replace1901_1_0 -> 722 in the cache */
   /* replace1901_1_1 -> 154 in the cache */
   /* replace1901_1_2 -> 155 in the cache */
   /* replace1901_1 -> 1252 in the cache */
   /* replace1901 -> 5359 in the cache */

   /* ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c'), ('imax', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1902_0 -> 722 in the cache */
   /* search1902_1_0 -> 152 in the cache */
   /* search1902_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1902_2_0 -> 152 in the cache */
   /* search1902_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5369, 5370 },
      -1,
   } },

   /* replace1902_0 -> 152 in the cache */
   /* replace1902_1_0 -> 722 in the cache */
   /* replace1902_1_1 -> 154 in the cache */
   /* replace1902_1_2 -> 155 in the cache */
   /* replace1902_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('imax(is_used_once)', 'b', 'c(is_not_const)'), ('imax', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1903_0 -> 722 in the cache */
   /* search1903_1_0 -> 152 in the cache */
   /* search1903_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1903_2_0 -> 152 in the cache */
   /* search1903_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5373, 5374 },
      -1,
   } },

   /* replace1903_0 -> 152 in the cache */
   /* replace1903_1_0 -> 722 in the cache */
   /* replace1903_1_1 -> 154 in the cache */
   /* replace1903_1_2 -> 155 in the cache */
   /* replace1903_1 -> 1252 in the cache */
   /* replace1903 -> 5372 in the cache */

   /* ('bcsel', 'a', ('imax', 'b', 'c(is_not_const)'), ('imax(is_used_once)', 'b', 'd')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1904_0 -> 722 in the cache */
   /* search1904_1_0 -> 152 in the cache */
   /* search1904_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1904_2_0 -> 152 in the cache */
   /* search1904_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5376, 5377 },
      -1,
   } },

   /* replace1904_0 -> 152 in the cache */
   /* replace1904_1_0 -> 722 in the cache */
   /* replace1904_1_1 -> 154 in the cache */
   /* replace1904_1_2 -> 155 in the cache */
   /* replace1904_1 -> 1252 in the cache */
   /* replace1904 -> 5372 in the cache */

   /* ('bcsel', 'a', ('imax', 'b', 'c'), ('imax(is_used_once)', 'b', 'd(is_not_const)')) => ('imax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1905_0 -> 722 in the cache */
   /* search1905_1_0 -> 152 in the cache */
   /* search1905_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1905_2_0 -> 152 in the cache */
   /* search1905_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_imax,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5379, 5380 },
      -1,
   } },

   /* replace1905_0 -> 152 in the cache */
   /* replace1905_1_0 -> 722 in the cache */
   /* replace1905_1_1 -> 154 in the cache */
   /* replace1905_1_2 -> 155 in the cache */
   /* replace1905_1 -> 1252 in the cache */
   /* replace1905 -> 5372 in the cache */

   /* ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c'), ('umin', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1906_0 -> 722 in the cache */
   /* search1906_1_0 -> 152 in the cache */
   /* search1906_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1906_2_0 -> 152 in the cache */
   /* search1906_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5382, 5383 },
      -1,
   } },

   /* replace1906_0 -> 152 in the cache */
   /* replace1906_1_0 -> 722 in the cache */
   /* replace1906_1_1 -> 154 in the cache */
   /* replace1906_1_2 -> 155 in the cache */
   /* replace1906_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('umin(is_used_once)', 'b', 'c(is_not_const)'), ('umin', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1907_0 -> 722 in the cache */
   /* search1907_1_0 -> 152 in the cache */
   /* search1907_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1907_2_0 -> 152 in the cache */
   /* search1907_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5386, 5387 },
      -1,
   } },

   /* replace1907_0 -> 152 in the cache */
   /* replace1907_1_0 -> 722 in the cache */
   /* replace1907_1_1 -> 154 in the cache */
   /* replace1907_1_2 -> 155 in the cache */
   /* replace1907_1 -> 1252 in the cache */
   /* replace1907 -> 5385 in the cache */

   /* ('bcsel', 'a', ('umin', 'b', 'c(is_not_const)'), ('umin(is_used_once)', 'b', 'd')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1908_0 -> 722 in the cache */
   /* search1908_1_0 -> 152 in the cache */
   /* search1908_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1908_2_0 -> 152 in the cache */
   /* search1908_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5389, 5390 },
      -1,
   } },

   /* replace1908_0 -> 152 in the cache */
   /* replace1908_1_0 -> 722 in the cache */
   /* replace1908_1_1 -> 154 in the cache */
   /* replace1908_1_2 -> 155 in the cache */
   /* replace1908_1 -> 1252 in the cache */
   /* replace1908 -> 5385 in the cache */

   /* ('bcsel', 'a', ('umin', 'b', 'c'), ('umin(is_used_once)', 'b', 'd(is_not_const)')) => ('umin', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1909_0 -> 722 in the cache */
   /* search1909_1_0 -> 152 in the cache */
   /* search1909_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1909_2_0 -> 152 in the cache */
   /* search1909_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5392, 5393 },
      -1,
   } },

   /* replace1909_0 -> 152 in the cache */
   /* replace1909_1_0 -> 722 in the cache */
   /* replace1909_1_1 -> 154 in the cache */
   /* replace1909_1_2 -> 155 in the cache */
   /* replace1909_1 -> 1252 in the cache */
   /* replace1909 -> 5385 in the cache */

   /* ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c'), ('umax', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1910_0 -> 722 in the cache */
   /* search1910_1_0 -> 152 in the cache */
   /* search1910_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 152, 154 },
      0,
   } },
   /* search1910_2_0 -> 152 in the cache */
   /* search1910_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 152, 5252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5395, 5396 },
      -1,
   } },

   /* replace1910_0 -> 152 in the cache */
   /* replace1910_1_0 -> 722 in the cache */
   /* replace1910_1_1 -> 154 in the cache */
   /* replace1910_1_2 -> 155 in the cache */
   /* replace1910_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('umax(is_used_once)', 'b', 'c(is_not_const)'), ('umax', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1911_0 -> 722 in the cache */
   /* search1911_1_0 -> 152 in the cache */
   /* search1911_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 152, 3240 },
      0,
   } },
   /* search1911_2_0 -> 152 in the cache */
   /* search1911_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5399, 5400 },
      -1,
   } },

   /* replace1911_0 -> 152 in the cache */
   /* replace1911_1_0 -> 722 in the cache */
   /* replace1911_1_1 -> 154 in the cache */
   /* replace1911_1_2 -> 155 in the cache */
   /* replace1911_1 -> 1252 in the cache */
   /* replace1911 -> 5398 in the cache */

   /* ('bcsel', 'a', ('umax', 'b', 'c(is_not_const)'), ('umax(is_used_once)', 'b', 'd')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1912_0 -> 722 in the cache */
   /* search1912_1_0 -> 152 in the cache */
   /* search1912_1_1 -> 3240 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 152, 3240 },
      -1,
   } },
   /* search1912_2_0 -> 152 in the cache */
   /* search1912_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5402, 5403 },
      -1,
   } },

   /* replace1912_0 -> 152 in the cache */
   /* replace1912_1_0 -> 722 in the cache */
   /* replace1912_1_1 -> 154 in the cache */
   /* replace1912_1_2 -> 155 in the cache */
   /* replace1912_1 -> 1252 in the cache */
   /* replace1912 -> 5398 in the cache */

   /* ('bcsel', 'a', ('umax', 'b', 'c'), ('umax(is_used_once)', 'b', 'd(is_not_const)')) => ('umax', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1913_0 -> 722 in the cache */
   /* search1913_1_0 -> 152 in the cache */
   /* search1913_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      0, 1,
      { 152, 154 },
      -1,
   } },
   /* search1913_2_0 -> 152 in the cache */
   /* search1913_2_1 -> 5252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_umax,
      1, 1,
      { 152, 5252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 722, 5405, 5406 },
      -1,
   } },

   /* replace1913_0 -> 152 in the cache */
   /* replace1913_1_0 -> 722 in the cache */
   /* replace1913_1_1 -> 154 in the cache */
   /* replace1913_1_2 -> 155 in the cache */
   /* replace1913_1 -> 1252 in the cache */
   /* replace1913 -> 5398 in the cache */

   /* ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1914_0 -> 722 in the cache */
   /* search1914_1_0 -> 152 in the cache */
   /* search1914_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 152, 154 },
      0,
   } },
   /* search1914_2_0 -> 152 in the cache */
   /* search1914_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 152, 155 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5408, 5409 },
      -1,
   } },

   /* replace1914_0 -> 152 in the cache */
   /* replace1914_1_0 -> 722 in the cache */
   /* replace1914_1_1 -> 154 in the cache */
   /* replace1914_1_2 -> 155 in the cache */
   /* replace1914_1 -> 1252 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 152, 1252 },
      -1,
   } },

   /* ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'b', 'd')) => ('fpow', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search1915_0 -> 722 in the cache */
   /* search1915_1_0 -> 152 in the cache */
   /* search1915_1_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 152, 154 },
      -1,
   } },
   /* search1915_2_0 -> 152 in the cache */
   /* search1915_2_1 -> 155 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 152, 155 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5412, 5413 },
      -1,
   } },

   /* replace1915_0 -> 152 in the cache */
   /* replace1915_1_0 -> 722 in the cache */
   /* replace1915_1_1 -> 154 in the cache */
   /* replace1915_1_2 -> 155 in the cache */
   /* replace1915_1 -> 1252 in the cache */
   /* replace1915 -> 5411 in the cache */

   /* ('bcsel', 'a', ('fpow(is_used_once)', 'b', 'c'), ('fpow', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c') */
   /* search1916_0 -> 722 in the cache */
   /* search1916_1_0 -> 129 in the cache */
   /* search1916_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 129, 131 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1916_2_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 5416, 131 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5415, 5417 },
      -1,
   } },

   /* replace1916_0_0 -> 722 in the cache */
   /* replace1916_0_1 -> 129 in the cache */
   /* replace1916_0_2 -> 5416 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 129, 5416 },
      -1,
   } },
   /* replace1916_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 5419, 131 },
      -1,
   } },

   /* ('bcsel', 'a', ('fpow', 'b', 'c'), ('fpow(is_used_once)', 'd', 'c')) => ('fpow', ('bcsel', 'a', 'b', 'd'), 'c') */
   /* search1917_0 -> 722 in the cache */
   /* search1917_1_0 -> 129 in the cache */
   /* search1917_1_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 129, 131 },
      -1,
   } },
   /* search1917_2_0 -> 5416 in the cache */
   /* search1917_2_1 -> 131 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 5416, 131 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 722, 5421, 5422 },
      -1,
   } },

   /* replace1917_0_0 -> 722 in the cache */
   /* replace1917_0_1 -> 129 in the cache */
   /* replace1917_0_2 -> 5416 in the cache */
   /* replace1917_0 -> 5419 in the cache */
   /* replace1917_1 -> 131 in the cache */
   /* replace1917 -> 5420 in the cache */

   /* ('bcsel', 'c', ('frcp(is_used_once)', 'a'), ('frcp(is_used_once)', 'b')) => ('frcp', ('bcsel', 'c', 'a', 'b')) */
   /* search1918_0 -> 506 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 5424 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5425, 5427 },
      -1,
   } },

   /* replace1918_0_0 -> 506 in the cache */
   /* replace1918_0_1 -> 5424 in the cache */
   /* replace1918_0_2 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5424, 5426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('frsq(is_used_once)', 'a'), ('frsq(is_used_once)', 'b')) => ('frsq', ('bcsel', 'c', 'a', 'b')) */
   /* search1919_0 -> 506 in the cache */
   /* search1919_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1919_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5431, 5432 },
      -1,
   } },

   /* replace1919_0_0 -> 506 in the cache */
   /* replace1919_0_1 -> 5424 in the cache */
   /* replace1919_0_2 -> 5426 in the cache */
   /* replace1919_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsqrt(is_used_once)', 'a'), ('fsqrt(is_used_once)', 'b')) => ('fsqrt', ('bcsel', 'c', 'a', 'b')) */
   /* search1920_0 -> 506 in the cache */
   /* search1920_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1920_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5435, 5436 },
      -1,
   } },

   /* replace1920_0_0 -> 506 in the cache */
   /* replace1920_0_1 -> 5424 in the cache */
   /* replace1920_0_2 -> 5426 in the cache */
   /* replace1920_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fexp2(is_used_once)', 'a'), ('fexp2(is_used_once)', 'b')) => ('fexp2', ('bcsel', 'c', 'a', 'b')) */
   /* search1921_0 -> 506 in the cache */
   /* search1921_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1921_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5439, 5440 },
      -1,
   } },

   /* replace1921_0_0 -> 506 in the cache */
   /* replace1921_0_1 -> 5424 in the cache */
   /* replace1921_0_2 -> 5426 in the cache */
   /* replace1921_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('flog2(is_used_once)', 'a'), ('flog2(is_used_once)', 'b')) => ('flog2', ('bcsel', 'c', 'a', 'b')) */
   /* search1922_0 -> 506 in the cache */
   /* search1922_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1922_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5443, 5444 },
      -1,
   } },

   /* replace1922_0_0 -> 506 in the cache */
   /* replace1922_0_1 -> 5424 in the cache */
   /* replace1922_0_2 -> 5426 in the cache */
   /* replace1922_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b')) */
   /* search1923_0 -> 506 in the cache */
   /* search1923_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1923_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5447, 5448 },
      -1,
   } },

   /* replace1923_0_0 -> 506 in the cache */
   /* replace1923_0_1 -> 5424 in the cache */
   /* replace1923_0_2 -> 5426 in the cache */
   /* replace1923_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin(is_used_once)', 'a'), ('fsin(is_used_once)', 'b')) => ('fsin', ('bcsel', 'c', 'a', 'b')) */
   /* search1924_0 -> 506 in the cache */
   /* search1924_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1924_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5451, 5452 },
      -1,
   } },

   /* replace1924_0_0 -> 506 in the cache */
   /* replace1924_0_1 -> 5424 in the cache */
   /* replace1924_0_2 -> 5426 in the cache */
   /* replace1924_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos(is_used_once)', 'a'), ('fcos(is_used_once)', 'b')) => ('fcos', ('bcsel', 'c', 'a', 'b')) */
   /* search1925_0 -> 506 in the cache */
   /* search1925_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1925_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5455, 5456 },
      -1,
   } },

   /* replace1925_0_0 -> 506 in the cache */
   /* replace1925_0_1 -> 5424 in the cache */
   /* replace1925_0_2 -> 5426 in the cache */
   /* replace1925_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsin_amd(is_used_once)', 'a'), ('fsin_amd(is_used_once)', 'b')) => ('fsin_amd', ('bcsel', 'c', 'a', 'b')) */
   /* search1926_0 -> 506 in the cache */
   /* search1926_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1926_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5459, 5460 },
      -1,
   } },

   /* replace1926_0_0 -> 506 in the cache */
   /* replace1926_0_1 -> 5424 in the cache */
   /* replace1926_0_2 -> 5426 in the cache */
   /* replace1926_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fsin_amd,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fcos_amd(is_used_once)', 'a'), ('fcos_amd(is_used_once)', 'b')) => ('fcos_amd', ('bcsel', 'c', 'a', 'b')) */
   /* search1927_0 -> 506 in the cache */
   /* search1927_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1927_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5463, 5464 },
      -1,
   } },

   /* replace1927_0_0 -> 506 in the cache */
   /* replace1927_0_1 -> 5424 in the cache */
   /* replace1927_0_2 -> 5426 in the cache */
   /* replace1927_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fcos_amd,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fneg(is_used_once)', 'a'), ('fneg(is_used_once)', 'b')) => ('fneg', ('bcsel', 'c', 'a', 'b')) */
   /* search1928_0 -> 506 in the cache */
   /* search1928_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1928_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5467, 5468 },
      -1,
   } },

   /* replace1928_0_0 -> 506 in the cache */
   /* replace1928_0_1 -> 5424 in the cache */
   /* replace1928_0_2 -> 5426 in the cache */
   /* replace1928_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fabs(is_used_once)', 'a'), ('fabs(is_used_once)', 'b')) => ('fabs', ('bcsel', 'c', 'a', 'b')) */
   /* search1929_0 -> 506 in the cache */
   /* search1929_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5424 },
      0,
   } },
   /* search1929_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5426 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5471, 5472 },
      -1,
   } },

   /* replace1929_0_0 -> 506 in the cache */
   /* replace1929_0_1 -> 5424 in the cache */
   /* replace1929_0_2 -> 5426 in the cache */
   /* replace1929_0 -> 5429 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 5429 },
      -1,
   } },

   /* ('bcsel', 'c', ('fsign(is_used_once)', 'a'), ('fsign(is_used_once)', 'b')) => ('fsign', ('bcsel', 'c', 'a', 'b')) */
   /* search1930_0 -> 506 in the cache */
   /* search1930_1_0 -> 5424 in the cache */
   /* search1930_1 -> 5447 in the cache */
   /* search1930_2_0 -> 5426 in the cache */
   /* search1930_2 -> 5448 in the cache */
   /* search1930 -> 5449 in the cache */

   /* replace1930_0_0 -> 506 in the cache */
   /* replace1930_0_1 -> 5424 in the cache */
   /* replace1930_0_2 -> 5426 in the cache */
   /* replace1930_0 -> 5429 in the cache */
   /* replace1930 -> 5450 in the cache */

   /* ('ineg', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('ineg', 'a'), ('ineg', 'b')) */
   /* search1931_0_0 -> 506 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5475, 5476 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 5477 },
      -1,
   } },

   /* replace1931_0 -> 506 in the cache */
   /* replace1931_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 5424 },
      -1,
   } },
   /* replace1931_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 5426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5479, 5480 },
      -1,
   } },

   /* ('iabs', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('iabs', 'a'), ('iabs', 'b')) */
   /* search1932_0_0 -> 506 in the cache */
   /* search1932_0_1 -> 5475 in the cache */
   /* search1932_0_2 -> 5476 in the cache */
   /* search1932_0 -> 5477 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 5477 },
      -1,
   } },

   /* replace1932_0 -> 506 in the cache */
   /* replace1932_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 5424 },
      -1,
   } },
   /* replace1932_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 5426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5483, 5484 },
      -1,
   } },

   /* ('inot', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('inot', 'a'), ('inot', 'b')) */
   /* search1933_0_0 -> 506 in the cache */
   /* search1933_0_1 -> 5475 in the cache */
   /* search1933_0_2 -> 5476 in the cache */
   /* search1933_0 -> 5477 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 5477 },
      -1,
   } },

   /* replace1933_0 -> 506 in the cache */
   /* replace1933_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 5424 },
      -1,
   } },
   /* replace1933_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 5426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5487, 5488 },
      -1,
   } },

   /* ('isign', ('bcsel', 'c', '#a', '#b')) => ('bcsel', 'c', ('isign', 'a'), ('isign', 'b')) */
   /* search1934_0_0 -> 506 in the cache */
   /* search1934_0_1 -> 5475 in the cache */
   /* search1934_0_2 -> 5476 in the cache */
   /* search1934_0 -> 5477 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 5477 },
      -1,
   } },

   /* replace1934_0 -> 506 in the cache */
   /* replace1934_1_0 -> 5424 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 5424 },
      -1,
   } },
   /* replace1934_2_0 -> 5426 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isign,
      -1, 0,
      { 5426 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 506, 5491, 5492 },
      -1,
   } },

   /* ('fisnormal', 'a@16') => ('ult', 4095, ('iadd', ('ishl', 'a', 1), 2048)) */
   /* search1935_0 -> 11 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 11 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0xfff /* 4095 */ },
   } },
   /* replace1935_1_0_0 -> 11 in the cache */
   /* replace1935_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 11, 381 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x800 /* 2048 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5496, 5497 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 5495, 5498 },
      -1,
   } },

   /* ('fisnormal', 'a@32') => ('ult', 33554431, ('iadd', ('ishl', 'a', 1), 16777216)) */
   /* search1936_0 -> 16 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 16 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1ffffff /* 33554431 */ },
   } },
   /* replace1936_1_0_0 -> 16 in the cache */
   /* replace1936_1_0_1 -> 381 in the cache */
   /* replace1936_1_0 -> 953 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1000000 /* 16777216 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 953, 5502 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 5501, 5503 },
      -1,
   } },

   /* ('fisnormal', 'a@64') => ('ult', 18014398509481983, ('iadd', ('ishl', 'a', 1), 9007199254740992)) */
   /* search1937_0 -> 21 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisnormal,
      -1, 0,
      { 21 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x3fffffffffffff /* 18014398509481983 */ },
   } },
   /* replace1937_1_0_0 -> 21 in the cache */
   /* replace1937_1_0_1 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 21, 381 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x20000000000000 /* 9007199254740992 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 5507, 5508 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ult,
      -1, 1,
      { 5506, 5509 },
      -1,
   } },

   /* ('f2f16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2f16', 'a'), ('f2f16', 'b')) */
   /* search1938_0_0 -> 16 in the cache */
   /* search1938_0_1 -> 36 in the cache */
   /* search1938_0 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 2940 },
      -1,
   } },

   /* replace1938_0_0 -> 16 in the cache */
   /* replace1938_0 -> 3630 in the cache */
   /* replace1938_1_0 -> 36 in the cache */
   /* replace1938_1 -> 3627 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3630, 3627 },
      -1,
   } },

   /* ('f2fmp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('f2fmp', 'a'), ('f2fmp', 'b')) */
   /* search1939_0_0 -> 16 in the cache */
   /* search1939_0_1 -> 36 in the cache */
   /* search1939_0 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 2940 },
      -1,
   } },

   /* replace1939_0_0 -> 16 in the cache */
   /* replace1939_0 -> 1941 in the cache */
   /* replace1939_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 1941, 5514 },
      -1,
   } },

   /* ('u2u16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('u2u16', 'a'), ('u2u16', 'b')) */
   /* search1940_0_0 -> 16 in the cache */
   /* search1940_0_1 -> 36 in the cache */
   /* search1940_0 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 2940 },
      -1,
   } },

   /* replace1940_0_0 -> 16 in the cache */
   /* replace1940_0 -> 3037 in the cache */
   /* replace1940_1_0 -> 36 in the cache */
   /* replace1940_1 -> 4453 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3037, 4453 },
      -1,
   } },

   /* ('i2i16', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2i16', 'a'), ('i2i16', 'b')) */
   /* search1941_0_0 -> 16 in the cache */
   /* search1941_0_1 -> 36 in the cache */
   /* search1941_0 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 2940 },
      -1,
   } },

   /* replace1941_0_0 -> 16 in the cache */
   /* replace1941_0 -> 3939 in the cache */
   /* replace1941_1_0 -> 36 in the cache */
   /* replace1941_1 -> 4124 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 3939, 4124 },
      -1,
   } },

   /* ('i2imp', ('vec2', 'a@32', 'b@32')) => ('vec2', ('i2imp', 'a'), ('i2imp', 'b')) */
   /* search1942_0_0 -> 16 in the cache */
   /* search1942_0_1 -> 36 in the cache */
   /* search1942_0 -> 2940 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 2940 },
      -1,
   } },

   /* replace1942_0_0 -> 16 in the cache */
   /* replace1942_0 -> 1952 in the cache */
   /* replace1942_1_0 -> 36 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 1952, 5521 },
      -1,
   } },

   /* ('f2f16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c')) */
   /* search1943_0_0 -> 16 in the cache */
   /* search1943_0_1 -> 36 in the cache */
   /* search1943_0_2 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 16, 36, 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 5523 },
      -1,
   } },

   /* replace1943_0_0 -> 16 in the cache */
   /* replace1943_0 -> 3630 in the cache */
   /* replace1943_1_0 -> 36 in the cache */
   /* replace1943_1 -> 3627 in the cache */
   /* replace1943_2_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 3630, 3627, 5525 },
      -1,
   } },

   /* ('f2fmp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c')) */
   /* search1944_0_0 -> 16 in the cache */
   /* search1944_0_1 -> 36 in the cache */
   /* search1944_0_2 -> 137 in the cache */
   /* search1944_0 -> 5523 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 5523 },
      -1,
   } },

   /* replace1944_0_0 -> 16 in the cache */
   /* replace1944_0 -> 1941 in the cache */
   /* replace1944_1_0 -> 36 in the cache */
   /* replace1944_1 -> 5514 in the cache */
   /* replace1944_2_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 1941, 5514, 5528 },
      -1,
   } },

   /* ('u2u16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c')) */
   /* search1945_0_0 -> 16 in the cache */
   /* search1945_0_1 -> 36 in the cache */
   /* search1945_0_2 -> 137 in the cache */
   /* search1945_0 -> 5523 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5523 },
      -1,
   } },

   /* replace1945_0_0 -> 16 in the cache */
   /* replace1945_0 -> 3037 in the cache */
   /* replace1945_1_0 -> 36 in the cache */
   /* replace1945_1 -> 4453 in the cache */
   /* replace1945_2_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 3037, 4453, 5531 },
      -1,
   } },

   /* ('i2i16', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c')) */
   /* search1946_0_0 -> 16 in the cache */
   /* search1946_0_1 -> 36 in the cache */
   /* search1946_0_2 -> 137 in the cache */
   /* search1946_0 -> 5523 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5523 },
      -1,
   } },

   /* replace1946_0_0 -> 16 in the cache */
   /* replace1946_0 -> 3939 in the cache */
   /* replace1946_1_0 -> 36 in the cache */
   /* replace1946_1 -> 4124 in the cache */
   /* replace1946_2_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 3939, 4124, 5534 },
      -1,
   } },

   /* ('i2imp', ('vec3', 'a@32', 'b@32', 'c@32')) => ('vec3', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c')) */
   /* search1947_0_0 -> 16 in the cache */
   /* search1947_0_1 -> 36 in the cache */
   /* search1947_0_2 -> 137 in the cache */
   /* search1947_0 -> 5523 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 5523 },
      -1,
   } },

   /* replace1947_0_0 -> 16 in the cache */
   /* replace1947_0 -> 1952 in the cache */
   /* replace1947_1_0 -> 36 in the cache */
   /* replace1947_1 -> 5521 in the cache */
   /* replace1947_2_0 -> 137 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 137 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec3,
      -1, 0,
      { 1952, 5521, 5537 },
      -1,
   } },

   /* ('f2f16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2f16', 'a'), ('f2f16', 'b'), ('f2f16', 'c'), ('f2f16', 'd')) */
   /* search1948_0_0 -> 16 in the cache */
   /* search1948_0_1 -> 36 in the cache */
   /* search1948_0_2 -> 137 in the cache */
   /* search1948_0_3 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 16, 36, 137, 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 5539 },
      -1,
   } },

   /* replace1948_0_0 -> 16 in the cache */
   /* replace1948_0 -> 3630 in the cache */
   /* replace1948_1_0 -> 36 in the cache */
   /* replace1948_1 -> 3627 in the cache */
   /* replace1948_2_0 -> 137 in the cache */
   /* replace1948_2 -> 5525 in the cache */
   /* replace1948_3_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3630, 3627, 5525, 5541 },
      -1,
   } },

   /* ('f2fmp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'), ('f2fmp', 'd')) */
   /* search1949_0_0 -> 16 in the cache */
   /* search1949_0_1 -> 36 in the cache */
   /* search1949_0_2 -> 137 in the cache */
   /* search1949_0_3 -> 165 in the cache */
   /* search1949_0 -> 5539 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 5539 },
      -1,
   } },

   /* replace1949_0_0 -> 16 in the cache */
   /* replace1949_0 -> 1941 in the cache */
   /* replace1949_1_0 -> 36 in the cache */
   /* replace1949_1 -> 5514 in the cache */
   /* replace1949_2_0 -> 137 in the cache */
   /* replace1949_2 -> 5528 in the cache */
   /* replace1949_3_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 1941, 5514, 5528, 5544 },
      -1,
   } },

   /* ('u2u16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('u2u16', 'a'), ('u2u16', 'b'), ('u2u16', 'c'), ('u2u16', 'd')) */
   /* search1950_0_0 -> 16 in the cache */
   /* search1950_0_1 -> 36 in the cache */
   /* search1950_0_2 -> 137 in the cache */
   /* search1950_0_3 -> 165 in the cache */
   /* search1950_0 -> 5539 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 5539 },
      -1,
   } },

   /* replace1950_0_0 -> 16 in the cache */
   /* replace1950_0 -> 3037 in the cache */
   /* replace1950_1_0 -> 36 in the cache */
   /* replace1950_1 -> 4453 in the cache */
   /* replace1950_2_0 -> 137 in the cache */
   /* replace1950_2 -> 5531 in the cache */
   /* replace1950_3_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3037, 4453, 5531, 5547 },
      -1,
   } },

   /* ('i2i16', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2i16', 'a'), ('i2i16', 'b'), ('i2i16', 'c'), ('i2i16', 'd')) */
   /* search1951_0_0 -> 16 in the cache */
   /* search1951_0_1 -> 36 in the cache */
   /* search1951_0_2 -> 137 in the cache */
   /* search1951_0_3 -> 165 in the cache */
   /* search1951_0 -> 5539 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 5539 },
      -1,
   } },

   /* replace1951_0_0 -> 16 in the cache */
   /* replace1951_0 -> 3939 in the cache */
   /* replace1951_1_0 -> 36 in the cache */
   /* replace1951_1 -> 4124 in the cache */
   /* replace1951_2_0 -> 137 in the cache */
   /* replace1951_2 -> 5534 in the cache */
   /* replace1951_3_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 3939, 4124, 5534, 5550 },
      -1,
   } },

   /* ('i2imp', ('vec4', 'a@32', 'b@32', 'c@32', 'd@32')) => ('vec4', ('i2imp', 'a'), ('i2imp', 'b'), ('i2imp', 'c'), ('i2imp', 'd')) */
   /* search1952_0_0 -> 16 in the cache */
   /* search1952_0_1 -> 36 in the cache */
   /* search1952_0_2 -> 137 in the cache */
   /* search1952_0_3 -> 165 in the cache */
   /* search1952_0 -> 5539 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 5539 },
      -1,
   } },

   /* replace1952_0_0 -> 16 in the cache */
   /* replace1952_0 -> 1952 in the cache */
   /* replace1952_1_0 -> 36 in the cache */
   /* replace1952_1 -> 5521 in the cache */
   /* replace1952_2_0 -> 137 in the cache */
   /* replace1952_2 -> 5537 in the cache */
   /* replace1952_3_0 -> 165 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 165 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec4,
      -1, 0,
      { 1952, 5521, 5537, 5553 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_expression_cond[] = {
   (is_used_once),
   (is_only_used_as_float),
   (is_used_by_if),
   (is_not_used_by_if),
   (is_used_by_non_fsat),
   (no_signed_wrap),
   (only_lower_8_bits_used),
   (only_lower_16_bits_used),
};

static const nir_search_variable_cond nir_opt_algebraic_variable_cond[] = {
   (is_pos_power_of_two),
   (is_neg_power_of_two),
   (is_bitcount2),
   (is_finite_not_zero),
   (is_finite),
   (is_not_const),
   (is_not_const_zero),
   (is_ult_0xfffc07fc),
   (is_not_fmul),
   (is_first_5_bits_uge_2),
   (is_unsigned_multiple_of_4),
   (is_unsigned_multiple_of_2),
   (is_unsigned_multiple_of_8),
   (is_unsigned_multiple_of_16),
   (is_unsigned_multiple_of_32),
   (is_unsigned_multiple_of_64),
   (is_a_number),
   (is_gt_0_and_lt_1),
   (is_zero_to_one),
   (is_not_negative),
   (is_integral),
   (is_not_positive),
   (is_not_zero),
   (is_a_number_not_negative),
   (is_a_number_not_positive),
   (is_gt_zero),
   (is_lt_zero),
   (is_a_number_gt_zero),
   (is_a_number_lt_zero),
   (is_fmul),
   (is_lower_half_zero),
   (is_upper_half_zero),
};

static const struct transform nir_opt_algebraic_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2508, 1184, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2944, 3034, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 52, 54, 4 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 57, 4 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 58, 32, 0 },
   { 67, 68, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 32, 0 },
   { 69, 73, 1 },
   { 74, 77, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 61, 0 },
   { 78, 81, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 62, 61, 0 },
   { 64, 61, 0 },
   { 82, 81, 1 },
   { 83, 88, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 61, 0 },
   { 66, 61, 0 },
   { 89, 95, 1 },
   { 96, 97, 1 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2885, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 2772, 0 },
   { 2773, 2774, 0 },
   { 2775, 1188, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3415, 3428, 74 },
   { 3707, 3718, 105 },
   { 3727, 3738, 105 },
   { 3747, 3757, 105 },
   { 3415, 3772, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 120, 32, 0 },
   { 3415, 3428, 74 },
   { 3707, 3718, 105 },
   { 3727, 3738, 105 },
   { 3747, 3757, 105 },
   { 3415, 3772, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3429, 3437, 74 },
   { 3719, 3726, 105 },
   { 3739, 3746, 105 },
   { 3758, 3765, 105 },
   { 3429, 3777, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 121, 32, 0 },
   { 3429, 3437, 74 },
   { 3719, 3726, 105 },
   { 3739, 3746, 105 },
   { 3758, 3765, 105 },
   { 3429, 3777, 105 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3401, 3405, 71 },
   { 3406, 3408, 72 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 122, 32, 0 },
   { 3401, 3405, 71 },
   { 3406, 3408, 72 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3409, 3410, 72 },
   { 3411, 3414, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 32, 0 },
   { 3409, 3410, 72 },
   { 3411, 3414, 73 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 125, 16, 0 },
   { 127, 126, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 213, 117, 0 },
   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 233, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 273, 274, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 273, 274, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 257, 258, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 257, 258, 6 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 276, 277, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 257, 258, 6 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 276, 277, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 228, 124, 0 },
   { 229, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 239, 3643, 97 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 670, 672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 670, 672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 281, 253, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 670, 672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 281, 253, 0 },
   { 670, 672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 320, 396, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 320, 396, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 286, 36, 0 },
   { 320, 396, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 286, 36, 0 },
   { 320, 396, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 286, 36, 0 },
   { 337, 339, 0 },
   { 320, 396, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 286, 36, 0 },
   { 337, 339, 0 },
   { 320, 396, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 323, 412, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 323, 412, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 36, 0 },
   { 323, 412, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 36, 0 },
   { 323, 412, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 36, 0 },
   { 340, 342, 0 },
   { 323, 412, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 287, 36, 0 },
   { 340, 342, 0 },
   { 323, 412, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 441, 442, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 288, 36, 0 },
   { 441, 442, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 288, 36, 0 },
   { 352, 353, 7 },
   { 441, 442, 10 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 439, 440, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 438, 323, 0 },
   { 439, 440, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 36, 0 },
   { 439, 440, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 36, 0 },
   { 438, 323, 0 },
   { 439, 440, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 36, 0 },
   { 354, 355, 8 },
   { 439, 440, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 289, 36, 0 },
   { 354, 355, 8 },
   { 438, 323, 0 },
   { 439, 440, 11 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 332, 428, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 332, 428, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 290, 36, 0 },
   { 332, 428, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 290, 36, 0 },
   { 332, 428, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 290, 36, 0 },
   { 346, 348, 0 },
   { 332, 428, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 290, 36, 0 },
   { 346, 348, 0 },
   { 332, 428, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 335, 436, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 335, 436, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 36, 0 },
   { 335, 436, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 36, 0 },
   { 335, 436, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 36, 0 },
   { 349, 351, 0 },
   { 335, 436, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 291, 36, 0 },
   { 349, 351, 0 },
   { 335, 436, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 447, 448, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 292, 36, 0 },
   { 447, 448, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 292, 36, 0 },
   { 358, 359, 7 },
   { 447, 448, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 445, 446, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 36, 0 },
   { 445, 446, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 293, 36, 0 },
   { 360, 361, 8 },
   { 445, 446, 13 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 294, 36, 0 },
   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 294, 36, 0 },
   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 295, 36, 0 },
   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 295, 36, 0 },
   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 294, 36, 0 },
   { 295, 36, 0 },
   { 343, 345, 0 },
   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 294, 36, 0 },
   { 295, 36, 0 },
   { 343, 345, 0 },
   { 326, 420, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 443, 444, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 296, 36, 0 },
   { 443, 444, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 297, 36, 0 },
   { 443, 444, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 296, 36, 0 },
   { 297, 36, 0 },
   { 356, 357, 7 },
   { 443, 444, 12 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 673, 677, 20 },
   { 2880, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 678, 680, 21 },
   { 681, 683, 21 },
   { 684, 686, 22 },
   { 2878, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 679, 693, 25 },
   { 682, 694, 25 },
   { 685, 695, 26 },
   { 2882, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 696, 698, 27 },
   { 2879, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 730, 735, 34 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 766, 767, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 764, 765, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 762, 763, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 758, 761, 0 },
   { 762, 763, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 768, 761, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2514, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2514, 1184, 0 },
   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2513, 32, 0 },
   { 2967, 2968, 57 },
   { 2969, 391, 57 },
   { 2970, 2971, 57 },
   { 3001, 424, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2510, 1184, 0 },
   { 2512, 2511, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2510, 1184, 0 },
   { 2512, 2511, 0 },
   { 2513, 32, 0 },
   { 2967, 2968, 57 },
   { 2969, 391, 57 },
   { 2970, 2971, 57 },
   { 3001, 424, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 2932, 1039, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1280, 61, 0 },
   { 1281, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1283, 32, 0 },
   { 1284, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3554, 3557, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3549, 3553, 84 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3549, 3553, 84 },
   { 4692, 11, 0 },
   { 4745, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3558, 3562, 85 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3558, 3562, 85 },
   { 4797, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3563, 3566, 85 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3563, 3566, 85 },
   { 4770, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 2328, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 2328, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 2328, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2320, 2328, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 2329, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 2329, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 2329, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2322, 2329, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 2330, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2331, 46 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2350, 2357, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2358, 2360, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2361, 2363, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2364, 2367, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2368, 2372, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2373, 2375, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2376, 2381, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2382, 2388, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2389, 2395, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2396, 2399, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2405, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2406, 2410, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2411, 2414, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2415, 2420, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2421, 2425, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2426, 2429, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2430, 2435, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2436, 2440, 47 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3188, 3189, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3188, 3189, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2520, 2523, 49 },
   { 2527, 2530, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2533, 2535, 49 },
   { 2538, 2540, 49 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2545, 2548, 50 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2545, 2548, 50 },
   { 2570, 32, 0 },
   { 2571, 240, 0 },
   { 2572, 2569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2545, 2548, 50 },
   { 2573, 2543, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2545, 2548, 50 },
   { 2570, 32, 0 },
   { 2571, 240, 0 },
   { 2572, 2569, 0 },
   { 2573, 2543, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2562, 2616, 53 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2611, 2613, 52 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2610, 2591, 0 },
   { 2611, 2613, 52 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2618, 2700, 55 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2701, 2770, 55 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2881, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2933, 3029, 60 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2945, 3036, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2953, 3037, 63 },
   { 2953, 3037, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2955, 3038, 63 },
   { 2955, 3038, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3495, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2989, 397, 57 },
   { 2990, 400, 57 },
   { 2991, 404, 57 },
   { 2992, 407, 57 },
   { 3005, 429, 58 },
   { 3006, 432, 58 },
   { 3495, 124, 0 },
   { 3497, 3499, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3496, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2993, 378, 57 },
   { 2994, 382, 57 },
   { 2995, 387, 57 },
   { 2996, 391, 57 },
   { 3007, 421, 58 },
   { 3008, 424, 58 },
   { 3496, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3030, 3033, 61 },
   { 3030, 3635, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3040, 3633, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3047, 3634, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3075, 3632, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3075, 3632, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3075, 3632, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3075, 3632, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3184, 16, 0 },
   { 3185, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3186, 3187, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3190, 3192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3193, 3194, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3545, 3548, 83 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3530, 3534, 81 },
   { 3530, 3538, 82 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3540, 3544, 82 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3345, 3348, 64 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3349, 3352, 64 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3353, 3356, 65 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3357, 3358, 66 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3363, 3367, 67 },
   { 3363, 3467, 77 },
   { 3363, 3469, 78 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3368, 3371, 68 },
   { 3381, 3385, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3372, 3374, 68 },
   { 3386, 3388, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3375, 3377, 68 },
   { 3389, 3393, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3378, 3380, 68 },
   { 3394, 3396, 69 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3397, 3398, 70 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3399, 3400, 70 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3473, 3476, 79 },
   { 3473, 3523, 80 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3477, 3479, 79 },
   { 3477, 3529, 80 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3480, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3568, 3574, 86 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3575, 3580, 87 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3581, 3588, 88 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3589, 3594, 89 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3595, 3600, 90 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3601, 3608, 91 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3609, 3616, 92 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3617, 3626, 93 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3636, 3638, 95 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3644, 913, 98 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3645, 3649, 99 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3650, 3654, 100 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3655, 3656, 101 },
   { 3655, 49, 98 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3657, 3650, 102 },
   { 3657, 3659, 100 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3660, 3645, 103 },
   { 3660, 49, 99 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3661, 3656, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3662, 540, 0 },
   { 3663, 3664, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3661, 3656, 0 },
   { 3662, 540, 0 },
   { 3663, 3664, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3662, 540, 0 },
   { 3663, 3664, 0 },
   { 3665, 3667, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3661, 3656, 0 },
   { 3662, 540, 0 },
   { 3663, 3664, 0 },
   { 3665, 3667, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3669, 3671, 1 },
   { 3673, 3677, 1 },
   { 3678, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3829, 722, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3830, 6, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3831, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3832, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4823, 4843, 108 },
   { 4845, 4863, 108 },
   { 4865, 4886, 108 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5205, 113, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5206, 113, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5207, 113, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5208, 113, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5209, 113, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5210, 113, 110 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5212, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5215, 137, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5213, 137, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5213, 137, 0 },
   { 5215, 137, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5494, 5499, 111 },
   { 5500, 5504, 111 },
   { 5505, 5510, 111 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4681, 4679, 0 },
   { 4734, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4706, 4679, 0 },
   { 4759, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 3220, 3221, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 3220, 3221, 0 },
   { 3264, 3221, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 3203, 3204, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 986, 989, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 5201, 5204, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4681, 4679, 0 },
   { 4734, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4706, 4679, 0 },
   { 4759, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4681, 4679, 0 },
   { 4734, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4706, 4679, 0 },
   { 4759, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2459, 2460, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4681, 4679, 0 },
   { 4706, 4679, 0 },
   { 4734, 49, 0 },
   { 4759, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4681, 4679, 0 },
   { 4734, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4681, 4679, 0 },
   { 4734, 49, 0 },
   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4706, 4679, 0 },
   { 4759, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4706, 4679, 0 },
   { 4759, 49, 0 },
   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4786, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3310, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 32, 0 },
   { 69, 73, 1 },
   { 74, 77, 1 },
   { 3310, 0, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 195, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 3250, 3251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 3250, 3251, 0 },
   { 3292, 3251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 301, 303, 0 },
   { 319, 320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 305, 306, 0 },
   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 312, 313, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 315, 316, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 308, 310, 0 },
   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 5197, 5200, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 181, 183, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 198, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 198, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 319, 320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4703, 3729, 0 },
   { 4731, 3749, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4783, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3074, 3075, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2603, 2605, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3702, 3684, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2603, 2605, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3338, 103, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2780, 722, 0 },
   { 2781, 1087, 0 },
   { 2784, 2785, 0 },
   { 2786, 2787, 0 },
   { 2790, 2791, 0 },
   { 2792, 2793, 0 },
   { 2794, 2795, 0 },
   { 2796, 2797, 0 },
   { 2799, 2800, 0 },
   { 2801, 2802, 0 },
   { 2803, 2804, 56 },
   { 2805, 2806, 56 },
   { 2809, 2810, 56 },
   { 2811, 2812, 56 },
   { 2813, 3, 0 },
   { 2815, 2198, 0 },
   { 2816, 2818, 0 },
   { 2820, 1183, 0 },
   { 2821, 2817, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5280, 5281, 0 },
   { 5284, 5281, 0 },
   { 5287, 5281, 0 },
   { 5290, 5281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5469, 5470, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5473, 5474, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5267, 5268, 0 },
   { 5271, 5268, 0 },
   { 5274, 5268, 0 },
   { 5277, 5268, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5254, 5255, 0 },
   { 5258, 5255, 0 },
   { 5261, 5255, 0 },
   { 5264, 5255, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5239, 5241, 0 },
   { 5244, 5241, 0 },
   { 5247, 5241, 0 },
   { 5250, 5241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5293, 5294, 0 },
   { 5297, 5294, 0 },
   { 5300, 5294, 0 },
   { 5303, 5294, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5306, 5307, 0 },
   { 5310, 5307, 0 },
   { 5313, 5307, 0 },
   { 5316, 5307, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5449, 5450, 0 },
   { 5449, 5450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5218, 5219, 0 },
   { 5222, 5219, 0 },
   { 5224, 5226, 0 },
   { 5228, 5226, 0 },
   { 5231, 5233, 0 },
   { 5235, 5233, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1250, 1251, 0 },
   { 1253, 1251, 0 },
   { 1256, 1258, 0 },
   { 1261, 1258, 0 },
   { 1265, 1267, 0 },
   { 1270, 1267, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3339, 3340, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1721, 1723, 0 },
   { 1828, 1830, 0 },
   { 1911, 1913, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5332, 5333, 0 },
   { 5336, 5333, 0 },
   { 5339, 5333, 0 },
   { 5342, 5333, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5345, 5346, 0 },
   { 5349, 5346, 0 },
   { 5352, 5346, 0 },
   { 5355, 5346, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5358, 5359, 0 },
   { 5362, 5359, 0 },
   { 5365, 5359, 0 },
   { 5368, 5359, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5371, 5372, 0 },
   { 5375, 5372, 0 },
   { 5378, 5372, 0 },
   { 5381, 5372, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5384, 5385, 0 },
   { 5388, 5385, 0 },
   { 5391, 5385, 0 },
   { 5394, 5385, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5397, 5398, 0 },
   { 5401, 5398, 0 },
   { 5404, 5398, 0 },
   { 5407, 5398, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4673, 4668, 0 },
   { 4726, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4698, 4668, 0 },
   { 4751, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4803, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4778, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5319, 5320, 0 },
   { 5323, 5320, 0 },
   { 5326, 5320, 0 },
   { 5329, 5320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5441, 5442, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5445, 5446, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5410, 5411, 0 },
   { 5414, 5411, 0 },
   { 5418, 5420, 0 },
   { 5423, 5420, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5437, 5438, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5428, 5430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5433, 5434, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5453, 5454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5457, 5458, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5461, 5462, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 5465, 5466, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3320, 3311, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3322, 3314, 0 },
   { 3325, 3314, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3323, 3317, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2080, 2081, 0 },
   { 2111, 2112, 0 },
   { 2141, 908, 0 },
   { 2171, 2172, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2080, 2081, 0 },
   { 2111, 2112, 0 },
   { 2141, 908, 0 },
   { 2171, 2172, 0 },
   { 2508, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 785, 787, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 789, 798, 0 },
   { 811, 819, 0 },
   { 832, 839, 0 },
   { 852, 860, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 982, 984, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 952, 953, 0 },
   { 959, 960, 0 },
   { 966, 967, 0 },
   { 975, 976, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 872, 875, 0 },
   { 880, 883, 0 },
   { 888, 891, 0 },
   { 896, 899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 872, 875, 0 },
   { 880, 883, 0 },
   { 888, 891, 0 },
   { 896, 899, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 872, 875, 0 },
   { 880, 883, 0 },
   { 888, 891, 0 },
   { 896, 899, 0 },
   { 2080, 2081, 0 },
   { 2111, 2112, 0 },
   { 2141, 908, 0 },
   { 2171, 2172, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 872, 875, 0 },
   { 880, 883, 0 },
   { 888, 891, 0 },
   { 896, 899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 876, 875, 0 },
   { 884, 883, 0 },
   { 892, 891, 0 },
   { 900, 899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 876, 875, 0 },
   { 884, 883, 0 },
   { 892, 891, 0 },
   { 900, 899, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 876, 875, 0 },
   { 884, 883, 0 },
   { 892, 891, 0 },
   { 900, 899, 0 },
   { 2080, 2081, 0 },
   { 2111, 2112, 0 },
   { 2141, 908, 0 },
   { 2171, 2172, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 876, 875, 0 },
   { 884, 883, 0 },
   { 892, 891, 0 },
   { 900, 899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { 3069, 3070, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 912, 913, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 907, 908, 0 },
   { 954, 953, 0 },
   { 961, 960, 0 },
   { 969, 967, 0 },
   { 978, 976, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 917, 919, 0 },
   { 923, 925, 0 },
   { 929, 931, 0 },
   { 935, 937, 0 },
   { 941, 943, 0 },
   { 947, 949, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2500, 2501, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2500, 2501, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 3294, 3296, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2499, 3, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 956, 955, 0 },
   { 963, 962, 0 },
   { 971, 970, 0 },
   { 980, 979, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2009, 2010, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 2500, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 186, 188, 0 },
   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 5107, 5108, 0 },
   { 5110, 5111, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 5096, 5097, 0 },
   { 5100, 5101, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 5102, 5103, 0 },
   { 5105, 5106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1038, 1039, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2461, 2460, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2000, 2002, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2005, 2006, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2504, 2505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 1588, 1574, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 1698, 1699, 0 },
   { 1700, 1699, 0 },
   { 1805, 1806, 0 },
   { 1807, 1806, 0 },
   { 1888, 1889, 0 },
   { 1890, 1889, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 1588, 1574, 0 },
   { 1698, 1699, 0 },
   { 1700, 1699, 0 },
   { 1805, 1806, 0 },
   { 1807, 1806, 0 },
   { 1888, 1889, 0 },
   { 1890, 1889, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 1593, 1577, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 1593, 1577, 0 },
   { 1698, 1699, 0 },
   { 1700, 1699, 0 },
   { 1805, 1806, 0 },
   { 1807, 1806, 0 },
   { 1888, 1889, 0 },
   { 1890, 1889, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1041, 1039, 0 },
   { 1572, 1574, 0 },
   { 1575, 1577, 0 },
   { 1588, 1574, 0 },
   { 1593, 1577, 0 },
   { 1698, 1699, 0 },
   { 1700, 1699, 0 },
   { 1805, 1806, 0 },
   { 1807, 1806, 0 },
   { 1888, 1889, 0 },
   { 1890, 1889, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 1596, 1580, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1216, 1217, 37 },
   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 1701, 1702, 0 },
   { 1703, 1702, 0 },
   { 1808, 1809, 0 },
   { 1810, 1809, 0 },
   { 1891, 1892, 0 },
   { 1893, 1892, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1216, 1217, 37 },
   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 1596, 1580, 0 },
   { 1701, 1702, 0 },
   { 1703, 1702, 0 },
   { 1808, 1809, 0 },
   { 1810, 1809, 0 },
   { 1891, 1892, 0 },
   { 1893, 1892, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 1599, 1583, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1216, 1217, 37 },
   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 1599, 1583, 0 },
   { 1701, 1702, 0 },
   { 1703, 1702, 0 },
   { 1808, 1809, 0 },
   { 1810, 1809, 0 },
   { 1891, 1892, 0 },
   { 1893, 1892, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1216, 1217, 37 },
   { 1578, 1580, 0 },
   { 1581, 1583, 0 },
   { 1596, 1580, 0 },
   { 1599, 1583, 0 },
   { 1701, 1702, 0 },
   { 1703, 1702, 0 },
   { 1808, 1809, 0 },
   { 1810, 1809, 0 },
   { 1891, 1892, 0 },
   { 1893, 1892, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1046, 1039, 0 },
   { 1640, 1641, 0 },
   { 1642, 1643, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1048, 1039, 0 },
   { 1648, 1649, 0 },
   { 1650, 1651, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1644, 1645, 0 },
   { 1646, 1647, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1652, 1653, 0 },
   { 1654, 1655, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1044, 1039, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2062, 2064, 42 },
   { 2062, 2070, 44 },
   { 2093, 2095, 42 },
   { 2093, 2101, 44 },
   { 2123, 2125, 42 },
   { 2123, 2131, 44 },
   { 2153, 2155, 42 },
   { 2153, 2161, 44 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2074, 2075, 45 },
   { 2105, 2106, 45 },
   { 2135, 2136, 45 },
   { 2165, 2166, 45 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1988, 1989, 0 },
   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4709, 1971, 0 },
   { 4737, 1979, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1983, 1985, 0 },
   { 2484, 32, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1992, 1993, 0 },
   { 2484, 32, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4789, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1996, 1997, 0 },
   { 2484, 32, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 2498, 2262, 0 },
   { 3298, 3300, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2033, 2034, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2462, 2463, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2028, 2030, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2031, 2032, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 2506, 2507, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1681, 1014, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1560, 1541, 0 },
   { 1681, 1014, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1681, 1014, 0 },
   { 1684, 1686, 0 },
   { 1689, 1686, 0 },
   { 1791, 1793, 0 },
   { 1796, 1793, 0 },
   { 1874, 1876, 0 },
   { 1879, 1876, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1560, 1541, 0 },
   { 1681, 1014, 0 },
   { 1684, 1686, 0 },
   { 1689, 1686, 0 },
   { 1791, 1793, 0 },
   { 1796, 1793, 0 },
   { 1874, 1876, 0 },
   { 1879, 1876, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1565, 1547, 0 },
   { 1681, 1014, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1565, 1547, 0 },
   { 1681, 1014, 0 },
   { 1684, 1686, 0 },
   { 1689, 1686, 0 },
   { 1791, 1793, 0 },
   { 1796, 1793, 0 },
   { 1874, 1876, 0 },
   { 1879, 1876, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1539, 1541, 0 },
   { 1545, 1547, 0 },
   { 1560, 1541, 0 },
   { 1565, 1547, 0 },
   { 1681, 1014, 0 },
   { 1684, 1686, 0 },
   { 1689, 1686, 0 },
   { 1791, 1793, 0 },
   { 1796, 1793, 0 },
   { 1874, 1876, 0 },
   { 1879, 1876, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 1568, 1552, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 1692, 1694, 0 },
   { 1697, 1694, 0 },
   { 1799, 1801, 0 },
   { 1804, 1801, 0 },
   { 1882, 1884, 0 },
   { 1887, 1884, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 1568, 1552, 0 },
   { 1692, 1694, 0 },
   { 1697, 1694, 0 },
   { 1799, 1801, 0 },
   { 1804, 1801, 0 },
   { 1882, 1884, 0 },
   { 1887, 1884, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 1571, 1557, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 1571, 1557, 0 },
   { 1692, 1694, 0 },
   { 1697, 1694, 0 },
   { 1799, 1801, 0 },
   { 1804, 1801, 0 },
   { 1882, 1884, 0 },
   { 1887, 1884, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1550, 1552, 0 },
   { 1555, 1557, 0 },
   { 1568, 1552, 0 },
   { 1571, 1557, 0 },
   { 1692, 1694, 0 },
   { 1697, 1694, 0 },
   { 1799, 1801, 0 },
   { 1804, 1801, 0 },
   { 1882, 1884, 0 },
   { 1887, 1884, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1602, 1604, 0 },
   { 1607, 1609, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1622, 1624, 0 },
   { 1627, 1629, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1612, 1614, 0 },
   { 1617, 1619, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1632, 1634, 0 },
   { 1637, 1639, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2188, 2189, 0 },
   { 2192, 2193, 0 },
   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2071, 2073, 45 },
   { 2102, 2104, 45 },
   { 2132, 2134, 45 },
   { 2162, 2164, 45 },
   { 2184, 2185, 0 },
   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2067, 2068, 43 },
   { 2067, 2076, 44 },
   { 2098, 2099, 43 },
   { 2098, 2107, 44 },
   { 2128, 2129, 43 },
   { 2128, 2137, 44 },
   { 2158, 2159, 43 },
   { 2158, 2167, 44 },
   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2022, 2023, 0 },
   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4712, 2013, 0 },
   { 4740, 2017, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4740, 2017, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2019, 2021, 0 },
   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4712, 2013, 0 },
   { 4765, 2017, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2024, 2025, 0 },
   { 2487, 32, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4792, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2026, 2027, 0 },
   { 2487, 32, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 3076, 3075, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4743, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4715, 2037, 0 },
   { 4768, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4690, 2037, 0 },
   { 4743, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4715, 2037, 0 },
   { 4768, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2493, 3, 0 },
   { 2494, 2196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2493, 3, 0 },
   { 2494, 2196, 0 },
   { 3302, 3304, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2057, 2058, 0 },
   { 2490, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2052, 2054, 0 },
   { 2490, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2055, 2056, 0 },
   { 2490, 61, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2046, 2047, 0 },
   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4743, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4715, 2037, 0 },
   { 4743, 2041, 0 },
   { 4768, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4743, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4743, 2041, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4690, 2037, 0 },
   { 4743, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2043, 2045, 0 },
   { 2490, 61, 0 },
   { 4715, 2037, 0 },
   { 4768, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4715, 2037, 0 },
   { 4768, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4715, 2037, 0 },
   { 4768, 2041, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4715, 2037, 0 },
   { 4768, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2048, 2049, 0 },
   { 2490, 61, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4795, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2050, 2051, 0 },
   { 2490, 61, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2084, 2085, 0 },
   { 2115, 2116, 0 },
   { 2144, 2145, 0 },
   { 2175, 2176, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2084, 2085, 0 },
   { 2115, 2116, 0 },
   { 2144, 2145, 0 },
   { 2175, 2176, 0 },
   { 2514, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2514, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 877, 879, 0 },
   { 885, 887, 0 },
   { 893, 895, 0 },
   { 901, 903, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 877, 879, 0 },
   { 885, 887, 0 },
   { 893, 895, 0 },
   { 901, 903, 0 },
   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 2999, 429, 58 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 877, 879, 0 },
   { 885, 887, 0 },
   { 893, 895, 0 },
   { 901, 903, 0 },
   { 2084, 2085, 0 },
   { 2115, 2116, 0 },
   { 2144, 2145, 0 },
   { 2175, 2176, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 877, 879, 0 },
   { 885, 887, 0 },
   { 893, 895, 0 },
   { 901, 903, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 800, 802, 0 },
   { 821, 823, 0 },
   { 841, 843, 0 },
   { 862, 864, 0 },
   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { 3071, 3068, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2082, 2083, 0 },
   { 2113, 2114, 0 },
   { 2142, 2143, 0 },
   { 2173, 2174, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2082, 2083, 0 },
   { 2113, 2114, 0 },
   { 2142, 2143, 0 },
   { 2173, 2174, 0 },
   { 2510, 1184, 0 },
   { 2512, 2511, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2513, 32, 0 },
   { 2967, 2968, 57 },
   { 2969, 391, 57 },
   { 2970, 2971, 57 },
   { 3000, 421, 58 },
   { 3001, 424, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2513, 32, 0 },
   { 2967, 2968, 57 },
   { 2969, 391, 57 },
   { 2970, 2971, 57 },
   { 3001, 424, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2082, 2083, 0 },
   { 2113, 2114, 0 },
   { 2142, 2143, 0 },
   { 2173, 2174, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 804, 809, 0 },
   { 825, 830, 0 },
   { 845, 850, 0 },
   { 866, 871, 0 },
   { 2513, 32, 0 },
   { 2967, 2968, 57 },
   { 2969, 391, 57 },
   { 2970, 2971, 57 },
   { 3001, 424, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3144, 3145, 0 },
   { 3147, 404, 0 },
   { 3148, 400, 0 },
   { 3149, 397, 0 },
   { 3151, 3122, 0 },
   { 3153, 3117, 0 },
   { 3155, 3112, 0 },
   { 3157, 3109, 0 },
   { 3159, 3105, 0 },
   { 3161, 3101, 0 },
   { 3163, 3164, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3094, 2962, 0 },
   { 3096, 400, 0 },
   { 3097, 404, 0 },
   { 3098, 407, 0 },
   { 3100, 3101, 0 },
   { 3103, 3105, 0 },
   { 3107, 3109, 0 },
   { 3110, 3112, 0 },
   { 3115, 3117, 0 },
   { 3120, 3122, 0 },
   { 3123, 2966, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3124, 2962, 0 },
   { 3126, 400, 0 },
   { 3127, 404, 0 },
   { 3128, 407, 0 },
   { 3130, 3101, 0 },
   { 3132, 3105, 0 },
   { 3134, 3109, 0 },
   { 3136, 3112, 0 },
   { 3138, 3117, 0 },
   { 3140, 3122, 0 },
   { 3141, 2966, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2998, 2982, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2997, 2982, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3142, 2976, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3086, 3087, 0 },
   { 3088, 3089, 0 },
   { 3090, 3091, 0 },
   { 3092, 3093, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3165, 3166, 0 },
   { 3167, 387, 0 },
   { 3168, 382, 0 },
   { 3169, 378, 0 },
   { 3170, 3171, 0 },
   { 3172, 3173, 0 },
   { 3174, 3175, 0 },
   { 3176, 3177, 0 },
   { 3178, 3179, 0 },
   { 3180, 3181, 0 },
   { 3182, 3183, 0 },
   { 3549, 3553, 84 },
   { 4692, 11, 0 },
   { 4745, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3078, 3079, 0 },
   { 3080, 3081, 0 },
   { 3082, 3083, 0 },
   { 3084, 3085, 0 },
   { 3549, 3553, 84 },
   { 4692, 11, 0 },
   { 4745, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2946, 21, 0 },
   { 2933, 3029, 60 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 48, 49, 0 },
   { 2944, 3034, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 51, 49, 0 },
   { 2944, 3034, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2944, 3034, 62 },
   { 3044, 2944, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2934, 16, 0 },
   { 2944, 3034, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2937, 2351, 0 },
   { 2944, 3034, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 99, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 102, 103, 0 },
   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 105, 104, 0 },
   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2321, 2320, 0 },
   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2323, 2322, 0 },
   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2325, 2324, 0 },
   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2327, 2326, 0 },
   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2884, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 109, 110, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3799, 722, 0 },
   { 3802, 6, 0 },
   { 3806, 11, 0 },
   { 3808, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 3235, 3237, 0 },
   { 3239, 3237, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 3235, 3237, 0 },
   { 3239, 3237, 0 },
   { 3266, 3237, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 3242, 3244, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 3242, 3244, 0 },
   { 3273, 3275, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 3246, 3248, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 3246, 3248, 0 },
   { 3283, 3285, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 5184, 5187, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 203, 3, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 157, 159, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 139, 141, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 173, 168, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 3208, 3210, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 3208, 3210, 0 },
   { 3253, 3210, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 3216, 3218, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 3216, 3218, 0 },
   { 3257, 3218, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 5188, 5191, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { 5188, 5191, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2464, 1223, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2601, 2602, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2606, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 3197, 3198, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 3333, 3334, 0 },
   { 3336, 3337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 213, 117, 0 },
   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 233, 16, 0 },
   { 3197, 3198, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 213, 117, 0 },
   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 233, 16, 0 },
   { 3212, 3214, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 213, 117, 0 },
   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 233, 16, 0 },
   { 3212, 3214, 0 },
   { 3255, 3214, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 213, 117, 0 },
   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 233, 16, 0 },
   { 5193, 5196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 213, 117, 0 },
   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 233, 16, 0 },
   { 3333, 3334, 0 },
   { 3336, 3337, 0 },
   { 5193, 5196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 3197, 3198, 6 },
   { 3333, 3334, 0 },
   { 3336, 3337, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 215, 216, 6 },
   { 219, 216, 0 },
   { 220, 221, 0 },
   { 2608, 2609, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 162, 164, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 273, 274, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 273, 274, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 162, 164, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { 3260, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { 3260, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 162, 164, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { 3260, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3226, 3228, 0 },
   { 3260, 3228, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3231, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3231, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3231, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3231, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3200, 3201, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3200, 3201, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3200, 3201, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { 3200, 3201, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 162, 164, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 168, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 168, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 176, 178, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 276, 277, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 168, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 276, 277, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 257, 258, 6 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 176, 178, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 168, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { 3261, 3230, 0 },
   { 3262, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 257, 258, 6 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { 3261, 3230, 0 },
   { 3262, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 176, 178, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { 3261, 3230, 0 },
   { 3262, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 168, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 250, 251, 0 },
   { 262, 263, 6 },
   { 268, 270, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3229, 3230, 0 },
   { 3261, 3230, 0 },
   { 3262, 3232, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3202, 377, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 257, 258, 6 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3202, 377, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 141, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3202, 377, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 167, 168, 0 },
   { 222, 223, 6 },
   { 224, 223, 0 },
   { 225, 226, 0 },
   { 377, 277, 29 },
   { 710, 277, 32 },
   { 3202, 377, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 1474, 206, 0 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1469, 1471, 0 },
   { 206, 1461, 39 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1473, 1472, 0 },
   { 206, 1461, 39 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1146, 1147, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1067, 1068, 0 },
   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1189, 1190, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1146, 1147, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1146, 1147, 0 },
   { 1165, 1167, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1078, 1017, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { 5113, 5116, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1181, 1087, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1200, 1190, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2443, 1011, 0 },
   { 2451, 1008, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2444, 1008, 0 },
   { 2452, 1011, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2441, 1017, 0 },
   { 2449, 1014, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2442, 1014, 0 },
   { 2450, 1017, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1053, 1054, 0 },
   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1146, 1147, 0 },
   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1119, 1118, 0 },
   { 1350, 1118, 0 },
   { 1354, 1118, 0 },
   { 1358, 1118, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2482, 1192, 0 },
   { 4600, 4602, 107 },
   { 4617, 4602, 107 },
   { 4625, 4626, 0 },
   { 4639, 4641, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1127, 674, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2229, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2223, 2224, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1059, 1060, 0 },
   { 1127, 674, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1209, 1210, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1196, 1197, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1072, 1074, 0 },
   { 1082, 1083, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { 5137, 5142, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2229, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1207, 1208, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2223, 2224, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1057, 1058, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1050, 1040, 0 },
   { 1127, 674, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1195, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1127, 674, 0 },
   { 1142, 1143, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 2229, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 2223, 2224, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 1151, 1154, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { 5129, 5132, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2225, 1040, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2225, 1040, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1127, 674, 0 },
   { 2225, 1040, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2225, 1040, 0 },
   { 2229, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2223, 2224, 0 },
   { 2225, 1040, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1205, 1206, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2230, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2230, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1127, 674, 0 },
   { 2230, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2229, 1039, 0 },
   { 2230, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2223, 2224, 0 },
   { 2230, 1039, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { 4607, 4608, 107 },
   { 4614, 4615, 107 },
   { 4620, 4608, 107 },
   { 4624, 4615, 107 },
   { 4631, 4632, 0 },
   { 4636, 4637, 0 },
   { 4646, 4647, 0 },
   { 4652, 4653, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1128, 1129, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2227, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2231, 2232, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1063, 1064, 0 },
   { 1128, 1129, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1213, 1214, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1193, 1194, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { 5147, 5150, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1086, 1087, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2227, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1203, 1204, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2231, 2232, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1061, 1062, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1051, 1052, 0 },
   { 1128, 1129, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1191, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1144, 1145, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1144, 1145, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1128, 1129, 0 },
   { 1144, 1145, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1144, 1145, 0 },
   { 2227, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1144, 1145, 0 },
   { 2231, 2232, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1144, 1145, 0 },
   { 1160, 1161, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1077, 1008, 0 },
   { 1081, 1064, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { 5143, 5146, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2233, 1052, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2233, 1052, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1128, 1129, 0 },
   { 2233, 1052, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2227, 1192, 0 },
   { 2233, 1052, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2231, 2232, 0 },
   { 2233, 1052, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1201, 1202, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2228, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2228, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1128, 1129, 0 },
   { 2228, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2227, 1192, 0 },
   { 2228, 1192, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2228, 1192, 0 },
   { 2231, 2232, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { 4605, 4606, 107 },
   { 4611, 4613, 107 },
   { 4619, 4606, 107 },
   { 4623, 4613, 107 },
   { 4629, 4630, 0 },
   { 4634, 4635, 0 },
   { 4644, 4645, 0 },
   { 4649, 4651, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1148, 1149, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1065, 1066, 0 },
   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1187, 1188, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1148, 1149, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1148, 1149, 0 },
   { 1173, 1174, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1079, 1014, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { 5117, 5120, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1182, 722, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1199, 1188, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2447, 1011, 0 },
   { 2455, 1008, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2448, 1008, 0 },
   { 2456, 1011, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2445, 1017, 0 },
   { 2453, 1014, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2446, 1014, 0 },
   { 2454, 1017, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1055, 1056, 0 },
   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1148, 1149, 0 },
   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1108, 1107, 0 },
   { 1351, 1335, 0 },
   { 1355, 1335, 0 },
   { 1359, 1335, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2483, 1039, 0 },
   { 4603, 4604, 107 },
   { 4618, 4604, 107 },
   { 4627, 4628, 0 },
   { 4642, 4643, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1451, 32, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1515, 1516, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1229, 1232, 0 },
   { 1275, 32, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1461, 206, 37 },
   { 1465, 1466, 37 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1225, 1223, 0 },
   { 1275, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1455, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1455, 103, 0 },
   { 1459, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1513, 1514, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1463, 206, 37 },
   { 1468, 1466, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1455, 103, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1455, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1219, 1220, 0 },
   { 1276, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 101, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5478, 5481, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2078, 6, 0 },
   { 2109, 11, 0 },
   { 2139, 16, 0 },
   { 2169, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 108, 106, 0 },
   { 2885, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 107, 106, 0 },
   { 2885, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2885, 32, 0 },
   { 5482, 5485, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 111, 112, 0 },
   { 2771, 2772, 0 },
   { 2773, 2774, 0 },
   { 2775, 1188, 0 },
   { 111, 112, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 2772, 0 },
   { 2773, 2774, 0 },
   { 2775, 1188, 0 },
   { 2849, 112, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 2772, 0 },
   { 2773, 2774, 0 },
   { 2775, 1188, 0 },
   { 2842, 11, 0 },
   { 2844, 16, 0 },
   { 3800, 722, 0 },
   { 3804, 6, 0 },
   { 2842, 11, 0 },
   { 2844, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 2772, 0 },
   { 2773, 2774, 0 },
   { 2775, 1188, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2771, 2772, 0 },
   { 2773, 2774, 0 },
   { 2775, 1188, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 3638, 3636, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1447, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1456, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1456, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 3640, 3636, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1456, 106, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1456, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1730, 11, 0 },
   { 1837, 16, 0 },
   { 1920, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3811, 2833, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2845, 2846, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1728, 11, 0 },
   { 1835, 16, 0 },
   { 1918, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1756, 1757, 0 },
   { 1779, 1780, 0 },
   { 1863, 1864, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1944, 1945, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1732, 11, 0 },
   { 1839, 16, 0 },
   { 1922, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3810, 2833, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2847, 2848, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1731, 11, 0 },
   { 1838, 16, 0 },
   { 1921, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1754, 1755, 0 },
   { 1777, 1778, 0 },
   { 1861, 1862, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1942, 1943, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1759, 1760, 0 },
   { 1782, 1760, 0 },
   { 1865, 1758, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1751, 1348, 0 },
   { 1774, 1348, 0 },
   { 1858, 1352, 0 },
   { 3809, 1472, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { 5540, 5542, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { 5524, 5526, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { 5511, 5512, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1762, 1763, 0 },
   { 1784, 1763, 0 },
   { 1866, 1761, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1746, 11, 0 },
   { 1769, 11, 0 },
   { 1853, 16, 0 },
   { 2876, 16, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2869, 16, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2874, 1938, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2875, 1935, 0 },
   { 3826, 11, 0 },
   { 3827, 16, 0 },
   { 3828, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1936, 1937, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2863, 1348, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5543, 5545, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5527, 5529, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5513, 5515, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1939, 1940, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2852, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2851, 1214, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5486, 5489, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1013, 1014, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2495, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1007, 1008, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1010, 1011, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1016, 1017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1019, 1020, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1022, 1023, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1025, 1026, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1028, 1029, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1031, 1032, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1034, 1035, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 451, 452, 14 },
   { 526, 527, 16 },
   { 592, 593, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 451, 452, 14 },
   { 526, 527, 16 },
   { 592, 593, 18 },
   { 670, 672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 283, 285, 0 },
   { 455, 457, 14 },
   { 528, 530, 16 },
   { 596, 598, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 283, 285, 0 },
   { 455, 457, 14 },
   { 528, 530, 16 },
   { 596, 598, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 283, 285, 0 },
   { 451, 452, 14 },
   { 455, 457, 14 },
   { 526, 527, 16 },
   { 528, 530, 16 },
   { 592, 593, 18 },
   { 596, 598, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 465, 468, 14 },
   { 537, 539, 16 },
   { 606, 609, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 465, 468, 14 },
   { 537, 539, 16 },
   { 606, 609, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 451, 452, 14 },
   { 465, 468, 14 },
   { 526, 527, 16 },
   { 537, 539, 16 },
   { 592, 593, 18 },
   { 606, 609, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 451, 452, 14 },
   { 526, 527, 16 },
   { 592, 593, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 451, 452, 14 },
   { 526, 527, 16 },
   { 592, 593, 18 },
   { 670, 672, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 283, 285, 0 },
   { 455, 457, 14 },
   { 528, 530, 16 },
   { 596, 598, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 281, 253, 0 },
   { 283, 285, 0 },
   { 455, 457, 14 },
   { 528, 530, 16 },
   { 596, 598, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 283, 285, 0 },
   { 451, 452, 14 },
   { 455, 457, 14 },
   { 526, 527, 16 },
   { 528, 530, 16 },
   { 592, 593, 18 },
   { 596, 598, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 465, 468, 14 },
   { 537, 539, 16 },
   { 606, 609, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 281, 253, 0 },
   { 465, 468, 14 },
   { 537, 539, 16 },
   { 606, 609, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 281, 253, 0 },
   { 451, 452, 14 },
   { 465, 468, 14 },
   { 526, 527, 16 },
   { 537, 539, 16 },
   { 592, 593, 18 },
   { 606, 609, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 283, 285, 0 },
   { 455, 457, 14 },
   { 461, 463, 14 },
   { 528, 530, 16 },
   { 533, 535, 16 },
   { 596, 598, 18 },
   { 602, 604, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 283, 285, 0 },
   { 455, 457, 14 },
   { 461, 463, 14 },
   { 528, 530, 16 },
   { 533, 535, 16 },
   { 596, 598, 18 },
   { 602, 604, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 283, 285, 0 },
   { 451, 452, 14 },
   { 455, 457, 14 },
   { 461, 463, 14 },
   { 526, 527, 16 },
   { 528, 530, 16 },
   { 533, 535, 16 },
   { 592, 593, 18 },
   { 596, 598, 18 },
   { 602, 604, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 465, 468, 14 },
   { 537, 539, 16 },
   { 606, 609, 18 },
   { 666, 668, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 0, 0 },
   { 279, 3, 0 },
   { 280, 0, 0 },
   { 465, 468, 14 },
   { 537, 539, 16 },
   { 606, 609, 18 },
   { 666, 668, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 280, 0, 0 },
   { 451, 452, 14 },
   { 465, 468, 14 },
   { 526, 527, 16 },
   { 537, 539, 16 },
   { 592, 593, 18 },
   { 606, 609, 18 },
   { 666, 668, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1177, 1178, 0 },
   { 2475, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1177, 1178, 0 },
   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1365, 1366, 0 },
   { 1370, 1369, 0 },
   { 1387, 1388, 0 },
   { 1392, 1391, 0 },
   { 1409, 1410, 0 },
   { 1414, 1413, 0 },
   { 1429, 1430, 0 },
   { 1434, 1433, 0 },
   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { 5121, 5124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1185, 1087, 0 },
   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1373, 1366, 0 },
   { 1375, 1369, 0 },
   { 1395, 1388, 0 },
   { 1397, 1391, 0 },
   { 1417, 1410, 0 },
   { 1418, 1413, 0 },
   { 1437, 1430, 0 },
   { 1439, 1433, 0 },
   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1378, 1366, 0 },
   { 1380, 1369, 0 },
   { 1400, 1388, 0 },
   { 1402, 1391, 0 },
   { 1421, 1410, 0 },
   { 1422, 1413, 0 },
   { 1442, 1430, 0 },
   { 1444, 1433, 0 },
   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { 3963, 3968, 106 },
   { 4010, 4015, 106 },
   { 4056, 4061, 106 },
   { 4123, 4128, 107 },
   { 4174, 4179, 107 },
   { 4252, 4257, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { 4306, 4309, 106 },
   { 4346, 4350, 106 },
   { 4388, 4393, 106 },
   { 4452, 4457, 107 },
   { 4497, 4502, 107 },
   { 4569, 4574, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { 3517, 3514, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { 3513, 3514, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1333, 1118, 0 },
   { 1338, 1118, 0 },
   { 1342, 1118, 0 },
   { 1346, 1118, 0 },
   { 2475, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 3945, 3947, 106 },
   { 4002, 3947, 106 },
   { 4047, 3947, 106 },
   { 4094, 4095, 107 },
   { 4108, 4110, 107 },
   { 4159, 4095, 107 },
   { 4167, 4110, 107 },
   { 4211, 4212, 0 },
   { 4224, 4225, 0 },
   { 4237, 4239, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 4290, 4292, 106 },
   { 4338, 4292, 106 },
   { 4380, 4292, 106 },
   { 4424, 4425, 107 },
   { 4437, 4439, 107 },
   { 4482, 4425, 107 },
   { 4490, 4439, 107 },
   { 4529, 4530, 0 },
   { 4542, 4543, 0 },
   { 4554, 4556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1179, 1180, 0 },
   { 2476, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1179, 1180, 0 },
   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1367, 1366, 0 },
   { 1368, 1369, 0 },
   { 1389, 1388, 0 },
   { 1390, 1391, 0 },
   { 1411, 1410, 0 },
   { 1412, 1413, 0 },
   { 1431, 1430, 0 },
   { 1432, 1433, 0 },
   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { 5125, 5128, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1186, 722, 0 },
   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1372, 1369, 0 },
   { 1376, 1366, 0 },
   { 1394, 1391, 0 },
   { 1398, 1388, 0 },
   { 1416, 1413, 0 },
   { 1419, 1410, 0 },
   { 1436, 1433, 0 },
   { 1440, 1430, 0 },
   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1377, 1369, 0 },
   { 1381, 1366, 0 },
   { 1399, 1391, 0 },
   { 1403, 1388, 0 },
   { 1420, 1413, 0 },
   { 1423, 1410, 0 },
   { 1441, 1433, 0 },
   { 1445, 1430, 0 },
   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { 3969, 3972, 106 },
   { 4016, 4019, 106 },
   { 4062, 4065, 106 },
   { 4129, 4132, 107 },
   { 4180, 4183, 107 },
   { 4258, 4261, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { 4310, 4313, 106 },
   { 4351, 4354, 106 },
   { 4394, 4397, 106 },
   { 4458, 4461, 107 },
   { 4503, 4506, 107 },
   { 4575, 4578, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { 3516, 3512, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { 3507, 3512, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1334, 1335, 0 },
   { 1339, 1335, 0 },
   { 1343, 1335, 0 },
   { 1347, 1335, 0 },
   { 2476, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 3948, 3949, 106 },
   { 4003, 3949, 106 },
   { 4048, 3949, 106 },
   { 4096, 4097, 107 },
   { 4111, 4112, 107 },
   { 4160, 4097, 107 },
   { 4168, 4112, 107 },
   { 4213, 4214, 0 },
   { 4226, 4227, 0 },
   { 4240, 4241, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 4293, 4294, 106 },
   { 4339, 4294, 106 },
   { 4381, 4294, 106 },
   { 4426, 4427, 107 },
   { 4440, 4441, 107 },
   { 4483, 4427, 107 },
   { 4491, 4441, 107 },
   { 4531, 4532, 0 },
   { 4544, 4545, 0 },
   { 4557, 4558, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2261, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2254, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2261, 1039, 0 },
   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { 5171, 5174, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2254, 1039, 0 },
   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2254, 1039, 0 },
   { 2306, 2307, 0 },
   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2283, 2285, 0 },
   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { 4320, 4325, 106 },
   { 4361, 4366, 106 },
   { 4405, 4411, 106 },
   { 4467, 4471, 107 },
   { 4513, 4518, 107 },
   { 4584, 4588, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2263, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2263, 1039, 0 },
   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2261, 1039, 0 },
   { 2263, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2254, 1039, 0 },
   { 2263, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2263, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { 5167, 5170, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2244, 1047, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2244, 1047, 0 },
   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2244, 1047, 0 },
   { 2261, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2244, 1047, 0 },
   { 2254, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2244, 1047, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2244, 1047, 0 },
   { 2287, 2289, 0 },
   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2255, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2255, 1039, 0 },
   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2255, 1039, 0 },
   { 2261, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2254, 1039, 0 },
   { 2255, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2242, 2243, 0 },
   { 2255, 1039, 0 },
   { 2477, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2255, 1039, 0 },
   { 2308, 2309, 0 },
   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2932, 1039, 0 },
   { 4314, 4319, 106 },
   { 4355, 4360, 106 },
   { 4398, 4404, 106 },
   { 4462, 4466, 107 },
   { 4507, 4512, 107 },
   { 4579, 4583, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 4297, 4298, 106 },
   { 4304, 4305, 106 },
   { 4341, 4298, 106 },
   { 4345, 4305, 106 },
   { 4383, 4298, 106 },
   { 4387, 4305, 106 },
   { 4430, 4431, 107 },
   { 4435, 4436, 107 },
   { 4444, 4445, 107 },
   { 4450, 4451, 107 },
   { 4485, 4431, 107 },
   { 4488, 4436, 107 },
   { 4493, 4445, 107 },
   { 4496, 4451, 107 },
   { 4535, 4536, 0 },
   { 4540, 4541, 0 },
   { 4548, 4549, 0 },
   { 4552, 4553, 0 },
   { 4561, 4562, 0 },
   { 4567, 4568, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2264, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2256, 2257, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2264, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { 5179, 5182, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 1192, 0 },
   { 2291, 2293, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2256, 2257, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2256, 2257, 0 },
   { 2310, 2311, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { 4332, 4337, 106 },
   { 4373, 4378, 106 },
   { 4418, 4423, 106 },
   { 4477, 4481, 107 },
   { 4524, 4528, 107 },
   { 4594, 4598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2265, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2264, 1192, 0 },
   { 2265, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 1192, 0 },
   { 2265, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2256, 2257, 0 },
   { 2265, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { 5175, 5178, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2258, 2259, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2258, 2259, 0 },
   { 2264, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 1192, 0 },
   { 2258, 2259, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2256, 2257, 0 },
   { 2258, 2259, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2258, 2259, 0 },
   { 2312, 2313, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1192, 0 },
   { 2264, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2246, 1192, 0 },
   { 2247, 1192, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1192, 0 },
   { 2256, 2257, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2247, 1192, 0 },
   { 2295, 2297, 0 },
   { 2478, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { 4326, 4331, 106 },
   { 4367, 4372, 106 },
   { 4412, 4417, 106 },
   { 4472, 4476, 107 },
   { 4519, 4523, 107 },
   { 4589, 4593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2478, 1192, 0 },
   { 4295, 4296, 106 },
   { 4301, 4303, 106 },
   { 4340, 4296, 106 },
   { 4344, 4303, 106 },
   { 4382, 4296, 106 },
   { 4386, 4303, 106 },
   { 4428, 4429, 107 },
   { 4433, 4434, 107 },
   { 4442, 4443, 107 },
   { 4447, 4449, 107 },
   { 4484, 4429, 107 },
   { 4487, 4434, 107 },
   { 4492, 4443, 107 },
   { 4495, 4449, 107 },
   { 4533, 4534, 0 },
   { 4538, 4539, 0 },
   { 4546, 4547, 0 },
   { 4550, 4551, 0 },
   { 4559, 4560, 0 },
   { 4564, 4566, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3482, 3483, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3480, 2501, 0 },
   { 3482, 3483, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 4654, 3034, 0 },
   { 4655, 3037, 0 },
   { 4657, 4515, 0 },
   { 4659, 1957, 0 },
   { 4660, 1973, 0 },
   { 4662, 4408, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1752, 1336, 0 },
   { 1775, 1336, 0 },
   { 1859, 1340, 0 },
   { 3815, 2833, 0 },
   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 5549, 5551, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 5533, 5535, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 5518, 5519, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1748, 11, 0 },
   { 1771, 11, 0 },
   { 1855, 16, 0 },
   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 3834, 3835, 0 },
   { 3842, 3835, 0 },
   { 3848, 3835, 0 },
   { 3854, 3835, 0 },
   { 3859, 3860, 0 },
   { 3863, 3833, 0 },
   { 3867, 3833, 0 },
   { 3871, 3833, 0 },
   { 3875, 3876, 0 },
   { 3879, 3876, 0 },
   { 3881, 3882, 0 },
   { 3885, 3841, 0 },
   { 3889, 3841, 0 },
   { 3893, 3894, 0 },
   { 3897, 3894, 0 },
   { 3899, 3900, 0 },
   { 3902, 3894, 0 },
   { 3904, 3900, 0 },
   { 3906, 1747, 0 },
   { 3908, 3847, 0 },
   { 3912, 3913, 0 },
   { 3916, 3913, 0 },
   { 3918, 3919, 0 },
   { 3922, 3913, 0 },
   { 3924, 3919, 0 },
   { 3926, 1770, 0 },
   { 3928, 3913, 0 },
   { 3930, 3919, 0 },
   { 3932, 1770, 0 },
   { 3934, 1854, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 3837, 3838, 0 },
   { 3844, 3838, 0 },
   { 3850, 3838, 0 },
   { 3856, 3838, 0 },
   { 3864, 3836, 0 },
   { 3868, 3836, 0 },
   { 3872, 3836, 0 },
   { 3886, 3843, 0 },
   { 3890, 3843, 0 },
   { 3909, 3849, 0 },
   { 4691, 11, 0 },
   { 4744, 16, 0 },
   { 4796, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2870, 16, 0 },
   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2872, 1949, 0 },
   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 3936, 3937, 0 },
   { 3936, 3937, 0 },
   { 3938, 3939, 0 },
   { 3938, 3939, 0 },
   { 3940, 3819, 0 },
   { 3940, 3819, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1753, 1336, 0 },
   { 1776, 1336, 0 },
   { 1860, 1340, 0 },
   { 3813, 2833, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 5546, 5548, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 5530, 5532, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 5516, 5517, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2979, 2981, 57 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 3839, 3835, 0 },
   { 3845, 3835, 0 },
   { 3851, 3835, 0 },
   { 3857, 3835, 0 },
   { 3865, 3833, 0 },
   { 3869, 3833, 0 },
   { 3873, 3833, 0 },
   { 3887, 3841, 0 },
   { 3891, 3841, 0 },
   { 3910, 3847, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1750, 11, 0 },
   { 1773, 11, 0 },
   { 1857, 16, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 3840, 3838, 0 },
   { 3846, 3838, 0 },
   { 3852, 3838, 0 },
   { 3858, 3838, 0 },
   { 3861, 3862, 0 },
   { 3866, 3836, 0 },
   { 3870, 3836, 0 },
   { 3874, 3836, 0 },
   { 3877, 3878, 0 },
   { 3880, 3878, 0 },
   { 3883, 3884, 0 },
   { 3888, 3843, 0 },
   { 3892, 3843, 0 },
   { 3895, 3896, 0 },
   { 3898, 3896, 0 },
   { 3901, 3009, 0 },
   { 3903, 3896, 0 },
   { 3905, 3009, 0 },
   { 3907, 1749, 0 },
   { 3911, 3849, 0 },
   { 3914, 3915, 0 },
   { 3917, 3915, 0 },
   { 3920, 3921, 0 },
   { 3923, 3915, 0 },
   { 3925, 3921, 0 },
   { 3927, 1772, 0 },
   { 3929, 3915, 0 },
   { 3931, 3921, 0 },
   { 3933, 1772, 0 },
   { 3935, 1856, 0 },
   { 4663, 11, 0 },
   { 4716, 16, 0 },
   { 4769, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2871, 16, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2975, 2977, 57 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2873, 1946, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 3941, 1973, 0 },
   { 3941, 1973, 0 },
   { 3942, 3037, 0 },
   { 3942, 3037, 0 },
   { 3943, 3824, 0 },
   { 3943, 3824, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2945, 3036, 62 },
   { 3051, 2945, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2935, 36, 0 },
   { 2945, 3036, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2938, 2354, 0 },
   { 2945, 3036, 62 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2953, 3037, 63 },
   { 3041, 2953, 0 },
   { 2953, 3037, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2953, 3037, 63 },
   { 3055, 2953, 0 },
   { 3056, 2955, 0 },
   { 2953, 3037, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2955, 3038, 63 },
   { 3048, 2955, 0 },
   { 2955, 3038, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3485, 3478, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3487, 3478, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3487, 3478, 0 },
   { 3495, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3485, 3478, 0 },
   { 3487, 3478, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3488, 3475, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3489, 3475, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3489, 3475, 0 },
   { 3496, 124, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3488, 3475, 0 },
   { 3489, 3475, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3039, 3040, 0 },
   { 3040, 3633, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3054, 3047, 0 },
   { 3040, 3633, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3052, 3040, 0 },
   { 3053, 3047, 0 },
   { 3040, 3633, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3046, 3047, 0 },
   { 3047, 3634, 94 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3491, 3492, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3494, 3492, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3491, 3492, 0 },
   { 3494, 3492, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2248, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2235, 2236, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 5155, 5158, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2248, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2248, 1039, 0 },
   { 2298, 2299, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2235, 2236, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2235, 2236, 0 },
   { 2267, 2269, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 3982, 3988, 106 },
   { 4028, 4033, 106 },
   { 4075, 4081, 106 },
   { 4141, 4146, 107 },
   { 4192, 4198, 107 },
   { 4270, 4276, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 1433, 3451, 76 },
   { 5151, 5154, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2237, 1045, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2237, 1045, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 3444, 3446, 75 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2237, 1045, 0 },
   { 2248, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2235, 2236, 0 },
   { 2237, 1045, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2237, 1045, 0 },
   { 2271, 2273, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 3444, 3446, 75 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2249, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2249, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 3439, 3442, 75 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2248, 1039, 0 },
   { 2249, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2235, 2236, 0 },
   { 2249, 1039, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2249, 1039, 0 },
   { 2300, 2301, 0 },
   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 3439, 3442, 75 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 1433, 3451, 76 },
   { 3973, 3980, 106 },
   { 4020, 4027, 106 },
   { 4066, 4073, 106 },
   { 4133, 4140, 107 },
   { 4184, 4191, 107 },
   { 4262, 4269, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 3952, 3953, 106 },
   { 3961, 3962, 106 },
   { 4005, 3953, 106 },
   { 4009, 3962, 106 },
   { 4050, 3953, 106 },
   { 4054, 3962, 106 },
   { 4100, 4101, 107 },
   { 4105, 4106, 107 },
   { 4115, 4116, 107 },
   { 4121, 4122, 107 },
   { 4162, 4101, 107 },
   { 4165, 4106, 107 },
   { 4170, 4116, 107 },
   { 4173, 4122, 107 },
   { 4217, 4218, 0 },
   { 4222, 4223, 0 },
   { 4230, 4231, 0 },
   { 4234, 4235, 0 },
   { 4244, 4245, 0 },
   { 4250, 4251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2239, 1192, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2250, 2251, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 5163, 5166, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2239, 1192, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2239, 1192, 0 },
   { 2275, 2277, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2250, 2251, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2250, 2251, 0 },
   { 2302, 2303, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 3995, 4000, 106 },
   { 4040, 4045, 106 },
   { 4088, 4093, 106 },
   { 4153, 4158, 107 },
   { 4205, 4210, 107 },
   { 4283, 4288, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { 5159, 5162, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 3449, 3450, 75 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2239, 1192, 0 },
   { 2252, 2253, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2250, 2251, 0 },
   { 2252, 2253, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2252, 2253, 0 },
   { 2304, 2305, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 3449, 3450, 75 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2240, 1192, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2240, 1192, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 3447, 3448, 75 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2239, 1192, 0 },
   { 2240, 1192, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2240, 1192, 0 },
   { 2250, 2251, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2240, 1192, 0 },
   { 2279, 2281, 0 },
   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 3447, 3448, 75 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { 3989, 3994, 106 },
   { 4034, 4039, 106 },
   { 4082, 4087, 106 },
   { 4147, 4152, 107 },
   { 4199, 4204, 107 },
   { 4277, 4282, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 3950, 3951, 106 },
   { 3958, 3960, 106 },
   { 4004, 3951, 106 },
   { 4008, 3960, 106 },
   { 4049, 3951, 106 },
   { 4053, 3960, 106 },
   { 4098, 4099, 107 },
   { 4103, 4104, 107 },
   { 4113, 4114, 107 },
   { 4118, 4120, 107 },
   { 4161, 4099, 107 },
   { 4164, 4104, 107 },
   { 4169, 4114, 107 },
   { 4172, 4120, 107 },
   { 4215, 4216, 0 },
   { 4220, 4221, 0 },
   { 4228, 4229, 0 },
   { 4232, 4233, 0 },
   { 4242, 4243, 0 },
   { 4247, 4249, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3636, 3638, 95 },
   { 5490, 5493, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2376, 2381, 47 },
   { 5066, 2411, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2376, 2381, 47 },
   { 5065, 2376, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2382, 2388, 47 },
   { 5072, 2415, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2382, 2388, 47 },
   { 5071, 2382, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2389, 2395, 47 },
   { 5076, 2421, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2389, 2395, 47 },
   { 5075, 2389, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5081, 5082, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5079, 5080, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5089, 5090, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5087, 5088, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2865, 1336, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2855, 2856, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2866, 1336, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2857, 2858, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3812, 1472, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1764, 1765, 0 },
   { 1785, 1786, 0 },
   { 1867, 1868, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1953, 1954, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3814, 1472, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1725, 1726, 0 },
   { 1832, 1833, 0 },
   { 1915, 1916, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1766, 1767, 0 },
   { 1787, 1788, 0 },
   { 1869, 1870, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1955, 1956, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2864, 1336, 0 },
   { 2864, 1336, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5552, 5554, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5536, 5538, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5520, 5522, 112 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1950, 1951, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1947, 1948, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2853, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2854, 11, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2867, 1348, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2859, 2860, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2868, 1348, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2861, 2862, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 738, 740, 35 },
   { 750, 751, 0 },
   { 766, 767, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 738, 740, 35 },
   { 745, 747, 0 },
   { 750, 751, 0 },
   { 766, 767, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 738, 740, 35 },
   { 742, 253, 0 },
   { 745, 747, 0 },
   { 750, 751, 0 },
   { 766, 767, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 755, 747, 0 },
   { 764, 765, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 753, 253, 0 },
   { 755, 747, 0 },
   { 764, 765, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 757, 253, 0 },
   { 762, 763, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 757, 253, 0 },
   { 758, 761, 0 },
   { 762, 763, 36 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2202, 2203, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2396, 2399, 47 },
   { 5070, 2426, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2396, 2399, 47 },
   { 5069, 2396, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2405, 47 },
   { 5074, 2430, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2400, 2405, 47 },
   { 5073, 2400, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2406, 2410, 47 },
   { 5078, 2436, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2406, 2410, 47 },
   { 5077, 2406, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5085, 5086, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5083, 5084, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5093, 5094, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 5091, 5092, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1296, 1297, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1280, 61, 0 },
   { 1281, 32, 0 },
   { 1296, 1297, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1280, 61, 0 },
   { 1281, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1296, 1297, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1288, 1289, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1283, 32, 0 },
   { 1284, 63, 0 },
   { 1288, 1289, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3024, 3025, 0 },
   { 3563, 3566, 85 },
   { 4770, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3026, 3025, 0 },
   { 3563, 3566, 85 },
   { 4770, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3072, 3068, 0 },
   { 3563, 3566, 85 },
   { 4770, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 2330, 46 },
   { 2336, 2320, 0 },
   { 2344, 2322, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 2330, 46 },
   { 2337, 2322, 0 },
   { 2345, 2320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 2330, 46 },
   { 2333, 2324, 0 },
   { 2342, 2326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2324, 2330, 46 },
   { 2335, 2326, 0 },
   { 2343, 2324, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2331, 46 },
   { 2340, 2320, 0 },
   { 2348, 2322, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2331, 46 },
   { 2341, 2322, 0 },
   { 2349, 2320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2331, 46 },
   { 2338, 2324, 0 },
   { 2346, 2326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2326, 2331, 46 },
   { 2339, 2326, 0 },
   { 2347, 2324, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { 3683, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { 3680, 3681, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { 3680, 3681, 104 },
   { 3683, 3684, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { 3686, 3687, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { 3689, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2814, 36, 0 },
   { 3686, 3687, 104 },
   { 3689, 3690, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2544, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2597, 2598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2589, 2590, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2592, 2593, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2595, 2596, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2580, 2582, 0 },
   { 2562, 2616, 53 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2583, 2584, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2615, 2594, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2614, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2617, 2562, 54 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2585, 2588, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2542, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2545, 2548, 50 },
   { 2570, 32, 0 },
   { 2571, 240, 0 },
   { 2572, 2569, 0 },
   { 2577, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2939, 2940, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2941, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2950, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2952, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2943, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 200, 3, 0 },
   { 3250, 3251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 200, 3, 0 },
   { 3250, 3251, 0 },
   { 3292, 3251, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 32, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 198, 3, 0 },
   { 200, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 198, 3, 0 },
   { 200, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 319, 320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 319, 320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 319, 320, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 322, 323, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 322, 323, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 331, 332, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 331, 332, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 334, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 334, 335, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 325, 326, 0 },
   { 329, 326, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 326, 0 },
   { 329, 326, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4677, 3729, 0 },
   { 4731, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4702, 3729, 0 },
   { 4731, 3749, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4677, 3729, 0 },
   { 4700, 3729, 0 },
   { 4731, 3749, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4677, 3729, 0 },
   { 4730, 3749, 0 },
   { 4731, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4677, 3729, 0 },
   { 4728, 3749, 0 },
   { 4731, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4703, 3729, 0 },
   { 4756, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4702, 3729, 0 },
   { 4703, 3729, 0 },
   { 4755, 3749, 0 },
   { 4756, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4700, 3729, 0 },
   { 4703, 3729, 0 },
   { 4753, 3749, 0 },
   { 4756, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4703, 3729, 0 },
   { 4730, 3749, 0 },
   { 4756, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4703, 3729, 0 },
   { 4728, 3749, 0 },
   { 4756, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { 4808, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4783, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4783, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { 4783, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4702, 3729, 0 },
   { 4755, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4700, 3729, 0 },
   { 4753, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4676, 3729, 0 },
   { 4730, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 200, 3, 0 },
   { 4674, 3729, 0 },
   { 4728, 3749, 0 },
   { 4780, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4700, 3729, 0 },
   { 4702, 3729, 0 },
   { 4753, 3749, 0 },
   { 4755, 3749, 0 },
   { 4805, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4702, 3729, 0 },
   { 4730, 3749, 0 },
   { 4755, 3749, 0 },
   { 4782, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4702, 3729, 0 },
   { 4728, 3749, 0 },
   { 4755, 3749, 0 },
   { 4780, 3749, 0 },
   { 4807, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4676, 3729, 0 },
   { 4700, 3729, 0 },
   { 4730, 3749, 0 },
   { 4753, 3749, 0 },
   { 4782, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4700, 3729, 0 },
   { 4728, 3749, 0 },
   { 4753, 3749, 0 },
   { 4780, 3749, 0 },
   { 4805, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4674, 3729, 0 },
   { 4676, 3729, 0 },
   { 4728, 3749, 0 },
   { 4730, 3749, 0 },
   { 4780, 3749, 0 },
   { 4782, 3749, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3065, 3066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4705, 4679, 0 },
   { 4758, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4704, 4679, 0 },
   { 4757, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4733, 49, 0 },
   { 4785, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4732, 49, 0 },
   { 4784, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 994, 998, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 986, 989, 0 },
   { 994, 998, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4705, 4679, 0 },
   { 4758, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4704, 4679, 0 },
   { 4757, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4680, 4679, 0 },
   { 4733, 49, 0 },
   { 4785, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2, 5, 1 },
   { 8, 10, 1 },
   { 13, 15, 1 },
   { 18, 20, 1 },
   { 23, 25, 1 },
   { 27, 31, 1 },
   { 40, 46, 3 },
   { 227, 61, 0 },
   { 235, 32, 0 },
   { 238, 100, 0 },
   { 4678, 4679, 0 },
   { 4732, 49, 0 },
   { 4784, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4705, 4679, 0 },
   { 4758, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4704, 4679, 0 },
   { 4757, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4680, 4679, 0 },
   { 4733, 49, 0 },
   { 4785, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3203, 3204, 0 },
   { 4678, 4679, 0 },
   { 4732, 49, 0 },
   { 4784, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4681, 4679, 0 },
   { 4705, 4679, 0 },
   { 4734, 49, 0 },
   { 4758, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4681, 4679, 0 },
   { 4704, 4679, 0 },
   { 4734, 49, 0 },
   { 4757, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4681, 4679, 0 },
   { 4733, 49, 0 },
   { 4734, 49, 0 },
   { 4785, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4681, 4679, 0 },
   { 4732, 49, 0 },
   { 4734, 49, 0 },
   { 4784, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4705, 4679, 0 },
   { 4706, 4679, 0 },
   { 4758, 49, 0 },
   { 4759, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4704, 4679, 0 },
   { 4706, 4679, 0 },
   { 4757, 49, 0 },
   { 4759, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4706, 4679, 0 },
   { 4733, 49, 0 },
   { 4759, 49, 0 },
   { 4785, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4706, 4679, 0 },
   { 4732, 49, 0 },
   { 4759, 49, 0 },
   { 4784, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4705, 4679, 0 },
   { 4758, 49, 0 },
   { 4810, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4704, 4679, 0 },
   { 4757, 49, 0 },
   { 4809, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4733, 49, 0 },
   { 4785, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4732, 49, 0 },
   { 4784, 49, 0 },
   { 4811, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4705, 4679, 0 },
   { 4758, 49, 0 },
   { 4786, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4704, 4679, 0 },
   { 4757, 49, 0 },
   { 4786, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4733, 49, 0 },
   { 4785, 49, 0 },
   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4732, 49, 0 },
   { 4784, 49, 0 },
   { 4786, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4704, 4679, 0 },
   { 4705, 4679, 0 },
   { 4757, 49, 0 },
   { 4758, 49, 0 },
   { 4809, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4705, 4679, 0 },
   { 4733, 49, 0 },
   { 4758, 49, 0 },
   { 4785, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4705, 4679, 0 },
   { 4732, 49, 0 },
   { 4758, 49, 0 },
   { 4784, 49, 0 },
   { 4810, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4680, 4679, 0 },
   { 4704, 4679, 0 },
   { 4733, 49, 0 },
   { 4757, 49, 0 },
   { 4785, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4704, 4679, 0 },
   { 4732, 49, 0 },
   { 4757, 49, 0 },
   { 4784, 49, 0 },
   { 4809, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4678, 4679, 0 },
   { 4680, 4679, 0 },
   { 4732, 49, 0 },
   { 4733, 49, 0 },
   { 4784, 49, 0 },
   { 4785, 49, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2508, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 872, 875, 0 },
   { 880, 883, 0 },
   { 888, 891, 0 },
   { 896, 899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 876, 875, 0 },
   { 884, 883, 0 },
   { 892, 891, 0 },
   { 900, 899, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 1000, 1003, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 778, 782, 0 },
   { 982, 984, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 982, 984, 0 },
   { 1000, 1003, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { 3059, 3061, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2088, 2090, 0 },
   { 2118, 2120, 0 },
   { 2148, 2150, 0 },
   { 2179, 2181, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2514, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2088, 2090, 0 },
   { 2118, 2120, 0 },
   { 2148, 2150, 0 },
   { 2179, 2181, 0 },
   { 2514, 1184, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 877, 879, 0 },
   { 885, 887, 0 },
   { 893, 895, 0 },
   { 901, 903, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 877, 879, 0 },
   { 885, 887, 0 },
   { 893, 895, 0 },
   { 901, 903, 0 },
   { 2088, 2090, 0 },
   { 2118, 2120, 0 },
   { 2148, 2150, 0 },
   { 2179, 2181, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { 3062, 3058, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1112, 1111, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1116, 1115, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1067, 1068, 0 },
   { 1169, 1172, 0 },
   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1110, 1111, 0 },
   { 1146, 1147, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1114, 1115, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1113, 1111, 0 },
   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1053, 1054, 0 },
   { 1120, 1118, 0 },
   { 2458, 1214, 0 },
   { 2482, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1097, 1094, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1104, 1102, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1065, 1066, 0 },
   { 1175, 1176, 0 },
   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1093, 1094, 0 },
   { 1148, 1149, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1101, 1102, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1099, 1094, 0 },
   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1055, 1056, 0 },
   { 1109, 1107, 0 },
   { 2457, 1210, 0 },
   { 2483, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { 5188, 5191, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 244, 242, 0 },
   { 3208, 3210, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 244, 242, 0 },
   { 3208, 3210, 0 },
   { 3253, 3210, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3195, 3196, 0 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 242, 0 },
   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 368, 369, 9 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 368, 369, 9 },
   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 504, 505, 0 },
   { 572, 573, 0 },
   { 644, 645, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 504, 505, 0 },
   { 572, 573, 0 },
   { 644, 645, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 504, 505, 0 },
   { 572, 573, 0 },
   { 644, 645, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 162, 164, 0 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 376, 377, 9 },
   { 504, 505, 0 },
   { 572, 573, 0 },
   { 644, 645, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 509, 510, 0 },
   { 576, 577, 0 },
   { 648, 649, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 509, 510, 0 },
   { 576, 577, 0 },
   { 648, 649, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 509, 510, 0 },
   { 576, 577, 0 },
   { 648, 649, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 509, 510, 0 },
   { 576, 577, 0 },
   { 648, 649, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 504, 505, 0 },
   { 509, 510, 0 },
   { 572, 573, 0 },
   { 576, 577, 0 },
   { 644, 645, 0 },
   { 648, 649, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 515, 517, 0 },
   { 581, 583, 0 },
   { 654, 656, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 374, 375, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 374, 375, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 374, 375, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 374, 375, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 374, 375, 9 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 252, 253, 0 },
   { 254, 255, 5 },
   { 256, 216, 6 },
   { 374, 375, 9 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 374, 375, 9 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 145, 147, 0 },
   { 374, 375, 9 },
   { 376, 377, 9 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2488, 32, 0 },
   { 2489, 1192, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2523, 2520, 48 },
   { 2530, 2527, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2535, 2533, 48 },
   { 2540, 2538, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2197, 63, 0 },
   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1660, 1663, 0 },
   { 1667, 1663, 0 },
   { 1672, 1675, 0 },
   { 1679, 1675, 0 },
   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2195, 1192, 0 },
   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4711, 2013, 0 },
   { 4740, 2017, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4687, 2013, 0 },
   { 4710, 2013, 0 },
   { 4740, 2017, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4687, 2013, 0 },
   { 4739, 2017, 0 },
   { 4740, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4687, 2013, 0 },
   { 4738, 2017, 0 },
   { 4740, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4712, 2013, 0 },
   { 4764, 2017, 0 },
   { 4765, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4712, 2013, 0 },
   { 4763, 2017, 0 },
   { 4765, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4712, 2013, 0 },
   { 4739, 2017, 0 },
   { 4765, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4712, 2013, 0 },
   { 4738, 2017, 0 },
   { 4765, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4816, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4815, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { 4817, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4711, 2013, 0 },
   { 4764, 2017, 0 },
   { 4792, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4763, 2017, 0 },
   { 4792, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { 4792, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2523, 2520, 48 },
   { 2530, 2527, 48 },
   { 2535, 2533, 48 },
   { 2540, 2538, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4739, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2824, 2825, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4710, 2013, 0 },
   { 4711, 2013, 0 },
   { 4763, 2017, 0 },
   { 4764, 2017, 0 },
   { 4815, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4711, 2013, 0 },
   { 4739, 2017, 0 },
   { 4764, 2017, 0 },
   { 4791, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4711, 2013, 0 },
   { 4738, 2017, 0 },
   { 4764, 2017, 0 },
   { 4790, 2017, 0 },
   { 4816, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4686, 2013, 0 },
   { 4710, 2013, 0 },
   { 4739, 2017, 0 },
   { 4763, 2017, 0 },
   { 4791, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4710, 2013, 0 },
   { 4738, 2017, 0 },
   { 4763, 2017, 0 },
   { 4790, 2017, 0 },
   { 4815, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4686, 2013, 0 },
   { 4738, 2017, 0 },
   { 4739, 2017, 0 },
   { 4790, 2017, 0 },
   { 4791, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4685, 2013, 0 },
   { 4738, 2017, 0 },
   { 4790, 2017, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 3067, 3066, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3785, 3786, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3791, 3792, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3797, 3798, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3780, 3783, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3788, 3789, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3794, 3795, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2499, 3, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2499, 3, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2835, 1472, 0 },
   { 2836, 2833, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1960, 1963, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 2500, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2500, 2501, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 186, 188, 0 },
   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 2504, 2505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2062, 2064, 42 },
   { 2062, 2070, 44 },
   { 2093, 2095, 42 },
   { 2093, 2101, 44 },
   { 2123, 2125, 42 },
   { 2123, 2131, 44 },
   { 2153, 2155, 42 },
   { 2153, 2161, 44 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1044, 1039, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4737, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4708, 1971, 0 },
   { 4737, 1979, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4684, 1971, 0 },
   { 4707, 1971, 0 },
   { 4737, 1979, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4684, 1971, 0 },
   { 4736, 1979, 0 },
   { 4737, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4684, 1971, 0 },
   { 4735, 1979, 0 },
   { 4737, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4709, 1971, 0 },
   { 4762, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4708, 1971, 0 },
   { 4709, 1971, 0 },
   { 4761, 1979, 0 },
   { 4762, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4709, 1971, 0 },
   { 4760, 1979, 0 },
   { 4762, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4709, 1971, 0 },
   { 4736, 1979, 0 },
   { 4762, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4709, 1971, 0 },
   { 4735, 1979, 0 },
   { 4762, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { 4814, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4789, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4789, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { 4789, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2499, 3, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 2822, 2823, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1362, 1363, 0 },
   { 1384, 1385, 0 },
   { 1406, 1407, 0 },
   { 1426, 1427, 0 },
   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4708, 1971, 0 },
   { 4761, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4760, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4736, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2062, 2064, 42 },
   { 2062, 2070, 44 },
   { 2093, 2095, 42 },
   { 2093, 2101, 44 },
   { 2123, 2125, 42 },
   { 2123, 2131, 44 },
   { 2153, 2155, 42 },
   { 2153, 2161, 44 },
   { 2484, 32, 0 },
   { 2958, 2960, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4707, 1971, 0 },
   { 4708, 1971, 0 },
   { 4760, 1979, 0 },
   { 4761, 1979, 0 },
   { 4812, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4708, 1971, 0 },
   { 4736, 1979, 0 },
   { 4761, 1979, 0 },
   { 4788, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4708, 1971, 0 },
   { 4735, 1979, 0 },
   { 4761, 1979, 0 },
   { 4787, 1979, 0 },
   { 4813, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4683, 1971, 0 },
   { 4707, 1971, 0 },
   { 4736, 1979, 0 },
   { 4760, 1979, 0 },
   { 4788, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4707, 1971, 0 },
   { 4735, 1979, 0 },
   { 4760, 1979, 0 },
   { 4787, 1979, 0 },
   { 4812, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4683, 1971, 0 },
   { 4735, 1979, 0 },
   { 4736, 1979, 0 },
   { 4787, 1979, 0 },
   { 4788, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2484, 32, 0 },
   { 4682, 1971, 0 },
   { 4735, 1979, 0 },
   { 4787, 1979, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2319, 2318, 0 },
   { 2885, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2948, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1124, 1094, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1057, 1058, 0 },
   { 1212, 1210, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1057, 1058, 0 },
   { 1156, 1159, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1057, 1058, 0 },
   { 2472, 722, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 1234, 1011, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 1151, 1154, 0 },
   { 1234, 1011, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1126, 1111, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1061, 1062, 0 },
   { 1215, 1214, 0 },
   { 1215, 1214, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1061, 1062, 0 },
   { 1162, 1163, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1061, 1062, 0 },
   { 1089, 1087, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1525, 1527, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1451, 32, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1446, 1211, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1226, 1227, 0 },
   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 1446, 1211, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1453, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1453, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1531, 1532, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1463, 206, 37 },
   { 1468, 1466, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1453, 32, 0 },
   { 1455, 103, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1455, 103, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1455, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { 1455, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1453, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1453, 32, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1453, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { 1453, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1222, 1224, 0 },
   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { 1457, 103, 0 },
   { 1519, 1522, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1286, 1238, 0 },
   { 1303, 1305, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 1477, 1479, 37 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 1482, 1484, 37 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 1487, 1489, 40 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 1491, 1495, 37 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 206, 1461, 39 },
   { 2465, 1220, 0 },
   { 2887, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 688, 678, 23 },
   { 690, 681, 23 },
   { 692, 684, 24 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 3, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { 3268, 3271, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { 3277, 3281, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { 3287, 3290, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { 772, 774, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { 688, 678, 23 },
   { 690, 681, 23 },
   { 692, 684, 24 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 205, 3, 0 },
   { 3235, 3237, 0 },
   { 3239, 3237, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 205, 3, 0 },
   { 3235, 3237, 0 },
   { 3239, 3237, 0 },
   { 3266, 3237, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 203, 3, 0 },
   { 205, 3, 0 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 203, 3, 0 },
   { 688, 678, 23 },
   { 690, 681, 23 },
   { 692, 684, 24 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 208, 209, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 203, 3, 0 },
   { 205, 3, 0 },
   { 688, 678, 23 },
   { 690, 681, 23 },
   { 692, 684, 24 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2545, 2548, 50 },
   { 2570, 32, 0 },
   { 2571, 240, 0 },
   { 2572, 2569, 0 },
   { 2579, 103, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2549, 2545, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2549, 2545, 51 },
   { 2561, 2562, 0 },
   { 2565, 240, 0 },
   { 2568, 2569, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { 1447, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { 1452, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 2838, 2825, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1454, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1313, 1315, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1454, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 3640, 3636, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1313, 1315, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1313, 1315, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1454, 32, 0 },
   { 1456, 106, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1456, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1313, 1315, 0 },
   { 1456, 106, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1454, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1454, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1313, 1315, 0 },
   { 1454, 32, 0 },
   { 1458, 106, 38 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1458, 106, 38 },
   { 2837, 2823, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1278, 32, 0 },
   { 1291, 1292, 0 },
   { 1313, 1315, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2473, 1039, 0 },
   { 2831, 722, 0 },
   { 2927, 1039, 0 },
   { 2928, 1192, 0 },
   { 2929, 1192, 0 },
   { 1433, 3451, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2477, 1039, 0 },
   { 2832, 722, 0 },
   { 2931, 1192, 0 },
   { 3458, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2474, 1192, 0 },
   { 2830, 1087, 0 },
   { 2924, 1192, 0 },
   { 2925, 1039, 0 },
   { 2926, 1039, 0 },
   { 1430, 3452, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2475, 1192, 0 },
   { 2776, 722, 0 },
   { 2779, 1087, 0 },
   { 2826, 1087, 0 },
   { 2827, 722, 0 },
   { 2923, 1039, 0 },
   { 3456, 3457, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2476, 1039, 0 },
   { 2777, 1087, 0 },
   { 2778, 722, 0 },
   { 2828, 722, 0 },
   { 2829, 1087, 0 },
   { 2922, 1192, 0 },
   { 3453, 3455, 76 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2201, 1335, 0 },
   { 2476, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1296, 1297, 0 },
   { 1323, 1325, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1280, 61, 0 },
   { 1281, 32, 0 },
   { 1296, 1297, 0 },
   { 1323, 1325, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1296, 1297, 0 },
   { 1323, 1325, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 2840, 2823, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1288, 1289, 0 },
   { 1308, 1310, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1283, 32, 0 },
   { 1284, 63, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1283, 32, 0 },
   { 1284, 63, 0 },
   { 1288, 1289, 0 },
   { 1308, 1310, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1288, 1289, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 2839, 2825, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1282, 32, 0 },
   { 1288, 1289, 0 },
   { 1308, 1310, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2782, 1073, 0 },
   { 2783, 669, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1245, 1246, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1247, 1248, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3702, 3684, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3702, 3684, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1235, 1236, 0 },
   { 1237, 1238, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3701, 3681, 104 },
   { 3702, 3684, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3706, 3690, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3706, 3690, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1239, 1240, 0 },
   { 1242, 1243, 0 },
   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3704, 3687, 104 },
   { 3706, 3690, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2316, 2317, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3694, 3695, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3694, 3695, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2316, 2317, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3692, 3693, 104 },
   { 3694, 3695, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3699, 3700, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3699, 3700, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4696, 4668, 0 },
   { 4749, 4721, 0 },
   { 4801, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4694, 4668, 0 },
   { 4747, 4721, 0 },
   { 4799, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4671, 4668, 0 },
   { 4724, 4721, 0 },
   { 4776, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1253, 1251, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3342, 3344, 0 },
   { 3696, 3697, 104 },
   { 3699, 3700, 104 },
   { 4667, 4668, 0 },
   { 4720, 4721, 0 },
   { 4773, 4721, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3313, 3311, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3316, 3314, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1272, 1094, 0 },
   { 1273, 1102, 0 },
   { 1274, 1102, 0 },
   { 2813, 3, 0 },
   { 3319, 3317, 0 },
   { 3696, 3697, 104 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3502, 3500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3505, 3500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1536, 1535, 0 },
   { 3554, 3557, 84 },
   { 4664, 11, 0 },
   { 4717, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4714, 2037, 0 },
   { 4767, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4713, 2037, 0 },
   { 4766, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4742, 2041, 0 },
   { 4794, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4741, 2041, 0 },
   { 4793, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4714, 2037, 0 },
   { 4767, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4713, 2037, 0 },
   { 4766, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4689, 2037, 0 },
   { 4742, 2041, 0 },
   { 4794, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2491, 32, 0 },
   { 2494, 2196, 0 },
   { 4688, 2037, 0 },
   { 4741, 2041, 0 },
   { 4793, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4714, 2037, 0 },
   { 4743, 2041, 0 },
   { 4767, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4690, 2037, 0 },
   { 4713, 2037, 0 },
   { 4743, 2041, 0 },
   { 4766, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4690, 2037, 0 },
   { 4742, 2041, 0 },
   { 4743, 2041, 0 },
   { 4794, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4690, 2037, 0 },
   { 4741, 2041, 0 },
   { 4743, 2041, 0 },
   { 4793, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4714, 2037, 0 },
   { 4715, 2037, 0 },
   { 4767, 2041, 0 },
   { 4768, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4713, 2037, 0 },
   { 4715, 2037, 0 },
   { 4766, 2041, 0 },
   { 4768, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4715, 2037, 0 },
   { 4742, 2041, 0 },
   { 4768, 2041, 0 },
   { 4794, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4715, 2037, 0 },
   { 4741, 2041, 0 },
   { 4768, 2041, 0 },
   { 4793, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4714, 2037, 0 },
   { 4767, 2041, 0 },
   { 4819, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4713, 2037, 0 },
   { 4766, 2041, 0 },
   { 4818, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4742, 2041, 0 },
   { 4794, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4741, 2041, 0 },
   { 4793, 2041, 0 },
   { 4820, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4714, 2037, 0 },
   { 4767, 2041, 0 },
   { 4795, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4713, 2037, 0 },
   { 4766, 2041, 0 },
   { 4795, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4742, 2041, 0 },
   { 4794, 2041, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4741, 2041, 0 },
   { 4793, 2041, 0 },
   { 4795, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4714, 2037, 0 },
   { 4767, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4713, 2037, 0 },
   { 4766, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4689, 2037, 0 },
   { 4742, 2041, 0 },
   { 4794, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 2493, 3, 0 },
   { 4688, 2037, 0 },
   { 4741, 2041, 0 },
   { 4793, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4713, 2037, 0 },
   { 4714, 2037, 0 },
   { 4766, 2041, 0 },
   { 4767, 2041, 0 },
   { 4818, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4714, 2037, 0 },
   { 4742, 2041, 0 },
   { 4767, 2041, 0 },
   { 4794, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4714, 2037, 0 },
   { 4741, 2041, 0 },
   { 4767, 2041, 0 },
   { 4793, 2041, 0 },
   { 4819, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4689, 2037, 0 },
   { 4713, 2037, 0 },
   { 4742, 2041, 0 },
   { 4766, 2041, 0 },
   { 4794, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4713, 2037, 0 },
   { 4741, 2041, 0 },
   { 4766, 2041, 0 },
   { 4793, 2041, 0 },
   { 4818, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4689, 2037, 0 },
   { 4741, 2041, 0 },
   { 4742, 2041, 0 },
   { 4793, 2041, 0 },
   { 4794, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2490, 61, 0 },
   { 4688, 2037, 0 },
   { 4741, 2041, 0 },
   { 4793, 2041, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1750, 11, 0 },
   { 1773, 11, 0 },
   { 1857, 16, 0 },
   { 1966, 1963, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 3840, 3838, 0 },
   { 3846, 3838, 0 },
   { 3852, 3838, 0 },
   { 3858, 3838, 0 },
   { 3861, 3862, 0 },
   { 3866, 3836, 0 },
   { 3870, 3836, 0 },
   { 3874, 3836, 0 },
   { 3877, 3878, 0 },
   { 3880, 3878, 0 },
   { 3883, 3884, 0 },
   { 3888, 3843, 0 },
   { 3892, 3843, 0 },
   { 3895, 3896, 0 },
   { 3898, 3896, 0 },
   { 3901, 3009, 0 },
   { 3903, 3896, 0 },
   { 3905, 3009, 0 },
   { 3907, 1749, 0 },
   { 3911, 3849, 0 },
   { 3914, 3915, 0 },
   { 3917, 3915, 0 },
   { 3920, 3921, 0 },
   { 3923, 3915, 0 },
   { 3925, 3921, 0 },
   { 3927, 1772, 0 },
   { 3929, 3915, 0 },
   { 3931, 3921, 0 },
   { 3933, 1772, 0 },
   { 3935, 1856, 0 },
   { 4663, 11, 0 },
   { 4716, 16, 0 },
   { 4769, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1750, 11, 0 },
   { 1773, 11, 0 },
   { 1857, 16, 0 },
   { 2987, 2988, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { 3840, 3838, 0 },
   { 3846, 3838, 0 },
   { 3852, 3838, 0 },
   { 3858, 3838, 0 },
   { 3861, 3862, 0 },
   { 3866, 3836, 0 },
   { 3870, 3836, 0 },
   { 3874, 3836, 0 },
   { 3877, 3878, 0 },
   { 3880, 3878, 0 },
   { 3883, 3884, 0 },
   { 3888, 3843, 0 },
   { 3892, 3843, 0 },
   { 3895, 3896, 0 },
   { 3898, 3896, 0 },
   { 3901, 3009, 0 },
   { 3903, 3896, 0 },
   { 3905, 3009, 0 },
   { 3907, 1749, 0 },
   { 3911, 3849, 0 },
   { 3914, 3915, 0 },
   { 3917, 3915, 0 },
   { 3920, 3921, 0 },
   { 3923, 3915, 0 },
   { 3925, 3921, 0 },
   { 3927, 1772, 0 },
   { 3929, 3915, 0 },
   { 3931, 3921, 0 },
   { 3933, 1772, 0 },
   { 3935, 1856, 0 },
   { 4663, 11, 0 },
   { 4716, 16, 0 },
   { 4769, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2984, 2986, 0 },
   { 3816, 722, 0 },
   { 3817, 6, 0 },
   { 3818, 11, 0 },
   { 3819, 16, 0 },
   { 3820, 21, 0 },
   { 3837, 3838, 0 },
   { 3844, 3838, 0 },
   { 3850, 3838, 0 },
   { 3856, 3838, 0 },
   { 3864, 3836, 0 },
   { 3868, 3836, 0 },
   { 3872, 3836, 0 },
   { 3886, 3843, 0 },
   { 3890, 3843, 0 },
   { 3909, 3849, 0 },
   { 4691, 11, 0 },
   { 4744, 16, 0 },
   { 4796, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2078, 6, 0 },
   { 2109, 11, 0 },
   { 2139, 16, 0 },
   { 2169, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2519, 2520, 48 },
   { 2526, 2527, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2532, 2533, 48 },
   { 2537, 2538, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2519, 2520, 48 },
   { 2526, 2527, 48 },
   { 2535, 2533, 48 },
   { 2540, 2538, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2523, 2520, 48 },
   { 2530, 2527, 48 },
   { 2532, 2533, 48 },
   { 2537, 2538, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2519, 2520, 48 },
   { 2526, 2527, 48 },
   { 2532, 2533, 48 },
   { 2537, 2538, 48 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 907, 908, 0 },
   { 954, 953, 0 },
   { 961, 960, 0 },
   { 969, 967, 0 },
   { 978, 976, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 956, 955, 0 },
   { 963, 962, 0 },
   { 971, 970, 0 },
   { 980, 979, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 486, 476, 15 },
   { 546, 547, 17 },
   { 556, 547, 17 },
   { 616, 617, 19 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 546, 547, 17 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 616, 617, 19 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 157, 159, 0 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 157, 159, 0 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { 688, 678, 23 },
   { 690, 681, 23 },
   { 692, 684, 24 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { 688, 678, 23 },
   { 690, 681, 23 },
   { 692, 684, 24 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 3, 0 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 205, 3, 0 },
   { 486, 476, 15 },
   { 496, 494, 14 },
   { 556, 547, 17 },
   { 565, 563, 16 },
   { 627, 617, 19 },
   { 636, 634, 18 },
   { 3308, 3237, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 486, 476, 15 },
   { 556, 547, 17 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 486, 476, 15 },
   { 546, 547, 17 },
   { 556, 547, 17 },
   { 616, 617, 19 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 194, 113, 0 },
   { 1736, 1737, 41 },
   { 1843, 1844, 41 },
   { 1926, 1927, 41 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2012, 2014, 0 },
   { 2016, 2018, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1970, 1972, 0 },
   { 1978, 1980, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2036, 2038, 0 },
   { 2040, 2042, 0 },
   { 3821, 722, 0 },
   { 3822, 6, 0 },
   { 3823, 11, 0 },
   { 3824, 16, 0 },
   { 3825, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 952, 953, 0 },
   { 959, 960, 0 },
   { 966, 967, 0 },
   { 975, 976, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1499, 1500, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1461, 206, 37 },
   { 1465, 1466, 37 },
   { 1499, 1500, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 2888, 206, 37 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1451, 32, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1451, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1449, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1499, 1500, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1525, 1527, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1294, 1240, 0 },
   { 1318, 1320, 0 },
   { 1446, 1211, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1275, 32, 0 },
   { 1446, 1211, 0 },
   { 1499, 1500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 373, 258, 9 },
   { 3327, 3328, 0 },
   { 3330, 3331, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 3195, 3196, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 499, 494, 0 },
   { 568, 563, 0 },
   { 639, 634, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 246, 3, 0 },
   { 248, 247, 5 },
   { 249, 36, 6 },
   { 260, 261, 0 },
   { 265, 266, 0 },
   { 499, 494, 0 },
   { 568, 563, 0 },
   { 639, 634, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 499, 494, 0 },
   { 568, 563, 0 },
   { 639, 634, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 162, 164, 0 },
   { 499, 494, 0 },
   { 568, 563, 0 },
   { 639, 634, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 499, 494, 0 },
   { 509, 510, 0 },
   { 568, 563, 0 },
   { 576, 577, 0 },
   { 639, 634, 0 },
   { 648, 649, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 149, 150, 0 },
   { 376, 377, 9 },
   { 499, 494, 0 },
   { 568, 563, 0 },
   { 639, 634, 0 },
   { 699, 701, 28 },
   { 223, 703, 29 },
   { 704, 706, 30 },
   { 707, 701, 31 },
   { 708, 703, 32 },
   { 709, 706, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1117, 1118, 0 },
   { 1146, 1147, 0 },
   { 2482, 1192, 0 },
   { 2904, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1106, 1107, 0 },
   { 1148, 1149, 0 },
   { 2483, 1039, 0 },
   { 2903, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2555, 2558, 51 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1057, 1058, 0 },
   { 1123, 1094, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1061, 1062, 0 },
   { 1125, 1111, 0 },
   { 2481, 1192, 0 },
   { 2907, 1192, 0 },
   { 2910, 1039, 0 },
   { 2913, 1039, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1504, 1505, 0 },
   { 3638, 3636, 96 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1452, 32, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1447, 1448, 0 },
   { 1450, 1448, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1277, 32, 0 },
   { 1299, 1300, 0 },
   { 1328, 1330, 0 },
   { 1504, 1505, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1509, 1510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1280, 61, 0 },
   { 1281, 32, 0 },
   { 1509, 1510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1296, 1297, 0 },
   { 1509, 1510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1279, 32, 0 },
   { 1296, 1297, 0 },
   { 1323, 1325, 0 },
   { 1509, 1510, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 61, 0 },
   { 1743, 1744, 41 },
   { 1850, 1851, 41 },
   { 1933, 1934, 41 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 2498, 2262, 0 },
   { 3020, 3022, 59 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 1151, 1154, 0 },
   { 2479, 1039, 0 },
   { 2892, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1142, 1143, 0 },
   { 1151, 1154, 0 },
   { 1234, 1011, 0 },
   { 2479, 1039, 0 },
   { 2896, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1138, 1141, 0 },
   { 2225, 1040, 0 },
   { 2479, 1039, 0 },
   { 2915, 1039, 0 },
   { 2918, 1192, 0 },
   { 2921, 1192, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 2899, 2891, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1276, 32, 0 },
   { 1531, 1532, 37 },
   { 2901, 2895, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 210, 113, 0 },
   { 211, 115, 5 },
   { 212, 117, 6 },
   { 231, 32, 0 },
   { 234, 32, 0 },
   { 237, 98, 0 },
   { 721, 729, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 114, 32, 0 },
   { 116, 11, 5 },
   { 118, 16, 6 },
   { 194, 113, 0 },
   { 523, 524, 15 },
   { 589, 590, 17 },
   { 662, 663, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 481, 482, 15 },
   { 552, 553, 17 },
   { 622, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 481, 482, 15 },
   { 486, 476, 15 },
   { 552, 553, 17 },
   { 556, 547, 17 },
   { 622, 623, 19 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 481, 482, 15 },
   { 486, 476, 15 },
   { 546, 547, 17 },
   { 552, 553, 17 },
   { 556, 547, 17 },
   { 616, 617, 19 },
   { 622, 623, 19 },
   { 627, 617, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 486, 476, 15 },
   { 493, 494, 14 },
   { 546, 547, 17 },
   { 556, 547, 17 },
   { 562, 563, 16 },
   { 616, 617, 19 },
   { 627, 617, 19 },
   { 633, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 486, 476, 15 },
   { 493, 494, 14 },
   { 496, 494, 14 },
   { 546, 547, 17 },
   { 556, 547, 17 },
   { 562, 563, 16 },
   { 565, 563, 16 },
   { 616, 617, 19 },
   { 627, 617, 19 },
   { 633, 634, 18 },
   { 636, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 481, 482, 15 },
   { 486, 476, 15 },
   { 493, 494, 14 },
   { 546, 547, 17 },
   { 552, 553, 17 },
   { 556, 547, 17 },
   { 562, 563, 16 },
   { 616, 617, 19 },
   { 622, 623, 19 },
   { 627, 617, 19 },
   { 633, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 481, 482, 15 },
   { 552, 553, 17 },
   { 622, 623, 19 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 133, 135, 0 },
   { 475, 476, 15 },
   { 486, 476, 15 },
   { 493, 494, 14 },
   { 546, 547, 17 },
   { 556, 547, 17 },
   { 562, 563, 16 },
   { 616, 617, 19 },
   { 627, 617, 19 },
   { 633, 634, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2208, 2211, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2216, 2220, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2208, 2211, 0 },
   { 2216, 2220, 0 },
   { 2484, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2499, 3, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 1271, 1094, 0 },
   { 1272, 1094, 0 },
   { 1274, 1102, 0 },
   { 1711, 1718, 0 },
   { 1818, 1825, 0 },
   { 1901, 1908, 0 },
   { 2813, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 907, 908, 0 },
   { 954, 953, 0 },
   { 961, 960, 0 },
   { 969, 967, 0 },
   { 978, 976, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 956, 955, 0 },
   { 963, 962, 0 },
   { 971, 970, 0 },
   { 980, 979, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 952, 953, 0 },
   { 959, 960, 0 },
   { 966, 967, 0 },
   { 975, 976, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2499, 3, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 907, 908, 0 },
   { 954, 953, 0 },
   { 961, 960, 0 },
   { 969, 967, 0 },
   { 978, 976, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 956, 955, 0 },
   { 963, 962, 0 },
   { 971, 970, 0 },
   { 980, 979, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 952, 953, 0 },
   { 959, 960, 0 },
   { 966, 967, 0 },
   { 975, 976, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2499, 3, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 907, 908, 0 },
   { 954, 953, 0 },
   { 961, 960, 0 },
   { 969, 967, 0 },
   { 978, 976, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 956, 955, 0 },
   { 963, 962, 0 },
   { 971, 970, 0 },
   { 980, 979, 0 },
   { 2467, 2468, 0 },
   { 2470, 2471, 0 },
   { 2484, 32, 0 },
   { 2485, 32, 0 },
   { 2486, 61, 0 },
   { 2972, 2973, 57 },
   { 3003, 3004, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 38, 2 },
   { 952, 953, 0 },
   { 959, 960, 0 },
   { 966, 967, 0 },
   { 975, 976, 0 },
   { 2509, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2515, 32, 0 },
   { 2961, 2962, 57 },
   { 979, 407, 57 },
   { 2964, 2966, 57 },
   { 970, 432, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 193, 0 },
   { 2487, 32, 0 },
   { 2497, 3, 0 },
   { 5063, 4988, 109 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 2487, 32, 0 },
   { 4987, 4988, 109 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_pass_op_table[nir_num_search_ops] = {
   [nir_op_imul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         10,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         13,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         5,
         13,
         5,
         5,
         5,
         5,
         0,
         12,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         15,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         2,
         3,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         3,
         3,
         354,
         355,
         356,
         357,
         358,
         3,
         359,
         360,
         361,
         362,
         2379,
         2380,
         2381,
         2382,
         2383,
         2384,
         2,
         354,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         2,
         355,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         349,
         356,
         349,
         349,
         349,
         349,
         349,
         349,
         363,
         364,
         365,
         366,
         349,
         349,
         2385,
         2386,
         2387,
         2388,
         2,
         357,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         2,
         358,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         2,
         3,
         2,
         2,
         349,
         2,
         2,
         367,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         350,
         359,
         350,
         350,
         363,
         350,
         350,
         350,
         350,
         368,
         369,
         370,
         350,
         350,
         2389,
         2390,
         2391,
         2392,
         351,
         360,
         351,
         351,
         364,
         351,
         351,
         351,
         368,
         351,
         371,
         372,
         351,
         351,
         2393,
         2394,
         2395,
         2396,
         352,
         361,
         352,
         352,
         365,
         352,
         352,
         352,
         369,
         371,
         352,
         373,
         352,
         352,
         2397,
         2398,
         2399,
         2400,
         353,
         362,
         353,
         353,
         366,
         353,
         353,
         353,
         370,
         372,
         373,
         353,
         353,
         353,
         2401,
         2402,
         2403,
         2404,
         2,
         2379,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         2,
         2380,
         2,
         2,
         349,
         2,
         2,
         2,
         350,
         351,
         352,
         353,
         2,
         2,
         2375,
         2376,
         2377,
         2378,
         2375,
         2381,
         2375,
         2375,
         2385,
         2375,
         2375,
         2375,
         2389,
         2393,
         2397,
         2401,
         2375,
         2375,
         2375,
         2405,
         2406,
         2407,
         2376,
         2382,
         2376,
         2376,
         2386,
         2376,
         2376,
         2376,
         2390,
         2394,
         2398,
         2402,
         2376,
         2376,
         2405,
         2376,
         2408,
         2409,
         2377,
         2383,
         2377,
         2377,
         2387,
         2377,
         2377,
         2377,
         2391,
         2395,
         2399,
         2403,
         2377,
         2377,
         2406,
         2408,
         2377,
         2410,
         2378,
         2384,
         2378,
         2378,
         2388,
         2378,
         2378,
         2378,
         2392,
         2396,
         2400,
         2404,
         2378,
         2378,
         2407,
         2409,
         2410,
         2378,
      },
   },
   [nir_op_ishl] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         0,
         3,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         4,
         14,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         4,
         14,
         4,
         4,
         4,
         4,
         0,
         11,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         17,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         3,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         6,
         19,
         0,
         5,
         5,
         3,
         6,
         20,
         15,
         3,
         6,
         21,
         5,
         5,
         3,
         6,
         22,
         15,
         3,
         6,
         23,
         5,
         5,
         3,
         6,
         0,
         15,
      },
      
      .num_filtered_states = 24,
      .table = (const uint16_t []) {
      
         4,
         5,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         6,
         7,
         6,
         6,
         6,
         1386,
         6,
         6,
         6,
         1387,
         6,
         6,
         6,
         2412,
         6,
         6,
         6,
         1386,
         6,
         1386,
         6,
         1386,
         6,
         1386,
         4,
         1388,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1389,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1390,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1391,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         1392,
         1393,
         1392,
         1392,
         1392,
         1394,
         1392,
         1392,
         1392,
         1395,
         1392,
         1392,
         1392,
         2413,
         1392,
         1392,
         1392,
         1394,
         1392,
         1394,
         1392,
         1394,
         1392,
         1394,
         1396,
         1397,
         1396,
         1396,
         1396,
         1398,
         1396,
         1396,
         1396,
         1399,
         1396,
         1396,
         1396,
         2414,
         1396,
         1396,
         1396,
         1398,
         1396,
         1398,
         1396,
         1398,
         1396,
         1398,
         4,
         1400,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         5,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1401,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         2415,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         2416,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1390,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         2417,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         2418,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         2419,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3100,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3191,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3197,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3201,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3206,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3210,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         3215,
         4,
         4,
         4,
         1384,
         4,
         4,
         4,
         1385,
         4,
         4,
         4,
         2411,
         4,
         4,
         4,
         1384,
         4,
         1384,
         4,
         1384,
         4,
         1384,
      },
   },
   [nir_op_unpack_64_2x32_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         8,
         1617,
         1618,
         1619,
         1620,
         1621,
      },
   },
   [nir_op_imul_2x32_64] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         9,
      },
   },
   [nir_op_umul_2x32_64] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_udiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         11,
         0,
         11,
      },
   },
   [nir_op_idiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         12,
         0,
         0,
         12,
         0,
         374,
         375,
         374,
      },
   },
   [nir_op_umod] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         13,
         0,
         13,
      },
   },
   [nir_op_imod] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         14,
         0,
         14,
      },
   },
   [nir_op_irem] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         15,
         0,
         15,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         8,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         14,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         15,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         10,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         17,
         5,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         9,
         9,
         9,
         9,
         9,
         9,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         5,
         5,
         5,
         5,
         5,
         5,
         15,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         17,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         5,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         16,
         1622,
         1623,
         1624,
         1625,
         1626,
         1627,
         1628,
         1629,
         1630,
         1631,
         1632,
         1633,
         1634,
         2746,
         2747,
         2748,
         3093,
      },
   },
   [nir_op_ineg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         4,
         6,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         17,
         1899,
         1900,
         1901,
         1902,
         1903,
         3060,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         18,
         1635,
         1636,
         1637,
         1638,
         1639,
         1640,
         1641,
         1642,
      },
   },
   [nir_search_op_u2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         19,
         2196,
         2197,
         2198,
      },
   },
   [nir_op_iabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         20,
         1904,
         1905,
         1906,
         2666,
      },
   },
   [nir_search_op_f2b] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         21,
         1643,
         1644,
      },
   },
   [nir_search_op_i2b] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         22,
         1907,
         1908,
         1909,
         1910,
         1911,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         7,
         8,
         7,
         8,
         7,
         8,
         9,
         10,
         9,
         10,
         9,
         10,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         12,
         12,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         14,
         2,
         2,
         15,
         16,
         17,
         18,
         19,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         3,
         21,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         20,
         3,
         3,
         3,
         3,
         3,
         5,
         22,
         5,
         5,
         23,
         24,
         5,
         5,
         5,
         5,
         5,
         22,
         5,
         5,
         5,
         23,
         24,
         22,
         5,
         5,
         23,
         24,
         5,
         22,
         22,
         22,
         25,
         26,
         5,
         23,
         24,
         23,
         24,
         23,
         27,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         9,
         9,
         9,
         10,
         9,
         9,
         9,
         10,
         9,
         9,
         28,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         30,
         31,
         5,
         5,
         30,
         31,
         5,
         5,
         5,
         30,
         31,
         5,
         22,
         32,
         33,
         34,
         22,
         5,
         30,
         31,
         5,
         30,
         31,
         5,
         23,
         35,
         36,
         23,
         24,
         37,
         38,
         24,
         5,
         5,
         30,
         31,
         5,
         30,
         31,
         31,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         8,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         39,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         28,
         28,
         28,
         11,
         11,
         11,
         20,
         3,
         21,
         21,
         21,
         21,
         21,
         21,
         4,
         4,
         20,
         20,
         3,
         20,
         20,
         3,
         0,
         2,
         40,
         41,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         21,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         20,
         20,
         3,
         3,
         3,
         3,
         20,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         43,
         44,
         43,
         44,
         43,
         44,
         43,
         45,
         43,
         44,
         43,
         44,
         46,
         47,
         48,
         49,
         43,
         44,
         50,
         51,
         52,
         44,
         43,
         44,
         53,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         21,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 54,
      .table = (const uint16_t []) {
      
         23,
         24,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         24,
         24,
         1647,
         1648,
         1649,
         24,
         24,
         1650,
         1651,
         1652,
         1653,
         24,
         1654,
         24,
         2731,
         2732,
         2733,
         2734,
         2735,
         2736,
         2737,
         2738,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         3071,
         3071,
         3072,
         3071,
         3071,
         3073,
         3073,
         3071,
         3071,
         3074,
         3075,
         2735,
         3178,
         24,
         3071,
         3071,
         24,
         3073,
         24,
         3071,
         3072,
         3071,
         3071,
         3071,
         1645,
         1647,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1646,
         1648,
         1655,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1655,
         1655,
         1655,
         1655,
         1655,
         2740,
         2730,
         2730,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         3078,
         3078,
         1646,
         3078,
         3078,
         3079,
         3079,
         3078,
         3078,
         1646,
         1655,
         1655,
         1655,
         1646,
         3078,
         3078,
         1646,
         3079,
         1646,
         3078,
         1646,
         3078,
         3078,
         3078,
         1646,
         1649,
         1655,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1655,
         1655,
         1655,
         1655,
         1655,
         2740,
         2730,
         2730,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         1646,
         3078,
         3078,
         1646,
         3078,
         3078,
         3079,
         3079,
         3078,
         3078,
         1646,
         1655,
         1655,
         1655,
         1646,
         3078,
         3078,
         1646,
         3079,
         1646,
         3078,
         1646,
         3078,
         3078,
         3078,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         1656,
         1656,
         1656,
         1656,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         1656,
         3081,
         3081,
         1656,
         3083,
         1656,
         3081,
         1656,
         3080,
         3081,
         3081,
         23,
         24,
         1645,
         1646,
         1646,
         23,
         1658,
         23,
         23,
         1659,
         1659,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         23,
         1650,
         1645,
         1646,
         1646,
         1657,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1657,
         1657,
         1657,
         1657,
         1657,
         1657,
         23,
         23,
         3084,
         3084,
         1657,
         3084,
         3084,
         3085,
         3085,
         3084,
         3084,
         23,
         1645,
         1645,
         1645,
         1657,
         3084,
         3084,
         1657,
         3085,
         1657,
         3084,
         1657,
         3084,
         3084,
         3084,
         23,
         1651,
         1645,
         1646,
         1646,
         1657,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1657,
         1657,
         1657,
         1657,
         1657,
         1657,
         23,
         23,
         3084,
         3084,
         1657,
         3084,
         3084,
         3085,
         3085,
         3084,
         3084,
         23,
         1645,
         1645,
         1645,
         1657,
         3084,
         3084,
         1657,
         3085,
         1657,
         3084,
         1657,
         3084,
         3084,
         3084,
         23,
         1652,
         1645,
         1646,
         1646,
         23,
         1659,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         23,
         1653,
         1645,
         1646,
         1646,
         23,
         1659,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         23,
         24,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         2741,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         23,
         1654,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         23,
         24,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1660,
         1645,
         1645,
         1645,
         1645,
         2742,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1660,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         2742,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         1645,
         2731,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1645,
         2732,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1645,
         2733,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1645,
         2734,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1645,
         2735,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2742,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2742,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         2729,
         2736,
         2729,
         2740,
         2740,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2743,
         2743,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         2729,
         3086,
         3086,
         2729,
         3086,
         3086,
         3087,
         3087,
         3086,
         3086,
         2729,
         2729,
         2729,
         2729,
         2729,
         3086,
         3086,
         2729,
         3087,
         2729,
         3086,
         2729,
         3086,
         3086,
         3086,
         2730,
         2737,
         2739,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2739,
         2739,
         2739,
         2739,
         2739,
         2743,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         3088,
         3088,
         2730,
         3088,
         3088,
         3089,
         3089,
         3088,
         3088,
         2730,
         2739,
         2739,
         2739,
         2730,
         3088,
         3088,
         2730,
         3089,
         2730,
         3088,
         2730,
         3088,
         3088,
         3088,
         2730,
         2738,
         2739,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2739,
         2739,
         2739,
         2739,
         2739,
         2743,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         2730,
         3088,
         3088,
         2730,
         3088,
         3088,
         3089,
         3089,
         3088,
         3088,
         2730,
         2739,
         2739,
         2739,
         2730,
         3088,
         3088,
         2730,
         3089,
         2730,
         3088,
         2730,
         3088,
         3088,
         3088,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         1656,
         1656,
         1656,
         1656,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         3179,
         3081,
         3081,
         3179,
         3083,
         3179,
         3081,
         3179,
         3180,
         3181,
         3181,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         1656,
         1656,
         1656,
         1656,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         1656,
         3182,
         3182,
         1656,
         3183,
         1656,
         3182,
         1656,
         3080,
         3081,
         3182,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         2744,
         1656,
         2744,
         2744,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3090,
         3091,
         23,
         1645,
         1645,
         1645,
         1656,
         3081,
         3081,
         1656,
         3083,
         2744,
         3091,
         1656,
         3080,
         3081,
         3081,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         1656,
         1656,
         1656,
         1656,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         3179,
         3182,
         3182,
         3179,
         3183,
         3179,
         3182,
         3179,
         3180,
         3181,
         3184,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         2744,
         1656,
         2744,
         2744,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3090,
         3091,
         23,
         1645,
         1645,
         1645,
         3179,
         3081,
         3081,
         3179,
         3083,
         3185,
         3091,
         3179,
         3180,
         3181,
         3181,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         2744,
         1656,
         2744,
         2744,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3090,
         3091,
         23,
         1645,
         1645,
         1645,
         1656,
         3182,
         3182,
         1656,
         3183,
         2744,
         3186,
         1656,
         3080,
         3081,
         3182,
         23,
         24,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         2741,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         2741,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         23,
         24,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1660,
         1645,
         1645,
         1645,
         1645,
         2742,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1660,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         3092,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         3069,
         3071,
         3076,
         3078,
         3078,
         3080,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3080,
         3080,
         3080,
         3080,
         3080,
         3080,
         3069,
         3069,
         3080,
         3081,
         3080,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         3069,
         3076,
         3076,
         3076,
         3080,
         3081,
         3081,
         3080,
         3083,
         3080,
         3081,
         3080,
         3080,
         3081,
         3081,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3081,
         3081,
         3081,
         3081,
         3081,
         3081,
         3069,
         3069,
         3081,
         3081,
         3081,
         3081,
         3081,
         3083,
         3083,
         3081,
         3081,
         3069,
         3076,
         3076,
         3076,
         3081,
         3081,
         3081,
         3081,
         3083,
         3081,
         3081,
         3081,
         3081,
         3081,
         3081,
         23,
         3072,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         1656,
         1656,
         1656,
         1656,
         1656,
         1656,
         23,
         23,
         3080,
         3081,
         1656,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         3179,
         3081,
         3081,
         3179,
         3083,
         3179,
         3081,
         3179,
         3180,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3080,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3080,
         3080,
         3080,
         3080,
         3080,
         3080,
         3069,
         3069,
         3080,
         3081,
         3080,
         3080,
         3081,
         3082,
         3083,
         3080,
         3081,
         3069,
         3076,
         3076,
         3076,
         3180,
         3081,
         3081,
         3180,
         3083,
         3180,
         3081,
         3180,
         3180,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3081,
         3081,
         3081,
         3081,
         3081,
         3081,
         3069,
         3069,
         3081,
         3081,
         3081,
         3081,
         3081,
         3083,
         3083,
         3081,
         3081,
         3069,
         3076,
         3076,
         3076,
         3181,
         3081,
         3081,
         3181,
         3083,
         3181,
         3081,
         3181,
         3181,
         3181,
         3181,
         3070,
         3073,
         3077,
         3079,
         3079,
         3082,
         3070,
         3085,
         3085,
         3070,
         3070,
         3070,
         3070,
         3070,
         3077,
         3077,
         3077,
         3077,
         3077,
         3087,
         3089,
         3089,
         3082,
         3082,
         3082,
         3082,
         3082,
         3082,
         3070,
         3070,
         3082,
         3083,
         3082,
         3082,
         3083,
         3082,
         3083,
         3082,
         3083,
         3070,
         3077,
         3077,
         3077,
         3082,
         3183,
         3183,
         3082,
         3183,
         3082,
         3183,
         3082,
         3082,
         3083,
         3183,
         3070,
         3073,
         3077,
         3079,
         3079,
         3083,
         3070,
         3085,
         3085,
         3070,
         3070,
         3070,
         3070,
         3070,
         3077,
         3077,
         3077,
         3077,
         3077,
         3087,
         3089,
         3089,
         3083,
         3083,
         3083,
         3083,
         3083,
         3083,
         3070,
         3070,
         3083,
         3083,
         3083,
         3083,
         3083,
         3083,
         3083,
         3083,
         3083,
         3070,
         3077,
         3077,
         3077,
         3083,
         3183,
         3183,
         3083,
         3183,
         3083,
         3183,
         3083,
         3083,
         3083,
         3183,
         3069,
         3071,
         3076,
         3078,
         3078,
         3080,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3080,
         3080,
         3090,
         3080,
         3090,
         3090,
         3069,
         3069,
         3080,
         3081,
         3080,
         3080,
         3081,
         3082,
         3083,
         3090,
         3091,
         3069,
         3076,
         3076,
         3076,
         3080,
         3081,
         3081,
         3080,
         3083,
         3090,
         3091,
         3080,
         3080,
         3081,
         3081,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3081,
         3081,
         3091,
         3081,
         3091,
         3091,
         3069,
         3069,
         3081,
         3081,
         3081,
         3081,
         3081,
         3083,
         3083,
         3091,
         3091,
         3069,
         3076,
         3076,
         3076,
         3081,
         3081,
         3081,
         3081,
         3083,
         3091,
         3091,
         3081,
         3081,
         3081,
         3081,
         23,
         3074,
         1645,
         1646,
         1646,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         3069,
         3069,
         23,
         3069,
         3069,
         3070,
         3070,
         3069,
         3069,
         23,
         1645,
         1645,
         1645,
         23,
         3069,
         3069,
         23,
         3070,
         23,
         3069,
         23,
         3069,
         3069,
         3069,
         1645,
         3075,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1645,
         2735,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2742,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3092,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         1645,
         3178,
         1645,
         1655,
         1655,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2739,
         2739,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3076,
         3076,
         3077,
         3077,
         3076,
         3076,
         1645,
         1645,
         1645,
         1645,
         1645,
         3076,
         3076,
         1645,
         3077,
         1645,
         3076,
         1645,
         3076,
         3076,
         3076,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         3179,
         1656,
         1656,
         3179,
         3179,
         1656,
         23,
         23,
         3080,
         3081,
         3179,
         3180,
         3181,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         3179,
         3081,
         3181,
         3179,
         3083,
         3179,
         3081,
         3179,
         3180,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3081,
         3182,
         3081,
         3182,
         3081,
         3182,
         3069,
         3069,
         3081,
         3081,
         3081,
         3081,
         3081,
         3183,
         3183,
         3081,
         3081,
         3069,
         3076,
         3076,
         3076,
         3081,
         3081,
         3081,
         3182,
         3183,
         3081,
         3081,
         3081,
         3081,
         3081,
         3081,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3081,
         3182,
         3081,
         3182,
         3081,
         3182,
         3069,
         3069,
         3081,
         3081,
         3081,
         3081,
         3081,
         3183,
         3183,
         3081,
         3081,
         3069,
         3076,
         3076,
         3076,
         3181,
         3081,
         3081,
         3184,
         3183,
         3181,
         3081,
         3181,
         3181,
         3181,
         3181,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         3179,
         1656,
         1656,
         3179,
         3179,
         1656,
         23,
         23,
         3080,
         3081,
         3179,
         3180,
         3181,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         3179,
         3182,
         3184,
         3179,
         3183,
         3179,
         3182,
         3179,
         3180,
         3181,
         3184,
         3070,
         3073,
         3077,
         3079,
         3079,
         3083,
         3070,
         3085,
         3085,
         3070,
         3070,
         3070,
         3070,
         3070,
         3077,
         3077,
         3077,
         3077,
         3077,
         3087,
         3089,
         3089,
         3083,
         3183,
         3083,
         3183,
         3083,
         3183,
         3070,
         3070,
         3083,
         3083,
         3083,
         3083,
         3083,
         3183,
         3183,
         3083,
         3083,
         3070,
         3077,
         3077,
         3077,
         3083,
         3183,
         3183,
         3183,
         3183,
         3083,
         3183,
         3083,
         3083,
         3083,
         3183,
         23,
         24,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         3179,
         1656,
         2744,
         3179,
         3185,
         2744,
         23,
         23,
         3080,
         3081,
         3179,
         3180,
         3181,
         3082,
         3083,
         3090,
         3091,
         23,
         1645,
         1645,
         1645,
         3179,
         3081,
         3181,
         3179,
         3083,
         3185,
         3091,
         3179,
         3180,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3081,
         3182,
         3091,
         3182,
         3091,
         3186,
         3069,
         3069,
         3081,
         3081,
         3081,
         3081,
         3081,
         3183,
         3183,
         3091,
         3091,
         3069,
         3076,
         3076,
         3076,
         3081,
         3081,
         3081,
         3182,
         3183,
         3091,
         3091,
         3081,
         3081,
         3081,
         3081,
         23,
         3072,
         1645,
         1646,
         1646,
         1656,
         23,
         1657,
         1657,
         23,
         23,
         23,
         23,
         23,
         1645,
         1645,
         1645,
         1645,
         1645,
         2729,
         2730,
         2730,
         3179,
         1656,
         1656,
         3179,
         3179,
         1656,
         23,
         23,
         3080,
         3081,
         3179,
         3180,
         3181,
         3082,
         3083,
         3080,
         3081,
         23,
         1645,
         1645,
         1645,
         3179,
         3081,
         3181,
         3179,
         3083,
         3179,
         3081,
         3179,
         3180,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3080,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3180,
         3080,
         3080,
         3180,
         3180,
         3080,
         3069,
         3069,
         3080,
         3081,
         3180,
         3180,
         3181,
         3082,
         3083,
         3080,
         3081,
         3069,
         3076,
         3076,
         3076,
         3180,
         3081,
         3181,
         3180,
         3083,
         3180,
         3081,
         3180,
         3180,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3181,
         3081,
         3081,
         3181,
         3181,
         3081,
         3069,
         3069,
         3081,
         3081,
         3181,
         3181,
         3181,
         3083,
         3083,
         3081,
         3081,
         3069,
         3076,
         3076,
         3076,
         3181,
         3081,
         3181,
         3181,
         3083,
         3181,
         3081,
         3181,
         3181,
         3181,
         3181,
         3069,
         3071,
         3076,
         3078,
         3078,
         3081,
         3069,
         3084,
         3084,
         3069,
         3069,
         3069,
         3069,
         3069,
         3076,
         3076,
         3076,
         3076,
         3076,
         3086,
         3088,
         3088,
         3181,
         3182,
         3081,
         3184,
         3181,
         3182,
         3069,
         3069,
         3081,
         3081,
         3181,
         3181,
         3181,
         3183,
         3183,
         3081,
         3081,
         3069,
         3076,
         3076,
         3076,
         3181,
         3081,
         3181,
         3184,
         3183,
         3181,
         3081,
         3181,
         3181,
         3181,
         3181,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         8,
         0,
         8,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         9,
         0,
         9,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         11,
         0,
         11,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         15,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         20,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         21,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         5,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         22,
         23,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         26,
         27,
         28,
         0,
         0,
         0,
         0,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         20,
         20,
         20,
         20,
         20,
         20,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         25,
         25,
         23,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 31,
      .table = (const uint16_t []) {
      
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         26,
         26,
         26,
         387,
         388,
         389,
         390,
         391,
         392,
         393,
         394,
         395,
         396,
         397,
         398,
         399,
         400,
         26,
         26,
         26,
         2290,
         2291,
         2292,
         2293,
         2294,
         2295,
         26,
         26,
         26,
         26,
         388,
         25,
         26,
         401,
         401,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         25,
         387,
         401,
         401,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         376,
         388,
         376,
         376,
         376,
         402,
         403,
         404,
         405,
         406,
         407,
         408,
         376,
         409,
         410,
         411,
         412,
         376,
         376,
         376,
         2296,
         2297,
         2298,
         2299,
         2300,
         2301,
         376,
         376,
         376,
         376,
         376,
         25,
         389,
         25,
         25,
         402,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         402,
         377,
         390,
         377,
         377,
         403,
         377,
         377,
         413,
         414,
         415,
         416,
         417,
         377,
         418,
         419,
         420,
         421,
         377,
         377,
         377,
         2285,
         2285,
         2302,
         2303,
         2304,
         2305,
         377,
         377,
         377,
         377,
         403,
         378,
         391,
         378,
         378,
         404,
         378,
         413,
         378,
         422,
         423,
         424,
         425,
         378,
         426,
         427,
         428,
         429,
         378,
         378,
         378,
         2306,
         2307,
         2308,
         2309,
         2310,
         2311,
         378,
         378,
         378,
         378,
         404,
         379,
         392,
         379,
         379,
         405,
         379,
         414,
         422,
         379,
         430,
         431,
         432,
         379,
         433,
         434,
         435,
         436,
         379,
         379,
         379,
         2312,
         2313,
         2314,
         2315,
         2316,
         2317,
         379,
         379,
         379,
         379,
         405,
         380,
         393,
         380,
         380,
         406,
         380,
         415,
         423,
         430,
         380,
         437,
         438,
         380,
         439,
         440,
         441,
         442,
         380,
         380,
         380,
         2318,
         2319,
         2320,
         2321,
         2322,
         2323,
         380,
         380,
         380,
         380,
         406,
         381,
         394,
         381,
         381,
         407,
         381,
         416,
         424,
         431,
         437,
         381,
         443,
         381,
         444,
         445,
         446,
         447,
         381,
         381,
         381,
         2324,
         2325,
         2326,
         2327,
         2328,
         2329,
         381,
         381,
         381,
         381,
         407,
         382,
         395,
         382,
         382,
         408,
         382,
         417,
         425,
         432,
         438,
         443,
         382,
         382,
         448,
         449,
         450,
         451,
         382,
         382,
         382,
         2330,
         2331,
         2332,
         2333,
         2334,
         2335,
         382,
         382,
         382,
         382,
         408,
         25,
         396,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         383,
         397,
         383,
         383,
         409,
         383,
         418,
         426,
         433,
         439,
         444,
         448,
         383,
         383,
         452,
         453,
         454,
         383,
         383,
         383,
         2336,
         2337,
         2338,
         2339,
         2340,
         2341,
         383,
         383,
         383,
         383,
         409,
         384,
         398,
         384,
         384,
         410,
         384,
         419,
         427,
         434,
         440,
         445,
         449,
         384,
         452,
         384,
         455,
         456,
         384,
         384,
         384,
         2342,
         2343,
         2344,
         2345,
         2346,
         2347,
         384,
         384,
         384,
         384,
         410,
         385,
         399,
         385,
         385,
         411,
         385,
         420,
         428,
         435,
         441,
         446,
         450,
         385,
         453,
         455,
         385,
         457,
         385,
         385,
         385,
         2348,
         2349,
         2350,
         2351,
         2352,
         2353,
         385,
         385,
         385,
         385,
         411,
         386,
         400,
         386,
         386,
         412,
         386,
         421,
         429,
         436,
         442,
         447,
         451,
         386,
         454,
         456,
         457,
         386,
         386,
         386,
         386,
         2354,
         2355,
         2356,
         2357,
         2358,
         2359,
         386,
         386,
         386,
         386,
         412,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         458,
         458,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         458,
         25,
         458,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         458,
         458,
         458,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         376,
         2284,
         2290,
         2284,
         2284,
         2296,
         2284,
         2285,
         2306,
         2312,
         2318,
         2324,
         2330,
         2284,
         2336,
         2342,
         2348,
         2354,
         2284,
         2284,
         2284,
         2284,
         2285,
         2360,
         2361,
         2362,
         2363,
         2284,
         2284,
         2284,
         2284,
         2296,
         2285,
         2291,
         2285,
         2285,
         2297,
         2285,
         2285,
         2307,
         2313,
         2319,
         2325,
         2331,
         2285,
         2337,
         2343,
         2349,
         2355,
         2285,
         2285,
         2285,
         2285,
         2285,
         2364,
         2365,
         2366,
         2367,
         2285,
         2285,
         2285,
         2285,
         2297,
         2286,
         2292,
         2286,
         2286,
         2298,
         2286,
         2302,
         2308,
         2314,
         2320,
         2326,
         2332,
         2286,
         2338,
         2344,
         2350,
         2356,
         2286,
         2286,
         2286,
         2360,
         2364,
         2286,
         2368,
         2369,
         2370,
         2286,
         2286,
         2286,
         2286,
         2298,
         2287,
         2293,
         2287,
         2287,
         2299,
         2287,
         2303,
         2309,
         2315,
         2321,
         2327,
         2333,
         2287,
         2339,
         2345,
         2351,
         2357,
         2287,
         2287,
         2287,
         2361,
         2365,
         2368,
         2287,
         2371,
         2372,
         2287,
         2287,
         2287,
         2287,
         2299,
         2288,
         2294,
         2288,
         2288,
         2300,
         2288,
         2304,
         2310,
         2316,
         2322,
         2328,
         2334,
         2288,
         2340,
         2346,
         2352,
         2358,
         2288,
         2288,
         2288,
         2362,
         2366,
         2369,
         2371,
         2288,
         2373,
         2288,
         2288,
         2288,
         2288,
         2300,
         2289,
         2295,
         2289,
         2289,
         2301,
         2289,
         2305,
         2311,
         2317,
         2323,
         2329,
         2335,
         2289,
         2341,
         2347,
         2353,
         2359,
         2289,
         2289,
         2289,
         2363,
         2367,
         2370,
         2372,
         2373,
         2289,
         2289,
         2289,
         2289,
         2289,
         2301,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         2374,
         2374,
         25,
         376,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         2374,
         25,
         2374,
         25,
         376,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         2374,
         2374,
         2374,
         25,
         376,
         25,
         26,
         25,
         25,
         376,
         25,
         377,
         378,
         379,
         380,
         381,
         382,
         25,
         383,
         384,
         385,
         386,
         25,
         25,
         25,
         2284,
         2285,
         2286,
         2287,
         2288,
         2289,
         25,
         25,
         25,
         25,
         3168,
         376,
         388,
         376,
         376,
         376,
         402,
         403,
         404,
         405,
         406,
         407,
         408,
         376,
         409,
         410,
         411,
         412,
         376,
         376,
         376,
         2296,
         2297,
         2298,
         2299,
         2300,
         2301,
         376,
         376,
         376,
         3168,
         376,
      },
   },
   [nir_op_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         27,
         28,
         28,
         28,
      },
   },
   [nir_op_isub_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         29,
         30,
         29,
         30,
      },
   },
   [nir_op_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         31,
         32,
         32,
         32,
      },
   },
   [nir_op_usub_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         33,
         34,
         33,
         34,
      },
   },
   [nir_op_usadd_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         35,
         35,
         35,
      },
   },
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         9,
         10,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         0,
         0,
         0,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         16,
         15,
         15,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         17,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         0,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         20,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         4,
         4,
         4,
         4,
         3,
         20,
         3,
         3,
         3,
         20,
         3,
         20,
         3,
         3,
         3,
         20,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         20,
         20,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         22,
         7,
         18,
         0,
         19,
         19,
         19,
         19,
         23,
         19,
         0,
         0,
         18,
         18,
         0,
         18,
         18,
         0,
         0,
         2,
         2,
         2,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         24,
         19,
         18,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         18,
         0,
         0,
         0,
         0,
         18,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         20,
         20,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 25,
      .table = (const uint16_t []) {
      
         36,
         37,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         37,
         37,
         1667,
         1668,
         1669,
         1670,
         1671,
         1672,
         1673,
         1674,
         1675,
         1676,
         37,
         1677,
         37,
         1674,
         2446,
         1667,
         2447,
         2448,
         2449,
         2450,
         3122,
         3123,
         3177,
         1661,
         1667,
         1661,
         1661,
         1661,
         1661,
         1661,
         1678,
         1679,
         1661,
         1680,
         1681,
         1661,
         1682,
         1661,
         1661,
         2451,
         1661,
         2452,
         2453,
         2454,
         2454,
         3124,
         3125,
         1661,
         36,
         1668,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1683,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         36,
         1669,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1683,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         36,
         1670,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         36,
         1671,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         1662,
         1672,
         1678,
         1662,
         1662,
         1662,
         1662,
         1684,
         1685,
         1662,
         1686,
         1687,
         1662,
         1688,
         1662,
         1662,
         2455,
         2456,
         2457,
         2458,
         2459,
         2459,
         3126,
         3127,
         1662,
         1663,
         1673,
         1679,
         1683,
         1683,
         1663,
         1663,
         1685,
         1663,
         1663,
         1689,
         1690,
         1663,
         1691,
         1663,
         1663,
         2460,
         1679,
         2461,
         2462,
         2463,
         2463,
         3128,
         3129,
         1663,
         36,
         1674,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         1664,
         1675,
         1680,
         1664,
         1664,
         1664,
         1664,
         1686,
         1689,
         1664,
         1664,
         1692,
         1664,
         1693,
         1664,
         1664,
         2442,
         1680,
         2464,
         2465,
         2466,
         2466,
         3130,
         3131,
         1664,
         1665,
         1676,
         1681,
         1665,
         1665,
         1665,
         1665,
         1687,
         1690,
         1665,
         1692,
         1694,
         1665,
         1695,
         1665,
         1665,
         2467,
         1681,
         2468,
         2469,
         2470,
         2470,
         3132,
         3133,
         1665,
         36,
         37,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         1696,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         1666,
         1677,
         1682,
         1666,
         1666,
         1666,
         1666,
         1688,
         1691,
         1666,
         1693,
         1695,
         1666,
         1666,
         1666,
         1666,
         2471,
         1682,
         2472,
         2473,
         2474,
         2474,
         3134,
         3135,
         1666,
         36,
         37,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         1697,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         36,
         1674,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         2475,
         2476,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
         2442,
         2446,
         2451,
         2442,
         2442,
         2442,
         2442,
         2455,
         2460,
         2442,
         2442,
         2467,
         2442,
         2471,
         2442,
         2476,
         2476,
         2451,
         2477,
         2478,
         2479,
         2479,
         3136,
         3137,
         2442,
         1661,
         1667,
         1661,
         1661,
         1661,
         1661,
         1661,
         2456,
         1679,
         1661,
         1680,
         1681,
         1661,
         1682,
         1661,
         1661,
         2451,
         1661,
         2452,
         2453,
         2454,
         2454,
         3138,
         3125,
         1661,
         2443,
         2447,
         2452,
         2443,
         2443,
         2443,
         2443,
         2457,
         2461,
         2443,
         2464,
         2468,
         2443,
         2472,
         2443,
         2443,
         2477,
         2452,
         2443,
         2444,
         2480,
         2480,
         3139,
         3121,
         2443,
         2444,
         2448,
         2453,
         2444,
         2444,
         2444,
         2444,
         2458,
         2462,
         2444,
         2465,
         2469,
         2444,
         2473,
         2444,
         2444,
         2478,
         2453,
         2444,
         2481,
         2482,
         2482,
         3140,
         3141,
         2444,
         2445,
         2449,
         2454,
         2445,
         2445,
         2445,
         2445,
         2459,
         2463,
         2445,
         2466,
         2470,
         2445,
         2474,
         2445,
         2445,
         2479,
         2454,
         2480,
         2482,
         2445,
         2445,
         3142,
         3143,
         2445,
         2445,
         2450,
         2454,
         2445,
         2445,
         2445,
         2445,
         2459,
         2463,
         2445,
         2466,
         2470,
         2445,
         2474,
         2445,
         2445,
         2479,
         2454,
         2480,
         2482,
         2445,
         2445,
         3142,
         3143,
         2445,
         3120,
         3122,
         3124,
         3120,
         3120,
         3120,
         3120,
         3126,
         3128,
         3120,
         3130,
         3132,
         3120,
         3134,
         3120,
         3120,
         3136,
         3138,
         3139,
         3140,
         3142,
         3142,
         3126,
         3144,
         3120,
         3121,
         3123,
         3125,
         3121,
         3121,
         3121,
         3121,
         3127,
         3129,
         3121,
         3131,
         3133,
         3121,
         3135,
         3121,
         3121,
         3137,
         3125,
         3121,
         3141,
         3143,
         3143,
         3144,
         3141,
         3121,
         36,
         3177,
         1661,
         36,
         36,
         36,
         36,
         1662,
         1663,
         36,
         1664,
         1665,
         36,
         1666,
         36,
         36,
         2442,
         1661,
         2443,
         2444,
         2445,
         2445,
         3120,
         3121,
         36,
      },
   },
   [nir_op_fmulz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         38,
         39,
         1698,
         38,
         38,
         38,
         1699,
         38,
         39,
         39,
         1700,
         1701,
         1702,
         1703,
         1704,
         39,
         1698,
         1700,
         1698,
         1698,
         1698,
         1698,
         1705,
         1698,
         38,
         1701,
         1698,
         38,
         38,
         38,
         1699,
         38,
         38,
         1702,
         1698,
         38,
         38,
         38,
         1699,
         38,
         38,
         1703,
         1698,
         38,
         38,
         38,
         1699,
         38,
         1699,
         1704,
         1705,
         1699,
         1699,
         1699,
         1699,
         1699,
         38,
         39,
         1698,
         38,
         38,
         38,
         1699,
         1706,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         11,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         12,
         4,
         3,
         3,
         3,
         11,
         3,
         3,
         3,
         3,
         3,
         11,
         3,
         3,
         11,
         3,
         11,
         3,
         11,
         11,
         3,
         3,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         11,
         11,
         11,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         14,
         13,
         13,
         13,
         13,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         10,
         10,
         10,
         15,
         10,
         0,
         0,
         10,
         10,
         0,
         10,
         10,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         10,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         11,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         6,
         6,
         13,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         44,
         45,
         44,
         1716,
         1716,
         44,
         1717,
         44,
         44,
         44,
         44,
         1716,
         1716,
         1717,
         1717,
         44,
         1718,
         1719,
         1718,
         1720,
         1720,
         1718,
         1721,
         1718,
         1718,
         1718,
         1718,
         1720,
         1720,
         1721,
         1721,
         1718,
         1722,
         1723,
         1722,
         1724,
         1724,
         1722,
         1725,
         1722,
         1722,
         1722,
         1722,
         1724,
         1724,
         1725,
         1725,
         1722,
         1726,
         1727,
         1726,
         1728,
         1728,
         1726,
         1729,
         1726,
         1726,
         1726,
         1726,
         1728,
         1728,
         1729,
         1729,
         1726,
         1730,
         1731,
         1730,
         1732,
         1732,
         1730,
         1733,
         1730,
         1730,
         1730,
         1730,
         1732,
         1732,
         1733,
         1733,
         1730,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         2488,
         42,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         2489,
         2490,
         2489,
         2491,
         2491,
         2489,
         2492,
         2489,
         2489,
         2489,
         2489,
         2491,
         2491,
         2492,
         2492,
         2489,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1722,
         1723,
         1722,
         1724,
         1724,
         1722,
         1725,
         1722,
         1722,
         1722,
         1722,
         1724,
         1724,
         1725,
         1725,
         1722,
         1726,
         1727,
         1726,
         1728,
         1728,
         1726,
         1729,
         1726,
         1726,
         1726,
         1726,
         1728,
         1728,
         1729,
         1729,
         1726,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         3146,
         3146,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         1718,
         1719,
         1718,
         1720,
         1720,
         1718,
         1721,
         1718,
         1718,
         1718,
         1718,
         1720,
         1720,
         1721,
         1721,
         1718,
         1734,
         1735,
         1734,
         1736,
         1736,
         1734,
         1737,
         1734,
         1734,
         1734,
         1734,
         1736,
         1736,
         1737,
         1737,
         1734,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         1739,
         1740,
         1739,
         1741,
         1741,
         1739,
         1742,
         1739,
         1739,
         1739,
         1739,
         1741,
         1741,
         1742,
         2493,
         1739,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         2494,
         2495,
         2494,
         2496,
         2496,
         2494,
         2497,
         2494,
         2494,
         2494,
         2494,
         2496,
         2496,
         2497,
         2497,
         2494,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         3147,
         3147,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         1722,
         1723,
         1722,
         1724,
         1724,
         1722,
         1725,
         1722,
         1722,
         1722,
         1722,
         1724,
         1724,
         1725,
         1725,
         1722,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         1726,
         1727,
         1726,
         1728,
         1728,
         1726,
         1729,
         1726,
         1726,
         1726,
         1726,
         1728,
         1728,
         1729,
         1729,
         1726,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         1730,
         1731,
         1730,
         1732,
         1732,
         1730,
         1733,
         1730,
         1730,
         1730,
         1730,
         1732,
         1732,
         1733,
         1733,
         1730,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         2488,
         42,
         1739,
         1740,
         1739,
         1741,
         1741,
         1739,
         1742,
         1739,
         1739,
         1739,
         1739,
         1741,
         1741,
         1742,
         2493,
         1739,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2499,
         2484,
         2500,
         2501,
         2500,
         2502,
         2502,
         2500,
         2503,
         2500,
         2500,
         2500,
         2500,
         2502,
         2502,
         2503,
         2504,
         2500,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         2505,
         2505,
         40,
         1708,
         40,
         40,
         40,
         40,
         2505,
         2505,
         1708,
         2483,
         40,
         40,
         41,
         40,
         2505,
         2505,
         40,
         1708,
         40,
         40,
         40,
         40,
         2505,
         2505,
         1708,
         2483,
         40,
         2500,
         2501,
         2500,
         2502,
         2502,
         2500,
         2503,
         2500,
         2500,
         2500,
         2500,
         3149,
         3149,
         2503,
         2504,
         2500,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         2483,
         40,
         2506,
         2507,
         2506,
         2508,
         2508,
         2506,
         2509,
         2506,
         2506,
         2506,
         2506,
         2508,
         2508,
         2509,
         2509,
         2506,
         2510,
         2511,
         2510,
         2512,
         2512,
         2510,
         2513,
         2510,
         2510,
         2510,
         2510,
         2512,
         2512,
         2513,
         2513,
         2510,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         2489,
         2490,
         2489,
         2491,
         2491,
         2489,
         2492,
         2489,
         2489,
         2489,
         2489,
         2491,
         2491,
         2492,
         2492,
         2489,
         2494,
         2495,
         2494,
         2496,
         2496,
         2494,
         2497,
         2494,
         2494,
         2494,
         2494,
         2496,
         2496,
         2497,
         2497,
         2494,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2499,
         2484,
         2510,
         2511,
         2510,
         2512,
         2512,
         2510,
         2513,
         2510,
         2510,
         2510,
         2510,
         2512,
         2512,
         2513,
         2513,
         2510,
         2510,
         2511,
         2510,
         2512,
         2512,
         2510,
         2513,
         2510,
         2510,
         2510,
         2510,
         2512,
         2512,
         2513,
         2513,
         2510,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         3150,
         3150,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2500,
         2501,
         2500,
         2502,
         2502,
         2500,
         2503,
         2500,
         2500,
         2500,
         2500,
         2502,
         2502,
         2503,
         2504,
         2500,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         1722,
         1723,
         1722,
         1724,
         1724,
         1722,
         1725,
         1722,
         1722,
         1722,
         1722,
         1724,
         1724,
         1725,
         1725,
         1722,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         1726,
         1727,
         1726,
         1728,
         1728,
         1726,
         1729,
         1726,
         1726,
         1726,
         1726,
         1728,
         1728,
         1729,
         1729,
         1726,
         1711,
         1712,
         1711,
         1738,
         1738,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1738,
         1738,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2498,
         2498,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2498,
         2498,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         2505,
         2505,
         40,
         1708,
         40,
         40,
         40,
         40,
         2505,
         2505,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         1709,
         1709,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         1713,
         1713,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         2505,
         2505,
         40,
         1708,
         40,
         40,
         40,
         40,
         2505,
         2505,
         1708,
         2483,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         2486,
         2486,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         1715,
         1715,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         1707,
         1707,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         42,
         43,
         42,
         1709,
         1709,
         42,
         1710,
         42,
         42,
         42,
         42,
         3146,
         3146,
         1710,
         1710,
         42,
         1711,
         1712,
         1711,
         1713,
         1713,
         1711,
         1714,
         1711,
         1711,
         1711,
         1711,
         3147,
         3147,
         1714,
         1714,
         1711,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         2500,
         2501,
         2500,
         2502,
         2502,
         2500,
         2503,
         2500,
         2500,
         2500,
         2500,
         3149,
         3149,
         2503,
         2504,
         2500,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         2484,
         2485,
         2484,
         2486,
         2486,
         2484,
         2487,
         2484,
         2484,
         2484,
         2484,
         3150,
         3150,
         2487,
         2487,
         2484,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1715,
         1715,
         40,
         1708,
         40,
         40,
         40,
         40,
         3148,
         3148,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
         40,
         41,
         40,
         1707,
         1707,
         40,
         1708,
         40,
         40,
         40,
         40,
         3145,
         3145,
         1708,
         1708,
         40,
      },
   },
   [nir_op_ffmaz] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         46,
         47,
         46,
         1743,
         1743,
         1744,
         48,
         49,
         48,
         1745,
         1745,
         1746,
         46,
         47,
         46,
         1743,
         1743,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1743,
         1743,
         1744,
         48,
         49,
         48,
         1745,
         1745,
         1746,
         50,
         51,
         50,
         1748,
         1748,
         1749,
         48,
         49,
         48,
         1745,
         1745,
         1746,
         1750,
         1751,
         1750,
         1752,
         1752,
         1753,
         1754,
         1755,
         1754,
         1756,
         1756,
         1757,
         48,
         49,
         48,
         1745,
         1745,
         1746,
         46,
         47,
         46,
         1743,
         1743,
         1744,
         48,
         49,
         48,
         1745,
         1745,
         1746,
         1758,
         1759,
         1758,
         1760,
         1760,
         1761,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1743,
         1743,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         1750,
         1751,
         1750,
         1752,
         1752,
         1753,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         1754,
         1755,
         1754,
         1756,
         1756,
         1757,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1743,
         1743,
         1744,
         48,
         49,
         48,
         1745,
         1745,
         1746,
         46,
         47,
         46,
         1743,
         1743,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1747,
         1747,
         1744,
         46,
         47,
         46,
         1743,
         1743,
         1744,
      },
   },
   [nir_op_iand] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         3,
         4,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         13,
         12,
         13,
         14,
         15,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         25,
         26,
         27,
         26,
         27,
         28,
         28,
         28,
         28,
         29,
         29,
         29,
         29,
         30,
         30,
         31,
         32,
         33,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         35,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         37,
         38,
         39,
         40,
         41,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         3,
         4,
         3,
         3,
         3,
         4,
         3,
         3,
         4,
         4,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         44,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         43,
         43,
         43,
         43,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         13,
         12,
         12,
         13,
         13,
         14,
         14,
         15,
         14,
         15,
         14,
         15,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         17,
         17,
         17,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         19,
         18,
         17,
         18,
         17,
         18,
         17,
         17,
         17,
         18,
         18,
         17,
         18,
         17,
         17,
         17,
         18,
         17,
         18,
         17,
         17,
         17,
         17,
         21,
         21,
         21,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         22,
         21,
         22,
         21,
         22,
         21,
         21,
         21,
         22,
         22,
         22,
         21,
         22,
         21,
         21,
         21,
         22,
         21,
         22,
         21,
         21,
         21,
         21,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         45,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         46,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         31,
         32,
         32,
         32,
         32,
         32,
         32,
         47,
         32,
         32,
         48,
         49,
         32,
         32,
         31,
         31,
         31,
         33,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         33,
         33,
         33,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         30,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         50,
         51,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         53,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         27,
         26,
         27,
         26,
         26,
         27,
         26,
         27,
         26,
         26,
         27,
         27,
         26,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         42,
         42,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         54,
         4,
         12,
         12,
         12,
         12,
         12,
         12,
         55,
         13,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         43,
         43,
         43,
         43,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         16,
         16,
         16,
         16,
         16,
         16,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         19,
         18,
         18,
         18,
         18,
         18,
         23,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         27,
         28,
         29,
         32,
         34,
         33,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         53,
         53,
         51,
         45,
         9,
         9,
         9,
         9,
         9,
         9,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         56,
         57,
         58,
         4,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         25,
         0,
         18,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         59,
         18,
         18,
         18,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         60,
         61,
         8,
         0,
         8,
         8,
         4,
         13,
         9,
         62,
         63,
         64,
         8,
         8,
         8,
         4,
         13,
         9,
         65,
         66,
         67,
         8,
         8,
         8,
         4,
         13,
         9,
         65,
      },
      
      .num_filtered_states = 68,
      .table = (const uint16_t []) {
      
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         53,
         53,
         1408,
         53,
         1409,
         1410,
         53,
         1411,
         1412,
         1413,
         53,
         53,
         53,
         1414,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         53,
         1415,
         53,
         1416,
         53,
         1417,
         53,
         1418,
         2587,
         2588,
         2589,
         2590,
         2591,
         53,
         53,
         2592,
         2593,
         2594,
         2595,
         3067,
         3068,
         53,
         53,
         53,
         3193,
         3195,
         3196,
         3203,
         3204,
         3205,
         3212,
         3213,
         3214,
         52,
         1408,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         1419,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         1419,
         52,
         52,
         52,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         52,
         1409,
         52,
         1419,
         1419,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         1419,
         52,
         52,
         52,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         52,
         1410,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         2596,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1402,
         1411,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1420,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1421,
         1402,
         1422,
         1402,
         1423,
         1402,
         1424,
         2597,
         1420,
         1402,
         2598,
         2599,
         1402,
         1402,
         2600,
         2601,
         2602,
         2603,
         1402,
         1402,
         1402,
         1402,
         1402,
         1420,
         1402,
         1402,
         1420,
         1402,
         1402,
         1420,
         1402,
         1402,
         1402,
         1412,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1420,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1402,
         1421,
         1402,
         1422,
         1402,
         1423,
         1402,
         1424,
         2597,
         1420,
         1402,
         2598,
         2599,
         1402,
         1402,
         2600,
         2601,
         2602,
         2603,
         1402,
         1402,
         1402,
         1402,
         1402,
         1420,
         1402,
         1402,
         1420,
         1402,
         1402,
         1420,
         1402,
         1402,
         1403,
         1413,
         1403,
         1403,
         1403,
         1403,
         1403,
         1420,
         1420,
         1425,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1426,
         1403,
         1427,
         1403,
         1428,
         1403,
         1429,
         2604,
         1425,
         1403,
         2605,
         2606,
         1403,
         1403,
         2607,
         2608,
         2609,
         2610,
         1403,
         1403,
         1403,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1430,
         52,
         52,
         52,
         52,
         52,
         52,
         1431,
         1431,
         1431,
         1431,
         1432,
         1432,
         1432,
         1432,
         1433,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         1434,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         1435,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         1435,
         52,
         52,
         52,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         52,
         1414,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         1435,
         1435,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         1435,
         52,
         52,
         52,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         1436,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         1436,
         1436,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         1437,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2611,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         1437,
         1437,
         1437,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1431,
         52,
         52,
         52,
         52,
         52,
         52,
         1438,
         1438,
         1438,
         1438,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1431,
         52,
         52,
         52,
         52,
         52,
         52,
         1438,
         1439,
         1440,
         1441,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1431,
         52,
         52,
         52,
         52,
         52,
         52,
         1438,
         1440,
         1442,
         1443,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1431,
         52,
         52,
         52,
         52,
         52,
         52,
         1438,
         1441,
         1443,
         1444,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1432,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1445,
         1445,
         1445,
         1445,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1432,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1445,
         1446,
         1447,
         1448,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1432,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1445,
         1447,
         1449,
         1450,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1432,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1445,
         1448,
         1450,
         1451,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         1433,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1452,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1453,
         1453,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1453,
         1453,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         3187,
         3188,
         3189,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1454,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1455,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1456,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1457,
         1457,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1458,
         1457,
         1457,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2612,
         1458,
         1458,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1457,
         1457,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2613,
         1457,
         1457,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1457,
         1457,
         52,
         1459,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2613,
         1457,
         1457,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1460,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1404,
         1415,
         1404,
         1404,
         1404,
         1404,
         1404,
         1421,
         1421,
         1426,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1404,
         1460,
         1460,
         1404,
         1461,
         1404,
         1462,
         1404,
         1463,
         2614,
         1426,
         1404,
         2615,
         2616,
         1404,
         1404,
         2617,
         2618,
         2619,
         2620,
         1404,
         1404,
         1404,
         1404,
         1404,
         1426,
         1404,
         1404,
         1426,
         1404,
         1404,
         1426,
         1404,
         1404,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1464,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1405,
         1416,
         1405,
         1405,
         1405,
         1405,
         1405,
         1422,
         1422,
         1427,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1405,
         1461,
         1464,
         1464,
         1405,
         1465,
         1405,
         1466,
         2621,
         1427,
         1405,
         2622,
         2623,
         1405,
         1405,
         2624,
         2625,
         2626,
         2627,
         1405,
         1405,
         1405,
         1405,
         1405,
         1427,
         1405,
         1405,
         1427,
         1405,
         1405,
         1427,
         1405,
         1405,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1467,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1406,
         1417,
         1406,
         1406,
         1406,
         1406,
         1406,
         1423,
         1423,
         1428,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1406,
         1462,
         1406,
         1465,
         1467,
         1467,
         1406,
         1468,
         2628,
         1428,
         1406,
         2629,
         2630,
         1406,
         1406,
         2631,
         2632,
         2633,
         2634,
         1406,
         1406,
         1406,
         1406,
         1406,
         1428,
         1406,
         1406,
         1428,
         1406,
         1406,
         1428,
         1406,
         1406,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1469,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1407,
         1418,
         1407,
         1407,
         1407,
         1407,
         1407,
         1424,
         1424,
         1429,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1407,
         1463,
         1407,
         1466,
         1407,
         1468,
         1469,
         1469,
         2635,
         1429,
         1407,
         2636,
         2637,
         1407,
         1407,
         2638,
         2639,
         2640,
         2641,
         1407,
         1407,
         1407,
         1407,
         1407,
         1429,
         1407,
         1407,
         1429,
         1407,
         1407,
         1429,
         1407,
         1407,
         2580,
         2587,
         2580,
         2580,
         2580,
         2580,
         2580,
         2597,
         2597,
         2604,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2580,
         2614,
         2580,
         2621,
         2580,
         2628,
         2580,
         2635,
         2604,
         2604,
         2580,
         2642,
         2643,
         2580,
         2580,
         2644,
         2645,
         2646,
         2647,
         2580,
         2580,
         2580,
         2580,
         2580,
         2604,
         2580,
         2580,
         2604,
         2580,
         2580,
         2604,
         2580,
         2580,
         1403,
         2588,
         1403,
         1403,
         1403,
         1403,
         1403,
         1420,
         1420,
         1425,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1426,
         1403,
         1427,
         1403,
         1428,
         1403,
         1429,
         2604,
         1425,
         1403,
         2605,
         2606,
         1403,
         1403,
         2607,
         2608,
         2609,
         2610,
         1403,
         1403,
         1403,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         52,
         2589,
         52,
         52,
         52,
         52,
         2596,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         2648,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         2581,
         2590,
         2581,
         2581,
         2581,
         2581,
         2581,
         2598,
         2598,
         2605,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2611,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2581,
         2615,
         2581,
         2622,
         2581,
         2629,
         2581,
         2636,
         2642,
         2605,
         2581,
         2611,
         2649,
         2581,
         2581,
         2650,
         2651,
         2652,
         2653,
         2581,
         2581,
         2611,
         2611,
         2611,
         2605,
         2581,
         2581,
         2605,
         2581,
         2581,
         2605,
         2581,
         2581,
         2582,
         2591,
         2582,
         2582,
         2582,
         2582,
         2582,
         2599,
         2599,
         2606,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2582,
         2612,
         2613,
         2613,
         2582,
         2616,
         2582,
         2623,
         2582,
         2630,
         2582,
         2637,
         2643,
         2606,
         2582,
         2649,
         2612,
         2612,
         2612,
         2654,
         2655,
         2656,
         2657,
         2582,
         2582,
         2582,
         2582,
         2582,
         2606,
         2582,
         2582,
         2606,
         2582,
         2582,
         2606,
         2582,
         2582,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1458,
         1457,
         1457,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2612,
         1458,
         2658,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1458,
         1457,
         1457,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2612,
         2658,
         1458,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         52,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         2583,
         2592,
         2583,
         2583,
         2583,
         2583,
         2583,
         2600,
         2600,
         2607,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2583,
         2617,
         2583,
         2624,
         2583,
         2631,
         2583,
         2638,
         2644,
         2607,
         2583,
         2650,
         2654,
         2583,
         2583,
         2583,
         2659,
         2660,
         2661,
         2583,
         2583,
         2583,
         2583,
         2583,
         2607,
         2583,
         2583,
         2607,
         2583,
         2583,
         2607,
         2583,
         2583,
         2584,
         2593,
         2584,
         2584,
         2584,
         2584,
         2584,
         2601,
         2601,
         2608,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2584,
         2618,
         2584,
         2625,
         2584,
         2632,
         2584,
         2639,
         2645,
         2608,
         2584,
         2651,
         2655,
         2584,
         2584,
         2659,
         2584,
         2662,
         2663,
         2584,
         2584,
         2584,
         2584,
         2584,
         2608,
         2584,
         2584,
         2608,
         2584,
         2584,
         2608,
         2584,
         2584,
         2585,
         2594,
         2585,
         2585,
         2585,
         2585,
         2585,
         2602,
         2602,
         2609,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2585,
         2619,
         2585,
         2626,
         2585,
         2633,
         2585,
         2640,
         2646,
         2609,
         2585,
         2652,
         2656,
         2585,
         2585,
         2660,
         2662,
         2585,
         2664,
         2585,
         2585,
         2585,
         2585,
         2585,
         2609,
         2585,
         2585,
         2609,
         2585,
         2585,
         2609,
         2585,
         2585,
         2586,
         2595,
         2586,
         2586,
         2586,
         2586,
         2586,
         2603,
         2603,
         2610,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2586,
         2620,
         2586,
         2627,
         2586,
         2634,
         2586,
         2641,
         2647,
         2610,
         2586,
         2653,
         2657,
         2586,
         2586,
         2661,
         2663,
         2664,
         2665,
         2586,
         2586,
         2586,
         2586,
         2586,
         2610,
         2586,
         2586,
         2610,
         2586,
         2586,
         2610,
         2586,
         2586,
         52,
         3067,
         52,
         1419,
         1419,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         1419,
         52,
         52,
         52,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         52,
         3068,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         1435,
         1435,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         1435,
         52,
         52,
         52,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         1437,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         3187,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2611,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         1437,
         1437,
         1437,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         1437,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         3188,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2611,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         1437,
         1437,
         1437,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         52,
         53,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         1437,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         3189,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2611,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         52,
         1437,
         1437,
         1437,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         52,
         52,
         1403,
         3193,
         1403,
         1403,
         1403,
         1403,
         1403,
         1420,
         1420,
         1425,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1426,
         1403,
         1427,
         1403,
         1428,
         1403,
         1429,
         2604,
         1425,
         1403,
         2605,
         2606,
         1403,
         1403,
         2607,
         2608,
         2609,
         2610,
         1403,
         1403,
         1403,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         52,
         3195,
         52,
         1419,
         1419,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         1419,
         52,
         52,
         52,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         52,
         3196,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         1435,
         1435,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         1435,
         52,
         52,
         52,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         3203,
         1403,
         1403,
         1403,
         1403,
         1403,
         1420,
         1420,
         1425,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1426,
         1403,
         1427,
         1403,
         1428,
         1403,
         1429,
         2604,
         1425,
         1403,
         2605,
         2606,
         1403,
         1403,
         2607,
         2608,
         2609,
         2610,
         1403,
         1403,
         1403,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         52,
         3204,
         52,
         1419,
         1419,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         1419,
         52,
         52,
         52,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         52,
         3205,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         1435,
         1435,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         1435,
         52,
         52,
         52,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         3212,
         1403,
         1403,
         1403,
         1403,
         1403,
         1420,
         1420,
         1425,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1403,
         1426,
         1403,
         1427,
         1403,
         1428,
         1403,
         1429,
         2604,
         1425,
         1403,
         2605,
         2606,
         1403,
         1403,
         2607,
         2608,
         2609,
         2610,
         1403,
         1403,
         1403,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         1425,
         1403,
         1403,
         52,
         3213,
         52,
         1419,
         1419,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         1419,
         52,
         52,
         52,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         1403,
         1419,
         52,
         52,
         3214,
         52,
         52,
         52,
         52,
         52,
         1402,
         1402,
         1403,
         52,
         52,
         1435,
         1435,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         52,
         1404,
         52,
         1405,
         52,
         1406,
         52,
         1407,
         2580,
         1403,
         52,
         2581,
         2582,
         52,
         52,
         2583,
         2584,
         2585,
         2586,
         52,
         1435,
         52,
         52,
         52,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
         1403,
         52,
         1435,
      },
   },
   [nir_op_ior] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         3,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         9,
         10,
         11,
         12,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         22,
         23,
         23,
         23,
         23,
         24,
         24,
         24,
         24,
         25,
         25,
         25,
         25,
         26,
         27,
         0,
         28,
         0,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         30,
         31,
         0,
         0,
         0,
         0,
         32,
         0,
         0,
         33,
         34,
         35,
         36,
         37,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         39,
         40,
         41,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         42,
         2,
         42,
         3,
         3,
         3,
         3,
         2,
         3,
         2,
         42,
         2,
         3,
         2,
         42,
         43,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         44,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         45,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         44,
         5,
         5,
         5,
         5,
         5,
         5,
         44,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         44,
         5,
         5,
         5,
         5,
         5,
         44,
         44,
         44,
         44,
         44,
         44,
         44,
         44,
         44,
         44,
         44,
         44,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         46,
         9,
         46,
         9,
         10,
         9,
         46,
         10,
         10,
         11,
         11,
         12,
         11,
         12,
         11,
         12,
         31,
         31,
         31,
         31,
         31,
         31,
         31,
         34,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         14,
         14,
         14,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         15,
         14,
         15,
         14,
         15,
         14,
         14,
         14,
         15,
         15,
         14,
         15,
         14,
         14,
         14,
         15,
         14,
         15,
         14,
         14,
         14,
         14,
         18,
         18,
         18,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         19,
         18,
         19,
         18,
         19,
         18,
         18,
         18,
         19,
         19,
         19,
         18,
         19,
         18,
         18,
         18,
         19,
         18,
         19,
         18,
         18,
         18,
         18,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         47,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         28,
         0,
         0,
         0,
         0,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         29,
         0,
         0,
         0,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         26,
         26,
         26,
         27,
         27,
         27,
         27,
         27,
         27,
         27,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         26,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         48,
         49,
         0,
         0,
         0,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         50,
         51,
         32,
         32,
         32,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         23,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         25,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         53,
         54,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         38,
         38,
         38,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         55,
         55,
         55,
         55,
         3,
         3,
         3,
         3,
         3,
         56,
         9,
         56,
         9,
         56,
         9,
         10,
         10,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         22,
         22,
         22,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         44,
         44,
         44,
         44,
         44,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         57,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         57,
         57,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         13,
         13,
         13,
         13,
         13,
         13,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         16,
         15,
         15,
         15,
         15,
         15,
         20,
         19,
         19,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         23,
         24,
         25,
         28,
         29,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         31,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         51,
         51,
         49,
         47,
         5,
         5,
         5,
         58,
         5,
         5,
         59,
         60,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         32,
         32,
         32,
         13,
         13,
         13,
         61,
         62,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         22,
         0,
         15,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         3,
         10,
         4,
         0,
         63,
         64,
         65,
         66,
         5,
         5,
         3,
         10,
         4,
         67,
         68,
         69,
         70,
         5,
         5,
         3,
         10,
         4,
         71,
         72,
         73,
         74,
         5,
         5,
      },
      
      .num_filtered_states = 75,
      .table = (const uint16_t []) {
      
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         55,
         55,
         55,
         55,
         1479,
         1480,
         1481,
         55,
         55,
         55,
         55,
         55,
         55,
         1482,
         1483,
         1483,
         1483,
         1483,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         55,
         1484,
         55,
         55,
         1485,
         55,
         55,
         1486,
         55,
         1487,
         55,
         1488,
         55,
         55,
         55,
         55,
         55,
         1480,
         1481,
         55,
         55,
         2518,
         2519,
         2520,
         2521,
         55,
         55,
         55,
         55,
         55,
         1480,
         1480,
         1479,
         1479,
         55,
         55,
         1479,
         1479,
         55,
         55,
         1479,
         1479,
         55,
         55,
         1479,
         1479,
         55,
         55,
         54,
         55,
         54,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         3061,
         1471,
         1471,
         1470,
         1470,
         1489,
         54,
         1470,
         1470,
         1489,
         54,
         1470,
         1470,
         1489,
         54,
         1470,
         1470,
         1489,
         54,
         54,
         55,
         1489,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1490,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         1489,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         1489,
         3061,
         1471,
         1471,
         1470,
         1470,
         1489,
         1490,
         1470,
         1470,
         1489,
         1490,
         1470,
         1470,
         1489,
         1490,
         1470,
         1470,
         1489,
         1490,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1471,
         1480,
         1471,
         1471,
         1492,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1500,
         1501,
         1501,
         1501,
         1501,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1502,
         1471,
         1471,
         1503,
         1471,
         1471,
         1504,
         1471,
         1505,
         1471,
         1506,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         2527,
         2528,
         2529,
         2530,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1471,
         1481,
         1471,
         1471,
         1492,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1500,
         1501,
         1501,
         1501,
         1501,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1502,
         1471,
         1471,
         1503,
         1471,
         1471,
         1504,
         1471,
         1505,
         1471,
         1506,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         2527,
         2528,
         2529,
         2530,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         1507,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         1508,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         1509,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2531,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3062,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         1510,
         1470,
         1470,
         54,
         1510,
         1470,
         1470,
         54,
         1510,
         1470,
         1470,
         54,
         1510,
         54,
         55,
         54,
         1490,
         1470,
         1471,
         1471,
         54,
         54,
         1510,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2531,
         1490,
         1471,
         1471,
         1510,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3062,
         1510,
         1471,
         1471,
         1470,
         1470,
         1490,
         1510,
         1470,
         1470,
         1490,
         1510,
         1470,
         1470,
         1490,
         1510,
         1470,
         1470,
         1490,
         1510,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         1511,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         1511,
         1511,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1472,
         1482,
         1472,
         1472,
         1493,
         1500,
         1500,
         1472,
         1472,
         1472,
         1472,
         1472,
         1472,
         1512,
         1513,
         1513,
         1513,
         1513,
         1472,
         1472,
         1472,
         1472,
         1472,
         1472,
         1472,
         1472,
         1472,
         1472,
         1514,
         1472,
         1472,
         1515,
         1472,
         1472,
         1516,
         1472,
         1517,
         1472,
         1518,
         1472,
         1472,
         1472,
         1472,
         1472,
         1500,
         1500,
         1472,
         1472,
         2532,
         2533,
         2534,
         2535,
         1472,
         1472,
         1472,
         1472,
         1472,
         1500,
         1500,
         1493,
         1493,
         1472,
         1472,
         1493,
         1493,
         1472,
         1472,
         1493,
         1493,
         1472,
         1472,
         1493,
         1493,
         1472,
         1472,
         1473,
         1483,
         1473,
         1473,
         1494,
         1501,
         1501,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1513,
         1519,
         1519,
         1519,
         1519,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1520,
         1473,
         1473,
         1521,
         1473,
         1473,
         1522,
         1473,
         1523,
         1473,
         1524,
         1473,
         1473,
         1473,
         1473,
         1473,
         2536,
         2536,
         1473,
         1473,
         2537,
         2538,
         2539,
         2540,
         1473,
         1473,
         1473,
         1473,
         1473,
         1501,
         1501,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1473,
         1483,
         1473,
         1473,
         1494,
         1501,
         1501,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1513,
         1519,
         1525,
         1526,
         1527,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1520,
         1473,
         1473,
         1521,
         1473,
         1473,
         1522,
         1473,
         1523,
         1473,
         1524,
         1473,
         1473,
         1473,
         1473,
         1473,
         2536,
         2536,
         1473,
         1473,
         2537,
         2538,
         2539,
         2540,
         1473,
         1473,
         1473,
         1473,
         1473,
         1501,
         1501,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1473,
         1483,
         1473,
         1473,
         1494,
         1501,
         1501,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1513,
         1519,
         1526,
         1528,
         1529,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1520,
         1473,
         1473,
         1521,
         1473,
         1473,
         1522,
         1473,
         1523,
         1473,
         1524,
         1473,
         1473,
         1473,
         1473,
         1473,
         2536,
         2536,
         1473,
         1473,
         2537,
         2538,
         2539,
         2540,
         1473,
         1473,
         1473,
         1473,
         1473,
         1501,
         1501,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1473,
         1483,
         1473,
         1473,
         1494,
         1501,
         1501,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1513,
         1519,
         1527,
         1529,
         1530,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1473,
         1520,
         1473,
         1473,
         1521,
         1473,
         1473,
         1522,
         1473,
         1523,
         1473,
         1524,
         1473,
         1473,
         1473,
         1473,
         1473,
         2536,
         2536,
         1473,
         1473,
         2537,
         2538,
         2539,
         2540,
         1473,
         1473,
         1473,
         1473,
         1473,
         1501,
         1501,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         1494,
         1494,
         1473,
         1473,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         1531,
         1531,
         1531,
         1531,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         1531,
         1532,
         1533,
         1534,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         1531,
         1533,
         1535,
         1536,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         1531,
         1534,
         1536,
         1537,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         1508,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         1538,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         1539,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         1540,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1541,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1542,
         1542,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1542,
         1542,
         1543,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1474,
         1484,
         1474,
         1474,
         1495,
         1502,
         1502,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1514,
         1520,
         1520,
         1520,
         1520,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1474,
         1543,
         1544,
         1474,
         1474,
         1545,
         1474,
         1474,
         1546,
         1474,
         1547,
         1474,
         1548,
         1474,
         1474,
         1474,
         1474,
         1474,
         1502,
         1502,
         1474,
         1474,
         2541,
         2542,
         2543,
         2544,
         1474,
         1474,
         1474,
         1474,
         1474,
         1502,
         1502,
         1495,
         1495,
         1474,
         1474,
         1495,
         1495,
         1474,
         1474,
         1495,
         1495,
         1474,
         1474,
         1495,
         1495,
         1474,
         1474,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         1549,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1550,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1475,
         1485,
         1475,
         1475,
         1496,
         1503,
         1503,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1515,
         1521,
         1521,
         1521,
         1521,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1475,
         1545,
         1475,
         1550,
         1550,
         1475,
         1475,
         1551,
         1475,
         1552,
         1475,
         1553,
         1475,
         1475,
         1475,
         1475,
         1475,
         1503,
         1503,
         1475,
         1475,
         2545,
         2546,
         2547,
         2548,
         1475,
         1475,
         1475,
         1475,
         1475,
         1503,
         1503,
         1496,
         1496,
         1475,
         1475,
         1496,
         1496,
         1475,
         1475,
         1496,
         1496,
         1475,
         1475,
         1496,
         1496,
         1475,
         1475,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         2549,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1554,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1476,
         1486,
         1476,
         1476,
         1497,
         1504,
         1504,
         1476,
         1476,
         1476,
         1476,
         1476,
         1476,
         1516,
         1522,
         1522,
         1522,
         1522,
         1476,
         1476,
         1476,
         1476,
         1476,
         1476,
         1476,
         1476,
         1476,
         1476,
         1546,
         1476,
         1476,
         1551,
         1476,
         1554,
         1554,
         1476,
         1555,
         1476,
         1556,
         1476,
         1476,
         1476,
         1476,
         1476,
         1504,
         1504,
         1476,
         1476,
         2550,
         2551,
         2552,
         2553,
         1476,
         1476,
         1476,
         1476,
         1476,
         1504,
         1504,
         1497,
         1497,
         1476,
         1476,
         1497,
         1497,
         1476,
         1476,
         1497,
         1497,
         1476,
         1476,
         1497,
         1497,
         1476,
         1476,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1557,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1477,
         1487,
         1477,
         1477,
         1498,
         1505,
         1505,
         1477,
         1477,
         1477,
         1477,
         1477,
         1477,
         1517,
         1523,
         1523,
         1523,
         1523,
         1477,
         1477,
         1477,
         1477,
         1477,
         1477,
         1477,
         1477,
         1477,
         1477,
         1547,
         1477,
         1477,
         1552,
         1477,
         1477,
         1555,
         1557,
         1557,
         1477,
         1558,
         1477,
         1477,
         1477,
         1477,
         1477,
         1505,
         1505,
         1477,
         1477,
         2554,
         2555,
         2556,
         2557,
         1477,
         1477,
         1477,
         1477,
         1477,
         1505,
         1505,
         1498,
         1498,
         1477,
         1477,
         1498,
         1498,
         1477,
         1477,
         1498,
         1498,
         1477,
         1477,
         1498,
         1498,
         1477,
         1477,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1559,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1478,
         1488,
         1478,
         1478,
         1499,
         1506,
         1506,
         1478,
         1478,
         1478,
         1478,
         1478,
         1478,
         1518,
         1524,
         1524,
         1524,
         1524,
         1478,
         1478,
         1478,
         1478,
         1478,
         1478,
         1478,
         1478,
         1478,
         1478,
         1548,
         1478,
         1478,
         1553,
         1478,
         1478,
         1556,
         1478,
         1558,
         1559,
         1559,
         1478,
         1478,
         1478,
         1478,
         1478,
         1506,
         1506,
         1478,
         1478,
         2558,
         2559,
         2560,
         2561,
         1478,
         1478,
         1478,
         1478,
         1478,
         1506,
         1506,
         1499,
         1499,
         1478,
         1478,
         1499,
         1499,
         1478,
         1478,
         1499,
         1499,
         1478,
         1478,
         1499,
         1499,
         1478,
         1478,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         1560,
         1560,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         1560,
         54,
         1560,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         1560,
         1560,
         1560,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         2531,
         2531,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         1489,
         1471,
         1471,
         2562,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         3063,
         1471,
         1471,
         1470,
         1470,
         1489,
         2531,
         1470,
         1470,
         1489,
         2531,
         1470,
         1470,
         1489,
         2531,
         1470,
         1470,
         1489,
         2531,
         54,
         55,
         1489,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1490,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         2549,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         1489,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2563,
         2564,
         54,
         54,
         54,
         1489,
         3061,
         3064,
         3169,
         1470,
         1470,
         1489,
         1490,
         1470,
         1470,
         1489,
         1490,
         1470,
         1470,
         1489,
         1490,
         1470,
         1470,
         1489,
         1490,
         1471,
         1480,
         1471,
         1471,
         1492,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1500,
         2536,
         2536,
         2536,
         2536,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1502,
         1471,
         1471,
         1503,
         1471,
         1471,
         1504,
         1471,
         1505,
         1471,
         1506,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         2527,
         2528,
         2529,
         2530,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1471,
         1481,
         1471,
         1471,
         1492,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1500,
         2536,
         2536,
         2536,
         2536,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1502,
         1471,
         1471,
         1503,
         1471,
         1471,
         1504,
         1471,
         1505,
         1471,
         1506,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         2527,
         2528,
         2529,
         2530,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         54,
         55,
         2522,
         2522,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2562,
         2522,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3065,
         54,
         1471,
         1471,
         1470,
         1470,
         2522,
         1510,
         1470,
         1470,
         2522,
         1510,
         1470,
         1470,
         2522,
         1510,
         1470,
         1470,
         2522,
         1510,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         2565,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         2514,
         2518,
         2514,
         2514,
         2523,
         2527,
         2527,
         2514,
         2514,
         2514,
         2514,
         2514,
         2514,
         2532,
         2537,
         2537,
         2537,
         2537,
         2514,
         2514,
         2514,
         2514,
         2514,
         2514,
         2514,
         2514,
         2514,
         2514,
         2541,
         2514,
         2514,
         2545,
         2514,
         2514,
         2550,
         2514,
         2554,
         2514,
         2558,
         2514,
         2514,
         2514,
         2514,
         2514,
         2527,
         2527,
         2514,
         2514,
         2514,
         2566,
         2567,
         2568,
         2514,
         2514,
         2514,
         2514,
         2514,
         2527,
         2527,
         2523,
         2523,
         2514,
         2514,
         2523,
         2523,
         2514,
         2514,
         2523,
         2523,
         2514,
         2514,
         2523,
         2523,
         2514,
         2514,
         2515,
         2519,
         2515,
         2515,
         2524,
         2528,
         2528,
         2515,
         2515,
         2515,
         2515,
         2515,
         2515,
         2533,
         2538,
         2538,
         2538,
         2538,
         2515,
         2515,
         2515,
         2515,
         2515,
         2515,
         2515,
         2515,
         2515,
         2515,
         2542,
         2515,
         2515,
         2546,
         2515,
         2515,
         2551,
         2515,
         2555,
         2515,
         2559,
         2515,
         2515,
         2515,
         2515,
         2515,
         2528,
         2528,
         2515,
         2515,
         2566,
         2515,
         2569,
         2570,
         2515,
         2515,
         2515,
         2515,
         2515,
         2528,
         2528,
         2524,
         2524,
         2515,
         2515,
         2524,
         2524,
         2515,
         2515,
         2524,
         2524,
         2515,
         2515,
         2524,
         2524,
         2515,
         2515,
         2516,
         2520,
         2516,
         2516,
         2525,
         2529,
         2529,
         2516,
         2516,
         2516,
         2516,
         2516,
         2516,
         2534,
         2539,
         2539,
         2539,
         2539,
         2516,
         2516,
         2516,
         2516,
         2516,
         2516,
         2516,
         2516,
         2516,
         2516,
         2543,
         2516,
         2516,
         2547,
         2516,
         2516,
         2552,
         2516,
         2556,
         2516,
         2560,
         2516,
         2516,
         2516,
         2516,
         2563,
         2529,
         2529,
         2516,
         2516,
         2567,
         2569,
         2516,
         2571,
         2516,
         2516,
         2516,
         2516,
         2516,
         2529,
         2529,
         2525,
         2525,
         2516,
         2516,
         2525,
         2525,
         2516,
         2516,
         2525,
         2525,
         2516,
         2516,
         2525,
         2525,
         2516,
         2516,
         2517,
         2521,
         2517,
         2517,
         2526,
         2530,
         2530,
         2517,
         2517,
         2517,
         2517,
         2517,
         2517,
         2535,
         2540,
         2540,
         2540,
         2540,
         2517,
         2517,
         2517,
         2517,
         2517,
         2517,
         2517,
         2517,
         2517,
         2517,
         2544,
         2517,
         2517,
         2548,
         2517,
         2517,
         2553,
         2517,
         2557,
         2517,
         2561,
         2517,
         2517,
         2517,
         2517,
         2564,
         2530,
         2530,
         2517,
         2517,
         2568,
         2570,
         2571,
         2572,
         2517,
         2517,
         2517,
         2517,
         2517,
         2530,
         2530,
         2526,
         2526,
         2517,
         2517,
         2526,
         2526,
         2517,
         2517,
         2526,
         2526,
         2517,
         2517,
         2526,
         2526,
         2517,
         2517,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         2573,
         2573,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         2573,
         54,
         2573,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         54,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         54,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         54,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         2573,
         2573,
         2573,
         54,
         54,
         1471,
         1471,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         1470,
         1470,
         54,
         54,
         54,
         55,
         54,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         3062,
         3062,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         54,
         1489,
         1471,
         1471,
         3065,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         54,
         3066,
         1471,
         1471,
         1470,
         1470,
         1489,
         3062,
         1470,
         1470,
         1489,
         3062,
         1470,
         1470,
         1489,
         3062,
         1470,
         1470,
         1489,
         3062,
         54,
         55,
         3061,
         3061,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         3063,
         3061,
         1471,
         1471,
         54,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3066,
         54,
         1471,
         1471,
         1470,
         1470,
         3061,
         1510,
         1470,
         1470,
         3061,
         1510,
         1470,
         1470,
         3061,
         1510,
         1470,
         1470,
         3061,
         1510,
         1471,
         1480,
         1471,
         1471,
         1492,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1500,
         1501,
         1501,
         1501,
         1501,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1502,
         1471,
         1471,
         1503,
         1471,
         1471,
         1504,
         1471,
         1505,
         1471,
         1506,
         1471,
         1471,
         1471,
         1471,
         3064,
         1471,
         1471,
         1471,
         1471,
         2527,
         2528,
         2529,
         2530,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1471,
         1480,
         1471,
         1471,
         1492,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1500,
         1501,
         1501,
         1501,
         1501,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1502,
         1471,
         1471,
         1503,
         1471,
         1471,
         1504,
         1471,
         1505,
         1471,
         1506,
         1471,
         1471,
         1471,
         1471,
         3169,
         1471,
         1471,
         1471,
         1471,
         2527,
         2528,
         2529,
         2530,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1492,
         1492,
         1471,
         1471,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         3170,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         3170,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         54,
         55,
         1489,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1490,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         1489,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         1489,
         3061,
         1471,
         1471,
         1470,
         1470,
         1489,
         3171,
         1470,
         1470,
         1489,
         3171,
         1470,
         1470,
         1489,
         3171,
         1470,
         1470,
         1489,
         3171,
         54,
         55,
         54,
         1490,
         1470,
         1471,
         1471,
         54,
         54,
         1510,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2531,
         1490,
         1471,
         1471,
         1510,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3062,
         1510,
         1471,
         1471,
         1470,
         1470,
         3171,
         1510,
         1470,
         1470,
         3171,
         1510,
         1470,
         1470,
         3171,
         1510,
         1470,
         1470,
         3171,
         1510,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         1491,
         3199,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         3199,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         54,
         55,
         1489,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1490,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         1489,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         1489,
         3061,
         1471,
         1471,
         1470,
         1470,
         1489,
         3171,
         1470,
         1470,
         1489,
         3200,
         1470,
         1470,
         1489,
         3200,
         1470,
         1470,
         1489,
         3200,
         54,
         55,
         54,
         1490,
         1470,
         1471,
         1471,
         54,
         54,
         1510,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2531,
         1490,
         1471,
         1471,
         1510,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3062,
         1510,
         1471,
         1471,
         1470,
         1470,
         3171,
         1510,
         1470,
         1470,
         3200,
         1510,
         1470,
         1470,
         3200,
         1510,
         1470,
         1470,
         3200,
         1510,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         3208,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         3208,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         54,
         55,
         1489,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1490,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         1489,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         1489,
         3061,
         1471,
         1471,
         1470,
         1470,
         1489,
         3171,
         1470,
         1470,
         1489,
         3200,
         1470,
         1470,
         1489,
         3209,
         1470,
         1470,
         1489,
         3209,
         54,
         55,
         54,
         1490,
         1470,
         1471,
         1471,
         54,
         54,
         1510,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2531,
         1490,
         1471,
         1471,
         1510,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3062,
         1510,
         1471,
         1471,
         1470,
         1470,
         3171,
         1510,
         1470,
         1470,
         3200,
         1510,
         1470,
         1470,
         3209,
         1510,
         1470,
         1470,
         3209,
         1510,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         3217,
         1470,
         1470,
         1470,
         1479,
         1470,
         1470,
         1491,
         1492,
         1492,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1493,
         1494,
         1494,
         1494,
         1494,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1470,
         1495,
         1470,
         1470,
         1496,
         1470,
         1470,
         1497,
         1470,
         1498,
         1470,
         1499,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1470,
         1470,
         2523,
         2524,
         2525,
         2526,
         1470,
         1470,
         1470,
         1470,
         1470,
         1492,
         1492,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         1491,
         1491,
         1470,
         1470,
         3217,
         1491,
         1470,
         1470,
         54,
         55,
         1489,
         1489,
         1470,
         1471,
         1471,
         54,
         54,
         54,
         1490,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         1489,
         1489,
         1471,
         1471,
         2522,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         1489,
         3061,
         1471,
         1471,
         1470,
         1470,
         1489,
         3171,
         1470,
         1470,
         1489,
         3200,
         1470,
         1470,
         1489,
         3209,
         1470,
         1470,
         1489,
         3218,
         54,
         55,
         54,
         1490,
         1470,
         1471,
         1471,
         54,
         54,
         1510,
         1510,
         54,
         54,
         1472,
         1473,
         1473,
         1473,
         1473,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         54,
         1474,
         54,
         54,
         1475,
         54,
         54,
         1476,
         54,
         1477,
         54,
         1478,
         54,
         54,
         54,
         2531,
         1490,
         1471,
         1471,
         1510,
         54,
         2514,
         2515,
         2516,
         2517,
         54,
         54,
         54,
         3062,
         1510,
         1471,
         1471,
         1470,
         1470,
         3171,
         1510,
         1470,
         1470,
         3200,
         1510,
         1470,
         1470,
         3209,
         1510,
         1470,
         1470,
         3218,
         1510,
      },
   },
   [nir_op_fsat] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         5,
         5,
         6,
         7,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         56,
         1762,
         1763,
         1764,
         1765,
         2723,
         2724,
         2725,
         2726,
         2727,
         2728,
      },
   },
   [nir_op_umul_unorm_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         57,
         57,
         57,
      },
   },
   [nir_op_fsign] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         58,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         59,
         60,
         59,
         59,
         1980,
         61,
         62,
         61,
         61,
         1981,
         1982,
         1983,
         1982,
         1982,
         1984,
         1985,
         1986,
         1985,
         1985,
         1987,
         59,
         60,
         59,
         59,
         1980,
         63,
         64,
         63,
         63,
         1988,
         65,
         66,
         65,
         65,
         1989,
         1990,
         1991,
         1990,
         1990,
         1992,
         1993,
         1994,
         1993,
         1993,
         1995,
         63,
         64,
         63,
         63,
         1988,
         59,
         60,
         59,
         59,
         1980,
         61,
         62,
         61,
         61,
         1981,
         1996,
         1997,
         1996,
         1996,
         1998,
         1985,
         1986,
         1985,
         1985,
         1987,
         59,
         60,
         59,
         59,
         1980,
         59,
         60,
         59,
         59,
         1980,
         61,
         62,
         61,
         61,
         1981,
         1982,
         1983,
         1982,
         1982,
         1984,
         1999,
         2000,
         1999,
         1999,
         2001,
         59,
         60,
         59,
         59,
         1980,
         59,
         60,
         59,
         59,
         1980,
         61,
         62,
         61,
         61,
         1981,
         1982,
         1983,
         1982,
         1982,
         1984,
         1985,
         1986,
         1985,
         1985,
         1987,
         59,
         60,
         59,
         59,
         1980,
      },
   },
   [nir_op_sdot_4x8_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         67,
         68,
         69,
         70,
         69,
         70,
         71,
         72,
      },
   },
   [nir_op_udot_4x8_uadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         73,
         74,
         75,
         76,
         75,
         76,
         77,
         78,
      },
   },
   [nir_op_sdot_4x8_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         79,
         79,
         80,
         80,
         80,
         80,
         81,
         81,
      },
   },
   [nir_op_udot_4x8_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         82,
         83,
         84,
         85,
         84,
         85,
         86,
         87,
      },
   },
   [nir_op_sdot_2x16_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         88,
         89,
         90,
         91,
         90,
         91,
         92,
         93,
      },
   },
   [nir_op_udot_2x16_uadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         94,
         95,
         96,
         97,
         96,
         97,
         98,
         99,
      },
   },
   [nir_op_sdot_2x16_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         100,
         100,
         101,
         101,
         101,
         101,
         102,
         102,
      },
   },
   [nir_op_udot_2x16_uadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         103,
         103,
         104,
         104,
         104,
         104,
         105,
         105,
      },
   },
   [nir_op_sudot_4x8_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         106,
         107,
         108,
         109,
         110,
         111,
         112,
         113,
      },
   },
   [nir_op_sudot_4x8_iadd_sat] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         114,
         114,
         115,
         115,
         116,
         116,
         117,
         117,
      },
   },
   [nir_op_bcsel] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         4,
         0,
         0,
         0,
         5,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         10,
         0,
         0,
         11,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         0,
         0,
         0,
         26,
         27,
         28,
         0,
         0,
         0,
         0,
         0,
         29,
         30,
         31,
         0,
         0,
         0,
         0,
         0,
         32,
         33,
         33,
         0,
         34,
         34,
         35,
         35,
         36,
         36,
         37,
         37,
         38,
         38,
         0,
         39,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         40,
         0,
         41,
         0,
         42,
         43,
         43,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         44,
         45,
         46,
         46,
         46,
         46,
         47,
         48,
         49,
         0,
         0,
         50,
         51,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         53,
         54,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         55,
         56,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         39,
         39,
         39,
         39,
         39,
         39,
         39,
         40,
         40,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         18,
         18,
         18,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         20,
         19,
         18,
         19,
         18,
         19,
         18,
         18,
         18,
         19,
         19,
         18,
         19,
         18,
         18,
         18,
         19,
         18,
         19,
         18,
         18,
         18,
         18,
         22,
         22,
         22,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         24,
         23,
         22,
         23,
         22,
         23,
         22,
         22,
         22,
         23,
         23,
         23,
         22,
         23,
         22,
         22,
         22,
         23,
         22,
         23,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         0,
         0,
         0,
         0,
         57,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         17,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         32,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         58,
         59,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         60,
         61,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         27,
         27,
         27,
         27,
         27,
         27,
         26,
         0,
         26,
         0,
         0,
         26,
         0,
         26,
         0,
         0,
         26,
         26,
         0,
         0,
         0,
         30,
         30,
         30,
         30,
         30,
         30,
         29,
         0,
         29,
         0,
         0,
         29,
         0,
         29,
         0,
         0,
         29,
         29,
         0,
         62,
         63,
         64,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         37,
         37,
         37,
         37,
         37,
         38,
         38,
         42,
         42,
         42,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         45,
         45,
         45,
         45,
         45,
         47,
         48,
         48,
         48,
         48,
         49,
         44,
         46,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         17,
         17,
         17,
         17,
         17,
         17,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         20,
         19,
         19,
         19,
         19,
         19,
         24,
         23,
         23,
         23,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         34,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         46,
         3,
         3,
         3,
         44,
         44,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         26,
         0,
         29,
         0,
         32,
         0,
         37,
         37,
         37,
         37,
         38,
         38,
         38,
         38,
         38,
         38,
         38,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         39,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         43,
         61,
         61,
         59,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         17,
         17,
         17,
         0,
         0,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         33,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         0,
         44,
         19,
         23,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         35,
         37,
         37,
         37,
         37,
         6,
         10,
         10,
         10,
         19,
         19,
         19,
         34,
         34,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         65,
         0,
         0,
         9,
         13,
         9,
         9,
         0,
         0,
         10,
         10,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         10,
         10,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         10,
         10,
      },
      
      .num_filtered_states = 66,
      .table = (const uint16_t []) {
      
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         124,
         125,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         507,
         124,
         508,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         124,
         122,
         123,
         509,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         510,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         511,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         512,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         513,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         514,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         515,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         516,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         517,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         518,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         519,
         520,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         521,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         519,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         522,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         523,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         524,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         525,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         526,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         527,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         528,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         529,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         530,
         531,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         532,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         530,
         533,
         534,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         535,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         533,
         536,
         537,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         538,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         536,
         539,
         540,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         541,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         539,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         542,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         543,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         544,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         545,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         546,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         547,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         548,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         549,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         550,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         551,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         552,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         2810,
         2811,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2812,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2810,
         2813,
         2814,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2815,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2813,
         2816,
         2817,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2818,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2816,
         2819,
         2820,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2821,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         2819,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         123,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         506,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         122,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         556,
         557,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         558,
         556,
         559,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         556,
         553,
         554,
         560,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         561,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         562,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         563,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         564,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         565,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         566,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         567,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         568,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         569,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         570,
         571,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         572,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         570,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         573,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         574,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         575,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         576,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         577,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         578,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         579,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         580,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         581,
         582,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         583,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         581,
         584,
         585,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         586,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         584,
         587,
         588,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         589,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         587,
         590,
         591,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         592,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         590,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         593,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         594,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         595,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         596,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         597,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         598,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         599,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         600,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         601,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         602,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         603,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         2822,
         2823,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2824,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2822,
         2825,
         2826,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2827,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2825,
         2828,
         2829,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2830,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2828,
         2831,
         2832,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2833,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         2831,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         554,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         555,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         553,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         604,
         605,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         606,
         604,
         607,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         604,
         3194,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         611,
         612,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         613,
         611,
         614,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         611,
         608,
         609,
         615,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         616,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         617,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         618,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         619,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         620,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         621,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         622,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         623,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         624,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         625,
         626,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         627,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         625,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         628,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         629,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         630,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         631,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         632,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         633,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         634,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         635,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         636,
         637,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         638,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         636,
         639,
         640,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         641,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         639,
         642,
         643,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         644,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         642,
         645,
         646,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         647,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         645,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         648,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         649,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         650,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         651,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         652,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         653,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         654,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         655,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         656,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         657,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         658,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         2834,
         2835,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2836,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2834,
         2837,
         2838,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2839,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2837,
         2840,
         2841,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2842,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2840,
         2843,
         2844,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2845,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         2843,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         609,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         610,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         608,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         662,
         663,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         664,
         662,
         665,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         662,
         659,
         660,
         666,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         667,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         668,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         669,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         670,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         671,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         672,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         673,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         674,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         675,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         676,
         677,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         678,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         676,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         679,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         680,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         681,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         682,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         683,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         684,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         685,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         686,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         687,
         688,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         689,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         687,
         690,
         691,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         692,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         690,
         693,
         694,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         695,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         693,
         696,
         697,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         698,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         696,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         699,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         700,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         701,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         702,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         703,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         704,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         705,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         706,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         707,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         708,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         709,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         2846,
         2847,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2848,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2846,
         2849,
         2850,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2851,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2849,
         2852,
         2853,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2854,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2852,
         2855,
         2856,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2857,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         2855,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         660,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         661,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         659,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         713,
         714,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         715,
         713,
         716,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         717,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         713,
         710,
         711,
         718,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         719,
         720,
         719,
         721,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         722,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         719,
         710,
         711,
         710,
         710,
         723,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         724,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         725,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         726,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         727,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         728,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         729,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         730,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         731,
         732,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         733,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         731,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         734,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         735,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         736,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         737,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         738,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         739,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         740,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         741,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         742,
         743,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         744,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         742,
         745,
         746,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         747,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         745,
         748,
         749,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         750,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         748,
         751,
         752,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         753,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         751,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         754,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         755,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         756,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         757,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         758,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         759,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         760,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         761,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         762,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         763,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         764,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         2858,
         2859,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2860,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2858,
         2861,
         2862,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2863,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2861,
         2864,
         2865,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2866,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2864,
         2867,
         2868,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2869,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         2867,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         711,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         712,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         710,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         768,
         769,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         770,
         768,
         771,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         768,
         765,
         766,
         772,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         773,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         774,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         775,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         776,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         777,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         778,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         779,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         780,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         781,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         782,
         783,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         784,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         782,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         785,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         786,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         787,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         788,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         789,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         790,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         791,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         792,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         793,
         794,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         795,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         793,
         796,
         797,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         798,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         796,
         799,
         800,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         801,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         799,
         802,
         803,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         804,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         802,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         805,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         806,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         807,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         808,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         809,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         810,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         811,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         812,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         813,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         814,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         815,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         2870,
         2871,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2872,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2870,
         2873,
         2874,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2875,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2873,
         2876,
         2877,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2878,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2876,
         2879,
         2880,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2881,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         2879,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         766,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         767,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         765,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         819,
         820,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         821,
         819,
         822,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         823,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         819,
         816,
         817,
         824,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         825,
         826,
         825,
         827,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         828,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         825,
         816,
         817,
         816,
         816,
         829,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         830,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         831,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         832,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         833,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         834,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         835,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         836,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         837,
         838,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         839,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         837,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         840,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         841,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         842,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         843,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         844,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         845,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         846,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         847,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         848,
         849,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         850,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         848,
         851,
         852,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         853,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         851,
         854,
         855,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         856,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         854,
         857,
         858,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         859,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         857,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         860,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         861,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         862,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         863,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         864,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         865,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         866,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         867,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         868,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         869,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         870,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         2882,
         2883,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2884,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2882,
         2885,
         2886,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2887,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2885,
         2888,
         2889,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2890,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2888,
         2891,
         2892,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2893,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         2891,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         817,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         818,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         816,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         874,
         875,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         876,
         874,
         877,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         874,
         871,
         872,
         878,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         879,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         880,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         881,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         882,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         883,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         884,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         885,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         886,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         887,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         888,
         889,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         890,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         888,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         891,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         892,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         893,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         894,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         895,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         896,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         897,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         898,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         899,
         900,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         901,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         899,
         902,
         903,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         904,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         902,
         905,
         906,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         907,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         905,
         908,
         909,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         910,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         908,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         911,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         912,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         913,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         914,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         915,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         916,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         917,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         918,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         919,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         920,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         921,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         2894,
         2895,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2896,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2894,
         2897,
         2898,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2899,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2897,
         2900,
         2901,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2902,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2900,
         2903,
         2904,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2905,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         2903,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         872,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         873,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         871,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         925,
         926,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         927,
         925,
         928,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         925,
         922,
         923,
         929,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         930,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         931,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         932,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         933,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         934,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         935,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         936,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         937,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         938,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         939,
         940,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         941,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         939,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         942,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         943,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         944,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         945,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         946,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         947,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         948,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         949,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         950,
         951,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         952,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         950,
         953,
         954,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         955,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         953,
         956,
         957,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         958,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         956,
         959,
         960,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         961,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         959,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         962,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         963,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         964,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         965,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         966,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         967,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         968,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         969,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         970,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         971,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         972,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         2906,
         2907,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2908,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2906,
         2909,
         2910,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2911,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2909,
         2912,
         2913,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2914,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2912,
         2915,
         2916,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2917,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         2915,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         923,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         924,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         922,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         976,
         977,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         978,
         976,
         979,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         976,
         973,
         974,
         980,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         981,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         982,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         983,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         984,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         985,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         986,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         987,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         988,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         989,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         990,
         991,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         992,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         990,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         993,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         994,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         995,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         996,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         997,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         998,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         999,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1000,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1001,
         1002,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1003,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1001,
         1004,
         1005,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1006,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1004,
         1007,
         1008,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1009,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1007,
         1010,
         1011,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1012,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         1010,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1013,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1014,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1015,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1016,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1017,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1018,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1019,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1020,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1021,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1022,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1023,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         2918,
         2919,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2920,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2918,
         2921,
         2922,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2923,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2921,
         2924,
         2925,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2926,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2924,
         2927,
         2928,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2929,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         2927,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         974,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         975,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         973,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1027,
         1028,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1029,
         1027,
         1030,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1027,
         1024,
         1025,
         1031,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1032,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1033,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1034,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1035,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1036,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1037,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1038,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1039,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1040,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1041,
         1042,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1043,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1041,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1044,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1045,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1046,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1047,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1048,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1049,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1050,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1051,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1052,
         1053,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1054,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1052,
         1055,
         1056,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1057,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1055,
         1058,
         1059,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1060,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1058,
         1061,
         1062,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1063,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1061,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1064,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1065,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1066,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1067,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1068,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1069,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1070,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1071,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1072,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1073,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1074,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         2930,
         2931,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2932,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2930,
         2933,
         2934,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2935,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2933,
         2936,
         2937,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2938,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2936,
         2939,
         2940,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2941,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         2939,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1025,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1026,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1024,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1078,
         1079,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1080,
         1078,
         1081,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1078,
         1075,
         1076,
         1082,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1083,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1084,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1085,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1086,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1087,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1088,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1089,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1090,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1091,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1092,
         1093,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1094,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1092,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1095,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1096,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1097,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1098,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1099,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1100,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1101,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1102,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1103,
         1104,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1105,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1103,
         1106,
         1107,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1108,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1106,
         1109,
         1110,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1111,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1109,
         1112,
         1113,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1114,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1112,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1115,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1116,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1117,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1118,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1119,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1120,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1121,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1122,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1123,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1124,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1125,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         2942,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         2943,
         2944,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2945,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2943,
         2946,
         2947,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2948,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2946,
         2949,
         2950,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2951,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2949,
         2952,
         2953,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2954,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         2952,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1076,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1077,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1075,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1129,
         1130,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1131,
         1129,
         1132,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1129,
         1126,
         1127,
         1133,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1134,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1135,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1136,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1137,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1138,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1139,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1140,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1141,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1142,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1143,
         1144,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1145,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1143,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1146,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1147,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1148,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1149,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1150,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1151,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1152,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1153,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1154,
         1155,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1156,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1154,
         1157,
         1158,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1159,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1157,
         1160,
         1161,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1162,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1160,
         1163,
         1164,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1165,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1163,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1166,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1167,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1168,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1169,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1170,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1171,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1172,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1173,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1174,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1175,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1176,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         2955,
         2956,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2957,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2955,
         2958,
         2959,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2960,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2958,
         2961,
         2962,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2963,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2961,
         2964,
         2965,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2966,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         2964,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1127,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1128,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1126,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1180,
         1181,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1182,
         1180,
         1183,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1180,
         1177,
         1178,
         1184,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1185,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1186,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1187,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1188,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1189,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1190,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1191,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1192,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1193,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1194,
         1195,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1196,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1194,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1197,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1198,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1199,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1200,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1201,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1202,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1203,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1204,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1205,
         1206,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1207,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1205,
         1208,
         1209,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1210,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1208,
         1211,
         1212,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1213,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1211,
         1214,
         1215,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1216,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1214,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1217,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1218,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1219,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1220,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1221,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1222,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1223,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1224,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1225,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1226,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1227,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         2967,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         2968,
         2969,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2970,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2968,
         2971,
         2972,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2973,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2971,
         2974,
         2975,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2976,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2974,
         2977,
         2978,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2979,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         2977,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1178,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1179,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1177,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1231,
         1232,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1233,
         1231,
         1234,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1228,
         1229,
         1235,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1236,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1237,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1238,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1239,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1240,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1241,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1242,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1243,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1244,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1245,
         1246,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1247,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1248,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1249,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1250,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1251,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1252,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1253,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1254,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1255,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1256,
         1257,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1258,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1259,
         1260,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1261,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1262,
         1263,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1264,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1265,
         1266,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1267,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1268,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1269,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1270,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1271,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1272,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1273,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1274,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1275,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1276,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1277,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1278,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         2980,
         2981,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2982,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2983,
         2984,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2985,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2986,
         2987,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2988,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2989,
         2990,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2991,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1282,
         1283,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1284,
         1282,
         1285,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1282,
         1279,
         1280,
         1286,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1287,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1288,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1289,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1290,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1291,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1292,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1293,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1294,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1295,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1296,
         1297,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1298,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1296,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1299,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1300,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1301,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1302,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1303,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1304,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1305,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1306,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1307,
         1308,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1309,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1307,
         1310,
         1311,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1312,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1310,
         1313,
         1314,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1315,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1313,
         1316,
         1317,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1318,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1316,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1319,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1320,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1321,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1322,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1323,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1324,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1325,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1326,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1327,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1328,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1329,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         2992,
         2993,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2994,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2992,
         2995,
         2996,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2997,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2995,
         2998,
         2999,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         3000,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         2998,
         3001,
         3002,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3003,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         3001,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1280,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1281,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1279,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1333,
         1334,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1335,
         1333,
         1336,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1333,
         1330,
         1331,
         1337,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1338,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1339,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1340,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1341,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1342,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1343,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1344,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1345,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1346,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1347,
         1348,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1349,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1347,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1350,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1351,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1352,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1353,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1354,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1355,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1356,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1357,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1358,
         1359,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1360,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1358,
         1361,
         1362,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1363,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1361,
         1364,
         1365,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1366,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1364,
         1367,
         1368,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1369,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1367,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1370,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1371,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1372,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1373,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1374,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1375,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1376,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1377,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1378,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1379,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1380,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         3004,
         3005,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3006,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3004,
         3007,
         3008,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3009,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3007,
         3010,
         3011,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3012,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3010,
         3013,
         3014,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3015,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         3013,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1331,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1332,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         1330,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         1381,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         1382,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         1383,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1231,
         1232,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1233,
         1231,
         1234,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1228,
         1229,
         1235,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1236,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1237,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1238,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1239,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1240,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1241,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1242,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1243,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1244,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1245,
         1246,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1247,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1248,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1249,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1250,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1251,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1252,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1253,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1254,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1255,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1256,
         1257,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1258,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1259,
         1260,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1261,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1262,
         1263,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1264,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1265,
         1266,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1267,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1268,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1269,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1270,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1271,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1272,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1273,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1274,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1275,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1276,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         3016,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1277,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1278,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         2980,
         2981,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2982,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2983,
         2984,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2985,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2986,
         2987,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2988,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2989,
         2990,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2991,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1231,
         1232,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1233,
         1231,
         1234,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1228,
         1229,
         1235,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1236,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1237,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1238,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1239,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1240,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1241,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1242,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1243,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1244,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1245,
         1246,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1247,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1248,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1249,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1250,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1251,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1252,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1253,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1254,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1255,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1256,
         1257,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1258,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1259,
         1260,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1261,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1262,
         1263,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1264,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1265,
         1266,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1267,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1268,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1269,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1270,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1271,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1272,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1273,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1274,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1275,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1276,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         3017,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1277,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1278,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         2980,
         2981,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2982,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2983,
         2984,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2985,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2986,
         2987,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2988,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2989,
         2990,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2991,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1231,
         1232,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1233,
         1231,
         1234,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1231,
         1228,
         1229,
         1235,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1236,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1237,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1238,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1239,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1240,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1241,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1242,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1243,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1244,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1245,
         1246,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1247,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1245,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1248,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1249,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1250,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1251,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1252,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1253,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1254,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1255,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1256,
         1257,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1258,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1256,
         1259,
         1260,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1261,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1259,
         1262,
         1263,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1264,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1262,
         1265,
         1266,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1267,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1265,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1268,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1269,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1270,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1271,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1272,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1273,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1274,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1275,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1276,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         3018,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1277,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1278,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         2980,
         2981,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2982,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2980,
         2983,
         2984,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2985,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2983,
         2986,
         2987,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2988,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2986,
         2989,
         2990,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2991,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         2989,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1229,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1230,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         1228,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         120,
         121,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         460,
         120,
         461,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         120,
         118,
         119,
         462,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         463,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         464,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         465,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         466,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         467,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         468,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         469,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         470,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         471,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         472,
         473,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         474,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         472,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         475,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         476,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         477,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         478,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         479,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         480,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         481,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         482,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         483,
         484,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         485,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         483,
         486,
         487,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         488,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         486,
         489,
         490,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         491,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         489,
         492,
         493,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         494,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         492,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         495,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         496,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         497,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         498,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         499,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         500,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         501,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         502,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         503,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         504,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         505,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         2798,
         2799,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2800,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2798,
         2801,
         2802,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2803,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2801,
         2804,
         2805,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2806,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2804,
         2807,
         2808,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2809,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         2807,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         119,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         459,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
         118,
      },
   },
   [nir_op_feq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         12,
         12,
         12,
         12,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         17,
         2,
         2,
         2,
         2,
         2,
         18,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         21,
         4,
         4,
         0,
         2,
         2,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         126,
         127,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         127,
         127,
         1768,
         1769,
         1770,
         1771,
         1772,
         1773,
         1774,
         1775,
         127,
         1776,
         1777,
         1778,
         1779,
         2428,
         2429,
         2430,
         1768,
         2431,
         2432,
         3151,
         1766,
         1768,
         1780,
         1766,
         1781,
         1781,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1780,
         1780,
         1781,
         1766,
         1781,
         126,
         1769,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         1767,
         1770,
         1781,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1781,
         1781,
         1767,
         1767,
         1767,
         1767,
         1771,
         1781,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1781,
         1781,
         1767,
         1767,
         1767,
         126,
         1772,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         1773,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         1774,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         1782,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         2433,
         1767,
         126,
         1767,
         126,
         1775,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         127,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         1783,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         1776,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         1777,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         1778,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         1779,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         2428,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         126,
         2429,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         1766,
         2430,
         1780,
         1766,
         1781,
         1781,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1780,
         1780,
         1781,
         1766,
         1781,
         1766,
         1768,
         1780,
         1766,
         1781,
         1781,
         1766,
         1766,
         2433,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1766,
         1780,
         2434,
         1781,
         1766,
         1781,
         1767,
         2431,
         1781,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1781,
         1781,
         1767,
         1767,
         1767,
         126,
         2432,
         1766,
         126,
         1767,
         1767,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         126,
         1766,
         1766,
         1767,
         126,
         1767,
         1767,
         3151,
         1781,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1767,
         1781,
         1781,
         1767,
         1767,
         1767,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         1,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         128,
         2225,
         2226,
         2227,
      },
   },
   [nir_op_ftrunc] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         129,
      },
   },
   [nir_op_ffloor] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         130,
      },
   },
   [nir_op_ffract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         131,
      },
   },
   [nir_op_fceil] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         132,
      },
   },
   [nir_search_op_b2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         1,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         1,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         133,
         2228,
         2229,
         2230,
         2231,
         3019,
         3020,
      },
   },
   [nir_op_fdph] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         134,
      },
   },
   [nir_op_fdot4] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         135,
         2215,
         2216,
         2217,
         2215,
         2215,
         2216,
         2217,
         2216,
         2216,
         2216,
         2217,
         2217,
         2217,
         2217,
         2217,
      },
   },
   [nir_op_vec4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         136,
         137,
         136,
         138,
         136,
         137,
         136,
         139,
         136,
         137,
         136,
         138,
         136,
         137,
         136,
         139,
      },
   },
   [nir_op_fdot3] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         140,
         2218,
         2219,
         2218,
         2218,
         2219,
         2219,
         2219,
         2219,
      },
   },
   [nir_op_vec3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         141,
         142,
         141,
         143,
         141,
         142,
         141,
         143,
      },
   },
   [nir_op_fdot2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         144,
         145,
         2220,
         145,
         145,
         2221,
         2220,
         2221,
         2220,
      },
   },
   [nir_op_vec2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         146,
         147,
         146,
         146,
         148,
         149,
         148,
         148,
         146,
         147,
         146,
         1615,
         146,
         147,
         146,
         146,
      },
   },
   [nir_op_fsum2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         150,
      },
   },
   [nir_op_ushr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         8,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         4,
         13,
         0,
         3,
         3,
         8,
         4,
         14,
         9,
         2,
         4,
         15,
         3,
         3,
         8,
         4,
         16,
         9,
         2,
         4,
         17,
         3,
         3,
         8,
         4,
         0,
         9,
      },
      
      .num_filtered_states = 18,
      .table = (const uint16_t []) {
      
         151,
         152,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         153,
         154,
         153,
         1591,
         153,
         1592,
         153,
         2422,
         2423,
         153,
         153,
         1591,
         153,
         1591,
         153,
         1591,
         153,
         1591,
         1593,
         1594,
         1593,
         1595,
         1593,
         1596,
         1593,
         2424,
         2425,
         1593,
         1593,
         1595,
         1593,
         1595,
         1593,
         1595,
         1593,
         1595,
         151,
         152,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1597,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         152,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1598,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         152,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         1593,
         1594,
         1593,
         1595,
         1593,
         1596,
         1593,
         2424,
         2425,
         1593,
         1593,
         1595,
         1593,
         1595,
         1593,
         1595,
         1593,
         1595,
         151,
         2426,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         2427,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3101,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3192,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3198,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3202,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3207,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3211,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         3216,
         151,
         1589,
         151,
         1590,
         151,
         2420,
         2421,
         151,
         151,
         1589,
         151,
         1589,
         151,
         1589,
         151,
         1589,
      },
   },
   [nir_op_ishr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         5,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         4,
         0,
         4,
         4,
         2,
         0,
         0,
         0,
         2,
         0,
         4,
         4,
         4,
         2,
         0,
         0,
         0,
         2,
         0,
         4,
         4,
         4,
         2,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         155,
         156,
         155,
         155,
         1599,
         155,
         157,
         158,
         157,
         157,
         1600,
         157,
         155,
         1601,
         155,
         155,
         1599,
         155,
         1602,
         1603,
         1602,
         1602,
         1604,
         1602,
         155,
         156,
         155,
         155,
         1599,
         155,
         155,
         1605,
         155,
         155,
         1599,
         155,
      },
   },
   [nir_op_inot] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         5,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         8,
         8,
         9,
         9,
         9,
         9,
         10,
         10,
         10,
         10,
         11,
         11,
         11,
         11,
         12,
         12,
         13,
         13,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         14,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         21,
         0,
         0,
         0,
         22,
         0,
         0,
         0,
         0,
         0,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         0,
         22,
         22,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         21,
         23,
         21,
         21,
         21,
         21,
         0,
         22,
         0,
         0,
         0,
         0,
         22,
         22,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         11,
         13,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         8,
         0,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 24,
      .table = (const uint16_t []) {
      
         159,
         1966,
         1967,
         1968,
         1969,
         1970,
         1971,
         1972,
         1973,
         1974,
         1975,
         1976,
         1977,
         1978,
         1979,
         2574,
         2575,
         2576,
         2577,
         2578,
         2579,
         3097,
         3098,
         3099,
      },
   },
   [nir_op_flt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         4,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         13,
         2,
         14,
         2,
         2,
         2,
         2,
         2,
         15,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         16,
         5,
         17,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         16,
         5,
         5,
         5,
         5,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         11,
         7,
         7,
         7,
         7,
         7,
         7,
         16,
         5,
         17,
         17,
         17,
         17,
         17,
         17,
         6,
         6,
         16,
         16,
         5,
         16,
         16,
         5,
         0,
         19,
         2,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         20,
         6,
         6,
         21,
         16,
         16,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         16,
         16,
         5,
         5,
         5,
         5,
         16,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         22,
         9,
         9,
         22,
         9,
         9,
         22,
         9,
         22,
         22,
         9,
         22,
         9,
         22,
         9,
         22,
         22,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         0,
         17,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 23,
      .table = (const uint16_t []) {
      
         160,
         161,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         162,
         163,
         1787,
         1788,
         1789,
         162,
         162,
         1790,
         1791,
         1792,
         1793,
         1794,
         162,
         1787,
         1787,
         1787,
         162,
         162,
         2668,
         1787,
         162,
         162,
         1792,
         160,
         1795,
         1796,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1796,
         1796,
         1796,
         160,
         160,
         160,
         1796,
         160,
         160,
         1785,
         160,
         161,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         160,
         1797,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         1798,
         1799,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         1798,
         1803,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         160,
         161,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         160,
         1804,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         1805,
         1806,
         1807,
         1805,
         1805,
         1805,
         1805,
         1805,
         1805,
         1808,
         1805,
         1809,
         1805,
         1807,
         1807,
         1807,
         1805,
         1805,
         1805,
         1807,
         1805,
         1805,
         1808,
         160,
         1810,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         1811,
         1812,
         1813,
         1811,
         1811,
         1811,
         1811,
         1811,
         1811,
         1814,
         1811,
         1815,
         1811,
         1813,
         1813,
         1813,
         1811,
         1811,
         1811,
         1813,
         1811,
         1811,
         1814,
         160,
         161,
         1784,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         1816,
         1784,
         1784,
         1784,
         160,
         160,
         160,
         1784,
         160,
         160,
         1785,
         160,
         2669,
         1796,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1796,
         1796,
         1796,
         160,
         160,
         160,
         1796,
         160,
         160,
         1785,
         160,
         2670,
         1796,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1796,
         1796,
         1796,
         160,
         160,
         160,
         1796,
         160,
         160,
         1785,
         160,
         2671,
         1796,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1796,
         1796,
         1796,
         160,
         160,
         160,
         1796,
         160,
         160,
         1785,
         1798,
         2672,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         1798,
         2673,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         1798,
         1799,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         160,
         3154,
         1796,
         160,
         160,
         160,
         160,
         160,
         160,
         1785,
         160,
         1786,
         160,
         1796,
         1796,
         1796,
         160,
         160,
         160,
         1796,
         160,
         160,
         1785,
         1798,
         3172,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         1798,
         3173,
         1800,
         1798,
         1798,
         1798,
         1798,
         1798,
         1798,
         1801,
         1798,
         1802,
         1798,
         1800,
         1800,
         1800,
         1798,
         1798,
         1798,
         1800,
         1798,
         1798,
         1801,
         1805,
         3174,
         1807,
         1805,
         1805,
         1805,
         1805,
         1805,
         1805,
         1808,
         1805,
         1809,
         1805,
         1807,
         1807,
         1807,
         1805,
         1805,
         1805,
         1807,
         1805,
         1805,
         1808,
      },
   },
   [nir_op_fge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         4,
         0,
         0,
         0,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         8,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         11,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         14,
         2,
         15,
         2,
         2,
         2,
         2,
         2,
         16,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         12,
         7,
         7,
         7,
         7,
         7,
         7,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         18,
         2,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         164,
         165,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         166,
         167,
         1820,
         1821,
         1822,
         166,
         166,
         166,
         1823,
         1824,
         1825,
         1826,
         1827,
         166,
         1820,
         1820,
         1820,
         2674,
         1820,
         164,
         1828,
         1829,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1829,
         1829,
         1829,
         164,
         1829,
         164,
         165,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         164,
         1830,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         1831,
         1832,
         1833,
         1831,
         1831,
         1831,
         1831,
         1831,
         1831,
         1831,
         1834,
         1831,
         1835,
         1831,
         1833,
         1833,
         1833,
         1831,
         1833,
         1831,
         1836,
         1833,
         1831,
         1831,
         1831,
         1831,
         1831,
         1831,
         1831,
         1834,
         1831,
         1835,
         1831,
         1833,
         1833,
         1833,
         1831,
         1833,
         164,
         1837,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         164,
         1838,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         164,
         165,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         1839,
         1840,
         1841,
         1839,
         1839,
         1839,
         1839,
         1839,
         1839,
         1839,
         1842,
         1839,
         1843,
         1839,
         1841,
         1841,
         1841,
         1839,
         1841,
         164,
         1844,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1817,
         1817,
         1817,
         164,
         1817,
         1845,
         1846,
         1847,
         1845,
         1845,
         1845,
         1845,
         1845,
         1845,
         1845,
         1848,
         1845,
         1849,
         1845,
         1847,
         1847,
         1847,
         1845,
         1847,
         164,
         165,
         1817,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         1850,
         1817,
         1817,
         1817,
         164,
         1817,
         164,
         2675,
         1829,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1829,
         1829,
         1829,
         164,
         1829,
         164,
         2676,
         1829,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1829,
         1829,
         1829,
         164,
         1829,
         164,
         2677,
         1829,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1829,
         1829,
         1829,
         164,
         1829,
         1831,
         1832,
         1833,
         1831,
         1831,
         1831,
         1831,
         1831,
         1831,
         1831,
         1834,
         1831,
         1835,
         1831,
         1833,
         1833,
         1833,
         1831,
         1833,
         164,
         3155,
         1829,
         164,
         164,
         164,
         164,
         164,
         164,
         164,
         1818,
         164,
         1819,
         164,
         1829,
         1829,
         1829,
         164,
         1829,
      },
   },
   [nir_op_fneu] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         7,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         11,
         12,
         12,
         12,
         12,
         13,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         17,
         2,
         2,
         2,
         2,
         2,
         18,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         13,
         13,
         13,
         13,
         14,
         14,
         14,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         21,
         4,
         4,
         0,
         2,
         2,
         18,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         5,
         5,
         5,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         21,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         168,
         169,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         169,
         169,
         1853,
         1854,
         1855,
         1856,
         1857,
         1858,
         1859,
         1860,
         169,
         1861,
         1862,
         1863,
         1864,
         2435,
         2436,
         2437,
         1853,
         2438,
         2439,
         3152,
         1851,
         1853,
         1865,
         1851,
         1866,
         1866,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1865,
         1865,
         1866,
         1851,
         1866,
         168,
         1854,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         1852,
         1855,
         1866,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1866,
         1866,
         1852,
         1852,
         1852,
         1852,
         1856,
         1866,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1866,
         1866,
         1852,
         1852,
         1852,
         168,
         1857,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         1858,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         1859,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         1867,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         2440,
         1852,
         168,
         1852,
         168,
         1860,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         169,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         1868,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         1861,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         1862,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         1863,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         1864,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         2435,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         168,
         2436,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         1851,
         2437,
         1865,
         1851,
         1866,
         1866,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1865,
         1865,
         1866,
         1851,
         1866,
         1851,
         1853,
         1865,
         1851,
         1866,
         1866,
         1851,
         1851,
         2440,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1851,
         1865,
         2441,
         1866,
         1851,
         1866,
         1852,
         2438,
         1866,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1866,
         1866,
         1852,
         1852,
         1852,
         168,
         2439,
         1851,
         168,
         1852,
         1852,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         168,
         1851,
         1851,
         1852,
         168,
         1852,
         1852,
         3152,
         1866,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1852,
         1866,
         1866,
         1852,
         1852,
         1852,
      },
   },
   [nir_op_ilt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         170,
         171,
         170,
         2136,
         2136,
         2137,
         2137,
         170,
         170,
         172,
         173,
         2138,
         2139,
         2140,
         2141,
         2142,
         2143,
         172,
         170,
         2144,
         170,
         2136,
         2136,
         2137,
         2137,
         170,
         170,
         2145,
         2146,
         2145,
         2147,
         2147,
         2148,
         2148,
         2145,
         2145,
         2145,
         2149,
         2145,
         2147,
         2147,
         2148,
         2148,
         2145,
         2145,
         2150,
         2151,
         2150,
         2152,
         2152,
         2153,
         2153,
         2150,
         2150,
         2150,
         2154,
         2150,
         2152,
         2152,
         2153,
         2153,
         2150,
         2150,
         170,
         2155,
         170,
         2136,
         2136,
         2137,
         2137,
         2156,
         170,
         170,
         2781,
         170,
         2136,
         2136,
         2137,
         2137,
         170,
         170,
      },
   },
   [nir_op_ult] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         5,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         7,
         7,
         8,
         8,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         10,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         3,
         3,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         3,
         3,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         3,
         3,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         174,
         175,
         2033,
         174,
         174,
         2034,
         2034,
         2035,
         2035,
         174,
         174,
         176,
         177,
         2036,
         176,
         2037,
         2038,
         2039,
         2040,
         2041,
         2042,
         2782,
         174,
         175,
         2033,
         174,
         174,
         2034,
         2034,
         2035,
         2035,
         174,
         174,
         2043,
         2044,
         2045,
         2043,
         2043,
         2046,
         2046,
         2047,
         2047,
         2043,
         2043,
         174,
         2048,
         2033,
         174,
         174,
         2034,
         2034,
         2035,
         2035,
         174,
         174,
         2049,
         2050,
         2051,
         2049,
         2049,
         2052,
         2052,
         2053,
         2053,
         2049,
         2049,
         2049,
         2054,
         2051,
         2049,
         2049,
         2052,
         2052,
         2053,
         2053,
         2049,
         2049,
         2055,
         2056,
         2057,
         2055,
         2055,
         2058,
         2058,
         2059,
         2059,
         2055,
         2055,
         2055,
         2060,
         2057,
         2055,
         2055,
         2058,
         2058,
         2059,
         2059,
         2055,
         2055,
         174,
         2061,
         2033,
         174,
         174,
         2034,
         2034,
         2035,
         2035,
         2062,
         174,
         174,
         175,
         2033,
         174,
         174,
         2034,
         2034,
         2035,
         2035,
         174,
         174,
      },
   },
   [nir_op_ige] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         178,
         179,
         178,
         2157,
         2157,
         2158,
         2158,
         178,
         178,
         178,
         178,
         178,
         180,
         181,
         2159,
         2160,
         2161,
         2162,
         2163,
         2164,
         180,
         180,
         180,
         180,
         178,
         2165,
         178,
         2157,
         2157,
         2158,
         2158,
         178,
         178,
         178,
         178,
         178,
         2166,
         2167,
         2166,
         2168,
         2168,
         2169,
         2169,
         2166,
         2166,
         2166,
         2166,
         2166,
         2166,
         2170,
         2166,
         2168,
         2168,
         2169,
         2169,
         2166,
         2166,
         2166,
         2166,
         2166,
         2171,
         2172,
         2171,
         2173,
         2173,
         2174,
         2174,
         2171,
         2171,
         2171,
         2171,
         2171,
         2171,
         2175,
         2171,
         2173,
         2173,
         2174,
         2174,
         2171,
         2171,
         2171,
         2171,
         2171,
         178,
         2176,
         178,
         2157,
         2157,
         2158,
         2158,
         2177,
         178,
         178,
         178,
         178,
         178,
         2178,
         178,
         2157,
         2157,
         2158,
         2158,
         178,
         178,
         178,
         178,
         178,
         178,
         2179,
         178,
         2157,
         2157,
         2158,
         2158,
         178,
         178,
         178,
         178,
         178,
         178,
         2180,
         178,
         2157,
         2157,
         2158,
         2158,
         178,
         178,
         178,
         178,
         178,
         178,
         2783,
         178,
         2157,
         2157,
         2158,
         2158,
         178,
         178,
         178,
         178,
         178,
      },
   },
   [nir_op_uge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         6,
         6,
         5,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         7,
         7,
         8,
         8,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         5,
         5,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         3,
         3,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         3,
         3,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         3,
         3,
      },
      
      .num_filtered_states = 10,
      .table = (const uint16_t []) {
      
         182,
         182,
         2063,
         182,
         182,
         2064,
         2064,
         2065,
         2065,
         182,
         183,
         183,
         2066,
         183,
         2067,
         2068,
         2069,
         2070,
         2071,
         2072,
         182,
         182,
         2063,
         182,
         182,
         2064,
         2064,
         2065,
         2065,
         182,
         2073,
         2073,
         2074,
         2073,
         2073,
         2075,
         2075,
         2076,
         2076,
         2073,
         182,
         2077,
         2063,
         182,
         182,
         2064,
         2064,
         2065,
         2065,
         182,
         2078,
         2078,
         2079,
         2078,
         2078,
         2080,
         2080,
         2081,
         2081,
         2078,
         2078,
         2082,
         2079,
         2078,
         2078,
         2080,
         2080,
         2081,
         2081,
         2078,
         2083,
         2083,
         2084,
         2083,
         2083,
         2085,
         2085,
         2086,
         2086,
         2083,
         2083,
         2087,
         2084,
         2083,
         2083,
         2085,
         2085,
         2086,
         2086,
         2083,
         182,
         2088,
         2063,
         182,
         182,
         2064,
         2064,
         2065,
         2065,
         2089,
      },
   },
   [nir_op_ieq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         12,
         0,
         13,
         0,
         14,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         11,
         11,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         9,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         0,
         3,
         3,
         0,
         6,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         0,
         6,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         0,
         6,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         184,
         185,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         185,
         185,
         2003,
         2004,
         2005,
         2006,
         2007,
         2008,
         2009,
         2010,
         2011,
         2012,
         2013,
         2014,
         2015,
         2784,
         2002,
         2003,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         2002,
         184,
         2004,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2005,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2006,
         2002,
         184,
         184,
         2016,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2007,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2008,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2009,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2010,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         2017,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2011,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2018,
         184,
         184,
         184,
         184,
         184,
         184,
         2012,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2013,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2014,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2015,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         2784,
         2002,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
         184,
      },
   },
   [nir_op_ine] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         6,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         7,
         0,
         7,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         8,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         0,
         3,
         3,
         0,
         6,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         0,
         6,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         0,
         6,
         0,
         0,
      },
      
      .num_filtered_states = 13,
      .table = (const uint16_t []) {
      
         186,
         187,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         187,
         187,
         2020,
         2021,
         2022,
         2023,
         2024,
         2025,
         2026,
         2027,
         2028,
         2029,
         2785,
         2019,
         2020,
         2019,
         2019,
         2019,
         2019,
         2019,
         2019,
         2019,
         2019,
         2019,
         2019,
         2019,
         186,
         2021,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2022,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2023,
         2019,
         186,
         186,
         2030,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2024,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2025,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2026,
         2019,
         186,
         186,
         186,
         186,
         186,
         2031,
         186,
         186,
         186,
         186,
         186,
         2027,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         2032,
         186,
         186,
         186,
         186,
         2028,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2029,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2785,
         2019,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         186,
         2786,
      },
   },
   [nir_op_fmin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         9,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         10,
         2,
         2,
         11,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         12,
         6,
         6,
         6,
         6,
         6,
         12,
         6,
         6,
         6,
         12,
         6,
         12,
         6,
         6,
         12,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         12,
         6,
         6,
         12,
         12,
         6,
         6,
         12,
         12,
         12,
         6,
         6,
         12,
         6,
         12,
         6,
         12,
         7,
         7,
         13,
         7,
         8,
         8,
         8,
         14,
         14,
         8,
         13,
         7,
         7,
         7,
         13,
         7,
         7,
         7,
         13,
         7,
         13,
         7,
         7,
         13,
         7,
         13,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         2,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         12,
         12,
         6,
         6,
         6,
         6,
         6,
         12,
         12,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         188,
         189,
         1869,
         1870,
         188,
         188,
         1871,
         1872,
         1872,
         2678,
         1869,
         2679,
         2680,
         3102,
         3102,
         3103,
         189,
         189,
         1873,
         1874,
         1875,
         1876,
         1877,
         1878,
         1879,
         2681,
         1873,
         2682,
         2683,
         3104,
         3105,
         3106,
         1869,
         1873,
         1869,
         1880,
         1869,
         1869,
         1881,
         1882,
         1882,
         2678,
         1869,
         2679,
         2684,
         3107,
         3107,
         3103,
         1870,
         1874,
         1880,
         1870,
         1870,
         1870,
         1883,
         1884,
         1884,
         2685,
         1880,
         2686,
         2687,
         3108,
         3108,
         3109,
         188,
         1875,
         1869,
         1870,
         188,
         188,
         1871,
         1872,
         1872,
         2678,
         1869,
         2679,
         2680,
         3102,
         3102,
         3103,
         188,
         1876,
         1869,
         1870,
         188,
         1885,
         1871,
         1872,
         1872,
         2678,
         1869,
         2679,
         2680,
         3102,
         3102,
         3103,
         1871,
         1877,
         1881,
         1883,
         1871,
         1871,
         1871,
         1886,
         1886,
         2688,
         1881,
         2689,
         2680,
         3110,
         3110,
         3111,
         1872,
         1878,
         1882,
         1884,
         1872,
         1872,
         1886,
         1872,
         1872,
         2690,
         1882,
         2691,
         2692,
         3102,
         3102,
         3112,
         1872,
         1879,
         1882,
         1884,
         1872,
         1872,
         1886,
         1872,
         1872,
         2690,
         1882,
         2691,
         2692,
         3102,
         3102,
         3112,
         2678,
         2681,
         2678,
         2685,
         2678,
         2678,
         2688,
         2690,
         2690,
         2678,
         2678,
         2693,
         2694,
         3113,
         3113,
         3114,
         1869,
         1873,
         1869,
         1880,
         1869,
         1869,
         1881,
         1882,
         1882,
         2678,
         2695,
         2679,
         2684,
         3107,
         3107,
         3103,
         2679,
         2682,
         2679,
         2686,
         2679,
         2679,
         2689,
         2691,
         2691,
         2693,
         2679,
         2679,
         2696,
         3115,
         3115,
         3116,
         2680,
         2683,
         2684,
         2687,
         2680,
         2680,
         2680,
         2692,
         2692,
         2694,
         2684,
         2696,
         2680,
         3117,
         3117,
         3118,
         3102,
         3104,
         3107,
         3108,
         3102,
         3102,
         3110,
         3102,
         3102,
         3113,
         3107,
         3115,
         3117,
         3102,
         3102,
         3119,
         3102,
         3105,
         3107,
         3108,
         3102,
         3102,
         3110,
         3102,
         3102,
         3113,
         3107,
         3115,
         3117,
         3102,
         3102,
         3119,
         3103,
         3106,
         3103,
         3109,
         3103,
         3103,
         3111,
         3112,
         3112,
         3114,
         3103,
         3116,
         3118,
         3119,
         3119,
         3103,
      },
   },
   [nir_search_op_i2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         190,
         2199,
         2200,
         2201,
         2202,
         3190,
      },
   },
   [nir_op_fmax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         8,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         9,
         2,
         10,
         2,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         6,
         6,
         6,
         6,
         6,
         13,
         13,
         0,
         0,
         12,
         0,
         12,
         0,
         12,
         7,
         7,
         14,
         7,
         7,
         7,
         7,
         14,
         7,
         14,
         14,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         12,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         14,
         7,
         7,
         7,
         7,
         7,
         14,
         7,
         14,
         7,
         7,
         7,
         14,
         14,
         14,
         14,
         7,
         7,
         14,
         7,
         7,
         14,
         14,
         4,
         4,
         4,
         4,
         4,
         4,
         0,
         0,
         11,
         11,
         11,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         13,
         13,
         6,
         12,
         12,
         0,
         12,
         0,
         12,
         12,
         0,
         12,
         0,
         12,
         0,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 17,
      .table = (const uint16_t []) {
      
         191,
         192,
         1887,
         1888,
         191,
         191,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
         192,
         192,
         1890,
         1891,
         1892,
         192,
         1893,
         1894,
         2701,
         1890,
         2702,
         2703,
         2704,
         2705,
         2706,
         3175,
         3176,
         1887,
         1890,
         1887,
         1895,
         1887,
         1887,
         1887,
         1896,
         2697,
         1887,
         2698,
         1887,
         2707,
         2707,
         2708,
         1887,
         1887,
         1888,
         1891,
         1895,
         1888,
         1888,
         1888,
         1888,
         1897,
         2709,
         1895,
         2710,
         1888,
         2711,
         2711,
         2712,
         1888,
         1888,
         191,
         1892,
         1887,
         1888,
         191,
         191,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
         191,
         192,
         1887,
         1888,
         191,
         1898,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
         191,
         1893,
         1887,
         1888,
         191,
         191,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
         1889,
         1894,
         1896,
         1897,
         1889,
         1889,
         1889,
         1889,
         2713,
         1896,
         2714,
         1889,
         2715,
         2715,
         2700,
         1889,
         1889,
         2697,
         2701,
         2697,
         2709,
         2697,
         2697,
         2697,
         2713,
         2697,
         2697,
         2716,
         2697,
         2717,
         2717,
         2718,
         2697,
         2697,
         1887,
         1890,
         1887,
         1895,
         1887,
         1887,
         1887,
         1896,
         2697,
         2719,
         2698,
         1887,
         2707,
         2707,
         2708,
         1887,
         1887,
         2698,
         2702,
         2698,
         2710,
         2698,
         2698,
         2698,
         2714,
         2716,
         2698,
         2698,
         2698,
         2720,
         2720,
         2721,
         2698,
         2698,
         191,
         2703,
         1887,
         1888,
         191,
         191,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
         2699,
         2704,
         2707,
         2711,
         2699,
         2699,
         2699,
         2715,
         2717,
         2707,
         2720,
         2699,
         2699,
         2699,
         2722,
         2699,
         2699,
         2699,
         2705,
         2707,
         2711,
         2699,
         2699,
         2699,
         2715,
         2717,
         2707,
         2720,
         2699,
         2699,
         2699,
         2722,
         2699,
         2699,
         2700,
         2706,
         2708,
         2712,
         2700,
         2700,
         2700,
         2700,
         2718,
         2708,
         2721,
         2700,
         2722,
         2722,
         2700,
         2700,
         2700,
         191,
         3175,
         1887,
         1888,
         191,
         191,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
         191,
         3176,
         1887,
         1888,
         191,
         191,
         191,
         1889,
         2697,
         1887,
         2698,
         191,
         2699,
         2699,
         2700,
         191,
         191,
      },
   },
   [nir_op_imin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         7,
         2,
         8,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         9,
         4,
         4,
         4,
         9,
         4,
         4,
         4,
         9,
         4,
         9,
         4,
         9,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         9,
         4,
         9,
         9,
         4,
         9,
         9,
         4,
         9,
         9,
         4,
         5,
         10,
         5,
         6,
         11,
         11,
         6,
         10,
         5,
         5,
         10,
         5,
         5,
         10,
         10,
         5,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         4,
         9,
         4,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         193,
         194,
         1912,
         1913,
         1914,
         1915,
         1915,
         2751,
         1912,
         2752,
         3156,
         3156,
         194,
         194,
         1916,
         1917,
         1918,
         1919,
         1920,
         2753,
         1916,
         2754,
         3157,
         3158,
         1912,
         1916,
         1912,
         1921,
         1922,
         1923,
         1923,
         2751,
         1912,
         2755,
         3159,
         3159,
         1913,
         1917,
         1921,
         1913,
         1924,
         1925,
         1925,
         2756,
         1921,
         2757,
         3160,
         3160,
         1914,
         1918,
         1922,
         1924,
         1914,
         1926,
         1926,
         2758,
         1922,
         2752,
         3161,
         3161,
         1915,
         1919,
         1923,
         1925,
         1926,
         1915,
         1915,
         2759,
         1923,
         2760,
         3156,
         3156,
         1915,
         1920,
         1923,
         1925,
         1926,
         1915,
         1915,
         2759,
         1923,
         2760,
         3156,
         3156,
         2751,
         2753,
         2751,
         2756,
         2758,
         2759,
         2759,
         2751,
         2751,
         2761,
         3162,
         3162,
         1912,
         1916,
         1912,
         1921,
         1922,
         1923,
         1923,
         2751,
         2762,
         2755,
         3159,
         3159,
         2752,
         2754,
         2755,
         2757,
         2752,
         2760,
         2760,
         2761,
         2755,
         2752,
         3163,
         3163,
         3156,
         3157,
         3159,
         3160,
         3161,
         3156,
         3156,
         3162,
         3159,
         3163,
         3156,
         3156,
         3156,
         3158,
         3159,
         3160,
         3161,
         3156,
         3156,
         3162,
         3159,
         3163,
         3156,
         3156,
      },
   },
   [nir_op_imax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         6,
         2,
         7,
         2,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         4,
         4,
         4,
         9,
         9,
         0,
         0,
         8,
         0,
         8,
         8,
         5,
         5,
         10,
         5,
         5,
         5,
         10,
         5,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         8,
         8,
         0,
         0,
         5,
         5,
         10,
         5,
         5,
         5,
         10,
         5,
         10,
         5,
         5,
         10,
         10,
         10,
         5,
         10,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         9,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         195,
         196,
         1927,
         1928,
         195,
         1929,
         2763,
         1927,
         2764,
         2764,
         2765,
         196,
         196,
         1930,
         1931,
         1932,
         1933,
         2766,
         1930,
         2767,
         2768,
         2769,
         1927,
         1930,
         1927,
         1934,
         1927,
         1935,
         2763,
         1927,
         2770,
         2770,
         2771,
         1928,
         1931,
         1934,
         1928,
         1928,
         1936,
         2772,
         1934,
         2773,
         2773,
         2774,
         195,
         1932,
         1927,
         1928,
         195,
         1929,
         2763,
         1927,
         2764,
         2764,
         2765,
         1929,
         1933,
         1935,
         1936,
         1929,
         1929,
         2775,
         1935,
         2776,
         2776,
         2765,
         2763,
         2766,
         2763,
         2772,
         2763,
         2775,
         2763,
         2763,
         2777,
         2777,
         2778,
         1927,
         1930,
         1927,
         1934,
         1927,
         1935,
         2763,
         2779,
         2770,
         2770,
         2771,
         2764,
         2767,
         2770,
         2773,
         2764,
         2776,
         2777,
         2770,
         2764,
         2764,
         2780,
         2764,
         2768,
         2770,
         2773,
         2764,
         2776,
         2777,
         2770,
         2764,
         2764,
         2780,
         2765,
         2769,
         2771,
         2774,
         2765,
         2765,
         2778,
         2771,
         2780,
         2780,
         2765,
      },
   },
   [nir_op_umin] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         5,
         2,
         5,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         2,
         6,
         3,
         6,
         3,
         6,
         3,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         197,
         198,
         2242,
         2243,
         197,
         2787,
         3164,
         198,
         198,
         2244,
         2245,
         198,
         2788,
         3165,
         2242,
         2244,
         2242,
         2246,
         2242,
         2787,
         3166,
         2243,
         2245,
         2246,
         2243,
         2243,
         2789,
         3164,
         197,
         198,
         2242,
         2243,
         2790,
         2787,
         3164,
         2787,
         2788,
         2787,
         2789,
         2787,
         2787,
         3167,
         3164,
         3165,
         3166,
         3164,
         3164,
         3167,
         3164,
      },
   },
   [nir_op_umax] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         2,
         5,
         2,
         5,
         5,
         2,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         199,
         200,
         2247,
         199,
         2791,
         2792,
         200,
         200,
         2248,
         200,
         2793,
         2794,
         2247,
         2248,
         2247,
         2247,
         2795,
         2792,
         199,
         200,
         2247,
         2796,
         2791,
         2792,
         2791,
         2793,
         2795,
         2791,
         2791,
         2797,
         2792,
         2794,
         2792,
         2792,
         2797,
         2792,
      },
   },
   [nir_op_extract_u8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         3,
         0,
         4,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         3,
         3,
         0,
         0,
         4,
         0,
         4,
         0,
         4,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
      },
      
      .num_filtered_states = 10,
      .table = (const uint16_t []) {
      
         201,
         202,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         202,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1606,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1607,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1608,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1609,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1610,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1611,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         1612,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         3021,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
         201,
      },
   },
   [nir_search_op_f2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         203,
         1937,
         1938,
         1939,
         1940,
         1941,
         1942,
      },
   },
   [nir_search_op_f2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         204,
         1943,
         1944,
         1945,
         1946,
         1947,
         1948,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         205,
         1949,
         1950,
         1951,
         1952,
         1953,
         1954,
         1955,
         1956,
         1957,
         1958,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         7,
         0,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         6,
         7,
         7,
         7,
         7,
         7,
         7,
         11,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 12,
      .table = (const uint16_t []) {
      
         206,
         2092,
         2093,
         2094,
         2095,
         2096,
         2097,
         2098,
         2099,
         2100,
         2101,
         3059,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         6,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         8,
         9,
         0,
         10,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         5,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         14,
         9,
         9,
         9,
         9,
         9,
         15,
         9,
         9,
         9,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         12,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         9,
         9,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         9,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         1,
         0,
         1,
         1,
         0,
         6,
         0,
         0,
         0,
         6,
         1,
         1,
         1,
         0,
         6,
         0,
         0,
         0,
         6,
         1,
         1,
         1,
         0,
         6,
         0,
         0,
      },
      
      .num_filtered_states = 19,
      .table = (const uint16_t []) {
      
         207,
         2102,
         2103,
         2104,
         2105,
         2106,
         2107,
         2108,
         2109,
         2110,
         2111,
         2112,
         2113,
         2114,
         3057,
         3058,
         3094,
         3095,
         3096,
      },
   },
   [nir_op_f2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         0,
         4,
         4,
         4,
         0,
         0,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         208,
         1959,
         1960,
         1961,
         1962,
         1963,
         1964,
         1965,
      },
   },
   [nir_op_i2imp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         3,
         3,
         3,
         0,
         0,
         4,
         4,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         6,
         0,
         7,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         6,
         6,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         7,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         8,
         8,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 9,
      .table = (const uint16_t []) {
      
         209,
         2203,
         2204,
         2205,
         2206,
         2207,
         2208,
         2209,
         2210,
      },
   },
   [nir_op_extract_i8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         210,
         211,
         210,
         210,
         210,
         211,
         210,
         210,
         210,
         1613,
         210,
         210,
         210,
         1614,
         210,
         210,
      },
   },
   [nir_op_extract_i16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         212,
         213,
         212,
         213,
      },
   },
   [nir_op_extract_u16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         214,
         215,
         214,
         214,
         214,
         214,
         215,
         214,
         214,
         214,
         214,
         2249,
         214,
         214,
         214,
         214,
         2250,
         214,
         214,
         214,
         214,
         2251,
         214,
         214,
         214,
      },
   },
   [nir_op_ixor] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         3,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         5,
         4,
         5,
         6,
         7,
         6,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         2,
         3,
         2,
         2,
         2,
         3,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         16,
         16,
         16,
         16,
         16,
         17,
         17,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         4,
         4,
         4,
         4,
         4,
         5,
         4,
         4,
         5,
         5,
         6,
         6,
         7,
         6,
         7,
         6,
         7,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         11,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         18,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         20,
         21,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         15,
         15,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         3,
         4,
         4,
         4,
         4,
         4,
         4,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         16,
         16,
         16,
         16,
         17,
         17,
         17,
         17,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         16,
         21,
         21,
         19,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
         3,
         5,
         0,
         0,
         0,
         3,
         5,
         0,
         0,
      },
      
      .num_filtered_states = 22,
      .table = (const uint16_t []) {
      
         216,
         217,
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         217,
         1566,
         217,
         1567,
         217,
         1568,
         217,
         1569,
         1570,
         1571,
         3026,
         3027,
         3028,
         3029,
         216,
         217,
         216,
         1572,
         216,
         216,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         216,
         217,
         1572,
         1572,
         216,
         216,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         216,
         217,
         216,
         216,
         216,
         1573,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         216,
         217,
         216,
         216,
         1573,
         1573,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         216,
         217,
         216,
         216,
         216,
         216,
         216,
         1574,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         216,
         217,
         216,
         216,
         216,
         216,
         1574,
         1574,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         216,
         217,
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         1575,
         216,
         1562,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         1561,
         1566,
         1561,
         1561,
         1561,
         1561,
         1561,
         1561,
         1575,
         1575,
         1561,
         1576,
         1561,
         1577,
         1561,
         1578,
         1579,
         1579,
         3030,
         3031,
         3032,
         3033,
         216,
         217,
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         1561,
         216,
         1580,
         216,
         1563,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         1562,
         1567,
         1562,
         1562,
         1562,
         1562,
         1562,
         1562,
         1562,
         1576,
         1580,
         1580,
         1562,
         1581,
         1562,
         1582,
         1583,
         1583,
         3034,
         3035,
         3036,
         3037,
         216,
         217,
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1584,
         216,
         1564,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         1563,
         1568,
         1563,
         1563,
         1563,
         1563,
         1563,
         1563,
         1563,
         1577,
         1563,
         1581,
         1584,
         1584,
         1563,
         1585,
         1586,
         1586,
         3038,
         3039,
         3040,
         3041,
         216,
         217,
         216,
         216,
         216,
         216,
         216,
         216,
         216,
         1561,
         216,
         1562,
         216,
         1563,
         216,
         1587,
         1565,
         1565,
         3022,
         3023,
         3024,
         3025,
         1564,
         1569,
         1564,
         1564,
         1564,
         1564,
         1564,
         1564,
         1564,
         1578,
         1564,
         1582,
         1564,
         1585,
         1587,
         1587,
         1588,
         1588,
         3042,
         3043,
         3044,
         3045,
         1565,
         1570,
         1565,
         1565,
         1565,
         1565,
         1565,
         1565,
         1565,
         1579,
         1565,
         1583,
         1565,
         1586,
         1565,
         1588,
         1565,
         1565,
         3046,
         3047,
         3048,
         3049,
         1565,
         1571,
         1565,
         1565,
         1565,
         1565,
         1565,
         1565,
         1565,
         1579,
         1565,
         1583,
         1565,
         1586,
         1565,
         1588,
         1565,
         1565,
         3046,
         3047,
         3048,
         3049,
         3022,
         3026,
         3022,
         3022,
         3022,
         3022,
         3022,
         3022,
         3022,
         3030,
         3022,
         3034,
         3022,
         3038,
         3022,
         3042,
         3046,
         3046,
         3022,
         3050,
         3051,
         3052,
         3023,
         3027,
         3023,
         3023,
         3023,
         3023,
         3023,
         3023,
         3023,
         3031,
         3023,
         3035,
         3023,
         3039,
         3023,
         3043,
         3047,
         3047,
         3050,
         3023,
         3053,
         3054,
         3024,
         3028,
         3024,
         3024,
         3024,
         3024,
         3024,
         3024,
         3024,
         3032,
         3024,
         3036,
         3024,
         3040,
         3024,
         3044,
         3048,
         3048,
         3051,
         3053,
         3024,
         3055,
         3025,
         3029,
         3025,
         3025,
         3025,
         3025,
         3025,
         3025,
         3025,
         3033,
         3025,
         3037,
         3025,
         3041,
         3025,
         3045,
         3049,
         3049,
         3052,
         3054,
         3055,
         3056,
      },
   },
   [nir_op_slt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         218,
         219,
         220,
         221,
      },
   },
   [nir_op_sge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         222,
         223,
         224,
         225,
      },
   },
   [nir_op_seq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         226,
         226,
         226,
         226,
         226,
         226,
         226,
         226,
         2252,
         2253,
         2254,
         2255,
         226,
         2252,
         226,
         226,
         226,
         226,
         226,
         2253,
         226,
         226,
         226,
         226,
         226,
         2254,
         226,
         226,
         226,
         226,
         226,
         2255,
         226,
         226,
         226,
         226,
      },
   },
   [nir_op_sne] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         3,
         3,
         3,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         4,
         4,
         5,
         5,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         227,
         227,
         227,
         227,
         227,
         227,
         227,
         227,
         2256,
         2257,
         2258,
         2259,
         227,
         2256,
         227,
         227,
         227,
         227,
         227,
         2257,
         227,
         227,
         227,
         227,
         227,
         2258,
         227,
         227,
         227,
         227,
         227,
         2259,
         227,
         227,
         227,
         227,
      },
   },
   [nir_op_fall_equal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         228,
      },
   },
   [nir_op_fall_equal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         229,
      },
   },
   [nir_op_fall_equal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         230,
      },
   },
   [nir_op_fany_nequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         231,
      },
   },
   [nir_op_fany_nequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         232,
      },
   },
   [nir_op_fany_nequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         233,
      },
   },
   [nir_op_ball_iequal2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         234,
         234,
         234,
         234,
         234,
         234,
         2182,
         2183,
         234,
         2182,
         234,
         234,
         234,
         2183,
         234,
         234,
      },
   },
   [nir_op_ball_iequal3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         235,
         235,
         235,
         235,
         235,
         235,
         2184,
         2185,
         235,
         2184,
         235,
         235,
         235,
         2185,
         235,
         235,
      },
   },
   [nir_op_ball_iequal4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         236,
         236,
         236,
         236,
         236,
         236,
         2186,
         2187,
         236,
         2186,
         236,
         236,
         236,
         2187,
         236,
         236,
      },
   },
   [nir_op_bany_inequal2] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         237,
         237,
         237,
         237,
         237,
         237,
         2232,
         2233,
         237,
         2232,
         237,
         237,
         237,
         2233,
         237,
         237,
      },
   },
   [nir_op_bany_inequal3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         238,
         238,
         238,
         238,
         238,
         238,
         2234,
         2235,
         238,
         2234,
         238,
         238,
         238,
         2235,
         238,
         238,
      },
   },
   [nir_op_bany_inequal4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         239,
         239,
         239,
         239,
         239,
         239,
         2236,
         2237,
         239,
         2236,
         239,
         239,
         239,
         2237,
         239,
         239,
      },
   },
   [nir_op_ball_fequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         240,
      },
   },
   [nir_op_ball_fequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         241,
      },
   },
   [nir_op_ball_fequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         242,
      },
   },
   [nir_op_bany_fnequal2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         243,
      },
   },
   [nir_op_bany_fnequal3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         244,
      },
   },
   [nir_op_bany_fnequal4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         245,
      },
   },
   [nir_op_isub] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         246,
         246,
         247,
         247,
      },
   },
   [nir_op_urol] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         248,
      },
   },
   [nir_op_uror] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         249,
      },
   },
   [nir_op_fexp2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         250,
         2277,
         2749,
         2750,
         3153,
      },
   },
   [nir_op_flog2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         2,
         2,
         2,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         4,
         4,
         4,
         5,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         251,
         2266,
         2267,
         2268,
         2269,
         2270,
      },
   },
   [nir_op_fpow] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         252,
         253,
         252,
         252,
         254,
         255,
         254,
         254,
         252,
         2278,
         252,
         252,
         252,
         2745,
         252,
         252,
      },
   },
   [nir_op_fsqrt] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         256,
         2271,
      },
   },
   [nir_op_frcp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         3,
         3,
         4,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 5,
      .table = (const uint16_t []) {
      
         257,
         2272,
         2273,
         2274,
         2275,
      },
   },
   [nir_op_frsq] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         258,
         2276,
      },
   },
   [nir_op_fdiv] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         259,
         259,
         260,
         260,
      },
   },
   [nir_op_fsin] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         261,
      },
   },
   [nir_op_fcos] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         262,
      },
   },
   [nir_op_fcsel] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         4,
         5,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         263,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2260,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2261,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2262,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2263,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2264,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
         2265,
      },
   },
   [nir_op_f2imp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         264,
         2192,
         2193,
      },
   },
   [nir_op_f2ump] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         265,
         2194,
         2195,
      },
   },
   [nir_op_i2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         266,
         2211,
         2212,
      },
   },
   [nir_op_u2fmp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         267,
         2213,
         2214,
      },
   },
   [nir_op_fround_even] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         268,
      },
   },
   [nir_op_pack_64_2x32_split] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         269,
         269,
         269,
         269,
         269,
         1616,
         269,
         269,
         269,
      },
   },
   [nir_op_unpack_64_2x32_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         270,
         2115,
         2116,
         2117,
      },
   },
   [nir_op_pack_64_2x32] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         271,
         2281,
         2667,
      },
   },
   [nir_op_unpack_64_2x32] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         272,
         2279,
         2280,
      },
   },
   [nir_op_unpack_double_2x32_dxil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         273,
         2283,
      },
   },
   [nir_op_pack_double_2x32_dxil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         274,
         2282,
      },
   },
   [nir_op_unpack_32_2x16_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         275,
         2118,
         2119,
      },
   },
   [nir_op_unpack_32_2x16_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         276,
         2120,
      },
   },
   [nir_op_ubfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         277,
         2121,
         0,
         278,
         2121,
         2122,
         2123,
         2124,
         0,
         277,
         2121,
         0,
         278,
         2121,
         2122,
         2123,
         2124,
         0,
         277,
         2121,
         0,
         278,
         2121,
         2122,
         2123,
         2124,
      },
   },
   [nir_op_ibfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         279,
         2125,
         0,
         280,
         2125,
         2126,
         2127,
         2128,
         0,
         279,
         2125,
         0,
         280,
         2125,
         2126,
         2127,
         2128,
         0,
         279,
         2125,
         0,
         280,
         2125,
         2126,
         2127,
         2128,
      },
   },
   [nir_op_pack_32_2x16_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         281,
      },
   },
   [nir_op_unpack_half_2x16_split_x] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         1,
         0,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
         0,
         2,
         1,
         1,
         1,
         0,
         2,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         282,
         2129,
         2130,
         2131,
      },
   },
   [nir_op_unpack_half_2x16_split_y] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         283,
         2132,
      },
   },
   [nir_op_pack_half_2x16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         2222,
         2223,
         2224,
      },
   },
   [nir_op_pack_half_2x16_split] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         284,
         285,
         286,
         287,
      },
   },
   [nir_op_pack_32_4x8_split] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         288,
      },
   },
   [nir_op_ussub_4x8_vc4] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         289,
         0,
         289,
      },
   },
   [nir_op_fsub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         290,
      },
   },
   [nir_op_uabs_usub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         291,
      },
   },
   [nir_op_uabs_isub] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         292,
      },
   },
   [nir_op_find_lsb] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         293,
      },
   },
   [nir_op_ifind_msb] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         294,
      },
   },
   [nir_op_ufind_msb] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         295,
      },
   },
   [nir_op_fmod] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         296,
      },
   },
   [nir_op_frem] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         297,
      },
   },
   [nir_op_uadd_carry] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         298,
      },
   },
   [nir_op_usub_borrow] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         299,
      },
   },
   [nir_op_bitfield_insert] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         300,
      },
   },
   [nir_op_ihadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         301,
      },
   },
   [nir_op_uhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         302,
      },
   },
   [nir_op_irhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         303,
      },
   },
   [nir_op_urhadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         304,
      },
   },
   [nir_op_imul_32x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         305,
      },
   },
   [nir_op_umul_32x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         306,
      },
   },
   [nir_op_ibitfield_extract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         307,
      },
   },
   [nir_op_ubitfield_extract] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         308,
      },
   },
   [nir_op_bitfield_select] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         2,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         309,
         0,
         0,
         309,
         0,
         2090,
         2091,
         2090,
         0,
         309,
         0,
         0,
         309,
         0,
         2090,
         2091,
         2090,
         0,
         309,
         0,
         0,
         309,
         0,
         2090,
         2091,
         2090,
      },
   },
   [nir_op_bfm] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         2133,
         2134,
         2135,
      },
   },
   [nir_op_pack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         310,
      },
   },
   [nir_op_pack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         311,
      },
   },
   [nir_op_pack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         312,
      },
   },
   [nir_op_pack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         313,
      },
   },
   [nir_op_unpack_unorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         314,
      },
   },
   [nir_op_unpack_unorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         315,
      },
   },
   [nir_op_unpack_snorm_2x16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         316,
      },
   },
   [nir_op_unpack_snorm_4x8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         317,
      },
   },
   [nir_op_isign] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         318,
         2181,
      },
   },
   [nir_op_amul] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         319,
      },
   },
   [nir_op_umul24] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         320,
      },
   },
   [nir_op_umad24] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         321,
      },
   },
   [nir_op_imul24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         322,
      },
   },
   [nir_op_umad24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         323,
      },
   },
   [nir_op_umul24_relaxed] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         324,
      },
   },
   [nir_op_imad24_ir3] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         325,
         326,
         327,
         326,
         327,
         328,
         329,
      },
   },
   [nir_op_imul24] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         330,
         330,
         330,
      },
   },
   [nir_op_b2b1] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         331,
      },
   },
   [nir_op_b2b8] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         332,
      },
   },
   [nir_op_b2b16] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         333,
      },
   },
   [nir_op_b2b32] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         334,
      },
   },
   [nir_op_ldexp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         335,
      },
   },
   [nir_op_ball_iequal8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         2188,
         2189,
         0,
         2188,
         0,
         0,
         0,
         2189,
         0,
         0,
      },
   },
   [nir_op_bany_inequal8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         2238,
         2239,
         0,
         2238,
         0,
         0,
         0,
         2239,
         0,
         0,
      },
   },
   [nir_op_ball_iequal16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         2190,
         2191,
         0,
         2190,
         0,
         0,
         0,
         2191,
         0,
         0,
      },
   },
   [nir_op_bany_inequal16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         2240,
         2241,
         0,
         2240,
         0,
         0,
         0,
         2241,
         0,
         0,
      },
   },
   [nir_op_fddx] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         336,
      },
   },
   [nir_op_fddx_fine] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         337,
      },
   },
   [nir_op_fddx_coarse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         338,
      },
   },
   [nir_op_fddy] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         339,
      },
   },
   [nir_op_fddy_fine] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         340,
      },
   },
   [nir_op_fddy_coarse] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         341,
      },
   },
   [nir_op_umul_low] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         342,
         342,
         342,
      },
   },
   [nir_op_imadsh_mix16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         343,
         343,
         344,
         344,
         345,
         345,
      },
   },
   [nir_op_fsin_amd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         346,
      },
   },
   [nir_op_fcos_amd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         347,
      },
   },
   [nir_op_fisnormal] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         348,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_transform_offsets[] = {
   0,
   0,
   0,
   1,
   0,
   12,
   15,
   17,
   21,
   23,
   25,
   27,
   30,
   34,
   37,
   42,
   0,
   0,
   47,
   0,
   49,
   0,
   51,
   0,
   55,
   0,
   59,
   61,
   67,
   74,
   80,
   87,
   90,
   94,
   97,
   101,
   0,
   104,
   111,
   115,
   121,
   128,
   138,
   150,
   165,
   178,
   194,
   200,
   207,
   216,
   226,
   236,
   247,
   249,
   257,
   259,
   263,
   266,
   269,
   271,
   273,
   277,
   280,
   285,
   288,
   293,
   297,
   303,
   305,
   307,
   310,
   313,
   317,
   321,
   323,
   325,
   328,
   331,
   335,
   339,
   341,
   344,
   348,
   350,
   353,
   356,
   360,
   364,
   369,
   371,
   373,
   376,
   379,
   383,
   387,
   389,
   391,
   394,
   397,
   401,
   405,
   407,
   410,
   414,
   416,
   419,
   423,
   425,
   427,
   430,
   433,
   436,
   439,
   444,
   449,
   451,
   454,
   457,
   462,
   466,
   471,
   476,
   500,
   506,
   513,
   520,
   546,
   548,
   0,
   551,
   554,
   559,
   564,
   0,
   567,
   569,
   0,
   0,
   0,
   0,
   571,
   0,
   0,
   0,
   573,
   575,
   0,
   0,
   0,
   0,
   578,
   0,
   580,
   586,
   588,
   0,
   595,
   601,
   604,
   0,
   612,
   617,
   622,
   627,
   632,
   637,
   642,
   647,
   652,
   654,
   657,
   662,
   668,
   673,
   679,
   681,
   684,
   688,
   693,
   698,
   704,
   709,
   715,
   717,
   719,
   721,
   727,
   729,
   735,
   737,
   0,
   740,
   742,
   744,
   746,
   748,
   750,
   752,
   754,
   758,
   760,
   764,
   766,
   0,
   0,
   770,
   774,
   780,
   0,
   0,
   786,
   788,
   792,
   794,
   797,
   799,
   802,
   804,
   808,
   810,
   812,
   814,
   816,
   818,
   820,
   822,
   824,
   826,
   828,
   830,
   832,
   834,
   836,
   838,
   840,
   842,
   844,
   846,
   848,
   850,
   852,
   854,
   856,
   858,
   860,
   862,
   864,
   866,
   868,
   871,
   0,
   0,
   874,
   876,
   881,
   884,
   890,
   0,
   0,
   892,
   894,
   897,
   899,
   901,
   0,
   0,
   0,
   0,
   903,
   905,
   907,
   0,
   0,
   0,
   0,
   909,
   912,
   915,
   917,
   926,
   928,
   936,
   939,
   941,
   943,
   945,
   947,
   949,
   0,
   951,
   954,
   956,
   958,
   960,
   962,
   965,
   967,
   969,
   971,
   973,
   975,
   979,
   982,
   985,
   988,
   991,
   993,
   995,
   998,
   1001,
   1003,
   1005,
   1007,
   1009,
   1011,
   1013,
   1015,
   1017,
   1019,
   1021,
   1023,
   1025,
   1027,
   1030,
   1033,
   1036,
   1038,
   1041,
   1045,
   1049,
   1054,
   1058,
   1060,
   1062,
   1064,
   1066,
   1070,
   1072,
   1074,
   1076,
   1078,
   1080,
   1082,
   1084,
   1086,
   1088,
   0,
   0,
   1091,
   1095,
   1097,
   1100,
   1103,
   1105,
   1107,
   1119,
   1132,
   1144,
   1156,
   1168,
   1181,
   1194,
   1206,
   1218,
   1222,
   1226,
   1229,
   1232,
   1234,
   1239,
   1243,
   1247,
   1251,
   1255,
   1258,
   1260,
   1265,
   0,
   1267,
   1269,
   1271,
   1273,
   1275,
   1278,
   1281,
   1284,
   1286,
   1288,
   1290,
   1293,
   1296,
   1300,
   1304,
   1308,
   1312,
   1316,
   1321,
   1324,
   1328,
   1332,
   1335,
   1338,
   1340,
   1343,
   1346,
   1349,
   1352,
   1355,
   1358,
   1362,
   1366,
   1370,
   1373,
   1376,
   1378,
   1380,
   1382,
   1384,
   1387,
   1390,
   1393,
   1395,
   1397,
   1400,
   1403,
   1406,
   1410,
   1414,
   1418,
   1421,
   1424,
   1427,
   1430,
   1434,
   1438,
   1442,
   1445,
   1448,
   1451,
   1455,
   1459,
   1463,
   1466,
   1469,
   1473,
   1477,
   1481,
   1484,
   1487,
   1492,
   1497,
   1501,
   1505,
   1510,
   1514,
   1518,
   1522,
   1526,
   1529,
   1531,
   1537,
   1544,
   1549,
   1557,
   1562,
   1567,
   1575,
   1583,
   1591,
   1599,
   1607,
   1613,
   1623,
   1629,
   1636,
   1648,
   1653,
   1660,
   1668,
   1676,
   1684,
   1692,
   1700,
   1708,
   1714,
   1721,
   1729,
   1735,
   1742,
   1750,
   1755,
   1761,
   1768,
   1773,
   1779,
   1786,
   1794,
   1799,
   1804,
   1812,
   1817,
   1822,
   1827,
   1832,
   1837,
   1842,
   1847,
   1855,
   1864,
   1871,
   1881,
   1888,
   1895,
   1905,
   1915,
   1925,
   1935,
   1945,
   1953,
   1965,
   1973,
   1982,
   1996,
   2003,
   2012,
   2022,
   2032,
   2042,
   2052,
   2062,
   2072,
   2080,
   2089,
   2099,
   2107,
   2116,
   2126,
   2133,
   2141,
   2150,
   2157,
   2165,
   2174,
   2184,
   2191,
   2198,
   2208,
   2215,
   2222,
   2229,
   2236,
   2243,
   2250,
   2257,
   2262,
   2268,
   2275,
   2281,
   2306,
   2314,
   2320,
   2329,
   2335,
   2341,
   2350,
   2359,
   2368,
   2377,
   2386,
   2393,
   2404,
   2411,
   2419,
   2432,
   2438,
   2446,
   2455,
   2464,
   2473,
   2482,
   2491,
   2500,
   2507,
   2515,
   2524,
   2531,
   2539,
   2548,
   2554,
   2561,
   2569,
   2575,
   2582,
   2590,
   2599,
   2605,
   2611,
   2620,
   2626,
   2632,
   2638,
   2644,
   2650,
   2656,
   2662,
   2667,
   2691,
   2698,
   2703,
   2708,
   2714,
   2721,
   2727,
   2752,
   2760,
   2766,
   2775,
   2781,
   2787,
   2796,
   2805,
   2814,
   2823,
   2832,
   2839,
   2850,
   2857,
   2865,
   2878,
   2884,
   2892,
   2901,
   2910,
   2919,
   2928,
   2937,
   2946,
   2953,
   2961,
   2970,
   2977,
   2985,
   2994,
   3000,
   3007,
   3015,
   3021,
   3028,
   3036,
   3045,
   3051,
   3057,
   3066,
   3072,
   3078,
   3084,
   3090,
   3096,
   3102,
   3108,
   3114,
   3121,
   3129,
   3136,
   3162,
   3171,
   3178,
   3188,
   3195,
   3202,
   3212,
   3222,
   3232,
   3242,
   3252,
   3260,
   3272,
   3280,
   3289,
   3303,
   3310,
   3319,
   3329,
   3339,
   3349,
   3359,
   3369,
   3379,
   3387,
   3396,
   3406,
   3414,
   3423,
   3433,
   3440,
   3448,
   3457,
   3464,
   3472,
   3481,
   3491,
   3498,
   3505,
   3515,
   3522,
   3529,
   3536,
   3543,
   3550,
   3557,
   3564,
   3571,
   3579,
   3588,
   3596,
   3623,
   3633,
   3641,
   3650,
   3661,
   3669,
   3678,
   3687,
   3697,
   3705,
   3716,
   3727,
   3738,
   3749,
   3760,
   3769,
   3782,
   3791,
   3801,
   3816,
   3824,
   3834,
   3845,
   3856,
   3867,
   3878,
   3889,
   3900,
   3909,
   3919,
   3930,
   3939,
   3949,
   3960,
   3968,
   3977,
   3987,
   3995,
   4004,
   4014,
   4025,
   4033,
   4041,
   4052,
   4060,
   4068,
   4076,
   4084,
   4092,
   4100,
   4108,
   4115,
   4123,
   4132,
   4140,
   4167,
   4177,
   4185,
   4196,
   4204,
   4212,
   4223,
   4234,
   4245,
   4256,
   4267,
   4276,
   4289,
   4298,
   4308,
   4323,
   4331,
   4341,
   4352,
   4363,
   4374,
   4385,
   4396,
   4407,
   4416,
   4426,
   4437,
   4446,
   4456,
   4467,
   4475,
   4484,
   4494,
   4502,
   4511,
   4521,
   4532,
   4540,
   4548,
   4559,
   4567,
   4575,
   4583,
   4591,
   4599,
   4607,
   4615,
   4623,
   4632,
   4642,
   4651,
   4679,
   4690,
   4699,
   4709,
   4721,
   4730,
   4740,
   4750,
   4761,
   4770,
   4782,
   4794,
   4806,
   4818,
   4830,
   4840,
   4854,
   4864,
   4875,
   4891,
   4900,
   4911,
   4923,
   4935,
   4947,
   4959,
   4971,
   4983,
   4993,
   5004,
   5016,
   5026,
   5037,
   5049,
   5058,
   5068,
   5079,
   5088,
   5098,
   5109,
   5121,
   5130,
   5139,
   5151,
   5160,
   5169,
   5178,
   5187,
   5196,
   5205,
   5214,
   5220,
   5227,
   5235,
   5242,
   5268,
   5277,
   5284,
   5294,
   5301,
   5308,
   5318,
   5328,
   5338,
   5348,
   5358,
   5366,
   5378,
   5386,
   5395,
   5409,
   5416,
   5425,
   5435,
   5445,
   5455,
   5465,
   5475,
   5485,
   5493,
   5502,
   5512,
   5520,
   5529,
   5539,
   5546,
   5554,
   5563,
   5570,
   5578,
   5587,
   5597,
   5604,
   5611,
   5621,
   5628,
   5635,
   5642,
   5649,
   5656,
   5663,
   5670,
   5677,
   5685,
   5694,
   5702,
   5729,
   5739,
   5747,
   5758,
   5766,
   5774,
   5785,
   5796,
   5807,
   5818,
   5829,
   5838,
   5851,
   5860,
   5870,
   5885,
   5893,
   5903,
   5914,
   5925,
   5936,
   5947,
   5958,
   5969,
   5978,
   5988,
   5999,
   6008,
   6018,
   6029,
   6037,
   6046,
   6056,
   6064,
   6073,
   6083,
   6094,
   6102,
   6110,
   6121,
   6129,
   6137,
   6145,
   6153,
   6161,
   6169,
   6177,
   6184,
   6192,
   6201,
   6209,
   6236,
   6246,
   6254,
   6265,
   6273,
   6281,
   6292,
   6303,
   6314,
   6325,
   6336,
   6345,
   6358,
   6367,
   6377,
   6392,
   6400,
   6410,
   6421,
   6432,
   6443,
   6454,
   6465,
   6476,
   6485,
   6495,
   6506,
   6515,
   6525,
   6536,
   6544,
   6553,
   6563,
   6571,
   6580,
   6590,
   6601,
   6609,
   6617,
   6628,
   6636,
   6644,
   6652,
   6660,
   6668,
   6676,
   6684,
   6692,
   6701,
   6711,
   6720,
   6748,
   6759,
   6768,
   6780,
   6789,
   6798,
   6810,
   6822,
   6834,
   6846,
   6858,
   6868,
   6882,
   6892,
   6903,
   6919,
   6928,
   6939,
   6951,
   6963,
   6975,
   6987,
   6999,
   7011,
   7021,
   7032,
   7044,
   7054,
   7065,
   7077,
   7086,
   7096,
   7107,
   7116,
   7126,
   7137,
   7149,
   7158,
   7167,
   7179,
   7188,
   7197,
   7206,
   7215,
   7224,
   7233,
   7242,
   7247,
   7253,
   7260,
   7266,
   7291,
   7299,
   7305,
   7314,
   7320,
   7326,
   7335,
   7344,
   7353,
   7362,
   7371,
   7378,
   7389,
   7396,
   7404,
   7417,
   7423,
   7431,
   7440,
   7449,
   7458,
   7467,
   7476,
   7485,
   7492,
   7500,
   7509,
   7516,
   7524,
   7533,
   7539,
   7546,
   7554,
   7560,
   7567,
   7575,
   7584,
   7590,
   7596,
   7605,
   7611,
   7617,
   7623,
   7629,
   7635,
   7641,
   7647,
   7652,
   7658,
   7665,
   7671,
   7696,
   7704,
   7710,
   7719,
   7725,
   7731,
   7740,
   7749,
   7758,
   7767,
   7776,
   7783,
   7794,
   7801,
   7809,
   7822,
   7828,
   7836,
   7845,
   7854,
   7863,
   7872,
   7881,
   7890,
   7897,
   7905,
   7914,
   7921,
   7929,
   7938,
   7944,
   7951,
   7959,
   7965,
   7972,
   7980,
   7989,
   7995,
   8001,
   8010,
   8016,
   8022,
   8028,
   8034,
   8040,
   8046,
   8052,
   8058,
   8065,
   8073,
   8080,
   8106,
   8115,
   8122,
   8132,
   8139,
   8146,
   8156,
   8166,
   8176,
   8186,
   8196,
   8204,
   8216,
   8224,
   8233,
   8247,
   8254,
   8263,
   8273,
   8283,
   8293,
   8303,
   8313,
   8323,
   8331,
   8340,
   8350,
   8358,
   8367,
   8377,
   8384,
   8392,
   8401,
   8408,
   8416,
   8425,
   8435,
   8442,
   8449,
   8459,
   8466,
   8473,
   8480,
   8487,
   8494,
   8501,
   8508,
   8513,
   8519,
   8526,
   8532,
   8557,
   8565,
   8571,
   8580,
   8586,
   8592,
   8601,
   8610,
   8619,
   8628,
   8637,
   8644,
   8655,
   8662,
   8670,
   8683,
   8689,
   8697,
   8706,
   8715,
   8724,
   8733,
   8742,
   8751,
   8758,
   8766,
   8775,
   8782,
   8790,
   8799,
   8805,
   8812,
   8820,
   8826,
   8833,
   8841,
   8850,
   8856,
   8862,
   8871,
   8877,
   8883,
   8889,
   8895,
   8901,
   8907,
   8913,
   8918,
   8924,
   8931,
   8937,
   8962,
   8970,
   8976,
   8985,
   8991,
   8997,
   9006,
   9015,
   9024,
   9033,
   9042,
   9049,
   9060,
   9067,
   9075,
   9088,
   9094,
   9102,
   9111,
   9120,
   9129,
   9138,
   9147,
   9156,
   9163,
   9171,
   9180,
   9187,
   9195,
   9204,
   9210,
   9217,
   9225,
   9231,
   9238,
   9246,
   9255,
   9261,
   9267,
   9276,
   9282,
   9288,
   9294,
   9300,
   9306,
   9312,
   9318,
   9324,
   9331,
   9339,
   9346,
   9372,
   9381,
   9388,
   9398,
   9405,
   9412,
   9422,
   9432,
   9442,
   9452,
   9462,
   9470,
   9482,
   9490,
   9499,
   9513,
   9520,
   9529,
   9539,
   9549,
   9559,
   9569,
   9579,
   9589,
   9597,
   9606,
   9616,
   9624,
   9633,
   9643,
   9650,
   9658,
   9667,
   9674,
   9682,
   9691,
   9701,
   9708,
   9715,
   9725,
   9732,
   9739,
   9746,
   9753,
   9760,
   9767,
   9774,
   9780,
   9787,
   9793,
   0,
   9798,
   9804,
   9806,
   9810,
   9817,
   9821,
   9828,
   9833,
   9840,
   9849,
   9854,
   9859,
   9866,
   9875,
   9880,
   9883,
   9887,
   9890,
   9893,
   9897,
   9901,
   9904,
   9907,
   9916,
   9929,
   9943,
   9952,
   9962,
   9971,
   9983,
   9993,
   10003,
   10012,
   10021,
   10024,
   10028,
   10033,
   10038,
   10042,
   10046,
   10050,
   10055,
   10060,
   10064,
   10068,
   10072,
   10076,
   10080,
   10083,
   10086,
   10089,
   10092,
   10095,
   10100,
   10106,
   10117,
   10129,
   10135,
   10147,
   10160,
   10164,
   10169,
   10180,
   10192,
   10197,
   10209,
   10222,
   10224,
   10229,
   10234,
   10238,
   10242,
   10245,
   10255,
   10261,
   10266,
   10272,
   10277,
   10282,
   10287,
   10292,
   10297,
   10301,
   10305,
   10309,
   10312,
   10315,
   10318,
   10320,
   10323,
   10327,
   10331,
   10334,
   10337,
   10342,
   10347,
   10353,
   10358,
   10362,
   10367,
   10373,
   10379,
   10384,
   10389,
   10392,
   10394,
   10398,
   10402,
   10406,
   10409,
   10413,
   10418,
   10423,
   10427,
   10431,
   10435,
   10438,
   10442,
   10447,
   10452,
   10456,
   10460,
   10462,
   10464,
   10467,
   10470,
   10473,
   10477,
   10480,
   10484,
   10489,
   10494,
   10498,
   10502,
   10507,
   10510,
   10514,
   10518,
   10521,
   10524,
   10530,
   10541,
   10553,
   10559,
   10571,
   10584,
   10588,
   10593,
   10603,
   10614,
   10619,
   10630,
   10642,
   10644,
   10648,
   10652,
   10656,
   10660,
   10665,
   10673,
   10678,
   10683,
   10687,
   10691,
   10701,
   10706,
   10712,
   10717,
   10722,
   10727,
   10732,
   10737,
   10741,
   10745,
   10749,
   10752,
   10756,
   10760,
   10763,
   10766,
   10769,
   10775,
   10781,
   10786,
   10791,
   10796,
   10802,
   10805,
   10808,
   10811,
   10816,
   10822,
   10827,
   10832,
   10837,
   10842,
   10847,
   10852,
   10857,
   10861,
   10865,
   10869,
   10873,
   10877,
   0,
   10882,
   10888,
   10890,
   10895,
   10906,
   10915,
   10920,
   10930,
   10937,
   10942,
   10949,
   0,
   10956,
   10962,
   10967,
   10977,
   10992,
   11007,
   11022,
   11027,
   11032,
   11037,
   11045,
   11060,
   0,
   11068,
   11071,
   11074,
   11077,
   11080,
   11083,
   11086,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   11088,
   11091,
   11094,
   11096,
   11099,
   11102,
   11105,
   11108,
   11110,
   11112,
   11117,
   11119,
   11121,
   11126,
   11133,
   11141,
   11146,
   11152,
   11157,
   11163,
   11168,
   11172,
   11174,
   11176,
   11178,
   0,
   11180,
   0,
   0,
   11182,
   0,
   0,
   11185,
   11193,
   11201,
   11210,
   11218,
   11227,
   11234,
   11241,
   11249,
   11259,
   11266,
   11273,
   11275,
   11277,
   11281,
   11283,
   11285,
   0,
   0,
   11287,
   0,
   0,
   11290,
   0,
   0,
   11293,
   11296,
   11299,
   0,
   11301,
   11303,
   11305,
   11310,
   11316,
   11323,
   11330,
   11338,
   11345,
   11354,
   11361,
   11366,
   11374,
   11382,
   11395,
   11408,
   11415,
   11425,
   11433,
   11441,
   11450,
   11464,
   11478,
   11490,
   11505,
   11518,
   11531,
   11544,
   11560,
   11575,
   11589,
   11603,
   11620,
   11636,
   11651,
   11664,
   11680,
   11694,
   11708,
   11716,
   11727,
   11736,
   11745,
   11754,
   11761,
   11771,
   11779,
   11787,
   11794,
   11801,
   11811,
   11821,
   11829,
   11840,
   11851,
   11861,
   11872,
   11884,
   11895,
   11907,
   11920,
   11934,
   11947,
   11954,
   11962,
   11970,
   11978,
   11981,
   11985,
   11989,
   11993,
   11996,
   11999,
   12004,
   12008,
   12012,
   12017,
   12021,
   12025,
   12029,
   12033,
   12038,
   12043,
   12048,
   12053,
   12057,
   12061,
   12067,
   12073,
   12079,
   12085,
   12091,
   12098,
   12104,
   12110,
   12117,
   12123,
   12129,
   12135,
   12141,
   12147,
   12154,
   12160,
   12166,
   12172,
   12179,
   12186,
   12193,
   12200,
   12206,
   12212,
   12218,
   12225,
   12232,
   12239,
   12245,
   12251,
   12257,
   12264,
   12271,
   12278,
   12291,
   12297,
   12303,
   12309,
   12316,
   12322,
   12328,
   12334,
   12340,
   12346,
   12352,
   12358,
   12364,
   12371,
   12377,
   12383,
   12389,
   12396,
   12403,
   12410,
   12417,
   12424,
   12430,
   12436,
   12442,
   12449,
   12456,
   12463,
   12469,
   12475,
   12481,
   12488,
   12495,
   12502,
   12515,
   12518,
   12521,
   12526,
   12530,
   12534,
   12539,
   12543,
   12547,
   12551,
   12555,
   12560,
   12565,
   12570,
   12575,
   12579,
   12583,
   12589,
   12595,
   12598,
   12601,
   12604,
   12606,
   12610,
   12614,
   12618,
   12622,
   12626,
   12629,
   12634,
   12638,
   12642,
   12645,
   12649,
   12652,
   12655,
   12658,
   12661,
   12664,
   12667,
   12670,
   12674,
   12677,
   12681,
   12684,
   12688,
   12692,
   12696,
   12699,
   0,
   12701,
   12703,
   0,
   12708,
   12711,
   12714,
   12717,
   12723,
   12728,
   12738,
   12742,
   12746,
   12749,
   12752,
   12755,
   12757,
   12760,
   12763,
   12766,
   12768,
   12771,
   12775,
   12779,
   12782,
   12786,
   12789,
   12792,
   12795,
   12798,
   12801,
   12804,
   12807,
   12810,
   12813,
   12817,
   12821,
   12825,
   12829,
   12831,
   12833,
   12837,
   12841,
   12843,
   12847,
   12849,
   12851,
   12855,
   12859,
   12861,
   12868,
   12876,
   12881,
   12886,
   12891,
   12898,
   12906,
   12911,
   12916,
   12921,
   12923,
   12925,
   12927,
   12929,
   12931,
   12933,
   12935,
   12937,
   12939,
   0,
   12941,
   12943,
   12945,
   12947,
   12949,
   12951,
   12953,
   12955,
   12957,
   12959,
   12961,
   12966,
   12972,
   12978,
   12986,
   12995,
   13000,
   13007,
   13015,
   13021,
   13028,
   13035,
   13044,
   13054,
   13060,
   13068,
   13077,
   13086,
   13097,
   13109,
   13115,
   13123,
   13132,
   13135,
   13142,
   13156,
   13163,
   13170,
   13184,
   13198,
   13204,
   13216,
   13228,
   13234,
   13240,
   13247,
   13254,
   13260,
   13272,
   13284,
   13287,
   13294,
   13308,
   13315,
   13322,
   13336,
   13350,
   13362,
   13374,
   13381,
   13388,
   13394,
   13406,
   13418,
   13421,
   13424,
   13427,
   13432,
   13437,
   13442,
   13448,
   13453,
   13459,
   13469,
   13472,
   13476,
   13480,
   13484,
   13488,
   13492,
   13495,
   13499,
   13503,
   13507,
   13511,
   13516,
   13519,
   13523,
   13527,
   13531,
   13535,
   13540,
   13549,
   13571,
   13574,
   13577,
   13580,
   13583,
   13586,
   13589,
   13593,
   13596,
   13600,
   13608,
   13611,
   13615,
   13619,
   13623,
   13626,
   13629,
   13633,
   13637,
   13641,
   13645,
   13648,
   13652,
   13656,
   13660,
   13664,
   13672,
   13694,
   13696,
   13699,
   13711,
   13721,
   13728,
   13735,
   13742,
   13781,
   13800,
   13807,
   13814,
   13826,
   13832,
   13842,
   13849,
   13856,
   13863,
   13870,
   13876,
   13892,
   13934,
   13941,
   13948,
   13955,
   13967,
   13970,
   13973,
   13976,
   13980,
   13985,
   13989,
   13991,
   13993,
   13996,
   13999,
   14001,
   14003,
   14006,
   14009,
   14012,
   14015,
   14019,
   14022,
   14024,
   14026,
   14029,
   14035,
   14041,
   14047,
   14053,
   14060,
   14066,
   14073,
   14084,
   14091,
   14097,
   14105,
   14112,
   14119,
   14128,
   14134,
   14142,
   14149,
   14156,
   14165,
   14177,
   14202,
   14208,
   14214,
   14220,
   14226,
   14233,
   14239,
   14246,
   14257,
   14264,
   14270,
   14278,
   14285,
   14292,
   14301,
   14307,
   14315,
   14322,
   14329,
   14338,
   14350,
   14375,
   14381,
   14387,
   14393,
   14396,
   14399,
   14402,
   14405,
   14408,
   14411,
   14414,
   14416,
   14418,
   14420,
   14422,
   14424,
   14426,
   14428,
   14430,
   14432,
   14436,
   14438,
   14440,
   14444,
   14448,
   14450,
   14453,
   14455,
   14457,
   14459,
   14461,
   14463,
   14465,
   14467,
   14469,
   14471,
   14473,
   14475,
   14479,
   14484,
   14490,
   14493,
   14497,
   14500,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   14504,
   14506,
   14509,
   14512,
   14515,
   14518,
   14521,
   14524,
   14526,
   14528,
   14530,
   14532,
   14535,
   14537,
   14542,
   14546,
   14549,
   14552,
   14557,
   14561,
   14565,
   14569,
   14573,
   14577,
   14581,
   14585,
   14589,
   14593,
   14597,
   14601,
   14604,
   14607,
   14611,
   14614,
   14617,
   14621,
   14623,
   14625,
   14627,
   14629,
   14631,
   14634,
   14636,
   14638,
   14640,
   14642,
   14644,
   14646,
   14652,
   14654,
   14656,
   14658,
   14660,
   14662,
   14664,
   14666,
   14670,
   14674,
   14678,
   14682,
   14686,
   14691,
   14696,
   14701,
   14706,
   14711,
   14715,
   14719,
   14724,
   14729,
   14734,
   14739,
   14743,
   14747,
   14751,
   14755,
   14758,
   14761,
   14766,
   14771,
   14776,
   14781,
   14784,
   14787,
   14792,
   14797,
   14802,
   14807,
   14810,
   14813,
   14818,
   14823,
   14828,
   14833,
   14836,
   14839,
   14844,
   14849,
   14854,
   14859,
   14863,
   14867,
   14873,
   14879,
   14885,
   14891,
   14895,
   14899,
   14905,
   14911,
   14917,
   14923,
   14927,
   14931,
   14937,
   14943,
   14949,
   14955,
   14958,
   14961,
   14966,
   14971,
   14976,
   14981,
   14984,
   14987,
   14992,
   14997,
   15002,
   15007,
   15012,
   15017,
   15022,
   15027,
   15032,
   15037,
   15042,
   15047,
   15054,
   15061,
   15068,
   15075,
   15082,
   15089,
   15091,
   15095,
   15099,
   15103,
   15107,
   15119,
   15132,
   15146,
   15160,
   15174,
   15188,
   15193,
   15198,
   15203,
   15208,
   15214,
   15220,
   15226,
   15232,
   15238,
   15244,
   15250,
   15256,
   15261,
   15266,
   15271,
   15276,
   15281,
   15286,
   15291,
   15296,
   15303,
   15310,
   15317,
   15324,
   15331,
   0,
   15338,
   15340,
   15345,
   15350,
   15354,
   15359,
   15364,
   15367,
   0,
   15371,
   15376,
   15378,
   15384,
   15389,
   15398,
   15404,
   15411,
   15415,
   15419,
   15425,
   15430,
   15434,
   15438,
   15443,
   15447,
   15451,
   15457,
   15462,
   15466,
   15470,
   15475,
   0,
   0,
   15479,
   15481,
   15492,
   15499,
   15506,
   15515,
   15525,
   15530,
   15532,
   15534,
   15537,
   15541,
   0,
   0,
   15543,
   15545,
   0,
   0,
   15549,
   15552,
   15555,
   15558,
   15562,
   0,
   0,
   15566,
   15568,
   0,
   0,
   15572,
   15574,
   15576,
   15581,
   15585,
   15589,
   15594,
   0,
   15596,
   15598,
   15609,
   15617,
   15628,
   15637,
   15646,
   15662,
   15675,
   15691,
   15705,
   15719,
   15730,
   15738,
   15749,
   15758,
   15767,
   15777,
   15789,
   15799,
   15812,
   15823,
   15834,
   15848,
   15859,
   15867,
   15878,
   15887,
   15896,
   15905,
   15917,
   15927,
   15937,
   15942,
   15947,
   15952,
   15957,
   15964,
   15971,
   15978,
   15985,
   15989,
   15995,
   16001,
   16007,
   16013,
   16019,
   16025,
   16031,
   16037,
   16041,
   16047,
   16053,
   16059,
   16065,
   16072,
   16077,
   16082,
   16087,
   16092,
   16098,
   16104,
   16110,
   16116,
   16123,
   16130,
   16137,
   16144,
   16146,
   16153,
   16160,
   16167,
   16174,
   16180,
   16186,
   16192,
   16198,
   16204,
   16210,
   16216,
   16222,
   16228,
   16233,
   16238,
   16241,
   16249,
   16257,
   16265,
   16273,
   16281,
   16289,
   16294,
   16297,
   16299,
   16301,
   16303,
   16305,
   16307,
   16309,
   16312,
   16318,
   16320,
   16325,
   16330,
   16335,
   16340,
   16349,
   16358,
   16368,
   16380,
   16388,
   16399,
   16410,
   16421,
   16433,
   16435,
   16439,
   16446,
   16449,
   16455,
   16461,
   16467,
   16473,
   16477,
   16484,
   16487,
   16493,
   16499,
   16505,
   16511,
   16518,
   16528,
   16531,
   16536,
   16544,
   16548,
   16555,
   16562,
   16569,
   16576,
   16581,
   16589,
   16593,
   16600,
   16607,
   16614,
   16621,
   16625,
   16632,
   16635,
   16641,
   16647,
   16653,
   16659,
   16663,
   16670,
   16673,
   16679,
   16685,
   16691,
   16697,
   16704,
   16707,
   16713,
   16719,
   16725,
   16731,
   16734,
   16740,
   16749,
   16758,
   16767,
   16776,
   16781,
   16786,
   16791,
   16796,
   16807,
   16815,
   16823,
   16831,
   16839,
   16847,
   16855,
   16860,
   16863,
   16865,
   16871,
   16878,
   16885,
   16892,
   16899,
   16907,
   16913,
   16921,
   16928,
   16935,
   16939,
   16943,
   16947,
   16952,
   16957,
   16962,
   16967,
   16972,
   16977,
   16982,
   16987,
   16992,
   16996,
   17000,
   17004,
   17009,
   17015,
   17019,
   17025,
   17029,
   17033,
   17035,
   17039,
   17043,
   17047,
   17050,
   17052,
   17056,
   17060,
   17063,
   17068,
   17073,
   17078,
   17081,
   17086,
   17091,
   17096,
   17099,
   17104,
   17108,
   17114,
   17118,
   17122,
   17128,
   17132,
   17136,
   17140,
   17144,
   17148,
   17151,
   17155,
   17160,
   17163,
   17169,
   17175,
   17181,
   17187,
   17192,
   17200,
   17208,
   17217,
   17222,
   17229,
   17231,
   17233,
   17241,
   17243,
   0,
   0,
   0,
   17249,
   17251,
   17256,
   17260,
   17264,
   17268,
   17272,
   17277,
   17282,
   17287,
   17292,
   17296,
   17300,
   17306,
   17310,
   17314,
   17316,
   17320,
   17324,
   17326,
   17329,
   17333,
   17336,
   17341,
   17346,
   17349,
   17354,
   17359,
   17362,
   17366,
   17372,
   17376,
   17380,
   17387,
   17392,
   17399,
   17407,
   17415,
   17418,
   17422,
   17428,
   17432,
   17435,
   17437,
   17441,
   17445,
   17451,
   17454,
   17457,
   17461,
   17468,
   17476,
   17485,
   17492,
   17500,
   17509,
   17516,
   17524,
   17533,
   17540,
   17548,
   17557,
   17566,
   17576,
   17587,
   17596,
   17606,
   17617,
   17626,
   17636,
   17647,
   17656,
   17666,
   17677,
   17685,
   17694,
   17704,
   17712,
   17721,
   17731,
   17739,
   17748,
   17758,
   17766,
   17775,
   17785,
   17793,
   17802,
   17812,
   17820,
   17829,
   17839,
   17847,
   17856,
   17866,
   17874,
   17883,
   17893,
   17902,
   17912,
   17923,
   17932,
   17942,
   17953,
   17962,
   17972,
   17983,
   17992,
   18002,
   18013,
   18023,
   18034,
   18046,
   18056,
   18067,
   18079,
   18089,
   18100,
   18112,
   18122,
   18133,
   18145,
   18155,
   18166,
   18178,
   18188,
   18199,
   18211,
   18221,
   18232,
   18244,
   18254,
   18265,
   18277,
   18288,
   18300,
   18313,
   18324,
   18336,
   18349,
   18360,
   18372,
   18385,
   18396,
   18408,
   18421,
   18430,
   18440,
   18451,
   18460,
   18470,
   18481,
   18490,
   18500,
   18511,
   18520,
   18530,
   18541,
   18551,
   18562,
   18574,
   18584,
   18595,
   18607,
   18617,
   18628,
   18640,
   18650,
   18661,
   18673,
   18683,
   18694,
   18706,
   18716,
   18727,
   18739,
   18749,
   18760,
   18772,
   18782,
   18793,
   18805,
   18816,
   18828,
   18841,
   18852,
   18864,
   18877,
   18888,
   18900,
   18913,
   18924,
   18936,
   18949,
   18955,
   18963,
   18972,
   18982,
   18990,
   18999,
   19009,
   19017,
   19026,
   19036,
   19044,
   19053,
   19063,
   19071,
   19080,
   19090,
   19098,
   19107,
   19117,
   19125,
   19134,
   19144,
   19152,
   19161,
   19171,
   19178,
   19187,
   19197,
   19208,
   19217,
   19227,
   19238,
   19247,
   19257,
   19268,
   19277,
   19287,
   19298,
   19306,
   19315,
   19325,
   19333,
   19342,
   19352,
   19360,
   19369,
   19379,
   19387,
   19396,
   19406,
   19414,
   19423,
   19433,
   19441,
   19450,
   19460,
   19468,
   19477,
   19487,
   19495,
   19504,
   19514,
   19523,
   19533,
   19544,
   19553,
   19563,
   19574,
   19583,
   19593,
   19604,
   19613,
   19623,
   19634,
   19641,
   19648,
   19655,
   19657,
   19659,
   19664,
   19669,
   19674,
   19679,
   19684,
   19691,
   19698,
   19705,
   19712,
   19719,
   19726,
   19733,
   19740,
   19747,
   19754,
   19761,
   19768,
   19774,
   19780,
   19786,
   19792,
   19798,
   19804,
   19810,
   19816,
   19822,
   19828,
   19834,
   19840,
   19848,
   19856,
   19864,
   19872,
   19880,
   19888,
   19893,
   19936,
   19979,
   19999,
   20004,
   20008,
   20012,
   20018,
   20021,
   20027,
   20033,
   20046,
   20058,
   20062,
   20069,
   20076,
   20080,
   20090,
   20094,
   20099,
   20104,
   20112,
   20117,
   20125,
   20130,
   20138,
   20146,
   20157,
   20162,
   20170,
   20178,
   20189,
   20195,
   20204,
   20209,
   20217,
   0,
   20222,
   20230,
   20238,
   0,
   0,
   0,
   20246,
   20253,
   20259,
   20262,
   20265,
   20269,
   20275,
   20279,
   20283,
   20287,
   20291,
   20295,
   20299,
   20302,
   20307,
   20311,
   20316,
   20320,
   20325,
   20330,
   0,
   0,
   20334,
   20341,
   20348,
   20350,
   0,
   0,
   0,
   0,
   20352,
   20355,
   0,
   0,
   0,
   0,
   20358,
   20362,
   20366,
   0,
   0,
   0,
   20368,
   20370,
   0,
   20372,
   20383,
   20399,
   20410,
   20422,
   20436,
   20448,
   20453,
   20458,
   20460,
   20467,
   20474,
   20477,
   20480,
   20484,
   20488,
   20492,
   20496,
   20501,
   20506,
   20509,
   20514,
   20518,
   20523,
   20528,
   20532,
   20536,
   20538,
   20546,
   20555,
   20562,
   20565,
   20569,
   20577,
   20585,
   20590,
   20598,
   20609,
   20620,
   20634,
   20648,
   20653,
   20664,
   20667,
   20670,
   0,
   20674,
   20677,
   20683,
   20692,
   20700,
   20713,
   20725,
   20732,
   20738,
   20742,
   20744,
   20747,
   20753,
   20762,
   20775,
   20787,
   20794,
   20800,
   20804,
   20806,
   20809,
   20815,
   20824,
   20837,
   20849,
   20856,
   20862,
   20867,
};

static const nir_algebraic_table nir_opt_algebraic_table = {
   .transforms = nir_opt_algebraic_transforms,
   .transform_offsets = nir_opt_algebraic_transform_offsets,
   .pass_op_table = nir_opt_algebraic_pass_op_table,
   .values = nir_opt_algebraic_values,
   .expression_cond = nir_opt_algebraic_expression_cond,
   .variable_cond = nir_opt_algebraic_variable_cond,
};

bool
nir_opt_algebraic(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[113];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   STATIC_ASSERT(5555 == ARRAY_SIZE(nir_opt_algebraic_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ffloor;
   condition_flags[22] = options->lower_ffloor || ((options->lower_doubles_options & nir_lower_dfloor) && !(options->lower_doubles_options & nir_lower_dfloor));
   condition_flags[23] = !options->lower_ffloor;
   condition_flags[24] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[25] = options->lower_ffract;
   condition_flags[26] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[27] = options->lower_fceil;
   condition_flags[28] = options->lower_ffma16;
   condition_flags[29] = options->lower_ffma32;
   condition_flags[30] = options->lower_ffma64;
   condition_flags[31] = options->fuse_ffma16;
   condition_flags[32] = options->fuse_ffma32;
   condition_flags[33] = options->fuse_ffma64;
   condition_flags[34] = options->lower_fdph;
   condition_flags[35] = !options->lower_fdph;
   condition_flags[36] = options->lower_fdot;
   condition_flags[37] = !options->lower_fsat;
   condition_flags[38] = !options->lower_iabs;
   condition_flags[39] = options->lower_fsat;
   condition_flags[40] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[41] = !options->lower_fsign;
   condition_flags[42] = options->lower_umax;
   condition_flags[43] = options->lower_umin;
   condition_flags[44] = !options->lower_umax;
   condition_flags[45] = !options->lower_umin;
   condition_flags[46] = options->lower_scmp;
   condition_flags[47] = options->lower_vector_cmp;
   condition_flags[48] = !options->lower_rotate;
   condition_flags[49] = options->lower_rotate;
   condition_flags[50] = options->lower_fpow;
   condition_flags[51] = !options->lower_fpow;
   condition_flags[52] = options->lower_fdiv;
   condition_flags[53] = options->lower_fsqrt;
   condition_flags[54] = !options->lower_fsqrt;
   condition_flags[55] = options->lower_sincos;
   condition_flags[56] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[57] = !options->lower_extract_byte;
   condition_flags[58] = !options->lower_extract_word;
   condition_flags[59] = options->has_pack_32_4x8;
   condition_flags[60] = options->lower_pack_64_2x32_split;
   condition_flags[61] = options->lower_pack_32_2x16_split;
   condition_flags[62] = options->lower_unpack_64_2x32_split;
   condition_flags[63] = options->lower_unpack_32_2x16_split;
   condition_flags[64] = options->lower_fmod;
   condition_flags[65] = options->lower_uadd_carry;
   condition_flags[66] = options->lower_usub_borrow;
   condition_flags[67] = options->lower_bitfield_insert;
   condition_flags[68] = options->lower_hadd;
   condition_flags[69] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[70] = options->lower_mul_32x16;
   condition_flags[71] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[72] = options->lower_uadd_sat;
   condition_flags[73] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[74] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[75] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[76] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[77] = options->lower_bitfield_insert_to_shifts;
   condition_flags[78] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[79] = options->lower_bitfield_extract;
   condition_flags[80] = options->lower_bitfield_extract_to_shifts;
   condition_flags[81] = options->lower_ifind_msb;
   condition_flags[82] = options->lower_find_msb_to_reverse;
   condition_flags[83] = options->lower_find_lsb;
   condition_flags[84] = options->lower_extract_byte;
   condition_flags[85] = options->lower_extract_word;
   condition_flags[86] = options->lower_pack_unorm_2x16;
   condition_flags[87] = options->lower_pack_unorm_4x8;
   condition_flags[88] = options->lower_pack_snorm_2x16;
   condition_flags[89] = options->lower_pack_snorm_4x8;
   condition_flags[90] = options->lower_unpack_unorm_2x16;
   condition_flags[91] = options->lower_unpack_unorm_4x8;
   condition_flags[92] = options->lower_unpack_snorm_2x16;
   condition_flags[93] = options->lower_unpack_snorm_4x8;
   condition_flags[94] = options->lower_pack_split;
   condition_flags[95] = options->lower_isign;
   condition_flags[96] = !options->lower_isign;
   condition_flags[97] = options->lower_fsign;
   condition_flags[98] = !options->has_imul24;
   condition_flags[99] = !options->has_umul24;
   condition_flags[100] = !options->has_umad24;
   condition_flags[101] = options->has_imul24;
   condition_flags[102] = options->has_umad24;
   condition_flags[103] = options->has_umul24;
   condition_flags[104] = options->has_fused_comp_and_csel;
   condition_flags[105] = options->lower_iadd_sat;
   condition_flags[106] = options->support_8bit_alu;
   condition_flags[107] = options->support_16bit_alu;
   condition_flags[108] = options->lower_ldexp;
   condition_flags[109] = !options->lower_bitfield_reverse;
   condition_flags[110] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[111] = options->lower_fisnormal;
   condition_flags[112] = !options->vectorize_vec2_16bit;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_table);
      }
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 15
 * transforms:
 *    ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fmul', ('fmul', 'a', 'c'), 'b')
 *    ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('imul', ('imul', 'a', 'c'), 'b')
 *    ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fadd', ('fadd', 'a', 'c'), 'b')
 *    ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('iadd', ('iadd', 'a', 'c'), 'b')
 *    ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c'))
 *    ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c'))
 *    ('~fadd', ('fneg', 'a'), 'a') => 0.0
 *    ('iadd', ('ineg', 'a'), 'a') => 0
 *    ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b
 *    ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b
 *    ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b
 *    ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b
 *    ('~flrp', ('fadd(is_used_once)', 'a', -1.0), ('fadd(is_used_once)', 'a', 1.0), 'd') => ('fadd', ('flrp', -1.0, 1.0, 'd'), 'a')
 *    ('~flrp', ('fadd(is_used_once)', 'a', 1.0), ('fadd(is_used_once)', 'a', -1.0), 'd') => ('fadd', ('flrp', 1.0, -1.0, 'd'), 'a')
 *    ('~flrp', ('fadd(is_used_once)', 'a', '#b'), ('fadd(is_used_once)', 'a', '#c'), 'd') => ('fadd', ('fmul', 'd', ('fadd', 'c', ('fneg', 'b'))), ('fadd', 'a', 'b'))
 */


static const nir_search_value_union nir_opt_algebraic_before_ffma_values[] = {
   /* ('~fmul(is_used_once)', ('fmul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fmul', ('fmul', 'a', 'c'), 'b') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 0, 1 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 2, 3 },
      0,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 5, 6 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 7, 8 },
      -1,
   } },

   /* ('imul(is_used_once)', ('imul(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('imul', ('imul', 'a', 'c'), 'b') */
   /* search1954_0_0 -> 0 in the cache */
   /* search1954_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search1954_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 10, 3 },
      0,
   } },

   /* replace1954_0_0 -> 5 in the cache */
   /* replace1954_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace1954_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 12, 8 },
      -1,
   } },

   /* ('~fadd(is_used_once)', ('fadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('fadd', ('fadd', 'a', 'c'), 'b') */
   /* search1955_0_0 -> 0 in the cache */
   /* search1955_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search1955_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 14, 3 },
      0,
   } },

   /* replace1955_0_0 -> 5 in the cache */
   /* replace1955_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace1955_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 16, 8 },
      -1,
   } },

   /* ('iadd(is_used_once)', ('iadd(is_used_once)', 'a(is_not_const)', '#b'), 'c(is_not_const)') => ('iadd', ('iadd', 'a', 'c'), 'b') */
   /* search1956_0_0 -> 0 in the cache */
   /* search1956_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search1956_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 18, 3 },
      0,
   } },

   /* replace1956_0_0 -> 5 in the cache */
   /* replace1956_0_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 5, 6 },
      -1,
   } },
   /* replace1956_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 20, 8 },
      -1,
   } },

   /* ('~fadd', ('fmul', 'a', 'b'), ('fmul', 'a', 'c')) => ('fmul', 'a', ('fadd', 'b', 'c')) */
   /* search1957_0_0 -> 5 in the cache */
   /* search1957_0_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 5, 8 },
      -1,
   } },
   /* search1957_1_0 -> 5 in the cache */
   /* search1957_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 5, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 22, 23 },
      -1,
   } },

   /* replace1957_0 -> 5 in the cache */
   /* replace1957_1_0 -> 8 in the cache */
   /* replace1957_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 8, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 5, 25 },
      -1,
   } },

   /* ('iadd', ('imul', 'a', 'b'), ('imul', 'a', 'c')) => ('imul', 'a', ('iadd', 'b', 'c')) */
   /* search1958_0_0 -> 5 in the cache */
   /* search1958_0_1 -> 8 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      1, 1,
      { 5, 8 },
      -1,
   } },
   /* search1958_1_0 -> 5 in the cache */
   /* search1958_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      2, 1,
      { 5, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 3,
      { 27, 28 },
      -1,
   } },

   /* replace1958_0 -> 5 in the cache */
   /* replace1958_1_0 -> 8 in the cache */
   /* replace1958_1_1 -> 6 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 8, 6 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_imul,
      0, 2,
      { 5, 30 },
      -1,
   } },

   /* ('~fadd', ('fneg', 'a'), 'a') => 0.0 */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1959_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 33, 32 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },

   /* ('iadd', ('ineg', 'a'), 'a') => 0 */
   /* search1960_0_0 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 32 },
      -1,
   } },
   /* search1960_1 -> 32 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 36, 32 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },

   /* ('iadd', ('ineg', 'a'), ('iadd', 'a', 'b')) => b */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 39 },
      -1,
   } },
   /* search1961_1_0 -> 39 in the cache */
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 39, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 40, 42 },
      -1,
   } },

   /* replace1961 -> 41 in the cache */

   /* ('iadd', 'a', ('iadd', ('ineg', 'a'), 'b')) => b */
   /* search1962_0 -> 39 in the cache */
   /* search1962_1_0_0 -> 39 in the cache */
   /* search1962_1_0 -> 40 in the cache */
   /* search1962_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 40, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 39, 44 },
      -1,
   } },

   /* replace1962 -> 41 in the cache */

   /* ('~fadd', ('fneg', 'a'), ('fadd', 'a', 'b')) => b */
   /* search1963_0_0 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 39 },
      -1,
   } },
   /* search1963_1_0 -> 39 in the cache */
   /* search1963_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 46, 47 },
      -1,
   } },

   /* replace1963 -> 41 in the cache */

   /* ('~fadd', 'a', ('fadd', ('fneg', 'a'), 'b')) => b */
   /* search1964_0 -> 39 in the cache */
   /* search1964_1_0_0 -> 39 in the cache */
   /* search1964_1_0 -> 46 in the cache */
   /* search1964_1_1 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 46, 41 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 39, 49 },
      -1,
   } },

   /* replace1964 -> 41 in the cache */

   /* ('~flrp', ('fadd(is_used_once)', 'a', -1.0), ('fadd(is_used_once)', 'a', 1.0), 'd') => ('fadd', ('flrp', -1.0, 1.0, 'd'), 'a') */
   /* search1965_0_0 -> 39 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 39, 51 },
      0,
   } },
   /* search1965_1_0 -> 39 in the cache */
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 53 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 52, 54, 55 },
      -1,
   } },

   /* replace1965_0_0 -> 51 in the cache */
   /* replace1965_0_1 -> 53 in the cache */
   /* replace1965_0_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 51, 53, 55 },
      -1,
   } },
   /* replace1965_1 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 57, 39 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', 1.0), ('fadd(is_used_once)', 'a', -1.0), 'd') => ('fadd', ('flrp', 1.0, -1.0, 'd'), 'a') */
   /* search1966_0_0 -> 39 in the cache */
   /* search1966_0_1 -> 53 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 39, 53 },
      0,
   } },
   /* search1966_1_0 -> 39 in the cache */
   /* search1966_1_1 -> 51 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 39, 51 },
      0,
   } },
   /* search1966_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 59, 60, 55 },
      -1,
   } },

   /* replace1966_0_0 -> 53 in the cache */
   /* replace1966_0_1 -> 51 in the cache */
   /* replace1966_0_2 -> 55 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 53, 51, 55 },
      -1,
   } },
   /* replace1966_1 -> 39 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 62, 39 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', '#b'), ('fadd(is_used_once)', 'a', '#c'), 'd') => ('fadd', ('fmul', 'd', ('fadd', 'c', ('fneg', 'b'))), ('fadd', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 64, 65 },
      0,
   } },
   /* search1967_1_0 -> 64 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 64, 67 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 66, 68, 69 },
      -1,
   } },

   /* replace1967_0_0 -> 69 in the cache */
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 72 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 71, 73 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 69, 74 },
      -1,
   } },
   /* replace1967_1_0 -> 64 in the cache */
   /* replace1967_1_1 -> 72 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 64, 72 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 75, 76 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_before_ffma_expression_cond[] = {
   (is_used_once),
};

static const nir_search_variable_cond nir_opt_algebraic_before_ffma_variable_cond[] = {
   (is_not_const),
};

static const struct transform nir_opt_algebraic_before_ffma_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 24, 26, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { 48, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 34, 35, 0 },
   { 48, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 9, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 9, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 29, 31, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { 43, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 37, 38, 0 },
   { 43, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 13, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 56, 58, 0 },
   { 61, 63, 0 },
   { 70, 77, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { 48, 41, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 17, 0 },
   { 34, 35, 0 },
   { 48, 41, 0 },
   { 50, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { 43, 41, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 19, 21, 0 },
   { 37, 38, 0 },
   { 43, 41, 0 },
   { 45, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_before_ffma_pass_op_table[nir_num_search_ops] = {
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         2,
         3,
         19,
         3,
         3,
         20,
         19,
         20,
         19,
      },
   },
   [nir_op_imul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         4,
         5,
         28,
         5,
         5,
         29,
         28,
         29,
         28,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         3,
         4,
         0,
         0,
         5,
         0,
         3,
         6,
         4,
         7,
         3,
         6,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         6,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         6,
         7,
         6,
         6,
         12,
         13,
         31,
         32,
         7,
         7,
         7,
         7,
         14,
         15,
         33,
         34,
         6,
         7,
         16,
         6,
         12,
         13,
         31,
         32,
         6,
         7,
         6,
         6,
         12,
         17,
         31,
         32,
         12,
         14,
         12,
         12,
         12,
         18,
         32,
         32,
         13,
         15,
         13,
         17,
         18,
         13,
         35,
         36,
         31,
         33,
         31,
         31,
         32,
         35,
         31,
         32,
         32,
         34,
         32,
         32,
         32,
         36,
         32,
         32,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         2,
         0,
         0,
         3,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         6,
         4,
         7,
         3,
         6,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         4,
         4,
         6,
         6,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         8,
         9,
         8,
         8,
         21,
         22,
         37,
         38,
         9,
         9,
         9,
         9,
         23,
         24,
         39,
         40,
         8,
         9,
         25,
         8,
         21,
         22,
         37,
         38,
         8,
         9,
         8,
         8,
         21,
         26,
         37,
         38,
         21,
         23,
         21,
         21,
         21,
         27,
         38,
         38,
         22,
         24,
         22,
         26,
         27,
         22,
         41,
         42,
         37,
         39,
         37,
         37,
         38,
         41,
         37,
         38,
         38,
         40,
         38,
         38,
         38,
         42,
         38,
         38,
      },
   },
   [nir_op_fneg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_ineg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         30,
         30,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_before_ffma_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   14,
   18,
   20,
   22,
   24,
   26,
   28,
   30,
   32,
   35,
   39,
   41,
   43,
   47,
   49,
   52,
   54,
   57,
   61,
   66,
   68,
   71,
   73,
   76,
   80,
};

static const nir_algebraic_table nir_opt_algebraic_before_ffma_table = {
   .transforms = nir_opt_algebraic_before_ffma_transforms,
   .transform_offsets = nir_opt_algebraic_before_ffma_transform_offsets,
   .pass_op_table = nir_opt_algebraic_before_ffma_pass_op_table,
   .values = nir_opt_algebraic_before_ffma_values,
   .expression_cond = nir_opt_algebraic_before_ffma_expression_cond,
   .variable_cond = nir_opt_algebraic_before_ffma_variable_cond,
};

bool
nir_opt_algebraic_before_ffma(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[113];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   STATIC_ASSERT(78 == ARRAY_SIZE(nir_opt_algebraic_before_ffma_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ffloor;
   condition_flags[22] = options->lower_ffloor || ((options->lower_doubles_options & nir_lower_dfloor) && !(options->lower_doubles_options & nir_lower_dfloor));
   condition_flags[23] = !options->lower_ffloor;
   condition_flags[24] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[25] = options->lower_ffract;
   condition_flags[26] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[27] = options->lower_fceil;
   condition_flags[28] = options->lower_ffma16;
   condition_flags[29] = options->lower_ffma32;
   condition_flags[30] = options->lower_ffma64;
   condition_flags[31] = options->fuse_ffma16;
   condition_flags[32] = options->fuse_ffma32;
   condition_flags[33] = options->fuse_ffma64;
   condition_flags[34] = options->lower_fdph;
   condition_flags[35] = !options->lower_fdph;
   condition_flags[36] = options->lower_fdot;
   condition_flags[37] = !options->lower_fsat;
   condition_flags[38] = !options->lower_iabs;
   condition_flags[39] = options->lower_fsat;
   condition_flags[40] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[41] = !options->lower_fsign;
   condition_flags[42] = options->lower_umax;
   condition_flags[43] = options->lower_umin;
   condition_flags[44] = !options->lower_umax;
   condition_flags[45] = !options->lower_umin;
   condition_flags[46] = options->lower_scmp;
   condition_flags[47] = options->lower_vector_cmp;
   condition_flags[48] = !options->lower_rotate;
   condition_flags[49] = options->lower_rotate;
   condition_flags[50] = options->lower_fpow;
   condition_flags[51] = !options->lower_fpow;
   condition_flags[52] = options->lower_fdiv;
   condition_flags[53] = options->lower_fsqrt;
   condition_flags[54] = !options->lower_fsqrt;
   condition_flags[55] = options->lower_sincos;
   condition_flags[56] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[57] = !options->lower_extract_byte;
   condition_flags[58] = !options->lower_extract_word;
   condition_flags[59] = options->has_pack_32_4x8;
   condition_flags[60] = options->lower_pack_64_2x32_split;
   condition_flags[61] = options->lower_pack_32_2x16_split;
   condition_flags[62] = options->lower_unpack_64_2x32_split;
   condition_flags[63] = options->lower_unpack_32_2x16_split;
   condition_flags[64] = options->lower_fmod;
   condition_flags[65] = options->lower_uadd_carry;
   condition_flags[66] = options->lower_usub_borrow;
   condition_flags[67] = options->lower_bitfield_insert;
   condition_flags[68] = options->lower_hadd;
   condition_flags[69] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[70] = options->lower_mul_32x16;
   condition_flags[71] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[72] = options->lower_uadd_sat;
   condition_flags[73] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[74] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[75] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[76] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[77] = options->lower_bitfield_insert_to_shifts;
   condition_flags[78] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[79] = options->lower_bitfield_extract;
   condition_flags[80] = options->lower_bitfield_extract_to_shifts;
   condition_flags[81] = options->lower_ifind_msb;
   condition_flags[82] = options->lower_find_msb_to_reverse;
   condition_flags[83] = options->lower_find_lsb;
   condition_flags[84] = options->lower_extract_byte;
   condition_flags[85] = options->lower_extract_word;
   condition_flags[86] = options->lower_pack_unorm_2x16;
   condition_flags[87] = options->lower_pack_unorm_4x8;
   condition_flags[88] = options->lower_pack_snorm_2x16;
   condition_flags[89] = options->lower_pack_snorm_4x8;
   condition_flags[90] = options->lower_unpack_unorm_2x16;
   condition_flags[91] = options->lower_unpack_unorm_4x8;
   condition_flags[92] = options->lower_unpack_snorm_2x16;
   condition_flags[93] = options->lower_unpack_snorm_4x8;
   condition_flags[94] = options->lower_pack_split;
   condition_flags[95] = options->lower_isign;
   condition_flags[96] = !options->lower_isign;
   condition_flags[97] = options->lower_fsign;
   condition_flags[98] = !options->has_imul24;
   condition_flags[99] = !options->has_umul24;
   condition_flags[100] = !options->has_umad24;
   condition_flags[101] = options->has_imul24;
   condition_flags[102] = options->has_umad24;
   condition_flags[103] = options->has_umul24;
   condition_flags[104] = options->has_fused_comp_and_csel;
   condition_flags[105] = options->lower_iadd_sat;
   condition_flags[106] = options->support_8bit_alu;
   condition_flags[107] = options->support_16bit_alu;
   condition_flags[108] = options->lower_ldexp;
   condition_flags[109] = !options->lower_bitfield_reverse;
   condition_flags[110] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[111] = options->lower_fisnormal;
   condition_flags[112] = !options->vectorize_vec2_16bit;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_before_ffma_table);
      }
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 187
 * transforms:
 *    ('flt', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('flt', 'a', ('fneg', 'b'))
 *    ('flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('flt', ('fneg', 'a'), 'b')
 *    ('flt', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('flt', ('fneg', 'a'), 'b')
 *    ('flt', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('flt', 'a', ('fneg', 'b'))
 *    ('~fge', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b'))
 *    ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b')
 *    ('~fge', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('fge', ('fneg', 'a'), 'b')
 *    ('~fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fge', 'a', ('fneg', 'b'))
 *    ('~feq', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b'))
 *    ('~fneu', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b'))
 *    ('fge', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b'))
 *    ('fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b')
 *    ('fge', 0.0, ('fadd(is_used_once)', 'a(is_finite)', 'b')) => ('fge', ('fneg', 'a'), 'b')
 *    ('fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b'))) => ('fge', 'a', ('fneg', 'b'))
 *    ('feq', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b'))
 *    ('fneu', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b'))
 *    ('iand', ('fneu', 'a', 'b'), ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'b'))) => ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a'))
 *    ('iand', ('fneu', 'a', 0.0), ('feq', 'a', 'a')) => ('!flt', 0.0, ('fabs', 'a'))
 *    ('ior', ('feq', 'a', 'b'), ('ior', ('fneu', 'a', 'a'), ('fneu', 'b', 'b'))) => ('inot', ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a')))
 *    ('ior', ('feq', 'a', 0.0), ('fneu', 'a', 'a')) => ('inot', ('!flt', 0.0, ('fabs', 'a')))
 *    ('fneg', ('fneg', 'a')) => a
 *    ('~fadd@16', ('fmul', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmul', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@64', ('fmul', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('~fadd@32', ('fmulz', 'a', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c')
 *    ('fadd@8', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@16', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@32', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fadd@64', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b')
 *    ('fneg', 'a') => ('fmul', 'a', -1.0)
 *    ('iadd', 'a', ('ineg', 'b')) => ('isub', 'a', 'b')
 *    ('ineg', 'a') => ('isub', 0, 'a')
 *    ('iabs', 'a') => ('imax', 'a', ('ineg', 'a'))
 *    ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c')
 *    ('iadd', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), 'c')
 *    ('isub', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), ('ineg', 'c'))
 *    ('vec2(is_only_used_as_float)', ('fneg@16', 'a'), 'b') => ('fmul', ('vec2', 'a', 'b'), ('vec2', -1.0, 1.0))
 *    ('vec2(is_only_used_as_float)', 'a', ('fneg@16', 'b')) => ('fmul', ('vec2', 'a', 'b'), ('vec2', 1.0, -1.0))
 *    ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a')
 *    ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b')
 *    ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b')
 *    ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b')
 *    ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0)
 *    ('~fge', ('fmin(is_used_once)', ('fadd(is_used_once)', 'a', 'b'), ('fadd', 'c', 'd')), 0.0) => ('iand', ('fge', 'a', ('fneg', 'b')), ('fge', 'c', ('fneg', 'd')))
 *    ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a')
 *    ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a')
 *    ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a')
 *    ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a')
 *    ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a')
 *    ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0)
 *    ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a')
 *    ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0)
 *    ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a')
 *    ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0)
 *    ('ior', 'a', 'a') => a
 *    ('iand', 'a', 'a') => a
 *    ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a')))
 *    ('fdot2', 'a', 'b') => ('fdot2_replicated', 'a', 'b')
 *    ('fdot3', 'a', 'b') => ('fdot3_replicated', 'a', 'b')
 *    ('fdot4', 'a', 'b') => ('fdot4_replicated', 'a', 'b')
 *    ('fdph', 'a', 'b') => ('fdph_replicated', 'a', 'b')
 *    ('~flrp', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a')
 *    ('fround_even', 'a') => ('bcsel', ('feq', ('ffract', 'a'), 0.5), ('fadd', ('ffloor', ('fadd', 'a', 0.5)), 1.0), ('ffloor', ('fadd', 'a', 0.5)))
 *    ('~ffma@32', 'a', 2.0, -1.0) => ('flrp', -1.0, 1.0, 'a')
 *    ('~ffma@32', 'a', -2.0, -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a'))
 *    ('~ffma@32', 'a', -2.0, 1.0) => ('flrp', 1.0, -1.0, 'a')
 *    ('~ffma@32', 'a', 2.0, 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a'))
 *    ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, 'a')
 *    ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a'))
 *    ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, 'a')
 *    ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a'))
 *    ('~ffma@32', ('fadd', 'b', ('fneg', 'a')), 'a', 'a') => ('flrp', 'a', 'b', 'a')
 *    ('~ffma@32', 'a', 2.0, ('fneg', ('fmul', 'a', 'a'))) => ('flrp', 'a', 1.0, 'a')
 *    ('~ffma@32', 'a', 2.0, ('fmul', ('fneg', 'a'), 'a')) => ('flrp', 'a', 1.0, 'a')
 *    ('~ffma@32', 'a', ('fneg', 'a'), ('fmul', 2.0, 'a')) => ('flrp', 'a', 1.0, 'a')
 *    ('~fmul@32', 'a', ('fadd', 2.0, ('fneg', 'a'))) => ('flrp', 'a', 1.0, 'a')
 *    ('fmin', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmin', 'a', 'b'))
 *    ('fmax', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmax', 'a', 'b'))
 *    ('bcsel', ('feq', ('fsqrt', 'a(is_not_negative)'), 0.0), 3.4028234663852886e+38, ('frsq', 'a')) => ('fmin', ('frsq', 'a'), 3.4028234663852886e+38)
 *    ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', ('ffma', 'e', 'f', 'g')))
 *    ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('fmul(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', 'e'))
 *    ('~fadd', ('fneg', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffma', ('fneg', 'a'), 'b', ('ffma', ('fneg', 'c'), 'd', ('ffma', ('fneg', 'e'), 'f', 'g')))
 *    ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', ('ffmaz', 'e', 'f', 'g')))
 *    ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('fmulz(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', 'e'))
 *    ('~fadd', ('fneg', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffmaz', ('fneg', 'a'), 'b', ('ffmaz', ('fneg', 'c'), 'd', ('ffmaz', ('fneg', 'e'), 'f', 'g')))
 *    ('ubfe', 'a', 'b', 0) => 0
 *    ('ibfe', 'a', 'b', 0) => 0
 *    ('ubfe', 'a', '#b', '#c') => ('iand', ('ushr', 4294967295, ('ineg', 'c')), ('ushr', 'a', 'b'))
 *    ('ibfe', 'a', '#b', '#c') => ('ishr', ('ishl', 'a', ('ineg', ('iadd', 'b', 'c'))), ('ineg', 'c'))
 *    ('ishl', 'a', 0) => a
 *    ('ishl', 'a', -32) => a
 *    ('ishr', 'a', 0) => a
 *    ('ishr', 'a', -32) => a
 *    ('ushr', 'a', 0) => a
 *    ('extract_i8', ('extract_i8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b')
 *    ('extract_i8', ('extract_u8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b')
 *    ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b')
 *    ('u2u16', ('u2u8', 'a@16')) => ('extract_u8', 'a', 0)
 *    ('u2u16', ('i2i8', 'a@16')) => ('extract_u8', 'a', 0)
 *    ('i2i16', ('u2u8', 'a@16')) => ('extract_i8', 'a', 0)
 *    ('i2i16', ('i2i8', 'a@16')) => ('extract_i8', 'a', 0)
 *    ('u2u32', ('u2u8', 'a@32')) => ('extract_u8', 'a', 0)
 *    ('u2u32', ('i2i8', 'a@32')) => ('extract_u8', 'a', 0)
 *    ('u2u32', ('u2u16', 'a@32')) => ('extract_u16', 'a', 0)
 *    ('u2u32', ('i2i16', 'a@32')) => ('extract_u16', 'a', 0)
 *    ('i2i32', ('u2u8', 'a@32')) => ('extract_i8', 'a', 0)
 *    ('i2i32', ('i2i8', 'a@32')) => ('extract_i8', 'a', 0)
 *    ('i2i32', ('u2u16', 'a@32')) => ('extract_i16', 'a', 0)
 *    ('i2i32', ('i2i16', 'a@32')) => ('extract_i16', 'a', 0)
 *    ('ishl', ('extract_u8', 'a@32', 0), 8) => ('insert_u8', 'a', 1)
 *    ('ishl', ('extract_u8', 'a@32', 0), 16) => ('insert_u8', 'a', 2)
 *    ('ishl', ('extract_u8', 'a@32', 0), 24) => ('insert_u8', 'a', 3)
 *    ('iand', ('ishl', 'a@32', 8), 65280) => ('insert_u8', 'a', 1)
 *    ('iand', ('ishl', 'a@32', 16), 16711680) => ('insert_u8', 'a', 2)
 *    ('iand', ('ishl', 'a@32', 24), 4278190080) => ('insert_u8', 'a', 3)
 *    ('ishl', 'a@32', 24) => ('insert_u8', 'a', 3)
 *    ('ishl', 'a@32', 16) => ('insert_u16', 'a', 1)
 *    ('insert_u8', ('extract_u8', 'a', 0), 'b') => ('insert_u8', 'a', 'b')
 *    ('insert_u16', ('extract_u16', 'a', 0), 'b') => ('insert_u16', 'a', 'b')
 *    ('iand', ('ine(is_used_once)', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('iand', ('ine(is_used_once)', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0)
 *    ('ior', ('ieq(is_used_once)', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0)
 *    ('~fadd@16', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f16', ('inot', 'b@bool'))) => ('b2f16', ('inot', ('ior', 'a', 'b')))
 *    ('~fadd@32', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f32', ('inot', 'b@bool'))) => ('b2f32', ('inot', ('ior', 'a', 'b')))
 *    ('~fadd@64', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c'))
 *    ('bcsel', 'a', 0, ('b2f64', ('inot', 'b@bool'))) => ('b2f64', ('inot', ('ior', 'a', 'b')))
 *    ('bcsel', 'a', ('fadd(is_used_once)', 'b', 'c'), ('fadd', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('fadd', 'b', 'c'), ('fadd(is_used_once)', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd'))
 *    ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e'))
 *    ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd')
 *    ('~f2f32', ('fabs', ('f2fmp', 'a'))) => ('fabs', 'a')
 *    ('~f2f32', ('fceil', ('f2fmp', 'a'))) => ('fceil', 'a')
 *    ('~f2f32', ('fcos', ('f2fmp', 'a'))) => ('fcos', 'a')
 *    ('~f2f32', ('fddx', ('f2fmp', 'a'))) => ('fddx', 'a')
 *    ('~f2f32', ('fddx_coarse', ('f2fmp', 'a'))) => ('fddx_coarse', 'a')
 *    ('~f2f32', ('fddx_fine', ('f2fmp', 'a'))) => ('fddx_fine', 'a')
 *    ('~f2f32', ('fddy', ('f2fmp', 'a'))) => ('fddy', 'a')
 *    ('~f2f32', ('fddy_coarse', ('f2fmp', 'a'))) => ('fddy_coarse', 'a')
 *    ('~f2f32', ('fddy_fine', ('f2fmp', 'a'))) => ('fddy_fine', 'a')
 *    ('~f2f32', ('fexp2', ('f2fmp', 'a'))) => ('fexp2', 'a')
 *    ('~f2f32', ('ffloor', ('f2fmp', 'a'))) => ('ffloor', 'a')
 *    ('~f2f32', ('ffract', ('f2fmp', 'a'))) => ('ffract', 'a')
 *    ('~f2f32', ('flog2', ('f2fmp', 'a'))) => ('flog2', 'a')
 *    ('~f2f32', ('fneg', ('f2fmp', 'a'))) => ('fneg', 'a')
 *    ('~f2f32', ('frcp', ('f2fmp', 'a'))) => ('frcp', 'a')
 *    ('~f2f32', ('fround_even', ('f2fmp', 'a'))) => ('fround_even', 'a')
 *    ('~f2f32', ('frsq', ('f2fmp', 'a'))) => ('frsq', 'a')
 *    ('~f2f32', ('fsat', ('f2fmp', 'a'))) => ('fsat', 'a')
 *    ('~f2f32', ('fsign', ('f2fmp', 'a'))) => ('fsign', 'a')
 *    ('~f2f32', ('fsin', ('f2fmp', 'a'))) => ('fsin', 'a')
 *    ('~f2f32', ('fsqrt', ('f2fmp', 'a'))) => ('fsqrt', 'a')
 *    ('~f2f32', ('fadd', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fadd', 'a', 'b')
 *    ('~f2f32', ('fdiv', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fdiv', 'a', 'b')
 *    ('~f2f32', ('fmax', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmax', 'a', 'b')
 *    ('~f2f32', ('fmin', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmin', 'a', 'b')
 *    ('~f2f32', ('fmod', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmod', 'a', 'b')
 *    ('~f2f32', ('fmul', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmul', 'a', 'b')
 *    ('~f2f32', ('fpow', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fpow', 'a', 'b')
 *    ('~f2f32', ('frem', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('frem', 'a', 'b')
 *    ('~f2f32', ('ffma', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('ffma', 'a', 'b', 'c')
 *    ('~f2f32', ('flrp', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('flrp', 'a', 'b', 'c')
 *    ('~feq', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('feq', 'a', 'b')
 *    ('~fge', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fge', 'a', 'b')
 *    ('~flt', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('flt', 'a', 'b')
 *    ('~fneu', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fneu', 'a', 'b')
 *    ('f2fmp', 'a') => ('f2f16', 'a')
 *    ('f2imp', 'a') => ('f2i16', 'a')
 *    ('f2ump', 'a') => ('f2u16', 'a')
 *    ('i2imp', 'a') => ('i2i16', 'a')
 *    ('i2fmp', 'a') => ('i2f16', 'a')
 *    ('i2imp', 'a') => ('u2u16', 'a')
 *    ('u2fmp', 'a') => ('u2f16', 'a')
 *    ('fisfinite', 'a') => ('flt', ('fabs', 'a'), inf)
 */


static const nir_search_value_union nir_opt_algebraic_late_values[] = {
   /* ('flt', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('flt', 'a', ('fneg', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 1 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 2, 3 },
      -1,
   } },

   /* replace1968_0 -> 0 in the cache */
   /* replace1968_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 5 },
      -1,
   } },

   /* ('flt', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('flt', ('fneg', 'a'), 'b') */
   /* search1969_0_0_0 -> 0 in the cache */
   /* search1969_0_0_1 -> 1 in the cache */
   /* search1969_0_0 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 2 },
      0,
   } },
   /* search1969_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 7, 3 },
      -1,
   } },

   /* replace1969_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace1969_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 9, 1 },
      -1,
   } },

   /* ('flt', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('flt', ('fneg', 'a'), 'b') */
   /* search1970_0 -> 3 in the cache */
   /* search1970_1_0 -> 0 in the cache */
   /* search1970_1_1 -> 1 in the cache */
   /* search1970_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3, 2 },
      -1,
   } },

   /* replace1970_0_0 -> 0 in the cache */
   /* replace1970_0 -> 9 in the cache */
   /* replace1970_1 -> 1 in the cache */
   /* replace1970 -> 10 in the cache */

   /* ('flt', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('flt', 'a', ('fneg', 'b')) */
   /* search1971_0 -> 3 in the cache */
   /* search1971_1_0_0 -> 0 in the cache */
   /* search1971_1_0_1 -> 1 in the cache */
   /* search1971_1_0 -> 2 in the cache */
   /* search1971_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 1,
      { 3, 7 },
      -1,
   } },

   /* replace1971_0 -> 0 in the cache */
   /* replace1971_1_0 -> 1 in the cache */
   /* replace1971_1 -> 5 in the cache */
   /* replace1971 -> 6 in the cache */

   /* ('~fge', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b')) */
   /* search1972_0_0 -> 0 in the cache */
   /* search1972_0_1 -> 1 in the cache */
   /* search1972_0 -> 2 in the cache */
   /* search1972_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 2, 3 },
      -1,
   } },

   /* replace1972_0 -> 0 in the cache */
   /* replace1972_1_0 -> 1 in the cache */
   /* replace1972_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 5 },
      -1,
   } },

   /* ('~fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b') */
   /* search1973_0_0_0 -> 0 in the cache */
   /* search1973_0_0_1 -> 1 in the cache */
   /* search1973_0_0 -> 2 in the cache */
   /* search1973_0 -> 7 in the cache */
   /* search1973_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 7, 3 },
      -1,
   } },

   /* replace1973_0_0 -> 0 in the cache */
   /* replace1973_0 -> 9 in the cache */
   /* replace1973_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 9, 1 },
      -1,
   } },

   /* ('~fge', 0.0, ('fadd(is_used_once)', 'a', 'b')) => ('fge', ('fneg', 'a'), 'b') */
   /* search1974_0 -> 3 in the cache */
   /* search1974_1_0 -> 0 in the cache */
   /* search1974_1_1 -> 1 in the cache */
   /* search1974_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 2 },
      -1,
   } },

   /* replace1974_0_0 -> 0 in the cache */
   /* replace1974_0 -> 9 in the cache */
   /* replace1974_1 -> 1 in the cache */
   /* replace1974 -> 16 in the cache */

   /* ('~fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a', 'b'))) => ('fge', 'a', ('fneg', 'b')) */
   /* search1975_0 -> 3 in the cache */
   /* search1975_1_0_0 -> 0 in the cache */
   /* search1975_1_0_1 -> 1 in the cache */
   /* search1975_1_0 -> 2 in the cache */
   /* search1975_1 -> 7 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 7 },
      -1,
   } },

   /* replace1975_0 -> 0 in the cache */
   /* replace1975_1_0 -> 1 in the cache */
   /* replace1975_1 -> 5 in the cache */
   /* replace1975 -> 14 in the cache */

   /* ('~feq', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b')) */
   /* search1976_0_0 -> 0 in the cache */
   /* search1976_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 0, 1 },
      0,
   } },
   /* search1976_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 19, 3 },
      -1,
   } },

   /* replace1976_0 -> 0 in the cache */
   /* replace1976_1_0 -> 1 in the cache */
   /* replace1976_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 5 },
      -1,
   } },

   /* ('~fneu', ('fadd(is_used_once)', 'a', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b')) */
   /* search1977_0_0 -> 0 in the cache */
   /* search1977_0_1 -> 1 in the cache */
   /* search1977_0 -> 19 in the cache */
   /* search1977_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 19, 3 },
      -1,
   } },

   /* replace1977_0 -> 0 in the cache */
   /* replace1977_1_0 -> 1 in the cache */
   /* replace1977_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 5 },
      -1,
   } },

   /* ('fge', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fge', 'a', ('fneg', 'b')) */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      0,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search1978_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 24, 1 },
      0,
   } },
   /* search1978_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 25, 3 },
      -1,
   } },

   /* replace1978_0 -> 0 in the cache */
   /* replace1978_1_0 -> 1 in the cache */
   /* replace1978_1 -> 5 in the cache */
   /* replace1978 -> 14 in the cache */

   /* ('fge', ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b')), 0.0) => ('fge', ('fneg', 'a'), 'b') */
   /* search1979_0_0_0 -> 24 in the cache */
   /* search1979_0_0_1 -> 1 in the cache */
   /* search1979_0_0 -> 25 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 25 },
      0,
   } },
   /* search1979_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 27, 3 },
      -1,
   } },

   /* replace1979_0_0 -> 0 in the cache */
   /* replace1979_0 -> 9 in the cache */
   /* replace1979_1 -> 1 in the cache */
   /* replace1979 -> 16 in the cache */

   /* ('fge', 0.0, ('fadd(is_used_once)', 'a(is_finite)', 'b')) => ('fge', ('fneg', 'a'), 'b') */
   /* search1980_0 -> 3 in the cache */
   /* search1980_1_0 -> 24 in the cache */
   /* search1980_1_1 -> 1 in the cache */
   /* search1980_1 -> 25 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 25 },
      -1,
   } },

   /* replace1980_0_0 -> 0 in the cache */
   /* replace1980_0 -> 9 in the cache */
   /* replace1980_1 -> 1 in the cache */
   /* replace1980 -> 16 in the cache */

   /* ('fge', 0.0, ('fneg(is_used_once)', ('fadd(is_used_once)', 'a(is_finite)', 'b'))) => ('fge', 'a', ('fneg', 'b')) */
   /* search1981_0 -> 3 in the cache */
   /* search1981_1_0_0 -> 24 in the cache */
   /* search1981_1_0_1 -> 1 in the cache */
   /* search1981_1_0 -> 25 in the cache */
   /* search1981_1 -> 27 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 1,
      { 3, 27 },
      -1,
   } },

   /* replace1981_0 -> 0 in the cache */
   /* replace1981_1_0 -> 1 in the cache */
   /* replace1981_1 -> 5 in the cache */
   /* replace1981 -> 14 in the cache */

   /* ('feq', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('feq', 'a', ('fneg', 'b')) */
   /* search1982_0_0 -> 24 in the cache */
   /* search1982_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 24, 1 },
      0,
   } },
   /* search1982_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 2,
      { 31, 3 },
      -1,
   } },

   /* replace1982_0 -> 0 in the cache */
   /* replace1982_1_0 -> 1 in the cache */
   /* replace1982_1 -> 5 in the cache */
   /* replace1982 -> 21 in the cache */

   /* ('fneu', ('fadd(is_used_once)', 'a(is_finite)', 'b'), 0.0) => ('fneu', 'a', ('fneg', 'b')) */
   /* search1983_0_0 -> 24 in the cache */
   /* search1983_0_1 -> 1 in the cache */
   /* search1983_0 -> 31 in the cache */
   /* search1983_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 2,
      { 31, 3 },
      -1,
   } },

   /* replace1983_0 -> 0 in the cache */
   /* replace1983_1_0 -> 1 in the cache */
   /* replace1983_1 -> 5 in the cache */
   /* replace1983 -> 23 in the cache */

   /* ('iand', ('fneu', 'a', 'b'), ('iand', ('feq', 'a', 'a'), ('feq', 'b', 'b'))) => ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a')) */
   /* search1984_0_0 -> 0 in the cache */
   /* search1984_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search1984_1_0_0 -> 0 in the cache */
   /* search1984_1_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search1984_1_1_0 -> 1 in the cache */
   /* search1984_1_1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 1, 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      2, 1,
      { 35, 36 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 34, 37 },
      -1,
   } },

   /* replace1984_0_0 -> 0 in the cache */
   /* replace1984_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   /* replace1984_1_0 -> 1 in the cache */
   /* replace1984_1_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 1, 0 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 39, 40 },
      -1,
   } },

   /* ('iand', ('fneu', 'a', 0.0), ('feq', 'a', 'a')) => ('!flt', 0.0, ('fabs', 'a')) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      1, 1,
      { 42, 43 },
      -1,
   } },
   /* search1985_1_0 -> 42 in the cache */
   /* search1985_1_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      -1, 0,
      { 42, 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 2,
      { 44, 45 },
      -1,
   } },

   /* replace1985_0 -> 43 in the cache */
   /* replace1985_1_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      true,
      false,
      nir_op_flt,
      -1, 0,
      { 43, 47 },
      -1,
   } },

   /* ('ior', ('feq', 'a', 'b'), ('ior', ('fneu', 'a', 'a'), ('fneu', 'b', 'b'))) => ('inot', ('ior', ('!flt', 'a', 'b'), ('!flt', 'b', 'a'))) */
   /* search1986_0_0 -> 0 in the cache */
   /* search1986_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 0, 1 },
      -1,
   } },
   /* search1986_1_0_0 -> 0 in the cache */
   /* search1986_1_0_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 0, 0 },
      -1,
   } },
   /* search1986_1_1_0 -> 1 in the cache */
   /* search1986_1_1_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 1, 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      2, 1,
      { 50, 51 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 49, 52 },
      -1,
   } },

   /* replace1986_0_0_0 -> 0 in the cache */
   /* replace1986_0_0_1 -> 1 in the cache */
   /* replace1986_0_0 -> 39 in the cache */
   /* replace1986_0_1_0 -> 1 in the cache */
   /* replace1986_0_1_1 -> 0 in the cache */
   /* replace1986_0_1 -> 40 in the cache */
   /* replace1986_0 -> 41 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 41 },
      -1,
   } },

   /* ('ior', ('feq', 'a', 0.0), ('fneu', 'a', 'a')) => ('inot', ('!flt', 0.0, ('fabs', 'a'))) */
   /* search1987_0_0 -> 42 in the cache */
   /* search1987_0_1 -> 43 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      1, 1,
      { 42, 43 },
      -1,
   } },
   /* search1987_1_0 -> 42 in the cache */
   /* search1987_1_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      -1, 0,
      { 42, 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 2,
      { 55, 56 },
      -1,
   } },

   /* replace1987_0_0 -> 43 in the cache */
   /* replace1987_0_1_0 -> 42 in the cache */
   /* replace1987_0_1 -> 47 in the cache */
   /* replace1987_0 -> 48 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 48 },
      -1,
   } },

   /* ('fneg', ('fneg', 'a')) => a */
   /* search1988_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 59 },
      -1,
   } },

   /* replace1988 -> 42 in the cache */

   /* ('~fadd@16', ('fmul', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 61, 62 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 16 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 63, 64 },
      -1,
   } },

   /* replace1989_0 -> 61 in the cache */
   /* replace1989_1 -> 62 in the cache */
   /* replace1989_2 -> 64 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 61, 62, 64 },
      -1,
   } },

   /* ('~fadd@32', ('fmul', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 67, 68 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 69, 70 },
      -1,
   } },

   /* replace1990_0 -> 67 in the cache */
   /* replace1990_1 -> 68 in the cache */
   /* replace1990_2 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 67, 68, 70 },
      -1,
   } },

   /* ('~fadd@64', ('fmul', 'a', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 73, 74 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 75, 76 },
      -1,
   } },

   /* replace1991_0 -> 73 in the cache */
   /* replace1991_1 -> 74 in the cache */
   /* replace1991_2 -> 76 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 73, 74, 76 },
      -1,
   } },

   /* ('~fadd@32', ('fmulz', 'a', 'b'), 'c') => ('ffmaz', 'a', 'b', 'c') */
   /* search1992_0_0 -> 67 in the cache */
   /* search1992_0_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      1, 1,
      { 67, 68 },
      -1,
   } },
   /* search1992_1 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 79, 70 },
      -1,
   } },

   /* replace1992_0 -> 67 in the cache */
   /* replace1992_1 -> 68 in the cache */
   /* replace1992_2 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 67, 68, 70 },
      -1,
   } },

   /* ('fadd@8', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, 8 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 8 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 83 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 82, 84 },
      -1,
   } },

   /* replace1993_0 -> 82 in the cache */
   /* replace1993_1 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 82, 83 },
      -1,
   } },

   /* ('fadd@16', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search1994_0 -> 61 in the cache */
   /* search1994_1_0 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 62 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 61, 87 },
      -1,
   } },

   /* replace1994_0 -> 61 in the cache */
   /* replace1994_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 61, 62 },
      -1,
   } },

   /* ('fadd@32', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search1995_0 -> 67 in the cache */
   /* search1995_1_0 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 68 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 67, 90 },
      -1,
   } },

   /* replace1995_0 -> 67 in the cache */
   /* replace1995_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('fadd@64', 'a', ('fneg', 'b')) => ('fsub', 'a', 'b') */
   /* search1996_0 -> 73 in the cache */
   /* search1996_1_0 -> 74 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 74 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 73, 93 },
      -1,
   } },

   /* replace1996_0 -> 73 in the cache */
   /* replace1996_1 -> 74 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fsub,
      -1, 0,
      { 73, 74 },
      -1,
   } },

   /* ('fneg', 'a') => ('fmul', 'a', -1.0) */
   /* search1997_0 -> 42 in the cache */
   /* search1997 -> 59 in the cache */

   /* replace1997_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 42, 96 },
      -1,
   } },

   /* ('iadd', 'a', ('ineg', 'b')) => ('isub', 'a', 'b') */
   /* search1998_0 -> 0 in the cache */
   /* search1998_1_0 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 1 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 0, 98 },
      -1,
   } },

   /* replace1998_0 -> 0 in the cache */
   /* replace1998_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('ineg', 'a') => ('isub', 0, 'a') */
   /* search1999_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 42 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   /* replace1999_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 102, 42 },
      -1,
   } },

   /* ('iabs', 'a') => ('imax', 'a', ('ineg', 'a')) */
   /* search2000_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iabs,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2000_0 -> 42 in the cache */
   /* replace2000_1_0 -> 42 in the cache */
   /* replace2000_1 -> 101 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_imax,
      0, 1,
      { 42, 101 },
      -1,
   } },

   /* ('iadd', ('iadd(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      1, 1,
      { 106, 107 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 2,
      { 108, 109 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 111, 112, 113 },
      -1,
   } },

   /* ('iadd', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), 'c') */
   /* search2002_0_0 -> 106 in the cache */
   /* search2002_0_1 -> 107 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 106, 107 },
      0,
   } },
   /* search2002_1 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 115, 109 },
      -1,
   } },

   /* replace2002_0 -> 111 in the cache */
   /* replace2002_1_0 -> 112 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 112 },
      -1,
   } },
   /* replace2002_2 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 111, 117, 113 },
      -1,
   } },

   /* ('isub', ('isub(is_used_once)', 'a(is_not_const)', 'b(is_not_const)'), 'c(is_not_const)') => ('iadd3', 'a', ('ineg', 'b'), ('ineg', 'c')) */
   /* search2003_0_0 -> 106 in the cache */
   /* search2003_0_1 -> 107 in the cache */
   /* search2003_0 -> 115 in the cache */
   /* search2003_1 -> 109 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_isub,
      -1, 0,
      { 115, 109 },
      -1,
   } },

   /* replace2003_0 -> 111 in the cache */
   /* replace2003_1_0 -> 112 in the cache */
   /* replace2003_1 -> 117 in the cache */
   /* replace2003_2_0 -> 113 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 113 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_iadd3,
      0, 1,
      { 111, 117, 120 },
      -1,
   } },

   /* ('vec2(is_only_used_as_float)', ('fneg@16', 'a'), 'b') => ('fmul', ('vec2', 'a', 'b'), ('vec2', -1.0, 1.0)) */
   /* search2004_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 61 },
      -1,
   } },
   /* search2004_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 122, 62 },
      1,
   } },

   /* replace2004_0_0 -> 61 in the cache */
   /* replace2004_0_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 61, 62 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 125, 126 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 124, 127 },
      -1,
   } },

   /* ('vec2(is_only_used_as_float)', 'a', ('fneg@16', 'b')) => ('fmul', ('vec2', 'a', 'b'), ('vec2', 1.0, -1.0)) */
   /* search2005_0 -> 61 in the cache */
   /* search2005_1_0 -> 62 in the cache */
   /* search2005_1 -> 87 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 61, 87 },
      1,
   } },

   /* replace2005_0_0 -> 61 in the cache */
   /* replace2005_0_1 -> 62 in the cache */
   /* replace2005_0 -> 124 in the cache */
   /* replace2005_1_0 -> 126 in the cache */
   /* replace2005_1_1 -> 125 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_vec2,
      -1, 0,
      { 126, 125 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 124, 130 },
      -1,
   } },

   /* ('flt', '#b(is_gt_0_and_lt_1)', ('fsat(is_used_once)', 'a')) => ('flt', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 133 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 132, 134 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace2006_1 -> 133 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 136, 133 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fge', 'a', 'b') */
   /* search2007_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 0 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      true,
      nir_type_invalid,
      2,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 138, 139 },
      -1,
   } },

   /* replace2007_0 -> 0 in the cache */
   /* replace2007_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('feq', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('feq', 'a', 'b') */
   /* search2008_0_0 -> 0 in the cache */
   /* search2008_0 -> 138 in the cache */
   /* search2008_1 -> 139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 138, 139 },
      -1,
   } },

   /* replace2008_0 -> 0 in the cache */
   /* replace2008_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fneu', ('fsat(is_used_once)', 'a'), '#b(is_gt_0_and_lt_1)') => ('fneu', 'a', 'b') */
   /* search2009_0_0 -> 0 in the cache */
   /* search2009_0 -> 138 in the cache */
   /* search2009_1 -> 139 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 138, 139 },
      -1,
   } },

   /* replace2009_0 -> 0 in the cache */
   /* replace2009_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fge', ('fsat(is_used_once)', 'a'), 1.0) => ('fge', 'a', 1.0) */
   /* search2010_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 42 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 146, 147 },
      -1,
   } },

   /* replace2010_0 -> 42 in the cache */
   /* replace2010_1 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 42, 147 },
      -1,
   } },

   /* ('~fge', ('fmin(is_used_once)', ('fadd(is_used_once)', 'a', 'b'), ('fadd', 'c', 'd')), 0.0) => ('iand', ('fge', 'a', ('fneg', 'b')), ('fge', 'c', ('fneg', 'd'))) */
   { .variable = {
      { nir_search_value_variable, -4 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 150, 151 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -4 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 153, 154 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 152, 155 },
      0,
   } },
   { .constant = {
      { nir_search_value_constant, -4 },
      nir_type_float, { 0x0 /* 0.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 3,
      { 156, 157 },
      -1,
   } },

   /* replace2011_0_0 -> 150 in the cache */
   /* replace2011_0_1_0 -> 151 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 151 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 150, 159 },
      -1,
   } },
   /* replace2011_1_0 -> 153 in the cache */
   /* replace2011_1_1_0 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 154 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 153, 161 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 160, 162 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), ('fneg', 'b')) => ('flt', 'b', 'a') */
   /* search2012_0_0 -> 0 in the cache */
   /* search2012_0 -> 9 in the cache */
   /* search2012_1_0 -> 1 in the cache */
   /* search2012_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 9, 5 },
      -1,
   } },

   /* replace2012_0 -> 1 in the cache */
   /* replace2012_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 1, 0 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), ('fneg', 'b')) => ('fge', 'b', 'a') */
   /* search2013_0_0 -> 0 in the cache */
   /* search2013_0 -> 9 in the cache */
   /* search2013_1_0 -> 1 in the cache */
   /* search2013_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 9, 5 },
      -1,
   } },

   /* replace2013_0 -> 1 in the cache */
   /* replace2013_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 1, 0 },
      -1,
   } },

   /* ('feq', ('fneg', 'a'), ('fneg', 'b')) => ('feq', 'b', 'a') */
   /* search2014_0_0 -> 0 in the cache */
   /* search2014_0 -> 9 in the cache */
   /* search2014_1_0 -> 1 in the cache */
   /* search2014_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 9, 5 },
      -1,
   } },

   /* replace2014_0 -> 1 in the cache */
   /* replace2014_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 1, 0 },
      -1,
   } },

   /* ('fneu', ('fneg', 'a'), ('fneg', 'b')) => ('fneu', 'b', 'a') */
   /* search2015_0_0 -> 0 in the cache */
   /* search2015_0 -> 9 in the cache */
   /* search2015_1_0 -> 1 in the cache */
   /* search2015_1 -> 5 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 9, 5 },
      -1,
   } },

   /* replace2015_0 -> 1 in the cache */
   /* replace2015_1 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 1, 0 },
      -1,
   } },

   /* ('flt', ('fneg', 'a'), -1.0) => ('flt', 1.0, 'a') */
   /* search2016_0_0 -> 42 in the cache */
   /* search2016_0 -> 59 in the cache */
   /* search2016_1 -> 96 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 59, 96 },
      -1,
   } },

   /* replace2016_0 -> 147 in the cache */
   /* replace2016_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 147, 42 },
      -1,
   } },

   /* ('flt', -1.0, ('fneg', 'a')) => ('flt', 'a', 1.0) */
   /* search2017_0 -> 96 in the cache */
   /* search2017_1_0 -> 42 in the cache */
   /* search2017_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 96, 59 },
      -1,
   } },

   /* replace2017_0 -> 42 in the cache */
   /* replace2017_1 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 42, 147 },
      -1,
   } },

   /* ('fge', ('fneg', 'a'), -1.0) => ('fge', 1.0, 'a') */
   /* search2018_0_0 -> 42 in the cache */
   /* search2018_0 -> 59 in the cache */
   /* search2018_1 -> 96 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 59, 96 },
      -1,
   } },

   /* replace2018_0 -> 147 in the cache */
   /* replace2018_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 147, 42 },
      -1,
   } },

   /* ('fge', -1.0, ('fneg', 'a')) => ('fge', 'a', 1.0) */
   /* search2019_0 -> 96 in the cache */
   /* search2019_1_0 -> 42 in the cache */
   /* search2019_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 96, 59 },
      -1,
   } },

   /* replace2019_0 -> 42 in the cache */
   /* replace2019_1 -> 147 in the cache */
   /* replace2019 -> 149 in the cache */

   /* ('fneu', ('fneg', 'a'), -1.0) => ('fneu', 1.0, 'a') */
   /* search2020_0_0 -> 42 in the cache */
   /* search2020_0 -> 59 in the cache */
   /* search2020_1 -> 96 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 59, 96 },
      -1,
   } },

   /* replace2020_0 -> 147 in the cache */
   /* replace2020_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 147, 42 },
      -1,
   } },

   /* ('feq', -1.0, ('fneg', 'a')) => ('feq', 'a', 1.0) */
   /* search2021_0 -> 96 in the cache */
   /* search2021_1_0 -> 42 in the cache */
   /* search2021_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 96, 59 },
      -1,
   } },

   /* replace2021_0 -> 42 in the cache */
   /* replace2021_1 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 42, 147 },
      -1,
   } },

   /* ('ior', 'a', 'a') => a */
   /* search2022_0 -> 42 in the cache */
   /* search2022_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ior,
      -1, 0,
      { 42, 42 },
      -1,
   } },

   /* replace2022 -> 42 in the cache */

   /* ('iand', 'a', 'a') => a */
   /* search2023_0 -> 42 in the cache */
   /* search2023_1 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_iand,
      -1, 0,
      { 42, 42 },
      -1,
   } },

   /* replace2023 -> 42 in the cache */

   /* ('~fadd', ('fneg(is_used_once)', ('fsat(is_used_once)', 'a(is_not_fmul)')), 1.0) => ('fsat', ('fadd', 1.0, ('fneg', 'a'))) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      3,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 185 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 186 },
      0,
   } },
   /* search2024_1 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 187, 147 },
      -1,
   } },

   /* replace2024_0_0 -> 147 in the cache */
   /* replace2024_0_1_0 -> 42 in the cache */
   /* replace2024_0_1 -> 59 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 147, 59 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 1,
      { 189 },
      -1,
   } },

   /* ('fdot2', 'a', 'b') => ('fdot2_replicated', 'a', 'b') */
   /* search2025_0 -> 0 in the cache */
   /* search2025_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2025_0 -> 0 in the cache */
   /* replace2025_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdot3', 'a', 'b') => ('fdot3_replicated', 'a', 'b') */
   /* search2026_0 -> 0 in the cache */
   /* search2026_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2026_0 -> 0 in the cache */
   /* replace2026_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdot4', 'a', 'b') => ('fdot4_replicated', 'a', 'b') */
   /* search2027_0 -> 0 in the cache */
   /* search2027_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* replace2027_0 -> 0 in the cache */
   /* replace2027_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 1 },
      -1,
   } },

   /* ('fdph', 'a', 'b') => ('fdph_replicated', 'a', 'b') */
   /* search2028_0 -> 0 in the cache */
   /* search2028_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* replace2028_0 -> 0 in the cache */
   /* replace2028_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('~flrp', ('fadd(is_used_once)', 'a', 'b'), ('fadd(is_used_once)', 'a', 'c'), 'd') => ('fadd', ('flrp', 'b', 'c', 'd'), 'a') */
   /* search2029_0_0 -> 150 in the cache */
   /* search2029_0_1 -> 151 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 150, 151 },
      0,
   } },
   /* search2029_1_0 -> 150 in the cache */
   /* search2029_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 150, 153 },
      0,
   } },
   /* search2029_2 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      true,
      false,
      false,
      nir_op_flrp,
      -1, 2,
      { 199, 200, 154 },
      -1,
   } },

   /* replace2029_0_0 -> 151 in the cache */
   /* replace2029_0_1 -> 153 in the cache */
   /* replace2029_0_2 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 151, 153, 154 },
      -1,
   } },
   /* replace2029_1 -> 150 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 202, 150 },
      -1,
   } },

   /* ('fround_even', 'a') => ('bcsel', ('feq', ('ffract', 'a'), 0.5), ('fadd', ('ffloor', ('fadd', 'a', 0.5)), 1.0), ('ffloor', ('fadd', 'a', 0.5))) */
   /* search2030_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2030_0_0_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 42 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x3fe0000000000000 /* 0.5 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 205, 206 },
      -1,
   } },
   /* replace2030_1_0_0_0 -> 42 in the cache */
   /* replace2030_1_0_0_1 -> 206 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 42, 206 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 208 },
      -1,
   } },
   /* replace2030_1_1 -> 147 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 2,
      { 209, 147 },
      -1,
   } },
   /* replace2030_2_0_0 -> 42 in the cache */
   /* replace2030_2_0_1 -> 206 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fadd,
      3, 1,
      { 42, 206 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 1,
      { 211 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 4,
      { 207, 210, 212 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, -1.0) => ('flrp', -1.0, 1.0, 'a') */
   /* search2031_0 -> 67 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x4000000000000000 /* 2.0 */ },
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 67, 214, 215 },
      -1,
   } },

   /* replace2031_0 -> 215 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   /* replace2031_2 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 215, 217, 67 },
      -1,
   } },

   /* ('~ffma@32', 'a', -2.0, -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a')) */
   /* search2032_0 -> 67 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_float, { 0xc000000000000000 /* -2.0 */ },
   } },
   /* search2032_2 -> 215 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 67, 219, 215 },
      -1,
   } },

   /* replace2032_0 -> 215 in the cache */
   /* replace2032_1 -> 217 in the cache */
   /* replace2032_2_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 215, 217, 221 },
      -1,
   } },

   /* ('~ffma@32', 'a', -2.0, 1.0) => ('flrp', 1.0, -1.0, 'a') */
   /* search2033_0 -> 67 in the cache */
   /* search2033_1 -> 219 in the cache */
   /* search2033_2 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 67, 219, 217 },
      -1,
   } },

   /* replace2033_0 -> 217 in the cache */
   /* replace2033_1 -> 215 in the cache */
   /* replace2033_2 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 217, 215, 67 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a')) */
   /* search2034_0 -> 67 in the cache */
   /* search2034_1 -> 214 in the cache */
   /* search2034_2 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 67, 214, 217 },
      -1,
   } },

   /* replace2034_0 -> 217 in the cache */
   /* replace2034_1 -> 215 in the cache */
   /* replace2034_2_0 -> 67 in the cache */
   /* replace2034_2 -> 221 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 217, 215, 221 },
      -1,
   } },

   /* ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, 'a') */
   /* search2035_0_0 -> 214 in the cache */
   /* search2035_0_1 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 214, 67 },
      0,
   } },
   /* search2035_1 -> 215 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 227, 215 },
      -1,
   } },

   /* replace2035_0 -> 215 in the cache */
   /* replace2035_1 -> 217 in the cache */
   /* replace2035_2 -> 67 in the cache */
   /* replace2035 -> 218 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), -1.0) => ('flrp', -1.0, 1.0, ('fneg', 'a')) */
   /* search2036_0_0 -> 219 in the cache */
   /* search2036_0_1 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 219, 67 },
      0,
   } },
   /* search2036_1 -> 215 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 229, 215 },
      -1,
   } },

   /* replace2036_0 -> 215 in the cache */
   /* replace2036_1 -> 217 in the cache */
   /* replace2036_2_0 -> 67 in the cache */
   /* replace2036_2 -> 221 in the cache */
   /* replace2036 -> 222 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', -2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, 'a') */
   /* search2037_0_0 -> 219 in the cache */
   /* search2037_0_1 -> 67 in the cache */
   /* search2037_0 -> 229 in the cache */
   /* search2037_1 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 229, 217 },
      -1,
   } },

   /* replace2037_0 -> 217 in the cache */
   /* replace2037_1 -> 215 in the cache */
   /* replace2037_2 -> 67 in the cache */
   /* replace2037 -> 224 in the cache */

   /* ('~fadd@32', ('fmul(is_used_once)', 2.0, 'a'), 1.0) => ('flrp', 1.0, -1.0, ('fneg', 'a')) */
   /* search2038_0_0 -> 214 in the cache */
   /* search2038_0_1 -> 67 in the cache */
   /* search2038_0 -> 227 in the cache */
   /* search2038_1 -> 217 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 227, 217 },
      -1,
   } },

   /* replace2038_0 -> 217 in the cache */
   /* replace2038_1 -> 215 in the cache */
   /* replace2038_2_0 -> 67 in the cache */
   /* replace2038_2 -> 221 in the cache */
   /* replace2038 -> 226 in the cache */

   /* ('~ffma@32', ('fadd', 'b', ('fneg', 'a')), 'a', 'a') => ('flrp', 'a', 'b', 'a') */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 234 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 233, 235 },
      -1,
   } },
   /* search2039_1 -> 234 in the cache */
   /* search2039_2 -> 234 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 236, 234, 234 },
      -1,
   } },

   /* replace2039_0 -> 234 in the cache */
   /* replace2039_1 -> 233 in the cache */
   /* replace2039_2 -> 234 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 234, 233, 234 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, ('fneg', ('fmul', 'a', 'a'))) => ('flrp', 'a', 1.0, 'a') */
   /* search2040_0 -> 67 in the cache */
   /* search2040_1 -> 214 in the cache */
   /* search2040_2_0_0 -> 67 in the cache */
   /* search2040_2_0_1 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      -1, 0,
      { 67, 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 239 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 67, 214, 240 },
      -1,
   } },

   /* replace2040_0 -> 67 in the cache */
   /* replace2040_1 -> 217 in the cache */
   /* replace2040_2 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 67, 217, 67 },
      -1,
   } },

   /* ('~ffma@32', 'a', 2.0, ('fmul', ('fneg', 'a'), 'a')) => ('flrp', 'a', 1.0, 'a') */
   /* search2041_0 -> 67 in the cache */
   /* search2041_1 -> 214 in the cache */
   /* search2041_2_0_0 -> 67 in the cache */
   /* search2041_2_0 -> 221 in the cache */
   /* search2041_2_1 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 221, 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 67, 214, 243 },
      -1,
   } },

   /* replace2041_0 -> 67 in the cache */
   /* replace2041_1 -> 217 in the cache */
   /* replace2041_2 -> 67 in the cache */
   /* replace2041 -> 242 in the cache */

   /* ('~ffma@32', 'a', ('fneg', 'a'), ('fmul', 2.0, 'a')) => ('flrp', 'a', 1.0, 'a') */
   /* search2042_0 -> 67 in the cache */
   /* search2042_1_0 -> 67 in the cache */
   /* search2042_1 -> 221 in the cache */
   /* search2042_2_0 -> 214 in the cache */
   /* search2042_2_1 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 1,
      { 214, 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 67, 221, 245 },
      -1,
   } },

   /* replace2042_0 -> 67 in the cache */
   /* replace2042_1 -> 217 in the cache */
   /* replace2042_2 -> 67 in the cache */
   /* replace2042 -> 242 in the cache */

   /* ('~fmul@32', 'a', ('fadd', 2.0, ('fneg', 'a'))) => ('flrp', 'a', 1.0, 'a') */
   /* search2043_0 -> 67 in the cache */
   /* search2043_1_0 -> 214 in the cache */
   /* search2043_1_1_0 -> 67 in the cache */
   /* search2043_1_1 -> 221 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 214, 221 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fmul,
      0, 2,
      { 67, 247 },
      -1,
   } },

   /* replace2043_0 -> 67 in the cache */
   /* replace2043_1 -> 217 in the cache */
   /* replace2043_2 -> 67 in the cache */
   /* replace2043 -> 242 in the cache */

   /* ('fmin', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmin', 'a', 'b')) */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 249, 250 },
      0,
   } },
   /* search2044_1_0 -> 249 in the cache */
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 249, 252 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 3,
      { 251, 253 },
      -1,
   } },

   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* replace2044_1_0 -> 250 in the cache */
   /* replace2044_1_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmin,
      1, 1,
      { 250, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 255, 256 },
      -1,
   } },

   /* ('fmax', ('fadd(is_used_once)', '#c', 'a'), ('fadd(is_used_once)', '#c', 'b')) => ('fadd', 'c', ('fmax', 'a', 'b')) */
   /* search2045_0_0 -> 249 in the cache */
   /* search2045_0_1 -> 250 in the cache */
   /* search2045_0 -> 251 in the cache */
   /* search2045_1_0 -> 249 in the cache */
   /* search2045_1_1 -> 252 in the cache */
   /* search2045_1 -> 253 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 3,
      { 251, 253 },
      -1,
   } },

   /* replace2045_0 -> 255 in the cache */
   /* replace2045_1_0 -> 250 in the cache */
   /* replace2045_1_1 -> 252 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fmax,
      1, 1,
      { 250, 252 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 2,
      { 255, 259 },
      -1,
   } },

   /* ('bcsel', ('feq', ('fsqrt', 'a(is_not_negative)'), 0.0), 3.4028234663852886e+38, ('frsq', 'a')) => ('fmin', ('frsq', 'a'), 3.4028234663852886e+38) */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      4,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 261 },
      -1,
   } },
   /* search2046_0_1 -> 43 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 262, 43 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x47efffffe0000000 /* 3.4028234663852886e+38 */ },
   } },
   /* search2046_2_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 42 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 1,
      { 263, 264, 265 },
      -1,
   } },

   /* replace2046_0_0 -> 42 in the cache */
   /* replace2046_0 -> 265 in the cache */
   /* replace2046_1 -> 264 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 265, 264 },
      -1,
   } },

   /* ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', ('ffma', 'e', 'f', 'g'))) */
   { .variable = {
      { nir_search_value_variable, -7 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      4, /* e */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      5, /* f */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fmul,
      3, 1,
      { 272, 273 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 2,
      { 270, 271, 274 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 3,
      { 268, 269, 275 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      6, /* g */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 276, 277 },
      -1,
   } },

   /* replace2047_0 -> 268 in the cache */
   /* replace2047_1 -> 269 in the cache */
   /* replace2047_2_0 -> 270 in the cache */
   /* replace2047_2_1 -> 271 in the cache */
   { .variable = {
      { nir_search_value_variable, -7 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      5, /* f */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -7 },
      6, /* g */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 279, 280, 281 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 270, 271, 282 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 268, 269, 283 },
      -1,
   } },

   /* ('~fadd', ('ffma(is_used_once)', 'a', 'b', ('fmul(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffma', 'a', 'b', ('ffma', 'c', 'd', 'e')) */
   { .variable = {
      { nir_search_value_variable, -5 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 287, 288 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 285, 286, 289 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 290, 291 },
      -1,
   } },

   /* replace2048_0 -> 285 in the cache */
   /* replace2048_1 -> 286 in the cache */
   { .variable = {
      { nir_search_value_variable, -5 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, -5 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 293, 294, 295 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 2,
      { 285, 286, 296 },
      -1,
   } },

   /* ('~fadd', ('fneg', ('ffma(is_used_once)', 'a', 'b', ('ffma', 'c', 'd', ('fmul(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffma', ('fneg', 'a'), 'b', ('ffma', ('fneg', 'c'), 'd', ('ffma', ('fneg', 'e'), 'f', 'g'))) */
   /* search2049_0_0_0 -> 268 in the cache */
   /* search2049_0_0_1 -> 269 in the cache */
   /* search2049_0_0_2_0 -> 270 in the cache */
   /* search2049_0_0_2_1 -> 271 in the cache */
   /* search2049_0_0_2_2_0 -> 272 in the cache */
   /* search2049_0_0_2_2_1 -> 273 in the cache */
   /* search2049_0_0_2_2 -> 274 in the cache */
   /* search2049_0_0_2 -> 275 in the cache */
   /* search2049_0_0 -> 276 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 276 },
      -1,
   } },
   /* search2049_1 -> 277 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 298, 277 },
      -1,
   } },

   /* replace2049_0_0 -> 268 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 268 },
      -1,
   } },
   /* replace2049_1 -> 269 in the cache */
   /* replace2049_2_0_0 -> 270 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 270 },
      -1,
   } },
   /* replace2049_2_1 -> 271 in the cache */
   /* replace2049_2_2_0_0 -> 279 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 279 },
      -1,
   } },
   /* replace2049_2_2_1 -> 280 in the cache */
   /* replace2049_2_2_2 -> 281 in the cache */
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      2, 1,
      { 302, 280, 281 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 2,
      { 301, 271, 303 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -7 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 3,
      { 300, 269, 304 },
      -1,
   } },

   /* ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)'))), 'g(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', ('ffmaz', 'e', 'f', 'g'))) */
   /* search2050_0_0 -> 67 in the cache */
   /* search2050_0_1 -> 68 in the cache */
   /* search2050_0_2_0 -> 70 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      5, /* f */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      3, 1,
      { 307, 308 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 2,
      { 70, 306, 309 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 3,
      { 67, 68, 310 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      6, /* g */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 311, 312 },
      -1,
   } },

   /* replace2050_0 -> 67 in the cache */
   /* replace2050_1 -> 68 in the cache */
   /* replace2050_2_0 -> 70 in the cache */
   /* replace2050_2_1 -> 306 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      5, /* f */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      6, /* g */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 314, 315, 316 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 70, 306, 317 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 67, 68, 318 },
      -1,
   } },

   /* ('~fadd', ('ffmaz(is_used_once)', 'a', 'b', ('fmulz(is_used_once)', 'c(is_not_const_and_not_fsign)', 'd(is_not_const_and_not_fsign)')), 'e(is_not_const)') => ('ffmaz', 'a', 'b', ('ffmaz', 'c', 'd', 'e')) */
   /* search2051_0_0 -> 67 in the cache */
   /* search2051_0_1 -> 68 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      3, /* d */
      false,
      nir_type_invalid,
      5,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmulz,
      2, 1,
      { 320, 321 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 67, 68, 322 },
      0,
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      4, /* e */
      false,
      nir_type_invalid,
      1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 323, 324 },
      -1,
   } },

   /* replace2051_0 -> 67 in the cache */
   /* replace2051_1 -> 68 in the cache */
   /* replace2051_2_0 -> 70 in the cache */
   /* replace2051_2_1 -> 306 in the cache */
   /* replace2051_2_2 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 70, 306, 314 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 2,
      { 67, 68, 326 },
      -1,
   } },

   /* ('~fadd', ('fneg', ('ffmaz(is_used_once)', 'a', 'b', ('ffmaz', 'c', 'd', ('fmulz(is_used_once)', 'e(is_not_const_and_not_fsign)', 'f(is_not_const_and_not_fsign)')))), 'g(is_not_const)') => ('ffmaz', ('fneg', 'a'), 'b', ('ffmaz', ('fneg', 'c'), 'd', ('ffmaz', ('fneg', 'e'), 'f', 'g'))) */
   /* search2052_0_0_0 -> 67 in the cache */
   /* search2052_0_0_1 -> 68 in the cache */
   /* search2052_0_0_2_0 -> 70 in the cache */
   /* search2052_0_0_2_1 -> 306 in the cache */
   /* search2052_0_0_2_2_0 -> 307 in the cache */
   /* search2052_0_0_2_2_1 -> 308 in the cache */
   /* search2052_0_0_2_2 -> 309 in the cache */
   /* search2052_0_0_2 -> 310 in the cache */
   /* search2052_0_0 -> 311 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 3,
      { 311 },
      -1,
   } },
   /* search2052_1 -> 312 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 4,
      { 328, 312 },
      -1,
   } },

   /* replace2052_0_0 -> 67 in the cache */
   /* replace2052_0 -> 221 in the cache */
   /* replace2052_1 -> 68 in the cache */
   /* replace2052_2_0_0 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 70 },
      -1,
   } },
   /* replace2052_2_1 -> 306 in the cache */
   /* replace2052_2_2_0_0 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 314 },
      -1,
   } },
   /* replace2052_2_2_1 -> 315 in the cache */
   /* replace2052_2_2_2 -> 316 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      2, 1,
      { 331, 315, 316 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 2,
      { 330, 306, 332 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 3,
      { 221, 68, 333 },
      -1,
   } },

   /* ('ubfe', 'a', 'b', 0) => 0 */
   /* search2053_0 -> 67 in the cache */
   /* search2053_1 -> 68 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 67, 68, 335 },
      -1,
   } },

   /* replace2053 -> 335 in the cache */

   /* ('ibfe', 'a', 'b', 0) => 0 */
   /* search2054_0 -> 67 in the cache */
   /* search2054_1 -> 68 in the cache */
   /* search2054_2 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 67, 68, 335 },
      -1,
   } },

   /* replace2054 -> 335 in the cache */

   /* ('ubfe', 'a', '#b', '#c') => ('iand', ('ushr', 4294967295, ('ineg', 'c')), ('ushr', 'a', 'b')) */
   /* search2055_0 -> 67 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      1, /* b */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .variable = {
      { nir_search_value_variable, 32 },
      2, /* c */
      true,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ubfe,
      -1, 0,
      { 67, 338, 339 },
      -1,
   } },

   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xffffffff /* 4294967295 */ },
   } },
   /* replace2055_0_1_0 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 0,
      { 70 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 341, 342 },
      -1,
   } },
   /* replace2055_1_0 -> 67 in the cache */
   /* replace2055_1_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 67, 68 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 343, 344 },
      -1,
   } },

   /* ('ibfe', 'a', '#b', '#c') => ('ishr', ('ishl', 'a', ('ineg', ('iadd', 'b', 'c'))), ('ineg', 'c')) */
   /* search2056_0 -> 67 in the cache */
   /* search2056_1 -> 338 in the cache */
   /* search2056_2 -> 339 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ibfe,
      -1, 0,
      { 67, 338, 339 },
      -1,
   } },

   /* replace2056_0_0 -> 67 in the cache */
   /* replace2056_0_1_0_0 -> 68 in the cache */
   /* replace2056_0_1_0_1 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iadd,
      0, 1,
      { 68, 70 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ineg,
      -1, 1,
      { 347 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 1,
      { 67, 348 },
      -1,
   } },
   /* replace2056_1_0 -> 70 in the cache */
   /* replace2056_1 -> 342 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 1,
      { 349, 342 },
      -1,
   } },

   /* ('ishl', 'a', 0) => a */
   /* search2057_0 -> 42 in the cache */
   /* search2057_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 42, 335 },
      -1,
   } },

   /* replace2057 -> 42 in the cache */

   /* ('ishl', 'a', -32) => a */
   /* search2058_0 -> 42 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { -0x20 /* -32 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 42, 352 },
      -1,
   } },

   /* replace2058 -> 42 in the cache */

   /* ('ishr', 'a', 0) => a */
   /* search2059_0 -> 42 in the cache */
   /* search2059_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 335 },
      -1,
   } },

   /* replace2059 -> 42 in the cache */

   /* ('ishr', 'a', -32) => a */
   /* search2060_0 -> 42 in the cache */
   /* search2060_1 -> 352 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ishr,
      -1, 0,
      { 42, 352 },
      -1,
   } },

   /* replace2060 -> 42 in the cache */

   /* ('ushr', 'a', 0) => a */
   /* search2061_0 -> 42 in the cache */
   /* search2061_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_ushr,
      -1, 0,
      { 42, 335 },
      -1,
   } },

   /* replace2061 -> 42 in the cache */

   /* ('extract_i8', ('extract_i8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b') */
   /* search2062_0_0 -> 0 in the cache */
   /* search2062_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, -2 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 357, 358 },
      -1,
   } },

   /* replace2062_0 -> 0 in the cache */
   /* replace2062_1 -> 1 in the cache */
   /* replace2062 -> 357 in the cache */

   /* ('extract_i8', ('extract_u8', 'a', 'b'), 0) => ('extract_i8', 'a', 'b') */
   /* search2063_0_0 -> 0 in the cache */
   /* search2063_0_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },
   /* search2063_1 -> 358 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 360, 358 },
      -1,
   } },

   /* replace2063_0 -> 0 in the cache */
   /* replace2063_1 -> 1 in the cache */
   /* replace2063 -> 357 in the cache */

   /* ('extract_u8', ('extract_i8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search2064_0_0 -> 0 in the cache */
   /* search2064_0_1 -> 1 in the cache */
   /* search2064_0 -> 357 in the cache */
   /* search2064_1 -> 358 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 357, 358 },
      -1,
   } },

   /* replace2064_0 -> 0 in the cache */
   /* replace2064_1 -> 1 in the cache */
   /* replace2064 -> 360 in the cache */

   /* ('extract_u8', ('extract_u8', 'a', 'b'), 0) => ('extract_u8', 'a', 'b') */
   /* search2065_0_0 -> 0 in the cache */
   /* search2065_0_1 -> 1 in the cache */
   /* search2065_0 -> 360 in the cache */
   /* search2065_1 -> 358 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 360, 358 },
      -1,
   } },

   /* replace2065_0 -> 0 in the cache */
   /* replace2065_1 -> 1 in the cache */
   /* replace2065 -> 360 in the cache */

   /* ('u2u16', ('u2u8', 'a@16')) => ('extract_u8', 'a', 0) */
   /* search2066_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 61 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 364 },
      -1,
   } },

   /* replace2066_0 -> 61 in the cache */
   { .constant = {
      { nir_search_value_constant, 16 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 61, 366 },
      -1,
   } },

   /* ('u2u16', ('i2i8', 'a@16')) => ('extract_u8', 'a', 0) */
   /* search2067_0_0 -> 61 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 61 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 368 },
      -1,
   } },

   /* replace2067_0 -> 61 in the cache */
   /* replace2067_1 -> 366 in the cache */
   /* replace2067 -> 367 in the cache */

   /* ('i2i16', ('u2u8', 'a@16')) => ('extract_i8', 'a', 0) */
   /* search2068_0_0 -> 61 in the cache */
   /* search2068_0 -> 364 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 364 },
      -1,
   } },

   /* replace2068_0 -> 61 in the cache */
   /* replace2068_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 61, 366 },
      -1,
   } },

   /* ('i2i16', ('i2i8', 'a@16')) => ('extract_i8', 'a', 0) */
   /* search2069_0_0 -> 61 in the cache */
   /* search2069_0 -> 368 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 368 },
      -1,
   } },

   /* replace2069_0 -> 61 in the cache */
   /* replace2069_1 -> 366 in the cache */
   /* replace2069 -> 371 in the cache */

   /* ('u2u32', ('u2u8', 'a@32')) => ('extract_u8', 'a', 0) */
   /* search2070_0_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_u2u8,
      -1, 0,
      { 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 373 },
      -1,
   } },

   /* replace2070_0 -> 67 in the cache */
   /* replace2070_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 67, 335 },
      -1,
   } },

   /* ('u2u32', ('i2i8', 'a@32')) => ('extract_u8', 'a', 0) */
   /* search2071_0_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_i2i8,
      -1, 0,
      { 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 376 },
      -1,
   } },

   /* replace2071_0 -> 67 in the cache */
   /* replace2071_1 -> 335 in the cache */
   /* replace2071 -> 375 in the cache */

   /* ('u2u32', ('u2u16', 'a@32')) => ('extract_u16', 'a', 0) */
   /* search2072_0_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2u16,
      -1, 0,
      { 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 378 },
      -1,
   } },

   /* replace2072_0 -> 67 in the cache */
   /* replace2072_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 67, 335 },
      -1,
   } },

   /* ('u2u32', ('i2i16', 'a@32')) => ('extract_u16', 'a', 0) */
   /* search2073_0_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2i16,
      -1, 0,
      { 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_u2u32,
      -1, 0,
      { 381 },
      -1,
   } },

   /* replace2073_0 -> 67 in the cache */
   /* replace2073_1 -> 335 in the cache */
   /* replace2073 -> 380 in the cache */

   /* ('i2i32', ('u2u8', 'a@32')) => ('extract_i8', 'a', 0) */
   /* search2074_0_0 -> 67 in the cache */
   /* search2074_0 -> 373 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 373 },
      -1,
   } },

   /* replace2074_0 -> 67 in the cache */
   /* replace2074_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i8,
      -1, 0,
      { 67, 335 },
      -1,
   } },

   /* ('i2i32', ('i2i8', 'a@32')) => ('extract_i8', 'a', 0) */
   /* search2075_0_0 -> 67 in the cache */
   /* search2075_0 -> 376 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 376 },
      -1,
   } },

   /* replace2075_0 -> 67 in the cache */
   /* replace2075_1 -> 335 in the cache */
   /* replace2075 -> 384 in the cache */

   /* ('i2i32', ('u2u16', 'a@32')) => ('extract_i16', 'a', 0) */
   /* search2076_0_0 -> 67 in the cache */
   /* search2076_0 -> 378 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 378 },
      -1,
   } },

   /* replace2076_0 -> 67 in the cache */
   /* replace2076_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_extract_i16,
      -1, 0,
      { 67, 335 },
      -1,
   } },

   /* ('i2i32', ('i2i16', 'a@32')) => ('extract_i16', 'a', 0) */
   /* search2077_0_0 -> 67 in the cache */
   /* search2077_0 -> 381 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_i2i32,
      -1, 0,
      { 381 },
      -1,
   } },

   /* replace2077_0 -> 67 in the cache */
   /* replace2077_1 -> 335 in the cache */
   /* replace2077 -> 387 in the cache */

   /* ('ishl', ('extract_u8', 'a@32', 0), 8) => ('insert_u8', 'a', 1) */
   /* search2078_0_0 -> 67 in the cache */
   /* search2078_0_1 -> 335 in the cache */
   /* search2078_0 -> 375 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x8 /* 8 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 375, 389 },
      -1,
   } },

   /* replace2078_0 -> 67 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x1 /* 1 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 67, 391 },
      -1,
   } },

   /* ('ishl', ('extract_u8', 'a@32', 0), 16) => ('insert_u8', 'a', 2) */
   /* search2079_0_0 -> 67 in the cache */
   /* search2079_0_1 -> 335 in the cache */
   /* search2079_0 -> 375 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x10 /* 16 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 375, 393 },
      -1,
   } },

   /* replace2079_0 -> 67 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x2 /* 2 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 67, 395 },
      -1,
   } },

   /* ('ishl', ('extract_u8', 'a@32', 0), 24) => ('insert_u8', 'a', 3) */
   /* search2080_0_0 -> 67 in the cache */
   /* search2080_0_1 -> 335 in the cache */
   /* search2080_0 -> 375 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x18 /* 24 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 375, 397 },
      -1,
   } },

   /* replace2080_0 -> 67 in the cache */
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0x3 /* 3 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 67, 399 },
      -1,
   } },

   /* ('iand', ('ishl', 'a@32', 8), 65280) => ('insert_u8', 'a', 1) */
   /* search2081_0_0 -> 67 in the cache */
   /* search2081_0_1 -> 389 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 67, 389 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff00 /* 65280 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 401, 402 },
      -1,
   } },

   /* replace2081_0 -> 67 in the cache */
   /* replace2081_1 -> 391 in the cache */
   /* replace2081 -> 392 in the cache */

   /* ('iand', ('ishl', 'a@32', 16), 16711680) => ('insert_u8', 'a', 2) */
   /* search2082_0_0 -> 67 in the cache */
   /* search2082_0_1 -> 393 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 67, 393 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff0000 /* 16711680 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 404, 405 },
      -1,
   } },

   /* replace2082_0 -> 67 in the cache */
   /* replace2082_1 -> 395 in the cache */
   /* replace2082 -> 396 in the cache */

   /* ('iand', ('ishl', 'a@32', 24), 4278190080) => ('insert_u8', 'a', 3) */
   /* search2083_0_0 -> 67 in the cache */
   /* search2083_0_1 -> 397 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ishl,
      -1, 0,
      { 67, 397 },
      -1,
   } },
   { .constant = {
      { nir_search_value_constant, 32 },
      nir_type_int, { 0xff000000 /* 4278190080 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_iand,
      0, 1,
      { 407, 408 },
      -1,
   } },

   /* replace2083_0 -> 67 in the cache */
   /* replace2083_1 -> 399 in the cache */
   /* replace2083 -> 400 in the cache */

   /* ('ishl', 'a@32', 24) => ('insert_u8', 'a', 3) */
   /* search2084_0 -> 67 in the cache */
   /* search2084_1 -> 397 in the cache */
   /* search2084 -> 407 in the cache */

   /* replace2084_0 -> 67 in the cache */
   /* replace2084_1 -> 399 in the cache */
   /* replace2084 -> 400 in the cache */

   /* ('ishl', 'a@32', 16) => ('insert_u16', 'a', 1) */
   /* search2085_0 -> 67 in the cache */
   /* search2085_1 -> 393 in the cache */
   /* search2085 -> 404 in the cache */

   /* replace2085_0 -> 67 in the cache */
   /* replace2085_1 -> 391 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 67, 391 },
      -1,
   } },

   /* ('insert_u8', ('extract_u8', 'a', 0), 'b') => ('insert_u8', 'a', 'b') */
   /* search2086_0_0 -> 0 in the cache */
   /* search2086_0_1 -> 358 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u8,
      -1, 0,
      { 0, 358 },
      -1,
   } },
   /* search2086_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 411, 1 },
      -1,
   } },

   /* replace2086_0 -> 0 in the cache */
   /* replace2086_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u8,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('insert_u16', ('extract_u16', 'a', 0), 'b') => ('insert_u16', 'a', 'b') */
   /* search2087_0_0 -> 0 in the cache */
   /* search2087_0_1 -> 358 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_extract_u16,
      -1, 0,
      { 0, 358 },
      -1,
   } },
   /* search2087_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 414, 1 },
      -1,
   } },

   /* replace2087_0 -> 0 in the cache */
   /* replace2087_1 -> 1 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_insert_u16,
      -1, 0,
      { 0, 1 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@8', 0), ('ine', 'b@8', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2088_0_0 -> 82 in the cache */
   { .constant = {
      { nir_search_value_constant, 8 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 82, 417 },
      0,
   } },
   /* search2088_1_0 -> 83 in the cache */
   /* search2088_1_1 -> 417 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 83, 417 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 418, 419 },
      -1,
   } },

   /* replace2088_0_0 -> 82 in the cache */
   /* replace2088_0_1 -> 83 in the cache */
   { .expression = {
      { nir_search_value_expression, 8 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 82, 83 },
      -1,
   } },
   /* replace2088_1 -> 417 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 421, 417 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@8', 0), ('ieq', 'b@8', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2089_0_0 -> 82 in the cache */
   /* search2089_0_1 -> 417 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 82, 417 },
      0,
   } },
   /* search2089_1_0 -> 83 in the cache */
   /* search2089_1_1 -> 417 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 83, 417 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 423, 424 },
      -1,
   } },

   /* replace2089_0_0 -> 82 in the cache */
   /* replace2089_0_1 -> 83 in the cache */
   /* replace2089_0 -> 421 in the cache */
   /* replace2089_1 -> 417 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 421, 417 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@16', 0), ('ine', 'b@16', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2090_0_0 -> 61 in the cache */
   /* search2090_0_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 61, 366 },
      0,
   } },
   /* search2090_1_0 -> 62 in the cache */
   /* search2090_1_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 62, 366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 427, 428 },
      -1,
   } },

   /* replace2090_0_0 -> 61 in the cache */
   /* replace2090_0_1 -> 62 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 61, 62 },
      -1,
   } },
   /* replace2090_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 430, 366 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@16', 0), ('ieq', 'b@16', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2091_0_0 -> 61 in the cache */
   /* search2091_0_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 61, 366 },
      0,
   } },
   /* search2091_1_0 -> 62 in the cache */
   /* search2091_1_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 62, 366 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 432, 433 },
      -1,
   } },

   /* replace2091_0_0 -> 61 in the cache */
   /* replace2091_0_1 -> 62 in the cache */
   /* replace2091_0 -> 430 in the cache */
   /* replace2091_1 -> 366 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 430, 366 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@32', 0), ('ine', 'b@32', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2092_0_0 -> 67 in the cache */
   /* search2092_0_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 67, 335 },
      0,
   } },
   /* search2092_1_0 -> 68 in the cache */
   /* search2092_1_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 68, 335 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 436, 437 },
      -1,
   } },

   /* replace2092_0_0 -> 67 in the cache */
   /* replace2092_0_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 67, 68 },
      -1,
   } },
   /* replace2092_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 439, 335 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@32', 0), ('ieq', 'b@32', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2093_0_0 -> 67 in the cache */
   /* search2093_0_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 67, 335 },
      0,
   } },
   /* search2093_1_0 -> 68 in the cache */
   /* search2093_1_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 68, 335 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 441, 442 },
      -1,
   } },

   /* replace2093_0_0 -> 67 in the cache */
   /* replace2093_0_1 -> 68 in the cache */
   /* replace2093_0 -> 439 in the cache */
   /* replace2093_1 -> 335 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 439, 335 },
      -1,
   } },

   /* ('iand', ('ine(is_used_once)', 'a@64', 0), ('ine', 'b@64', 0)) => ('ine', ('umin', 'a', 'b'), 0) */
   /* search2094_0_0 -> 73 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_int, { 0x0 /* 0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      1, 1,
      { 73, 445 },
      0,
   } },
   /* search2094_1_0 -> 74 in the cache */
   /* search2094_1_1 -> 445 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      2, 1,
      { 74, 445 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_iand,
      0, 3,
      { 446, 447 },
      -1,
   } },

   /* replace2094_0_0 -> 73 in the cache */
   /* replace2094_0_1 -> 74 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_umin,
      1, 1,
      { 73, 74 },
      -1,
   } },
   /* replace2094_1 -> 445 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ine,
      0, 2,
      { 449, 445 },
      -1,
   } },

   /* ('ior', ('ieq(is_used_once)', 'a@64', 0), ('ieq', 'b@64', 0)) => ('ieq', ('umin', 'a', 'b'), 0) */
   /* search2095_0_0 -> 73 in the cache */
   /* search2095_0_1 -> 445 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      1, 1,
      { 73, 445 },
      0,
   } },
   /* search2095_1_0 -> 74 in the cache */
   /* search2095_1_1 -> 445 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      2, 1,
      { 74, 445 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 3,
      { 451, 452 },
      -1,
   } },

   /* replace2095_0_0 -> 73 in the cache */
   /* replace2095_0_1 -> 74 in the cache */
   /* replace2095_0 -> 449 in the cache */
   /* replace2095_1 -> 445 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ieq,
      0, 2,
      { 449, 445 },
      -1,
   } },

   /* ('~fadd@16', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   /* search2096_0 -> 126 in the cache */
   { .variable = {
      { nir_search_value_variable, 16 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2096_1_1_0 -> 62 in the cache */
   /* search2096_1_1_1 -> 125 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 62, 125 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 455, 456 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 126, 457 },
      -1,
   } },

   /* replace2096_0_0 -> 126 in the cache */
   /* replace2096_0_1_0 -> 455 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 455 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 126, 459 },
      -1,
   } },
   /* replace2096_1_0 -> 62 in the cache */
   /* replace2096_1_1 -> 455 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 62, 455 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 460, 461 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f16', ('inot', 'b@bool'))) => ('b2f16', ('inot', ('ior', 'a', 'b'))) */
   { .variable = {
      { nir_search_value_variable, 1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2097_1 -> 366 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_bool,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 0,
      { 464 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 0,
      { 465 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 366, 466 },
      -1,
   } },

   /* replace2097_0_0_0 -> 463 in the cache */
   { .variable = {
      { nir_search_value_variable, 1 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_ior,
      0, 1,
      { 463, 468 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_inot,
      -1, 1,
      { 469 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_b2f16,
      -1, 1,
      { 470 },
      -1,
   } },

   /* ('~fadd@32', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   /* search2098_0 -> 217 in the cache */
   { .variable = {
      { nir_search_value_variable, 32 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2098_1_1_0 -> 68 in the cache */
   /* search2098_1_1_1 -> 215 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 68, 215 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 472, 473 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 217, 474 },
      -1,
   } },

   /* replace2098_0_0 -> 217 in the cache */
   /* replace2098_0_1_0 -> 472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 472 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 217, 476 },
      -1,
   } },
   /* replace2098_1_0 -> 68 in the cache */
   /* replace2098_1_1 -> 472 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 68, 472 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 477, 478 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f32', ('inot', 'b@bool'))) => ('b2f32', ('inot', ('ior', 'a', 'b'))) */
   /* search2099_0 -> 463 in the cache */
   /* search2099_1 -> 335 in the cache */
   /* search2099_2_0_0 -> 464 in the cache */
   /* search2099_2_0 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 0,
      { 465 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 335, 480 },
      -1,
   } },

   /* replace2099_0_0_0 -> 463 in the cache */
   /* replace2099_0_0_1 -> 468 in the cache */
   /* replace2099_0_0 -> 469 in the cache */
   /* replace2099_0 -> 470 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_b2f32,
      -1, 1,
      { 470 },
      -1,
   } },

   /* ('~fadd@64', 1.0, ('fmul(is_used_once)', 'c', ('fadd', 'b', -1.0))) => ('fadd', ('fadd', 1.0, ('fneg', 'c')), ('fmul', 'b', 'c')) */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0x3ff0000000000000 /* 1.0 */ },
   } },
   { .variable = {
      { nir_search_value_variable, 64 },
      0, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   /* search2100_1_1_0 -> 74 in the cache */
   { .constant = {
      { nir_search_value_constant, 64 },
      nir_type_float, { 0xbff0000000000000 /* -1.0 */ },
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      2, 1,
      { 74, 485 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      1, 2,
      { 484, 486 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 483, 487 },
      -1,
   } },

   /* replace2100_0_0 -> 483 in the cache */
   /* replace2100_0_1_0 -> 484 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 483, 489 },
      -1,
   } },
   /* replace2100_1_0 -> 74 in the cache */
   /* replace2100_1_1 -> 484 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fmul,
      2, 1,
      { 74, 484 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 3,
      { 490, 491 },
      -1,
   } },

   /* ('bcsel', 'a', 0, ('b2f64', ('inot', 'b@bool'))) => ('b2f64', ('inot', ('ior', 'a', 'b'))) */
   /* search2101_0 -> 463 in the cache */
   /* search2101_1 -> 445 in the cache */
   /* search2101_2_0_0 -> 464 in the cache */
   /* search2101_2_0 -> 465 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 0,
      { 465 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 445, 493 },
      -1,
   } },

   /* replace2101_0_0_0 -> 463 in the cache */
   /* replace2101_0_0_1 -> 468 in the cache */
   /* replace2101_0_0 -> 469 in the cache */
   /* replace2101_0 -> 470 in the cache */
   { .expression = {
      { nir_search_value_expression, 64 },
      false,
      false,
      false,
      nir_op_b2f64,
      -1, 1,
      { 470 },
      -1,
   } },

   /* ('bcsel', 'a', ('fadd(is_used_once)', 'b', 'c'), ('fadd', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2102_0 -> 463 in the cache */
   /* search2102_1_0 -> 151 in the cache */
   /* search2102_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 151, 153 },
      0,
   } },
   /* search2102_2_0 -> 151 in the cache */
   /* search2102_2_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 151, 154 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 496, 497 },
      -1,
   } },

   /* replace2102_0 -> 151 in the cache */
   /* replace2102_1_0 -> 463 in the cache */
   /* replace2102_1_1 -> 153 in the cache */
   /* replace2102_1_2 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 153, 154 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 151, 499 },
      -1,
   } },

   /* ('bcsel', 'a', ('fadd', 'b', 'c'), ('fadd(is_used_once)', 'b', 'd')) => ('fadd', 'b', ('bcsel', 'a', 'c', 'd')) */
   /* search2103_0 -> 463 in the cache */
   /* search2103_1_0 -> 151 in the cache */
   /* search2103_1_1 -> 153 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 151, 153 },
      -1,
   } },
   /* search2103_2_0 -> 151 in the cache */
   /* search2103_2_1 -> 154 in the cache */
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_fadd,
      1, 1,
      { 151, 154 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -4 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 501, 502 },
      -1,
   } },

   /* replace2103_0 -> 151 in the cache */
   /* replace2103_1_0 -> 463 in the cache */
   /* replace2103_1_1 -> 153 in the cache */
   /* replace2103_1_2 -> 154 in the cache */
   /* replace2103_1 -> 499 in the cache */
   /* replace2103 -> 500 in the cache */

   /* ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2104_0 -> 463 in the cache */
   /* search2104_1_0 -> 286 in the cache */
   /* search2104_1_1 -> 293 in the cache */
   /* search2104_1_2 -> 294 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 286, 293, 294 },
      0,
   } },
   /* search2104_2_0 -> 286 in the cache */
   /* search2104_2_1 -> 293 in the cache */
   /* search2104_2_2 -> 295 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 286, 293, 295 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 504, 505 },
      -1,
   } },

   /* replace2104_0 -> 286 in the cache */
   /* replace2104_1 -> 293 in the cache */
   /* replace2104_2_0 -> 463 in the cache */
   /* replace2104_2_1 -> 294 in the cache */
   /* replace2104_2_2 -> 295 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 294, 295 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 286, 293, 507 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'c', 'e')) => ('ffma', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2105_0 -> 463 in the cache */
   /* search2105_1_0 -> 286 in the cache */
   /* search2105_1_1 -> 293 in the cache */
   /* search2105_1_2 -> 294 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 286, 293, 294 },
      -1,
   } },
   /* search2105_2_0 -> 286 in the cache */
   /* search2105_2_1 -> 293 in the cache */
   /* search2105_2_2 -> 295 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 286, 293, 295 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 509, 510 },
      -1,
   } },

   /* replace2105_0 -> 286 in the cache */
   /* replace2105_1 -> 293 in the cache */
   /* replace2105_2_0 -> 463 in the cache */
   /* replace2105_2_1 -> 294 in the cache */
   /* replace2105_2_2 -> 295 in the cache */
   /* replace2105_2 -> 507 in the cache */
   /* replace2105 -> 508 in the cache */

   /* ('bcsel', 'a', ('ffma(is_used_once)', 'b', 'c', 'd'), ('ffma', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2106_0 -> 463 in the cache */
   /* search2106_1_0 -> 286 in the cache */
   /* search2106_1_1 -> 293 in the cache */
   /* search2106_1_2 -> 294 in the cache */
   /* search2106_1 -> 504 in the cache */
   /* search2106_2_0 -> 286 in the cache */
   /* search2106_2_1 -> 295 in the cache */
   /* search2106_2_2 -> 294 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 286, 295, 294 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 504, 512 },
      -1,
   } },

   /* replace2106_0 -> 286 in the cache */
   /* replace2106_1_0 -> 463 in the cache */
   /* replace2106_1_1 -> 293 in the cache */
   /* replace2106_1_2 -> 295 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 293, 295 },
      -1,
   } },
   /* replace2106_2 -> 294 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 286, 514, 294 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffma', 'b', 'c', 'd'), ('ffma(is_used_once)', 'b', 'e', 'd')) => ('ffma', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2107_0 -> 463 in the cache */
   /* search2107_1_0 -> 286 in the cache */
   /* search2107_1_1 -> 293 in the cache */
   /* search2107_1_2 -> 294 in the cache */
   /* search2107_1 -> 509 in the cache */
   /* search2107_2_0 -> 286 in the cache */
   /* search2107_2_1 -> 295 in the cache */
   /* search2107_2_2 -> 294 in the cache */
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_ffma,
      1, 1,
      { 286, 295, 294 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -5 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 509, 516 },
      -1,
   } },

   /* replace2107_0 -> 286 in the cache */
   /* replace2107_1_0 -> 463 in the cache */
   /* replace2107_1_1 -> 293 in the cache */
   /* replace2107_1_2 -> 295 in the cache */
   /* replace2107_1 -> 514 in the cache */
   /* replace2107_2 -> 294 in the cache */
   /* replace2107 -> 515 in the cache */

   /* ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2108_0 -> 463 in the cache */
   /* search2108_1_0 -> 68 in the cache */
   /* search2108_1_1 -> 70 in the cache */
   /* search2108_1_2 -> 306 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 68, 70, 306 },
      0,
   } },
   /* search2108_2_0 -> 68 in the cache */
   /* search2108_2_1 -> 70 in the cache */
   /* search2108_2_2 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 68, 70, 314 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 518, 519 },
      -1,
   } },

   /* replace2108_0 -> 68 in the cache */
   /* replace2108_1 -> 70 in the cache */
   /* replace2108_2_0 -> 463 in the cache */
   /* replace2108_2_1 -> 306 in the cache */
   /* replace2108_2_2 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 306, 314 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 68, 70, 521 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'c', 'e')) => ('ffmaz', 'b', 'c', ('bcsel', 'a', 'd', 'e')) */
   /* search2109_0 -> 463 in the cache */
   /* search2109_1_0 -> 68 in the cache */
   /* search2109_1_1 -> 70 in the cache */
   /* search2109_1_2 -> 306 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 68, 70, 306 },
      -1,
   } },
   /* search2109_2_0 -> 68 in the cache */
   /* search2109_2_1 -> 70 in the cache */
   /* search2109_2_2 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 68, 70, 314 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 523, 524 },
      -1,
   } },

   /* replace2109_0 -> 68 in the cache */
   /* replace2109_1 -> 70 in the cache */
   /* replace2109_2_0 -> 463 in the cache */
   /* replace2109_2_1 -> 306 in the cache */
   /* replace2109_2_2 -> 314 in the cache */
   /* replace2109_2 -> 521 in the cache */
   /* replace2109 -> 522 in the cache */

   /* ('bcsel', 'a', ('ffmaz(is_used_once)', 'b', 'c', 'd'), ('ffmaz', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2110_0 -> 463 in the cache */
   /* search2110_1_0 -> 68 in the cache */
   /* search2110_1_1 -> 70 in the cache */
   /* search2110_1_2 -> 306 in the cache */
   /* search2110_1 -> 518 in the cache */
   /* search2110_2_0 -> 68 in the cache */
   /* search2110_2_1 -> 314 in the cache */
   /* search2110_2_2 -> 306 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 68, 314, 306 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 518, 526 },
      -1,
   } },

   /* replace2110_0 -> 68 in the cache */
   /* replace2110_1_0 -> 463 in the cache */
   /* replace2110_1_1 -> 70 in the cache */
   /* replace2110_1_2 -> 314 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 0,
      { 463, 70, 314 },
      -1,
   } },
   /* replace2110_2 -> 306 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      0, 1,
      { 68, 528, 306 },
      -1,
   } },

   /* ('bcsel', 'a', ('ffmaz', 'b', 'c', 'd'), ('ffmaz(is_used_once)', 'b', 'e', 'd')) => ('ffmaz', 'b', ('bcsel', 'a', 'c', 'e'), 'd') */
   /* search2111_0 -> 463 in the cache */
   /* search2111_1_0 -> 68 in the cache */
   /* search2111_1_1 -> 70 in the cache */
   /* search2111_1_2 -> 306 in the cache */
   /* search2111_1 -> 523 in the cache */
   /* search2111_2_0 -> 68 in the cache */
   /* search2111_2_1 -> 314 in the cache */
   /* search2111_2_2 -> 306 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffmaz,
      1, 1,
      { 68, 314, 306 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_bcsel,
      -1, 2,
      { 463, 523, 530 },
      -1,
   } },

   /* replace2111_0 -> 68 in the cache */
   /* replace2111_1_0 -> 463 in the cache */
   /* replace2111_1_1 -> 70 in the cache */
   /* replace2111_1_2 -> 314 in the cache */
   /* replace2111_1 -> 528 in the cache */
   /* replace2111_2 -> 306 in the cache */
   /* replace2111 -> 529 in the cache */

   /* ('~f2f32', ('fabs', ('f2fmp', 'a'))) => ('fabs', 'a') */
   /* search2112_0_0_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 67 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 533 },
      -1,
   } },

   /* replace2112_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fceil', ('f2fmp', 'a'))) => ('fceil', 'a') */
   /* search2113_0_0_0 -> 67 in the cache */
   /* search2113_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 536 },
      -1,
   } },

   /* replace2113_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fceil,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fcos', ('f2fmp', 'a'))) => ('fcos', 'a') */
   /* search2114_0_0_0 -> 67 in the cache */
   /* search2114_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 539 },
      -1,
   } },

   /* replace2114_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fcos,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fddx', ('f2fmp', 'a'))) => ('fddx', 'a') */
   /* search2115_0_0_0 -> 67 in the cache */
   /* search2115_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 542 },
      -1,
   } },

   /* replace2115_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddx,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fddx_coarse', ('f2fmp', 'a'))) => ('fddx_coarse', 'a') */
   /* search2116_0_0_0 -> 67 in the cache */
   /* search2116_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 545 },
      -1,
   } },

   /* replace2116_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddx_coarse,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fddx_fine', ('f2fmp', 'a'))) => ('fddx_fine', 'a') */
   /* search2117_0_0_0 -> 67 in the cache */
   /* search2117_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 548 },
      -1,
   } },

   /* replace2117_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddx_fine,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fddy', ('f2fmp', 'a'))) => ('fddy', 'a') */
   /* search2118_0_0_0 -> 67 in the cache */
   /* search2118_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 551 },
      -1,
   } },

   /* replace2118_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddy,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fddy_coarse', ('f2fmp', 'a'))) => ('fddy_coarse', 'a') */
   /* search2119_0_0_0 -> 67 in the cache */
   /* search2119_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 554 },
      -1,
   } },

   /* replace2119_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddy_coarse,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fddy_fine', ('f2fmp', 'a'))) => ('fddy_fine', 'a') */
   /* search2120_0_0_0 -> 67 in the cache */
   /* search2120_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 557 },
      -1,
   } },

   /* replace2120_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fddy_fine,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fexp2', ('f2fmp', 'a'))) => ('fexp2', 'a') */
   /* search2121_0_0_0 -> 67 in the cache */
   /* search2121_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 560 },
      -1,
   } },

   /* replace2121_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fexp2,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('ffloor', ('f2fmp', 'a'))) => ('ffloor', 'a') */
   /* search2122_0_0_0 -> 67 in the cache */
   /* search2122_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 563 },
      -1,
   } },

   /* replace2122_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffloor,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('ffract', ('f2fmp', 'a'))) => ('ffract', 'a') */
   /* search2123_0_0_0 -> 67 in the cache */
   /* search2123_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 566 },
      -1,
   } },

   /* replace2123_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_ffract,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('flog2', ('f2fmp', 'a'))) => ('flog2', 'a') */
   /* search2124_0_0_0 -> 67 in the cache */
   /* search2124_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 569 },
      -1,
   } },

   /* replace2124_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flog2,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fneg', ('f2fmp', 'a'))) => ('fneg', 'a') */
   /* search2125_0_0_0 -> 67 in the cache */
   /* search2125_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 572 },
      -1,
   } },

   /* replace2125_0 -> 67 in the cache */
   /* replace2125 -> 221 in the cache */

   /* ('~f2f32', ('frcp', ('f2fmp', 'a'))) => ('frcp', 'a') */
   /* search2126_0_0_0 -> 67 in the cache */
   /* search2126_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 574 },
      -1,
   } },

   /* replace2126_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frcp,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fround_even', ('f2fmp', 'a'))) => ('fround_even', 'a') */
   /* search2127_0_0_0 -> 67 in the cache */
   /* search2127_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 577 },
      -1,
   } },

   /* replace2127_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fround_even,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('frsq', ('f2fmp', 'a'))) => ('frsq', 'a') */
   /* search2128_0_0_0 -> 67 in the cache */
   /* search2128_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 580 },
      -1,
   } },

   /* replace2128_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frsq,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fsat', ('f2fmp', 'a'))) => ('fsat', 'a') */
   /* search2129_0_0_0 -> 67 in the cache */
   /* search2129_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 583 },
      -1,
   } },

   /* replace2129_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsat,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fsign', ('f2fmp', 'a'))) => ('fsign', 'a') */
   /* search2130_0_0_0 -> 67 in the cache */
   /* search2130_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 586 },
      -1,
   } },

   /* replace2130_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fsin', ('f2fmp', 'a'))) => ('fsin', 'a') */
   /* search2131_0_0_0 -> 67 in the cache */
   /* search2131_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 589 },
      -1,
   } },

   /* replace2131_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsin,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fsqrt', ('f2fmp', 'a'))) => ('fsqrt', 'a') */
   /* search2132_0_0_0 -> 67 in the cache */
   /* search2132_0_0 -> 532 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 532 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 592 },
      -1,
   } },

   /* replace2132_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fsqrt,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('~f2f32', ('fadd', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fadd', 'a', 'b') */
   /* search2133_0_0_0 -> 67 in the cache */
   /* search2133_0_0 -> 532 in the cache */
   /* search2133_0_1_0 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 68 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 596 },
      -1,
   } },

   /* replace2133_0 -> 67 in the cache */
   /* replace2133_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('fdiv', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fdiv', 'a', 'b') */
   /* search2134_0_0_0 -> 67 in the cache */
   /* search2134_0_0 -> 532 in the cache */
   /* search2134_0_1_0 -> 68 in the cache */
   /* search2134_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 599 },
      -1,
   } },

   /* replace2134_0 -> 67 in the cache */
   /* replace2134_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fdiv,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('fmax', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmax', 'a', 'b') */
   /* search2135_0_0_0 -> 67 in the cache */
   /* search2135_0_0 -> 532 in the cache */
   /* search2135_0_1_0 -> 68 in the cache */
   /* search2135_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 602 },
      -1,
   } },

   /* replace2135_0 -> 67 in the cache */
   /* replace2135_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('fmin', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmin', 'a', 'b') */
   /* search2136_0_0_0 -> 67 in the cache */
   /* search2136_0_0 -> 532 in the cache */
   /* search2136_0_1_0 -> 68 in the cache */
   /* search2136_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 605 },
      -1,
   } },

   /* replace2136_0 -> 67 in the cache */
   /* replace2136_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('fmod', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmod', 'a', 'b') */
   /* search2137_0_0_0 -> 67 in the cache */
   /* search2137_0_0 -> 532 in the cache */
   /* search2137_0_1_0 -> 68 in the cache */
   /* search2137_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 608 },
      -1,
   } },

   /* replace2137_0 -> 67 in the cache */
   /* replace2137_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmod,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('fmul', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fmul', 'a', 'b') */
   /* search2138_0_0_0 -> 67 in the cache */
   /* search2138_0_0 -> 532 in the cache */
   /* search2138_0_1_0 -> 68 in the cache */
   /* search2138_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 611 },
      -1,
   } },

   /* replace2138_0 -> 67 in the cache */
   /* replace2138_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('fpow', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('fpow', 'a', 'b') */
   /* search2139_0_0_0 -> 67 in the cache */
   /* search2139_0_0 -> 532 in the cache */
   /* search2139_0_1_0 -> 68 in the cache */
   /* search2139_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 614 },
      -1,
   } },

   /* replace2139_0 -> 67 in the cache */
   /* replace2139_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_fpow,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('frem', ('f2fmp', 'a'), ('f2fmp', 'b'))) => ('frem', 'a', 'b') */
   /* search2140_0_0_0 -> 67 in the cache */
   /* search2140_0_0 -> 532 in the cache */
   /* search2140_0_1_0 -> 68 in the cache */
   /* search2140_0_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 532, 595 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 617 },
      -1,
   } },

   /* replace2140_0 -> 67 in the cache */
   /* replace2140_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_frem,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('~f2f32', ('ffma', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('ffma', 'a', 'b', 'c') */
   /* search2141_0_0_0 -> 67 in the cache */
   /* search2141_0_0 -> 532 in the cache */
   /* search2141_0_1_0 -> 68 in the cache */
   /* search2141_0_1 -> 595 in the cache */
   /* search2141_0_2_0 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2fmp,
      -1, 0,
      { 70 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 532, 595, 620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 1,
      { 621 },
      -1,
   } },

   /* replace2141_0 -> 67 in the cache */
   /* replace2141_1 -> 68 in the cache */
   /* replace2141_2 -> 70 in the cache */
   /* replace2141 -> 72 in the cache */

   /* ('~f2f32', ('flrp', ('f2fmp', 'a'), ('f2fmp', 'b'), ('f2fmp', 'c'))) => ('flrp', 'a', 'b', 'c') */
   /* search2142_0_0_0 -> 67 in the cache */
   /* search2142_0_0 -> 532 in the cache */
   /* search2142_0_1_0 -> 68 in the cache */
   /* search2142_0_1 -> 595 in the cache */
   /* search2142_0_2_0 -> 70 in the cache */
   /* search2142_0_2 -> 620 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 532, 595, 620 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, 32 },
      true,
      false,
      false,
      nir_op_f2f32,
      -1, 0,
      { 623 },
      -1,
   } },

   /* replace2142_0 -> 67 in the cache */
   /* replace2142_1 -> 68 in the cache */
   /* replace2142_2 -> 70 in the cache */
   { .expression = {
      { nir_search_value_expression, 32 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 67, 68, 70 },
      -1,
   } },

   /* ('~feq', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('feq', 'a', 'b') */
   /* search2143_0_0 -> 67 in the cache */
   /* search2143_0 -> 532 in the cache */
   /* search2143_1_0 -> 68 in the cache */
   /* search2143_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 532, 595 },
      -1,
   } },

   /* replace2143_0 -> 67 in the cache */
   /* replace2143_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_feq,
      0, 1,
      { 67, 68 },
      -1,
   } },

   /* ('~fge', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fge', 'a', 'b') */
   /* search2144_0_0 -> 67 in the cache */
   /* search2144_0 -> 532 in the cache */
   /* search2144_1_0 -> 68 in the cache */
   /* search2144_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 532, 595 },
      -1,
   } },

   /* replace2144_0 -> 67 in the cache */
   /* replace2144_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fge,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('~flt', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('flt', 'a', 'b') */
   /* search2145_0_0 -> 67 in the cache */
   /* search2145_0 -> 532 in the cache */
   /* search2145_1_0 -> 68 in the cache */
   /* search2145_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 532, 595 },
      -1,
   } },

   /* replace2145_0 -> 67 in the cache */
   /* replace2145_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 67, 68 },
      -1,
   } },

   /* ('~fneu', ('f2fmp', 'a'), ('f2fmp', 'b')) => ('fneu', 'a', 'b') */
   /* search2146_0_0 -> 67 in the cache */
   /* search2146_0 -> 532 in the cache */
   /* search2146_1_0 -> 68 in the cache */
   /* search2146_1 -> 595 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      true,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 532, 595 },
      -1,
   } },

   /* replace2146_0 -> 67 in the cache */
   /* replace2146_1 -> 68 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fneu,
      0, 1,
      { 67, 68 },
      -1,
   } },

   /* ('f2fmp', 'a') => ('f2f16', 'a') */
   /* search2147_0 -> 67 in the cache */
   /* search2147 -> 532 in the cache */

   /* replace2147_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2f16,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('f2imp', 'a') => ('f2i16', 'a') */
   /* search2148_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2imp,
      -1, 0,
      { 67 },
      -1,
   } },

   /* replace2148_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2i16,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('f2ump', 'a') => ('f2u16', 'a') */
   /* search2149_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2ump,
      -1, 0,
      { 67 },
      -1,
   } },

   /* replace2149_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_f2u16,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('i2imp', 'a') => ('i2i16', 'a') */
   /* search2150_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2imp,
      -1, 0,
      { 67 },
      -1,
   } },

   /* replace2150_0 -> 67 in the cache */
   /* replace2150 -> 381 in the cache */

   /* ('i2fmp', 'a') => ('i2f16', 'a') */
   /* search2151_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2fmp,
      -1, 0,
      { 67 },
      -1,
   } },

   /* replace2151_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_i2f16,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('i2imp', 'a') => ('u2u16', 'a') */
   /* search2152_0 -> 67 in the cache */
   /* search2152 -> 639 in the cache */

   /* replace2152_0 -> 67 in the cache */
   /* replace2152 -> 378 in the cache */

   /* ('u2fmp', 'a') => ('u2f16', 'a') */
   /* search2153_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2fmp,
      -1, 0,
      { 67 },
      -1,
   } },

   /* replace2153_0 -> 67 in the cache */
   { .expression = {
      { nir_search_value_expression, 16 },
      false,
      false,
      false,
      nir_op_u2f16,
      -1, 0,
      { 67 },
      -1,
   } },

   /* ('fisfinite', 'a') => ('flt', ('fabs', 'a'), inf) */
   /* search2154_0 -> 42 in the cache */
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_fisfinite,
      -1, 0,
      { 42 },
      -1,
   } },

   /* replace2154_0_0 -> 42 in the cache */
   /* replace2154_0 -> 47 in the cache */
   { .constant = {
      { nir_search_value_constant, -1 },
      nir_type_float, { 0x7ff0000000000000 /* inf */ },
   } },
   { .expression = {
      { nir_search_value_expression, 1 },
      false,
      false,
      false,
      nir_op_flt,
      -1, 0,
      { 47, 645 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_late_expression_cond[] = {
   (is_used_once),
   (is_only_used_as_float),
};

static const nir_search_variable_cond nir_opt_algebraic_late_variable_cond[] = {
   (is_finite),
   (is_not_const),
   (is_gt_0_and_lt_1),
   (is_not_fmul),
   (is_not_negative),
   (is_not_const_and_not_fsign),
};

static const struct transform nir_opt_algebraic_late_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 184, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 183, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 101, 103, 117 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 104, 105, 118 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 191, 192, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 193, 194, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 195, 196, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 197, 198, 121 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 204, 213, 122 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 218, 17 },
   { 220, 222, 17 },
   { 223, 224, 17 },
   { 225, 226, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 336, 335, 0 },
   { 340, 345, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 337, 335, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 337, 335, 0 },
   { 346, 350, 124 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 351, 42, 0 },
   { 353, 42, 0 },
   { 407, 400, 125 },
   { 404, 410, 126 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 354, 42, 0 },
   { 355, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 356, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 532, 634, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 635, 636, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 637, 638, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 639, 381, 0 },
   { 639, 378, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 640, 641, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 642, 643, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 644, 646, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 10, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 174, 175, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 135, 137, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 6, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 172, 173, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 164, 165, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 630, 631, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 42, 0 },
   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 17, 16, 0 },
   { 29, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 178, 149, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 13, 14, 0 },
   { 26, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 176, 177, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 166, 167, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 140, 141, 0 },
   { 148, 149, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 628, 629, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 20, 21, 0 },
   { 32, 21, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 181, 182, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 142, 143, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 168, 169, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 626, 627, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 22, 23, 0 },
   { 33, 23, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 179, 180, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 144, 145, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 170, 171, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 632, 633, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 254, 257, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 201, 203, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 498, 500, 0 },
   { 503, 500, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 506, 508, 0 },
   { 511, 508, 0 },
   { 513, 515, 0 },
   { 517, 515, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 520, 522, 0 },
   { 525, 522, 0 },
   { 527, 529, 0 },
   { 531, 529, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 258, 260, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 228, 218, 17 },
   { 230, 222, 17 },
   { 231, 224, 17 },
   { 232, 226, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 80, 81, 32 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 129, 131, 120 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 128, 120 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 123, 128, 120 },
   { 129, 131, 120 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 246, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 184, 42, 0 },
   { 403, 392, 125 },
   { 406, 396, 125 },
   { 409, 400, 125 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 184, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 46, 48, 0 },
   { 184, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 184, 42, 0 },
   { 420, 422, 0 },
   { 429, 431, 0 },
   { 438, 440, 0 },
   { 448, 450, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 57, 58, 0 },
   { 183, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 183, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 183, 42, 0 },
   { 425, 426, 0 },
   { 434, 435, 0 },
   { 443, 444, 0 },
   { 453, 454, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 110, 114, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 99, 100, 116 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 116, 118, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 99, 100, 116 },
   { 110, 114, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 110, 114, 119 },
   { 116, 118, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 99, 100, 116 },
   { 116, 118, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 119, 121, 119 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 359, 357, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 361, 357, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 362, 360, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 363, 360, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 351, 42, 0 },
   { 353, 42, 0 },
   { 390, 392, 125 },
   { 394, 396, 125 },
   { 398, 400, 125 },
   { 407, 400, 125 },
   { 404, 410, 126 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 412, 413, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 365, 367, 57 },
   { 374, 375, 57 },
   { 379, 380, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 369, 367, 57 },
   { 377, 375, 57 },
   { 382, 380, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 370, 371, 57 },
   { 383, 384, 57 },
   { 386, 387, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 372, 371, 57 },
   { 385, 384, 57 },
   { 388, 387, 58 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 415, 416, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 204, 213, 122 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 12, 6, 0 },
   { 174, 175, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 8, 10, 0 },
   { 172, 173, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 18, 14, 0 },
   { 30, 14, 0 },
   { 178, 149, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 16, 0 },
   { 28, 16, 0 },
   { 176, 177, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 158, 163, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 241, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 244, 242, 17 },
   { 246, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 216, 218, 17 },
   { 220, 222, 17 },
   { 223, 224, 17 },
   { 225, 226, 17 },
   { 237, 238, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 17 },
   { 241, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 17 },
   { 244, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 237, 238, 17 },
   { 246, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 292, 297, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 188, 190, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 458, 462, 14 },
   { 475, 479, 16 },
   { 488, 492, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 228, 218, 17 },
   { 230, 222, 17 },
   { 231, 224, 17 },
   { 232, 226, 17 },
   { 458, 462, 14 },
   { 475, 479, 16 },
   { 488, 492, 18 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 292, 297, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 292, 297, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 292, 297, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 292, 297, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 292, 297, 123 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 573, 221, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 606, 607, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 624, 625, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 612, 613, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 603, 604, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 597, 598, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 622, 72, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 584, 585, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 578, 579, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 593, 594, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 581, 582, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 534, 535, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 537, 538, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 540, 541, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 543, 544, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 546, 547, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 549, 550, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 552, 553, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 555, 556, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 558, 559, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 561, 562, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 564, 565, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 567, 568, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 570, 571, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 575, 576, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 587, 588, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 590, 591, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 600, 601, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 609, 610, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 615, 616, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 618, 619, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 467, 471, 0 },
   { 481, 482, 0 },
   { 494, 495, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 266, 267, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 248, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 248, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 248, 242, 17 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 38, 41, 0 },
   { 184, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 53, 54, 0 },
   { 183, 42, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 284, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 284, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 278, 284, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 278, 284, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 278, 284, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 284, 123 },
   { 292, 297, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 292, 297, 123 },
   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 284, 123 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 313, 319, 123 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 278, 284, 123 },
   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 59, 97, 115 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 65, 66, 31 },
   { 71, 72, 32 },
   { 77, 78, 33 },
   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 80, 81, 32 },
   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 292, 297, 123 },
   { 299, 305, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 292, 297, 123 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { 325, 327, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 325, 327, 123 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 278, 284, 123 },
   { 299, 305, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 278, 284, 123 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { 313, 319, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 313, 319, 123 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 85, 86, 113 },
   { 88, 89, 113 },
   { 91, 92, 113 },
   { 94, 95, 114 },
   { 299, 305, 123 },
   { 329, 334, 123 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_late_pass_op_table[nir_num_search_ops] = {
   [nir_op_flt] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         52,
         53,
         54,
         0,
         166,
         0,
         55,
         0,
         0,
         0,
         0,
         0,
         0,
         56,
         0,
         57,
         0,
         0,
         57,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         58,
         0,
         0,
         167,
         0,
         57,
         0,
         0,
         57,
      },
   },
   [nir_op_fadd] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         7,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         3,
         9,
         4,
         8,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         11,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         12,
         0,
         10,
         0,
         10,
         12,
         10,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         8,
         0,
         0,
         13,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         14,
         15,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 16,
      .table = (const uint16_t []) {
      
         2,
         3,
         98,
         99,
         99,
         100,
         2,
         98,
         99,
         99,
         182,
         183,
         235,
         236,
         252,
         253,
         3,
         3,
         101,
         102,
         103,
         104,
         3,
         184,
         185,
         186,
         187,
         188,
         237,
         238,
         254,
         255,
         98,
         101,
         98,
         105,
         105,
         106,
         98,
         98,
         105,
         105,
         189,
         190,
         239,
         240,
         252,
         253,
         99,
         102,
         105,
         99,
         99,
         107,
         99,
         105,
         99,
         99,
         191,
         192,
         241,
         242,
         256,
         257,
         99,
         103,
         105,
         99,
         99,
         107,
         99,
         105,
         99,
         99,
         191,
         192,
         241,
         242,
         256,
         257,
         100,
         104,
         106,
         107,
         107,
         100,
         100,
         106,
         107,
         107,
         193,
         194,
         243,
         244,
         258,
         259,
         2,
         3,
         98,
         99,
         99,
         100,
         108,
         98,
         99,
         99,
         182,
         183,
         235,
         236,
         252,
         253,
         98,
         184,
         98,
         105,
         105,
         106,
         98,
         98,
         105,
         105,
         189,
         190,
         239,
         240,
         252,
         253,
         99,
         185,
         105,
         99,
         99,
         107,
         99,
         105,
         99,
         99,
         191,
         192,
         241,
         242,
         256,
         257,
         99,
         186,
         105,
         99,
         99,
         107,
         99,
         105,
         99,
         99,
         191,
         192,
         241,
         242,
         256,
         257,
         182,
         187,
         189,
         191,
         191,
         193,
         182,
         189,
         191,
         191,
         182,
         195,
         245,
         246,
         260,
         261,
         183,
         188,
         190,
         192,
         192,
         194,
         183,
         190,
         192,
         192,
         195,
         183,
         247,
         248,
         262,
         263,
         235,
         237,
         239,
         241,
         241,
         243,
         235,
         239,
         241,
         241,
         245,
         247,
         235,
         249,
         264,
         265,
         236,
         238,
         240,
         242,
         242,
         244,
         236,
         240,
         242,
         242,
         246,
         248,
         249,
         236,
         266,
         267,
         252,
         254,
         252,
         256,
         256,
         258,
         252,
         252,
         256,
         256,
         260,
         262,
         264,
         266,
         252,
         268,
         253,
         255,
         253,
         257,
         257,
         259,
         253,
         253,
         257,
         257,
         261,
         263,
         265,
         267,
         268,
         253,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         7,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         4,
         59,
         60,
         61,
         62,
         63,
         250,
         251,
      },
   },
   [nir_op_fge] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         64,
         65,
         0,
         0,
         168,
         0,
         0,
         66,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         67,
         0,
         68,
         0,
         0,
         68,
         0,
         0,
         69,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         70,
         0,
         0,
         0,
         169,
         0,
         68,
         0,
         0,
         68,
         0,
         0,
         170,
         0,
         0,
         0,
         0,
         0,
         0,
      },
   },
   [nir_op_feq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 7,
      .table = (const uint16_t []) {
      
         5,
         6,
         5,
         5,
         5,
         5,
         5,
         6,
         6,
         71,
         72,
         73,
         74,
         6,
         5,
         71,
         5,
         5,
         5,
         5,
         5,
         5,
         72,
         5,
         75,
         5,
         5,
         5,
         5,
         73,
         5,
         5,
         5,
         5,
         5,
         5,
         74,
         5,
         5,
         5,
         5,
         5,
         5,
         6,
         5,
         5,
         5,
         5,
         76,
      },
   },
   [nir_op_fneu] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         7,
         8,
         7,
         7,
         7,
         7,
         8,
         8,
         77,
         78,
         79,
         8,
         7,
         77,
         7,
         7,
         7,
         7,
         7,
         78,
         7,
         80,
         7,
         7,
         7,
         79,
         7,
         7,
         7,
         7,
         7,
         8,
         7,
         7,
         7,
         81,
      },
   },
   [nir_op_iand] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         2,
         2,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         4,
         4,
         4,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         115,
         9,
         9,
         9,
         9,
         116,
         9,
         117,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         232,
         9,
         9,
         117,
         9,
         9,
         9,
         9,
         232,
         9,
         115,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         118,
         9,
         9,
         9,
         9,
         232,
         232,
         9,
         9,
         9,
      },
   },
   [nir_op_ior] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         1,
         1,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         10,
         233,
         10,
         10,
         10,
         119,
         10,
         233,
         10,
         10,
         119,
         120,
         10,
         10,
         10,
         10,
         10,
         10,
         121,
         10,
         10,
         233,
         233,
         10,
         10,
         10,
      },
   },
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         2,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         0,
         5,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 6,
      .table = (const uint16_t []) {
      
         11,
         12,
         90,
         91,
         11,
         229,
         12,
         12,
         92,
         93,
         12,
         230,
         90,
         92,
         90,
         94,
         90,
         229,
         91,
         93,
         94,
         91,
         91,
         231,
         11,
         12,
         90,
         91,
         95,
         229,
         229,
         230,
         229,
         231,
         229,
         229,
      },
   },
   [nir_op_fmulz] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         13,
      },
   },
   [nir_op_iadd] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         14,
         123,
         124,
         125,
         123,
         123,
         126,
         127,
         124,
         126,
         124,
         128,
         125,
         127,
         128,
         125,
      },
   },
   [nir_op_ineg] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         15,
      },
   },
   [nir_op_iabs] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         16,
      },
   },
   [nir_op_isub] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         17,
         17,
         129,
         129,
      },
   },
   [nir_op_vec2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         109,
         110,
         111,
      },
   },
   [nir_op_fsat] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         18,
         142,
      },
   },
   [nir_op_fmin] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         2,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         1,
         1,
         2,
         2,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         82,
         82,
         0,
         0,
         82,
         83,
         0,
         0,
         0,
         0,
         84,
      },
   },
   [nir_op_fdot2] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         19,
      },
   },
   [nir_op_fdot3] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         20,
      },
   },
   [nir_op_fdot4] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         21,
      },
   },
   [nir_op_fdph] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         22,
      },
   },
   [nir_op_flrp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         85,
         85,
         85,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         86,
      },
   },
   [nir_op_fround_even] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         23,
         143,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         6,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         7,
         4,
         8,
         7,
         3,
         0,
         0,
         9,
         0,
         0,
         9,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         10,
         10,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         10,
         0,
         10,
         0,
         10,
         0,
         0,
         10,
         10,
         0,
         0,
         9,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         4,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         9,
         9,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
         9,
      },
      
      .num_filtered_states = 11,
      .table = (const uint16_t []) {
      
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         25,
         24,
         112,
         113,
         24,
         172,
         173,
         177,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         25,
         24,
         112,
         113,
         24,
         172,
         173,
         177,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         174,
         178,
         174,
         175,
         175,
         174,
         179,
         180,
         180,
         174,
         176,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         25,
         24,
         112,
         113,
         24,
         172,
         173,
         177,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         174,
         174,
         174,
         175,
         181,
         174,
         174,
         175,
         181,
         174,
         176,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         114,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         25,
         24,
         112,
         113,
         24,
         172,
         173,
         177,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         174,
         174,
         174,
         175,
         181,
         174,
         174,
         175,
         181,
         174,
         176,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         178,
         174,
         175,
         175,
         174,
         179,
         180,
         180,
         174,
         176,
         174,
         174,
         174,
         175,
         181,
         174,
         174,
         175,
         181,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         174,
         174,
         175,
         181,
         174,
         174,
         175,
         181,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         25,
         24,
         112,
         112,
         24,
         172,
         173,
         173,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         113,
         24,
         24,
         112,
         113,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
         174,
         174,
         174,
         175,
         175,
         174,
         174,
         175,
         175,
         174,
         176,
         24,
         24,
         24,
         112,
         112,
         24,
         24,
         112,
         112,
         24,
         171,
      },
   },
   [nir_op_fmax] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         96,
         0,
         0,
         0,
         97,
      },
   },
   [nir_op_bcsel] = {
      .filter = (const uint16_t []) {
         0,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         0,
         4,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         3,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         4,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         3,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         5,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
      },
      
      .num_filtered_states = 8,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         228,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         227,
         0,
         0,
         87,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         88,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         89,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
   },
   [nir_op_fsqrt] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         26,
         144,
      },
   },
   [nir_op_frsq] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         27,
         145,
      },
   },
   [nir_op_ffmaz] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
         28,
         122,
         234,
      },
   },
   [nir_op_ubfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         29,
         0,
         30,
         0,
         29,
         0,
         30,
      },
   },
   [nir_op_ibfe] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         31,
         0,
         32,
         0,
         31,
         0,
         32,
      },
   },
   [nir_op_ishl] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         33,
         0,
         0,
         33,
         0,
         0,
         134,
         0,
      },
   },
   [nir_op_ishr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         34,
         0,
         34,
      },
   },
   [nir_op_ushr] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         35,
         0,
         35,
      },
   },
   [nir_op_extract_i8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         36,
         130,
         36,
         36,
         36,
         131,
         36,
         36,
      },
   },
   [nir_op_extract_u8] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         2,
         3,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 4,
      .table = (const uint16_t []) {
      
         37,
         38,
         37,
         37,
         37,
         38,
         37,
         37,
         37,
         132,
         37,
         37,
         37,
         133,
         37,
         37,
      },
   },
   [nir_search_op_u2u] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         39,
         136,
         137,
      },
   },
   [nir_search_op_i2i] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         2,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         40,
         138,
         139,
      },
   },
   [nir_op_insert_u8] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         135,
         135,
      },
   },
   [nir_op_insert_u16] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         140,
         140,
      },
   },
   [nir_op_extract_u16] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         41,
         0,
         41,
      },
   },
   [nir_op_ine] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         42,
         42,
         42,
      },
   },
   [nir_op_ieq] = {
      .filter = (const uint16_t []) {
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         43,
         43,
         43,
      },
   },
   [nir_search_op_b2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         141,
      },
   },
   [nir_op_inot] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         44,
      },
   },
   [nir_search_op_f2f] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         3,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         4,
         0,
         5,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         6,
         0,
         0,
         0,
         0,
         0,
         7,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         8,
         9,
         10,
         11,
         12,
         13,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
         26,
         27,
         28,
         29,
         30,
         31,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 32,
      .table = (const uint16_t []) {
      
         0,
         196,
         197,
         198,
         199,
         200,
         201,
         202,
         203,
         204,
         205,
         206,
         207,
         208,
         209,
         210,
         211,
         212,
         213,
         214,
         215,
         216,
         217,
         218,
         219,
         220,
         221,
         222,
         223,
         224,
         225,
         226,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         146,
      },
   },
   [nir_op_f2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         45,
      },
   },
   [nir_op_fceil] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         147,
      },
   },
   [nir_op_fcos] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         148,
      },
   },
   [nir_op_fddx] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         149,
      },
   },
   [nir_op_fddx_coarse] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         150,
      },
   },
   [nir_op_fddx_fine] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         151,
      },
   },
   [nir_op_fddy] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         152,
      },
   },
   [nir_op_fddy_coarse] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         153,
      },
   },
   [nir_op_fddy_fine] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         154,
      },
   },
   [nir_op_fexp2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         155,
      },
   },
   [nir_op_ffloor] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         156,
      },
   },
   [nir_op_ffract] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         157,
      },
   },
   [nir_op_flog2] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         158,
      },
   },
   [nir_op_frcp] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         159,
      },
   },
   [nir_op_fsign] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         160,
      },
   },
   [nir_op_fsin] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         161,
      },
   },
   [nir_op_fdiv] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         162,
      },
   },
   [nir_op_fmod] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         163,
      },
   },
   [nir_op_fpow] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         164,
      },
   },
   [nir_op_frem] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         0,
         0,
         0,
         165,
      },
   },
   [nir_op_f2imp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         46,
      },
   },
   [nir_op_f2ump] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         47,
      },
   },
   [nir_op_i2imp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         48,
      },
   },
   [nir_op_i2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         49,
      },
   },
   [nir_op_u2fmp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         50,
      },
   },
   [nir_op_fisfinite] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         51,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_late_transform_offsets[] = {
   0,
   0,
   0,
   0,
   1,
   0,
   0,
   0,
   0,
   3,
   5,
   0,
   0,
   0,
   0,
   7,
   9,
   0,
   0,
   11,
   13,
   15,
   17,
   19,
   0,
   21,
   0,
   0,
   0,
   26,
   28,
   31,
   33,
   36,
   41,
   44,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   46,
   48,
   50,
   52,
   55,
   57,
   59,
   61,
   63,
   65,
   67,
   69,
   71,
   73,
   75,
   77,
   80,
   82,
   84,
   86,
   89,
   91,
   94,
   96,
   98,
   101,
   103,
   106,
   108,
   0,
   110,
   112,
   114,
   117,
   119,
   121,
   123,
   0,
   125,
   0,
   127,
   0,
   129,
   132,
   137,
   0,
   0,
   0,
   0,
   0,
   0,
   142,
   0,
   144,
   149,
   153,
   155,
   160,
   164,
   172,
   174,
   182,
   188,
   0,
   193,
   195,
   197,
   0,
   200,
   0,
   202,
   207,
   209,
   212,
   218,
   221,
   223,
   0,
   229,
   231,
   233,
   235,
   238,
   241,
   244,
   246,
   248,
   250,
   252,
   254,
   262,
   264,
   268,
   272,
   276,
   280,
   0,
   0,
   282,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   284,
   287,
   290,
   294,
   298,
   0,
   300,
   302,
   304,
   306,
   308,
   310,
   313,
   319,
   322,
   325,
   328,
   330,
   332,
   338,
   345,
   356,
   358,
   360,
   366,
   372,
   377,
   382,
   385,
   388,
   391,
   393,
   395,
   397,
   399,
   401,
   403,
   405,
   407,
   409,
   411,
   413,
   415,
   417,
   419,
   421,
   423,
   425,
   427,
   429,
   431,
   433,
   435,
   437,
   439,
   441,
   443,
   445,
   447,
   449,
   451,
   453,
   457,
   459,
   461,
   463,
   465,
   468,
   0,
   471,
   473,
   475,
   477,
   479,
   485,
   491,
   496,
   501,
   504,
   507,
   510,
   513,
   516,
   519,
   522,
   524,
   526,
   532,
   538,
   544,
   550,
   559,
   568,
   575,
   582,
   589,
   596,
   603,
   610,
   617,
   624,
   631,
   638,
};

static const nir_algebraic_table nir_opt_algebraic_late_table = {
   .transforms = nir_opt_algebraic_late_transforms,
   .transform_offsets = nir_opt_algebraic_late_transform_offsets,
   .pass_op_table = nir_opt_algebraic_late_pass_op_table,
   .values = nir_opt_algebraic_late_values,
   .expression_cond = nir_opt_algebraic_late_expression_cond,
   .variable_cond = nir_opt_algebraic_late_variable_cond,
};

bool
nir_opt_algebraic_late(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[127];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   STATIC_ASSERT(647 == ARRAY_SIZE(nir_opt_algebraic_late_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ffloor;
   condition_flags[22] = options->lower_ffloor || ((options->lower_doubles_options & nir_lower_dfloor) && !(options->lower_doubles_options & nir_lower_dfloor));
   condition_flags[23] = !options->lower_ffloor;
   condition_flags[24] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[25] = options->lower_ffract;
   condition_flags[26] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[27] = options->lower_fceil;
   condition_flags[28] = options->lower_ffma16;
   condition_flags[29] = options->lower_ffma32;
   condition_flags[30] = options->lower_ffma64;
   condition_flags[31] = options->fuse_ffma16;
   condition_flags[32] = options->fuse_ffma32;
   condition_flags[33] = options->fuse_ffma64;
   condition_flags[34] = options->lower_fdph;
   condition_flags[35] = !options->lower_fdph;
   condition_flags[36] = options->lower_fdot;
   condition_flags[37] = !options->lower_fsat;
   condition_flags[38] = !options->lower_iabs;
   condition_flags[39] = options->lower_fsat;
   condition_flags[40] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[41] = !options->lower_fsign;
   condition_flags[42] = options->lower_umax;
   condition_flags[43] = options->lower_umin;
   condition_flags[44] = !options->lower_umax;
   condition_flags[45] = !options->lower_umin;
   condition_flags[46] = options->lower_scmp;
   condition_flags[47] = options->lower_vector_cmp;
   condition_flags[48] = !options->lower_rotate;
   condition_flags[49] = options->lower_rotate;
   condition_flags[50] = options->lower_fpow;
   condition_flags[51] = !options->lower_fpow;
   condition_flags[52] = options->lower_fdiv;
   condition_flags[53] = options->lower_fsqrt;
   condition_flags[54] = !options->lower_fsqrt;
   condition_flags[55] = options->lower_sincos;
   condition_flags[56] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[57] = !options->lower_extract_byte;
   condition_flags[58] = !options->lower_extract_word;
   condition_flags[59] = options->has_pack_32_4x8;
   condition_flags[60] = options->lower_pack_64_2x32_split;
   condition_flags[61] = options->lower_pack_32_2x16_split;
   condition_flags[62] = options->lower_unpack_64_2x32_split;
   condition_flags[63] = options->lower_unpack_32_2x16_split;
   condition_flags[64] = options->lower_fmod;
   condition_flags[65] = options->lower_uadd_carry;
   condition_flags[66] = options->lower_usub_borrow;
   condition_flags[67] = options->lower_bitfield_insert;
   condition_flags[68] = options->lower_hadd;
   condition_flags[69] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[70] = options->lower_mul_32x16;
   condition_flags[71] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[72] = options->lower_uadd_sat;
   condition_flags[73] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[74] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[75] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[76] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[77] = options->lower_bitfield_insert_to_shifts;
   condition_flags[78] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[79] = options->lower_bitfield_extract;
   condition_flags[80] = options->lower_bitfield_extract_to_shifts;
   condition_flags[81] = options->lower_ifind_msb;
   condition_flags[82] = options->lower_find_msb_to_reverse;
   condition_flags[83] = options->lower_find_lsb;
   condition_flags[84] = options->lower_extract_byte;
   condition_flags[85] = options->lower_extract_word;
   condition_flags[86] = options->lower_pack_unorm_2x16;
   condition_flags[87] = options->lower_pack_unorm_4x8;
   condition_flags[88] = options->lower_pack_snorm_2x16;
   condition_flags[89] = options->lower_pack_snorm_4x8;
   condition_flags[90] = options->lower_unpack_unorm_2x16;
   condition_flags[91] = options->lower_unpack_unorm_4x8;
   condition_flags[92] = options->lower_unpack_snorm_2x16;
   condition_flags[93] = options->lower_unpack_snorm_4x8;
   condition_flags[94] = options->lower_pack_split;
   condition_flags[95] = options->lower_isign;
   condition_flags[96] = !options->lower_isign;
   condition_flags[97] = options->lower_fsign;
   condition_flags[98] = !options->has_imul24;
   condition_flags[99] = !options->has_umul24;
   condition_flags[100] = !options->has_umad24;
   condition_flags[101] = options->has_imul24;
   condition_flags[102] = options->has_umad24;
   condition_flags[103] = options->has_umul24;
   condition_flags[104] = options->has_fused_comp_and_csel;
   condition_flags[105] = options->lower_iadd_sat;
   condition_flags[106] = options->support_8bit_alu;
   condition_flags[107] = options->support_16bit_alu;
   condition_flags[108] = options->lower_ldexp;
   condition_flags[109] = !options->lower_bitfield_reverse;
   condition_flags[110] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[111] = options->lower_fisnormal;
   condition_flags[112] = !options->vectorize_vec2_16bit;
   condition_flags[113] = options->has_fsub;
   condition_flags[114] = options->has_fsub && !(options->lower_doubles_options & nir_lower_dsub);
   condition_flags[115] = options->lower_fneg;
   condition_flags[116] = options->has_isub || options->lower_ineg;
   condition_flags[117] = options->lower_ineg;
   condition_flags[118] = options->lower_iabs;
   condition_flags[119] = options->has_iadd3;
   condition_flags[120] = options->vectorize_vec2_16bit;
   condition_flags[121] = options->fdot_replicates;
   condition_flags[122] = options->lower_fround_even;
   condition_flags[123] = (info->stage != MESA_SHADER_VERTEX && info->stage != MESA_SHADER_GEOMETRY) && !options->intel_vec4;
   condition_flags[124] = options->avoid_ternary_with_two_constants;
   condition_flags[125] = !options->lower_insert_byte;
   condition_flags[126] = !options->lower_insert_word;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_late_table);
      }
   }

   return progress;
}


#include "nir.h"
#include "nir_builder.h"
#include "nir_search.h"
#include "nir_search_helpers.h"

/* What follows is NIR algebraic transform code for the following 19
 * transforms:
 *    ('fmul', ('fneg', 'a'), ('fneg', 'b')) => ('fmul', 'a', 'b')
 *    ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c')
 *    ('fdot2_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot2_replicated', 'a', 'b')
 *    ('fdot3_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot3_replicated', 'a', 'b')
 *    ('fdot4_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot4_replicated', 'a', 'b')
 *    ('fneg', ('fneg', 'a')) => a
 *    ('fneg', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fneg', 'a'), 'b')
 *    ('fabs', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fabs', 'a'), ('fabs', 'b'))
 *    ('fneg', ('ffma(is_used_once)', 'a', 'b', 'c')) => ('ffma', ('fneg', 'a'), 'b', ('fneg', 'c'))
 *    ('fneg', ('flrp(is_used_once)', 'a', 'b', 'c')) => ('flrp', ('fneg', 'a'), ('fneg', 'b'), 'c')
 *    ('fneg', ('~fadd(is_used_once)', 'a', 'b')) => ('fadd', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fmin(is_used_once)', 'a', 'b')) => ('fmax', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fmax(is_used_once)', 'a', 'b')) => ('fmin', ('fneg', 'a'), ('fneg', 'b'))
 *    ('fneg', ('fdot2_replicated(is_used_once)', 'a', 'b')) => ('fdot2_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdot3_replicated(is_used_once)', 'a', 'b')) => ('fdot3_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdot4_replicated(is_used_once)', 'a', 'b')) => ('fdot4_replicated', ('fneg', 'a'), 'b')
 *    ('fneg', ('fdph_replicated(is_used_once)', 'a', 'b')) => ('fdph_replicated', 'a', ('fneg', 'b'))
 *    ('fneg', ('fsign(is_used_once)', 'a')) => ('fsign', ('fneg', 'a'))
 *    ('fabs', ('fsign(is_used_once)', 'a')) => ('fsign', ('fabs', 'a'))
 */


static const nir_search_value_union nir_opt_algebraic_distribute_src_mods_values[] = {
   /* ('fmul', ('fneg', 'a'), ('fneg', 'b')) => ('fmul', 'a', 'b') */
   { .variable = {
      { nir_search_value_variable, -2 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 0 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -2 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 2 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2155_0 -> 0 in the cache */
   /* replace2155_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('ffma', ('fneg', 'a'), ('fneg', 'b'), 'c') => ('ffma', 'a', 'b', 'c') */
   { .variable = {
      { nir_search_value_variable, -3 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 6 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      1, /* b */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 8 },
      -1,
   } },
   { .variable = {
      { nir_search_value_variable, -3 },
      2, /* c */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 7, 9, 10 },
      -1,
   } },

   /* replace2156_0 -> 6 in the cache */
   /* replace2156_1 -> 8 in the cache */
   /* replace2156_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 6, 8, 10 },
      -1,
   } },

   /* ('fdot2_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot2_replicated', 'a', 'b') */
   /* search2157_0_0 -> 0 in the cache */
   /* search2157_0 -> 1 in the cache */
   /* search2157_1_0 -> 2 in the cache */
   /* search2157_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2157_0 -> 0 in the cache */
   /* replace2157_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fdot3_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot3_replicated', 'a', 'b') */
   /* search2158_0_0 -> 0 in the cache */
   /* search2158_0 -> 1 in the cache */
   /* search2158_1_0 -> 2 in the cache */
   /* search2158_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2158_0 -> 0 in the cache */
   /* replace2158_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fdot4_replicated', ('fneg', 'a'), ('fneg', 'b')) => ('fdot4_replicated', 'a', 'b') */
   /* search2159_0_0 -> 0 in the cache */
   /* search2159_0 -> 1 in the cache */
   /* search2159_1_0 -> 2 in the cache */
   /* search2159_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* replace2159_0 -> 0 in the cache */
   /* replace2159_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 2 },
      -1,
   } },

   /* ('fneg', ('fneg', 'a')) => a */
   { .variable = {
      { nir_search_value_variable, -1 },
      0, /* a */
      false,
      nir_type_invalid,
      -1,
      {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15},
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 20 },
      -1,
   } },

   /* replace2160 -> 19 in the cache */

   /* ('fneg', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fneg', 'a'), 'b') */
   /* search2161_0_0 -> 0 in the cache */
   /* search2161_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 22 },
      -1,
   } },

   /* replace2161_0_0 -> 0 in the cache */
   /* replace2161_0 -> 1 in the cache */
   /* replace2161_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fabs', ('fmul(is_used_once)', 'a', 'b')) => ('fmul', ('fabs', 'a'), ('fabs', 'b')) */
   /* search2162_0_0 -> 0 in the cache */
   /* search2162_0_1 -> 2 in the cache */
   /* search2162_0 -> 22 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 1,
      { 22 },
      -1,
   } },

   /* replace2162_0_0 -> 0 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 0 },
      -1,
   } },
   /* replace2162_1_0 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 2 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmul,
      0, 1,
      { 26, 27 },
      -1,
   } },

   /* ('fneg', ('ffma(is_used_once)', 'a', 'b', 'c')) => ('ffma', ('fneg', 'a'), 'b', ('fneg', 'c')) */
   /* search2163_0_0 -> 6 in the cache */
   /* search2163_0_1 -> 8 in the cache */
   /* search2163_0_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 6, 8, 10 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 29 },
      -1,
   } },

   /* replace2163_0_0 -> 6 in the cache */
   /* replace2163_0 -> 7 in the cache */
   /* replace2163_1 -> 8 in the cache */
   /* replace2163_2_0 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 10 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_ffma,
      0, 1,
      { 7, 8, 31 },
      -1,
   } },

   /* ('fneg', ('flrp(is_used_once)', 'a', 'b', 'c')) => ('flrp', ('fneg', 'a'), ('fneg', 'b'), 'c') */
   /* search2164_0_0 -> 6 in the cache */
   /* search2164_0_1 -> 8 in the cache */
   /* search2164_0_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 6, 8, 10 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 33 },
      -1,
   } },

   /* replace2164_0_0 -> 6 in the cache */
   /* replace2164_0 -> 7 in the cache */
   /* replace2164_1_0 -> 8 in the cache */
   /* replace2164_1 -> 9 in the cache */
   /* replace2164_2 -> 10 in the cache */
   { .expression = {
      { nir_search_value_expression, -3 },
      false,
      false,
      false,
      nir_op_flrp,
      -1, 0,
      { 7, 9, 10 },
      -1,
   } },

   /* ('fneg', ('~fadd(is_used_once)', 'a', 'b')) => ('fadd', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2165_0_0 -> 0 in the cache */
   /* search2165_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      true,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 36 },
      -1,
   } },

   /* replace2165_0_0 -> 0 in the cache */
   /* replace2165_0 -> 1 in the cache */
   /* replace2165_1_0 -> 2 in the cache */
   /* replace2165_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fadd,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fmin(is_used_once)', 'a', 'b')) => ('fmax', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2166_0_0 -> 0 in the cache */
   /* search2166_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 39 },
      -1,
   } },

   /* replace2166_0_0 -> 0 in the cache */
   /* replace2166_0 -> 1 in the cache */
   /* replace2166_1_0 -> 2 in the cache */
   /* replace2166_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fmax(is_used_once)', 'a', 'b')) => ('fmin', ('fneg', 'a'), ('fneg', 'b')) */
   /* search2167_0_0 -> 0 in the cache */
   /* search2167_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmax,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 42 },
      -1,
   } },

   /* replace2167_0_0 -> 0 in the cache */
   /* replace2167_0 -> 1 in the cache */
   /* replace2167_1_0 -> 2 in the cache */
   /* replace2167_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fmin,
      0, 1,
      { 1, 3 },
      -1,
   } },

   /* ('fneg', ('fdot2_replicated(is_used_once)', 'a', 'b')) => ('fdot2_replicated', ('fneg', 'a'), 'b') */
   /* search2168_0_0 -> 0 in the cache */
   /* search2168_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 45 },
      -1,
   } },

   /* replace2168_0_0 -> 0 in the cache */
   /* replace2168_0 -> 1 in the cache */
   /* replace2168_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot2_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdot3_replicated(is_used_once)', 'a', 'b')) => ('fdot3_replicated', ('fneg', 'a'), 'b') */
   /* search2169_0_0 -> 0 in the cache */
   /* search2169_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 48 },
      -1,
   } },

   /* replace2169_0_0 -> 0 in the cache */
   /* replace2169_0 -> 1 in the cache */
   /* replace2169_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot3_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdot4_replicated(is_used_once)', 'a', 'b')) => ('fdot4_replicated', ('fneg', 'a'), 'b') */
   /* search2170_0_0 -> 0 in the cache */
   /* search2170_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 1,
      { 51 },
      -1,
   } },

   /* replace2170_0_0 -> 0 in the cache */
   /* replace2170_0 -> 1 in the cache */
   /* replace2170_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdot4_replicated,
      0, 1,
      { 1, 2 },
      -1,
   } },

   /* ('fneg', ('fdph_replicated(is_used_once)', 'a', 'b')) => ('fdph_replicated', 'a', ('fneg', 'b')) */
   /* search2171_0_0 -> 0 in the cache */
   /* search2171_0_1 -> 2 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 2 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 54 },
      -1,
   } },

   /* replace2171_0 -> 0 in the cache */
   /* replace2171_1_0 -> 2 in the cache */
   /* replace2171_1 -> 3 in the cache */
   { .expression = {
      { nir_search_value_expression, -2 },
      false,
      false,
      false,
      nir_op_fdph_replicated,
      -1, 0,
      { 0, 3 },
      -1,
   } },

   /* ('fneg', ('fsign(is_used_once)', 'a')) => ('fsign', ('fneg', 'a')) */
   /* search2172_0_0 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 19 },
      0,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fneg,
      -1, 0,
      { 57 },
      -1,
   } },

   /* replace2172_0_0 -> 19 in the cache */
   /* replace2172_0 -> 20 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 20 },
      -1,
   } },

   /* ('fabs', ('fsign(is_used_once)', 'a')) => ('fsign', ('fabs', 'a')) */
   /* search2173_0_0 -> 19 in the cache */
   /* search2173_0 -> 57 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 57 },
      -1,
   } },

   /* replace2173_0_0 -> 19 in the cache */
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fabs,
      -1, 0,
      { 19 },
      -1,
   } },
   { .expression = {
      { nir_search_value_expression, -1 },
      false,
      false,
      false,
      nir_op_fsign,
      -1, 0,
      { 61 },
      -1,
   } },

};

static const nir_search_expression_cond nir_opt_algebraic_distribute_src_mods_expression_cond[] = {
   (is_used_once),
};


static const struct transform nir_opt_algebraic_distribute_src_mods_transforms[] = {
   { ~0, ~0, ~0 }, /* Sentinel */

   { 23, 24, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 21, 19, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 30, 32, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 46, 47, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 49, 50, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 52, 53, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 34, 35, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 37, 38, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 40, 41, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 43, 44, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 55, 56, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 58, 59, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 25, 28, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 60, 62, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 4, 5, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 11, 12, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 13, 14, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 15, 16, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

   { 17, 18, 0 },
   { ~0, ~0, ~0 }, /* Sentinel */

};

static const struct per_op_table nir_opt_algebraic_distribute_src_mods_pass_op_table[nir_num_search_ops] = {
   [nir_op_fmul] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         2,
         2,
         2,
         28,
      },
   },
   [nir_op_fneg] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         2,
         3,
         4,
         5,
         6,
         7,
         8,
         9,
         10,
         11,
         12,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         2,
         0,
         0,
         1,
         3,
         4,
         5,
         6,
      },
      
      .num_filtered_states = 13,
      .table = (const uint16_t []) {
      
         3,
         14,
         15,
         16,
         17,
         18,
         19,
         20,
         21,
         22,
         23,
         24,
         25,
      },
   },
   [nir_op_ffma] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         4,
         4,
         4,
         4,
         4,
         4,
         29,
         29,
      },
   },
   [nir_op_fdot2_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         5,
         5,
         5,
         30,
      },
   },
   [nir_op_fdot3_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         6,
         6,
         6,
         31,
      },
   },
   [nir_op_fdot4_replicated] = {
      .filter = (const uint16_t []) {
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 2,
      .table = (const uint16_t []) {
      
         7,
         7,
         7,
         32,
      },
   },
   [nir_op_fabs] = {
      .filter = (const uint16_t []) {
         0,
         0,
         1,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         2,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         0,
         1,
         0,
         0,
         0,
         0,
      },
      
      .num_filtered_states = 3,
      .table = (const uint16_t []) {
      
         0,
         26,
         27,
      },
   },
   [nir_op_flrp] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         8,
      },
   },
   [nir_op_fadd] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         9,
      },
   },
   [nir_op_fmin] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         10,
      },
   },
   [nir_op_fmax] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         11,
      },
   },
   [nir_op_fdph_replicated] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         12,
      },
   },
   [nir_op_fsign] = {
      .filter = NULL,
      
      .num_filtered_states = 1,
      .table = (const uint16_t []) {
      
         13,
      },
   },
};

/* Mapping from state index to offset in transforms (0 being no transforms) */
static const uint16_t nir_opt_algebraic_distribute_src_mods_transform_offsets[] = {
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   0,
   1,
   3,
   5,
   7,
   9,
   11,
   13,
   15,
   17,
   19,
   21,
   23,
   25,
   27,
   29,
   31,
   33,
   35,
   37,
};

static const nir_algebraic_table nir_opt_algebraic_distribute_src_mods_table = {
   .transforms = nir_opt_algebraic_distribute_src_mods_transforms,
   .transform_offsets = nir_opt_algebraic_distribute_src_mods_transform_offsets,
   .pass_op_table = nir_opt_algebraic_distribute_src_mods_pass_op_table,
   .values = nir_opt_algebraic_distribute_src_mods_values,
   .expression_cond = nir_opt_algebraic_distribute_src_mods_expression_cond,
   .variable_cond = NULL,
};

bool
nir_opt_algebraic_distribute_src_mods(nir_shader *shader)
{
   bool progress = false;
   bool condition_flags[127];
   const nir_shader_compiler_options *options = shader->options;
   const shader_info *info = &shader->info;
   (void) options;
   (void) info;

   STATIC_ASSERT(63 == ARRAY_SIZE(nir_opt_algebraic_distribute_src_mods_values));
   condition_flags[0] = true;
   condition_flags[1] = !options->lower_bitops;
   condition_flags[2] = options->lower_bitops;
   condition_flags[3] = !options->lower_bitops && (options->lower_int64_options & (nir_lower_imul64 | nir_lower_shift64)) == nir_lower_imul64;
   condition_flags[4] = options->lower_mul_2x32_64;
   condition_flags[5] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 16);
   condition_flags[6] = !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[7] = !options->lower_iadd_sat;
   condition_flags[8] = !options->lower_uadd_sat;
   condition_flags[9] = options->has_fmulz && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[10] = !options->has_sdot_4x8;
   condition_flags[11] = !options->has_udot_4x8;
   condition_flags[12] = !options->has_sudot_4x8;
   condition_flags[13] = !options->has_dot_2x16;
   condition_flags[14] = options->lower_flrp16;
   condition_flags[15] = !options->lower_flrp16;
   condition_flags[16] = options->lower_flrp32;
   condition_flags[17] = !options->lower_flrp32;
   condition_flags[18] = options->lower_flrp64;
   condition_flags[19] = !options->lower_flrp64;
   condition_flags[20] = options->lower_ftrunc;
   condition_flags[21] = options->lower_ffloor;
   condition_flags[22] = options->lower_ffloor || ((options->lower_doubles_options & nir_lower_dfloor) && !(options->lower_doubles_options & nir_lower_dfloor));
   condition_flags[23] = !options->lower_ffloor;
   condition_flags[24] = !options->lower_ffloor && !(options->lower_doubles_options & nir_lower_dfloor);
   condition_flags[25] = options->lower_ffract;
   condition_flags[26] = options->lower_ffract || (options->lower_doubles_options & nir_lower_dfract);
   condition_flags[27] = options->lower_fceil;
   condition_flags[28] = options->lower_ffma16;
   condition_flags[29] = options->lower_ffma32;
   condition_flags[30] = options->lower_ffma64;
   condition_flags[31] = options->fuse_ffma16;
   condition_flags[32] = options->fuse_ffma32;
   condition_flags[33] = options->fuse_ffma64;
   condition_flags[34] = options->lower_fdph;
   condition_flags[35] = !options->lower_fdph;
   condition_flags[36] = options->lower_fdot;
   condition_flags[37] = !options->lower_fsat;
   condition_flags[38] = !options->lower_iabs;
   condition_flags[39] = options->lower_fsat;
   condition_flags[40] = !options->lower_fsat && !nir_is_float_control_signed_zero_inf_nan_preserve(info->float_controls_execution_mode, 32);
   condition_flags[41] = !options->lower_fsign;
   condition_flags[42] = options->lower_umax;
   condition_flags[43] = options->lower_umin;
   condition_flags[44] = !options->lower_umax;
   condition_flags[45] = !options->lower_umin;
   condition_flags[46] = options->lower_scmp;
   condition_flags[47] = options->lower_vector_cmp;
   condition_flags[48] = !options->lower_rotate;
   condition_flags[49] = options->lower_rotate;
   condition_flags[50] = options->lower_fpow;
   condition_flags[51] = !options->lower_fpow;
   condition_flags[52] = options->lower_fdiv;
   condition_flags[53] = options->lower_fsqrt;
   condition_flags[54] = !options->lower_fsqrt;
   condition_flags[55] = options->lower_sincos;
   condition_flags[56] = !(options->lower_doubles_options & nir_lower_fp64_full_software);
   condition_flags[57] = !options->lower_extract_byte;
   condition_flags[58] = !options->lower_extract_word;
   condition_flags[59] = options->has_pack_32_4x8;
   condition_flags[60] = options->lower_pack_64_2x32_split;
   condition_flags[61] = options->lower_pack_32_2x16_split;
   condition_flags[62] = options->lower_unpack_64_2x32_split;
   condition_flags[63] = options->lower_unpack_32_2x16_split;
   condition_flags[64] = options->lower_fmod;
   condition_flags[65] = options->lower_uadd_carry;
   condition_flags[66] = options->lower_usub_borrow;
   condition_flags[67] = options->lower_bitfield_insert;
   condition_flags[68] = options->lower_hadd;
   condition_flags[69] = options->lower_hadd64 || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[70] = options->lower_mul_32x16;
   condition_flags[71] = options->lower_uadd_sat || (options->lower_int64_options & nir_lower_iadd64) != 0;
   condition_flags[72] = options->lower_uadd_sat;
   condition_flags[73] = (options->lower_int64_options & nir_lower_usub_sat64) != 0;
   condition_flags[74] = (options->lower_int64_options & nir_lower_iadd_sat64) != 0;
   condition_flags[75] = (options->lower_int64_options & nir_lower_minmax64) != 0;
   condition_flags[76] = (options->lower_int64_options & nir_lower_icmp64) != 0;
   condition_flags[77] = options->lower_bitfield_insert_to_shifts;
   condition_flags[78] = options->lower_bitfield_insert_to_bitfield_select;
   condition_flags[79] = options->lower_bitfield_extract;
   condition_flags[80] = options->lower_bitfield_extract_to_shifts;
   condition_flags[81] = options->lower_ifind_msb;
   condition_flags[82] = options->lower_find_msb_to_reverse;
   condition_flags[83] = options->lower_find_lsb;
   condition_flags[84] = options->lower_extract_byte;
   condition_flags[85] = options->lower_extract_word;
   condition_flags[86] = options->lower_pack_unorm_2x16;
   condition_flags[87] = options->lower_pack_unorm_4x8;
   condition_flags[88] = options->lower_pack_snorm_2x16;
   condition_flags[89] = options->lower_pack_snorm_4x8;
   condition_flags[90] = options->lower_unpack_unorm_2x16;
   condition_flags[91] = options->lower_unpack_unorm_4x8;
   condition_flags[92] = options->lower_unpack_snorm_2x16;
   condition_flags[93] = options->lower_unpack_snorm_4x8;
   condition_flags[94] = options->lower_pack_split;
   condition_flags[95] = options->lower_isign;
   condition_flags[96] = !options->lower_isign;
   condition_flags[97] = options->lower_fsign;
   condition_flags[98] = !options->has_imul24;
   condition_flags[99] = !options->has_umul24;
   condition_flags[100] = !options->has_umad24;
   condition_flags[101] = options->has_imul24;
   condition_flags[102] = options->has_umad24;
   condition_flags[103] = options->has_umul24;
   condition_flags[104] = options->has_fused_comp_and_csel;
   condition_flags[105] = options->lower_iadd_sat;
   condition_flags[106] = options->support_8bit_alu;
   condition_flags[107] = options->support_16bit_alu;
   condition_flags[108] = options->lower_ldexp;
   condition_flags[109] = !options->lower_bitfield_reverse;
   condition_flags[110] = info->stage == MESA_SHADER_COMPUTE && info->cs.derivative_group == DERIVATIVE_GROUP_NONE;
   condition_flags[111] = options->lower_fisnormal;
   condition_flags[112] = !options->vectorize_vec2_16bit;
   condition_flags[113] = options->has_fsub;
   condition_flags[114] = options->has_fsub && !(options->lower_doubles_options & nir_lower_dsub);
   condition_flags[115] = options->lower_fneg;
   condition_flags[116] = options->has_isub || options->lower_ineg;
   condition_flags[117] = options->lower_ineg;
   condition_flags[118] = options->lower_iabs;
   condition_flags[119] = options->has_iadd3;
   condition_flags[120] = options->vectorize_vec2_16bit;
   condition_flags[121] = options->fdot_replicates;
   condition_flags[122] = options->lower_fround_even;
   condition_flags[123] = (info->stage != MESA_SHADER_VERTEX && info->stage != MESA_SHADER_GEOMETRY) && !options->intel_vec4;
   condition_flags[124] = options->avoid_ternary_with_two_constants;
   condition_flags[125] = !options->lower_insert_byte;
   condition_flags[126] = !options->lower_insert_word;

   nir_foreach_function(function, shader) {
      if (function->impl) {
         progress |= nir_algebraic_impl(function->impl, condition_flags,
                                        &nir_opt_algebraic_distribute_src_mods_table);
      }
   }

   return progress;
}

